package io.confluent.command;

import com.google.common.collect.ImmutableMap;

import org.apache.kafka.clients.producer.KafkaProducer;
import org.apache.kafka.clients.producer.ProducerRecord;
import org.apache.kafka.clients.producer.RecordMetadata;
import org.apache.kafka.common.utils.SystemTime;
import org.apache.kafka.streams.KafkaStreams;
import org.apache.kafka.streams.StreamsConfig;
import org.apache.kafka.streams.kstream.KStream;
import org.apache.kafka.streams.kstream.KStreamBuilder;
import org.apache.kafka.streams.kstream.Predicate;
import org.apache.kafka.streams.processor.AbstractProcessor;
import org.apache.kafka.streams.processor.Processor;
import org.apache.kafka.streams.processor.ProcessorContext;
import org.apache.kafka.streams.processor.ProcessorSupplier;
import org.apache.kafka.streams.processor.StateStore;
import org.apache.kafka.streams.processor.StateStoreSupplier;
import org.apache.kafka.streams.state.KeyValueStore;
import org.apache.kafka.streams.state.QueryableStoreTypes;
import org.apache.kafka.streams.state.ReadOnlyKeyValueStore;
import org.apache.kafka.streams.state.internals.MeteredKeyValueStore;
import org.apache.kafka.streams.state.internals.RocksDBStore;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.HashSet;
import java.util.Map;
import java.util.Properties;
import java.util.Set;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.Future;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.TimeoutException;
import java.util.concurrent.atomic.AtomicLong;

import io.confluent.command.record.Command;
import io.confluent.command.record.Command.CommandConfigType;
import io.confluent.command.record.Command.CommandKey;
import io.confluent.command.record.Command.CommandMessage;
import io.confluent.serializers.ProtoSerde;
import io.confluent.serializers.UberSerde;

public class CommandStore {

  public static final String DEFAULT_COMMAND_TOPIC = "_confluent-command";
  private static final Logger log = LoggerFactory.getLogger(CommandStore.class);
  private static final UberSerde<CommandKey> commandKeySerde = new ProtoSerde<>(CommandKey.getDefaultInstance());
  private static final UberSerde<CommandMessage> commandMessageSerde = new ProtoSerde<>(CommandMessage.getDefaultInstance());
  private static final long DEFAULT_START_TIMEOUT = TimeUnit.SECONDS.toMillis(30);
  private static final long DEFAULT_AWAIT_TIMEOUT = TimeUnit.SECONDS.toMillis(15);

  protected static String storeName(final CommandConfigType type) {
    return "type-" + type.name();
  }

  private final String commandTopic;
  private final KafkaStreams streamsJob;
  private final KafkaProducer<CommandKey, CommandMessage> producer;
  private final Set<CommandConfigType> types;
  private final AtomicLong highOffset;

  protected CommandStore(
      String commandTopic,
      KafkaStreams streamsJob,
      KafkaProducer<CommandKey, CommandMessage> producer,
      Set<CommandConfigType> types,
      AtomicLong highOffset
  ) {
    this.commandTopic = commandTopic;
    this.streamsJob = streamsJob;
    this.producer = producer;
    this.types = types;
    this.highOffset = highOffset;
  }

  public String getCommandTopic() {
    return commandTopic;
  }

  public static UberSerde<CommandKey> getCommandKeySerde() {
    return commandKeySerde;
  }

  public static UberSerde<CommandMessage> getCommandMessageSerde() {
    return commandMessageSerde;
  }

  public void start() throws TimeoutException, InterruptedException {
    start(DEFAULT_START_TIMEOUT);
  }

  public void start(long allowance) throws TimeoutException, InterruptedException {
    long timeout = System.currentTimeMillis() + allowance;
    streamsJob.start();
    for (CommandConfigType type : types) {
      ReadOnlyKeyValueStore<CommandKey, CommandMessage> store = null;
      while (store == null) {
        Thread.sleep(allowance / 10);
        try {
          store = type(type);
        } catch (Exception e) {
          log.info("unable to get store type={} {}", type, log.isDebugEnabled() ? e : "");
        }
        if (timeout < System.currentTimeMillis()) {
          log.warn("unable to start with allowance={}", allowance);
          throw new TimeoutException();
        }
      }
    }
  }

  public void close() {
    try {
      streamsJob.close();
    } catch (Exception e) {
      log.error("closing streams job", e);
    }
    try {
      producer.close();
    } catch (Exception e) {
      log.error("closing producer", e);
    }
  }

  public Future<RecordMetadata> sendCommand(CommandKey key, CommandMessage command) {
    return producer.send(new ProducerRecord<>(commandTopic, 0, key, command));
  }

  public RecordMetadata awaitCommand(CommandKey key, CommandMessage command) throws InterruptedException, TimeoutException, ExecutionException {
    return awaitCommand(key, command, DEFAULT_AWAIT_TIMEOUT);
  }

  public RecordMetadata awaitCommand(CommandKey key, CommandMessage command, long timeout) throws InterruptedException, TimeoutException, ExecutionException {
    long endTime = System.currentTimeMillis() + TimeUnit.SECONDS.toMillis(timeout);
    RecordMetadata recordMetadata = sendCommand(key, command).get(timeout, TimeUnit.SECONDS);

    synchronized (highOffset) {
      while (
          highOffset.get() < recordMetadata.offset() &&
          endTime > System.currentTimeMillis()
          ) {
        // TODO(norwood): think about just returning a future here and having the future be completed in the Processor
        highOffset.wait(Math.max(endTime - System.currentTimeMillis(), 0));
      }
    }
    if (highOffset.get() < recordMetadata.offset()) {
      throw new TimeoutException("Timed-out waiting to see " + key + " update at offset " + recordMetadata.offset());
    }
    return recordMetadata;
  }

  public ReadOnlyKeyValueStore<CommandKey, CommandMessage> type(CommandConfigType type) {
    if (!types.contains(type)) {
      log.warn("not configured for type={}", type);
    }
    return streamsJob.store(storeName(type), QueryableStoreTypes.<CommandKey, CommandMessage>keyValueStore());
  }

  public static class Builder {

    private static final Map<String, Object> DEFAULT_PROPERTIES =
        ImmutableMap.<String, Object>of(
            "application.id", "test-commander",
            "bootstrap.servers", "localhost:9092"
        );

    private String commandTopic = DEFAULT_COMMAND_TOPIC;
    private StreamsConfig streamsConfig = null;
    private Map<String, Object> producerConfig = DEFAULT_PROPERTIES;
    private Set<CommandConfigType> types = new HashSet<>();

    public Builder() {
    }

    public Builder topic(String topic) {
      this.commandTopic = topic;
      return this;
    }

    public Builder streamsConfig(Properties streamsConfigs) {
      this.streamsConfig = new StreamsConfig(streamsConfigs);
      return this;
    }

    public Builder streamsConfig(StreamsConfig streamsConfigs) {
      this.streamsConfig = streamsConfigs;
      return this;
    }

    public Builder producerConfig(Map<String, Object> producerConfig) {
      this.producerConfig = producerConfig;
      return this;
    }

    public Builder store(CommandConfigType type) {
      this.types.add(type);
      return this;
    }

    public CommandStore build() {
      final AtomicLong highOffset = new AtomicLong();
      KStreamBuilder builder = new KStreamBuilder();
      KStream<CommandKey, CommandMessage> commandStream = builder.stream(commandKeySerde, commandMessageSerde, commandTopic);
      for (final CommandConfigType type : types) {
        StateStoreSupplier stateStoreSupplier = new RocksDBSupplier(storeName(type));
        builder.addStateStore(stateStoreSupplier);
        commandStream.filter(
                new Predicate<CommandKey, CommandMessage>() {
                  @Override
                  public boolean test(CommandKey key, CommandMessage value) {
                    return type.equals(key.getConfigType());
                  }
                }
            ).process(
            new ProcessorSupplier<CommandKey, CommandMessage>() {
              @Override
              public Processor<CommandKey, CommandMessage> get() {
                return new AbstractProcessor<CommandKey, CommandMessage>() {
                  private KeyValueStore<CommandKey, CommandMessage> store;

                  @Override
                  public void init(ProcessorContext context) {
                    super.init(context);
                    store = (KeyValueStore<CommandKey, CommandMessage>) context.getStateStore(storeName(type));
                  }

                  @Override
                  public void process(CommandKey key, CommandMessage value) {
                    store.put(key, value);
                    highOffset.set(context().offset());
                    synchronized (highOffset) {
                      highOffset.notifyAll();
                    }
                  }
                };
              }
            }, storeName(type)
        );
      }
      return new CommandStore(
          commandTopic,
          new KafkaStreams(builder, streamsConfig == null ? new StreamsConfig(DEFAULT_PROPERTIES) : streamsConfig),
          new KafkaProducer<>(
              producerConfig,
              commandKeySerde,
              commandMessageSerde
          ),
          types,
          highOffset
      );
    }
  }

  public static class RocksDBSupplier implements StateStoreSupplier {

    private final String storeName;

    public RocksDBSupplier(String storeName) {
      this.storeName = storeName;
    }

    @Override
    public String name() {
      return storeName;
    }

    @Override
    public StateStore get() {
      return new MeteredKeyValueStore<>(
          new RocksDBStore<>(storeName, commandKeySerde, commandMessageSerde),
          "commander-" + storeName + "-rocksdb-state",
          new SystemTime()
      );
    }

    @Override
    public Map<String, String> logConfig() {
      return null;
    }

    @Override
    public boolean loggingEnabled() {
      return false;
    }
  }

  public static void main(String[] args) throws Throwable {
    CommandStore commander = new CommandStore.Builder()
        .store(CommandConfigType.LICENSE_INFO)
        .store(CommandConfigType.ACTION_CONFIG)
        .build();
    try {
      commander.start();
      CommandKey key = CommandKey.newBuilder()
          .setConfigType(CommandConfigType.LICENSE_INFO)
          .setGuid("guid")
          .build();
      commander.awaitCommand(
          key,
          CommandMessage.newBuilder()
              .setLicenseInfo(Command.LicenseInfo.newBuilder().setJwt("jwt 9").build())
              .build()
      );
      log.info("{}", commander.type(CommandConfigType.LICENSE_INFO).get(key));
    } finally {
      commander.close();
    }
  }
}
