// (Copyright) [2016 - 2016] Confluent, Inc.
package io.confluent.metrics.reporter.integration;

import io.confluent.metrics.record.ConfluentMetric;
import io.confluent.metrics.reporter.ConfluentMetricsReporterConfig;
import io.confluent.serializers.ProtoSerde;
import org.apache.kafka.clients.consumer.ConsumerConfig;
import org.apache.kafka.clients.consumer.ConsumerRecord;
import org.apache.kafka.clients.consumer.ConsumerRecords;
import org.apache.kafka.clients.consumer.KafkaConsumer;
import org.apache.kafka.common.serialization.ByteArrayDeserializer;
import org.junit.After;
import org.junit.Before;
import org.junit.Test;

import java.util.Collections;
import java.util.Properties;

import static io.confluent.metrics.record.ConfluentMetric.MetricType.BROKER;
import static org.junit.Assert.*;

public class MetricsReporterTest extends MetricReporterClusterTestHarness {
    protected final ProtoSerde<ConfluentMetric.MetricsMessage> serdes = new ProtoSerde<>(ConfluentMetric.MetricsMessage.getDefaultInstance());
    protected KafkaConsumer<byte[], byte[]> consumer;

    @Before
    public void setUp() throws Exception {
        super.setUp();
        consumer = createNewConsumer();
        consumer.subscribe(Collections.singleton(ConfluentMetricsReporterConfig.DEFAULT_TOPIC_CONFIG));
    }

    @After
    public void tearDown() throws Exception {
        consumer.close();
        super.tearDown();
    }

    @Test
    public void testMetricsReporter() {
        Result latestResult = null;

        long startMs = System.currentTimeMillis();
        while (System.currentTimeMillis() - startMs < 20000) {
            ConsumerRecords<byte[], byte[]> records = consumer.poll(200);
            for (ConsumerRecord<byte[], byte[]> record : records) {
                latestResult = verify(record);
                if (latestResult.verified())
                    return;
            }
        }

        if (latestResult == null)
            fail("No records have been verified");
        else
            fail(String.format("One of the following is false : has Kafka measurable(%b), has Yammer gauge(%b), " +
                    "has Yammer Meter(%b), has Yammer histogram(%b), has Yammer Timer(%b)", latestResult.hasKafkaMeasurable,
                    latestResult.hasYammerGauge, latestResult.hasYammerMeter, latestResult.hasYammerHistogram, latestResult.hasYammerTimer));
    }

    protected Result verify(ConsumerRecord<byte[], byte[]> record) {
        Result result = new Result();
        ConfluentMetric.MetricsMessage metricsMessage = serdes.deserialize(record.value());

        assertEquals("metric type should be consumer", BROKER, metricsMessage.getMetricType());

        assertTrue("clusterId should be set", !metricsMessage.getClusterId().isEmpty());
        assertEquals("brokerId should match", 1, metricsMessage.getBrokerId());
        assertTrue("clientId should not be set", metricsMessage.getClientId().isEmpty());
        assertTrue("groupId should not be set", metricsMessage.getGroupId().isEmpty());

        result.hasKafkaMeasurable = metricsMessage.getKafkaMeasurableCount() > 0;
        result.hasYammerGauge = metricsMessage.getYammerGaugeCount() > 0;
        result.hasYammerMeter = metricsMessage.getYammerMeterCount() > 0;
        result.hasYammerHistogram = metricsMessage.getYammerHistogramCount() > 0;
        result.hasYammerTimer = metricsMessage.getYammerTimerCount() > 0;

        return result;
    }

    private KafkaConsumer<byte[], byte[]> createNewConsumer() {
        Properties properties = new Properties();
        properties.put(ConsumerConfig.BOOTSTRAP_SERVERS_CONFIG, brokerList);
        properties.put(ConsumerConfig.GROUP_ID_CONFIG, "metric-reporter-consumer");
        // The metric topic may not be there initially. So, we need to refresh metadata more frequently to pick it up once created.
        properties.put(ConsumerConfig.METADATA_MAX_AGE_CONFIG, "400");
        return new KafkaConsumer<>(properties, new ByteArrayDeserializer(), new ByteArrayDeserializer());
    }

    public static class Result {
        public boolean hasKafkaMeasurable;
        public boolean hasYammerGauge;
        public boolean hasYammerMeter;
        public boolean hasYammerHistogram;
        public boolean hasYammerTimer;

        public boolean verified() {
            return hasKafkaMeasurable && hasYammerGauge && hasYammerMeter && hasYammerHistogram && hasYammerTimer;
        }
    }
}
