// (Copyright) [2016 - 2016] Confluent, Inc.
package io.confluent.kafka.databalancing;

import io.confluent.kafka.databalancing.exception.ValidationException;
import io.confluent.kafka.databalancing.topology.Broker;
import io.confluent.kafka.databalancing.topology.ReplicaAssignment;
import io.confluent.kafka.databalancing.topology.TopologyUtils;
import kafka.admin.BrokerMetadata;
import kafka.common.AdminCommandFailedException;
import org.apache.kafka.common.TopicPartition;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

public abstract class AbstractRebalancer implements Rebalancer {

    protected AdminCommandFailedException partitionsBeingReassignedException(Set<TopicPartition> partitionsBeingReassigned) {
        StringBuilder sb = new StringBuilder();
        for (String line : TopologyUtils.partitionsByTopicToLines(partitionsBeingReassigned))
            sb.append("\t" + line + "\n");

        return new AdminCommandFailedException(String.format("A rebalance is currently in progress for:\n%sExiting.",
                sb.toString()));
    }

    protected static List<Broker> brokersToRemove(List<BrokerMetadata> allBrokers, ReplicaAssignment currentAssignment,
                                                  List<Integer> brokerIdsToRemove) {
        List<Broker> brokersToRemove = new ArrayList<>();

        Collection<Integer> removalSuggestions = suggestBrokersToRemove(allBrokers, currentAssignment, brokerIdsToRemove);

        if (!removalSuggestions.isEmpty()) {
            throw new ValidationException("Brokers with ids: " + removalSuggestions.toString() + " exist in the" +
                    " current assignment but are not online. Please run this tool with `--remove-broker-ids "
                    + Utils.mkString(removalSuggestions, ",")
                    + "` to move partitions away from these brokers.");
        }

        for (int brokerId : brokerIdsToRemove) {
            BrokerMetadata broker = TopologyUtils.findBrokerMetadata(allBrokers, brokerId);
            if (broker == null && !currentAssignment.brokers().contains(new Broker(brokerId)))
                throw new ValidationException("Broker with id "+ brokerId + " which has been requested for removal" +
                        " does not exist.");
            brokersToRemove.add(new Broker(brokerId));
        }
        return brokersToRemove;
    }

    private static Collection<Integer> suggestBrokersToRemove(List<BrokerMetadata> allBrokers, ReplicaAssignment currentAssignment,
                                                              List<Integer> brokerIdsToRemove) {
        Set<Integer> currentBrokerIdSet = new HashSet<>(allBrokers.size());
        Set<Integer> removeBrokerIdSet = new HashSet<>(brokerIdsToRemove);
        for (BrokerMetadata brokerMetadata : allBrokers) {
            currentBrokerIdSet.add(brokerMetadata.id());
        }
        List<Integer> suggestedBrokers = new ArrayList<>();
        for (Broker broker : currentAssignment.brokers()) {
            if (!(currentBrokerIdSet.contains(broker.id()) || removeBrokerIdSet.contains(broker.id()))) {
                suggestedBrokers.add(broker.id());
            }
        }
        return suggestedBrokers;
    }

}
