// (Copyright) [2016 - 2016] Confluent, Inc.
package io.confluent.kafka.databalancing.license;

import kafka.utils.ZkUtils;
import kafka.utils.ZkUtils$;
import org.I0Itec.zkclient.exception.ZkNodeExistsException;
import org.apache.kafka.common.security.JaasUtils;
import org.apache.kafka.common.utils.Time;
import org.apache.zookeeper.data.Stat;
import scala.Option;
import scala.Tuple2;

import java.util.concurrent.TimeUnit;

/**
 * Keeps track of a users trial period, with the state information stored on zookeeper.
 */
class ZkTrialPeriod {
    private static final String TRIAL_PERIOD_ZK_PATH = "/confluent-rebalancer/license-trial";
    private static final int ZK_SESSION_TIMEOUT_MS = 30000;
    private static final int ZK_CONNECT_TIMEOUT_MS = 7000;
    private static final long TRIAL_LIMIT_MS = TimeUnit.DAYS.toMillis(30);

    private final String zkConnect;
    private final Time time;

    ZkTrialPeriod(String zkConnect, Time time) {
        this.zkConnect = zkConnect;
        this.time = time;
    }

    /**
     * Starts or verifies a trial period.
     * @return the time remaining on the trial in milliseconds, or 0 if the trial has expired.
     */
    long startOrVerify() {
        ZkUtils zkUtils = ZkUtils.apply(zkConnect, ZK_SESSION_TIMEOUT_MS, ZK_CONNECT_TIMEOUT_MS,
                JaasUtils.isZkSecurityEnabled());
        try {
            zkUtils.createPersistentPath(TRIAL_PERIOD_ZK_PATH, "", zkUtils.DefaultAcls());
            return TRIAL_LIMIT_MS;
        } catch (ZkNodeExistsException e) {
            // The node already exists, which means that the trial has already started.
        }

        Stat stat = zkUtils.readData(TRIAL_PERIOD_ZK_PATH)._2;

        long elapsedTime = Math.max(time.milliseconds() - stat.getCtime(), 0);  // the Math.max is to compensate for clock skews.
        long remainingTime = Math.max(TRIAL_LIMIT_MS - elapsedTime, 0);
        return remainingTime;

    }
}
