// (Copyright) [2016 - 2016] Confluent, Inc.
package io.confluent.kafka.databalancing.report;

import io.confluent.kafka.databalancing.Utils;
import io.confluent.kafka.databalancing.metric.Metrics;
import io.confluent.kafka.databalancing.topology.Broker;
import io.confluent.kafka.databalancing.topology.Replica;
import io.confluent.kafka.databalancing.topology.ReplicaAssignment;
import kafka.admin.BrokerMetadata;
import org.apache.kafka.common.TopicPartition;

import java.io.PrintStream;
import java.text.DecimalFormat;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

public class RebalanceReport {

    private static final Comparator<Broker> BROKER_COMPARATOR = new Comparator<Broker>() {
        @Override
        public int compare(Broker b1, Broker b2) {
            return Integer.compare(b1.id(), b2.id());
        }
    };

    private final List<Replica> replicasAdded;
    private final Set<Integer> targetBrokers;
    private final long movingReplicasSize;
    private final Metrics metrics;
    private final Collection<Broker> brokersToRemove;
    private final List<String> allRacks;
    private final Map<Broker, Collection<TopicPartition>> currentBrokerToLeaders;
    private final Map<Broker, Collection<TopicPartition>> proposedBrokerToLeaders;

    private int leaderChanges;
    private int partitionChanges;

    private final Map<Broker, Collection<TopicPartition>> currentBrokerToReplicas;
    private final Map<Broker, Collection<TopicPartition>> proposedBrokerToReplicas;
    private final Set<Broker> proposedBrokersWithNoReplicas;
    private final List<Broker> allBrokers;

    private final Map<Broker, Long> brokerToPositiveSizeDiffDuringRebalance;
    private final Map<Broker, Long> brokerToSizeDiffAfterRebalance;

    private final Map<String, Long> currentRackToReplicaCount;
    private final Map<String, Long> currentRackToLeaderCount;
    private final Map<String, Long> currentRackToSize;
    private final Map<String, Long> proposedRackToReplicaCount;
    private final Map<String, Long> proposedRackToLeaderCount;
    private final Map<String, Long> proposedRackToSize;

    public RebalanceReport(Collection<BrokerMetadata> registeredBrokers, ReplicaAssignment currentAssignment,
                           ReplicaAssignment proposedAssignment, Metrics metrics, Collection<Broker> brokersToRemove) {
        this.metrics = metrics;
        this.brokersToRemove = brokersToRemove;
        currentBrokerToReplicas = currentAssignment.brokerToReplicas();
        currentBrokerToLeaders = currentAssignment.brokerToLeaders();
        proposedBrokerToReplicas = proposedAssignment.brokerToReplicas();
        proposedBrokerToLeaders = proposedAssignment.brokerToLeaders();
        this.allRacks = allRacks(registeredBrokers);
        allBrokers = allBrokers(registeredBrokers, currentAssignment, proposedAssignment);

        replicasAdded = new ArrayList<>();
        targetBrokers = new HashSet<>();

        for (TopicPartition tp : currentAssignment.asMap().keySet()) {
            List<Integer> currentBrokers = currentAssignment.asMap().get(tp);
            List<Integer> proposedBrokers = proposedAssignment.asMap().get(tp);

            if (!currentBrokers.equals(proposedBrokers))
                partitionChanges += 1;

            List<Integer> brokersAdded = new ArrayList<>(proposedBrokers);
            brokersAdded.removeAll(currentBrokers);

            for (Integer brokerAdded : brokersAdded)
                replicasAdded.add(new Replica(tp, new Broker(brokerAdded)));

            List<Integer> brokersRemoved = new ArrayList<>(currentBrokers);
            brokersRemoved.removeAll(proposedBrokers);

            targetBrokers.addAll(brokersAdded);

            Integer currentLeader = currentBrokers.isEmpty() ? null : currentBrokers.get(0);
            Integer newLeader = proposedBrokers.isEmpty() ? null : proposedBrokers.get(0);
            if (currentLeader == null && newLeader != null || currentLeader != null && !currentLeader.equals(newLeader))
                leaderChanges += 1;
        }

        movingReplicasSize = Math.round(replicasSize(replicasAdded) / (double) 1000 * 1000);

        brokerToPositiveSizeDiffDuringRebalance = new HashMap<>();
        brokerToSizeDiffAfterRebalance = new HashMap<>();
        populateBrokerToDiffMaps();

        proposedBrokersWithNoReplicas = new HashSet<>(currentAssignment.brokers());
        proposedBrokersWithNoReplicas.removeAll(proposedBrokerToReplicas.keySet());

        currentRackToLeaderCount = new HashMap<>();
        currentRackToReplicaCount = new HashMap<>();
        currentRackToSize = new HashMap<>();
        proposedRackToLeaderCount = new HashMap<>();
        proposedRackToReplicaCount = new HashMap<>();
        proposedRackToSize = new HashMap<>();
        populateRackMaps(registeredBrokers);
    }

    private void populateRackMaps(Collection<BrokerMetadata> registeredBrokers) {
        for (BrokerMetadata brokerMeta : registeredBrokers) {
            if (brokerMeta.rack().isDefined()) {
                String rack = brokerMeta.rack().get();
                Broker broker = new Broker(brokerMeta.id());

                long currentLeaderCount = getOrEmpty(currentBrokerToLeaders, broker).size();
                currentRackToLeaderCount.put(rack, getOrZero(currentRackToLeaderCount, rack) + currentLeaderCount);

                Collection<TopicPartition> currentReplicas = getOrEmpty(currentBrokerToReplicas, broker);
                currentRackToReplicaCount.put(rack, getOrZero(currentRackToReplicaCount, rack) + currentReplicas.size());
                currentRackToSize.put(rack, getOrZero(currentRackToSize, rack) + partitionsSize(currentReplicas));

                long proposedLeaderCount = getOrEmpty(proposedBrokerToLeaders, broker).size();
                proposedRackToLeaderCount.put(rack, getOrZero(proposedRackToLeaderCount, rack) + proposedLeaderCount);

                Collection<TopicPartition> proposedReplicas = getOrEmpty(proposedBrokerToReplicas, broker);
                proposedRackToReplicaCount.put(rack, getOrZero(proposedRackToReplicaCount, rack) + proposedReplicas.size());
                proposedRackToSize.put(rack, getOrZero(proposedRackToSize, rack) + partitionsSize(proposedReplicas));
            }
        }
    }

    private void populateBrokerToDiffMaps() {
        for (Broker broker : proposedBrokerToReplicas.keySet()) {
            Collection<TopicPartition> currentPartitions = getOrEmpty(currentBrokerToReplicas, broker);
            Collection<TopicPartition> proposedPartitions = getOrEmpty(proposedBrokerToReplicas, broker);

            Set<TopicPartition> partitionsAdded = new HashSet<>(proposedPartitions);
            partitionsAdded.removeAll(currentPartitions);
            long partitionsAddedSize = partitionsSize(partitionsAdded);
            if (partitionsAddedSize > 0) {
                brokerToPositiveSizeDiffDuringRebalance.put(broker, partitionsAddedSize);

                Set<TopicPartition> partitionsRemoved = new HashSet<>(currentPartitions);
                partitionsRemoved.removeAll(proposedPartitions);

                brokerToSizeDiffAfterRebalance.put(broker, partitionsAddedSize - partitionsSize(partitionsRemoved));
            }
        }
    }

    private static List<String> allRacks(Collection<BrokerMetadata> registeredBrokers) {
        Set<String> racks = new HashSet<>();
        for (BrokerMetadata b : registeredBrokers) {
            if (b.rack().isDefined())
                racks.add(b.rack().get());
        }
        return Utils.sorted(racks);
    }

    private static List<Broker> allBrokers(Collection<BrokerMetadata> registeredBrokers, ReplicaAssignment currentAssignment,
                                    ReplicaAssignment proposedAssignment) {
        Set<Broker> allBrokersSet = new HashSet<>();
        for (BrokerMetadata b : registeredBrokers)
            allBrokersSet.add(new Broker(b.id()));
        allBrokersSet.addAll(proposedAssignment.brokers());
        allBrokersSet.addAll(currentAssignment.brokers());
        return sortedBrokers(allBrokersSet);
    }

    public void print(PrintStream out, boolean verbose) {
        out.println("You are about to move " + replicasAdded.size() + " replica(s) for " + partitions(replicasAdded).size() +
                " partitions to " + targetBrokers.size() + " broker(s) with total size " + toMb(movingReplicasSize) + " MB.");
        out.println("The preferred leader for " + leaderChanges + " partition(s) will be changed.");
        out.println("In total, the assignment for " + partitionChanges + " partitions will be changed.");

        if (!brokersToRemove.isEmpty()) {
            out.println("\nYou have requested all replicas to be moved out of " + brokersToRemove.size() + " broker(s) with id(s): " +
                    brokersString(brokersToRemove) + ".");

            if (proposedBrokersWithNoReplicas.containsAll(brokersToRemove))
                out.println("After the rebalance, these broker(s) will have no replicas.");
            else {
                Set<Broker> brokers = new HashSet<>(brokersToRemove);
                brokers.removeAll(proposedBrokersWithNoReplicas);
                out.println("However, the following brokers will still have at least one replica after the rebalance: " +
                        brokersString(brokers));
            }
        }

        printBrokersRequiringMoreSpace(out);

        printMinMaxBrokers(out);

        if (allRacks.isEmpty())
            out.println("No racks are defined.");
        else {
            out.println("\nRack stats (before -> after):");
            printDetailedRackStats(out);
        }

        if (verbose) {
            out.println("\nBroker stats (before -> after):");
            printDetailedBrokerStats(out);
        }

        out.println();

    }

    private Collection<TopicPartition> partitions(List<Replica> replicas) {
        Set<TopicPartition> partitions = new HashSet<>();
        for (Replica replica : replicas)
            partitions.add(replica.topicPartition());
        return partitions;
    }

    private void printDetailedRackStats(PrintStream out) {
        String format = "\t%-10s %-15s %-15s %-20s";
        out.println(String.format(format, "Rack", "Leader Count", "Replica Count", "Size (MB)"));
        for (String rack : allRacks) {
            String line = String.format(format, rack,
                    getOrZero(currentRackToLeaderCount, rack) + " -> " + getOrZero(proposedRackToLeaderCount, rack),
                    getOrZero(currentRackToReplicaCount, rack) + " -> " +  getOrZero(proposedRackToReplicaCount, rack),
                    toMb(getOrZero(currentRackToSize, rack)) + " -> " + toMb(getOrZero(proposedRackToSize, rack)));
            out.println(line);
        }
    }

    private <K, E> K maxCount(Collection<K> keys, Map<K, Collection<E>> map) {
        return min(keys, map, new Comparator<Collection<E>>() {
            @Override
            public int compare(Collection<E> o1, Collection<E> o2) {
                return Integer.compare(o2.size(), o1.size());
            }
        }, Collections.<E>emptyList());
    }

    private <K> K maxSize(Collection<K> keys, Map<K, Collection<TopicPartition>> map) {
        return min(keys, map, new Comparator<Collection<TopicPartition>>() {
            @Override
            public int compare(Collection<TopicPartition> o1, Collection<TopicPartition> o2) {
                return Long.compare(partitionsSize(o2), partitionsSize(o1));
            }
        }, Collections.<TopicPartition>emptyList());
    }

    private <K, V> K min(Collection<K> keys, Map<K, V> map, Comparator<V> comparator, V defaultValue) {
        K result = null;
        V minValue = null;
        for (K key : keys) {
            V value = map.get(key);
            if (value == null)
                value = defaultValue;
            if (result == null || comparator.compare(value, minValue) == -1) {
                result = key;
                minValue = value;
            }
        }
        return result;
    }

    private <K, E> K minCount(Collection<K> keys, Map<K, Collection<E>> map) {
        return min(keys, map, new Comparator<Collection<E>>() {
            @Override
            public int compare(Collection<E> o1, Collection<E> o2) {
                return Integer.compare(o1.size(), o2.size());
            }
        }, Collections.<E>emptyList());
    }

    private <K> K minSize(Collection<K> keys, Map<K, Collection<TopicPartition>> map) {
        return min(keys, map, new Comparator<Collection<TopicPartition>>() {
            @Override
            public int compare(Collection<TopicPartition> o1, Collection<TopicPartition> o2) {
                return Long.compare(partitionsSize(o1), partitionsSize(o2));
            }
        }, Collections.<TopicPartition>emptyList());
    }

    private void printMinMaxBrokers(PrintStream out) {
        Broker currentBrokerWithMinSize = minSize(allBrokers, currentBrokerToReplicas);
        long currentMinSize = partitionsSize(getOrEmpty(currentBrokerToReplicas, currentBrokerWithMinSize));
        Broker currentBrokerWithMaxSize = maxSize(allBrokers, currentBrokerToReplicas);
        long currentMaxSize = partitionsSize(getOrEmpty(currentBrokerToReplicas, currentBrokerWithMaxSize));

        Broker proposedBrokerWithMinSize = minSize(allBrokers, proposedBrokerToReplicas);
        long proposedMinSize = partitionsSize(getOrEmpty(proposedBrokerToReplicas, proposedBrokerWithMinSize));
        Broker proposedBrokerWithMaxSize = maxSize(allBrokers, proposedBrokerToReplicas);
        long proposedMaxSize = partitionsSize(getOrEmpty(proposedBrokerToReplicas, proposedBrokerWithMaxSize));

        Broker currentBrokerWithMinLeaders = minCount(allBrokers, currentBrokerToLeaders);
        int currentMinLeaderCount = getOrEmpty(currentBrokerToLeaders, currentBrokerWithMinLeaders).size();
        Broker currentBrokerWithMaxLeaders = maxCount(allBrokers, currentBrokerToLeaders);
        int currentMaxLeaderCount = getOrEmpty(currentBrokerToLeaders, currentBrokerWithMaxLeaders).size();

        Broker proposedBrokerWithMinLeaders = minCount(allBrokers, proposedBrokerToLeaders);
        int proposedMinLeaderCount = getOrEmpty(proposedBrokerToLeaders, proposedBrokerWithMinLeaders).size();
        Broker proposedBrokerWithMaxLeaders = maxCount(allBrokers, proposedBrokerToLeaders);
        int proposedMaxLeaderCount = getOrEmpty(proposedBrokerToLeaders, proposedBrokerWithMaxLeaders).size();

        Broker currentBrokerWithMinReplicas = minCount(allBrokers, currentBrokerToReplicas);
        int currentMinReplicaCount = getOrEmpty(currentBrokerToReplicas, currentBrokerWithMinReplicas).size();
        Broker currentBrokerWithMaxReplicas = maxCount(allBrokers, currentBrokerToReplicas);
        int currentMaxReplicaCount = getOrEmpty(currentBrokerToReplicas, currentBrokerWithMaxReplicas).size();

        Broker proposedBrokerWithMinReplicas = minCount(allBrokers, proposedBrokerToReplicas);
        int proposedMinReplicaCount = getOrEmpty(proposedBrokerToReplicas, proposedBrokerWithMinReplicas).size();
        Broker proposedBrokerWithMaxReplicas = maxCount(allBrokers, proposedBrokerToReplicas);
        int proposedMaxReplicaCount = getOrEmpty(proposedBrokerToReplicas, proposedBrokerWithMaxReplicas).size();

        out.println("\nMin/max stats for brokers (before -> after):");
        String format = "\t%-5s %-28s %-28s %-35s";
        out.println(String.format(format, "Type", "Leader Count", "Replica Count", "Size (MB)"));

        String minLeaderCount = minOrMaxLine(currentMinLeaderCount, currentBrokerWithMinLeaders, proposedMinLeaderCount,
                proposedBrokerWithMinLeaders);
        String minReplicaCount = minOrMaxLine(currentMinReplicaCount, currentBrokerWithMinReplicas,
                proposedMinReplicaCount, proposedBrokerWithMinReplicas);
        String minSize = minOrMaxLine(toMb(currentMinSize), currentBrokerWithMinSize, toMb(proposedMinSize),
                proposedBrokerWithMinSize);

        String maxLeaderCount = minOrMaxLine(currentMaxLeaderCount, currentBrokerWithMaxLeaders, proposedMaxLeaderCount,
                proposedBrokerWithMaxLeaders);
        String maxReplicaCount = minOrMaxLine(currentMaxReplicaCount, currentBrokerWithMaxReplicas,
                proposedMaxReplicaCount, proposedBrokerWithMaxReplicas);;
        String maxSize = minOrMaxLine(toMb(currentMaxSize), currentBrokerWithMaxSize, toMb(proposedMaxSize),
                proposedBrokerWithMaxSize);

        out.println(String.format(format, "Min", minLeaderCount, minReplicaCount, minSize));
        out.println(String.format(format, "Max", maxLeaderCount, maxReplicaCount, maxSize));
    }

    private String minOrMaxLine(long currentValue, Broker currentBroker, long proposedValue, Broker proposedBroker) {
        return minOrMaxLine(Long.toString(currentValue), currentBroker, Long.toString(proposedValue), proposedBroker);
    }

    private String minOrMaxLine(String currentValue, Broker currentBroker, String proposedValue, Broker proposedBroker) {
        return currentValue + " (id: " + currentBroker.id() + ") -> " + proposedValue + " (id: " + proposedBroker.id() + ")";
    }

    private void printBrokersRequiringMoreSpace(PrintStream out) {
        if (brokerToPositiveSizeDiffDuringRebalance.isEmpty())
            return;
        out.println("\nThe following brokers will require more disk space during the rebalance and, in some cases, after the rebalance:");
        String format = "\t%-10s %-15s %-22s %-15s";
        out.println(String.format(format, "Broker", "Current (MB)", "During Rebalance (MB)", "After Rebalance (MB)"));
        for (Broker broker : sortedBrokers(brokerToPositiveSizeDiffDuringRebalance.keySet())) {
            long currentSize = partitionsSize(getOrEmpty(currentBrokerToReplicas, broker));
            long duringSize = currentSize + getOrZero(brokerToPositiveSizeDiffDuringRebalance, broker);
            long afterSize = currentSize + getOrZero(brokerToSizeDiffAfterRebalance, broker);
            String line = String.format(format, broker.id(),
                    toMb(currentSize),
                    toMb(duringSize),
                    toMb(afterSize));
            out.println(line);
        }
    }

    private String toMb(long value) {
        return Utils.bytesToFormattedMb(value);
    }

    private void printDetailedBrokerStats(PrintStream out) {
        String format = "\t%-10s %-15s %-15s %-20s";
        out.println(String.format(format, "Broker", "Leader Count", "Replica Count", "Size (MB)"));
        for (Broker broker : allBrokers) {
            Collection<TopicPartition> currentReplicas = getOrEmpty(currentBrokerToReplicas, broker);
            Collection<TopicPartition> proposedReplicas = getOrEmpty(proposedBrokerToReplicas, broker);
            long currentSize = partitionsSize(currentReplicas);
            long proposedSize = partitionsSize(proposedReplicas);
            String line = String.format(format, broker.id(),
                    getOrEmpty(currentBrokerToLeaders, broker).size() + " -> " + getOrEmpty(proposedBrokerToLeaders, broker).size(),
                    currentReplicas.size() + " -> " + proposedReplicas.size(),
                    toMb(currentSize) + " -> " + toMb(proposedSize));
            out.println(line);
        }
    }

    private static List<Broker> sortedBrokers(Collection<Broker> brokers) {
        return Utils.sorted(brokers, BROKER_COMPARATOR);
    }

    private <K> long getOrZero(Map<K, Long> map, K key) {
        Long result = map.get(key);
        return result == null ? 0 : result;
    }

    private <K, E> Collection<E> getOrEmpty(Map<K, Collection<E>> map, K key) {
        Collection<E> result = map.get(key);
        return result == null ? Collections.<E>emptyList() : result;
    }

    private String brokersString(Collection<Broker> brokers) {
        List<Integer> brokerIds = new ArrayList<>();
        for (Broker broker : brokers)
            brokerIds.add(broker.id());
        Collections.sort(brokerIds);
        return Utils.mkString(brokerIds, ",");
    }

    private long replicasSize(Collection<Replica> replicas) {
        long result = 0;
        for (Replica replica : replicas)
            result += metrics.partitionSize(replica.topicPartition());
        return result;
    }

    private long partitionsSize(Collection<TopicPartition> partitions) {
        long result = 0;
        for (TopicPartition partition : partitions)
            result += metrics.partitionSize(partition);
        return result;
    }

    public int partitionsChanges() {
        return partitionChanges;
    }
}
