package io.confluent.kafka.databalancing.throttle;

import io.confluent.kafka.databalancing.topology.Broker;
import io.confluent.kafka.databalancing.topology.ReplicaAssignment;
import kafka.admin.AdminUtilities;
import kafka.log.LogConfig;
import kafka.server.ConfigType;
import kafka.utils.ZkUtils;
import org.apache.kafka.common.TopicPartition;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import scala.collection.Seq;

import java.util.Collection;
import java.util.List;
import java.util.Map;
import java.util.Properties;

import static java.lang.String.valueOf;
import static java.util.Arrays.asList;
import static kafka.server.DynamicConfig.Broker$.MODULE$;
import static scala.collection.JavaConversions.asJavaCollection;
import static scala.collection.JavaConversions.asScalaBuffer;

public class Throttler implements Throttle {
    private static final Logger logger = LoggerFactory.getLogger(Throttler.class);
    static final String LEADER_RATE_PROP = MODULE$.LeaderReplicationThrottledRateProp();
    static final String FOLLOWER_RATE_PROP = MODULE$.FollowerReplicationThrottledRateProp();
    static final String LEADER_REPLICAS_PROP = LogConfig.LeaderReplicationThrottledReplicasProp();
    static final String FOLLOWER_REPLICAS_PROP = LogConfig.FollowerReplicationThrottledReplicasProp();
    private final ZkUtils zkUtils;
    private final AdminUtilities admin;

    public Throttler(ZkUtils zkUtils) {
        this(zkUtils, new KafkaAdmin());
    }

    /** Visible for testing */
    public Throttler(ZkUtils zkUtils, AdminUtilities admin) {
        this.zkUtils = zkUtils;
        this.admin = admin;
    }

    @Override
    public void engage(long quota, ReplicaAssignment current, ReplicaAssignment proposed) {
        AssignmentPair assignments = new AssignmentPair(current, proposed);

        // If there is a rebalance in progress just update the throttle
        if (zkUtils.pathExists(ZkUtils.ReassignPartitionsPath())) {
            limit(quota, assignments.brokers());
        } else {
            //Otherwise clear and existing settings and engage a new throttle
            disengage();
            engage(quota, assignments);
        }
    }

    void engage(long quota, AssignmentPair assignments) {
        limit(quota, assignments.brokers());
        throttleReplicas(assignments);
    }

    /**
     * Remove the throttle limit from all brokers in the cluster and the throttled replicas from all topics to ensure
     * we clean-up entities left behind from previous invocations of the rebalance tool.
     */
    @Override
    public boolean disengage() {
        return removeReplicas() | removeRates();
    }

    private void limit(long quota, Collection<Broker> brokers) {
        for (Broker broker : brokers) {
            //Get the existing config from ZK
            Properties existing = copyOf(admin.fetchEntityConfig(zkUtils, ConfigType.Broker(), valueOf(broker.id())));

            boolean leaderUpdated = updateIfChanged(valueOf(quota), existing, LEADER_RATE_PROP);
            boolean followerUpdated = updateIfChanged(valueOf(quota), existing, FOLLOWER_RATE_PROP);

            if (leaderUpdated || followerUpdated) {
                admin.changeBrokerConfig(zkUtils, toSeq(broker.id()), existing);
                logger.info("Updated broker config for broker [{}] to be [{}]", broker, existing);
            }
        }
    }

    /**
     * Set throttles to replicas that are moving. Note: this command should only be used when the assignment is initiated.
     */
    private void throttleReplicas(AssignmentPair assignments) {
        for (String topic : assignments.topics()) {
            //Apply the leader throttle to all replicas that exist before the re-balance.
            String throttledLeaders = format(assignments.moveSources(topic));

            //Apply a follower throttle to all "move destinations".
            String throttledFollowers = format(assignments.moveDestinations(topic));

            //Ensure we keep any existing properties
            Properties existing = copyOf(admin.fetchEntityConfig(zkUtils, ConfigType.Topic(), topic));
            boolean leaderUpdated = updateIfChanged(throttledLeaders, existing, LEADER_REPLICAS_PROP);
            boolean followerUpdated = updateIfChanged(throttledFollowers, existing, FOLLOWER_REPLICAS_PROP);

            if (leaderUpdated || followerUpdated) {
                admin.changeTopicConfig(zkUtils, topic, existing);
                logger.info("Updated topic config for topic [{}] to be [{}]", topic, existing);
            }
        }
    }

    private boolean removeReplicas() {
        boolean changed = false;
        for (String topic : asJavaCollection(zkUtils.getAllTopics())) {
            Properties configs = copyOf(admin.fetchEntityConfig(zkUtils, ConfigType.Topic(), topic));
            // bitwise OR as we don't want to short-circuit
            if (configs.remove(LEADER_REPLICAS_PROP) != null
                    | configs.remove(FOLLOWER_REPLICAS_PROP) != null) {
                admin.changeTopicConfig(zkUtils, topic, configs);
                changed = true;
            }
        }
        return changed;
    }

    private boolean removeRates() {
        boolean changed = false;
        for (kafka.cluster.Broker broker : asJavaCollection(zkUtils.getAllBrokersInCluster())) {
            Properties configs = copyOf(admin.fetchEntityConfig(zkUtils, ConfigType.Broker(), valueOf(broker.id())));
            // bitwise OR as we don't want to short-circuit
            if (configs.remove(LEADER_RATE_PROP) != null
                    | configs.remove(FOLLOWER_RATE_PROP) != null) {
                admin.changeBrokerConfig(zkUtils, toSeq(broker.id()), configs);
                changed = true;
            }
        }
        return changed;
    }

    private boolean updateIfChanged(String value, Properties properties, String key) {
        String propValue = properties.getProperty(key);
        boolean changed = value == null ? propValue != null : !value.equals(propValue);
        if (changed)
            properties.put(key, value);
        return changed;
    }

    private String format(Map<TopicPartition, List<Integer>> moves) {
        StringBuilder formatted = new StringBuilder();
        for (Map.Entry<TopicPartition, List<Integer>> entry : moves.entrySet()) {
            TopicPartition tp = entry.getKey();
            for (int brokerId : entry.getValue())
                formatted.append(String.format("%s:%s,", tp.partition(), brokerId));
        }
        if (formatted.length() > 0)
            formatted.deleteCharAt(formatted.length() - 1);

        return formatted.toString();
    }

    private Properties copyOf(Properties t) {
        Properties configs = new Properties();
        configs.putAll(t);
        return configs;
    }

    private Seq<Object> toSeq(int val) {
        return asScalaBuffer(asList((Object) val)).toSeq();
    }
}
