// (Copyright) [2016 - 2016] Confluent, Inc.
package io.confluent.kafka.databalancing.topology;

import io.confluent.kafka.databalancing.ScalaUtils;
import kafka.common.TopicAndPartition;
import kafka.utils.ZkUtils;
import org.apache.kafka.common.TopicPartition;
import scala.collection.JavaConversions;
import scala.collection.Seq;

import java.util.Collection;
import java.util.HashSet;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.Set;

public class ReplicaAssignment {

    private final Map<TopicPartition, List<Integer>> underlying;

    public ReplicaAssignment(Map<TopicPartition, List<Integer>> assignmentMap) {
        this.underlying = Objects.requireNonNull(assignmentMap, "assignmentMap cannot be null");
    }

    public static ReplicaAssignment fromScala(scala.collection.Map<TopicAndPartition, Seq<Object>> assignment) {
        return new ReplicaAssignment(fromScalaMap(assignment));
    }

    public boolean isEmpty() {
        return underlying.isEmpty();
    }

    public Map<TopicPartition, List<Integer>> asMap() {
        return Collections.unmodifiableMap(underlying);
    }

    public Collection<Broker> brokers() {
        Set<Broker> result = new HashSet<>();
        for (List<Integer> replicaIds : underlying.values())
            for (Integer brokerId : replicaIds)
                result.add(new Broker(brokerId));
        return result;
    }

    public Collection<String> topics() {
        Set<String> result = new HashSet<>();
        for (TopicPartition partition : underlying.keySet())
            result.add(partition.topic());
        return result;
    }

    public Collection<Integer> partitionsForTopic(String topic) {
        List<Integer> result = new ArrayList<>();
        for (TopicPartition tp : underlying.keySet()) {
            if (tp.topic() == topic)
                result.add(tp.partition());
        }
        return result;
    }

    public Map<Broker, Collection<TopicPartition>> brokerToLeaders() {
        Map<Broker, Collection<TopicPartition>> result = new HashMap<>();
        for (Map.Entry<TopicPartition, List<Integer>> entry : underlying.entrySet()) {
            Broker leader = new Broker(entry.getValue().get(0));
            Collection<TopicPartition> partitions = result.get(leader);
            if (partitions == null) {
                partitions = new ArrayList<>();
                result.put(leader, partitions);
            }
            partitions.add(entry.getKey());
        }
        return result;
    }

    public Map<Broker, Collection<TopicPartition>> brokerToReplicas() {
        Map<Broker, Collection<TopicPartition>> result = new HashMap<>();
        for (Map.Entry<TopicPartition, List<Integer>> entry : underlying.entrySet()) {
            for (Integer brokerId : entry.getValue()) {
                Broker broker = new Broker(brokerId);
                Collection<TopicPartition> partitions = result.get(broker);
                if (partitions == null) {
                    partitions = new ArrayList<>();
                    result.put(broker, partitions);
                }
                partitions.add(entry.getKey());
            }
        }
        return result;
    }

    public String toJson() {
        return ZkUtils.formatAsReassignmentJson(toScalaMap(underlying));
    }

    public Set<TopicPartition> topicPartitions() {
        return Collections.unmodifiableSet(underlying.keySet());
    }

    public List<Integer> replicas(TopicPartition topicPartition) {
        List<Integer> brokers = underlying.get(topicPartition);
        return brokers == null ? Collections.<Integer>emptyList() : Collections.unmodifiableList(brokers);
    }

    /**
     * Emulates the implementation in ReassignPartitionsCommand and picks the number of replicas of one of the
     * partitions of each topic as its replication factor.
     * TODO Consider a more explicit way of defining replication factor
     */
    public Map<String,Integer> replicationFactors() {
        Map<String, Integer> result = new HashMap<>();
        for (Map.Entry<TopicPartition, List<Integer>> entry : underlying.entrySet()) {
            String topic = entry.getKey().topic();
            if (!result.containsKey(topic)) {
                int replicaCount = entry.getValue().size();
                result.put(topic, replicaCount);
            }
        }
        return result;
    }

    public int size() {
        return underlying.size();
    }

    @Override
    public boolean equals(Object o) {
        if (this == o)
            return true;
        if (o == null || getClass() != o.getClass())
            return false;

        ReplicaAssignment that = (ReplicaAssignment) o;

        return underlying.equals(that.underlying);

    }

    @Override
    public int hashCode() {
        return underlying.hashCode();
    }

    @Override
    public String toString() {
        return "ReplicaAssignment(" + underlying + ")";
    }

    private static scala.collection.Map<TopicAndPartition, Seq<Object>> toScalaMap(
            Map<TopicPartition, List<Integer>> assignment) {
        Map<TopicAndPartition, Seq<Object>> result = new HashMap<>();
        for (Map.Entry<TopicPartition, List<Integer>> entry : assignment.entrySet()) {
            TopicPartition tp = entry.getKey();
            List<Object> brokerIds = new ArrayList<>();
            for (Integer brokerId : entry.getValue())
                brokerIds.add(brokerId);
            result.put(new TopicAndPartition(tp.topic(), tp.partition()), JavaConversions.asScalaBuffer(brokerIds));
        }
        return JavaConversions.mapAsScalaMap(result);
    }

    private static Map<TopicPartition, List<Integer>> fromScalaMap(
            scala.collection.Map<TopicAndPartition, Seq<Object>> assignment) {
        Map<TopicPartition, List<Integer>> result = new HashMap<>();
        for (Map.Entry<TopicAndPartition, Seq<Object>> entry : JavaConversions.mapAsJavaMap(assignment).entrySet()) {
            TopicAndPartition tp = entry.getKey();
            List<Integer> brokerIds = ScalaUtils.fromScalaSeqInt(entry.getValue());
            result.put(new TopicPartition(tp.topic(), tp.partition()), brokerIds);
        }
        return result;
    }

}
