// (Copyright) [2016 - 2016] Confluent, Inc.
package io.confluent.kafka.databalancing.view;

import io.confluent.kafka.databalancing.CompareResult;
import io.confluent.kafka.databalancing.RebalanceContext;
import io.confluent.kafka.databalancing.fairness.LeaderFairness;
import io.confluent.kafka.databalancing.fairness.LeaderSizeFairness;
import io.confluent.kafka.databalancing.fairness.ReplicaFairness;
import io.confluent.kafka.databalancing.fairness.ReplicaSizeFairness;
import io.confluent.kafka.databalancing.topology.Broker;
import io.confluent.kafka.databalancing.topology.Replica;
import io.confluent.kafka.databalancing.topology.TopologyUtils;
import org.apache.kafka.common.TopicPartition;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.List;
import java.util.Map;

import static io.confluent.kafka.databalancing.topology.TopologyUtils.sortBrokersByLeadersSize;
import static io.confluent.kafka.databalancing.topology.TopologyUtils.sortBrokersBySize;

public class RackSizeFairView extends RackFairView {

    // Created lazily
    private Map<String, Long> rackLeadersSizeMap;
    private Map<String, Long> rackSizeMap;

    public RackSizeFairView(RebalanceContext context, String topic) {
        super(context, topic);
    }

    @Override
    protected ReplicaFairness createReplicaFairness(RebalanceContext context, String topic) {
        return new ReplicaSizeFairness(context, null, topic);
    }

    @Override
    protected LeaderFairness createLeaderFairness(RebalanceContext context, String topic) {
        return new LeaderSizeFairness(context, null, topic);
    }

    @Override
    protected List<TopicPartition> createLeadersOnAboveParRacks(RebalanceContext context, LeaderFairness fairness, String topic) {
        List<TopicPartition> result = new ArrayList<>();
        for (String rack : fairness.aboveParRacks()) {
            List<TopicPartition> replicas = TopologyUtils.availableLeadersOnRack(context, rack, topic);
            sortRackLeaders(context, fairness, rack, replicas);
            result.addAll(replicas);

        }
        return result;
    }

    /**
     * Sorts in place rack leader partitions by the ones that have the most impact in reducing the variance.
     * In the optimal case we move a leader that causes the rack size to be the same as the rack fair value.
     * In case of ties, we choose the smaller partition (less data movement). If there are still ties at this point,
     * we compare topic names and partition ids to ensure determinism.
     */
    private void sortRackLeaders(final RebalanceContext context, LeaderFairness fairness, String rack,
                                 List<TopicPartition> replicas) {
        final double deviation = rackLeadersSizeMap().get(rack) - fairness.rackFairValue();
        Collections.sort(replicas, new Comparator<TopicPartition>() {
            @Override
            public int compare(TopicPartition tp1, TopicPartition tp2) {
                long c1 = context.partitionSize(tp1);
                long c2 = context.partitionSize(tp2);
                int result = Double.compare(Math.abs(c1 - deviation), Math.abs(c2 - deviation));
                if (result != 0)
                    return result;
                if (c1 != c2)
                    return Long.compare(c1, c2);
                return TopologyUtils.topicPartitionComparator.compare(tp1, tp2);
            }
        });
    }

    private Map<String, Long> rackLeadersSizeMap() {
        if (rackLeadersSizeMap == null)
            rackLeadersSizeMap = TopologyUtils.rackLeadersSizes(context, null, topic);
        return rackLeadersSizeMap;
    }

    @Override
    protected List<Replica> createReplicasOnAboveParRacks(final RebalanceContext context, ReplicaFairness fairness, String topic) {
        List<Replica> result = new ArrayList<>();
        for (String rack : fairness.aboveParRacks()) {
            List<Replica> replicas = TopologyUtils.availableReplicasOnRack(context, rack, topic);
            sortRackReplicas(context, fairness, rack, replicas);
            result.addAll(replicas);
        }
        return result;
    }

    /**
     * Sorts in-place rack replicas by the ones that have the most impact in reducing the variance. In the optimal case
     * we move a replica that causes the rack size to be the same as the rack fair value. In case of ties, we
     * choose the smaller partition (less data movement). If there are still ties at this point, we compare broker ids,
     * topic names and partition ids to ensure determinism.
     */
    private void sortRackReplicas(final RebalanceContext context, ReplicaFairness fairness, String rack,
                                  List<Replica> replicas) {
        final double deviation = rackSizeMap().get(rack) - fairness.rackFairValue();
        Collections.sort(replicas, new Comparator<Replica>() {
            @Override
            public int compare(Replica r1, Replica r2) {
                long c1 = context.partitionSize(r1.topicPartition());
                long c2 = context.partitionSize(r2.topicPartition());
                int result = Double.compare(Math.abs(c1 - deviation), Math.abs(c2 - deviation));
                if (result != 0)
                    return result;
                if (c1 != c2)
                    return Long.compare(c1, c2);
                if (r1.broker().id() != r2.broker().id())
                    return Integer.compare(r1.broker().id(), r2.broker().id());
                return TopologyUtils.topicPartitionComparator.compare(r1.topicPartition(), r2.topicPartition());
            }
        });
    }

    private Map<String, Long> rackSizeMap() {
        if (rackSizeMap == null)
            this.rackSizeMap = TopologyUtils.rackSizes(context, null, topic);
        return rackSizeMap;
    }

    @Override
    protected List<Broker> createBrokersWithBelowParLeaderMetric(RebalanceContext context, LeaderFairness fairness, String topic) {
        List<Broker> result = new ArrayList<>();
        for (String rack : fairness.belowParRacks())
            result.addAll(sortBrokersByLeadersSize(context, TopologyUtils.availableBrokersOnRack(context, rack), topic));
        return result;
    }

    @Override
    protected List<Broker> createBrokersWithBelowParReplicaMetric(RebalanceContext context, ReplicaFairness fairness, String topic) {
        List<Broker> result = new ArrayList<>();
        for (String rack : fairness.belowParRacks())
            result.addAll(sortBrokersBySize(context, TopologyUtils.availableBrokersOnRack(context, rack), topic));
        return result;
    }

    @Override
    public ClusterView refresh(RebalanceContext context) {
        return new RackSizeFairView(context, topic);
    }

}
