package io.confluent.kafka.databalancing.view;

import io.confluent.kafka.databalancing.RebalanceContext;

import java.util.HashMap;
import java.util.Map;

/**
 * A supplier of topic views that caches them until `clear` is invoked. Users should invoke `clear` whenever a partition
 * is moved (the equivalent of `ClusterView.refresh` in a sense).
 *
 * This class helps in cases where we may need the view for the same topic more than once before a partition is moved
 * while not having static knowledge of which topics may be needed.
 */
public abstract class TopicViewSupplier<V extends ClusterView> {

    public static TopicViewSupplier<BrokerFairView> brokerCount(final RebalanceContext context, final String rack) {
        return new TopicViewSupplier<BrokerFairView>() {
            @Override
            public BrokerFairView create(String topic) {
                return new BrokerCountFairView(context, rack, topic);
            }
        };
    }

    public static TopicViewSupplier<BrokerFairView> brokerSize(final RebalanceContext context, final String rack) {
        return new TopicViewSupplier() {
            @Override
            public BrokerFairView create(String topic) {
                return new BrokerSizeFairView(context, rack, topic);
            }
        };
    }

    public static TopicViewSupplier<RackFairView> rackCount(final RebalanceContext context) {
        return new TopicViewSupplier() {
            @Override
            public RackFairView create(String topic) {
                return new RackCountFairView(context, topic);
            }
        };
    }

    public static TopicViewSupplier<RackFairView> rackSize(final RebalanceContext context) {
        return new TopicViewSupplier() {
            @Override
            public RackFairView create(String topic) {
                return new RackSizeFairView(context, topic);
            }
        };
    }

    private final Map<String, V> topicToView = new HashMap<>();

    public V get(String topic) {
        V view = topicToView.get(topic);
        if (view == null) {
            view = create(topic);
            topicToView.put(topic, view);
        }
        return view;
    }

    public void clear() {
        topicToView.clear();
    }

    protected abstract V create(String topic);
}
