// (Copyright) [2016 - 2016] Confluent, Inc.
package io.confluent.kafka.databalancing;

import io.confluent.kafka.databalancing.fairness.BaseFairness;
import io.confluent.kafka.databalancing.fairness.LeaderCountFairness;
import io.confluent.kafka.databalancing.fairness.LeaderSizeFairness;
import io.confluent.kafka.databalancing.fairness.ReplicaCountFairness;
import io.confluent.kafka.databalancing.fairness.ReplicaSizeFairness;
import io.confluent.kafka.databalancing.topology.Broker;
import kafka.admin.BrokerMetadata;
import org.apache.kafka.common.TopicPartition;
import org.junit.Test;

import java.util.List;

import static io.confluent.kafka.databalancing.TestUtils.bk;
import static io.confluent.kafka.databalancing.TestUtils.createMap;
import static io.confluent.kafka.databalancing.TestUtils.createRebalanceContext;
import static io.confluent.kafka.databalancing.TestUtils.entry;
import static io.confluent.kafka.databalancing.TestUtils.replicationFactors;
import static java.util.Arrays.asList;
import static org.junit.Assert.assertEquals;

public class BaseFairnessTest {

    @Test
    public void shouldCompareFairnessCorrectlyWhenPassedDifferentTopic() {
        List<BrokerMetadata> brokers = asList(bk(100, "rack1"), bk(101, "rack2"));
        Broker broker0 = new Broker(brokers.get(0).id());
        Broker broker1 = new Broker(brokers.get(1).id());

        RebalanceContext context = createRebalanceContext(brokers,
                createMap(entry(new TopicPartition("topic-a", 0), asList(100)),
                        entry(new TopicPartition("topic-b", 0), asList(100)),
                        entry(new TopicPartition("topic-b", 1), asList(100)),
                        entry(new TopicPartition("topic-a", 1), asList(100))),
                replicationFactors(asList("topic-a", "topic-b"), 1));
        List<BaseFairness> fairnessList = asList(
                new LeaderCountFairness(context, null, "topic-a"),
                new ReplicaCountFairness(context, null, "topic-a"),
                new LeaderSizeFairness(context, null, "topic-a"),
                new ReplicaSizeFairness(context, null, "topic-a")
        );

        for (BaseFairness fairness : fairnessList) {
            // Move partition from a different topic
            assertEquals(CompareResult.EQUAL, fairness.compareBrokerFairness(new TopicPartition("topic-b", 0), broker0, broker1));
            assertEquals(CompareResult.EQUAL, fairness.compareBrokerFairness(new TopicPartition("topic-b", 0), broker1, broker1));
            assertEquals(CompareResult.EQUAL, fairness.compareRackFairness(new TopicPartition("topic-b", 0), broker0, broker1));
            assertEquals(CompareResult.EQUAL, fairness.compareRackFairness(new TopicPartition("topic-b", 0), broker1, broker1));

            // Switch partitions from a different topic
            assertEquals(CompareResult.EQUAL, fairness.compareBrokerFairness(broker0, new TopicPartition("topic-b", 0),
                    broker1, new TopicPartition("topic-b", 1)));
            assertEquals(CompareResult.EQUAL, fairness.compareBrokerFairness(broker1, new TopicPartition("topic-b", 0),
                    broker0, new TopicPartition("topic-b", 1)));
            assertEquals(CompareResult.EQUAL, fairness.compareRackFairness(broker0, new TopicPartition("topic-b", 0),
                    broker1, new TopicPartition("topic-b", 1)));
            assertEquals(CompareResult.EQUAL, fairness.compareRackFairness(broker1, new TopicPartition("topic-b", 0),
                    broker0, new TopicPartition("topic-b", 1)));

            // Switch partitions from different topic and fairness topic
            assertEquals(CompareResult.LESS, fairness.compareBrokerFairness(broker0, new TopicPartition("topic-b", 0),
                    broker1, new TopicPartition("topic-a", 1)));
            assertEquals(CompareResult.GREATER, fairness.compareBrokerFairness(broker1, new TopicPartition("topic-b", 0),
                    broker0, new TopicPartition("topic-a", 1)));
            assertEquals(CompareResult.LESS, fairness.compareRackFairness(broker0, new TopicPartition("topic-b", 0),
                    broker1, new TopicPartition("topic-a", 1)));
            assertEquals(CompareResult.GREATER, fairness.compareRackFairness(broker1, new TopicPartition("topic-b", 0),
                    broker0, new TopicPartition("topic-a", 1)));

            // Switch partitions from fairness topic and different topic
            assertEquals(CompareResult.GREATER, fairness.compareBrokerFairness(broker0, new TopicPartition("topic-a", 0),
                    broker1, new TopicPartition("topic-b", 1)));
            assertEquals(CompareResult.LESS, fairness.compareBrokerFairness(broker1, new TopicPartition("topic-a", 0),
                    broker0, new TopicPartition("topic-b", 1)));
            assertEquals(CompareResult.GREATER, fairness.compareRackFairness(broker0, new TopicPartition("topic-a", 0),
                    broker1, new TopicPartition("topic-b", 1)));
            assertEquals(CompareResult.LESS, fairness.compareRackFairness(broker1, new TopicPartition("topic-a", 0),
                    broker0, new TopicPartition("topic-b", 1)));


            // Move partition from fairness topic
            assertEquals(CompareResult.GREATER, fairness.compareBrokerFairness(new TopicPartition("topic-a", 0), broker0, broker1));
            assertEquals(CompareResult.LESS, fairness.compareBrokerFairness(new TopicPartition("topic-a", 0), broker1, broker0));
            assertEquals(CompareResult.GREATER, fairness.compareRackFairness(new TopicPartition("topic-a", 0), broker0, broker1));
            assertEquals(CompareResult.LESS, fairness.compareRackFairness(new TopicPartition("topic-a", 0), broker1, broker0));

            // Switch partitions from fairness topic
            assertEquals(CompareResult.EQUAL, fairness.compareBrokerFairness(broker0, new TopicPartition("topic-a", 0),
                    broker1, new TopicPartition("topic-a", 1)));
            assertEquals(CompareResult.EQUAL, fairness.compareBrokerFairness(broker1, new TopicPartition("topic-a", 0),
                    broker0, new TopicPartition("topic-a", 1)));
            assertEquals(CompareResult.EQUAL, fairness.compareRackFairness(broker0, new TopicPartition("topic-a", 0),
                    broker1, new TopicPartition("topic-a", 1)));
            assertEquals(CompareResult.EQUAL, fairness.compareRackFairness(broker1, new TopicPartition("topic-a", 0),
                    broker0, new TopicPartition("topic-a", 1)));
        }
    }
}
