// (Copyright) [2016 - 2016] Confluent, Inc.
package io.confluent.kafka.databalancing;

import io.airlift.airline.ParseOptionMissingException;
import io.confluent.kafka.databalancing.exception.NoRebalanceInProgressException;
import io.confluent.kafka.databalancing.exception.RebalanceInProgressException;
import io.confluent.kafka.databalancing.exception.ValidationException;
import io.confluent.kafka.databalancing.license.LicenseValidator;
import io.confluent.kafka.databalancing.license.NoopLicenseValidator;
import io.confluent.kafka.databalancing.metric.Metrics;
import io.confluent.kafka.databalancing.topology.ReplicaAssignment;
import kafka.admin.BrokerMetadata;
import kafka.utils.ZkUtils;
import org.apache.kafka.common.TopicPartition;
import org.junit.Test;
import scala.Console;
import scala.Option;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.InputStream;
import java.io.PrintStream;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.concurrent.Callable;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;

import static io.confluent.kafka.databalancing.TestUtils.createMap;
import static io.confluent.kafka.databalancing.TestUtils.createPartitions;
import static io.confluent.kafka.databalancing.TestUtils.createSimpleMetrics;
import static io.confluent.kafka.databalancing.TestUtils.entry;
import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertTrue;
import static org.junit.Assert.fail;

public class ConfluentRebalancerCommandTest {

    private static final String CONFIRMATION_STRING = "Would you like to continue? (y/n):";

    private final List<BrokerMetadata> allBrokers = Arrays.asList(new BrokerMetadata(1, Option.apply("rack1")),
            new BrokerMetadata(2, Option.apply("rack1")), new BrokerMetadata(3, Option.apply("rack2")));
    private final ReplicaAssignment currentAssignment = new ReplicaAssignment(createPartitions(
            Arrays.asList("t1", "t2"), 2, Arrays.asList(1, 2)));

    private final RebalancerFactory rebalancerFactory = new RebalancerFactory() {
        @Override
        public Rebalancer create(String zkUrl) {
            Metrics metrics = createSimpleMetrics(currentAssignment.topicPartitions());
            RebalanceStatus status = new RebalanceStatus(new ReplicaAssignment(Collections.<TopicPartition, List<Integer>>emptyMap()));
            return new InMemoryRebalancer(allBrokers, currentAssignment, metrics, status);
        }
    };

    RebalancerFactory rebalancerFactoryWithRebalanceInProgress = new RebalancerFactory() {
        @Override
        public Rebalancer create(String zkUrl) {
            Metrics metrics = createSimpleMetrics(currentAssignment.topicPartitions());
            Map<TopicPartition, List<Integer>> inProgressAssignment = new HashMap<>();
            inProgressAssignment.put(new TopicPartition("t1", 0), Arrays.asList(3, 1));
            inProgressAssignment.put(new TopicPartition("t2", 1), Arrays.asList(1));
            inProgressAssignment.put(new TopicPartition("t2", 0), Arrays.asList(2));
            RebalanceStatus status = new RebalanceStatus(new ReplicaAssignment(inProgressAssignment));
            return new InMemoryRebalancer(allBrokers, currentAssignment, metrics, status);
        }
    };

    private final LicenseValidator licenseValidator = new NoopLicenseValidator();

    @Test(expected = ParseOptionMissingException.class)
    public void testRunStatusWithMissingZookeeperParam() {
        ConfluentRebalancerCommand.run(rebalancerFactory, licenseValidator, "status");
    }

    @Test
    public void testMainNoCommandPrintsUsage() {
        String out = captureMainStandardOut();
        assertTrue("Unexpected output: " + out, out.startsWith("usage: confluent-rebalancer <command> [<args>]"));
    }

    @Test
    public void testRunExecute() {
        String out = captureRunStandardOut("execute", "--zookeeper", "localhost:2181",
                "--metrics-bootstrap-server", "localhost:9092", "--throttle", "1000", "--force", "--verbose");
        assertTrue(out.contains("The rebalance has been started"));
    }

    @Test
    public void testRunExecuteWithRemovedBrokers() {
        String out = captureRunStandardOut("execute", "--zookeeper", "localhost:2181",
                "--metrics-bootstrap-server", "localhost:9092", "--throttle", "1000", "--force", "--verbose",
                "--remove-broker-ids", "2");
        assertTrue(out.contains("The rebalance has been started"));
    }

    @Test(expected = ValidationException.class)
    public void testRunExecuteWithInvalidRemovedBroker() {
        ConfluentRebalancerCommand.run(rebalancerFactory, licenseValidator, "execute", "--zookeeper", "localhost:2181",
                "--metrics-bootstrap-server", "localhost:9092", "--throttle", "1000", "--force", "--verbose",
                "--remove-broker-ids", "4");
    }

    @Test(expected = ValidationException.class)
    public void testRunExecuteWithInvalidRemovedBrokerSyntax() {
        ConfluentRebalancerCommand.run(rebalancerFactory, licenseValidator, "execute", "--zookeeper", "localhost:2181",
                "--metrics-bootstrap-server", "localhost:9092", "--throttle", "1000", "--force", "--verbose",
                "--remove-broker-ids", "1;2");
    }

    @Test
    public void testRunExecuteWithRebalanceInProgress() {
        String out = captureRunStandardOut(rebalancerFactoryWithRebalanceInProgress, "execute", "--zookeeper", "localhost:2181",
                "--metrics-bootstrap-server", "localhost:9092", "--throttle", "7700", "--force");
        assertEquals("The throttle rate was updated to 7700 bytes/sec.", out);
    }

    @Test
    public void testRunExecuteWithMultipleRemovedBrokers() {
        String out = captureRunStandardOut(rebalancerFactory, "execute", "--zookeeper", "localhost:2181",
                "--metrics-bootstrap-server", "localhost:9092", "--throttle", "1000", "--force", "--verbose",
                "--remove-broker-ids", "1,2");
        assertTrue(out.contains("The rebalance has been started"));
    }

    @Test
    public void testRunExecuteWithEnterFollowedByYOnConfirmationPrompt() {
        String out = runExecuteWithConfirmationPrompt("\ny");
        assertEquals(2, countSubstring(out, CONFIRMATION_STRING));
        assertTrue(out.contains("The rebalance has been started"));
    }

    @Test
    public void testRunExecuteWithYOnConfirmationPrompt() {
        String out = runExecuteWithConfirmationPrompt("y");
        assertEquals(1, countSubstring(out, CONFIRMATION_STRING));
        assertTrue(out.contains("The rebalance has been started"));
    }

    @Test
    public void testRunExecuteWithNOnConfirmationPrompt() {
        String out = runExecuteWithConfirmationPrompt("n");
        assertEquals(1, countSubstring(out, CONFIRMATION_STRING));
        assertTrue(out.endsWith("OK, exiting"));
    }

    @Test
    public void testRunExecuteWithBadInputFollowedByYOnConfirmationPrompt() {
        String out = runExecuteWithConfirmationPrompt("dsndsd\nnd\ny");
        assertEquals(3, countSubstring(out, CONFIRMATION_STRING));
        assertTrue(out.contains("The rebalance has been started"));
    }

    @Test
    public void testRunExecuteWithBadInputFollowedByNOnConfirmationPrompt() {
        String out = runExecuteWithConfirmationPrompt("fhfhya\nya\nn");
        assertEquals(3, countSubstring(out, CONFIRMATION_STRING));
        assertTrue(out.endsWith("OK, exiting"));
    }

    private int countSubstring(String s, String substring) {
        int count = 0;
        int index = 0;
        while (true) {
            index = s.indexOf(substring, index);
            if (index == -1)
                break;
            index += 1;
            count += 1;
        }
        return count;
    }

    private String runExecuteWithConfirmationPrompt(String input) {
        InputStream savedSystemIn = System.in;
        ExecutorService executor = Executors.newSingleThreadExecutor();
        System.setIn(new ByteArrayInputStream(input.getBytes()));
        Console.setIn(System.in);
        String out;
        try {
            out = executor.submit(new Callable<String>() {
                @Override
                public String call() throws Exception {
                    return captureRunStandardOut("execute", "--zookeeper", "localhost:2181",
                            "--metrics-bootstrap-server", "localhost:9092", "--throttle", "1000");
                }
            }).get();
        } catch (Exception e) {
            throw new RuntimeException(e);
        } finally {
            executor.shutdown();
            System.setIn(savedSystemIn);
            Console.setIn(System.in);
        }
        return out;
    }

    @Test
    public void testRunProposedAssignment() {
        String out = captureRunStandardOut("proposed-assignment", "--zookeeper", "localhost:2181",
                "--metrics-bootstrap-server", "localhost:9092");
        ReplicaAssignment assignment = ReplicaAssignment.fromScala(ZkUtils.parsePartitionReassignmentData(out));
        assertEquals(currentAssignment.topicPartitions(), assignment.topicPartitions());
        Map<TopicPartition, List<Integer>> expected = createMap(
                entry(new TopicPartition("t1", 0), Arrays.asList(2, 3)),
                entry(new TopicPartition("t1", 1), Arrays.asList(1, 3)),
                entry(new TopicPartition("t2", 0), Arrays.asList(1, 3)),
                entry(new TopicPartition("t2", 1), Arrays.asList(3, 2))
        );
        assertEquals(expected, assignment.asMap());
    }

    @Test
    public void testRunProposedAssignmentWithRemovedBrokers() {
        String out = captureRunStandardOut("proposed-assignment", "--zookeeper", "localhost:2181",
                "--metrics-bootstrap-server", "localhost:9092", "--remove-broker-ids", "2");
        ReplicaAssignment assignment = ReplicaAssignment.fromScala(ZkUtils.parsePartitionReassignmentData(out));
        assertEquals(currentAssignment.topicPartitions(), assignment.topicPartitions());
        Map<TopicPartition, List<Integer>> expected = createMap(
                entry(new TopicPartition("t1", 0), Arrays.asList(3, 1)),
                entry(new TopicPartition("t1", 1), Arrays.asList(1, 3)),
                entry(new TopicPartition("t2", 0), Arrays.asList(3, 1)),
                entry(new TopicPartition("t2", 1), Arrays.asList(1, 3))
        );
        assertEquals(expected, assignment.asMap());
    }

    @Test
    public void testRunStatusWithRebalanceInProgress() {
        String out = captureRunStandardOut(rebalancerFactoryWithRebalanceInProgress, "status", "--zookeeper", "localhost:2181");
        String expected = "Partitions being rebalanced:\n" +
            "\tTopic t1: 0\n" +
            "\tTopic t2: 0,1";
        assertEquals(expected, out);
    }

    @Test
    public void testRunStatusWithNoRebalanceInProgress() {
        try {
            ConfluentRebalancerCommand.run(rebalancerFactory, licenseValidator, "status", "--zookeeper", "localhost:2181");
            fail("Should have thrown exception");
        } catch (NoRebalanceInProgressException e) {
            assertEquals("No rebalance is currently in progress. If you have called `status` after a rebalance was " +
                    "started successfully, the rebalance has completed. Run the `execute` command to check if the " +
                    "cluster is balanced.", e.getMessage());
        }
    }

    @Test
    public void testRunFinishWithRebalanceInProgress() {
        try {
            ConfluentRebalancerCommand.run(rebalancerFactoryWithRebalanceInProgress, licenseValidator, "finish", "--zookeeper", "localhost:2181");
            fail("Should have thrown exception");
        } catch (RebalanceInProgressException e) {
            assertEquals("3 partitions are being rebalanced", e.getMessage());
        }
    }

    @Test
    public void testRunFinishWithNoRebalanceInProgress() {
        String out = captureRunStandardOut(rebalancerFactory, "finish", "--zookeeper", "localhost:2181");
        String expected = "The rebalance has completed and throttling has been disabled";
        assertEquals(expected, out);
    }

    private String captureRunStandardOut(RebalancerFactory factory, String... args) {
        return captureStandardStream(false, runRunnable(factory, args));
    }

    private String captureRunStandardOut(String... args) {
        return captureRunStandardOut(rebalancerFactory, args);
    }

    private static Runnable runRunnable(final RebalancerFactory factory, final String... args) {
        return new Runnable() {
            @Override
            public void run() {
                ConfluentRebalancerCommand.run(factory, new NoopLicenseValidator(), args);
            }
        };
    }

    private String captureMainStandardErr(String... args) {
        return captureStandardStream(true, mainRunnable(args));
    }

    private String captureMainStandardOut(String... args) {
        return captureStandardStream(false, mainRunnable(args));
    }

    private Runnable mainRunnable(final String... args) {
        return new Runnable() {
            @Override
            public void run() {
                ConfluentRebalancerCommand.main(args);
            }
        };
    }

    private String captureStandardStream(boolean isErr, Runnable runnable) {
        ByteArrayOutputStream outputStream = new ByteArrayOutputStream();
        PrintStream currentStream = isErr ? System.err : System.out;
        PrintStream tempStream = new PrintStream(outputStream);
        if (isErr)
            System.setErr(tempStream);
        else
            System.setOut(tempStream);
        try {
            runnable.run();
            return new String(outputStream.toByteArray()).trim();
        } finally {
            if (isErr)
                System.setErr(currentStream);
            else
                System.setOut(currentStream);
        }
    }
}
