// (Copyright) [2016 - 2016] Confluent, Inc.
package io.confluent.kafka.databalancing;

import io.confluent.kafka.databalancing.metric.Metrics;
import io.confluent.kafka.databalancing.report.RebalanceReport;
import io.confluent.kafka.databalancing.throttle.Throttler;
import io.confluent.kafka.databalancing.topology.Broker;
import io.confluent.kafka.databalancing.topology.ReplicaAssignment;
import kafka.utils.ZkUtils;
import org.I0Itec.zkclient.ZkClient;
import org.I0Itec.zkclient.ZkConnection;
import org.apache.kafka.common.TopicPartition;
import org.apache.zookeeper.data.Stat;
import org.junit.Before;
import org.junit.Test;

import java.util.Collections;

import static io.confluent.kafka.databalancing.TestUtils.createBrokers;
import static io.confluent.kafka.databalancing.TestUtils.createMap;
import static io.confluent.kafka.databalancing.TestUtils.createSimpleMetrics;
import static io.confluent.kafka.databalancing.TestUtils.entry;
import static java.util.Arrays.asList;
import static org.junit.Assert.assertEquals;
import static org.mockito.Mockito.*;

public class DefaultRebalancerTest {

    private final ZkClient zkClient = mock(ZkClient.class);
    private final ZkConnection zkConnection = mock(ZkConnection.class);
    private final ZkUtils zkUtils = new ZkUtils(zkClient, zkConnection, false);

    private final Throttler throttler = mock(Throttler.class);
    private final Rebalancer rebalancer = new DefaultRebalancer(zkUtils, throttler);

    private final ReplicaAssignment currentAssignment = new ReplicaAssignment(createMap(
            entry(new TopicPartition("t-a", 0), asList(100, 101, 102)),
            entry(new TopicPartition("t-a", 1), asList(100, 101, 102)),
            entry(new TopicPartition("t-b", 0), asList(100, 101)),
            entry(new TopicPartition("t-b", 1), asList(103, 104)),
            entry(new TopicPartition("t-c", 0), asList(101, 100))));

    private final ReplicaAssignment proposedAssignment = new ReplicaAssignment(createMap(
            entry(new TopicPartition("t-a", 0), asList(100, 102, 101)),
            entry(new TopicPartition("t-a", 1), asList(100, 101, 102)),
            entry(new TopicPartition("t-b", 0), asList(100, 101)),
            entry(new TopicPartition("t-b", 1), asList(103, 104)),
            entry(new TopicPartition("t-c", 0), asList(101, 105))));

    private final String expectedAssignmentJson = "{\"version\":1,\"partitions\":[" +
            "{\"topic\":\"t-a\",\"partition\":0,\"replicas\":[100,102,101]}," +
            "{\"topic\":\"t-c\",\"partition\":0,\"replicas\":[101,105]}]}";

    @Before
    public void setUp() {
        when(zkClient.exists(eq("/"))).thenReturn(true);
    }

    /**
     * Verifies that `DefaultRebalancer.startRebalance()` stores the correct JSON value into
     * the `ReassignPartitionsPath` znode. In particular, it should only store the changed partitions to avoid
     * unnecessary logging of errors by the controller.
     */
    @Test
    public void testStartRebalance() {
        rebalancer.startRebalance(currentAssignment, proposedAssignment, 1000);
        verify(zkClient).createPersistent(eq(ZkUtils.ReassignPartitionsPath()), eq(expectedAssignmentJson),
                eq(zkUtils.DefaultAcls()));
    }

    /**
     * Verifies that status only returns changed partitions and that the output is consistent with rebalance report.
     */
    @Test
    public void testStatus() {
        when(zkClient.readData(eq(ZkUtils.ReassignPartitionsPath()), any(Stat.class))).thenReturn(expectedAssignmentJson);
        assertEquals(2, rebalancer.status().inProgressAssignment().size());
        RebalanceReport report = new RebalanceReport(createBrokers(100, 106, "rack1"),
                currentAssignment, proposedAssignment, createSimpleMetrics(currentAssignment.topicPartitions()),
                Collections.<Broker>emptyList());
        assertEquals(rebalancer.status().inProgressAssignment().size(), report.partitionsChanges());
    }

}
