// (Copyright) [2016 - 2016] Confluent, Inc.
package io.confluent.kafka.databalancing;

import io.confluent.kafka.databalancing.topology.Broker;
import kafka.admin.BrokerMetadata;
import org.apache.kafka.common.TopicPartition;
import org.junit.Test;

import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;

import static io.confluent.kafka.databalancing.TestUtils.*;
import static java.util.Arrays.asList;
import static org.junit.Assert.assertEquals;

public class RebalanceContextTest {

    private static final String topic = "topic";

    private static TopicPartition tp(int partition) {
        return new TopicPartition(topic, partition);
    }

    private static Map<String, Integer> replicationFactors(int replicationFactor) {
        return TestUtils.replicationFactors(asList(topic), replicationFactor);
    }

    @Test
    public void shouldFindFollowers() {
        //Given
        List<BrokerMetadata> brokers = concat(createBrokers(100, 102, "rack1"), createBrokers(102, 104, "rack2"));
        Map<TopicPartition, List<Integer>> partitions = new HashMap<>();
        partitions.put(new TopicPartition("t1", 0), asList(100, 101, 102));
        partitions.put(new TopicPartition("t2", 1), asList(100, 101, 102));
        partitions.put(new TopicPartition("t3", 2), asList(100, 101, 102));
        partitions.put(new TopicPartition("t4", 3), asList(100, 101, 102));
        Map<String, Integer> reps = new HashMap<>();
        reps.put("t1", 4);
        reps.put("t2", 4);
        reps.put("t3", 2);
        reps.put("t4", 2);

        RebalanceContext context = createRebalanceContext(brokers, partitions, reps);
        assertEquals(0, context.followers(new Broker(100)).size());
        assertEquals(4, context.followers(new Broker(101)).size());
        assertEquals(4, context.followers(new Broker(102)).size());
        assertEquals(0, context.followers(new Broker(103)).size());
    }

    @Test
    public void shouldCreateSimpleClusterTopologyOfBrokersToReplicas() {
        List<BrokerMetadata> brokers = asList(bk(100, "rack1"), bk(101, "rack2"));
        Map<TopicPartition, List<Integer>> partitions = createMap(
                entry(tp(0), asList(100, 101)),
                entry(tp(1), asList(100)));


        RebalanceContext context = createRebalanceContext(brokers, partitions, replicationFactors(2));

        assertEquals(new HashSet<>(asList(new Broker(100), new Broker(101))), context.allBrokers());


        assertEquals(new HashSet<>(asList(tp(0), tp(1))), new HashSet<>(context.replicas(new Broker(100))));
        assertEquals(new HashSet<>(asList(tp(0))), new HashSet<>(context.replicas(new Broker(101))));
    }

    @Test
    public void testAddReplica() {
        //Given
        List<BrokerMetadata> brokers = createBrokers(100, 102, "rack1");
        Map<TopicPartition, List<Integer>> underReplicated = createMap(
                entry(tp(0), asList(100)),
                entry(tp(1), Collections.<Integer>emptyList()));
        Map<String, Integer> reps = replicationFactors(2);
        MutableRebalanceContext context = createRebalanceContext(brokers, underReplicated, reps);

        //When
        context.addReplica(tp(0), new Broker(101));

        //Then
        assertEquals(new Broker(100), context.leader(tp(0)));
        assertEquals(asList(new Broker(100), new Broker(101)), context.brokers(tp(0)));
        assertEquals(asList(tp(0)), context.leaders(new Broker(100)));
        assertEquals(Collections.emptyList(), context.leaders(new Broker(101)));
        assertEquals(Collections.emptyList(), context.followers(new Broker(100)));
        assertEquals(asList(tp(0)), context.followers(new Broker(101)));

        //When
        context.addReplica(tp(1), new Broker(101));

        //Then
        assertEquals(new Broker(101), context.leader(tp(1)));
        assertEquals(asList(new Broker(101)), context.brokers(tp(1)));
        assertEquals(asList(tp(0)), context.leaders(new Broker(100)));
        assertEquals(asList(tp(1)), context.leaders(new Broker(101)));
        assertEquals(Collections.emptyList(), context.followers(new Broker(100)));
        assertEquals(asList(tp(0)), context.followers(new Broker(101)));
    }

    @Test
    public void makeLeader() {
        //Given
        List<BrokerMetadata> brokers = createBrokers(100, 102, "rack1");
        Map<TopicPartition, List<Integer>> underReplicated = createMap(
                entry(tp(0), asList(100, 101)));
        Map<String, Integer> reps = replicationFactors(1);
        MutableRebalanceContext context = createRebalanceContext(brokers, underReplicated, reps);

        //When
        context.makeLeader(tp(0), new Broker(101));

        //Then
        assertEquals(new Broker(101), context.leader(tp(0)));
        assertEquals(asList(new Broker(101), new Broker(100)), context.brokers(tp(0)));
        assertEquals(Collections.emptyList(), context.leaders(new Broker(100)));
        assertEquals(asList(tp(0)), context.leaders(new Broker(101)));
        assertEquals(asList(tp(0)), context.followers(new Broker(100)));
        assertEquals(Collections.emptyList(), context.followers(new Broker(101)));
    }

}
