// (Copyright) [2016 - 2016] Confluent, Inc.
package io.confluent.kafka.databalancing;

import io.confluent.kafka.databalancing.metric.Metrics;
import io.confluent.kafka.databalancing.topology.Broker;
import io.confluent.kafka.databalancing.topology.ReplicaAssignment;
import kafka.admin.BrokerMetadata;
import org.apache.kafka.common.TopicPartition;
import org.junit.Test;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import scala.Option;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Random;
import java.util.Set;

import static java.util.Arrays.asList;
import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertTrue;
import static io.confluent.kafka.databalancing.TestUtils.*;

public class RebalancingAcceptanceTest {

    private static final Logger LOG = LoggerFactory.getLogger(RebalancingAcceptanceTest.class);

    private Map<TopicPartition, Long> createRandomPartitionSizes(Set<TopicPartition> partitions, int minSize, int maxSize) {
        Random random = new Random(0);
        Map<TopicPartition, Long> result = new HashMap<>();
        for (TopicPartition tp : partitions)
            result.put(tp, (long) random.nextInt(maxSize - minSize) + minSize);
        return result;
    }

    @Test
    public void shouldRebalanceSmallFourTopicClusterSpreadOverTwoRacks() {
        MovesOptimisedRebalancePolicy policy = new MovesOptimisedRebalancePolicy();
        int broker0 = 100;

        //Given
        int brokerCount = 100;
        int partitionCount = 100;
        List<Integer> replicas = asList(100, 101, 102);
        List<String> topics = asList("t1", "t2", "t3", "t4");

        Map<String, Integer> reps = replicationFactors(topics, replicas.size());
        List<BrokerMetadata> brokers = concat(
                createBrokers(broker0, broker0 + brokerCount / 2, "rack1"),
                createBrokers(broker0 + brokerCount / 2, broker0 + brokerCount, "rack2")
        );

        Map<TopicPartition, List<Integer>> partitions = createPartitions(topics, partitionCount, replicas);
        ReplicaAssignment assignment = new ReplicaAssignment(partitions);
        Metrics metrics = new Metrics(createRandomPartitionSizes(partitions.keySet(), 10, 200));

        //When
        ReplicaAssignment reassigned = policy.rebalancePartitions(brokers, assignment, reps, metrics,
                Collections.<Broker>emptyList());

        List<Integer> reassignedBrokers = brokers(reassigned.asMap());
        Map<Integer, List<TopicPartition>> reassignedBrokerToPartitions = brokerToPartitions(reassigned.asMap());

        //Then counts should match
        assertEquals(topics.size() * partitionCount, reassigned.size());
        assertEquals(topics.size() * partitionCount * replicas.size(), reassignedBrokers.size());

        long clusterSize = sum(metrics.partitionToSize().values()) * replicas.size();
        double expectedBrokerSize = clusterSize / (double) brokers.size();

        for (int brokerId = broker0; brokerId < broker0 + brokerCount; ++brokerId) {
            //Then replicas should be somewhat evenly spread (the replica size optimisation, which runs after the replica
            //count optimisation means that we cannot expect an optimal spread)
            int expected = topics.size() * partitionCount * replicas.size() / brokerCount;
            assertEquals("Partition count outside expected range for broker " + brokerId,
                    expected, filter(reassignedBrokers, brokerId).size(), 7.0);

            long brokerSize = partitionSizeSum(metrics.partitionToSize(), reassignedBrokerToPartitions.get(brokerId));
            double tolerance = expectedBrokerSize * 0.035;
            assertEquals("Broker size outside expected range for broker " + brokerId, expectedBrokerSize,
                    brokerSize, tolerance);

            //Then leaders should be evenly spread
            expected = partitionCount * topics.size() / brokerCount;
            assertEquals("Broker leadership is uneven for " + brokerId, expected,
                    filter(leaders(reassigned.asMap()).values(), brokerId).size());
        }

        checkTopicFairness(topics, brokers, reps, metrics, reassigned.asMap(), reassignedBrokerToPartitions, 3.0, 1.2, 2.0);
    }

    @Test
    public void shouldRebalanceClusterWithDifferentSizedTopicsAndReplicaCountsButSingleRack() {
        MovesOptimisedRebalancePolicy policy = new MovesOptimisedRebalancePolicy();
        int broker0 = 100;

        //Given
        int brokerCount = 91;
        int partitionCount = 71;
        List<Integer> replicas1 = asList(100, 101, 102);
        List<Integer> replicas2 = asList(100, 101);
        List<String> topics1 = asList("t1", "t2");
        List<String> topics2 = asList("t5", "t6");
        List<String> allTopics = concat(topics1, topics2);

        Map<String, Integer> reps = merged(replicationFactors(topics1, replicas1.size()),
                replicationFactors(topics2, replicas2.size()));
        List<BrokerMetadata> brokers = createBrokers(broker0, broker0 + brokerCount, "rack1");

        Map<TopicPartition, List<Integer>> partitions1 = createPartitions(topics1, partitionCount, replicas1);
        Map<TopicPartition, List<Integer>> partitions2 = createPartitions(topics2, partitionCount, replicas2);
        ReplicaAssignment assignment = new ReplicaAssignment(merged(partitions1, partitions2));
        Metrics metrics = new Metrics(createRandomPartitionSizes(assignment.topicPartitions(), 40, 275));

        //When
        ReplicaAssignment reassigned = policy.rebalancePartitions(brokers, assignment, reps, metrics,
                Collections.<Broker>emptyList());

        List<Integer> reassignedBrokers = brokers(reassigned.asMap());
        Map<Integer, List<TopicPartition>> reassignedBrokerToPartitions = brokerToPartitions(reassigned.asMap());

        int expectedNumPartitions = allTopics.size() * partitionCount;
        int expectedNumReplicas = topics1.size() * partitionCount * replicas1.size() + topics2.size() *
                partitionCount * replicas2.size();

        //Then counts should match
        assertEquals(expectedNumPartitions, reassigned.size());
        assertEquals(expectedNumReplicas, reassignedBrokers.size());

        long clusterSize = sum(getAll(metrics.partitionToSize(), partitions1.keySet()).values()) * replicas1.size() +
                sum(getAll(metrics.partitionToSize(), partitions2.keySet()).values()) * replicas2.size();
        double expectedBrokerSize = clusterSize / (double) brokers.size();

        for (int brokerId = broker0; brokerId < broker0 + brokerCount; ++brokerId) {
            //Then replicas should be somewhat evenly spread (the replica size optimisation, which runs after the replica
            //count optimisation means that we cannot expect an optimal spread)
            double expected = Math.round(expectedNumReplicas / (double) brokerCount);
            assertEquals("Partition count outside expected range for broker " + brokerId, expected,
                    filter(reassignedBrokers, brokerId).size(), 4.0);

            long brokerSize = partitionSizeSum(metrics.partitionToSize(), reassignedBrokerToPartitions.get(brokerId));
            double tolerance = expectedBrokerSize * 0.1;
            assertEquals("Broker size outside expected range for broker " + brokerId, expectedBrokerSize,
                    brokerSize, tolerance);

            //Then leaders should be evenly spread
            int expectedLowerBound = (int) Math.floor(expectedNumPartitions / brokerCount);
            int actual = filter(leaders(reassigned.asMap()).values(), brokerId).size();
            assertTrue("expected:" + expectedLowerBound + "/+1 actual:" + actual,
                    expectedLowerBound == actual || expectedLowerBound + 1 == actual);
        }

        checkTopicFairness(allTopics, brokers, reps, metrics, reassigned.asMap(), reassignedBrokerToPartitions, 2.0, 1.2, 3.0);

    }

    @Test
    public void shouldRebalanceClusterWithDifferentSizedTopicsAndReplicaCountsAndRacks() {
        MovesOptimisedRebalancePolicy policy = new MovesOptimisedRebalancePolicy();
        int broker0 = 100;

        //Given
        int brokerCount = 171;
        int partitionCount = 139;
        List<Integer> replicas1 = asList(100, 101, 102);
        List<Integer> replicas2 = asList(100, 101);
        List<String> topics1 = asList("t1", "t2");
        List<String> topics2 = asList("t5", "t6");
        List<String> allTopics = concat(topics1, topics2);

        Map<String, Integer> reps = merged(replicationFactors(topics1, replicas1.size()),
                replicationFactors(topics2, replicas2.size()));
        List<BrokerMetadata> brokers = concat(
                createBrokers(broker0, broker0 + brokerCount / 2, "rack1"),
                createBrokers(broker0 + brokerCount / 2, broker0 + brokerCount, "rack2")
        );

        Map<TopicPartition, List<Integer>> partitions1 = createPartitions(topics1, partitionCount, replicas1);
        Map<TopicPartition, List<Integer>> partitions2 = createPartitions(topics2, partitionCount, replicas2);
        ReplicaAssignment assignment = new ReplicaAssignment(merged(partitions1, partitions2));
        Metrics metrics = new Metrics(createRandomPartitionSizes(assignment.topicPartitions(), 35, 150));

        //When
        ReplicaAssignment reassigned = policy.rebalancePartitions(brokers, assignment, reps, metrics,
                Collections.<Broker>emptyList());

        List<Integer> reassignedBrokers = brokers(reassigned.asMap());
        Map<Integer, List<TopicPartition>> reassignedBrokerToPartitions = brokerToPartitions(reassigned.asMap());

        int expectedNumPartitions = allTopics.size() * partitionCount;
        int expectedNumReplicas = topics1.size() * partitionCount * replicas1.size() + topics2.size() *
                partitionCount * replicas2.size();
        //Then counts should match
        assertEquals(expectedNumPartitions, reassigned.size());
        assertEquals(expectedNumReplicas, reassignedBrokers.size());

        int expectedNumReplicasInRack = (int) Math.round(expectedNumReplicas / 2.0);

        long clusterSize = sum(getAll(metrics.partitionToSize(), partitions1.keySet()).values()) * replicas1.size() +
                sum(getAll(metrics.partitionToSize(), partitions2.keySet()).values()) * replicas2.size();
        double expectedRackSize = clusterSize / 2.0;
        double expectedBrokerSize = clusterSize / (double) brokers.size();

        for (String rack : asList("rack2", "rack1")) {
            List<BrokerMetadata> brokersInRack = filterBrokersByRack(brokers, rack);

            int rackReplicaCount = rackReplicaCount(brokersInRack, reassignedBrokerToPartitions);
            assertEquals("Replica count outside expected range for rack " + rack, expectedNumReplicasInRack,
                    rackReplicaCount, 20.0);

            long rackSize = rackSize(metrics, brokersInRack, reassignedBrokerToPartitions);
            double tolerance = expectedRackSize * 0.03;
            assertEquals("Rack size outside expected range for rack " + rack, expectedRackSize,
                    rackSize, tolerance);
        }

        for (int brokerId = broker0; brokerId < broker0 + brokerCount; ++brokerId) {
            //Then replicas should be somewhat evenly spread (the replica size optimisation, which runs after the replica
            //count optimisation means that we cannot expect an optimal spread)
            double expected = Math.round(expectedNumReplicas / (double) brokerCount);
            assertEquals("Partition count outside expected range for broker " + brokerId, expected,
                    filter(reassignedBrokers, brokerId).size(), 4.0);

            long brokerSize = partitionSizeSum(metrics.partitionToSize(), reassignedBrokerToPartitions.get(brokerId));
            double tolerance = expectedBrokerSize * 0.05;
            assertEquals("Broker size outside expected range for broker " + brokerId, expectedBrokerSize,
                    brokerSize, tolerance);

            //Then leaders should be evenly spread
            int expectedLeaders = (int) Math.floor(expectedNumPartitions / brokerCount);
            int actual = filter(leaders(reassigned.asMap()).values(), brokerId).size();
            assertTrue("expected:" + expectedLeaders + "/+1 actual:" + actual,
                    expectedLeaders == actual || expectedLeaders + 1 == actual);
        }

        checkTopicFairness(allTopics, brokers, reps, metrics, reassigned.asMap(), reassignedBrokerToPartitions, 3.0, 1.4, 3.0);
    }

    @Test
    public void fairnessWithSparseTopics() {
        checkFairRebalance(new MovesOptimisedRebalancePolicy(), 133, 17, 37, asList(2, 3),
                asList(5, 7, 8, 10, 15), 35, 150);
    }

    @Test
    public void fairnessWithDenseTopics() {
        checkFairRebalance(new MovesOptimisedRebalancePolicy(), 40, 5, 40, asList(3, 4, 5),
                asList(25, 40, 60, 100), 35, 150);
    }

    @Test
    public void fairnessWithNormalTopics() {
        checkFairRebalance(new MovesOptimisedRebalancePolicy(), 100, 10, 20, asList(3, 4),
                asList(4, 5, 6), 35, 150);
    }

    @Test
    public void fairnessWithLargeTopics() {
        checkFairRebalance(new MovesOptimisedRebalancePolicy(), 100, 10, 20, asList(3, 4),
                asList(4, 5, 6), 1_000_000_000, 2_000_000_000);
    }

    private void checkFairRebalance(RebalancePolicy policy, int brokerCount, int rackCount, int topicCount,
                                    List<Integer> replicationFactors, List<Integer> partitionCounts, int minSize,
                                    int maxSize) {
        int broker0 = 100;

        List<String> allTopics = new ArrayList<>();
        // These are the topics whose partitionCount * replicationFactor is greater than the number of brokers.
        List<String> topicsWhichShouldBeEvenlyDistributed = new ArrayList<>();

        for (int i = 0; i < topicCount; ++i) {
            allTopics.add("topic" + i);
        }

        List<BrokerMetadata> brokers = new ArrayList<>();
        List<String> racks = new ArrayList<>();
        for (int i = 0; i < rackCount; ++i) {
            racks.add("rack" + i);
        }
        int expectedNumPartitions = 0;
        int expectedNumReplicas = 0;
        Map<String, Integer> reps =  new HashMap<>();
        Map<TopicPartition, List<Integer>> assignmentsMap = new HashMap<>();
        for (int i = 0; i < allTopics.size(); ++i) {
            int replicationFactor = replicationFactors.get(i % replicationFactors.size());
            int partitionCount = partitionCounts.get(i % partitionCounts.size());
            reps.put(allTopics.get(i), replicationFactor);
            List<Integer> replicas = new ArrayList<>();
            for (int j = 0; j < replicationFactor; ++j) {
                replicas.add(broker0 + j);
            }
            assignmentsMap.putAll(createPartitions(Collections.singletonList(allTopics.get(i)), partitionCount,
                    replicas));
            expectedNumPartitions += partitionCount;
            expectedNumReplicas += partitionCount * replicationFactor;
            if (brokerCount * 1.5 <= partitionCount * replicationFactor) {
                topicsWhichShouldBeEvenlyDistributed.add(allTopics.get(i));
            }
        }
        for (int i = 0; i < brokerCount; ++i) {
            int brokerId = broker0 + i;
            brokers.add(new BrokerMetadata(brokerId, Option.apply(racks.get(i % rackCount))));
        }

        ReplicaAssignment assignment = new ReplicaAssignment(assignmentsMap);
        Metrics metrics = new Metrics(createRandomPartitionSizes(assignment.topicPartitions(), minSize, maxSize));

        //When
        long startTime = System.currentTimeMillis();
        ReplicaAssignment reassigned = policy.rebalancePartitions(brokers, assignment, reps, metrics,
                Collections.<Broker>emptyList());
        LOG.debug("Rebalance took : {} ms", (System.currentTimeMillis() - startTime));

        List<Integer> reassignedBrokers = brokers(reassigned.asMap());
        Map<Integer, List<TopicPartition>> reassignedBrokerToPartitions = brokerToPartitions(reassigned.asMap());

        //Then counts should match
        assertEquals(expectedNumPartitions, reassigned.size());
        assertEquals(expectedNumReplicas, reassignedBrokers.size());

        int expectedNumReplicasInRack = Math.round(expectedNumReplicas / rackCount);

        long clusterSize = 0;
        Map<TopicPartition, Long> partitionToSize = metrics.partitionToSize();
        for (Map.Entry<TopicPartition, List<Integer>> replicaAssigment : assignmentsMap.entrySet()) {
            clusterSize += partitionToSize.get(replicaAssigment.getKey()) * replicaAssigment.getValue().size();
        }
        double expectedRackSize = clusterSize / rackCount;
        double expectedBrokerSize = clusterSize / (double) brokers.size();

        for (String rack : racks) {
            List<BrokerMetadata> brokersInRack = filterBrokersByRack(brokers, rack);

            int rackReplicaCount = rackReplicaCount(brokersInRack, reassignedBrokerToPartitions);
            assertEquals("Replica count outside expected range for rack " + rack, expectedNumReplicasInRack,
                    rackReplicaCount, Math.ceil(expectedNumReplicasInRack * 0.10));

            long rackSize = rackSize(metrics, brokersInRack, reassignedBrokerToPartitions);
            double tolerance = expectedRackSize * 0.03;
            assertEquals("Rack size outside expected range for rack " + rack, expectedRackSize,
                    rackSize, tolerance);
        }

        for (int brokerId = broker0; brokerId < broker0 + brokerCount; ++brokerId) {
            //Then replicas should be somewhat evenly spread (the replica size optimisation, which runs after the replica
            //count optimisation means that we cannot expect an optimal spread)
            double expected = Math.round(expectedNumReplicas / (double) brokerCount);
            // Sparse tests have very few partitions per broker, and dense tests have many. Hence the tolerance has to
            // be absolute in the former case, and relative in the latter. This should be achieved by the following.
            double partitionCountTolerance = Math.max(4, expected * 0.05);
            assertEquals("Partition count outside expected range for broker " + brokerId, expected,
                    filter(reassignedBrokers, brokerId).size(), partitionCountTolerance);

            long brokerSize = partitionSizeSum(metrics.partitionToSize(), reassignedBrokerToPartitions.get(brokerId));
            double tolerance = expectedBrokerSize * 0.15;
            assertEquals("Broker size outside expected range for broker " + brokerId, expectedBrokerSize,
                    brokerSize, tolerance);

            //Then leaders should be evenly spread
            int expectedLeaders = (int) Math.floor(expectedNumPartitions / brokerCount);
            int actual = filter(leaders(reassigned.asMap()).values(), brokerId).size();
            assertTrue("expected:" + expectedLeaders + "/+1 actual:" + actual,
                    expectedLeaders == actual || expectedLeaders + 1 == actual);
        }
        // We only want to do this when the number of TopicPartitions is far in excess of the number of available brokers.
        checkTopicFairness(topicsWhichShouldBeEvenlyDistributed, brokers, reps, metrics, reassigned.asMap(),
                reassignedBrokerToPartitions, 4.0, 1.4, 3.0);
    }

    private long rackSize(Metrics metrics, List<BrokerMetadata> brokersInRack, Map<Integer,
            List<TopicPartition>> brokerToPartitions) {
        long rackSize = 0;
        for (BrokerMetadata broker : brokersInRack) {
            List<TopicPartition> brokerPartitions = brokerToPartitions.get(broker.id());
            rackSize += sum(getAll(metrics.partitionToSize(), new HashSet<>(brokerPartitions)).values());
        }
        return rackSize;
    }

    private int rackReplicaCount(List<BrokerMetadata> brokersInRack, Map<Integer, List<TopicPartition>> brokerToPartitions) {
        int rackReplicaCount = 0;
        for (BrokerMetadata broker : brokersInRack) {
            List<TopicPartition> brokerPartitions = brokerToPartitions.get(broker.id());
            rackReplicaCount += brokerPartitions.size();
        }
        return rackReplicaCount;
    }

    private void checkTopicFairness(List<String> topics, List<BrokerMetadata> brokers, Map<String, Integer> reps, Metrics metrics,
                                    Map<TopicPartition, List<Integer>> reassigned, Map<Integer, List<TopicPartition>> reassignedBrokerToPartitions,
                                    double expectedPartitionCountTolerance, double expectedBrokerSizeMultiplier, double expectedLeaderCountTolerance) {

        for (String topic : topics) {

            List<TopicPartition> topicPartitions = filterPartitionsByTopic(reassigned.keySet(), topic);
            Map<TopicPartition, Long> topicPartitionToSize = new HashMap<>(metrics.partitionToSize());
            topicPartitionToSize.keySet().retainAll(topicPartitions);

            Map<TopicPartition, List<Integer>> reassignedForTopic = new HashMap<>(reassigned);
            reassignedForTopic.keySet().retainAll(topicPartitions);

            int replicationFactor = reps.get(topic);
            long topicSize = sum(topicPartitionToSize.values()) * replicationFactor;
            int partitionCount = topicPartitions.size();
            double expectedBrokerSizeForTopic = topicSize / (double) brokers.size();

            for (BrokerMetadata broker : brokers) {
                int brokerId = broker.id();
                List<TopicPartition> brokerPartitions = reassignedBrokerToPartitions.get(brokerId);
                assertNotNull("Broker should have non-null partitions " + brokerId, brokerPartitions);
                List<TopicPartition> brokerPartitionsForTopic = filterPartitionsByTopic(brokerPartitions, topic);

                //Then replicas should be somewhat evenly spread (the replica size optimisation, which runs after the replica
                //count optimisation means that we cannot expect an optimal spread)
                int expected = Math.round((float) partitionCount * replicationFactor / brokers.size());
                assertEquals("Partition count outside expected range for broker " + brokerId + " and topic " + topic,
                        expected, brokerPartitionsForTopic.size(), expectedPartitionCountTolerance);

                long brokerSize = partitionSizeSum(metrics.partitionToSize(), brokerPartitionsForTopic);
                double tolerance = expectedBrokerSizeForTopic * expectedBrokerSizeMultiplier;
                assertEquals("Broker topic size for topic " + topic + " outside expected range for broker " + brokerId,
                        expectedBrokerSizeForTopic, brokerSize, tolerance);

                //Then leaders should be evenly spread
                expected = partitionCount / brokers.size();
                assertEquals("Broker topic leadership for topic " + topic + " is uneven for " + brokerId, expected,
                        filter(leaders(reassignedForTopic).values(), brokerId).size(), expectedLeaderCountTolerance);
            }

        }
    }


}
