// (Copyright) [2016 - 2016] Confluent, Inc.
package io.confluent.kafka.databalancing;

import io.confluent.kafka.databalancing.constraint.Constraints;
import io.confluent.kafka.databalancing.exception.ValidationException;
import io.confluent.kafka.databalancing.metric.Metrics;
import io.confluent.kafka.databalancing.topology.Broker;
import io.confluent.kafka.databalancing.topology.ReplicaAssignment;
import kafka.admin.BrokerMetadata;

import kafka.common.Topic;
import org.apache.kafka.common.TopicPartition;
import org.junit.Test;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import static io.confluent.kafka.databalancing.TestUtils.*;
import static java.util.Arrays.asList;
import static java.util.Collections.emptyList;
import static org.apache.kafka.common.utils.Utils.sorted;
import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNotEquals;
import static org.junit.Assert.assertTrue;

public class RebalancingTest {

    //TODO test that should take a sample cluster and increase the replication factor?? really??
    //TODO tests that test number of moves is minimised (we should mock the move and make leader commands)

    private static final String topic = "topic";

    private static TopicPartition tp(int partition) {
        return new TopicPartition(topic, partition);
    }

    private static Map<String, Integer> replicationFactors(int replicationFactor) {
        return TestUtils.replicationFactors(asList(topic), replicationFactor);
    }

    private Map<TopicPartition,List<Integer>> rebalancePartitions(RebalancePolicy policy,
                                                                  List<BrokerMetadata> brokers,
                                                                  Map<TopicPartition, List<Integer>> partitions,
                                                                  Map<String, Integer> reps) {
        Metrics metrics = createSimpleMetrics(partitions.keySet());
        return rebalancePartitions(policy, brokers, partitions, reps, metrics.partitionToSize());
    }

    private Map<TopicPartition,List<Integer>> rebalancePartitions(RebalancePolicy policy,
                                                                  List<BrokerMetadata> brokers,
                                                                  Map<TopicPartition, List<Integer>> partitions,
                                                                  Map<String, Integer> reps,
                                                                  Map<TopicPartition, Long> partitionToSize) {
        return rebalancePartitions(policy, brokers, partitions, reps, partitionToSize, Collections.<Broker>emptyList());
    }

    private Map<TopicPartition,List<Integer>> rebalancePartitions(RebalancePolicy policy,
                                                                  List<BrokerMetadata> brokers,
                                                                  Map<TopicPartition, List<Integer>> partitions,
                                                                  Map<String, Integer> reps,
                                                                  Map<TopicPartition, Long> partitionToSize,
                                                                  List<Broker> brokersToRemove) {
        Metrics metrics = new Metrics(partitionToSize);
        return policy.rebalancePartitions(brokers, new ReplicaAssignment(partitions), reps, metrics, brokersToRemove).asMap();
    }

    private Map<TopicPartition,List<Integer>> rebalancePartitions(RebalancePolicy policy,
                                                                  List<BrokerMetadata> brokers,
                                                                  Map<TopicPartition, List<Integer>> partitions,
                                                                  Map<String, Integer> reps,
                                                                  List<Broker> brokersToRemove) {
        Metrics metrics = createSimpleMetrics(partitions.keySet());
        return policy.rebalancePartitions(brokers, new ReplicaAssignment(partitions), reps, metrics, brokersToRemove).asMap();
    }

    /**
     * Step 1: Ensure partitions are fully replicated
     */

    @Test
    public void shouldReReplicateOneUnderReplicatedPartition() {
        MovesOptimisedRebalancePolicy policy = new MovesOptimisedRebalancePolicy();;

        //Given
        List<BrokerMetadata> brokers = createBrokers(100, 105, "rack1");
        Map<TopicPartition, List<Integer>> underReplicated = createMap(entry(tp(0), asList(100, 101, 102)));
        Map<String, Integer> reps = replicationFactors(4);

        //When
        Map<TopicPartition, List<Integer>> reassigned = rebalancePartitions(policy, brokers, underReplicated, reps);

        //Then there should be four values. They should be on different Brokers
        assertEquals(4, reassigned.values().iterator().next().size());
        assertEquals(4, new HashSet<>(reassigned.values().iterator().next()).size());
    }

    @Test
    public void shouldFullyReplicatePartitions() {
        MovesOptimisedRebalancePolicy policy = new MovesOptimisedRebalancePolicy();;

        //Given two partitions. One under-replicated by 2 replicas. 2 empty brokers 103/104
        List<BrokerMetadata> brokers = createBrokers(100, 104, "rack1");
        Map<TopicPartition, List<Integer>> underReplicated = createMap(entry(tp(0), asList(100)));
        Map<String, Integer> reps = replicationFactors(4);

        //When
        MutableRebalanceContext context = createRebalanceContext(brokers, underReplicated, reps);
        Constraints constraints = new Constraints(context, null);
        policy.fullyReplicate(context, constraints);

        //Then p1 should have two new replicas on the two empty brokers, 103, 104
        assertEquals(4, context.brokers(tp(0)).size());
    }

    @Test
    public void shouldCreateMultipleReplicasPerPartitionIfNecessary() {
        MovesOptimisedRebalancePolicy policy = new MovesOptimisedRebalancePolicy();;

        //Given two partitions. One under-replicated by 2 replicas. 2 empty brokers 103/104
        List<BrokerMetadata> brokers = createBrokers(100, 104, "rack1");
        Map<TopicPartition, List<Integer>> underReplicated = createMap(
                entry(tp(0), asList(100, 101, 102)),
                entry(tp(1), asList(100, 101)),
                entry(tp(2), asList(100)),
                entry(tp(3), Collections.<Integer>emptyList()));
        int replicationFactor = 3;
        Map<String, Integer> reps = replicationFactors(replicationFactor);

        //When
        Map<TopicPartition, List<Integer>> reassigned = rebalancePartitions(policy, brokers, underReplicated, reps);
        //Then p1 should have two new replicas on the two empty brokers, 103, 104
        for (int partitionId = 0; partitionId <= 3; ++partitionId)
            assertEquals(replicationFactor, reassigned.get(tp(partitionId)).size());
    }

    @Test
    public void shouldPickLeastLoadedBrokerWhenReReplicatingUnderReplicatedPartitions() {
        MovesOptimisedRebalancePolicy policy = new MovesOptimisedRebalancePolicy();;

        //Given broker 102 is least loaded and p[4] is under-replicated
        List<BrokerMetadata> brokers = createBrokers(100, 104, "rack1");
        Map<TopicPartition, List<Integer>> underReplicated = createMap(
                entry(tp(0), asList(100, 101, 102)),
                entry(tp(1), asList(100, 102, 103)),
                entry(tp(2), asList(100, 101, 103)),
                entry(tp(3), asList(100, 101, 103)),
                entry(tp(4), asList(100, 101)));
        Map<String, Integer> reps = replicationFactors(3);
        Map<TopicPartition, Long> partitionToSize = createMap(
                entry(tp(0), 45L),
                entry(tp(1), 15L),
                entry(tp(2), 12L),
                entry(tp(3), 10L),
                entry(tp(4), 5L));
        MutableRebalanceContext context = createRebalanceContext(brokers, underReplicated, reps, partitionToSize);
        Constraints constraints = new Constraints(context, null);

        //When
        policy.fullyReplicate(context, constraints);

        //then p[4] should have a new replica on broker 103, the least loaded in terms of replica size (but not count)
        assertEquals(3, context.brokers(tp(4)).size());
        assertTrue(context.brokers(tp(4)).contains(new Broker(103)));
    }

    @Test
    public void shouldFavourDifferentRackWhenReReplicatingUnderreplicatedPartitions() {
        MovesOptimisedRebalancePolicy policy = new MovesOptimisedRebalancePolicy();;

        //Given
        List<BrokerMetadata> brokers = asList(bk(100, "rack1"), bk(101, "rack1"), bk(102, "rack2"), bk(103, "rack1"));
        Map<TopicPartition, List<Integer>> underReplicated = createMap(
                entry(tp(0), asList(100)));
        Map<String, Integer> reps = replicationFactors(2);

        //When we create a new replica for the under-replicated partition
        Map<TopicPartition, List<Integer>> reassigned = rebalancePartitions(policy, brokers, underReplicated, reps);

        //Then it should be created on the broker on a different rack (102)
        assertEquals(asList(100, 102), reassigned.get(tp(0)));
    }

    @Test
    public void shouldPreferLeastLoadedBrokersOnOtherRacksWhenReReplicatingUnderReplicatedPartitions() {
        MovesOptimisedRebalancePolicy policy = new MovesOptimisedRebalancePolicy();

        //Given broker 100 is least loaded and p[4] is under-replicated
        List<BrokerMetadata> brokers = concat(
                createBrokers(100, 103, "rack1"),
                createBrokers(103, 104, "rack2"));
        Map<TopicPartition, List<Integer>> underReplicated = createMap(
                entry(tp(0), asList(100, 101, 102)), //102 has 2 replicas, 103 has 3, 101 has 4, 100 has 5
                entry(tp(1), asList(100, 102, 103)),
                entry(tp(2), asList(100, 101, 103)),
                entry(tp(3), asList(100, 101, 103)),
                entry(tp(4), asList(100, 101)));
        Map<String, Integer> reps = replicationFactors(3);

        //When
        Map<TopicPartition, List<Integer>> reassigned = rebalancePartitions(policy, brokers, underReplicated, reps);

        //then p[4] should include a new replica. 102 is the least loaded,
        //but we should have picked 103 as it's on a different rack
        assertTrue(reassigned.get(tp(4)).contains(103));
    }

    @Test
    public void shouldNotReReplicateIfNoBrokerAvailableWithoutExistingReplica() {
        MovesOptimisedRebalancePolicy policy = new MovesOptimisedRebalancePolicy();
        List<BrokerMetadata> brokers = createBrokers(100, 103, "rack1");

        //Given
        Map<TopicPartition, List<Integer>> topology = createMap(entry(tp(0), asList(100, 101, 102)));
        Map<String, Integer> reps = replicationFactors(3);

        //When
        Map<TopicPartition, List<Integer>> reassigned = rebalancePartitions(policy, brokers, topology, reps);

        //Then nothing should have changed
        assertEquals(asList(100, 101, 102), reassigned.get(tp(0)));
    }

    @Test
    public void shouldContinueUnaffectedIfMoreReplicasThanReplicationFactor() {
        MovesOptimisedRebalancePolicy policy = new MovesOptimisedRebalancePolicy();
        List<BrokerMetadata> brokers = createBrokers(100, 103, "rack1");

        //Given
        Map<TopicPartition, List<Integer>> partitionWithThreeReplicas = createMap(entry(tp(0), asList(100, 101, 102)));
        Map<String, Integer> reps = replicationFactors(1);

        //When
        Map<TopicPartition, List<Integer>> reassigned = rebalancePartitions(policy, brokers, partitionWithThreeReplicas, reps);

        //Then nothing should have changed
        assertEquals(asList(100, 101, 102), reassigned.get(tp(0)));
    }


    @Test
    public void shouldContinueEvenIfNoAvailableBrokersForCreatingFullyReplicatedPartitions() {
        MovesOptimisedRebalancePolicy policy = new MovesOptimisedRebalancePolicy();

        //Given three brokers, single partition and a replication factor of 4
        Map<TopicPartition, List<Integer>> partitionWithThreeReplicas = createMap(entry(tp(0), asList(100, 101, 102)));
        List<BrokerMetadata> brokers = createBrokers(100, 103, "rack1");
        Map<String, Integer> reps = replicationFactors(4);

        //When
        Map<TopicPartition, List<Integer>> reassigned = rebalancePartitions(policy, brokers, partitionWithThreeReplicas, reps);

        //Then we should still have only 3 replicas (hence still under-replicated)
        assertEquals(3, reassigned.get(tp(0)).size());
    }


    /**
     * Step 2.1: Optimise for replica fairness across racks
     */
    @Test
    public void shouldOptimiseForEvenReplicaPlacementAcrossTwoSimpleRacks() {
        MovesOptimisedRebalancePolicy policy = new MovesOptimisedRebalancePolicy();

        //Given
        List<BrokerMetadata> brokers = asList(bk(100, "rack1"), bk(101, "rack2"));
        Map<TopicPartition, List<Integer>> partitions = createMap(
                entry(tp(0), asList(100)),
                entry(tp(1), asList(100)));
        Map<String, Integer> reps = replicationFactors(1);

        //When
        Map<TopicPartition, List<Integer>> reassigned = rebalancePartitions(policy, brokers, partitions, reps);

        //Then should be one per rack
        assertEquals(createMap(entry(tp(0), asList(101)), entry(tp(1), asList(100))), reassigned);
    }

    @Test
    public void shouldOptimiseForEvenReplicaPlacementAcrossManyRacks() {
        MovesOptimisedRebalancePolicy policy = new MovesOptimisedRebalancePolicy();

        //Given all replicas are on one (of 3) racks
        List<BrokerMetadata> brokers = asList(bk(100, "rack1"), bk(101, "rack1"), bk(102, "rack2"), bk(103, "rack2"),
                bk(104, "rack3"), bk(105, "rack3"));
        Map<TopicPartition, List<Integer>> partitions = createMap(
                entry(tp(0), asList(100)),
                entry(tp(1), asList(101)),
                entry(tp(2), asList(100)),
                entry(tp(3), asList(101)),
                entry(tp(4), asList(100)),
                entry(tp(5), asList(101)));
        Map<String, Integer> reps = replicationFactors(1);

        //When
        Map<TopicPartition, List<Integer>> reassigned = rebalancePartitions(policy, brokers, partitions, reps);

        //Then should end evenly spread one replica per broker and hence two per rack
        assertEquals(asList(100, 101, 102, 103, 104, 105), sorted(brokers(reassigned)));
    }

    @Test
    public void shouldOptimiseForFairnessWithinAndAcrossRacks() {
        MovesOptimisedRebalancePolicy policy = new MovesOptimisedRebalancePolicy();

        //Given all replicas are on one (of 3) racks
        List<BrokerMetadata> brokers = asList(bk(100, "rack1"), bk(101, "rack1"), bk(102, "rack1"), bk(103, "rack2"),
                bk(104, "rack2"), bk(105, "rack2"));
        Map<TopicPartition, List<Integer>> partitions = createMap(
                entry(tp(0), asList(100, 101, 102)),
                entry(tp(1), asList(100, 101, 102)),
                entry(tp(2), asList(100, 101, 102)),
                entry(tp(3), asList(100, 101, 102)),
                entry(tp(4), asList(100, 101, 102)),
                entry(tp(5), asList(100, 101, 102)));
        Map<String, Integer> reps = replicationFactors(3);

        //When
        Map<TopicPartition, List<Integer>> reassigned = rebalancePartitions(policy, brokers, partitions, reps);

        //Then should end evenly spread one replica per broker and hence two per rack
        for (int brokerId = 100; brokerId <= 105; brokerId++)
            assertEquals(3, filter(brokers(reassigned), brokerId).size());

        //Should have one leaders each
        for (int brokerId = 100; brokerId <= 105; brokerId++)
            assertEquals(1, filter(leaders(reassigned).values(), brokerId).size());

    }

    @Test
    public void shouldOptimiseForSizeFairnessWithinAndAcrossRacks() {
        MovesOptimisedRebalancePolicy policy = new MovesOptimisedRebalancePolicy();

        //Given all replicas are on one (of 3) racks
        List<BrokerMetadata> brokers = asList(bk(100, "rack1"), bk(101, "rack1"), bk(102, "rack1"), bk(103, "rack2"),
                bk(104, "rack2"), bk(105, "rack2"));
        Map<TopicPartition, List<Integer>> partitions = createMap(
                entry(tp(0), asList(100, 101, 102)),
                entry(tp(1), asList(100, 101, 102)),
                entry(tp(2), asList(100, 101, 102)),
                entry(tp(3), asList(100, 101, 102)),
                entry(tp(4), asList(100, 101, 102)),
                entry(tp(5), asList(100, 101, 102)));
        Map<String, Integer> reps = replicationFactors(3);
        Map<TopicPartition, Long> partitionToSize = createMap(
                entry(tp(0), 5L),
                entry(tp(1), 3L),
                entry(tp(2), 2L),
                entry(tp(3), 10L),
                entry(tp(4), 6L),
                entry(tp(5), 4L)
        );

        //When
        Map<TopicPartition, List<Integer>> reassigned = rebalancePartitions(policy, brokers, partitions, reps,
                partitionToSize);

        //Then should end evenly spread one replica per broker and hence two per rack
        for (int brokerId = 100; brokerId <= 105; brokerId++) {
            long sum = partitionSizeSum(partitionToSize, brokerPartitions(reassigned, brokerId));
            long expected = Math.round(sum(partitionToSize.values()) * 3 / (double) brokers.size());
            assertEquals("Unexpected replica size for broker " + brokerId, expected, sum, 1.0);
        }

        //Should have one leaders each
        for (int brokerId = 100; brokerId <= 105; brokerId++)
            assertEquals(1, filter(leaders(reassigned).values(), brokerId).size());

    }

    @Test
    public void shouldFindFairnessWhereBrokersPerRacksAreUnevenWithTwoReplicas() {
        MovesOptimisedRebalancePolicy policy = new MovesOptimisedRebalancePolicy();

        //Given 6 partitions, 2 replicas each, all starting out on the first two (of 3) brokers (rack1)
        Map<TopicPartition, List<Integer>> partitions = new HashMap<>();
        for (int x = 0; x < 6; ++x)
            partitions.put(tp(x), asList(100, 101));
        List<BrokerMetadata> brokers = asList(bk(100, "rack1"), bk(101, "rack1"), bk(102, "rack2"));
        Map<String, Integer> reps = replicationFactors(2);

        //When
        Map<TopicPartition, List<Integer>> reassigned = rebalancePartitions(policy, brokers, partitions, reps);

        //Then we should still have 12 replicas
        assertEquals(12, brokers(reassigned).size());

        //Broker 100/102 should have 3 replicas each. 102 should have 6 replicas
        assertEquals(3, filter(brokers(reassigned), 100).size()); //rack1
        assertEquals(3, filter(brokers(reassigned), 101).size()); //rack1
        assertEquals(6, filter(brokers(reassigned), 102).size()); //rack2
    }

    @Test
    public void shouldObeyRackFairnessStrictlyEvenAtTheCostOfReplicaFairnessAcrossBrokers() {
        MovesOptimisedRebalancePolicy policy = new MovesOptimisedRebalancePolicy();

        //Given two racks and three brokers, Par is two replicas per rack.
        List<BrokerMetadata> brokers = asList(bk(100, "rack1"), bk(101, "rack1"), bk(102, "rack2"));
        Map<TopicPartition, List<Integer>> partitions = createMap(
                entry(tp(0), asList(100)),
                entry(tp(1), asList(100)),
                entry(tp(2), asList(101)),
                entry(tp(3), asList(101)));
        Map<String, Integer> reps = replicationFactors(1);

        //When
        Map<TopicPartition, List<Integer>> reassigned = rebalancePartitions(policy, brokers, partitions, reps);

        //Then broker 102 (the only broker on rack2) should end up with
        //two partitions (to obey rack fairness)
        assertEquals(2, filter(brokers(reassigned), 102).size());
        assertEquals(1, filter(brokers(reassigned), 100).size());
        assertEquals(1, filter(brokers(reassigned), 101).size());
        for (Map.Entry<TopicPartition, List<Integer>> entry : reassigned.entrySet())
            assertEquals("Partition " + entry.getKey() + " should have one replica", 1, entry.getValue().size());
        assertEquals(4, reassigned.size());
    }

    @Test
    public void shouldAchieveFairnessAcrossRacksWithMultipleTopics() {
        MovesOptimisedRebalancePolicy policy = new MovesOptimisedRebalancePolicy();

        //Given
        List<BrokerMetadata> brokers = asList(bk(100, "rack1"), bk(101, "rack2"));
        Map<TopicPartition, List<Integer>> partitions = createMap(
                entry(new TopicPartition("sales", 0), asList(100)),
                entry(new TopicPartition("sales", 1), asList(100)),
                entry(new TopicPartition("orders", 0), asList(100)),
                entry(new TopicPartition("orders", 1), asList(100)));
        Map<String, Integer> reps = createMap(entry("sales", 1), entry("orders", 1));

        //When
        Map<TopicPartition, List<Integer>> reassigned = rebalancePartitions(policy, brokers, partitions, reps);

        //Then should be one per rack
        for (List<Integer> replicaAssignment : reassigned.values())
            assertEquals(1, replicaAssignment.size());

        //The two partitions of each topic should be on different brokers
        for (String topic : asList("sales", "orders"))
            assertNotEquals(reassigned.get(new TopicPartition(topic, 0)), reassigned.get(new TopicPartition(topic, 1)));
    }

    @Test
    public void shouldAchieveTopicFairnessAcrossRacks() {
        MovesOptimisedRebalancePolicy policy = new MovesOptimisedRebalancePolicy();

        //Given
        List<BrokerMetadata> brokers = concat(createBrokers(100, 102, "rack1"),
                createBrokers(102, 104, "rack2"));
        Map<TopicPartition, List<Integer>> partitions = createMap(
                entry(new TopicPartition("t1", 0), asList(100)),
                entry(new TopicPartition("t1", 1), asList(101)),
                entry(new TopicPartition("t2", 0), asList(102)),
                entry(new TopicPartition("t2", 1), asList(103)));
        Map<String, Integer> reps = createMap(entry("t1", 1), entry("t2", 1));
        Map<TopicPartition, Long> partitionToSize = createMap(
                entry(new TopicPartition("t1", 0), 16L),
                entry(new TopicPartition("t1", 1), 12L),
                entry(new TopicPartition("t2", 0), 12L),
                entry(new TopicPartition("t2", 1), 18L));

        //When
        Map<TopicPartition, List<Integer>> reassigned = rebalancePartitions(policy, brokers, partitions, reps,
                partitionToSize);

        assertEquals(4, reassigned.size());
        assertEquals(4, new HashSet<>(brokers(reassigned)).size());

        //Each broker should have one partition
        for (List<Integer> replicaAssignment : reassigned.values())
            assertEquals(1, replicaAssignment.size());

        //Partitions for topics `t1` and `t2` should not be on the same rack
        for (String topic : asList("t1", "t2")) {
            Integer b1 = reassigned.get(new TopicPartition(topic, 0)).get(0);
            String rack1 = brokerRack(brokers, b1);
            Integer b2 = reassigned.get(new TopicPartition(topic, 1)).get(0);
            String rack2 = brokerRack(brokers, b2);
            assertNotEquals("Partitions for topic " + topic + " should not be in the same rack", rack1, rack2);
        }
    }


    /**
     * Step 2.2: Optimise for leader fairness across racks
     */
    @Test
    public void shouldOptimiseForLeaderFairnessAcrossRacks() {
        MovesOptimisedRebalancePolicy policy = new MovesOptimisedRebalancePolicy();

        //Given
        List<BrokerMetadata> brokers = asList(bk(100, "rack1"), bk(101, "rack2"));
        Map<TopicPartition, List<Integer>> partitions = createMap(
                entry(tp(0), asList(100, 101)),
                entry(tp(1), asList(100, 101)));
        Map<String, Integer> reps = replicationFactors(2);

        //When
        Map<TopicPartition, List<Integer>> reassigned = rebalancePartitions(policy, brokers, partitions, reps);

        //Then should be one per rack
        assertEquals(createMap(entry(tp(0), asList(101, 100)), entry(tp(1), asList(100, 101))), reassigned);
    }

    @Test
    public void shouldOptimiseForLeaderFairnessAcrossThreeRacks() {
        MovesOptimisedRebalancePolicy policy = new MovesOptimisedRebalancePolicy();

        //Given
        List<BrokerMetadata> brokers = asList(bk(100, "rack1"), bk(101, "rack2"), bk(102, "rack3"));
        Map<TopicPartition, List<Integer>> partitions = createMap(
                entry(tp(0), asList(100, 101)),
                entry(tp(1), asList(100, 102)),
                entry(tp(2), asList(101, 102)));
        Map<String, Integer> reps = replicationFactors(2);

        //When
        Map<TopicPartition, List<Integer>> reassigned = rebalancePartitions(policy, brokers, partitions, reps);

        //Then should be one per rack
        assertEquals(createMap(entry(tp(0), asList(100, 101)), entry(tp(1), asList(102, 100)),
                entry(tp(2), asList(101, 102))), reassigned);
    }

    @Test
    public void shouldNotOptimiseIfAlreadyFair() {
        MovesOptimisedRebalancePolicy policy = new MovesOptimisedRebalancePolicy();

        //Given
        List<BrokerMetadata> brokers = asList(bk(100, "rack1"), bk(101, "rack2"));
        Map<TopicPartition, List<Integer>> partitions = createMap(
                entry(tp(0), asList(100, 101)),
                entry(tp(1), asList(101, 100)));
        Map<String, Integer> reps = replicationFactors(2);

        //When
        Map<TopicPartition, List<Integer>> reassigned = rebalancePartitions(policy, brokers, partitions, reps);

        //Then should be one per rack
        assertEquals(createMap(entry(tp(0), asList(100, 101)), entry(tp(1), asList(101, 100))), reassigned);
    }

    /**
     * Step 3.1: Optimise for replica fairness across brokers
     */

    @Test
    public void shouldOptimiseForEvenReplicaPlacementAcrossBrokersSimple() {
        MovesOptimisedRebalancePolicy policy = new MovesOptimisedRebalancePolicy();

        //Given
        List<BrokerMetadata> brokers = asList(bk(100, "rack1"), bk(101, "rack1"));
        Map<TopicPartition, List<Integer>> partitions = createMap(
                entry(tp(0), asList(100)),
                entry(tp(1), asList(100)));
        Map<String, Integer> reps = replicationFactors(1);

        //When
        Map<TopicPartition, List<Integer>> reassigned = rebalancePartitions(policy, brokers, partitions, reps);

        //Then should be one per rack
        assertEquals(createMap(entry(tp(0), asList(101)), entry(tp(1), asList(100))), reassigned);
    }

    @Test
    public void shouldOptimiseForEvenReplicaPlacementAcrossBrokersWithNoRacks() {
        MovesOptimisedRebalancePolicy policy = new MovesOptimisedRebalancePolicy();

        //Given
        List<BrokerMetadata> brokers = asList(bk(100, null), bk(101, null), bk(102, null));
        Map<TopicPartition, List<Integer>> partitions = createMap(
                entry(tp(0), asList(100, 101)),
                entry(tp(1), asList(101, 100)),
                entry(tp(2), asList(100)));
        Map<String, Integer> reps = replicationFactors(2);

        //When
        Map<TopicPartition, List<Integer>> reassigned = rebalancePartitions(policy, brokers, partitions, reps);

        //Then
        assertEquals(createMap(entry(tp(0), asList(102, 101)), entry(tp(1), asList(101, 100)),
                entry(tp(2), asList(100, 102))), reassigned);
    }

    @Test
    public void shouldOptimiseForEvenReplicaPlacementAcrossBrokersOnSingleRack() {
        MovesOptimisedRebalancePolicy policy = new MovesOptimisedRebalancePolicy();

        //Given
        List<BrokerMetadata> brokers = asList(bk(100, "rack1"), bk(101, "rack1"), bk(102, "rack1"), bk(103, "rack1"));
        Map<TopicPartition, List<Integer>> partitions = createMap(
                entry(tp(0), asList(100, 101)),
                entry(tp(1), asList(100, 101)),
                entry(tp(2), asList(100, 101)),
                entry(tp(3), asList(100, 101)));
        Map<String, Integer> reps = replicationFactors(2);

        //When
        Map<TopicPartition, List<Integer>> reassigned = rebalancePartitions(policy, brokers, partitions, reps);

        //Then should be two per broker
        for (int brokerId = 100; brokerId <= 103; brokerId++)
            assertEquals(2, filter(brokers(reassigned), brokerId).size());
    }

    @Test //TODO is this test redundant?
    public void shouldMoveReplicasToLeastLoadedBroker() {
        MovesOptimisedRebalancePolicy policy = new MovesOptimisedRebalancePolicy();

        //Given
        List<BrokerMetadata> brokers = asList(bk(100, "rack1"), bk(101, "rack1"), bk(102, "rack1"), bk(103, "rack1"));
        //100, 2 replicas, 101-> 3 replicas, 102, 2 replicas, 103, 1 replica
        Map<TopicPartition, List<Integer>> partitions = createMap(
                entry(tp(0), asList(100, 101)),
                entry(tp(1), asList(100, 102)),
                entry(tp(2), asList(102, 101)),
                entry(tp(3), asList(103, 101)));
        Map<String, Integer> reps = replicationFactors(2);

        //When
        Map<TopicPartition, List<Integer>> reassigned = rebalancePartitions(policy, brokers, partitions, reps);

        //Then should have moved one replica from 101, 100
        for (int brokerId = 100; brokerId <= 103; brokerId++)
            assertEquals(2, filter(brokers(reassigned), brokerId).size());
    }

    @Test
    public void shouldNotMoveReplicaIfBreaksRackConstraint() {
        MovesOptimisedRebalancePolicy policy = new MovesOptimisedRebalancePolicy();

        //Given
        List<BrokerMetadata> brokers = asList(bk(100, "rack1"), bk(101, "rack2"), bk(102, "rack2"));
        Map<TopicPartition, List<Integer>> partitions = createMap(
                entry(tp(0), asList(100, 101)),
                entry(tp(1), asList(100, 101)),
                entry(tp(2), asList(100, 101)),
                entry(tp(3), asList(100, 101)),
                entry(tp(4), asList(100, 101)),
                entry(tp(5), asList(100, 101)));
        Map<String, Integer> reps = replicationFactors(2);

        //When
        Map<TopicPartition, List<Integer>> reassigned = rebalancePartitions(policy, brokers, partitions, reps);

        //Then All replicas on 100 should remain there (i.e. on rack 1)
        assertEquals(6, filter(brokers(reassigned), 100).size());
    }

    @Test
    public void shouldNotMoveReplicaFromOverParBrokerWhereReplicasDoNotBalanceEvenly() {
        MovesOptimisedRebalancePolicy policy = new MovesOptimisedRebalancePolicy();

        //Given 5 brokers and 6 partitions with all replicas on broker 100
        List<BrokerMetadata> brokers = createBrokers(100, 105, "rack1");
        Map<TopicPartition, List<Integer>> partitions = createMap(
                entry(tp(0), asList(100)),
                entry(tp(1), asList(100)),
                entry(tp(2), asList(100)),
                entry(tp(3), asList(100)),
                entry(tp(4), asList(100)),
                entry(tp(5), asList(100)));
        Map<String, Integer> reps = replicationFactors(1);

        //When
        Map<TopicPartition, List<Integer>> reassigned = rebalancePartitions(policy, brokers, partitions, reps);

        //One broker will have an extra replica. It should be 100
        assertEquals(2, filter(brokers(reassigned), 100).size());
        //The others should have just one replica
        for (int id = 101; id <= 104; ++id)
            assertEquals(1, filter(brokers(reassigned), id).size());
        //We should have made only 4 moves
        assertEquals(4, policy.replicasMoved());
    }


    /**
     * Step 3.2: Optimise for leader fairness across brokers
     */
    @Test
    public void shouldOptimiseForLeaderFairnessAcrossBrokersSimple() {
        MovesOptimisedRebalancePolicy policy = new MovesOptimisedRebalancePolicy();

        //Given
        List<BrokerMetadata> brokers = asList(bk(100, "rack1"), bk(101, "rack1"));
        Map<TopicPartition, List<Integer>> partitions = createMap(
                entry(tp(0), asList(100, 101)),
                entry(tp(1), asList(100, 101)));
        Map<String, Integer> reps = replicationFactors(2);

        //When
        Map<TopicPartition, List<Integer>> reassigned = rebalancePartitions(policy, brokers, partitions, reps);

        //Then should be one per broker
        Collection<Integer> leaders = leaders(reassigned).values();
        assertEquals(2, new HashSet<>(leaders).size());
    }

    @Test
    public void shouldOptimiseForLeaderFairnessAcrossBrokers() {
        MovesOptimisedRebalancePolicy policy = new MovesOptimisedRebalancePolicy();

        //Given
        List<BrokerMetadata> brokers = asList(bk(100, "rack1"), bk(101, "rack1"), bk(102, "rack1"));
        Map<TopicPartition, List<Integer>> partitions = createMap(
                entry(tp(0), asList(100, 101, 102)),
                entry(tp(1), asList(100, 101, 102)),
                entry(tp(3), asList(100, 101, 102)));
        Map<String, Integer> reps = replicationFactors(3);

        //When
        Map<TopicPartition, List<Integer>> reassigned = rebalancePartitions(policy, brokers, partitions, reps);

        //Then should be one per broker
        Collection<Integer> leaders = leaders(reassigned).values();
        assertEquals(3, new HashSet<>(leaders).size());
    }

    @Test
    public void shouldOptimiseForLeaderFairnessAcrossNoRacks() {
        MovesOptimisedRebalancePolicy policy = new MovesOptimisedRebalancePolicy();

        //Given
        List<BrokerMetadata> brokers = asList(bk(100, null), bk(101, null), bk(102, null));
        Map<TopicPartition, List<Integer>> partitions = createMap(
                entry(tp(0), asList(100, 101, 102)),
                entry(tp(1), asList(100, 101, 102)),
                entry(tp(3), asList(100, 101, 102)));
        Map<String, Integer> reps = replicationFactors(3);

        //When
        Map<TopicPartition, List<Integer>> reassigned = rebalancePartitions(policy, brokers, partitions, reps);

        //Then should be one per broker
        Collection<Integer> leaders = leaders(reassigned).values();
        assertEquals(3, new HashSet<>(leaders).size());
    }

    @Test
    public void shouldOptimiseLeaderFairnessWithinRacks() {
        MovesOptimisedRebalancePolicy policy = new MovesOptimisedRebalancePolicy();

        //Given
        List<BrokerMetadata> brokers = asList(bk(100, "rack1"), bk(101, "rack2"), bk(102, "rack2"));
        Map<Integer, String> brokerToRack = new HashMap<>();
        for (BrokerMetadata broker : brokers) {
            if (broker.rack().isDefined())
                brokerToRack.put(broker.id(), broker.rack().get());
        }

        Map<TopicPartition, List<Integer>> partitions = createMap(
                entry(tp(0), asList(100, 101)),
                entry(tp(1), asList(100, 101)),
                entry(tp(2), asList(100, 101)),
                entry(tp(3), asList(100, 101)),
                entry(tp(4), asList(100, 101)),
                entry(tp(5), asList(100, 101)));

        Map<String, Integer> reps = replicationFactors(2);

        //When
        Map<TopicPartition, List<Integer>> reassigned = rebalancePartitions(policy, brokers, partitions, reps);

        //rack 1 & 2 should get 6 replicas despite having different numbers of brokers
        assertEquals(6, rackReplicaCount(brokers, brokers(reassigned), "rack1"));
        assertEquals(6, rackReplicaCount(brokers, brokers(reassigned), "rack2"));

        //each broker should have 2 leaders
        for (int id = 100; id <= 102; ++id)
            assertEquals(2, filter(leaders(reassigned).values(), id).size());

    }

    private int rackReplicaCount(List<BrokerMetadata> allBrokers, List<Integer> brokerIds, String rack) {
        int count = 0;
        for (Integer brokerId : brokerIds) {
            String brokerRack = brokerRack(allBrokers, brokerId);
            if (brokerRack == null ? rack == null : brokerRack.equals(rack))
                ++count;
        }
        return count;
    }

    private String brokerRack(List<BrokerMetadata> allBrokers, int brokerId) {
        for (BrokerMetadata broker : allBrokers) {
            if (broker.id() == brokerId) {
                if (broker.rack().isDefined())
                    return broker.rack().get();
                return null;
            }
        }
        return null;
    }

    @Test
    public void shouldNotMoveLeadershipFromOverParBrokerWhereLeadersDoNotBalanceEvenly() {
        MovesOptimisedRebalancePolicy policy = new MovesOptimisedRebalancePolicy();

        //Given 5 brokers and 6 partitions with all replicas on brokers 100 and 101
        List<BrokerMetadata> brokers = createBrokers(100, 105, "rack1");
        Map<TopicPartition, List<Integer>> partitions = createMap(
                entry(tp(0), asList(100, 101)),
                entry(tp(1), asList(100, 101)),
                entry(tp(2), asList(100, 101)),
                entry(tp(3), asList(100, 101)),
                entry(tp(4), asList(100, 101)),
                entry(tp(5), asList(100, 101)));
        Map<String, Integer> reps = replicationFactors(2);

        //When
        Map<TopicPartition, List<Integer>> reassigned = rebalancePartitions(policy, brokers, partitions, reps);

        //100/101 will have an extra replica
        assertEquals(3, filter(brokers(reassigned), 100).size());
        assertEquals(3, filter(brokers(reassigned), 101).size());

        for (int id  = 102; id <= 104; id++)
            assertEquals("Unexpected number of replicas for broker id " + id, 2,
                    filter(brokers(reassigned), id).size());

        assertEquals(6, policy.replicasMoved());

        //One broker will have an extra leader. It should be 100 (i.e. it stayed there)
        assertEquals(2, filter(leaders(reassigned).values(), 100).size());
        //The others should have just one leader
        for (int id : asList(101, 102, 103, 104))
            assertEquals(1, filter(leaders(reassigned).values(), id).size());

        assertEquals(6, policy.replicasMoved());
    }

    /**
     * Tests for multiple topics
     */

    @Test
    public void shouldBalanceLeadersOverMultipleTopicsSingleRack() {
        MovesOptimisedRebalancePolicy policy = new MovesOptimisedRebalancePolicy();

        //Given
        List<BrokerMetadata> brokers = createBrokers(100, 106, "rack1");
        Map<TopicPartition, List<Integer>> partitions = createMap(
                entry(new TopicPartition("t1", 0), asList(100, 101, 102)),
                entry(new TopicPartition("t1", 1), asList(100, 101, 102)),
                entry(new TopicPartition("t2", 0), asList(100, 101, 102)),
                entry(new TopicPartition("t2", 1), asList(103, 104, 105)),
                entry(new TopicPartition("t3", 0), asList(103, 104, 105)),
                entry(new TopicPartition("t3", 1), asList(103, 104, 105)));
        Map<String, Integer> reps = createMap(entry("t1", 3), entry("t2", 3), entry("t3", 3));

        //When
        Map<TopicPartition, List<Integer>> reassigned = rebalancePartitions(policy, brokers, partitions, reps);

        //Then should be one leader per broker
        Collection<Integer> leaders = leaders(reassigned).values();
        assertEquals(6, leaders.size());

        //The two partitions of each topic should have different brokers as leaders
        for (String topic : reps.keySet())
            assertNotEquals(reassigned.get(new TopicPartition(topic, 0)).get(0),
                    reassigned.get(new TopicPartition(topic, 1)).get(0));
    }

    @Test
    public void shouldBalanceLeadersOverMultipleTopicsAndMultipleRacks() {
        MovesOptimisedRebalancePolicy policy = new MovesOptimisedRebalancePolicy();

        //Given
        List<BrokerMetadata> brokers = asList(bk(100, "rack1"), bk(101, "rack1"), bk(102, "rack2"), bk(103, "rack2"));
        Map<TopicPartition, List<Integer>> partitions = createMap(
                entry(new TopicPartition("t1", 0), asList(100, 101, 102)),
                entry(new TopicPartition("t2", 0), asList(100, 101, 102)),
                entry(new TopicPartition("t3", 0), asList(100, 101, 102)),
                entry(new TopicPartition("t3", 1), asList(100, 101, 102)));
        Map<String, Integer> reps = createMap(entry("t1", 3), entry("t2", 3), entry("t3", 3));

        //When
        Map<TopicPartition, List<Integer>> reassigned = rebalancePartitions(policy, brokers, partitions, reps);

        //Then should be one leader per broker
        Collection<Integer> leaders = leaders(reassigned).values();
        assertEquals(4, leaders.size());

        //The two partitions of `t3` should have different brokers as leaders
        assertNotEquals(reassigned.get(new TopicPartition("t3", 0)).get(0),
                reassigned.get(new TopicPartition("t3", 1)).get(0));
    }

    @Test
    public void shouldBalanceReplicasOverMultipleTopicsSingleRack() {
        MovesOptimisedRebalancePolicy policy = new MovesOptimisedRebalancePolicy();

        //Given
        List<BrokerMetadata> brokers = createBrokers(100, 106, "rack1");
        Map<TopicPartition, List<Integer>> partitions = createMap(
                entry(new TopicPartition("t1", 0), asList(100, 101)),
                entry(new TopicPartition("t1", 1), asList(100, 101)),
                entry(new TopicPartition("t2", 0), asList(100, 101)),
                entry(new TopicPartition("t2", 1), asList(100, 101)),
                entry(new TopicPartition("t3", 0), asList(103, 104)),
                entry(new TopicPartition("t3", 1), asList(103, 104)));
        Map<String, Integer> reps = createMap(entry("t1", 2), entry("t2", 2), entry("t3", 2));

        //When
        Map<TopicPartition, List<Integer>> reassigned = rebalancePartitions(policy, brokers, partitions, reps);

        //Two replicas per broker
        for (int brokerId = 100; brokerId <= 102; ++brokerId)
            assertEquals(2, filter(brokers(reassigned), brokerId).size());
        //One Leader per broker
        for (int brokerId = 100; brokerId <= 102; ++brokerId)
            assertEquals(1, filter(leaders(reassigned).values(), brokerId).size());

        //Leader and Partitions for the same topic should not be on the same broker
        for (String topic : reps.keySet()) {
            assertNotEquals(reassigned.get(new TopicPartition(topic, 0)).get(0),
                    reassigned.get(new TopicPartition(topic, 1)).get(0));

            Set<Integer> intersection = intersection(reassigned.get(new TopicPartition(topic, 0)),
                    reassigned.get(new TopicPartition(topic, 1)));
            assertEquals("The same broker(s) appeared in both partitions of topic " + topic, Collections.emptySet(),
                    intersection);
        }
    }

    private <E> Set<E> intersection(Collection<E> c1, Collection<E> c2) {
        Set<E> intersection = new HashSet<>(c1);
        intersection.retainAll(c2);
        return intersection;
    }

    @Test
    public void shouldBalanceReplicasOverMultipleTopicsAndMultipleRacks() {
        MovesOptimisedRebalancePolicy policy = new MovesOptimisedRebalancePolicy();

        //Given
        List<BrokerMetadata> brokers = asList(bk(100, "rack1"), bk(101, "rack1"), bk(102, "rack2"), bk(103, "rack2"));
        Map<TopicPartition, List<Integer>> partitions = createMap(
                entry(new TopicPartition("t1", 0), asList(100, 101)),
                entry(new TopicPartition("t1", 1), asList(100, 101)),
                entry(new TopicPartition("t2", 0), asList(100, 101)),
                entry(new TopicPartition("t2", 1), asList(100, 101)),
                entry(new TopicPartition("t3", 0), asList(100, 101)),
                entry(new TopicPartition("t3", 1), asList(100, 101)),
                entry(new TopicPartition("t4", 0), asList(100, 101)),
                entry(new TopicPartition("t4", 1), asList(100, 101)));
        Map<String, Integer> reps = createMap(entry("t1", 2), entry("t2", 2), entry("t3", 2), entry("t4", 2));

        //When
        Map<TopicPartition, List<Integer>> reassigned = rebalancePartitions(policy, brokers, partitions, reps);

        System.out.println("reassigned " + reassigned);

        //Three replicas per broker
        for (int brokerId = 100; brokerId <= 103; ++brokerId)
            assertEquals("Unexpected replica count for broker " + brokerId, 4,
                    filter(brokers(reassigned), brokerId).size());
        //One Leader per broker
        for (int brokerId = 100; brokerId <= 103; ++brokerId)
            assertEquals("Unexpected leader count for broker " + brokerId, 2, filter(leaders(reassigned).values(), brokerId).size());

        //Leader and Partitions for the same topic should not be on the same broker
        for (String topic : reps.keySet()) {
            assertNotEquals(reassigned.get(new TopicPartition(topic, 0)).get(0),
                    reassigned.get(new TopicPartition(topic, 1)).get(0));

            Set<Integer> intersection = intersection(reassigned.get(new TopicPartition(topic, 0)),
                    reassigned.get(new TopicPartition(topic, 1)));
            assertEquals("The same broker(s) appeared in both partitions of topic " + topic, Collections.emptySet(),
                    intersection);
        }

    }

    /* Because of the replication constraint, some size imbalance remains at the end */
    @Test
    public void shouldBalanceReplicaBySizeOverMultipleTopicsAndMultipleRacksWithSlightImbalanceAtTheEnd() {
        MovesOptimisedRebalancePolicy policy = new MovesOptimisedRebalancePolicy();

        //Given
        List<BrokerMetadata> brokers = asList(bk(100, "rack1"), bk(101, "rack1"), bk(102, "rack2"), bk(103, "rack2"));
        Map<TopicPartition, List<Integer>> partitions = createMap(
                entry(new TopicPartition("t1", 0), asList(100, 101, 102)),
                entry(new TopicPartition("t2", 0), asList(100, 101, 102)),
                entry(new TopicPartition("t2", 1), asList(100, 101, 102)),
                entry(new TopicPartition("t3", 0), asList(100, 101, 102)));
        Map<String, Integer> reps = createMap(entry("t1", 3), entry("t2", 3), entry("t3", 3));
        Map<TopicPartition, Long> partitionToSize = createMap(
                entry(new TopicPartition("t1", 0), 16L),
                entry(new TopicPartition("t2", 0), 15L),
                entry(new TopicPartition("t2", 1), 2L),
                entry(new TopicPartition("t3", 0), 28L));

        //When
        Map<TopicPartition, List<Integer>> reassigned = rebalancePartitions(policy, brokers, partitions, reps, partitionToSize);

        //Three replicas per broker
        for (int brokerId = 100; brokerId <= 103; ++brokerId)
            assertEquals(3, filter(brokers(reassigned), brokerId).size());

        assertEquals(45, partitionSizeSum(partitionToSize, brokerPartitions(reassigned, 100)));
        assertEquals(46, partitionSizeSum(partitionToSize, brokerPartitions(reassigned, 101)));
        assertEquals(59, partitionSizeSum(partitionToSize, brokerPartitions(reassigned, 102)));
        assertEquals(33, partitionSizeSum(partitionToSize, brokerPartitions(reassigned, 103)));

        //One Leader per broker
        for (int brokerId = 100; brokerId <= 103; ++brokerId)
            assertEquals(1, filter(leaders(reassigned).values(), brokerId).size());

        //Leader and Partitions for topic `t2` should not be on the same broker
        assertNotEquals(reassigned.get(new TopicPartition("t2", 0)).get(0),
                reassigned.get(new TopicPartition("t2", 1)).get(0));

        Set<Integer> intersection = intersection(reassigned.get(new TopicPartition("t2", 0)),
                reassigned.get(new TopicPartition("t2", 1)));
        // The number of replicas (6) is larger than the number of brokers (4), so the intersection will not be empty
        assertEquals(2, intersection.size());
    }

    @Test
    public void shouldBalanceReplicaBySizeOverMultipleTopicsAndMultipleRacks() {
        MovesOptimisedRebalancePolicy policy = new MovesOptimisedRebalancePolicy();

        //Given
        List<BrokerMetadata> brokers = asList(bk(100, "rack1"), bk(101, "rack1"), bk(102, "rack1"), bk(103, "rack2"),
                bk(104, "rack2"), bk(105, "rack2"));
        Map<TopicPartition, List<Integer>> partitions = createMap(
                entry(new TopicPartition("t1", 0), asList(100, 101)),
                entry(new TopicPartition("t2", 0), asList(103, 104)),
                entry(new TopicPartition("t2", 1), asList(104, 105)),
                entry(new TopicPartition("t3", 0), asList(105, 103)),
                entry(new TopicPartition("t3", 1), asList(100)),
                entry(new TopicPartition("t4", 0), asList(103)));
        Map<String, Integer> reps = createMap(entry("t1", 2), entry("t2", 2), entry("t3", 2), entry("t4", 2));
        Map<TopicPartition, Long> partitionToSize = createMap(
                entry(new TopicPartition("t1", 0), 16L),
                entry(new TopicPartition("t2", 0), 15L),
                entry(new TopicPartition("t2", 1), 10L),
                entry(new TopicPartition("t3", 0), 21L),
                entry(new TopicPartition("t3", 1), 12L),
                entry(new TopicPartition("t4", 0), 20L));

        //When
        Map<TopicPartition, List<Integer>> reassigned = rebalancePartitions(policy, brokers, partitions, reps, partitionToSize);

        //Then
        long expectedBrokerSize = Math.round(sum(partitionToSize.values()) * 2 / (double) brokers.size());
        for (int brokerId = 100; brokerId <= 105; ++brokerId) {
            assertEquals(3, filter(brokers(reassigned), brokerId).size(), 1.0);
            long brokerSize = partitionSizeSum(partitionToSize, brokerPartitions(reassigned, brokerId));
            assertEquals(expectedBrokerSize, brokerSize, 1.0);
        }

        //One Leader per broker
        for (int brokerId = 100; brokerId <= 105; ++brokerId)
            assertEquals(1, filter(leaders(reassigned).values(), brokerId).size());

        //Leader and Partitions for topics `t2` and `t3` should not be on the same broker
        for (String topic : asList("t2", "t3")) {
            assertNotEquals(reassigned.get(new TopicPartition(topic, 0)).get(0),
                    reassigned.get(new TopicPartition(topic, 1)).get(0));

            Set<Integer> intersection = intersection(reassigned.get(new TopicPartition(topic, 0)),
                    reassigned.get(new TopicPartition(topic, 1)));
            assertEquals("The same broker(s) appeared in both partitions of topic " + topic, Collections.emptySet(),
                    intersection);
        }
    }

    @Test
    public void shouldBalanceReplicaBySizeOverMultipleLargeTopicsAndMultipleRacks() {
        MovesOptimisedRebalancePolicy policy = new MovesOptimisedRebalancePolicy();

        //Given
        List<BrokerMetadata> brokers = asList(bk(100, "rack1"), bk(101, "rack1"),
                bk(102, "rack1"), bk(103, "rack2"), bk(104, "rack2"),
                bk(105, "rack2"));

        Map<TopicPartition, List<Integer>> partitions = createMap(
                entry(new TopicPartition("t1", 0), asList(100, 101)),
                entry(new TopicPartition("t2", 0), asList(103, 104)),
                entry(new TopicPartition("t2", 1), asList(104, 105)),
                entry(new TopicPartition("t3", 0), asList(105, 103)),
                entry(new TopicPartition("t3", 1), asList(100)),
                entry(new TopicPartition("t4", 0), asList(103)));
        Map<String, Integer> reps = createMap(entry("t1", 2), entry("t2", 2), entry("t3", 2), entry("t4", 2));
        Map<TopicPartition, Long> partitionToSize = createMap(
                entry(new TopicPartition("t1", 0), 6_235_823_523L),
                entry(new TopicPartition("t2", 0), 3_143_233_000L),
                entry(new TopicPartition("t2", 1), 2_000_000_000L),
                entry(new TopicPartition("t3", 0), 5_243_232_555L),
                entry(new TopicPartition("t3", 1), 3_122_111_000L),
                entry(new TopicPartition("t4", 0), 4_100_000_000L));

        //When
        Map<TopicPartition, List<Integer>> reassigned = rebalancePartitions(policy, brokers, partitions, reps, partitionToSize);

        //Then
        long expectedBrokerSize = Math.round(sum(partitionToSize.values()) * 2 / (double) brokers.size());
        for (int brokerId = 100; brokerId <= 105; ++brokerId) {
            long brokerSize = partitionSizeSum(partitionToSize, brokerPartitions(reassigned, brokerId));
            assertEquals(expectedBrokerSize, brokerSize, expectedBrokerSize * 0.2);
        }

        //One Leader per broker
        for (int brokerId = 100; brokerId <= 105; ++brokerId)
            assertEquals(1, filter(leaders(reassigned).values(), brokerId).size());

        //Leader and Partitions for topics `t2` and `t3` should not be on the same broker
        for (String topic : asList("t2", "t3")) {
            assertNotEquals(reassigned.get(new TopicPartition(topic, 0)).get(0),
                    reassigned.get(new TopicPartition(topic, 1)).get(0));

            Set<Integer> intersection = intersection(reassigned.get(new TopicPartition(topic, 0)),
                    reassigned.get(new TopicPartition(topic, 1)));
            assertEquals("The same broker(s) appeared in both partitions of topic " + topic, Collections.emptySet(),
                    intersection);
        }
    }

    @Test
    public void shouldBalanceReplicasOverMultipleTopicsWithMultipleReplicationFactors() {
        MovesOptimisedRebalancePolicy policy = new MovesOptimisedRebalancePolicy();

        //Given
        List<BrokerMetadata> brokers = asList(bk(100, "rack1"), bk(101, "rack1"), bk(102, "rack2"), bk(103, "rack2"));
        Map<TopicPartition, List<Integer>> partitions = createMap(
                entry(new TopicPartition("t1", 0), asList(100, 101, 102, 103)),
                entry(new TopicPartition("t2", 0), asList(100, 101, 102, 103)),
                entry(new TopicPartition("t3", 0), asList(100, 101)),
                entry(new TopicPartition("t3", 1), asList(100, 102)));
        Map<String, Integer> reps = createMap(entry("t1", 4), entry("t2", 4), entry("t3", 2));

        //When
        Map<TopicPartition, List<Integer>> reassigned = rebalancePartitions(policy, brokers, partitions, reps);

        //Then should be three replicas per broker
        for (int brokerId = 100; brokerId<= 103; ++brokerId)
            assertEquals(3, filter(brokers(reassigned), brokerId).size());
        // Then there should be one leader per broker
        for (int brokerId = 100; brokerId<= 103; ++brokerId)
            assertEquals(1, filter(leaders(reassigned).values(), brokerId).size());

        //Leader and Partitions for topic `t3` should not be on the same broker
        assertNotEquals(reassigned.get(new TopicPartition("t3", 0)).get(0),
                reassigned.get(new TopicPartition("t3", 1)).get(0));

        Set<Integer> intersection = intersection(reassigned.get(new TopicPartition("t3", 0)),
                reassigned.get(new TopicPartition("t3", 1)));
        assertEquals("The same broker(s) appeared in both partitions of topic t3", Collections.emptySet(),
                intersection);
    }

    @Test
    public void shouldAddBrokerToCluster() {
        MovesOptimisedRebalancePolicy policy = new MovesOptimisedRebalancePolicy();

        //Given
        List<BrokerMetadata> brokers = asList(bk(100, "rack1"), bk(101, "rack1"), bk(102, "rack2"), bk(103, "rack2"));
        Map<TopicPartition, List<Integer>> partitions = createMap(
                entry(tp(0), asList(100, 101, 102)),
                entry(tp(1), asList(100, 101, 102)),
                entry(tp(2), asList(100, 101, 102)),
                entry(tp(3), asList(100, 101, 102)));
        Map<String, Integer> reps = replicationFactors(3);

        //When
        Map<TopicPartition, List<Integer>> reassigned = rebalancePartitions(policy, brokers, partitions, reps);

        //Then should be thre replicas per broker
        for (int brokerId = 100; brokerId<= 103; ++brokerId)
            assertEquals(3, filter(brokers(reassigned), brokerId).size());
        //Then should be one leader per broker
        for (int brokerId = 100; brokerId<= 103; ++brokerId)
            assertEquals(1, filter(leaders(reassigned).values(), brokerId).size());
    }

    /**
     * Test Move & Leader Functions
     */

    @Test
    public void shouldMakeLeader() {
        MovesOptimisedRebalancePolicy policy = new MovesOptimisedRebalancePolicy();

        //Given
        List<BrokerMetadata> brokers = createBrokers(100, 102, "rack1");
        Map<TopicPartition, List<Integer>> partitions = createMap(entry(tp(0), asList(100, 101)));
        MutableRebalanceContext context = createRebalanceContext(brokers, partitions, replicationFactors(3));

        //When
        policy.makeLeader(context, tp(0), new Broker(101));

        //Then
        assertEquals(asList(new Broker(101), new Broker(100)), context.brokers(tp(0)));
    }

    @Test
    public void shouldMakeLeaderDoNothingIfMakingExistingLeaderTheLeader() {
        MovesOptimisedRebalancePolicy policy = new MovesOptimisedRebalancePolicy();

        //Given
        List<BrokerMetadata> brokers = createBrokers(100, 102, "rack1");
        Map<TopicPartition, List<Integer>> partitions = createMap(entry(tp(0), asList(100, 101)));
        MutableRebalanceContext context = createRebalanceContext(brokers, partitions, replicationFactors(3));

        //When
        policy.makeLeader(context, tp(0), new Broker(100));

        //Then
        assertEquals(asList(100, 101), partitions.get(tp(0)));
    }

    @Test
    public void shouldCatchSituationsWithBelowParBrokers() {
        MovesOptimisedRebalancePolicy policy = new MovesOptimisedRebalancePolicy();

        //Given 5 brokers with replicas (1, 2, 3, 3, 3)
        List<BrokerMetadata> brokers = asList(bk(100, "rack1"), bk(101, "rack1"), bk(102, "rack1"), bk(103, "rack1"),
                bk(104, "rack1"));
        Map<TopicPartition, List<Integer>> partitions = createMap(
                entry(tp(0), asList(100)),
                entry(tp(1), asList(101)),
                entry(tp(2), asList(102)),
                entry(tp(3), asList(102)),
                entry(tp(4), asList(103)),
                entry(tp(5), asList(103)),
                entry(tp(6), asList(103)),
                entry(tp(7), asList(104)),
                entry(tp(8), asList(104)),
                entry(tp(9), asList(104)));

        //When
        Map<TopicPartition, List<Integer>> result = rebalancePartitions(policy, brokers, partitions, replicationFactors(1));

        //Then replicas
        for (int brokerId = 100; brokerId <= 104; brokerId++)
            assertEquals(2, filter(brokers(result), brokerId).size());
        //Then should be two leaders per broker
        for (int brokerId = 100; brokerId <= 104; brokerId++)
            assertEquals(2, filter(leaders(result).values(), brokerId).size());
    }

    @Test
    public void shouldMovePartitionsAwayFromRemovedBroker() {
        MovesOptimisedRebalancePolicy policy = new MovesOptimisedRebalancePolicy();

        //Given
        List<BrokerMetadata> brokers = asList(bk(100, "rack1"), bk(101, "rack1"), bk(102, "rack1"));
        Map<TopicPartition, List<Integer>> partitions = createMap(
          entry(tp(0), asList(100, 102)),
          entry(tp(1), asList(101, 102)),
          entry(tp(2), asList(100, 101)),
          entry(tp(3), asList(102, 100)));
        Map<String, Integer> reps = replicationFactors(2);

        //When
        Map<TopicPartition, List<Integer>> reassigned = rebalancePartitions(policy, brokers, partitions, reps,
                asList(new Broker(102)));

        //Then should be three replicas per broker
        for (int brokerId = 100; brokerId <= 101; brokerId++)
          assertEquals(4, filter(brokers(reassigned), brokerId).size());
        //Then should be one leader per broker
        for (int brokerId = 100; brokerId <= 101; brokerId++)
          assertEquals(2, filter(leaders(reassigned).values(), brokerId).size());

        assertEquals(emptyList(), filter(brokers(reassigned), 102));
        assertEquals(emptyList(), filter(leaders(reassigned).values(), 102));
    }

    @Test
    public void shouldMovePartitionsAwayFromRemovedBrokers() {
        MovesOptimisedRebalancePolicy policy = new MovesOptimisedRebalancePolicy();

        List<BrokerMetadata> brokers = asList(bk(100, "rack1"), bk(101, "rack1"), bk(102, "rack1"), bk(103, "rack2"),
                bk(104, "rack2"), bk(105, "rack2"));
        Map<TopicPartition, List<Integer>> partitions = createMap(
                entry(tp(0), asList(100, 101)),
                entry(tp(1), asList(101, 102)),
                entry(tp(2), asList(102, 103)),
                entry(tp(3), asList(102, 103)),
                entry(tp(4), asList(103, 104)),
                entry(tp(5), asList(103, 104)),
                entry(tp(6), asList(103, 104)),
                entry(tp(7), asList(104)),
                entry(tp(8), asList(104)),
                entry(tp(9), asList(104)));

        //When
        List<Integer> removedBrokerIds = Arrays.asList(102, 104);
        List<Broker> removedBrokers = new ArrayList<>();
        for (Integer brokerId : removedBrokerIds)
            removedBrokers.add(new Broker(brokerId));

        Map<TopicPartition, List<Integer>> result = rebalancePartitions(policy, brokers, partitions,
                replicationFactors(2), removedBrokers);

        //Then replicas
        for (int brokerId = 100; brokerId <= 105; brokerId++) {
            if (!removedBrokerIds.contains(brokerId))
                assertEquals("Unexpected replica count for broker " + brokerId, 5,
                        filter(brokers(result), brokerId).size());
        }

        //Then leaders
        int totalLeaders = 0;
        for (int brokerId = 100; brokerId <= 105; brokerId++) {
            if (!removedBrokerIds.contains(brokerId)) {
                int expected = filter(leaders(result).values(), brokerId).size();
                totalLeaders += expected;
                assertTrue("Unexpected leader count for broker " + brokerId, expected == 2 || expected == 3);
            }
        }
        assertEquals(10, totalLeaders);

        //Then removed brokers
        for (int brokerId : removedBrokerIds) {
            assertEquals(emptyList(), filter(brokers(result), brokerId));
            assertEquals(emptyList(), filter(leaders(result).values(), brokerId));
        }
    }

    @Test(expected = ValidationException.class)
    public void shouldObeyRackConstraintWhenMovingPartitionsAwayFromBroker() {
        MovesOptimisedRebalancePolicy policy = new MovesOptimisedRebalancePolicy();

        //Given
        List<BrokerMetadata> brokers = asList(bk(100, "rack1"), bk(101, "rack1"), bk(102, "rack2"));
        Map<TopicPartition, List<Integer>> partitions = createMap(
                entry(tp(0), asList(100, 102)),
                entry(tp(1), asList(101, 102)),
                entry(tp(2), asList(102, 101)));
        Map<String, Integer> reps = replicationFactors(2);

        //When
        rebalancePartitions(policy, brokers, partitions, reps, asList(new Broker(102)));
    }

    @Test
    public void shouldObeyPartitionConstraintWhenMovingPartitionsAwayFromBroker() {
        MovesOptimisedRebalancePolicy policy = new MovesOptimisedRebalancePolicy();

        //Given
        List<BrokerMetadata> brokers = asList(bk(100, "rack1"), bk(101, "rack1"));
        Map<TopicPartition, List<Integer>> partitions = createMap(
                entry(tp(0), asList(100, 101)),
                entry(tp(1), asList(101, 100)));
        Map<String, Integer> reps = replicationFactors(2);

        //When
        Map<TopicPartition, List<Integer>> reassigned = rebalancePartitions(policy, brokers, partitions, reps,
                asList(new Broker(102)));

        //Then, no change after rebalance as it would not obey partition constraint
        assertEquals(partitions, reassigned);
    }

    @Test(expected = ValidationException.class)
    public void shouldNotRemoveBrokersWhenItViolatesPartitionConstraints() {
        MovesOptimisedRebalancePolicy policy = new MovesOptimisedRebalancePolicy();
        List<BrokerMetadata> brokers = asList(bk(100, "rack1"), bk(101, "rack1"), bk(102, "rack1"));
        Map<TopicPartition, List<Integer>> partitions = createMap(
                entry(tp(0), asList(100, 101, 102)),
                entry(tp(1), asList(100, 101, 102)),
                entry(tp(2), asList(100, 101, 102))
        );
        Map<String, Integer> reps = replicationFactors(3);

        rebalancePartitions(policy, brokers, partitions, reps, asList(new Broker(102)));
    }

    @Test(expected = ValidationException.class)
    public void shouldSuggestBrokersToRemoveForDeadBrokers() {
        MovesOptimisedRebalancePolicy policy = new MovesOptimisedRebalancePolicy();
        List<BrokerMetadata> brokers = asList(bk(100, "rack1"), bk(101, "rack1"));
        Map<TopicPartition, List<Integer>> partitions = createMap(
                entry(tp(0), asList(100, 101)),
                entry(tp(1), asList(101, 102)),
                entry(tp(2), asList(102, 100))
        );
        Map<String, Integer> reps = replicationFactors(2);
        rebalancePartitions(policy, brokers, partitions, reps);
   }

    @Test
    public void shouldMovePartitionsAwayFromDeadBrokers() {
        MovesOptimisedRebalancePolicy policy = new MovesOptimisedRebalancePolicy();
        List<BrokerMetadata> brokers = asList(bk(100, "rack1"), bk(101, "rack1"));
        Map<TopicPartition, List<Integer>> partitions = createMap(
                entry(tp(0), asList(100, 101)),
                entry(tp(1), asList(101, 102)),
                entry(tp(2), asList(102, 100))
        );
        Map<String, Integer> reps = replicationFactors(2);
        Map<TopicPartition, List<Integer>> reassigned = rebalancePartitions(policy, brokers, partitions, reps,
                asList(new Broker(102)));

        for (TopicPartition topicPartition : reassigned.keySet()) {
            assertTrue(!reassigned.get(topicPartition).contains(new Broker(102)));
        }

    }
}
