// (Copyright) [2016 - 2016] Confluent, Inc.
package io.confluent.kafka.databalancing;

import io.confluent.kafka.databalancing.topology.Broker;
import io.confluent.kafka.databalancing.view.BrokerCountFairView;
import io.confluent.kafka.databalancing.view.BrokerFairView;
import io.confluent.kafka.databalancing.view.BrokerSizeFairView;
import io.confluent.kafka.databalancing.view.RackSizeFairView;
import io.confluent.kafka.databalancing.view.RackFairView;
import kafka.admin.BrokerMetadata;
import org.apache.kafka.common.TopicPartition;
import org.junit.Test;

import java.util.Collections;
import java.util.List;

import static io.confluent.kafka.databalancing.TestUtils.concat;
import static io.confluent.kafka.databalancing.TestUtils.createBrokers;
import static io.confluent.kafka.databalancing.TestUtils.createMap;
import static io.confluent.kafka.databalancing.TestUtils.createRebalanceContext;
import static io.confluent.kafka.databalancing.TestUtils.entry;
import static java.util.Arrays.asList;
import static org.junit.Assert.assertEquals;

public class SizeFairnessTest {

    private static final String topic = "topic";

    private static TopicPartition tp(int partition) {
        return new TopicPartition(topic, partition);
    }

    @Test
    public void shouldCalculateReplicaFairnessForRacks() {
        List<BrokerMetadata> brokers = concat(createBrokers(100, 102, "rack1"), createBrokers(102, 104, "rack2"));

        //1 replica, 2 racks. rack1 is Below Par, rack2 is onPar.
        RackFairView view = new RackSizeFairView(createRebalanceContext(brokers,
                createMap(entry(tp(0), asList(101))), Collections.singletonMap("topic", 1),
                createMap(entry(tp(0), 49L))), null);
        assertEquals(24.5, view.replicaFairness().rackFairValue(), 0.001);
        assertEquals(asList("rack1"), view.replicaFairness().aboveParRacks());
        assertEquals(asList("rack2"), view.replicaFairness().belowParRacks());

        //2 replicas, 2 racks, both replicas on one rack
        view = new RackSizeFairView(createRebalanceContext(brokers,
                createMap(entry(tp(0), asList(103, 102))), Collections.singletonMap(topic, 1),
                createMap(entry(tp(0), 260L))), null);
        assertEquals(260, view.replicaFairness().rackFairValue(), 0.001);
        assertEquals(asList("rack2"), view.replicaFairness().aboveParRacks());
        assertEquals(asList("rack1"), view.replicaFairness().belowParRacks());

        //2 replicas with the same size, 2 racks, replicas on different racks
        view = new RackSizeFairView(createRebalanceContext(brokers,
                createMap(entry(tp(0), asList(100, 101)), entry(tp(1), asList(102, 103))),
                Collections.singletonMap(topic, 1), createMap(entry(tp(0), 315L), entry(tp(1), 315L))), null);
        assertEquals(630, view.replicaFairness().rackFairValue(), 0.001);
        assertEquals(asList(), view.replicaFairness().aboveParRacks());
        assertEquals(asList(), view.replicaFairness().belowParRacks());

        //3 replicas, 2 racks, replicas spread 2/1
        view = new RackSizeFairView(createRebalanceContext(brokers,
                createMap(entry(tp(0), asList(100, 101, 103))), Collections.singletonMap(topic, 1),
                createMap(entry(tp(0), 949L))), null);
        assertEquals(1423.5, view.replicaFairness().rackFairValue(), 0.001);
        assertEquals(asList("rack1"), view.replicaFairness().aboveParRacks());
        assertEquals(asList("rack2"), view.replicaFairness().belowParRacks());

        //3 replicas, 2 racks, replicas spread 3/0
        view = new RackSizeFairView(createRebalanceContext(brokers,
                createMap(entry(tp(0), asList(100, 101)), entry(tp(1), asList(100))),
                Collections.singletonMap(topic, 1), createMap(entry(tp(0), 100L), entry(tp(1), 150L))), null);
        assertEquals(175, view.replicaFairness().rackFairValue(), 0.001);
        assertEquals(asList("rack1"), view.replicaFairness().aboveParRacks());
        assertEquals(asList("rack2"), view.replicaFairness().belowParRacks());

        //4 replicas, 2 racks, spread 2/2
        view = new RackSizeFairView(createRebalanceContext(brokers,
                createMap(entry(tp(0), asList(103, 102, 101, 100))), Collections.singletonMap(topic, 1),
                createMap(entry(tp(0), 100L))), null);
        assertEquals(200, view.replicaFairness().rackFairValue(), 0.001);
        assertEquals(asList(), view.replicaFairness().aboveParRacks());
        assertEquals(asList(), view.replicaFairness().belowParRacks());

        //4 replicas, 2 racks, spread 3/1 with the single one being larger than the other 3 combined
        view = new RackSizeFairView(createRebalanceContext(brokers,
                createMap(entry(tp(0), asList(100)),
                        entry(tp(1), asList(100)),
                        entry(tp(2), asList(100)),
                        entry(tp(3), asList(103))),
                Collections.singletonMap(topic, 1),
                createMap(entry(tp(0), 10L), entry(tp(1), 20L), entry(tp(2), 15L), entry(tp(3), 100L))), null);
        assertEquals(72.5, view.replicaFairness().rackFairValue(), 0.001);
        assertEquals(asList("rack2"), view.replicaFairness().aboveParRacks());
        assertEquals(asList("rack1"), view.replicaFairness().belowParRacks());

        //4 replicas, 2 racks, spread 4/0
        view = new RackSizeFairView(createRebalanceContext(brokers,
                createMap(entry(tp(0), asList(100)),
                        entry(tp(1), asList(100)),
                        entry(tp(2), asList(100)),
                        entry(tp(3), asList(100))),
                Collections.singletonMap(topic, 1),
                createMap(entry(tp(0), 10_000_000_000L),
                        entry(tp(1), 20_000_000_000L),
                        entry(tp(2), 20_000_000_000L),
                        entry(tp(3), 13_000_000_000L))), null);
        assertEquals(31_500_000_000L, view.replicaFairness().rackFairValue(), 0.001);
        assertEquals(asList("rack1"), view.replicaFairness().aboveParRacks());
        assertEquals(asList("rack2"), view.replicaFairness().belowParRacks());
    }

    @Test
    public void shouldCalculateReplicaFairnessForBrokers() {
        List<BrokerMetadata> brokers = createBrokers(100, 102, "rack1");
        Broker broker0 = new Broker(brokers.get(0).id());
        Broker broker1 = new Broker(brokers.get(1).id());

        //1 replica, 2 brokers. bk1 is Below Par, bk2 is onPar.
        BrokerFairView view = new BrokerSizeFairView(createRebalanceContext(brokers,
                createMap(entry(tp(0), asList(101))), Collections.singletonMap(topic, 1),
                createMap(entry(tp(0), 49L))), "rack1", null);
        assertEquals(24.5, view.replicaFairness().brokerFairValue(), 0.001);
        assertEquals(asList(broker1), view.replicaFairness().aboveParBrokers());
        assertEquals(asList(broker0), view.replicaFairness().belowParBrokers());

        //2 replicas, 2 brokers, both replicas on one broker
        view = new BrokerSizeFairView(createRebalanceContext(brokers,
                createMap(entry(tp(0), asList(100)), entry(tp(1), asList(100))),
                Collections.singletonMap(topic, 1),
                createMap(entry(tp(0), 4L), entry(tp(1), 2L))), "rack1", null);
        assertEquals(3.0, view.replicaFairness().brokerFairValue(), 0.001);
        assertEquals(asList(broker0), view.replicaFairness().aboveParBrokers());
        assertEquals(asList(broker1), view.replicaFairness().belowParBrokers());

        //2 replicas, 2 brokers, replicas on each
        view = new BrokerSizeFairView(createRebalanceContext(brokers,
                createMap(entry(tp(0), asList(100)), entry(tp(1), asList(101))),
                Collections.singletonMap(topic, 1),
                createMap(entry(tp(0), 11L), entry(tp(1), 11L))), "rack1", null);
        assertEquals(11, view.replicaFairness().brokerFairValue(), 0.001);
        assertEquals(asList(), view.replicaFairness().aboveParBrokers());
        assertEquals(asList(), view.replicaFairness().belowParBrokers());

        //3 replicas, 2 brokers, replicas spread 2/1
        view = new BrokerSizeFairView(createRebalanceContext(brokers,
                createMap(entry(tp(0), asList(100, 101)), entry(tp(1), asList(100))),
                Collections.singletonMap(topic, 1),
                createMap(entry(tp(0), 3L), entry(tp(1), 4L))), "rack1", null);
        assertEquals(5, view.replicaFairness().brokerFairValue(), 0.001);
        assertEquals(asList(broker0), view.replicaFairness().aboveParBrokers());
        assertEquals(asList(broker1), view.replicaFairness().belowParBrokers());

        //3 replicas, 2 brokers, replicas spread 3/0
        view = new BrokerSizeFairView(createRebalanceContext(brokers,
                createMap(entry(tp(0), asList(100)), entry(tp(1), asList(100)), entry(tp(2), asList(100))),
                Collections.singletonMap(topic, 1),
                createMap(entry(tp(0), 49L), entry(tp(1), 22L), entry(tp(2), 11L))), "rack1", null);
        assertEquals(41, view.replicaFairness().brokerFairValue(), 0.001);
        assertEquals(asList(broker0), view.replicaFairness().aboveParBrokers());
        assertEquals(asList(broker1), view.replicaFairness().belowParBrokers());

        //4 replicas, 2 brokers, spread 2/2
        view = new BrokerSizeFairView(createRebalanceContext(brokers,
                createMap(entry(tp(0), asList(100, 101)), entry(tp(1), asList(100, 101))),
                Collections.singletonMap(topic, 1),
                createMap(entry(tp(0), 3L), entry(tp(1), 22L))), "rack1", null);
        assertEquals(25, view.replicaFairness().brokerFairValue(), 0.001);
        assertEquals(asList(), view.replicaFairness().aboveParBrokers());
        assertEquals(asList(), view.replicaFairness().belowParBrokers());

        //4 replicas, 2 brokers, spread 3/1
        view = new BrokerSizeFairView(createRebalanceContext(brokers,
                createMap(entry(tp(0), asList(100)),
                        entry(tp(1), asList(100)),
                        entry(tp(2), asList(100)),
                        entry(tp(3), asList(101))),
                Collections.singletonMap(topic, 1),
                createMap(entry(tp(0), 12L), entry(tp(1), 8L), entry(tp(2), 2L), entry(tp(3), 9L))),
                "rack1", null);
        assertEquals(15.5, view.replicaFairness().brokerFairValue(), 0.001);
        assertEquals(asList(broker0), view.replicaFairness().aboveParBrokers());
        assertEquals(asList(broker1), view.replicaFairness().belowParBrokers());

        //4 replicas, 2 brokers, spread 4/0
        view = new BrokerSizeFairView(createRebalanceContext(brokers,
                createMap(entry(tp(0), asList(100)),
                        entry(tp(1), asList(100)),
                        entry(tp(2), asList(100)),
                        entry(tp(3), asList(100))),
                Collections.singletonMap(topic, 1),
                createMap(entry(tp(0), 1_000_000_000L), entry(tp(1), 2_000_000_000L),
                        entry(tp(2), 6_000_000_000L), entry(tp(3), 1_000_000_000L))),
                "rack1", null);
        assertEquals(5_000_000_000L, view.replicaFairness().brokerFairValue(), 0.001);
        assertEquals(asList(broker0), view.replicaFairness().aboveParBrokers());
        assertEquals(asList(broker1), view.replicaFairness().belowParBrokers());
    }

    @Test
    public void shouldCalculateLeaderFairnessForRacks() {
        List<BrokerMetadata> brokers = concat(createBrokers(100, 102, "rack1"), createBrokers(102, 104, "rack2"));

        //1 leader, 2 racks. rack1 is Below Par, rack2 is onPar.
        RackFairView view = new RackSizeFairView(createRebalanceContext(brokers,
                createMap(entry(tp(0), asList(101))), Collections.singletonMap(topic, 1),
                createMap(entry(tp(0), 24L))), null);
        assertEquals(12, view.leaderFairness().rackFairValue(), 0.001);
        assertEquals(asList("rack1"), view.leaderFairness().aboveParRacks());
        assertEquals(asList("rack2"), view.leaderFairness().belowParRacks());

        //2 leaders, 2 racks, both leaders on one rack
        view = new RackSizeFairView(createRebalanceContext(brokers,
                createMap(entry(tp(0), asList(102)), entry(tp(1), asList(102))),
                Collections.singletonMap(topic, 1),
                createMap(entry(tp(0), 55L), entry(tp(1), 98L))), null);
        assertEquals(76.5, view.leaderFairness().rackFairValue(), 0.001);
        assertEquals(asList("rack2"), view.leaderFairness().aboveParRacks());
        assertEquals(asList("rack1"), view.leaderFairness().belowParRacks());

        //2 leaders, 2 racks, leaders on different racks
        view = new RackSizeFairView(createRebalanceContext(brokers,
                createMap(entry(tp(0), asList(100)), entry(tp(1), asList(102))),
                Collections.singletonMap(topic, 1),
                createMap(entry(tp(0), 23L), entry(tp(1), 23L))), null);
        assertEquals(23.0, view.leaderFairness().rackFairValue(), 0.001);
        assertEquals(asList(), view.leaderFairness().aboveParRacks());
        assertEquals(asList(), view.leaderFairness().belowParRacks());

        //3 leaders, 2 racks, replicas spread 2/1
        view = new RackSizeFairView(createRebalanceContext(brokers,
                createMap(entry(tp(0), asList(100)), entry(tp(1), asList(100)), entry(tp(2), asList(102))),
                Collections.singletonMap(topic, 1),
                createMap(entry(tp(0), 5L), entry(tp(1), 7L), entry(tp(2), 6L))), null);
        assertEquals(9.0, view.leaderFairness().rackFairValue(), 0.001);
        assertEquals(asList("rack1"), view.leaderFairness().aboveParRacks());
        assertEquals(asList("rack2"), view.leaderFairness().belowParRacks());

        //3 replicas, 2 racks, replicas spread 3/0
        view = new RackSizeFairView(createRebalanceContext(brokers,
                createMap(entry(tp(0), asList(100)), entry(tp(1), asList(100)), entry(tp(2), asList(100))),
                Collections.singletonMap(topic, 1),
                createMap(entry(tp(0), 9L), entry(tp(1), 7L), entry(tp(2), 6L))), null);
        assertEquals(11.0, view.leaderFairness().rackFairValue(), 0.001);
        assertEquals(asList("rack1"), view.leaderFairness().aboveParRacks());
        assertEquals(asList("rack2"), view.leaderFairness().belowParRacks());

        //4 leader, 2 racks, spread 2/2
        view = new RackSizeFairView(createRebalanceContext(brokers,
                createMap(entry(tp(0), asList(100)),
                        entry(tp(1), asList(100)),
                        entry(tp(2), asList(103)),
                        entry(tp(3), asList(103))),
                Collections.singletonMap(topic, 1),
                createMap(entry(tp(0), 49L), entry(tp(1), 47L), entry(tp(2), 48L), entry(tp(3), 48L))), null);
        assertEquals(96.0, view.leaderFairness().rackFairValue(), 0.001);
        assertEquals(asList(), view.leaderFairness().aboveParRacks());
        assertEquals(asList(), view.leaderFairness().belowParRacks());

        //4 leader, 2 racks, spread 3/1
        view = new RackSizeFairView(createRebalanceContext(brokers,
                createMap(entry(tp(0), asList(100)),
                        entry(tp(1), asList(100)),
                        entry(tp(2), asList(100)),
                        entry(tp(3), asList(103))),
                Collections.singletonMap(topic, 1),
                createMap(entry(tp(0), 3L), entry(tp(1), 4L), entry(tp(2), 5L), entry(tp(3), 22L))), null);
        assertEquals(17.0, view.leaderFairness().rackFairValue(), 0.001);
        assertEquals(asList("rack2"), view.leaderFairness().aboveParRacks());
        assertEquals(asList("rack1"), view.leaderFairness().belowParRacks());

        //4 leader, 2 racks, spread 4/0
        view = new RackSizeFairView(createRebalanceContext(brokers,
                createMap(entry(tp(0), asList(100)),
                        entry(tp(1), asList(100)),
                        entry(tp(2), asList(100)),
                        entry(tp(3), asList(100))),
                Collections.singletonMap(topic, 1),
                createMap(entry(tp(0), 5_000_000_000L), entry(tp(1), 7_000_000_000L),
                        entry(tp(2), 2_000_000_000L), entry(tp(3), 10_000_000_000L))), null);
        assertEquals(12_000_000_000L, view.leaderFairness().rackFairValue(), 0.001);
        assertEquals(asList("rack1"), view.leaderFairness().aboveParRacks());
        assertEquals(asList("rack2"), view.leaderFairness().belowParRacks());
    }

    @Test
    public void shouldCalculateLeaderFairnessForBrokers() {
        List<BrokerMetadata> brokers = createBrokers(100, 102, "rack1");
        Broker broker0 = new Broker(brokers.get(0).id());
        Broker broker1 = new Broker(brokers.get(1).id());

        //1 leader, 2 brokers. bk1 is Below Par, bk2 is onPar.
        BrokerFairView view = new BrokerSizeFairView(createRebalanceContext(brokers,
                createMap(entry(tp(0), asList(101))), Collections.singletonMap(topic, 1),
                createMap(entry(tp(0), 23L))), "rack1", null);
        assertEquals(11.5, view.leaderFairness().brokerFairValue(), 0.001);
        assertEquals(asList(broker1), view.leaderFairness().aboveParBrokers());
        assertEquals(asList(broker0), view.leaderFairness().belowParBrokers());

        //2 leaders, 2 brokers, both replicas on one broker
        view = new BrokerSizeFairView(createRebalanceContext(brokers,
                createMap(entry(tp(0), asList(100)), entry(tp(1), asList(100))),
                Collections.singletonMap(topic, 1),
                createMap(entry(tp(0), 7L), entry(tp(1), 9L))), "rack1", null);
        assertEquals(8, view.leaderFairness().brokerFairValue(), 0.001);
        assertEquals(asList(broker0), view.leaderFairness().aboveParBrokers());
        assertEquals(asList(broker1), view.leaderFairness().belowParBrokers());

        //2 leaders, 2 brokers, replicas on each
        view = new BrokerSizeFairView(createRebalanceContext(brokers,
                createMap(entry(tp(0), asList(100)), entry(tp(1), asList(101))),
                Collections.singletonMap(topic, 1),
                createMap(entry(tp(0), 11L), entry(tp(1), 11L))), "rack1", null);
        assertEquals(11, view.leaderFairness().brokerFairValue(), 0.001);
        assertEquals(asList(), view.leaderFairness().aboveParBrokers());
        assertEquals(asList(), view.leaderFairness().belowParBrokers());

        //3 leaders, 2 brokers, replicas spread 2/1
        view = new BrokerSizeFairView(createRebalanceContext(brokers,
                createMap(entry(tp(0), asList(100)), entry(tp(1), asList(100)), entry(tp(2), asList(101))),
                Collections.singletonMap(topic, 1),
                createMap(entry(tp(0), 11L), entry(tp(1), 12L), entry(tp(2), 11L))), "rack1", null);
        assertEquals(17, view.leaderFairness().brokerFairValue(), 0.001);
        assertEquals(asList(broker0), view.leaderFairness().aboveParBrokers());
        assertEquals(asList(broker1), view.leaderFairness().belowParBrokers());

        //3 leaders, 2 brokers, replicas spread 3/0
        view = new BrokerSizeFairView(createRebalanceContext(brokers,
                createMap(entry(tp(0), asList(100)), entry(tp(1), asList(100)), entry(tp(2), asList(100))),
                Collections.singletonMap(topic, 1),
                createMap(entry(tp(0), 11L), entry(tp(1), 9L), entry(tp(2), 18L))), "rack1", null);
        assertEquals(19, view.leaderFairness().brokerFairValue(), 0.001);
        assertEquals(asList(broker0), view.leaderFairness().aboveParBrokers());
        assertEquals(asList(broker1), view.leaderFairness().belowParBrokers());

        //4 leaders, 2 brokers, spread 2/2
        view = new BrokerSizeFairView(createRebalanceContext(brokers,
                createMap(entry(tp(0), asList(100)),
                        entry(tp(1), asList(100)),
                        entry(tp(2), asList(101)),
                        entry(tp(3), asList(101))),
                Collections.singletonMap(topic, 1),
                createMap(entry(tp(0), 5L), entry(tp(1), 9L), entry(tp(2), 11L), entry(tp(3), 3L))), "rack1", null);
        assertEquals(14, view.leaderFairness().brokerFairValue(), 0.001);
        assertEquals(asList(), view.leaderFairness().aboveParBrokers());
        assertEquals(asList(), view.leaderFairness().belowParBrokers());

        //4 leaders, 2 brokers, spread 3/1
        view = new BrokerSizeFairView(createRebalanceContext(brokers,
                createMap(entry(tp(0), asList(100)),
                        entry(tp(1), asList(100)),
                        entry(tp(2), asList(100)),
                        entry(tp(3), asList(101))),
                Collections.singletonMap(topic, 1),
                createMap(entry(tp(0), 20L), entry(tp(1), 9L), entry(tp(2), 18L), entry(tp(3), 10L))), "rack1", null);
        assertEquals(28.5, view.leaderFairness().brokerFairValue(), 0.001);
        assertEquals(asList(broker0), view.leaderFairness().aboveParBrokers());
        assertEquals(asList(broker1), view.leaderFairness().belowParBrokers());

        //4 leaders, 2 brokers, spread 4/0
        view = new BrokerSizeFairView(createRebalanceContext(brokers,
                createMap(entry(tp(0), asList(100)),
                        entry(tp(1), asList(100)),
                        entry(tp(2), asList(100)),
                        entry(tp(3), asList(100))),
                Collections.singletonMap(topic, 1),
                createMap(entry(tp(0), 2_000_000_000L), entry(tp(1), 2_000_000_000L),
                        entry(tp(2), 3_000_000_000L), entry(tp(3), 3_000_000_000L))), "rack1", null);
        assertEquals(5_000_000_000L, view.leaderFairness().brokerFairValue(), 0.001);
        assertEquals(asList(broker0), view.leaderFairness().aboveParBrokers());
        assertEquals(asList(broker1), view.leaderFairness().belowParBrokers());
    }
}
