// (Copyright) [2016 - 2016] Confluent, Inc.
package io.confluent.kafka.databalancing.integration;

import io.confluent.kafka.databalancing.integration.utils.EmbeddedKafkaServer;
import io.confluent.kafka.databalancing.metric.MetricsCollector;
import io.confluent.kafka.databalancing.metric.Metrics;
import kafka.server.KafkaConfig$;
import kafka.utils.ZkUtils;
import kafka.zk.EmbeddedZookeeper;
import org.apache.kafka.clients.consumer.ConsumerConfig;
import org.apache.kafka.clients.producer.KafkaProducer;
import org.apache.kafka.clients.producer.ProducerConfig;
import org.apache.kafka.clients.producer.ProducerRecord;
import org.apache.kafka.common.TopicPartition;
import org.apache.kafka.common.security.JaasUtils;
import org.apache.kafka.common.serialization.StringSerializer;
import org.apache.kafka.test.TestCondition;
import org.apache.kafka.test.TestUtils;
import org.junit.After;
import org.junit.Before;
import org.junit.Test;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.IOException;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.Random;

import static java.util.Arrays.asList;
import static org.junit.Assert.assertTrue;

/**
 * This class does an end-to-end integration test for @MetricsReader.
 * In particular, it creates a Kafka cluster with the @ConfluentMetricsReporter, writes messages to this cluster, and
 * then tries to read the metrics through the @MetricsReader class.
 */
public class MetricsCollectorIntegrationTest {
    private static final Logger LOG = LoggerFactory.getLogger(MetricsCollectorIntegrationTest.class);
    private EmbeddedZookeeper zookeeper;
    private int brokerPort;
    private EmbeddedKafkaServer kafkaEmbedded;
    private int metricsReporterPublishMs;
    private ZkUtils zkUtils;

    @Before
    public void setUp() throws IOException {
        zookeeper = new EmbeddedZookeeper();
        metricsReporterPublishMs = 1500;
        // We pick a random port each time to avoid `BindException` in case a port we are trying to use is already
        // taken on the system. In general, ports 60000+ are not used very much. We can't use `0` for the port as we
        // need its value for the metrics reporter configuration before we start the broker
        Random portGenerator = new Random();
        // pick a port in the range 60000 - 64000;
        brokerPort = portGenerator.nextInt(4000) + 60000;

        Properties brokerConfig = new Properties();
        brokerConfig.put(KafkaConfig$.MODULE$.BrokerIdProp(), "0");
        brokerConfig.put(KafkaConfig$.MODULE$.ZkConnectProp(), zKConnectString());
        brokerConfig.put(KafkaConfig$.MODULE$.PortProp(), brokerPort);
        brokerConfig.put(KafkaConfig$.MODULE$.MetricReporterClassesProp(),
                "io.confluent.metrics.reporter.ConfluentMetricsReporter");
        brokerConfig.put("confluent.metrics.reporter.bootstrap.servers", "localhost:" + brokerPort);
        brokerConfig.put("confluent.metrics.reporter.zookeeper.connect", zKConnectString());
        brokerConfig.put("confluent.metrics.reporter.topic.replicas", "1");
        brokerConfig.put("confluent.metrics.reporter.publish.ms", Integer.toString(metricsReporterPublishMs));
        kafkaEmbedded = new EmbeddedKafkaServer(brokerConfig);
        zkUtils = ZkUtils.apply("localhost:" + zookeeper.port(), 30000, 7000, JaasUtils.isZkSecurityEnabled());
    }

    @After
    public void tearDown() {
        kafkaEmbedded.stop();
        zookeeper.shutdown();
        zkUtils.close();
    }

    private String zKConnectString() {
        return "localhost:" + zookeeper.port();
    }

    @Test
    public void basicTest() throws Exception {

        final String firstTopic = "myfirsttopic";
        final String secondTopic = "mysecondtopic";

        kafkaEmbedded.createTopic(firstTopic, 1, 1);
        kafkaEmbedded.createTopic(secondTopic, 1, 1);

        KafkaProducer<String, String> producer = createProducer();

        for (int i = 0; i < 1000; ++i) {
            producer.send(new ProducerRecord<>(firstTopic, Integer.toString(i), Integer.toString(i)));
            producer.send(new ProducerRecord<>(secondTopic, Integer.toString(i), Integer.toString(i)));
        }
        producer.flush();

        // The topics whose metrics we are interested in receiving. These have been seeded externally.
        TopicPartition tp0 = new TopicPartition(firstTopic, 0);
        TopicPartition tp1 = new TopicPartition(secondTopic, 0);
        List<TopicPartition> topicPartitions = asList(tp0, tp1);

        Properties metricConsumerProps = new Properties();
        metricConsumerProps.put(ConsumerConfig.BOOTSTRAP_SERVERS_CONFIG, "localhost:" + brokerPort);
        metricConsumerProps.put(ConsumerConfig.METADATA_MAX_AGE_CONFIG, "100");
        MetricPropagatedCondition metricsPropagatedCondition = new MetricPropagatedCondition(
                zkUtils.getClusterId().get(), metricConsumerProps, topicPartitions, 0, 0);
        TestUtils.waitForCondition(metricsPropagatedCondition, 60000,
                "Metrics were not published within the expected timeframe.");

        Map<TopicPartition, Long> topicPartitionSizeMap = metricsPropagatedCondition.metrics().partitionToSize();

        assertTrue(topicPartitionSizeMap.containsKey(tp0));
        long firstPartitionSize = topicPartitionSizeMap.get(tp0);
        assertTrue(firstPartitionSize > 0);

        assertTrue(topicPartitionSizeMap.containsKey(tp1));
        long secondPartitionSize = topicPartitionSizeMap.get(tp1);
        assertTrue(secondPartitionSize > 0);

        LOG.info(firstTopic + " -> " + firstPartitionSize);
        LOG.info(secondTopic + " -> " + secondPartitionSize);

        for (int i = 0; i < 1000; ++i) {
            producer.send(new ProducerRecord<>(firstTopic, Integer.toString(i), Integer.toString(i)));
            producer.send(new ProducerRecord<>(secondTopic, Integer.toString(i), Integer.toString(i)));
        }
        producer.close();

        // Sleep for the seek duration + publish duration to make sure that we get the new metrics of metrics.
        long waitDuration = MetricsCollector.CONSUMER_REPLAY_PERIOD_MS + metricsReporterPublishMs;

        TestUtils.waitForCondition(new MetricPropagatedCondition(zkUtils.getClusterId().get(),
                        metricConsumerProps, topicPartitions,  firstPartitionSize, secondPartitionSize), waitDuration,
                "New metrics were not propagated within the expected timeframe.");

   }

   private class MetricPropagatedCondition implements TestCondition {
       private final Properties metricConsumerProps;
       private final List<TopicPartition> topicPartitions;
       private final long firstPartitionSize;
       private final long secondPartitionSize;
       private final String clusterId;
       private Metrics metrics;

       MetricPropagatedCondition(String clusterId, Properties props, List<TopicPartition> topicPartitionList,
                                 long firstPartSize, long secondPartSize) {
           this.clusterId = clusterId;
           metricConsumerProps = props;
           topicPartitions = topicPartitionList;
           firstPartitionSize = firstPartSize;
           secondPartitionSize = secondPartSize;
       }
       @Override
       public boolean conditionMet() {
           metrics = MetricsCollector.collectMetrics(null, metricConsumerProps, topicPartitions, clusterId);
           long newFirstPartitionSize = metrics.partitionToSize().get(topicPartitions.get(0));
           long newSecondPartitionSize = metrics.partitionToSize().get(topicPartitions.get(1));
           return newFirstPartitionSize > firstPartitionSize && newSecondPartitionSize > secondPartitionSize;
       }

       public Metrics metrics() {
           return metrics;
       }
   }

    private KafkaProducer<String, String> createProducer() {
        Properties props = new Properties();
        props.put(ProducerConfig.BOOTSTRAP_SERVERS_CONFIG, "localhost:" + brokerPort);
        props.put(ProducerConfig.ACKS_CONFIG, "all");
        return new KafkaProducer<>(props, new StringSerializer(), new StringSerializer());
    }
}
