// (Copyright) [2016 - 2016] Confluent, Inc.
package io.confluent.kafka.databalancing.integration.utils;

import kafka.common.Topic;
import kafka.server.KafkaConfig;
import kafka.server.KafkaConfig$;
import kafka.server.KafkaServer;
import kafka.utils.TestUtils;
import kafka.utils.ZkUtils;
import kafka.zk.EmbeddedZookeeper;
import org.apache.kafka.common.security.JaasUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import scala.collection.JavaConversions$;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.Properties;

public class EmbeddedKafkaCluster {

    public static final String METRICS_TOPIC = "_confluent-metrics";
    public static final int METRICS_TOPIC_PARTITIONS = 10;

    private static final Logger LOG = LoggerFactory.getLogger(EmbeddedKafkaCluster.class);

    private final EmbeddedZookeeper zookeeper;
    private final int origNumBrokers;
    private final List<EmbeddedKafkaServer> brokers;
    private final EmbeddedKafkaCluster metricsCluster;
    private int newNumBrokers;
    private boolean brokerCountChanged;

    public final ZkUtils zkUtils;

    /**
     * Setup a cluster without any metrics reporting.
     * @param numBrokers the number of brokers in the cluster.
     */
    public EmbeddedKafkaCluster(int numBrokers) throws IOException {
        this(numBrokers, null);
    }

    /**
     * Setup a cluster which reports metrics to the specified cluster.
     * @param numBrokers The number of brokers in the cluster.
     * @param metricsCluster the cluster where metrics are to be reported.
     * @throws IOException
     */
    public EmbeddedKafkaCluster(int numBrokers, EmbeddedKafkaCluster metricsCluster) throws IOException {
        brokerCountChanged = false;
        origNumBrokers = numBrokers;
        newNumBrokers = 0;
        zookeeper = new EmbeddedZookeeper();
        this.metricsCluster = metricsCluster;
        zkUtils = ZkUtils.apply(zkConnectString(), 30000, 7000, JaasUtils.isZkSecurityEnabled());
        brokers = new ArrayList<>();
        for (int i = 0; i < numBrokers; ++i) {
            brokers.add(createBroker(i));
        }

        LOG.info("creating offset topic");
        waitUntilOffsetTopicIsCreated(brokers.get(0).kafkaServer().config(),
                brokers.get(0).kafkaServer().groupCoordinator().offsetsTopicConfigs(),
                brokers);
   }

    public void growTo(int newNumBrokers) throws IOException {
        // To keep the code simple, we allow only one cluster topology change per test.
        if (brokerCountChanged) {
            throw new IllegalStateException("Cannot change cluster size more than once.");
        }
        this.newNumBrokers = newNumBrokers;
        brokerCountChanged = true;

        for (int i = origNumBrokers; i < this.newNumBrokers; ++i) {
            brokers.add(createBroker(i));
        }

    }

    public void createTopic(String topicName, int partitions, int replicationFactor) {
        createTopic(topicName, partitions, replicationFactor, new Properties());
    }

    public void createTopic(String topicName, int partitions, int replicationFactor, Properties topicProps) {
        brokers.get(0).createTopic(topicName, partitions, replicationFactor, topicProps);
        ensureTopicCreated(topicName, partitions);
    }

    public void ensureTopicCreated(String topicName, int partitions) {
        waitUntilTopicIsCreated(topicName, partitions, brokers, zkUtils);
    }

    public static void ensureTopicCreated(String topicName, int partitions, EmbeddedKafkaCluster cluster) {
        waitUntilTopicIsCreated(topicName, partitions, cluster.brokers, cluster.zkUtils);
    }

    public boolean removeBrokerWithId(int brokerId) {
        EmbeddedKafkaServer brokerToRemove = null;
        for (EmbeddedKafkaServer broker : brokers) {
            if (broker.kafkaServer().config().brokerId() == brokerId) {
                brokerToRemove = broker;
                break;
            }
        }
        if (brokerToRemove != null) {
            brokerToRemove.stop();
            brokers.remove(brokerToRemove);
            return true;
        }
        return false;
    }

    public void shutDown() {
        LOG.info("Shutting down the cluster");
        for (EmbeddedKafkaServer broker : brokers) {
            broker.stop();
        }
        zookeeper.shutdown();
        zkUtils.close();
    }

    public String brokerList() {
        return brokerList(0, brokers.size());
    }

    public String brokerList(int start, int end) {
        return brokerList(brokers, start, end);
    }

    private static String brokerList(List<EmbeddedKafkaServer> brokers, int start, int end) {
        StringBuilder builder = new StringBuilder();
        boolean first = true;
        for (int i = start; i < end; ++i) {
            if (!first) {
                builder.append(",");
            } else {
                first = false;
            }
            builder.append(brokers.get(i).brokerList());
        }
        return builder.toString();
    }

    public String zkConnectString() {
        return "localhost:" + zookeeper.port();
    }

    private EmbeddedKafkaServer createBroker(int brokerId) throws IOException {
        Properties brokerConfig = new Properties();
        brokerConfig.put(KafkaConfig$.MODULE$.BrokerIdProp(), Integer.toString(brokerId));
        brokerConfig.put(KafkaConfig$.MODULE$.ZkConnectProp(), zkConnectString());
        brokerConfig.put(KafkaConfig$.MODULE$.PortProp(), TestUtils.RandomPort());
        brokerConfig.put(KafkaConfig$.MODULE$.DeleteTopicEnableProp(), "true");
        brokerConfig.put(KafkaConfig$.MODULE$.ControlledShutdownEnableProp(), "false");
        brokerConfig.put(KafkaConfig$.MODULE$.LogCleanerEnableProp(), "false");
        if (metricsCluster != null) {
            brokerConfig.put(KafkaConfig$.MODULE$.MetricReporterClassesProp(),
                    "io.confluent.metrics.reporter.ConfluentMetricsReporter");
            brokerConfig.put(KafkaConfig$.MODULE$.OffsetsTopicReplicationFactorProp(), "2");
            brokerConfig.put(KafkaConfig$.MODULE$.OffsetsTopicPartitionsProp(), "10");
            brokerConfig.put("confluent.metrics.reporter.bootstrap.servers", metricsCluster.brokerList());
            brokerConfig.put("confluent.metrics.reporter.zookeeper.connect", metricsCluster.zkConnectString());
            brokerConfig.put("confluent.metrics.reporter.topic.replicas", "1");

        }
        return new EmbeddedKafkaServer(brokerConfig);
    }


    private static void waitUntilTopicIsCreated(String topic, int partitions, Collection<EmbeddedKafkaServer> servers,
                                                ZkUtils zkUtils) {
        List<KafkaServer> kafkaServers = new ArrayList<>();
        for (EmbeddedKafkaServer embedded : servers) {
            kafkaServers.add(embedded.kafkaServer());
        }
        for (int i = 0; i < partitions; ++i) {
            TestUtils.waitUntilMetadataIsPropagated(JavaConversions$.MODULE$.asScalaBuffer(kafkaServers), topic, i, 60000);
            final scala.Option<Object> None = scala.Option.apply(null);
            TestUtils.waitUntilLeaderIsElectedOrChanged(zkUtils, topic, i, 60000, None, None);
        }
    }

    private void waitUntilOffsetTopicIsCreated(KafkaConfig config, Properties topicConfigs,
                                               Collection<EmbeddedKafkaServer> servers) {
        Integer offsetTopicPartitions = config.getInt(KafkaConfig.OffsetsTopicPartitionsProp());
        Short offsetTopicReplicationFactor = config.getShort(KafkaConfig.OffsetsTopicReplicationFactorProp());
        String offsetTopic = Topic.GroupMetadataTopicName();
        createTopic(offsetTopic, offsetTopicPartitions, offsetTopicReplicationFactor, topicConfigs);
        LOG.info("OffsetTopic: {}, num partitions: {}, replication factor: {}, configs: {}", offsetTopic,
                offsetTopicPartitions.toString(), offsetTopicReplicationFactor, topicConfigs.toString());
        waitUntilTopicIsCreated(offsetTopic, offsetTopicPartitions, servers, zkUtils);
    }
}
