package io.confluent.kafka.databalancing.throttle;

import io.confluent.kafka.databalancing.topology.ReplicaAssignment;
import kafka.admin.AdminUtilities;
import kafka.server.ConfigType;
import kafka.utils.ZkUtils;
import org.apache.kafka.common.TopicPartition;
import org.junit.Before;
import org.junit.Test;
import org.mockito.ArgumentCaptor;
import scala.collection.Seq;

import java.util.HashMap;
import java.util.List;
import java.util.Properties;

import static io.confluent.kafka.databalancing.TestUtils.*;
import static io.confluent.kafka.databalancing.throttle.Throttler.*;
import static java.util.Arrays.asList;
import static junit.framework.TestCase.assertFalse;
import static junit.framework.TestCase.assertTrue;
import static org.junit.Assert.assertEquals;
import static org.mockito.ArgumentCaptor.forClass;
import static org.mockito.Mockito.*;
import static scala.collection.JavaConversions.asScalaBuffer;

public class ThrottlerTest {
    private static final String topic1 = "topic1";
    private static final String topic2 = "topic2";
    private AdminUtilities admin;
    private ZkUtils zk;
    private ArgumentCaptor<Properties> props;
    private Throttler throttler;

    @Before
    public void setUp() {
        zk = mock(ZkUtils.class);
        admin = mock(AdminUtilities.class);
        props = forClass(Properties.class);
        throttler = new Throttler(zk, admin);
        when(admin.fetchEntityConfig((ZkUtils) anyObject(), eq(ConfigType.Broker()), anyString())).thenReturn(new Properties());
        when(admin.fetchEntityConfig((ZkUtils) anyObject(), eq(ConfigType.Topic()), eq(topic1))).thenReturn(new Properties());
        when(admin.fetchEntityConfig((ZkUtils) anyObject(), eq(ConfigType.Topic()), eq(topic2))).thenReturn(new Properties());
    }

    @Test
    public void shouldSetLimitOnMovingReplicas() {
        //Given a move from 100 to 102, where 101 is passive
        ReplicaAssignment current = assignment(asList(topic1), 1, asList(100, 101));
        ReplicaAssignment proposed = assignment(asList(topic1), 1, asList(101, 102));

        //When
        throttler.engage(42L, new AssignmentPair(current, proposed));

        //Then assert we set the limit on all three brokers (we don't discriminate with limits, just throttled replicas)
        for (int broker : asList(100, 101, 102)) {
            verify(admin).changeBrokerConfig(eq(zk), eq(toSeq(broker)), props.capture());
            assertProperty(props, "42", LEADER_RATE_PROP);
            assertProperty(props, "42", FOLLOWER_RATE_PROP);
        }
    }

    @Test
    public void shouldSetReplicaListForMovingReplicas() {

        //Given a move from 100 to 102, where 101 is passive
        ReplicaAssignment current = assignment(asList(topic1), 1, asList(100, 101));
        ReplicaAssignment proposed = assignment(asList(topic1), 1, asList(101, 102));

        //When
        throttler.engage(42L, new AssignmentPair(current, proposed));

        //Then
        verify(admin).changeTopicConfig(eq(zk), eq(topic1), props.capture());
        //Leader throttles should be on the two original replicas (as either could be leader)
        assertProperty(props, "0:100,0:101", LEADER_REPLICAS_PROP);
        //Follower throttle should only be on the move destination: 102
        assertProperty(props, "0:102", FOLLOWER_REPLICAS_PROP);
    }

    @Test
    public void shouldThrottleMultipleTopics() {

        //Given topics 1 -> move from broker 100 -> 102, topics 2 -> move from broker 101 -> 100
        ReplicaAssignment current = new ReplicaAssignment(new HashMap() {{
            put(new TopicPartition(topic1, 0), asList(100, 101));
            put(new TopicPartition(topic2, 0), asList(101, 102));
        }});
        ReplicaAssignment proposed = new ReplicaAssignment(new HashMap() {{
            put(new TopicPartition(topic1, 0), asList(101, 102));
            put(new TopicPartition(topic2, 0), asList(100, 102));
        }});

        //When
        throttler.engage(42L, new AssignmentPair(current, proposed));

        //Then Assert we set the throttled replicas on the topic config
        verify(admin, times(1)).changeTopicConfig(eq(zk), eq(topic1), props.capture());
        assertProperty(props, "0:100,0:101", LEADER_REPLICAS_PROP);
        assertProperty(props, "0:102", FOLLOWER_REPLICAS_PROP);

        verify(admin, times(1)).changeTopicConfig(eq(zk), eq(topic2), props.capture());
        assertProperty(props, "0:101,0:102", LEADER_REPLICAS_PROP);
        assertProperty(props, "0:100", FOLLOWER_REPLICAS_PROP);

        //Then assert we also set the limit on all brokers
        for (int broker : asList(100, 101, 102)) {
            verify(admin).changeBrokerConfig(eq(zk), eq(toSeq(broker)), props.capture());
            assertProperty(props, "42", LEADER_RATE_PROP);
            assertProperty(props, "42", FOLLOWER_RATE_PROP);
        }
    }

    @Test
    public void shouldThrottleMultiplePartitions() {

        //Given partitions 0 & 1 moves from broker 100 -> 102. Partition 2 does not move.
        ReplicaAssignment current = new ReplicaAssignment(new HashMap() {{
            put(new TopicPartition(topic1, 0), asList(100, 101));
            put(new TopicPartition(topic1, 1), asList(100, 101));
        }});
        ReplicaAssignment proposed = new ReplicaAssignment(new HashMap() {{
            put(new TopicPartition(topic1, 0), asList(101, 102));
            put(new TopicPartition(topic1, 1), asList(101, 102));
        }});

        //When
        throttler.engage(42L, new AssignmentPair(current, proposed));

        //Then Assert we set the throttled replicas on the topic
        verify(admin, times(1)).changeTopicConfig(eq(zk), eq(topic1), props.capture());
        assertProperty(props, "0:100,0:101,1:100,1:101", LEADER_REPLICAS_PROP);
        assertProperty(props, "0:102,1:102", FOLLOWER_REPLICAS_PROP);

        //Then assert we also set the limit per broker
        for (int broker : asList(100, 101, 102)) {
            verify(admin).changeBrokerConfig(eq(zk), eq(toSeq(broker)), props.capture());
            assertProperty(props, "42", LEADER_RATE_PROP);
            assertProperty(props, "42", FOLLOWER_RATE_PROP);
        }
    }

    @Test
    public void shouldThrottleMultipleTopicsAndPartitions() {

        //Given
        ReplicaAssignment current = new ReplicaAssignment(new HashMap() {{
            put(new TopicPartition(topic1, 0), asList(100, 101));
            put(new TopicPartition(topic1, 1), asList(100, 101));
            put(new TopicPartition(topic2, 0), asList(101, 102));
            put(new TopicPartition(topic2, 1), asList(101, 102));

            //Control (doesn't move):
            put(new TopicPartition(topic2, 2), asList(100, 102));
        }});
        ReplicaAssignment proposed = new ReplicaAssignment(new HashMap() {{
            put(new TopicPartition(topic1, 0), asList(101, 102)); //moves to 102
            put(new TopicPartition(topic1, 1), asList(101, 102)); //moves to 102
            put(new TopicPartition(topic2, 0), asList(100, 102)); //moves to 100
            put(new TopicPartition(topic2, 1), asList(101, 100)); //moves to 100

            //Control (doesn't move):
            put(new TopicPartition(topic2, 2), asList(100, 102));
        }});

        //When
        throttler.engage(42L, new AssignmentPair(current, proposed));

        //Then Assert we set the throttled replicas on the topic
        verify(admin, times(1)).changeTopicConfig(eq(zk), eq(topic1), props.capture());
        assertProperty(props, "0:100,0:101,1:100,1:101", LEADER_REPLICAS_PROP);
        assertProperty(props, "0:102,1:102", FOLLOWER_REPLICAS_PROP);

        verify(admin, times(1)).changeTopicConfig(eq(zk), eq(topic2), props.capture());
        assertProperty(props, "0:101,0:102,1:101,1:102", LEADER_REPLICAS_PROP);
        assertProperty(props, "0:100,1:100", FOLLOWER_REPLICAS_PROP);

        //Then assert we also set the limit on all brokers
        for (int broker : asList(100, 101, 102)) {
            verify(admin).changeBrokerConfig(eq(zk), eq(toSeq(broker)), props.capture());
            assertProperty(props, "42", LEADER_RATE_PROP);
            assertProperty(props, "42", FOLLOWER_RATE_PROP);
        }
    }

    @Test
    public void shouldFindTwoMovingReplicasInSamePartition() {

        //Given topics 1 -> move from broker 100 -> 102, topics 2 -> move from broker 101 -> 100
        ReplicaAssignment current = new ReplicaAssignment(new HashMap() {{
            put(new TopicPartition(topic1, 0), asList(100, 101, 102, 103));
        }});
        ReplicaAssignment proposed = new ReplicaAssignment(new HashMap() {{
            put(new TopicPartition(topic1, 0), asList(100, 101, 104, 105));
        }});

        //When
        throttler.engage(42L, new AssignmentPair(current, proposed));

        //Then Assert we set the throttled replicas on the topic
        verify(admin, times(1)).changeTopicConfig(eq(zk), eq(topic1), props.capture());
        assertProperty(props, "0:100,0:101,0:102,0:103", LEADER_REPLICAS_PROP);
        assertProperty(props, "0:104,0:105", FOLLOWER_REPLICAS_PROP);

        //Then assert we also set the limit per broker
        for (int broker : asList(100, 101, 102)) {
            verify(admin).changeBrokerConfig(eq(zk), eq(toSeq(broker)), props.capture());
            assertProperty(props, "42", LEADER_RATE_PROP);
            assertProperty(props, "42", FOLLOWER_RATE_PROP);
        }
    }

    @Test
    public void shouldChangeLimitIfExistingAssignment() {

        //Given a move from 100 to 102, where 101 is passive
        ReplicaAssignment current = assignment(asList(topic1), 1, asList(100, 101));
        ReplicaAssignment proposed = assignment(asList(topic1), 1, asList(101, 102));

        //Given
        when(zk.pathExists(ZkUtils.ReassignPartitionsPath())).thenReturn(true);

        //When
        throttler.engage(42L, current, proposed);

        //Then assert we set the limit on all three brokers (we don't discriminate with limits, just throttled replicas)
        for (int broker : asList(100, 101, 102)) {
            verify(admin).changeBrokerConfig(eq(zk), eq(toSeq(broker)), props.capture());
            assertProperty(props, "42", LEADER_RATE_PROP);
            assertProperty(props, "42", FOLLOWER_RATE_PROP);
        }
    }

    @Test
    public void shouldRemoveQuotaLimitBrokerConfigOnDisengage() {
        when(zk.getAllTopics()).thenReturn(asScalaBuffer(asList(topic1)));
        int[] allBrokersInCluster = {100, 101, 102, 103, 104};

        //Given a larger list of brokers, and configured properties
        when(zk.getAllBrokersInCluster()).thenReturn(toBrokerSeq(allBrokersInCluster));
        for (int id : allBrokersInCluster)
            when(admin.fetchEntityConfig(eq(zk), eq(ConfigType.Broker()), eq(String.valueOf(id))))
                    .thenReturn(props(asList(LEADER_RATE_PROP, "42", FOLLOWER_RATE_PROP, "43")));

        //When
        boolean removed = throttler.disengage();

        //Then assert broker config should have been removed across all 5 brokers
        for (int broker : allBrokersInCluster) {
            verify(admin).changeBrokerConfig(eq(zk), eq(toSeq(broker)), props.capture());
            assertFalse(props.getValue().containsKey(LEADER_RATE_PROP));
            assertFalse(props.getValue().containsKey(FOLLOWER_RATE_PROP));
        }
        assertTrue(removed);
    }

    @Test
    public void shouldRemoveReplicaListTopicConfigOnDisengage() {
        when(zk.getAllTopics()).thenReturn(asScalaBuffer(asList(topic1)));
        when(zk.getAllBrokersInCluster()).thenReturn(toBrokerSeq(100));

        //Given there is an existing set of throttled replicas
        List<String> existing = asList(
                FOLLOWER_REPLICAS_PROP, "1:101,1:102",
                LEADER_REPLICAS_PROP, "1:101,1:102"
        );
        when(admin.fetchEntityConfig(eq(zk), eq(ConfigType.Topic()), eq(topic1))).thenReturn(props(existing));

        //When
        boolean removed = throttler.disengage();

        //Then assert broker config should have been removed across all 5 brokers
        verify(admin).changeTopicConfig(eq(zk), eq(topic1), props.capture());
        assertFalse(props.getValue().containsKey(FOLLOWER_REPLICAS_PROP));
        assertFalse(props.getValue().containsKey(LEADER_REPLICAS_PROP));
        assertTrue(removed);
    }

    @Test
    public void shouldReportSuccessEvenIfOnlyPartOfTheThrottleReplicasConfigWasChanged() {
        when(zk.getAllTopics()).thenReturn(asScalaBuffer(asList(topic1)));
        when(zk.getAllBrokersInCluster()).thenReturn(toBrokerSeq(100));

        //Given there is only follower replicas are throttled
        when(admin.fetchEntityConfig(eq(zk), eq(ConfigType.Topic()), eq(topic1)))
                .thenReturn(props(FOLLOWER_REPLICAS_PROP, "1:101,1:102"));

        //When
        boolean applied = throttler.disengage();

        //Then
        assertTrue(applied);
    }

    @Test
    public void shouldReportSuccessEvenIfOnlyPartOfTheThrottleLimitConfigWasChanged() {
        when(zk.getAllTopics()).thenReturn(asScalaBuffer(asList(topic1)));
        when(zk.getAllBrokersInCluster()).thenReturn(toBrokerSeq(100));

        //Given there is only leader limit applied
        when(admin.fetchEntityConfig(eq(zk), eq(ConfigType.Broker()), eq(String.valueOf(100))))
                .thenReturn(props(LEADER_RATE_PROP, "42"));

        //When
        boolean applied = throttler.disengage();

        //Then
        assertTrue(applied);
    }

    @Test
    public void shouldNotReportRemovalIfNothingRemoved() {
        when(zk.getAllTopics()).thenReturn(asScalaBuffer(asList(topic1)));
        when(zk.getAllBrokersInCluster()).thenReturn(toBrokerSeq(100));

        //Given there there is no throttle configured of any type

        //When
        boolean applied = throttler.disengage();

        //Then
        assertFalse(applied);
    }


    @Test
    public void shouldNotUpdateZkIfThrottleDidNotChange() {
        reset(admin);
        ReplicaAssignment current = assignment(asList(topic1), 1, asList(100));
        ReplicaAssignment proposed = assignment(asList(topic1), 1, asList(101));
        when(zk.getAllBrokersInCluster()).thenReturn(toBrokerSeq(100, 101));

        //Given the cluster is already correctly throttled
        when(admin.fetchEntityConfig(eq(zk), eq(ConfigType.Topic()), eq(topic1)))
                .thenReturn(props(asList(FOLLOWER_REPLICAS_PROP, "0:101", LEADER_REPLICAS_PROP, "0:100")));
        when(admin.fetchEntityConfig(eq(zk), eq(ConfigType.Broker()), eq(String.valueOf(100))))
                .thenReturn(props(asList(LEADER_RATE_PROP, "42", FOLLOWER_RATE_PROP, "42")));
        when(admin.fetchEntityConfig(eq(zk), eq(ConfigType.Broker()), eq(String.valueOf(101))))
                .thenReturn(props(asList(LEADER_RATE_PROP, "42", FOLLOWER_RATE_PROP, "42")));

        //When
        throttler.engage(42L, new AssignmentPair(current, proposed));

        //Then verify that no updates were made to ZK
        verify(admin, never()).changeTopicConfig(eq(zk), anyString(), any(Properties.class));
        verify(admin, never()).changeBrokerConfig(eq(zk), any(Seq.class), any(Properties.class));
    }


    @Test
    public void shouldUpdateThrottleValuesIfThereAreExistingOnesInZkThatHaveDifferentValues() {

        //Given a move
        ReplicaAssignment current = assignment(asList(topic1), 1, asList(100));
        ReplicaAssignment proposed = assignment(asList(topic1), 1, asList(101));

        //Given the cluster is already has a throttle configured (but different values)
        when(admin.fetchEntityConfig(eq(zk), eq(ConfigType.Topic()), eq(topic1)))
                .thenReturn(props(asList(FOLLOWER_REPLICAS_PROP, "0:104", LEADER_REPLICAS_PROP, "0:102")));
        when(admin.fetchEntityConfig(eq(zk), eq(ConfigType.Broker()), eq(String.valueOf(100))))
                .thenReturn(props(asList(LEADER_RATE_PROP, "42", FOLLOWER_RATE_PROP, "42")));
        when(admin.fetchEntityConfig(eq(zk), eq(ConfigType.Broker()), eq(String.valueOf(101))))
                .thenReturn(props(asList(LEADER_RATE_PROP, "42", FOLLOWER_RATE_PROP, "42")));

        //When we update the limit to something new
        throttler.engage(69L, new AssignmentPair(current, proposed));

        //Then Assert the replicas list was written correctly
        verify(admin, times(1)).changeTopicConfig(eq(zk), eq(topic1), props.capture());
        assertProperty(props, "0:100", LEADER_REPLICAS_PROP);
        assertProperty(props, "0:101", FOLLOWER_REPLICAS_PROP);

        //Then Assert the limit was written correctly
        for (int broker : asList(100, 101)) {
            verify(admin).changeBrokerConfig(eq(zk), eq(toSeq(broker)), props.capture());
            assertProperty(props, "69", LEADER_RATE_PROP);
            assertProperty(props, "69", FOLLOWER_RATE_PROP);
        }
    }

    @Test
    public void shouldNotOverwriteOtherTypesOfExistingConfig() {
        ReplicaAssignment current = new ReplicaAssignment(new HashMap() {{
            put(new TopicPartition(topic1, 0), asList(100, 101));
            put(new TopicPartition(topic2, 0), asList(101, 100));
        }});
        ReplicaAssignment proposed = new ReplicaAssignment(new HashMap() {{
            put(new TopicPartition(topic1, 0), asList(101, 100));
            put(new TopicPartition(topic2, 0), asList(100, 101));
        }});


        //Given we have some existing properties for each type of entity.
        when(admin.fetchEntityConfig((ZkUtils) anyObject(), eq(ConfigType.Topic()), eq(topic1)))
                .thenReturn(props("some-key", "some-topic1-value"));
        when(admin.fetchEntityConfig((ZkUtils) anyObject(), eq(ConfigType.Topic()), eq(topic2)))
                .thenReturn(props("some-key", "some-topic2-value"));
        when(admin.fetchEntityConfig((ZkUtils) anyObject(), eq(ConfigType.Broker()), eq("100")))
                .thenReturn(props("some-key", "some-broker-value-100"));
        when(admin.fetchEntityConfig((ZkUtils) anyObject(), eq(ConfigType.Broker()), eq("101")))
                .thenReturn(props("some-key", "some-broker-value-101"));

        //When
        throttler.engage(42L, new AssignmentPair(current, proposed));

        //Then the original broker config should not be overwritten
        verify(admin).changeBrokerConfig(eq(zk), eq(toSeq(100)), props.capture());
        assertProperty(props, "some-broker-value-100", "some-key");
        props = forClass(Properties.class);
        verify(admin).changeBrokerConfig(eq(zk), eq(toSeq(101)), props.capture());
        assertProperty(props, "some-broker-value-101", "some-key");


        //Then the original topic config should not be overwritten
        verify(admin).changeTopicConfig(eq(zk), eq(topic1), props.capture());
        assertProperty(props, "some-topic1-value", "some-key");
        verify(admin).changeTopicConfig(eq(zk), eq(topic2), props.capture());
        assertProperty(props, "some-topic2-value", "some-key");
    }

    private void assertProperty(ArgumentCaptor<Properties> props, String expected, String property) {
        assertEquals(expected, props.getValue().getProperty(property));
    }
}
    

