// (Copyright) [2016 - 2016] Confluent, Inc.
package io.confluent.kafka.databalancing.topology;

import io.confluent.kafka.databalancing.MutableRebalanceContext;
import io.confluent.kafka.databalancing.RebalanceContext;
import io.confluent.kafka.databalancing.TestUtils;
import io.confluent.kafka.databalancing.topology.Broker;
import io.confluent.kafka.databalancing.topology.TopologyUtils;
import kafka.admin.BrokerMetadata;
import org.apache.kafka.common.TopicPartition;
import org.junit.Test;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.List;
import java.util.Map;

import static io.confluent.kafka.databalancing.TestUtils.bk;
import static io.confluent.kafka.databalancing.TestUtils.concat;
import static io.confluent.kafka.databalancing.TestUtils.createBrokers;
import static io.confluent.kafka.databalancing.TestUtils.createMap;
import static io.confluent.kafka.databalancing.TestUtils.createRebalanceContext;
import static io.confluent.kafka.databalancing.TestUtils.entry;
import static io.confluent.kafka.databalancing.topology.TopologyUtils.downrank;
import static io.confluent.kafka.databalancing.topology.TopologyUtils.leastLoadedBrokersPreferringOtherRacks;
import static java.util.Arrays.asList;
import static org.junit.Assert.assertEquals;

public class TopologyUtilsTest {

    private static final String topic = "topic";

    private static TopicPartition tp(int partition) {
        return new TopicPartition(topic, partition);
    }

    private static Map<String, Integer> replicationFactors(int replicationFactor) {
        return TestUtils.replicationFactors(asList(topic), replicationFactor);
    }

    @Test
    public void shouldOrderBrokersByReplicaLoad() {
        //Given three brokers with increasing load: 101,102,100
        List<BrokerMetadata> brokers = createBrokers(100, 103, "rack1");
        Map<TopicPartition, List<Integer>> partitions = createMap(
                entry(tp(0), asList(100, 101, 102)),
                entry(tp(1), asList(100, 101, 102)),
                entry(tp(2), asList(100, 101, 102)),
                entry(tp(3), asList(100, 101)),
                entry(tp(4), asList(100)));

        //When
        RebalanceContext context = createRebalanceContext(brokers, partitions, replicationFactors(2));
        List<Broker> leastLoaded = TopologyUtils.leastLoadedBrokers(context);

        //Then
        assertEquals(asList(new Broker(102), new Broker(101), new Broker(100)), leastLoaded);
    }

    @Test
    public void shouldConsiderEmptyBrokers() {
        //Given two brokers, only one with replicas
        List<BrokerMetadata> brokers = createBrokers(100, 102, "rack1");
        Map<TopicPartition, List<Integer>> partitions = createMap(entry(tp(4), asList(101)));

        //When
        RebalanceContext context = createRebalanceContext(brokers, partitions, replicationFactors(2));
        List<Broker> leastLoaded = TopologyUtils.leastLoadedBrokers(context);

        //Then
        assertEquals(asList(new Broker(100), new Broker(101)), leastLoaded);
    }

    @Test
    public void shouldConsiderEmptyBrokersWithNoRack() {
        //Given two brokers, only one with replicas
        List<BrokerMetadata> brokers = createBrokers(100, 102, null);
        Map<TopicPartition, List<Integer>> partitions = createMap(entry(tp(4), asList(101)));

        //When
        RebalanceContext context = createRebalanceContext(brokers, partitions, replicationFactors(2));
        List<Broker> leastLoaded = TopologyUtils.leastLoadedBrokers(context);

        //Then
        assertEquals(asList(new Broker(100), new Broker(101)), leastLoaded);
    }

    @Test
    public void shouldPreferRacksThatAreNotPassedSoTheyAppearLast() {
        //Given
        List<BrokerMetadata> brokers = concat(
                createBrokers(100, 102, "rack1"),
                createBrokers(102, 104, "racl2"));
        Map<TopicPartition, List<Integer>> partitions = createMap(
                entry(tp(0), asList(103, 102, 101, 100)),
                entry(tp(1), asList(103, 102, 101)),
                entry(tp(2), asList(103, 102)),
                entry(tp(3), asList(103)));

        //When
        RebalanceContext context = createRebalanceContext(brokers, partitions, replicationFactors(2));
        List<Broker> leastLoaded = leastLoadedBrokersPreferringOtherRacks(context, Collections.singleton("rack1"));

        //Then least loaded would be 100, 101, 102, 103 (based purely on replica count, least loaded first)
        //but rack1 (100, 101) should drop in priority so they appear last
        assertEquals(asList(new Broker(102), new Broker(103), new Broker(100), new Broker(101)), leastLoaded);
    }

    @Test
    public void shouldSummariseLeaderCounts() {
        List<BrokerMetadata> brokers = asList(bk(100, "rack1"), bk(101, "rack2"));
        Map<TopicPartition, List<Integer>> partitions = createMap(
                entry(tp(0), asList(100, 101)),
                entry(tp(1), asList(100, 101)),
                entry(tp(2), asList(101, 100)),
                entry(tp(3), asList(101, 100)),
                entry(tp(4), asList(101, 100)));

        RebalanceContext context = createRebalanceContext(brokers, partitions, replicationFactors(2));
        Map<Broker, Long> leaderCounts = TopologyUtils.brokerLeaderCounts(context, context.allBrokers(), null);

        assertEquals(2, leaderCounts.get(new Broker(100)).longValue());
        assertEquals(3, leaderCounts.get(new Broker(101)).longValue());
    }

    @Test
    public void shouldDownrank() {
        List<Broker> brokers = new ArrayList<>(Arrays.asList(new Broker(103), new Broker(104)));
        downrank(brokers, Arrays.asList(new Broker(104), new Broker(103)));
        assertEquals(Arrays.asList(new Broker(103), new Broker(104)), brokers);
        brokers.add(new Broker(105));
        brokers.add(new Broker(106));
        downrank(brokers, Arrays.asList(new Broker(104), new Broker(103)));
        assertEquals(Arrays.asList(new Broker(105), new Broker(106), new Broker(103), new Broker(104)), brokers);
    }

}
