// (Copyright) [2016 - 2016] Confluent, Inc.
package io.confluent.controlcenter;

import com.google.common.collect.ImmutableSet;
import com.google.common.util.concurrent.ThreadFactoryBuilder;
import com.google.inject.Guice;
import com.google.inject.Injector;
import com.google.inject.Key;

import com.github.zafarkhaja.semver.Version;

import io.confluent.controlcenter.data.ClusterMetadataInitializer;
import org.apache.commons.cli.CommandLine;
import org.apache.commons.cli.CommandLineParser;
import org.apache.commons.cli.DefaultParser;
import org.apache.commons.cli.Option;
import org.apache.commons.cli.Options;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.ScheduledExecutorService;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.TimeoutException;

import io.confluent.command.CommandStore;
import io.confluent.controlcenter.command.CommandModule;
import io.confluent.controlcenter.connect.ConnectService;
import io.confluent.controlcenter.rest.ControlCenterApplication;
import io.confluent.controlcenter.rest.RestModule;
import io.confluent.controlcenter.serialization.SerializationModule;
import io.confluent.controlcenter.streams.KafkaStreamsManager;
import io.confluent.controlcenter.streams.StreamsConfigModule;
import io.confluent.controlcenter.streams.StreamsModule;
import io.confluent.controlcenter.streams.TopicStoreModule;
import io.confluent.controlcenter.streams.verify.MonitoringHeartbeatSender;
import io.confluent.controlcenter.license.LicenseModule;
import io.confluent.monitoring.common.TimeBucket;

public class ControlCenter {

  private static final Logger log = LoggerFactory.getLogger(ControlCenter.class);

  public static void main(String[] args) throws Exception {
    final String TOPICS_FOR_TASKS_OPT = "runTasksFor";

    Options options = new Options();
    Option topicsForTasksOpt = Option.builder(TOPICS_FOR_TASKS_OPT)
        .hasArg()
        .longOpt("run-tasks-for")
        .desc("Comma-separated list of topics. Run only tasks whose input topics include this set.")
        .build();
    options.addOption(topicsForTasksOpt);

    CommandLineParser parser = new DefaultParser();
    CommandLine cmdLine = parser.parse(options, args);

    if (cmdLine.getArgList().size() == 0) {
      System.err.println("You must provide a path to the config file");
      System.exit(1);
    }
    final ControlCenterConfig config = new ControlCenterConfig(cmdLine.getArgList().get(0));
    boolean enableSubTopologies = false;
    ImmutableSet<String> topicsForTasks = null;

    if (cmdLine.hasOption(TOPICS_FOR_TASKS_OPT)) {
      enableSubTopologies = true;
      topicsForTasks = ImmutableSet.copyOf(cmdLine.getOptionValue(TOPICS_FOR_TASKS_OPT).trim().split(","));
    }

    Injector injector = Guice.createInjector(
        new ControlCenterConfigModule(config, enableSubTopologies, topicsForTasks),
        new CommandModule(),
        new StreamsConfigModule(),
        new TopicStoreModule(),
        new StreamsModule(),
        new RestModule(),
        new SerializationModule(),
        new LicenseModule()
    );
    Version c3Version = injector.getInstance(Key.get(Version.class));
    log.info("Starting Control Center version={}", c3Version);
    KafkaHelper.ControlCenterPreconditions ccp = injector.getInstance(KafkaHelper.ControlCenterPreconditions.class);

    if (!ccp.call()) {
      log.error("failed to create necessary topics");
      return;
    }

    final KafkaStreamsManager kafkaMonitoringStreamsManager =
        injector.getInstance(KafkaStreamsManager.class);
    final CommandStore commandStore =
        injector.getInstance(CommandStore.class);
    final MonitoringHeartbeatSender monitoringHeartbeatSender =
        injector.getInstance(MonitoringHeartbeatSender.class);
    final ClusterMetadataInitializer clusterMetadataInitializer = injector.getInstance(ClusterMetadataInitializer.class);
    final ScheduledExecutorService ses = Executors.newScheduledThreadPool(
        2,
        new ThreadFactoryBuilder().setNameFormat("control-center-heartbeat-%d").build()
    );

    Runtime.getRuntime().addShutdownHook(new Thread() {
      @Override
      public void run() {
        shutdownTasksAndAwaitTermination(ses);
        monitoringHeartbeatSender.close();
        kafkaMonitoringStreamsManager.close();
        commandStore.close();
      }
    });

    try {
      log.info("Starting command topology");
      commandStore.start(TimeUnit.SECONDS.toMillis(60));
      ControlCenterApplication a = injector.getInstance(ControlCenterApplication.class);
      log.info("Starting monitoring topology");
      kafkaMonitoringStreamsManager.start(config.getInt(ControlCenterConfig.CONTROL_CENTER_REQUEST_TIMEOUT_MS_CONFIG) * 2);
      log.info("Initializing cluster metadata");
      clusterMetadataInitializer.call();
      ses.scheduleWithFixedDelay(monitoringHeartbeatSender, 500, TimeBucket.SIZE, TimeUnit.MILLISECONDS);
      ses.submit(injector.getInstance(ConnectService.ConnectInit.class));
      a.start();
      a.join();
    } catch (TimeoutException | InterruptedException e) {
      log.error("failed to start topology", e);
      commandStore.close();
    } catch (Exception e) {
      log.error("Unexpected error", e);
    } finally {
      shutdownTasksAndAwaitTermination(ses);
      kafkaMonitoringStreamsManager.close();
      commandStore.close();
    }
  }

  private static void shutdownTasksAndAwaitTermination(final ExecutorService es) {
    es.shutdown();
    try {
      if (!es.awaitTermination(10, TimeUnit.SECONDS)) {
        es.shutdownNow();
        if (!es.awaitTermination(60, TimeUnit.SECONDS)) {
          log.warn("Failed to shutdown heartbeat sender");
        }
      }
    } catch (InterruptedException ie) {
      log.warn("Failed to cleanly shutdown heartbeat sender");
      es.shutdownNow();
    }
  }

}
