/**
 * Copyright 2014 - 2016 Confluent Inc. 
 */
package io.confluent.controlcenter;

import org.apache.kafka.clients.CommonClientConfigs;
import org.apache.kafka.clients.consumer.ConsumerConfig;
import org.apache.kafka.clients.producer.ProducerConfig;
import org.apache.kafka.streams.StreamsConfig;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.List;
import java.util.Properties;
import java.util.concurrent.TimeUnit;

import io.confluent.command.CommandStore;
import io.confluent.common.config.AbstractConfig;
import io.confluent.common.config.ConfigDef;
import io.confluent.rest.RestConfig;

public class ControlCenterConfig extends AbstractConfig {

  public static final String CONFLUENT_CONTROLCENTER_PREFIX = "confluent.controlcenter.";
  public static final String CONFLUENT_CONTROLCENTER_KSTREAMS_PREFIX = ControlCenterConfig.CONFLUENT_CONTROLCENTER_PREFIX + "streams.";
  public static final String CONFLUENT_CONTROLCENTER_REST_PREFIX = ControlCenterConfig.CONFLUENT_CONTROLCENTER_PREFIX + "rest.";
  public static final String CONFLUENT_CONTROLCENTER_EMAIL_PREFIX = ControlCenterConfig.CONFLUENT_CONTROLCENTER_PREFIX + "mail.";
  public static final String CONFLUENT_CONTROLCENTER_CONNECT_PREFIX = ControlCenterConfig.CONFLUENT_CONTROLCENTER_PREFIX + "connect.";
  public static final String CONFLUENT_MONITORING_INTERCEPTOR_PREFIX = "confluent.monitoring.interceptor.";

  public static final String CONTROL_CENTER_NAME_CONFIG =
      ControlCenterConfig.CONFLUENT_CONTROLCENTER_PREFIX + "name";
  public static final String DEFAULT_CONTROL_CENTER_NAME = "_confluent-controlcenter-3-1-0";
  protected static final String CONTROL_CENTER_NAME_DOC =
      "Control Center Name. All internal Topics will be prepended with this value.";

  public static final String CONTROL_CENTER_ID_CONFIG =
      ControlCenterConfig.CONFLUENT_CONTROLCENTER_PREFIX + "id";
  public static final String DEFAULT_CONTROL_CENTER_ID = "1";
  protected static final String CONTROL_CENTER_ID_DOC = "Control Center ID";

  public static final String CONTROL_CENTER_TOPIC_CONFIG = ControlCenterConfig.CONFLUENT_MONITORING_INTERCEPTOR_PREFIX + "topic";
  public static final String DEFAULT_CONTROL_CENTER_TOPIC = "_confluent-monitoring";
  protected static final String CONTROL_CENTER_TOPIC_DOC =
      "The durable topic that acts as the durable log for the data";

  public static final String CONTROL_CENTER_INTERNAL_PARTITIONS_CONFIG =
      ControlCenterConfig.CONFLUENT_CONTROLCENTER_PREFIX + "internal.topics.partitions";
  public static final int DEFAULT_CONTROL_CENTER_INTERNAL_PARTITIONS = 4;
  protected static final String CONTROL_CENTER_INTERNAL_PARTITIONS_DOC =
      "Partition count for internal topics";

  public static final String CONTROL_CENTER_INTERNAL_REPLICATION_CONFIG =
      ControlCenterConfig.CONFLUENT_CONTROLCENTER_PREFIX + "internal.topics.replication";
  public static final int DEFAULT_CONTROL_CENTER_INTERNAL_REPLICATION = 3;
  protected static final String CONTROL_CENTER_INTERNAL_REPLICATION_DOC =
      "Replication factor for internal topics";

  public static final String CONTROL_CENTER_INTERNAL_RETENTION_MS_CONFIG =
      ControlCenterConfig.CONFLUENT_CONTROLCENTER_PREFIX + "internal.topics.retention.ms";
  public static final int DEFAULT_CONTROL_CENTER_INTERNAL_RETENTION_MS = 86400000;
  protected static final String CONTROL_CENTER_INTERNAL_RETENTION_MS_DOC =
      "Retention ms for internal topics";

  public static final String CONTROL_CENTER_INTERNAL_RETENTION_BYTES_CONFIG =
      ControlCenterConfig.CONFLUENT_CONTROLCENTER_PREFIX + "internal.topics.retention.bytes";
  public static final int DEFAULT_CONTROL_CENTER_INTERNAL_RETENTION_BYTES = -1;
  protected static final String CONTROL_CENTER_INTERNAL_RETENTION_BYTES_DOC =
      "Retention bytes for internal topics";

  public static final String CONTROL_CENTER_INTERNAL_CHANGELOG_SEGMENT_BYTES_CONFIG =
      ControlCenterConfig.CONFLUENT_CONTROLCENTER_PREFIX + "internal.topics.changelog.segment.bytes";
  //This must be set to 128MB for now to make sure all distinct keys for our changelogs fit in the log cleaner dedupe buffer (defaults to 128MB)
  //Otherwise, the log cleaner thread will die and prevent compaction for all topics
  public static final int DEFAULT_CONTROL_CENTER_INTERNAL_CHANGELOG_SEGMENT_BYTES = 134217728;
  protected static final String CONTROL_CENTER_INTERNAL_CHANGELOG_SEGMENT_BYTES_DOC =
      "Segment size in bytes for internal changelog topics";

  public static final String CONTROL_CENTER_MONITORING_TOPIC_PARTITIONS_CONFIG =
      CONFLUENT_MONITORING_INTERCEPTOR_PREFIX + "topic.partitions";
  public static final int DEFAULT_CONTROL_CENTER_MONITORING_TOPIC_PARTITIONS = 12;
  protected static final String CONTROL_CENTER_MONITORING_TOPIC_PARTITIONS_DOC =
      "Partition count for monitoring topic";

  public static final String CONTROL_CENTER_MONITORING_TOPIC_REPLICATION_CONFIG =
      CONFLUENT_MONITORING_INTERCEPTOR_PREFIX + "topic.replication";
  public static final int DEFAULT_CONTROL_CENTER_MONITORING_TOPIC_REPLICATION = 3;
  protected static final String CONTROL_CENTER_MONITORING_TOPIC_REPLICATION_DOC =
      "Replication factor for monitoring topic";

  public static final String CONTROL_CENTER_MONITORING_TOPIC_RETENTION_MS_CONFIG =
      CONFLUENT_MONITORING_INTERCEPTOR_PREFIX + "topic.retention.ms";
  public static final int DEFAULT_CONTROL_CENTER_MONITORING_TOPIC_RETENTION_MS = 259200000;
  protected static final String CONTROL_CENTER_MONITORING_TOPIC_RETENTION_MS_DOC =
      "Retention ms for monitoring topic";

  public static final String CONTROL_CENTER_MONITORING_TOPIC_RETENTION_BYTES_CONFIG =
      CONFLUENT_MONITORING_INTERCEPTOR_PREFIX + "topic.retention.bytes";
  public static final int DEFAULT_CONTROL_CENTER_MONITORING_TOPIC_RETENTION_BYTES = -1;
  protected static final String CONTROL_CENTER_MONITORING_TOPIC_RETENTION_BYTES_DOC =
      "Retention bytes for monitoring topic";

  public static final String CONTROL_CENTER_LICENSE_CONFIG = ControlCenterConfig.CONFLUENT_CONTROLCENTER_PREFIX + "license";
  protected static final String CONTROL_CENTER_LICENSE_DOC = "License";

  public static final String CONTROL_CENTER_BOOTSTRAP_SERVERS_CONFIG =
      ConsumerConfig.BOOTSTRAP_SERVERS_CONFIG;
  public static final String DEFAULT_CONTROL_CENTER_BOOTSTRAP_SERVERS_CONFIG = "localhost:9092";
  public static final String CONTROL_CENTER_BOOTSTRAP_SERVERS_DOC =
      CommonClientConfigs.BOOTSTRAP_SERVERS_DOC;

  public static final String CONTROL_CENTER_DATA_DIR = ControlCenterConfig.CONFLUENT_CONTROLCENTER_PREFIX + "data.dir";
  public static final String DEFAULT_CONTROL_CENTER_DATA_DIR = "/tmp/confluent/control-center";
  protected static final String CONTROL_CENTER_DATA_DIR_DOC =
      "Location for ControlCenter specific data";

  public static final String CONTROL_CENTER_MAX_POLL_RECORDS_CONFIG =
      CONFLUENT_CONTROLCENTER_KSTREAMS_PREFIX + StreamsConfig.consumerPrefix(ConsumerConfig.MAX_POLL_RECORDS_CONFIG);
  public static final int DEFAULT_CONTROL_CENTER_MAX_POLL_RECORDS_CONFIG = 10000;
  protected static final String CONTROL_CENTER_MAX_POLL_RECORDS_DOC =
      "Max # of records to read for each invocation of poll()";

  public static final String CONTROL_CENTER_SESSION_TIMEOUT_MS_CONFIG =
      CONFLUENT_CONTROLCENTER_KSTREAMS_PREFIX + StreamsConfig.consumerPrefix(ConsumerConfig.SESSION_TIMEOUT_MS_CONFIG);
  public static final int DEFAULT_CONTROL_CENTER_SESSION_TIMEOUT_MS_CONFIG = 30000;
  protected static final String CONTROL_CENTER_SESSION_TIMEOUT_MS_DOC =
      "The timeout used to detect a consumer failure";

  public static final String CONTROL_CENTER_REQUEST_TIMEOUT_MS_CONFIG =
      CONFLUENT_CONTROLCENTER_KSTREAMS_PREFIX + StreamsConfig.consumerPrefix(ConsumerConfig.REQUEST_TIMEOUT_MS_CONFIG);
  public static final int DEFAULT_CONTROL_CENTER_REQUEST_TIMEOUT_MS_CONFIG = 30001;
  protected static final String CONTROL_CENTER_REQUEST_TIMEOUT_MS_DOC =
      "The maximum amount of time the client will wait for the response of a request";

  public static final String CONTROL_CENTER_RETRIES_CONFIG =
      CONFLUENT_CONTROLCENTER_KSTREAMS_PREFIX + StreamsConfig.producerPrefix(ProducerConfig.RETRIES_CONFIG);
  public static final int DEFAULT_CONTROL_CENTER_RETRIES_CONFIG= Integer.MAX_VALUE;
  protected static final String CONTROL_CENTER_RETRIES_DOC =
      "Number of retries in case of production failure";

  public static final String CONTROL_CENTER_RETRY_BACKOFF_MS_CONFIG =
      CONFLUENT_CONTROLCENTER_KSTREAMS_PREFIX + StreamsConfig.producerPrefix(ProducerConfig.RETRY_BACKOFF_MS_CONFIG);
  public static final long DEFAULT_CONTROL_CENTER_RETRY_BACKOFF_MS_CONFIG = 100L;
  protected static final String CONTROL_CENTER_RETRY_BACKOFF_MS_DOC =
      "Time to wait before retrying in case of production failure";

  public static final String CONTROL_CENTER_COMPRESSION_TYPE_CONFIG =
      CONFLUENT_CONTROLCENTER_KSTREAMS_PREFIX + StreamsConfig.producerPrefix(ProducerConfig.COMPRESSION_TYPE_CONFIG);
  public static final String DEFAULT_CONTROL_CENTER_COMPRESSION_TYPE_CONFIG = "lz4";
  protected static final String CONTROL_CENTER_COMPRESSION_TYPE_DOC =
      "Compression type to use for internal topics";

  public static final String CONTROL_CENTER_PARALLELISM =
      CONFLUENT_CONTROLCENTER_KSTREAMS_PREFIX + StreamsConfig.NUM_STREAM_THREADS_CONFIG;
  public static final Integer DEFAULT_CONTROL_CENTER_PARALLELISM = 8;
  protected static final String CONTROL_CENTER_PARALLELISM_DOC =
      "The number of threads to execute stream processing";

  public static final String CONTROL_CENTER_PRODUCER_SECURITY_PROTOCOL =
      CONFLUENT_CONTROLCENTER_KSTREAMS_PREFIX + StreamsConfig.producerPrefix(CommonClientConfigs.SECURITY_PROTOCOL_CONFIG);
  public static final String CONTROL_CENTER_CONSUMER_SECURITY_PROTOCOL =
      CONFLUENT_CONTROLCENTER_KSTREAMS_PREFIX + StreamsConfig.consumerPrefix(CommonClientConfigs.SECURITY_PROTOCOL_CONFIG);
  public static final String DEFAULT_CONTROL_CENTER_SECURITY_PROTOCOL = CommonClientConfigs.DEFAULT_SECURITY_PROTOCOL;
  protected static final String CONTROL_CENTER_SECURITY_PROTOCOL_DOC = CommonClientConfigs.SECURITY_PROTOCOL_DOC;

  public static final String CONTROL_CENTER_REST_PORT =
      CONFLUENT_CONTROLCENTER_REST_PREFIX + RestConfig.PORT_CONFIG;
  public static final Integer DEFAULT_CONTROL_CENTER_REST_PORT = 9021;
  protected static final String CONTROL_CENTER_REST_PORT_DOC =
      "DEPRECATED: port to listen on for new HTTP connections. Use listeners instead.";

  public static final String CONTROL_CENTER_CONNECT_CLUSTER = ControlCenterConfig.CONFLUENT_CONTROLCENTER_CONNECT_PREFIX + "cluster";
  public static final List<String> DEFAULT_CONTROL_CENTER_CONNECT_CLUSTER = Collections.singletonList("localhost:8083");
  protected static final String CONTROL_CENTER_CONNECT_CLUSTER_DOC =
      "Comma-separated list of URLs for the Connect cluster";

  public static final String CONTROL_CENTER_CONNECT_ZOOKEEPER_CONNECT = ControlCenterConfig.CONFLUENT_CONTROLCENTER_CONNECT_PREFIX + ControlCenterConfig.ZOOKEEPER_CONNECT_CONFIG;
  protected static final String CONTROL_CENTER_CONNECT_ZOOKEEPER_CONNECT_DOC =
      "ZooKeeper connection string for the Kafka cluster backing the Connect cluster";

  public static final String CONTROL_CENTER_CONNECT_TIMEOUT = ControlCenterConfig.CONFLUENT_CONTROLCENTER_CONNECT_PREFIX + "timeout";
  public static final Long DEFAULT_CONTROL_CENTER_CONNECT_TIMEOUT = TimeUnit.SECONDS.toMillis(15);
  protected static final String CONTROL_CENTER_CONNECT_TIMEOUT_DOC =
      "Timeout in millis for calls to connect cluster";

  public static final String CONTROL_CENTER_DISPLAY_HOST_PREFIX =
      CONFLUENT_CONTROLCENTER_REST_PREFIX + "advertised.url";
  public static final String DEFAULT_CONTROL_CENTER_DISPLAY_HOST_PREFIX = "";
  protected static final String CONTROL_CENTER_DISPLAY_HOST_PREFIX_DOC =
      "Externally visible host. C3 uses this as an override to rest.listeners when generating urls for external communications, e.g. emails.";

  public static final String CONTROL_CENTER_EMAIL_ENABLE =
      CONFLUENT_CONTROLCENTER_EMAIL_PREFIX + "enabled";
  public static final boolean DEFAULT_CONTROL_CENTER_EMAIL_ENABLE = false;
  protected static final String CONTROL_CENTER_EMAIL_ENABLE_DOC =
      "Enable sending of emails.";

  public static final String CONTROL_CENTER_EMAIL_HOST_NAME =
      CONFLUENT_CONTROLCENTER_EMAIL_PREFIX + "host.name";
  public static final String DEFAULT_CONTROL_CENTER_EMAIL_HOST_NAME = "localhost";
  protected static final String CONTROL_CENTER_EMAIL_HOST_NAME_DOC =
      "Hostname of outgoing smtp server.";

  public static final String CONTROL_CENTER_EMAIL_PORT =
      CONFLUENT_CONTROLCENTER_EMAIL_PREFIX + "port";
  public static final Integer DEFAULT_CONTROL_CENTER_EMAIL_PORT = 587;
  protected static final String CONTROL_CENTER_EMAIL_PORT_DOC =
      "SMTP port open on " + CONTROL_CENTER_EMAIL_HOST_NAME;

  public static final String CONTROL_CENTER_EMAIL_FROM =
      CONFLUENT_CONTROLCENTER_EMAIL_PREFIX + "from";
  protected static final String CONTROL_CENTER_EMAIL_FROM_DOC =
      "The 'from' address for emails sent from Control Center.";

  public static final String CONTROL_CENTER_EMAIL_BOUNCE_ADDRESS =
      CONFLUENT_CONTROLCENTER_EMAIL_PREFIX + "bounce.address";
  protected static final String CONTROL_CENTER_EMAIL_BOUNCE_ADDRESS_DOC =
      "Override for " + CONTROL_CENTER_EMAIL_FROM + " config to send message bounce notifications.";

  public static final String CONTROL_CENTER_EMAIL_SSL_CHECKSERVERIDENTITY =
      CONFLUENT_CONTROLCENTER_EMAIL_PREFIX + "ssl.checkserveridentity";
  public static final Boolean DEFAULT_CONTROL_CENTER_EMAIL_SSL_CHECKSERVERIDENTITY = false;
  protected static final String CONTROL_CENTER_EMAIL_SSL_CHECKSERVERIDENTITY_DOC =
      "Forces validation of server's certificate when using STARTTLS or SSL.";

  public static final String CONTROL_CENTER_EMAIL_STARTTLS_REQUIRED =
      CONFLUENT_CONTROLCENTER_EMAIL_PREFIX + "starttls.required";
  public static final Boolean DEFAULT_CONTROL_CENTER_EMAIL_STARTTLS_REQUIRED = false;
  protected static final String CONTROL_CENTER_EMAIL_STARTTLS_REQUIRED_DOC =
      "Forces using STARTTLS.";

  public static final String CONTROL_CENTER_EMAIL_USERNAME =
      CONFLUENT_CONTROLCENTER_EMAIL_PREFIX + "username";
  public static final String DEFAULT_CONTROL_CENTER_EMAIL_USERNAME = "";
  protected static final String CONTROL_CENTER_EMAIL_USERNAME_DOC =
      "Username for username/pw authentication. If this value is not set there will be no authentication with your smtp server.";

  public static final String CONTROL_CENTER_EMAIL_PASSWORD =
      CONFLUENT_CONTROLCENTER_EMAIL_PREFIX + "password";
  public static final String DEFAULT_CONTROL_CENTER_EMAIL_PASSWORD = "";
  protected static final String CONTROL_CENTER_EMAIL_PASSWORD_DOC =
      "Password for username/pw authentication.";

  public static final String CONTROL_CENTER_MAX_TRIGGER_EVENTS_PER_ALERT_CONFIG =
      CONFLUENT_CONTROLCENTER_PREFIX + "alert.max.trigger.events";
  public static final int DEFAULT_CONTROL_CENTER_MAX_TRIGGER_EVENTS_PER_ALERT_CONFIG = 1000;
  protected static final String CONTROL_CENTER_MAX_TRIGGER_EVENTS_PER_ALERT_DOC =
      "The maximum number of trigger events in one alert";

  public static final String CONTROL_CENTER_COMMAND_TOPIC = CONFLUENT_CONTROLCENTER_PREFIX + "command.topic";
  public static final String DEFAULT_CONTROL_CENTER_COMMAND_TOPIC = CommandStore.DEFAULT_COMMAND_TOPIC;
  protected static final String CONTROL_CENTER_COMMAND_TOPIC_DOC =
      "Topic used for CP configuration";

  public static final String CONTROL_CENTER_COMMAND_TOPIC_REPLICATION_CONFIG =
      CONFLUENT_CONTROLCENTER_PREFIX + "command.topic.replication";
  public static final int DEFAULT_CONTROL_CENTER_COMMAND_TOPIC_REPLICATION = 3;
  protected static final String CONTROL_CENTER_COMMAND_TOPIC_REPLICATION_DOC =
      "Replication factor for command topic";

  public static final String CONTROL_CENTER_COMMAND_TOPIC_RETENTION_MS_CONFIG =
      CONFLUENT_CONTROLCENTER_PREFIX + "command.topic.retention.ms";
  public static final int DEFAULT_CONTROL_CENTER_COMMAND_TOPIC_RETENTION_MS = 259200000;
  protected static final String CONTROL_CENTER_COMMAND_TOPIC_RETENTION_MS_DOC =
      "Retention ms for command topic";

  public static final String CONTROL_CENTER_COMMAND_TOPIC_RETENTION_BYTES_CONFIG =
      CONFLUENT_CONTROLCENTER_PREFIX + "command.topic.retention.bytes";
  public static final int DEFAULT_CONTROL_CENTER_COMMAND_TOPIC_RETENTION_BYTES = -1;
  protected static final String CONTROL_CENTER_COMMAND_TOPIC_RETENTION_BYTES_DOC =
      "Retention bytes for command topic";

  private static final ConfigDef config;

  public static final String ZOOKEEPER_CONNECT_CONFIG = "zookeeper.connect";
  private static final String
      ZOOKEEPER_CONNECT_DOC =
      "Specifies the ZooKeeper connection string in the form "
      + "hostname:port where host and port are the host and port of a ZooKeeper server. To allow connecting "
      + "through other ZooKeeper nodes when that ZooKeeper machine is down you can also specify multiple hosts "
      + "in the form hostname1:port1,hostname2:port2,hostname3:port3.\n"
      + "\n"
      + "The server may also have a ZooKeeper chroot path as part of it's ZooKeeper connection string which puts "
      + "its data under some path in the global ZooKeeper namespace. If so the consumer should use the same "
      + "chroot path in its connection string. For example to give a chroot path of /chroot/path you would give "
      + "the connection string as hostname1:port1,hostname2:port2,hostname3:port3/chroot/path.";
  public static final String ZOOKEEPER_CONNECT_DEFAULT = "localhost:2181";

  static {
    config = (new ConfigDef())
        .define(
            CONTROL_CENTER_NAME_CONFIG,
            ConfigDef.Type.STRING,
            DEFAULT_CONTROL_CENTER_NAME,
            ConfigDef.Importance.HIGH,
            CONTROL_CENTER_NAME_DOC
        ).define(
            CONTROL_CENTER_ID_CONFIG,
            ConfigDef.Type.STRING,
            DEFAULT_CONTROL_CENTER_ID,
            ConfigDef.Importance.HIGH,
            CONTROL_CENTER_ID_DOC
        ).define(
            CONTROL_CENTER_PARALLELISM,
            ConfigDef.Type.INT,
            DEFAULT_CONTROL_CENTER_PARALLELISM,
            ConfigDef.Importance.HIGH,
            CONTROL_CENTER_PARALLELISM_DOC
        ).define(
            CONTROL_CENTER_PRODUCER_SECURITY_PROTOCOL,
            ConfigDef.Type.STRING,
            DEFAULT_CONTROL_CENTER_SECURITY_PROTOCOL,
            ConfigDef.Importance.HIGH,
            CONTROL_CENTER_SECURITY_PROTOCOL_DOC
        ).define(
            CONTROL_CENTER_CONSUMER_SECURITY_PROTOCOL,
            ConfigDef.Type.STRING,
            DEFAULT_CONTROL_CENTER_SECURITY_PROTOCOL,
            ConfigDef.Importance.HIGH,
            CONTROL_CENTER_SECURITY_PROTOCOL_DOC
        ).define(
            CONTROL_CENTER_TOPIC_CONFIG,
            ConfigDef.Type.STRING,
            DEFAULT_CONTROL_CENTER_TOPIC,
            ConfigDef.Importance.HIGH,
            CONTROL_CENTER_TOPIC_DOC
        ).define(
            ZOOKEEPER_CONNECT_CONFIG,
            ConfigDef.Type.STRING,
            ZOOKEEPER_CONNECT_DEFAULT,
            ConfigDef.Importance.HIGH,
            ZOOKEEPER_CONNECT_DOC
        ).define(
            CONTROL_CENTER_LICENSE_CONFIG,
            ConfigDef.Type.STRING,
            "",
            ConfigDef.Importance.HIGH,
            CONTROL_CENTER_LICENSE_DOC
        ).define(
            CONTROL_CENTER_BOOTSTRAP_SERVERS_CONFIG,
            ConfigDef.Type.LIST,
            DEFAULT_CONTROL_CENTER_BOOTSTRAP_SERVERS_CONFIG,
            ConfigDef.Importance.HIGH,
            CONTROL_CENTER_BOOTSTRAP_SERVERS_DOC
        ).define(
            CONTROL_CENTER_DATA_DIR,
            ConfigDef.Type.STRING,
            DEFAULT_CONTROL_CENTER_DATA_DIR,
            ConfigDef.Importance.HIGH,
            CONTROL_CENTER_DATA_DIR_DOC
        ).define(
            CONTROL_CENTER_INTERNAL_REPLICATION_CONFIG,
            ConfigDef.Type.INT,
            DEFAULT_CONTROL_CENTER_INTERNAL_REPLICATION,
            ConfigDef.Importance.LOW,
            CONTROL_CENTER_INTERNAL_REPLICATION_DOC
        ).define(
            CONTROL_CENTER_INTERNAL_PARTITIONS_CONFIG,
            ConfigDef.Type.INT,
            DEFAULT_CONTROL_CENTER_INTERNAL_PARTITIONS,
            ConfigDef.Importance.LOW,
            CONTROL_CENTER_INTERNAL_PARTITIONS_DOC
        ).define(
            CONTROL_CENTER_INTERNAL_RETENTION_MS_CONFIG,
            ConfigDef.Type.LONG,
            DEFAULT_CONTROL_CENTER_INTERNAL_RETENTION_MS,
            ConfigDef.Importance.LOW,
            CONTROL_CENTER_INTERNAL_RETENTION_MS_DOC
        ).define(
            CONTROL_CENTER_INTERNAL_CHANGELOG_SEGMENT_BYTES_CONFIG,
            ConfigDef.Type.LONG,
            DEFAULT_CONTROL_CENTER_INTERNAL_CHANGELOG_SEGMENT_BYTES,
            ConfigDef.Importance.LOW,
            CONTROL_CENTER_INTERNAL_CHANGELOG_SEGMENT_BYTES_DOC
        ).define(
            CONTROL_CENTER_INTERNAL_RETENTION_BYTES_CONFIG,
            ConfigDef.Type.LONG,
            DEFAULT_CONTROL_CENTER_INTERNAL_RETENTION_BYTES,
            ConfigDef.Importance.LOW,
            CONTROL_CENTER_INTERNAL_RETENTION_BYTES_DOC
        ).define(
            CONTROL_CENTER_MONITORING_TOPIC_PARTITIONS_CONFIG,
            ConfigDef.Type.INT,
            DEFAULT_CONTROL_CENTER_MONITORING_TOPIC_PARTITIONS,
            ConfigDef.Importance.LOW,
            CONTROL_CENTER_MONITORING_TOPIC_PARTITIONS_DOC
        ).define(
            CONTROL_CENTER_MONITORING_TOPIC_REPLICATION_CONFIG,
            ConfigDef.Type.INT,
            DEFAULT_CONTROL_CENTER_MONITORING_TOPIC_REPLICATION,
            ConfigDef.Importance.LOW,
            CONTROL_CENTER_MONITORING_TOPIC_REPLICATION_DOC
        ).define(
            CONTROL_CENTER_MONITORING_TOPIC_RETENTION_MS_CONFIG,
            ConfigDef.Type.LONG,
            DEFAULT_CONTROL_CENTER_MONITORING_TOPIC_RETENTION_MS,
            ConfigDef.Importance.LOW,
            CONTROL_CENTER_MONITORING_TOPIC_RETENTION_MS_DOC
        ).define(
            CONTROL_CENTER_MONITORING_TOPIC_RETENTION_BYTES_CONFIG,
            ConfigDef.Type.LONG,
            DEFAULT_CONTROL_CENTER_MONITORING_TOPIC_RETENTION_BYTES,
            ConfigDef.Importance.LOW,
            CONTROL_CENTER_MONITORING_TOPIC_RETENTION_BYTES_DOC
        ).define(
            CONTROL_CENTER_CONNECT_CLUSTER,
            ConfigDef.Type.LIST,
            DEFAULT_CONTROL_CENTER_CONNECT_CLUSTER,
            ConfigDef.Importance.HIGH,
            CONTROL_CENTER_CONNECT_CLUSTER_DOC
        ).define(
            CONTROL_CENTER_CONNECT_ZOOKEEPER_CONNECT,
            ConfigDef.Type.STRING,
            "",
            ConfigDef.Importance.HIGH,
            CONTROL_CENTER_CONNECT_ZOOKEEPER_CONNECT_DOC
        ).define(
            CONTROL_CENTER_CONNECT_TIMEOUT,
            ConfigDef.Type.LONG,
            DEFAULT_CONTROL_CENTER_CONNECT_TIMEOUT,
            ConfigDef.Importance.LOW,
            CONTROL_CENTER_CONNECT_TIMEOUT_DOC
        ).define(
            CONTROL_CENTER_MAX_POLL_RECORDS_CONFIG,
            ConfigDef.Type.INT,
            DEFAULT_CONTROL_CENTER_MAX_POLL_RECORDS_CONFIG,
            ConfigDef.Importance.MEDIUM,
            CONTROL_CENTER_MAX_POLL_RECORDS_DOC
        ).define(
            CONTROL_CENTER_SESSION_TIMEOUT_MS_CONFIG,
            ConfigDef.Type.INT,
            DEFAULT_CONTROL_CENTER_SESSION_TIMEOUT_MS_CONFIG,
            ConfigDef.Importance.MEDIUM,
            CONTROL_CENTER_SESSION_TIMEOUT_MS_DOC
        ).define(
            CONTROL_CENTER_REQUEST_TIMEOUT_MS_CONFIG,
            ConfigDef.Type.INT,
            DEFAULT_CONTROL_CENTER_REQUEST_TIMEOUT_MS_CONFIG,
            ConfigDef.Importance.MEDIUM,
            CONTROL_CENTER_REQUEST_TIMEOUT_MS_DOC
        ).define(
            CONTROL_CENTER_RETRIES_CONFIG,
            ConfigDef.Type.INT,
            DEFAULT_CONTROL_CENTER_RETRIES_CONFIG,
            ConfigDef.Importance.MEDIUM,
            CONTROL_CENTER_RETRIES_DOC
        ).define(
            CONTROL_CENTER_RETRY_BACKOFF_MS_CONFIG,
            ConfigDef.Type.LONG,
            DEFAULT_CONTROL_CENTER_RETRY_BACKOFF_MS_CONFIG,
            ConfigDef.Importance.MEDIUM,
            CONTROL_CENTER_RETRY_BACKOFF_MS_DOC
        ).define(
            CONTROL_CENTER_COMPRESSION_TYPE_CONFIG,
            ConfigDef.Type.STRING,
            DEFAULT_CONTROL_CENTER_COMPRESSION_TYPE_CONFIG,
            ConfigDef.Importance.LOW,
            CONTROL_CENTER_COMPRESSION_TYPE_DOC
        ).define(
            CONTROL_CENTER_REST_PORT,
            ConfigDef.Type.INT,
            DEFAULT_CONTROL_CENTER_REST_PORT,
            ConfigDef.Importance.LOW,
            CONTROL_CENTER_REST_PORT_DOC
        ).define(
            CONTROL_CENTER_DISPLAY_HOST_PREFIX,
            ConfigDef.Type.STRING,
            DEFAULT_CONTROL_CENTER_DISPLAY_HOST_PREFIX,
            ConfigDef.Importance.LOW,
            CONTROL_CENTER_DISPLAY_HOST_PREFIX_DOC
        ).define(
            CONTROL_CENTER_EMAIL_ENABLE,
            ConfigDef.Type.BOOLEAN,
            DEFAULT_CONTROL_CENTER_EMAIL_ENABLE,
            ConfigDef.Importance.LOW,
            CONTROL_CENTER_EMAIL_ENABLE_DOC
        ).define(
            CONTROL_CENTER_EMAIL_HOST_NAME,
            ConfigDef.Type.STRING,
            DEFAULT_CONTROL_CENTER_EMAIL_HOST_NAME,
            ConfigDef.Importance.LOW,
            CONTROL_CENTER_EMAIL_HOST_NAME_DOC
        ).define(
            CONTROL_CENTER_EMAIL_PORT,
            ConfigDef.Type.INT,
            DEFAULT_CONTROL_CENTER_EMAIL_PORT,
            ConfigDef.Importance.LOW,
            CONTROL_CENTER_EMAIL_PORT_DOC
        ).define(
            CONTROL_CENTER_EMAIL_FROM,
            ConfigDef.Type.STRING,
            "c3@confluent.io",
            ConfigDef.Importance.LOW,
            CONTROL_CENTER_EMAIL_FROM_DOC
        ).define(
            CONTROL_CENTER_EMAIL_BOUNCE_ADDRESS,
            ConfigDef.Type.STRING,
            "",
            ConfigDef.Importance.LOW,
            CONTROL_CENTER_EMAIL_BOUNCE_ADDRESS_DOC
        ).define(
            CONTROL_CENTER_EMAIL_SSL_CHECKSERVERIDENTITY,
            ConfigDef.Type.BOOLEAN,
            DEFAULT_CONTROL_CENTER_EMAIL_SSL_CHECKSERVERIDENTITY,
            ConfigDef.Importance.LOW,
            CONTROL_CENTER_EMAIL_SSL_CHECKSERVERIDENTITY_DOC
        ).define(
            CONTROL_CENTER_EMAIL_STARTTLS_REQUIRED,
            ConfigDef.Type.BOOLEAN,
            DEFAULT_CONTROL_CENTER_EMAIL_STARTTLS_REQUIRED,
            ConfigDef.Importance.LOW,
            CONTROL_CENTER_EMAIL_STARTTLS_REQUIRED_DOC
        ).define(
            CONTROL_CENTER_EMAIL_USERNAME,
            ConfigDef.Type.STRING,
            DEFAULT_CONTROL_CENTER_EMAIL_USERNAME,
            ConfigDef.Importance.LOW,
            CONTROL_CENTER_EMAIL_USERNAME_DOC
        ).define(
            CONTROL_CENTER_EMAIL_PASSWORD,
            ConfigDef.Type.STRING,
            DEFAULT_CONTROL_CENTER_EMAIL_PASSWORD,
            ConfigDef.Importance.LOW,
            CONTROL_CENTER_EMAIL_PASSWORD_DOC
        ).define(
            CONTROL_CENTER_MAX_TRIGGER_EVENTS_PER_ALERT_CONFIG,
            ConfigDef.Type.INT,
            DEFAULT_CONTROL_CENTER_MAX_TRIGGER_EVENTS_PER_ALERT_CONFIG,
            ConfigDef.Importance.LOW,
            CONTROL_CENTER_MAX_TRIGGER_EVENTS_PER_ALERT_DOC
        ).define(
            CONTROL_CENTER_COMMAND_TOPIC,
            ConfigDef.Type.STRING,
            DEFAULT_CONTROL_CENTER_COMMAND_TOPIC,
            ConfigDef.Importance.LOW,
            CONTROL_CENTER_COMMAND_TOPIC_DOC
        ).define(
            CONTROL_CENTER_COMMAND_TOPIC_REPLICATION_CONFIG,
            ConfigDef.Type.INT,
            DEFAULT_CONTROL_CENTER_COMMAND_TOPIC_REPLICATION,
            ConfigDef.Importance.LOW,
            CONTROL_CENTER_COMMAND_TOPIC_REPLICATION_DOC
        ).define(
            CONTROL_CENTER_COMMAND_TOPIC_RETENTION_MS_CONFIG,
            ConfigDef.Type.LONG,
            DEFAULT_CONTROL_CENTER_COMMAND_TOPIC_RETENTION_MS,
            ConfigDef.Importance.LOW,
            CONTROL_CENTER_COMMAND_TOPIC_RETENTION_MS_DOC
        ).define(
            CONTROL_CENTER_COMMAND_TOPIC_RETENTION_BYTES_CONFIG,
            ConfigDef.Type.LONG,
            DEFAULT_CONTROL_CENTER_COMMAND_TOPIC_RETENTION_BYTES,
            ConfigDef.Importance.LOW,
            CONTROL_CENTER_COMMAND_TOPIC_RETENTION_BYTES_DOC
        );
  }

  public ControlCenterConfig(String propsFile) {
    this(getPropsFromFile(propsFile));
  }

  public ControlCenterConfig(Properties props) {
    super(config, props);
  }

  public ControlCenterConfig() {
    this(new Properties());
  }

  public String getConnectZookeeperConnect() {
    String zkConnect = getString(CONTROL_CENTER_CONNECT_ZOOKEEPER_CONNECT);
    if (zkConnect.isEmpty()) {
      zkConnect = getString(ZOOKEEPER_CONNECT_CONFIG);
    }
    return zkConnect;
  }

  public static void main(String[] args) {
    System.out.println(config.toRst());
  }
}
