package io.confluent.controlcenter.connect;

import com.google.common.base.Function;
import com.google.common.base.Supplier;
import com.google.common.collect.ImmutableList;
import com.google.common.collect.Lists;
import com.google.inject.Inject;
import com.google.inject.Singleton;

import com.fasterxml.jackson.annotation.JsonProperty;

import org.apache.kafka.connect.runtime.rest.entities.ConnectorInfo;
import org.apache.kafka.connect.runtime.rest.entities.ConnectorPluginInfo;
import org.apache.kafka.connect.runtime.rest.entities.CreateConnectorRequest;
import org.apache.kafka.connect.util.ConnectorTaskId;
import org.glassfish.jersey.client.ClientProperties;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.List;
import java.util.Map;
import java.util.Random;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.TimeoutException;

import javax.ws.rs.client.Client;
import javax.ws.rs.client.ClientBuilder;
import javax.ws.rs.client.Entity;
import javax.ws.rs.client.WebTarget;
import javax.ws.rs.core.GenericType;
import javax.ws.rs.core.MediaType;
import javax.ws.rs.core.Response;

import io.confluent.controlcenter.ControlCenterConfig;
import io.confluent.controlcenter.rest.res.JsonStandard;

@JsonStandard
class StatusConnector {
  public String state;
  public String worker_id;
  public String trace;
}

@JsonStandard
class ConnectorStatus {
  public StatusConnector connector;
  public String name;
  public List<Object> tasks;
}

@Singleton
public class ConnectService {

  private static final Logger log = LoggerFactory.getLogger(ConnectService.class);

  public static class ConnectInit implements Runnable {

    private final ConnectService connectService;

    @Inject
    public ConnectInit(ConnectService connectService) {
      this.connectService = connectService;
    }

    @Override
    public void run() {
      try {
        this.connectService.getConnectorPlugins();
        this.connectService.getConnectors();
      } catch (Throwable t) {
        log.warn("failed to connect to kafka-connect {}", t.getMessage());
      }
    }
  }

  private final Supplier<WebTarget> connect;
  private final long timeout;

  @Inject
  public ConnectService(ControlCenterConfig config) {

    timeout = config.getLong(ControlCenterConfig.CONTROL_CENTER_CONNECT_TIMEOUT);

    final Client client = ClientBuilder.newClient();
    final List<WebTarget> connectTargets = ImmutableList.copyOf(
        Lists.transform(
            config.getList(ControlCenterConfig.CONTROL_CENTER_CONNECT_CLUSTER),
            new Function<String, WebTarget>() {
              @Override
              public WebTarget apply(String url) {
                return client.target(url.startsWith("http://") ? url : "http://" + url);
              }
            }
        )
    );

    if (connectTargets.isEmpty()) {
      log.warn(ControlCenterConfig.CONTROL_CENTER_CONNECT_CLUSTER + " configuration is empty.");
      connect = null;
      return;
    }

    final Random random = new Random();

    // pick a server at random among those provided
    connect = new Supplier<WebTarget>() {
      @Override
      public WebTarget get() {
        return connectTargets.get(random.nextInt(connectTargets.size()));
      }
    };
  }

  public enum ConnectorType {
    SOURCE, SINK;

    public static ConnectorType fromString(String s) {
      if (s != null) {
        for (ConnectorType type : values()) {
          if (s.toLowerCase().contains(type.name().toLowerCase())) {
            return type;
          }
        }
      }
      return null;
    }
  }

  @JsonStandard
  public static class TypedConnectorPluginInfo extends ConnectorPluginInfo {
    public TypedConnectorPluginInfo(@JsonProperty("class") String clazz) {
      super(clazz);
    }

    public ConnectorType getType() {
      return ConnectorType.fromString(clazz());
    }
  }


  public List<TypedConnectorPluginInfo> getConnectorPlugins() throws InterruptedException, ExecutionException, TimeoutException {
    return connect.get()
        .path("connector-plugins")
        .request()
        .accept(MediaType.APPLICATION_JSON_TYPE)
        .async()
        .get(new GenericType<List<TypedConnectorPluginInfo>>() {
        }).get(timeout, TimeUnit.MILLISECONDS);
  }

  public Response validateConnectorPlugin(String connType, Map<String, String> connectorConfig) throws InterruptedException, ExecutionException, TimeoutException {
    return connect.get()
        .path("connector-plugins")
        .path(connType)
        .path("config")
        .path("validate")
        .request()
        .async()
        .put(Entity.entity(connectorConfig, MediaType.APPLICATION_JSON_TYPE))
        .get(timeout, TimeUnit.MILLISECONDS);
  }


  @JsonStandard
  public static class Connector {
    public String name;
    public ConnectorType type;
    public int tasks;
    public String status;

    @JsonStandard
    public String topics;

    public Connector() {}

    public Connector(ConnectorInfo connectorInfo, ConnectorStatus connectorStatus) {
      name = connectorInfo.name();
      type = ConnectorType.fromString(connectorInfo.config().get("connector.class"));
      tasks = connectorInfo.tasks().size();
      topics = connectorInfo.config().get("topics");
      status = connectorStatus.connector.state;
    }
  }

  public List<Connector> getConnectors() throws InterruptedException, ExecutionException, TimeoutException {
    List<String> connectors = connect.get()
        .path("connectors")
        .request()
        .accept(MediaType.APPLICATION_JSON_TYPE)
        .async()
        .get(new GenericType<List<String>>() {
        }).get(timeout, TimeUnit.MILLISECONDS);
    List<Connector> out = Lists.newArrayList();
    for (String connector : connectors) {
      out.add(
          new Connector(
              connect.get()
                  .path("connectors")
                  .path(connector)
                  .request()
                  .accept(MediaType.APPLICATION_JSON_TYPE)
                  .async()
                  .get(new GenericType<ConnectorInfo>() {
                  }).get(timeout, TimeUnit.MILLISECONDS),
              connect.get()
                  .path("connectors")
                  .path(connector)
                  .path("status")
                  .request()
                  .accept(MediaType.APPLICATION_JSON_TYPE)
                  .async()
                  .get(new GenericType<ConnectorStatus>() {
                  }).get(timeout, TimeUnit.MILLISECONDS)
          )
      );
    }
    return out;
  }


  public Object createConnector(CreateConnectorRequest createRequest) throws InterruptedException, ExecutionException, TimeoutException {
    return connect.get()
        .path("connectors")
        .request()
        .async()
        .post(Entity.entity(createRequest, MediaType.APPLICATION_JSON_TYPE))
        .get(timeout, TimeUnit.MILLISECONDS);
  }

  @JsonStandard
  public static class TypedConnectorInfo extends ConnectorInfo {

    public TypedConnectorInfo(
        @JsonProperty("name") String name,
        @JsonProperty("config") Map<String, String> config,
        @JsonProperty("tasks") List<ConnectorTaskId> tasks
    ) {
      super(name, config, tasks);
    }

    public ConnectorType getType() {
      return ConnectorType.fromString(config().get("connector.class"));
    }
  }

  public TypedConnectorInfo getConnector(String connector) {
    return connect.get()
        .path("connectors")
        .path(connector)
        .request()
        .get(new GenericType<TypedConnectorInfo>() {});
  }

  public Object getConnectorStatus(String connector) throws InterruptedException, ExecutionException, TimeoutException {
    return connect.get()
        .path("connectors")
        .path(connector)
        .path("status")
        .request()
        .async()
        .get()
        .get(timeout, TimeUnit.MILLISECONDS);
  }

  public Object getTaskConfigs(String connector) throws InterruptedException, ExecutionException, TimeoutException {
    return connect.get()
        .path("connectors")
        .path(connector)
        .path("tasks")
        .request()
        .async()
        .get()
        .get(timeout, TimeUnit.MILLISECONDS);
  }

  public Object getConnectorConfig(String connector) throws InterruptedException, ExecutionException, TimeoutException {
    return connect.get()
        .path("connectors")
        .path(connector)
        .path("config")
        .request()
        .async()
        .get()
        .get(timeout, TimeUnit.MILLISECONDS);
  }

  public Object putConnectorConfig(String connector, Map<String, String> connectorConfig) throws InterruptedException, ExecutionException, TimeoutException {
    return connect.get()
        .path("connectors")
        .path(connector)
        .path("config")
        .request()
        .async()
        .put(Entity.entity(connectorConfig, MediaType.APPLICATION_JSON_TYPE))
        .get(timeout, TimeUnit.MILLISECONDS);
  }

  public Object putConnectorPause(String connector) throws InterruptedException, ExecutionException, TimeoutException {
    return connect.get()
        .path("connectors")
        .path(connector)
        .path("pause")
        .request()
        .property(ClientProperties.SUPPRESS_HTTP_COMPLIANCE_VALIDATION, true)
        .async()
        .put(null)
        .get(timeout, TimeUnit.MILLISECONDS);
  }

  public Object putConnectorResume(String connector) throws InterruptedException, ExecutionException, TimeoutException {
    return connect.get()
        .path("connectors")
        .path(connector)
        .path("resume")
        .request()
        .property(ClientProperties.SUPPRESS_HTTP_COMPLIANCE_VALIDATION, true)
        .async()
        .put(null)
        .get(timeout, TimeUnit.MILLISECONDS);
  }

  public Object putTaskConfigs(String connector, List<Map<String, String>> taskConfigs) throws InterruptedException, ExecutionException, TimeoutException {
    return connect.get()
        .path("connectors")
        .path(connector)
        .path("tasks")
        .request()
        .async()
        .put(Entity.entity(taskConfigs, MediaType.APPLICATION_JSON_TYPE))
        .get(timeout, TimeUnit.MILLISECONDS);
  }

  public Object getTaskStatus(String connector, Integer task) throws InterruptedException, ExecutionException, TimeoutException {
    return connect.get()
        .path("connectors")
        .path(connector)
        .path("tasks")
        .path(task.toString())
        .request()
        .async()
        .get()
        .get(timeout, TimeUnit.MILLISECONDS);
  }

  public Object destroyConnector(String connector) throws InterruptedException, ExecutionException, TimeoutException {
    return connect.get()
        .path("connectors")
        .path(connector)
        .request()
        .async()
        .delete()
        .get(timeout, TimeUnit.MILLISECONDS);
  }
}
