// (Copyright) [2016 - 2016] Confluent, Inc.
package io.confluent.controlcenter.data;

import com.google.inject.Inject;
import com.google.inject.Provider;
import com.google.protobuf.Descriptors.FieldDescriptor;
import io.confluent.command.CommandStore;
import io.confluent.command.cluster_metadata.CommandClusterMetadata.ClusterMetadata;
import io.confluent.command.cluster_metadata.CommandClusterMetadata.KafkaClusterMetadata;
import io.confluent.command.cluster_metadata.CommandClusterMetadata.KafkaConnectClusterMetadata;
import io.confluent.command.record.Command.CommandConfigType;
import io.confluent.command.record.Command.CommandKey;
import io.confluent.command.record.Command.CommandMessage;
import io.confluent.controlcenter.command.CommandModule.ClusterMetadataStore;
import io.confluent.controlcenter.record.Controlcenter.Cluster;
import io.confluent.controlcenter.record.Controlcenter.ClusterContext;
import io.confluent.controlcenter.record.Controlcenter.ClusterType;
import io.confluent.controlcenter.rest.res.ConnectCluster;
import io.confluent.controlcenter.rest.res.KafkaCluster;
import io.confluent.controlcenter.rest.res.KafkaClusterDisplay;
import io.confluent.controlcenter.rest.res.KafkaClusterListResponse;
import io.confluent.controlcenter.serialization.OrderedKeyPrefixedSerdeSupplier;
import io.confluent.controlcenter.streams.TopicStoreModule.ClusterStore;
import io.confluent.controlcenter.util.KVQuery;
import io.confluent.serializers.OrderedKeyPrefixedSerde;
import org.apache.kafka.common.utils.Bytes;
import org.apache.kafka.streams.KeyValue;
import org.apache.kafka.streams.state.KeyValueIterator;
import org.apache.kafka.streams.state.ReadOnlyKeyValueStore;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

public class ClusterMetadataDao {
  private static final String KAFKA_CLUSTER_DEFAULT_DISPLAY_ID = "default";
  private static final String KAFKA_CLUSTER_DEFAULT_DISPLAY_NAME = "Unknown";
  private static final FieldDescriptor kafkaField = ClusterMetadata.getDescriptor().findFieldByNumber(ClusterMetadata.KAFKA_FIELD_NUMBER);
  private static final FieldDescriptor connectField = ClusterMetadata.getDescriptor().findFieldByNumber(ClusterMetadata.CONNECT_FIELD_NUMBER);
  private static final Cluster KAFKA_CLUSTER_SCAN_KEY = Cluster.newBuilder()
      .setClusterType(ClusterType.KAFKA)
      .build();
  private final CommandKey.Builder cmdKeyBuilder = CommandKey.newBuilder();
  private final ReadOnlyKeyValueStore<CommandKey, CommandMessage> clusterMetadataStore;
  private final Provider<ReadOnlyKeyValueStore<Bytes, Cluster>> clusterStore;
  private final Bytes kafkaScanPrefix;
  private final CommandStore commandStore;

  @Inject
  public ClusterMetadataDao(
      @ClusterMetadataStore ReadOnlyKeyValueStore<CommandKey, CommandMessage> clusterMetadataStore,
      @ClusterStore Provider<ReadOnlyKeyValueStore<Bytes, Cluster>> clusterStore,
      OrderedKeyPrefixedSerdeSupplier<ClusterContext, Cluster> keySerdeSupplier,
      CommandStore commandStore
  ) {
    this.clusterMetadataStore = clusterMetadataStore;
    this.clusterStore = clusterStore;
    this.kafkaScanPrefix = getKafkaScanPrefix(keySerdeSupplier);
    this.commandStore = commandStore;
  }

  private Bytes getKafkaScanPrefix(OrderedKeyPrefixedSerdeSupplier<ClusterContext, Cluster> keySerdeSupplier) {
    OrderedKeyPrefixedSerde<ClusterContext, Cluster> monitoringClusterPrefixKeySerde = keySerdeSupplier.get(ClusterContext.STREAM_MONITORING).prefixKeySerde(1);
    return monitoringClusterPrefixKeySerde.key(KAFKA_CLUSTER_SCAN_KEY);
  }

  /**
   * Returns the set of all Kafka clusters that we know about, including
   *    - what we've seen in stream-monitoring messages
   *    - what we've seen in for Kafka Connect
   *
   *    Ids for clusters without ids are stored as empty string in RocksDB
   *    but returned as 'default' from this API
   */
  public List<KafkaCluster> getKafkaClusters() {
    List<KafkaCluster> rawClusters = getAllKafkaClusters();
    List<KafkaCluster> displayClusters = new ArrayList<>();
    for (KafkaCluster cluster : rawClusters) {
      displayClusters.add(new KafkaCluster(getExternalKafkaId(cluster.clusterId), cluster.displayName, cluster.zookeeperConnect, cluster.bootstrapServers));
    }
    return displayClusters;
  }

  private List<KafkaCluster> getAllKafkaClusters() {
    Map<String, Cluster> clustersFromMon = new HashMap<>();
    for (final Cluster cluster : getRawKafkaClustersForStreamMonitoring()) {
      clustersFromMon.put(cluster.getClusterId(), cluster);
    }

    final List<KafkaCluster> clusters = new ArrayList<>();
    Set<String> clusterIdsFromMetadata = new HashSet<>();
    try (final KeyValueIterator<CommandKey, CommandMessage> iter = clusterMetadataStore.all()) {
      while (iter.hasNext()) {
        final KeyValue<CommandKey, CommandMessage> entry = iter.next();
        final ClusterMetadata clusterMetadata = entry.value.getClusterMetadata();
        if (clusterMetadata.hasField(kafkaField)) {
          final String clusterId = entry.key.getGuid();
          final KafkaClusterMetadata kafkaMeta = clusterMetadata.getKafka();
          clusters.add(new KafkaCluster(
              clusterId,
              getKafkaClusterName(null, clusterMetadata),
              kafkaMeta.getZookeeperConnectList(),
              kafkaMeta.getBootstrapServersList()
          ));
          clusterIdsFromMetadata.add(clusterId);
        }
      }
    }

    //Add the remaining clusters
    clustersFromMon.keySet().removeAll(clusterIdsFromMetadata);
    for (String clusterId : clustersFromMon.keySet()) {
      clusters.add(new KafkaCluster(clusterId, getKafkaClusterName(clustersFromMon.get(clusterId), null)));
    }

    return clusters;
  }

  public List<ConnectCluster> getConnectClusters() {
    final List<ConnectCluster> clusters = new ArrayList<>();
    try (final KeyValueIterator<CommandKey, CommandMessage> iter = clusterMetadataStore.all()) {
      while (iter.hasNext()) {
        final KeyValue<CommandKey, CommandMessage> entry = iter.next();
        final ClusterMetadata clusterMetadata = entry.value.getClusterMetadata();
        if (clusterMetadata.hasField(connectField)) {
          final KafkaConnectClusterMetadata connectMetadata = clusterMetadata.getConnect();
          clusters.add(new ConnectCluster(
              connectMetadata.getRestUrlsList(),
              connectMetadata.getKafkaClusterId(),
              clusterMetadata.getDisplayName()
          ));
        }
      }
    }
    return clusters;
  }

  public KafkaClusterListResponse getKafkaClustersForStreamMonitoring() {
    final List<KafkaClusterDisplay> clusters = new ArrayList<>();
    for (final Cluster cluster : getRawKafkaClustersForStreamMonitoring()) {
      clusters.add(getKafkaClusterDisplay(cluster.getClusterId(), cluster));
    }

    String defaultSelectedClusterId = null;
    if (clusters.size() > 0) {
      defaultSelectedClusterId = clusters.get(0).clusterId;
    }
    return new KafkaClusterListResponse(defaultSelectedClusterId, clusters);
  }

  private List<Cluster> getRawKafkaClustersForStreamMonitoring() {
    final List<Cluster> clusters = new ArrayList<>();
    try (final KeyValueIterator<Bytes, Cluster> iter = KVQuery.prefixScan(clusterStore.get(), kafkaScanPrefix)) {
      while (iter.hasNext()) {
        clusters.add(iter.next().value);
      }
    }
    return clusters;
  }

  private KafkaClusterDisplay getKafkaClusterDisplay(final String clusterId, final Cluster cluster) {
    final CommandKey cmdKey = getKafkaCommandKey(clusterId);
    final CommandMessage cmdMsg = clusterMetadataStore.get(cmdKey);
    final ClusterMetadata clusterMetadata = cmdMsg != null? cmdMsg.getClusterMetadata() : null;
    return new KafkaClusterDisplay(
        getExternalKafkaId(clusterId),
        getKafkaDisplayName(clusterId, clusterMetadata, cluster)
    );
  }

  private CommandKey getKafkaCommandKey(String clusterId) {
    cmdKeyBuilder.clear();
    return cmdKeyBuilder
        .setConfigType(CommandConfigType.CLUSTER_METADATA)
        .setGuid(clusterId)
        .build();
  }

  public KafkaClusterListResponse getKafkaClustersForConnect() {
    final List<ConnectCluster> connectClusters = getConnectClusters();
    final Set<String> kafkaIds = new HashSet<>();
    for (final ConnectCluster connectCluster : connectClusters) {
      kafkaIds.add(connectCluster.kafkaClusterId);
    }
    final List<KafkaClusterDisplay> kafkaClusters = new ArrayList<>();
    for (final String kafkaClusterId : kafkaIds) {
      kafkaClusters.add(getKafkaClusterDisplay(kafkaClusterId, null));
    }

    String defaultDisplayId = null;
    if (kafkaClusters.size() > 0) {
      defaultDisplayId = kafkaClusters.get(0).clusterId;
    }

    return new KafkaClusterListResponse(
        defaultDisplayId,
        kafkaClusters
    );
  }

  /**
   * Missing cluster ids are represented internally as empty strings, externally as 'default'
   */
  public static String getExternalKafkaId(String internalClusterId) {
    if (!internalClusterId.isEmpty()) {
      return internalClusterId;
    }
    return KAFKA_CLUSTER_DEFAULT_DISPLAY_ID;
  }

  public static String getInternalKafkaId(String externalClusterId) {
    if (externalClusterId.equals(KAFKA_CLUSTER_DEFAULT_DISPLAY_ID)) {
      return "";
    }
    return externalClusterId;
  }

  private String getKafkaClusterName(Cluster cluster, ClusterMetadata clusterMetadata) {
    if (clusterMetadata != null && !clusterMetadata.getDisplayName().isEmpty()) {
      return clusterMetadata.getDisplayName();
    }
    if (cluster != null && !cluster.getClusterName().isEmpty()) {
      return cluster.getClusterName();
    }
    return "";
  }

  private String getKafkaDisplayName(final String clusterId, final ClusterMetadata clusterMetadata, final Cluster cluster) {
    final String clusterName = getKafkaClusterName(cluster, clusterMetadata);
    if (!clusterName.isEmpty()) {
      return clusterName;
    }
    if (!clusterId.isEmpty()) {
      return clusterId;
    }
    return KAFKA_CLUSTER_DEFAULT_DISPLAY_NAME;
  }

  public void updateKafkaCluster(final String externalClusterId, final KafkaCluster cluster, final long timeoutMs) throws Exception {
    final String clusterId = getInternalKafkaId(externalClusterId);
    CommandKey key = getKafkaCommandKey(clusterId);
    CommandMessage.Builder msgBuilder = getCmdMsgBuilder(key);
    updateClusterFields(msgBuilder, cluster);
    commandStore.awaitCommand(key, msgBuilder.build(), timeoutMs);
  }

  private CommandMessage.Builder getCmdMsgBuilder(CommandKey key) {
    CommandMessage msg = clusterMetadataStore.get(key);
    if (msg != null) {
      return CommandMessage.newBuilder(msg);
    }
    return CommandMessage.newBuilder();
  }

  public void updateClusterFields(CommandMessage.Builder msgBuilder, KafkaCluster cluster) {
    ClusterMetadata.Builder clusterMetadataBuilder = msgBuilder.getClusterMetadataBuilder();
    KafkaClusterMetadata.Builder kafkaBuilder = clusterMetadataBuilder.getKafkaBuilder();
    if (cluster.displayName != null) {
      clusterMetadataBuilder.setDisplayName(cluster.displayName);
    }
    if (cluster.zookeeperConnect != null) {
      kafkaBuilder.clearZookeeperConnect();
      kafkaBuilder.addAllZookeeperConnect(cluster.zookeeperConnect);
    }
    if (cluster.bootstrapServers != null) {
      kafkaBuilder.clearBootstrapServers();
      kafkaBuilder.addAllBootstrapServers(cluster.bootstrapServers);
    }
  }

}
