// (Copyright) [2016 - 2016] Confluent, Inc.
package io.confluent.controlcenter.rest;

import com.google.common.base.Strings;
import com.google.common.collect.Lists;
import com.google.common.collect.Maps;
import com.google.common.collect.Sets;
import com.google.inject.Inject;
import com.google.inject.Provider;
import io.confluent.controlcenter.Rollup;
import io.confluent.controlcenter.data.ClusterMetadataDao;
import io.confluent.controlcenter.keys.Keys.KeyType;
import io.confluent.controlcenter.record.Controlcenter.MemberInfo;
import io.confluent.controlcenter.record.Controlcenter.TopicPartition;
import io.confluent.controlcenter.record.Controlcenter.WindowedGrouping;
import io.confluent.controlcenter.rest.req.RangeRequest;
import io.confluent.controlcenter.rest.res.DeliveryResponse;
import io.confluent.controlcenter.rest.res.DeliveryResponse.DeliveryType;
import io.confluent.controlcenter.rest.res.DeliveryResponse.MemberResponse;
import io.confluent.controlcenter.serialization.OrderedKeyPrefixedSerdeSupplier;
import io.confluent.controlcenter.streams.TopicStoreModule.GroupStore;
import io.confluent.controlcenter.streams.TopicStoreModule.MonitoringMessageAggregatorWindowsStore;
import io.confluent.controlcenter.streams.TopicStoreModule.MonitoringStreamStore;
import io.confluent.license.License;
import io.confluent.monitoring.record.Monitoring.ClientType;
import io.confluent.monitoring.record.Monitoring.MonitoringMessage;
import org.apache.kafka.common.utils.Bytes;
import org.apache.kafka.streams.state.ReadOnlyWindowStore;
import org.apache.kafka.streams.state.WindowStoreIterator;
import org.jose4j.jwt.JwtClaims;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.ws.rs.BadRequestException;
import javax.ws.rs.ClientErrorException;
import javax.ws.rs.DefaultValue;
import javax.ws.rs.GET;
import javax.ws.rs.Path;
import javax.ws.rs.PathParam;
import javax.ws.rs.Produces;
import javax.ws.rs.QueryParam;
import javax.ws.rs.core.MediaType;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.Set;

@Path("/2.0/monitoring/{clusterId}")
@Produces(MediaType.APPLICATION_JSON)
public class MessageDeliveryResource {

  private static final Logger log = LoggerFactory.getLogger(MessageDeliveryResource.class);

  private static String validateClusterId(String clusterId) {
    if (clusterId == null) {
      throw new BadRequestException("missing clusterId");
    }
    return ClusterMetadataDao.getInternalKafkaId(clusterId);
  }

  private final OrderedKeyPrefixedSerdeSupplier<KeyType, MonitoringMessage> keySerdeSupplier;

  private Provider<Map<Rollup, ReadOnlyWindowStore<Bytes, WindowedGrouping>>> groupDb;
  private Provider<Map<Rollup, ReadOnlyWindowStore<Bytes, MonitoringMessage>>> expectedDb;
  private Provider<Map<Rollup, ReadOnlyWindowStore<Bytes, MonitoringMessage>>> actualDb;
  private Long expiration;

  private enum AggregationType {
    ConsumerGroup,
    Topic
  }

  @Inject
  public MessageDeliveryResource(
      @GroupStore Provider<Map<Rollup, ReadOnlyWindowStore<Bytes, WindowedGrouping>>> groupDb,
      @MonitoringMessageAggregatorWindowsStore Provider<Map<Rollup, ReadOnlyWindowStore<Bytes, MonitoringMessage>>> expectedDb,
      @MonitoringStreamStore Provider<Map<Rollup, ReadOnlyWindowStore<Bytes, MonitoringMessage>>> actualDb,
      OrderedKeyPrefixedSerdeSupplier<KeyType, MonitoringMessage> keySerdeSupplier,
      JwtClaims claims
  ) {
    this.groupDb = groupDb;
    this.expectedDb = expectedDb;
    this.actualDb = actualDb;
    this.keySerdeSupplier = keySerdeSupplier;
    this.expiration = License.getExpiration(claims);
  }

  private void checkParams(
      int limit,
      Rollup rollup,
      Long startTimeMs,
      Long stopTimeMs
  ) throws BadRequestException {
    if (System.currentTimeMillis() > expiration) {
      throw new ClientErrorException("License expired", 402);
    }
    if (limit > 6000) {
      throw new BadRequestException("limit must be < 6000");
    }
    if (startTimeMs == null) {
      throw new BadRequestException("missing startTimeMs");
    }
    if (stopTimeMs == null) {
      throw new BadRequestException("missing stopTimeMs");
    }
    if (rollup == null) {
      throw new BadRequestException("missing rollup");
    }
  }

  private WindowedGrouping getAggregateGroupInfo(
      RangeRequest rr,
      int limit,
      KeyType keyType
  ) {
    try(WindowStoreIterator<WindowedGrouping> groupInfos = groupDb.get()
        .get(rr.rollup)
        .fetch(
            keySerdeSupplier.get(keyType).key(rr.createStartKeySpec()),
            rr.getRollupAlignedStartTimeMs(),
            rr.getRollupAlignedStopTimeMs()
        )) {
      Set<MemberInfo> memberInfos = Sets.newHashSet();
      while (groupInfos.hasNext() && limit-- > 0) {
        WindowedGrouping wg = groupInfos.next().value;
        memberInfos.addAll(wg.getMembersList());
      }

      WindowedGrouping aggregate = WindowedGrouping
          .newBuilder()
          .addAllMembers(memberInfos)
          .setWindow(0)
          .build();
      return aggregate;
    }
  }

  private Map<Metric, ArrayList<Object>> getConsumedMessages(
      RangeRequest rr,
      int limit,
      long stepMs,
      KeyType keyType
  ) {
    rr.clientType = ClientType.CONSUMER;
    try (WindowStoreIterator<MonitoringMessage> expectedMessages = actualDb.get()
        .get(rr.rollup)
        .fetch(
            keySerdeSupplier.get(keyType).key(rr.createStartKeySpec()),
            rr.getRollupAlignedStartTimeMs(),
            rr.getRollupAlignedStopTimeMs()
        )) {
      List<MonitoringMessage> consumedMessages = Lists.newArrayList();
      while (expectedMessages.hasNext() && limit-- > 0) {
        consumedMessages.add(expectedMessages.next().value);
      }
      Map<Metric, ArrayList<Object>> consumed = Maps.newHashMap();
      for (Metric metric : Metric.values()) {
        consumed.put(
            metric,
            Util.extractValues(
                consumedMessages,
                stepMs, rr.getRollupAlignedStartTimeMs(), rr.getRollupAlignedStopTimeMs(),
                limit, metric
            )
        );
      }
      return consumed;
    }
  }

  private Map<Metric, ArrayList<Object>> getProducedMessages(
      RangeRequest rr,
      int limit,
      long stepMs,
      KeyType keyType
  ) {
    //TODO(norwood): don't modify rr
    rr.clientType = ClientType.PRODUCER;
    try (WindowStoreIterator<MonitoringMessage> expectedMessages = expectedDb.get()
        .get(rr.rollup)
        .fetch(
            keySerdeSupplier.get(keyType).key(rr.createStartKeySpec()),
            rr.getRollupAlignedStartTimeMs(),
            rr.getRollupAlignedStopTimeMs()
        )) {
      List<MonitoringMessage> producedMessages = Lists.newArrayList();
      while (expectedMessages.hasNext() && limit-- > 0) {
        producedMessages.add(expectedMessages.next().value);
      }
      Map<Metric, ArrayList<Object>> produced = Maps.newHashMap();
      for (Metric metric : new Metric[]{Metric.count, Metric.error, Metric.crc}) {
        produced.put(
            metric,
            Util.extractValues(
                producedMessages,
                stepMs, rr.getRollupAlignedStartTimeMs(), rr.getRollupAlignedStopTimeMs(),
                limit, metric
            )
        );
      }
      return produced;
    }
  }

  private DeliveryResponse addOverviewData(
      DeliveryResponse res,
      RangeRequest rr,
      DeliveryType deliveryType,
      int limit
  ) {
    res.produced =
        getProducedMessages(rr, limit, rr.rollup.getMillis(), deliveryType.baseProductionKeyType);
    res.consumed =
        getConsumedMessages(rr, limit, rr.rollup.getMillis(), deliveryType.baseConsumptionKeyType);

    if (deliveryType != DeliveryType.CONSUMER) {
      res.consumed.remove(Metric.group);
    }

    return res;
  }

  private void addMembers(
      DeliveryResponse res,
      RangeRequest rr,
      DeliveryType deliveryType,
      int limit,
      String clientId,
      String group,
      Long stepMs,
      RequestType requestType,
      AggregationType aggregateBy
  ) {

    // Notes:
    // (1) It is more efficient to query 1hr rollup data than 15s.
    // (2) For MEMBER_LIST requests, we aren't too concerned about including members that have no
    //     data for the particular period, so we will change 15s rollups to 1hr in some cases.
    // (3) The RangeRequest datatype automatically determines quantized endpoints appropriate to
    //     the given rollup.
    if (requestType == RequestType.MEMBER_LIST && rr.rollup == Rollup.FIFTEEN_SECONDS) {
      if (rr.getRollupAlignedStopTimeMs() - rr.getRollupAlignedStartTimeMs() > Rollup.ONE_HOUR.getMillis()) {
        rr.rollup = Rollup.ONE_HOUR;
      }
    }

    WindowedGrouping aggregateGroupInfo
        = getAggregateGroupInfo(rr, limit, deliveryType.subGroupingKeyType);
    res.members = Lists.newArrayList();
    res.sources = Lists.newArrayList();
    res.destinations = Lists.newArrayList();

    if (deliveryType.subGroupingKeyType != null) {
      for (MemberInfo member : aggregateGroupInfo.getMembersList()) {
        rr = Util.createMergedRangeRequest(member, rr, clientId, group);

        MemberResponse memberResponse = new MemberResponse(member);
        if (requestType == RequestType.ALL) {
          memberResponse.consumed
              = getConsumedMessages(rr, limit, stepMs, deliveryType.subConsumptionKeyType);
          memberResponse.consumed.remove(Metric.group);
          memberResponse.produced
              = getProducedMessages(rr, limit, stepMs, deliveryType.subProductionKeyType);
        }
        res.members.add(memberResponse);
        if (aggregateBy == AggregationType.Topic) {
          res.sources.add(new MemberResponse(member));
        } else {
          res.destinations.add(new MemberResponse(member));
        }
      }
    }

    if (deliveryType.secondarySubGroupingKeyType != null) {
      rr.clientType = ClientType.CONSUMER;
      aggregateGroupInfo =
          getAggregateGroupInfo(rr, limit, deliveryType.secondarySubGroupingKeyType);
      for (MemberInfo member : aggregateGroupInfo.getMembersList()) {
        if (aggregateBy == AggregationType.Topic) {
          res.destinations.add(new MemberResponse(member));
        } else {
          res.sources.add(new MemberResponse(member));
        }
      }
    }
  }

  /**
   * @param limit       max # of records to return
   * @param group       group to query on
   * @param rollup      rollup to use for gathering data
   * @param startTimeMs start of time range
   * @param stopTimeMs  end of time range
   */
  private DeliveryResponse constructResponse(
      int limit,
      String clusterId,
      String group,
      String clientId,
      Rollup rollup,
      Long startTimeMs,
      Long stopTimeMs,
      RequestType requestType,
      String memberGroup,
      String memberTopic,
      Integer memberPartition,
      String memberClientId,
      AggregationType groupBy
  ) {
    checkParams(limit, rollup, startTimeMs, stopTimeMs);

    RangeRequest rr = new RangeRequest(clusterId, startTimeMs, stopTimeMs);
    rr.group = group;
    rr.rollup = rollup;
    rr.clientId = clientId;
    rr.clientType = ClientType.CONSUMER;

    long stepMs = rollup.getMillis();

    DeliveryType deliveryType = groupBy == AggregationType.Topic
        ? DeliveryType.CONSUMER_GROUPS_TOPICS
        : DeliveryType.CONSUMER_GROUPS;

    if (group != null) {
      deliveryType = groupBy == AggregationType.Topic
          ? DeliveryType.CONSUMER_GROUP_TOPICS
          : DeliveryType.CONSUMER_GROUP;
    }

    if (rr.clientId != null) {
      // only occurs in partitionType == ConsumerGroup case.
      deliveryType = DeliveryType.CONSUMER;
    }

    DeliveryResponse res = new DeliveryResponse(
        rr.rollup,
        rr.getRollupAlignedStartTimeMs(),
        rr.rollup.getMillis(),
        rr.group,
        rr.clientId,
        deliveryType
    );

    if (requestType == RequestType.ALL || requestType == RequestType.OVERVIEW) {
      addOverviewData(res, rr, deliveryType, limit);
    }

    if (requestType == RequestType.OVERVIEW) {
      return res;
    }

    if (requestType == RequestType.ALL || requestType == RequestType.MEMBER_LIST) {
      try {
        addMembers(res, rr, deliveryType, limit, clientId, group, stepMs, requestType, groupBy);
        return res;
      } catch (IllegalArgumentException e) {
        log.error("request=range", e);
        throw new BadRequestException(e);
      }
    }

    // RequestType == RequestType.MEMBER

    String
        topic =
        deliveryType == DeliveryType.CONSUMER_GROUPS_TOPICS ? memberTopic : null;

    TopicPartition
        topicPartition =
        deliveryType == DeliveryType.CONSUMER_GROUPS_TOPICS
            ? TopicPartition.getDefaultInstance()
            : TopicPartition.newBuilder()
            .setTopic(Strings.nullToEmpty(memberTopic))
            .setPartition(memberPartition == null ? 0 : memberPartition)
            .build();

    MemberInfo memberInfo = MemberInfo.newBuilder()
        .setClientId(Strings.nullToEmpty(memberClientId))
        .setGroup(Strings.nullToEmpty(memberGroup))
        .setTopic(Strings.nullToEmpty(topic))
        .setTopicPartition(topicPartition)
        .build();

    rr = Util.createMergedRangeRequest(memberInfo, rr, clientId, group);

    MemberResponse
        memberResponse =
        new MemberResponse(memberInfo);

    memberResponse.consumed
        = getConsumedMessages(rr, limit, stepMs, deliveryType.subConsumptionKeyType);
    memberResponse.consumed.remove(Metric.group);
    memberResponse.produced
        = getProducedMessages(rr, limit, stepMs, deliveryType.subProductionKeyType);

    res.members = Lists.newArrayList();
    res.members.add(memberResponse);

    return res;
  }

  @GET
  @Path("/consumer")
  public DeliveryResponse consumer(
      @PathParam("clusterId") String clusterId,
      @QueryParam("limit") @DefaultValue("6000") int limit,
      @QueryParam("group") String group,
      @QueryParam("clientId") String clientId,
      @QueryParam("rollup") Rollup rollup,
      @QueryParam("startTimeMs") Long startTimeMs,
      @QueryParam("stopTimeMs") Long stopTimeMs,
      @QueryParam("type") @DefaultValue("ALL") RequestType type,
      @QueryParam("memberTopic") String memberTopic,
      @QueryParam("memberPartition") Integer memberPartition
  ) {
    if (clientId == null) {
      throw new BadRequestException("missing clientId");
    }
    return constructResponse(
        limit, validateClusterId(clusterId), group, clientId, rollup, startTimeMs, stopTimeMs, type,
        null, memberTopic, memberPartition, null,
        AggregationType.ConsumerGroup
    );
  }

  @GET
  @Path("/consumer_group")
  public DeliveryResponse consumerGroup(
      @PathParam("clusterId") String clusterId,
      @QueryParam("limit") @DefaultValue("6000") int limit,
      @QueryParam("group") String group,
      @QueryParam("rollup") Rollup rollup,
      @QueryParam("startTimeMs") Long startTimeMs,
      @QueryParam("stopTimeMs") Long stopTimeMs,
      @QueryParam("type") @DefaultValue("ALL") RequestType type,
      @QueryParam("memberClientId") String memberClientId
  ) {
    if (group == null) {
      throw new BadRequestException("missing group");
    }
    return constructResponse(
        limit, validateClusterId(clusterId), group, null, rollup, startTimeMs, stopTimeMs, type,
        null, null, null, memberClientId,
        AggregationType.ConsumerGroup
    );
  }

  @GET
  @Path("/consumer_groups")
  public DeliveryResponse consumerGroups(
      @PathParam("clusterId") String clusterId,
      @QueryParam("limit") @DefaultValue("6000") int limit,
      @QueryParam("group") String group,
      @QueryParam("clientId") String clientId,
      @QueryParam("rollup") Rollup rollup,
      @QueryParam("startTimeMs") Long startTimeMs,
      @QueryParam("stopTimeMs") Long stopTimeMs,
      @QueryParam("type") @DefaultValue("ALL") RequestType type,
      @QueryParam("memberGroup") String memberGroup
  ) {
    return constructResponse(
        limit, validateClusterId(clusterId), group, clientId, rollup, startTimeMs, stopTimeMs, type,
        memberGroup, null, null, null,
        AggregationType.ConsumerGroup
    );
  }

  @GET
  @Path("/consumer_group_topics")
  public DeliveryResponse consumerGroupTopics(
      @PathParam("clusterId") String clusterId,
      @QueryParam("limit") @DefaultValue("6000") int limit,
      @QueryParam("group") String group,
      @QueryParam("rollup") Rollup rollup,
      @QueryParam("startTimeMs") Long startTimeMs,
      @QueryParam("stopTimeMs") Long stopTimeMs,
      @QueryParam("type") @DefaultValue("ALL") RequestType type,
      @QueryParam("memberTopic") String memberTopic,
      @QueryParam("memberPartition") Integer memberPartition
  ) {
    return constructResponse(
        limit, validateClusterId(clusterId), group, null, rollup, startTimeMs, stopTimeMs, type,
        null, memberTopic, memberPartition, null,
        AggregationType.Topic
    );
  }

  @GET
  @Path("/consumer_groups_topics")
  public DeliveryResponse consumerGroupsTopics(
      @PathParam("clusterId") String clusterId,
      @QueryParam("limit") @DefaultValue("6000") int limit,
      @QueryParam("group") String group,
      @QueryParam("rollup") Rollup rollup,
      @QueryParam("startTimeMs") Long startTimeMs,
      @QueryParam("stopTimeMs") Long stopTimeMs,
      @QueryParam("type") @DefaultValue("ALL") RequestType type,
      @QueryParam("memberTopic") String memberTopic
  ) {
    return constructResponse(
        limit, validateClusterId(clusterId), group, null, rollup, startTimeMs, stopTimeMs, type,
        null, memberTopic, null, null,
        AggregationType.Topic
    );
  }
}
