// (Copyright) [2016 - 2016] Confluent, Inc.
package io.confluent.controlcenter.serialization;

import com.google.common.collect.ImmutableList;
import com.google.inject.Binder;
import com.google.inject.BindingAnnotation;
import com.google.inject.Module;
import com.google.inject.Provides;
import com.google.inject.Singleton;
import com.google.protobuf.Descriptors;

import org.apache.kafka.common.serialization.Deserializer;
import org.apache.kafka.common.serialization.Serializer;
import org.apache.kafka.common.utils.Bytes;

import java.lang.annotation.ElementType;
import java.lang.annotation.Retention;
import java.lang.annotation.Target;
import java.util.Map;

import io.confluent.command.record.alert.CommandAlert.ActionConfig;
import io.confluent.command.record.alert.CommandAlert.MonitoringTriggerConfig;
import io.confluent.controlcenter.alert.record.Alert;
import io.confluent.controlcenter.alert.record.Alert.AlertInfo;
import io.confluent.controlcenter.keys.Keys.KeyType;
import io.confluent.controlcenter.record.Controlcenter.TriggerEvent;
import io.confluent.controlcenter.record.Controlcenter.ClientGroup;
import io.confluent.controlcenter.record.Controlcenter.ClientGroupProductionAggregate;
import io.confluent.controlcenter.record.Controlcenter.ClientGroups;
import io.confluent.controlcenter.record.Controlcenter.Cluster;
import io.confluent.controlcenter.record.Controlcenter.ClusterContext;
import io.confluent.controlcenter.record.Controlcenter.TopicPartition;
import io.confluent.controlcenter.record.Controlcenter.TopicPartitionClientGroup;
import io.confluent.controlcenter.record.Controlcenter.TopicPartitionInfo;
import io.confluent.controlcenter.record.Controlcenter.TriggerActions;
import io.confluent.controlcenter.record.Controlcenter.TriggerEventKey;
import io.confluent.controlcenter.record.Controlcenter.TriggerMeasurement;
import io.confluent.controlcenter.record.Controlcenter.VerifiableMonitoringMessage;
import io.confluent.controlcenter.record.Controlcenter.VerifiableTriggerInfo;
import io.confluent.controlcenter.record.Controlcenter.WindowedClusterGroup;
import io.confluent.controlcenter.record.Controlcenter.WindowedGrouping;
import io.confluent.controlcenter.streams.verify.MonitoringVerifier;
import io.confluent.monitoring.record.Monitoring.MonitoringMessage;
import io.confluent.serializers.OrderedKeyProtoSerde;
import io.confluent.serializers.OrderedKeyUberSerde;
import io.confluent.serializers.OrderedKeyUberSerdeDelegate;
import io.confluent.serializers.ProtoSerde;
import io.confluent.serializers.StringSerde;
import io.confluent.serializers.UberSerde;

import static java.lang.annotation.RetentionPolicy.RUNTIME;

public class SerializationModule implements Module {

  @Override
  public void configure(Binder binder) {
  }

  @Provides
  @Singleton
  private OrderedKeyPrefixedSerdeSupplier<KeyType, MonitoringMessage> keyTypeMonitoringMessageOrderedKeyPrefixedSerdeSupplier() {
    return new OrderedKeyPrefixedMMSerdeSupplier();
  }

  @Provides
  @Singleton
  private OrderedKeyPrefixedSerdeSupplier<ClusterContext, Cluster> clusterContextClusterOrderedKeyPrefixedSerdeSupplier() {
    return new OrderedKeyPrefixedClusterSerdeSupplier();
  }

  @Provides
  @Singleton
  private UberSerde<ClientGroupProductionAggregate> clientGroupProductionAggregateSerde() {
    return new ProtoSerde<>(ClientGroupProductionAggregate.getDefaultInstance());
  }

  @Provides
  @Singleton
  private UberSerde<MonitoringMessage> monitoringMessageSerde() {
    return new ProtoSerde<>(MonitoringMessage.getDefaultInstance());
  }

  @Provides
  @Singleton
  private UberSerde<String> stringSerde() {
    return new StringSerde();
  }

  @Provides
  @Singleton
  private UberSerde<TopicPartitionClientGroup> topicPartitionClientGroupSerde() {
    return new ProtoSerde<>(TopicPartitionClientGroup.getDefaultInstance());
  }

  @Provides
  @Singleton
  private UberSerde<ClientGroup> clientGroupUberSerde() {
    return new ProtoSerde<>(ClientGroup.getDefaultInstance());
  }

  @Provides
  @Singleton
  private UberSerde<ClientGroups> clientGroupsUberSerde() {
    return new ProtoSerde<>(ClientGroups.getDefaultInstance());
  }

  @Provides
  @Singleton
  private UberSerde<TopicPartition> topicPartitionSerde() {
    return new ProtoSerde<>(TopicPartition.getDefaultInstance());
  }

  @Provides
  @Singleton
  private UberSerde<TopicPartitionInfo> topicPartitionInfoSerde() {
    return new ProtoSerde<>(TopicPartitionInfo.getDefaultInstance());
  }

  @Provides
  @Singleton
  private UberSerde<WindowedGrouping> windowedGroupingSerde() {
    return new ProtoSerde<>(WindowedGrouping.getDefaultInstance());
  }

  @Provides
  @Singleton
  private UberSerde<Cluster> cluserSerde() {
    return new ProtoSerde<>(Cluster.getDefaultInstance());
  }

  @Provides
  @Singleton
  private UberSerde<VerifiableMonitoringMessage> verifiableMonitoringMessageSerde() {
    return new ProtoSerde<>(VerifiableMonitoringMessage.getDefaultInstance());
  }

  @Provides
  @Singleton
  private UberSerde<MonitoringVerifier> monitoringVerifierSerde() {
    return new MonitoringVerifierSerde();
  }

  @Provides
  @Singleton
  private UberSerde<TriggerMeasurement> triggerMeasurementSerde() {
    return new ProtoSerde<>(TriggerMeasurement.getDefaultInstance());
  }

  @Provides
  @Singleton
  private UberSerde<MonitoringTriggerConfig> monitoringTriggerConfigUberSerde() {
    return new ProtoSerde<>(MonitoringTriggerConfig.getDefaultInstance());
  }

  @Provides
  @Singleton
  private UberSerde<ActionConfig> actionConfigUberSerde() {
    return new ProtoSerde<>(ActionConfig.getDefaultInstance());
  }

  @Provides
  @Singleton
  private UberSerde<WindowedClusterGroup> windowedClusterGroupSerde() {
    return new ProtoSerde<>(WindowedClusterGroup.getDefaultInstance());
  }

  @BindingAnnotation
  @Retention(RUNTIME)
  @Target({ElementType.FIELD, ElementType.PARAMETER, ElementType.METHOD})
  public @interface VerifierKeySerde {
  }

  @Provides
  @VerifierKeySerde
  @Singleton
  private OrderedKeyUberSerde<MonitoringMessage> getVerifierKeySerde() {
    return OrderedKeyProtoSerde.create(MonitoringMessage.getDefaultInstance(),
        MonitoringMessage.CLUSTERID_FIELD_NUMBER,
        MonitoringMessage.CLIENTTYPE_FIELD_NUMBER,
        MonitoringMessage.CLIENTID_FIELD_NUMBER,
        MonitoringMessage.TOPIC_FIELD_NUMBER,
        MonitoringMessage.PARTITION_FIELD_NUMBER,
        MonitoringMessage.SESSION_FIELD_NUMBER
    );
  }

  @BindingAnnotation
  @Retention(RUNTIME)
  @Target({ElementType.FIELD, ElementType.PARAMETER, ElementType.METHOD})
  public @interface MonitoringTriggerKeySerde {
  }

  @Provides
  @MonitoringTriggerKeySerde
  @Singleton
  private OrderedKeyUberSerde<WindowedClusterGroup> getMonitoringTriggerKeySerde() {
    return OrderedKeyProtoSerde.create(WindowedClusterGroup.getDefaultInstance(),
        WindowedClusterGroup.CLUSTERID_FIELD_NUMBER,
        WindowedClusterGroup.GROUP_FIELD_NUMBER,
        WindowedClusterGroup.WINDOW_FIELD_NUMBER
    );
  }

  @BindingAnnotation
  @Retention(RUNTIME)
  @Target({ElementType.FIELD, ElementType.PARAMETER, ElementType.METHOD})
  public @interface TriggerActionsKeySerde {
  }

  @Provides
  @TriggerActionsKeySerde
  @Singleton
  private OrderedKeyUberSerde<TriggerEventKey> getTriggerActionsKeySerde() {
    return OrderedKeyProtoSerde.create(TriggerEventKey.getDefaultInstance(),
        TriggerEventKey.TRIGGERGUID_FIELD_NUMBER,
        TriggerEventKey.TIMESTAMP_FIELD_NUMBER,
        TriggerEventKey.WINDOW_FIELD_NUMBER
    );
  }

  // used to invert ts when serializing alert key so we get reverse chrono sort
  // since this serde is only used to make keys, we only need to override key/toProto
  static class AlertInfoInvertedTimestampSerde extends OrderedKeyUberSerdeDelegate<AlertInfo> {
    AlertInfoInvertedTimestampSerde(OrderedKeyUberSerde<AlertInfo> delegate) {
      super(delegate);
    }

    @Override
    public Bytes key(AlertInfo message) {
      return super.key(
          AlertInfo.newBuilder(message)
              .setTimestamp(-message.getTimestamp())
              .build()
      );
    }

    @Override
    public AlertInfo toProto(Bytes key) {
      AlertInfo ai = super.toProto(key);
      return AlertInfo.newBuilder(ai)
          .setTimestamp(-ai.getTimestamp())
          .build();
    }
  }

  @BindingAnnotation
  @Retention(RUNTIME)
  @Target({ElementType.FIELD, ElementType.PARAMETER, ElementType.METHOD})
  public @interface TriggerEventsKeySerde {
  }

  @Provides
  @TriggerEventsKeySerde
  @Singleton
  private OrderedKeyUberSerde<AlertInfo> getTriggerEventsKeySerde() {
    return new AlertInfoInvertedTimestampSerde(
        OrderedKeyProtoSerde.create(
            AlertInfo.getDefaultInstance(),
            AlertInfo.TIMESTAMP_FIELD_NUMBER,
            AlertInfo.GUID_FIELD_NUMBER
        )
    );
  }

  @BindingAnnotation
  @Retention(RUNTIME)
  @Target({ElementType.FIELD, ElementType.PARAMETER, ElementType.METHOD})
  public @interface TriggerEventsKeyPrefixSerde {
  }

  @Provides
  @TriggerEventsKeyPrefixSerde
  @Singleton
  private OrderedKeyUberSerde<AlertInfo> getTriggerEventsKeyPrefixSerde(
      final @TriggerEventsKeySerde OrderedKeyUberSerde<AlertInfo> keySerde
  ) {
    return new AlertInfoInvertedTimestampSerde(
        OrderedKeyProtoSerde.create(
            AlertInfo.getDefaultInstance(),
            AlertInfo.TIMESTAMP_FIELD_NUMBER,
            AlertInfo.GUID_FIELD_NUMBER
        )
    );
  }

  @Provides
  @Singleton
  private UberSerde<TriggerActions> triggerActionsSerde() {
    return new ProtoSerde<>(TriggerActions.getDefaultInstance());
  }

  @Provides
  @Singleton
  private UberSerde<VerifiableTriggerInfo> verifiableTriggerSerde() {
    return new ProtoSerde<>(VerifiableTriggerInfo.getDefaultInstance());
  }

  @Provides
  @Singleton
  private UberSerde<AlertInfo> alertInfoSerde() {
    return new ProtoSerde<>(AlertInfo.getDefaultInstance());
  }

  @Provides
  @Singleton
  private UberSerde<TriggerEvent> triggerEventSerde() {
    return new ProtoSerde<>(TriggerEvent.getDefaultInstance());
  }
}
