// (Copyright) [2016 - 2016] Confluent, Inc.
package io.confluent.controlcenter.streams;

import com.google.common.collect.ImmutableSet;
import com.google.common.collect.Sets;
import com.google.common.util.concurrent.ThreadFactoryBuilder;
import com.google.inject.Inject;
import com.google.inject.Provider;

import org.apache.kafka.streams.KafkaStreams;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.Set;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.TimeoutException;

import io.confluent.controlcenter.Rollup;

public class KafkaStreamsManager {

  private static final Logger log = LoggerFactory.getLogger(KafkaStreamsManager.class);

  private final Provider<KafkaStreams> kafkaStreamsProvider;
  private final Thread.UncaughtExceptionHandler kafkaStreamsUncaughtExceptionHandler;
  private final ExecutorService es;
  private final ImmutableSet<TopicStoreMaster.Store> stores;
  private KafkaStreams currentStreams;

  private static final class KafkaStreamsUncaughtExceptionHandler
      implements Thread.UncaughtExceptionHandler {

    private final KafkaStreamsManager kafkaStreamsManager;

    public KafkaStreamsUncaughtExceptionHandler(KafkaStreamsManager kafkaStreamsManager) {
      this.kafkaStreamsManager = kafkaStreamsManager;
    }

    @Override
    public void uncaughtException(Thread t, Throwable e) {
      log.error("streams exception", e);
      try {
        kafkaStreamsManager.start(TimeUnit.SECONDS.toMillis(90));
      } catch (Exception ex) {
        log.error("Failed to restart Kafka Streams after failure", ex);
      }
    }
  }

  private static final class StreamCloser implements Runnable {

    private final KafkaStreams streamsToClose;

    private StreamCloser(KafkaStreams streamsToClose) {
      this.streamsToClose = streamsToClose;
    }

    @Override
    public void run() {
      try {
        streamsToClose.close();
      } catch (Throwable t) {
        log.error("closing streams", t);
      }
    }
  }

  @Inject
  public KafkaStreamsManager(
      Provider<KafkaStreams> kafkaStreamsProvider,
      @TopicStoreModule.AllStores ImmutableSet<TopicStoreMaster.Store> stores
  ) {
    this.kafkaStreamsProvider = kafkaStreamsProvider;
    this.stores = stores;
    this.kafkaStreamsUncaughtExceptionHandler = new KafkaStreamsUncaughtExceptionHandler(this);
    this.es = Executors.newSingleThreadScheduledExecutor(
        new ThreadFactoryBuilder().setNameFormat("stream-manager-%d").build()
    );
  }

  public synchronized void start(long allowance) throws Exception {
    long timeout = System.currentTimeMillis() + allowance;
    if (currentStreams != null) {
      es.submit(new StreamCloser(currentStreams)).get(allowance, TimeUnit.SECONDS);
    }
    currentStreams = kafkaStreamsProvider.get();
    if (currentStreams == null) {
      throw new RuntimeException("Kafka Stream failed to initialize");
    }
    currentStreams.setUncaughtExceptionHandler(kafkaStreamsUncaughtExceptionHandler);
    currentStreams.start();
    Set<TopicStoreMaster.Store> toCheck = Sets.newHashSet(stores);
    while (!toCheck.isEmpty()) {
      for (TopicStoreMaster.Store storeDef : stores) {
        if (!toCheck.contains(storeDef)) {
          continue;
        }
        String storeName = storeDef.rollup ? TopicStoreMaster.nameJoin(storeDef.name, Rollup.FIFTEEN_SECONDS.name()) : storeDef.name;
        try {
          Object store = currentStreams.store(
              storeName,
              storeDef.queryableWindowStoreType == null ?
              storeDef.queryableKeyValueStoreType :
              storeDef.queryableWindowStoreType
          );

          if (store != null) {
            toCheck.remove(storeDef);
          }
        } catch (Exception e) {
          log.debug("unable to get store name={} {}", storeDef.name, log.isTraceEnabled() ? e : "");
        }
        if (timeout < System.currentTimeMillis()) {
          log.warn("unable to start with allowance={}", allowance);
          throw new TimeoutException();
        }
      }
      Thread.sleep(Math.min(allowance / 10, 5000));
    }
  }

  public synchronized void close() {
    try {
      if (currentStreams != null) {
        if (es.submit(new StreamCloser(currentStreams)).get(30, TimeUnit.SECONDS) == null) {
          log.info("Successfully closed Kafka Streams");
        }
        currentStreams = null;
      }
    } catch (Exception e) {
      log.error("Closing Kafka Streams did not finish", e);
    }
  }
}
