package io.confluent.controlcenter.streams;

import com.google.inject.Inject;

import org.apache.kafka.streams.KeyValue;
import org.apache.kafka.streams.kstream.Transformer;
import org.apache.kafka.streams.kstream.TransformerSupplier;
import org.apache.kafka.streams.processor.ProcessorContext;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.ArrayList;
import java.util.List;

import io.confluent.controlcenter.ControlCenterConfig;
import io.confluent.monitoring.record.Monitoring;

/**
 * Fills in monitoring topic partition info in regular monitoring messages and transforms C3
 * heartbeat message to an array of C3 heartbeat messages for every partition in re-keyed topic.
 *
 * The verifier downstream uses timestamps in C3 heartbeat messages as a clock for deciding when
 * to expire sequences. To handle the case when one partition in monitoring topic is slower than
 * others, the 'clock' is per monitoring topic partition. This transformer is needed to correlate
 * C3 heartbeat messages to client/topic/partitions from the same monitoring topic partition.
 */
public class SourcePartitionTrackingTransformerSupplier implements
                                       TransformerSupplier<byte[], Monitoring.MonitoringMessage, KeyValue<byte[], Iterable<KeyValue<byte[], Monitoring.MonitoringMessage>>>> {

  private static final Logger log = LoggerFactory.getLogger(SourcePartitionTrackingTransformerSupplier.class);

  private final int numInternalPartitions;

  @Inject
  public SourcePartitionTrackingTransformerSupplier(ControlCenterConfig cfg) {
    this.numInternalPartitions = cfg.getInt(
        ControlCenterConfig.CONTROL_CENTER_INTERNAL_PARTITIONS_CONFIG);
  }

  @Override
  public Transformer<byte[], Monitoring.MonitoringMessage, KeyValue<byte[], Iterable<KeyValue<byte[], Monitoring.MonitoringMessage>>>> get() {
    return new SourcePartitionTrackingTransformer();
  }

  private class SourcePartitionTrackingTransformer implements
                                            Transformer<byte[], Monitoring.MonitoringMessage, KeyValue<byte[], Iterable<KeyValue<byte[], Monitoring.MonitoringMessage>>>> {
    private ProcessorContext context;

    @SuppressWarnings("unchecked")
    @Override
    public void init(ProcessorContext context) {
      this.context = context;
    }

    @Override
    public KeyValue<byte[], Iterable<KeyValue<byte[], Monitoring.MonitoringMessage>>> transform(
        byte[] key,
        Monitoring.MonitoringMessage monitoringMessage
    ) {
      List<KeyValue<byte[], Monitoring.MonitoringMessage>> ret = new ArrayList<>();

      if (monitoringMessage.getClientType() == Monitoring.ClientType.CONTROLCENTER) {
        // broadcast C3 heartbeat to every partition of re-keyed monitoring topic
        if (log.isTraceEnabled()) {
          log.trace(
              "Broadcasting C3 heartbeat from monitoring topic partition={} to all (total={}) "
              + "partitions of re-keyed monitoring topic", context.partition(),
              numInternalPartitions);
        }
        for (int partition = 0; partition < numInternalPartitions; ++partition) {
          // we are re-using 'partition' field to tell which partition of the re-keyed monitoring
          // topic to route this message
          Monitoring.MonitoringMessage msg = Monitoring.MonitoringMessage.newBuilder
              (monitoringMessage)
              .setMonitoringTopicPartition(context.partition())
              .setPartition(partition)
              .build();
          ret.add(KeyValue.pair(key, msg));
        }
      } else {
        Monitoring.MonitoringMessage msg = Monitoring.MonitoringMessage.newBuilder
            (monitoringMessage)
            .setMonitoringTopicPartition(context.partition())
            .build();
        ret.add(KeyValue.pair(key, msg));
      }
      return KeyValue.pair(null, (Iterable<KeyValue<byte[], Monitoring.MonitoringMessage>>) ret);
    }

    @Override
    public KeyValue<byte[], Iterable<KeyValue<byte[], Monitoring.MonitoringMessage>>> punctuate(long timestamp) {
      return null;
    }

    @Override
    public void close() {
    }
  }
}
