// (Copyright) [2016 - 2016] Confluent, Inc.
package io.confluent.controlcenter.streams;

import com.google.common.collect.ImmutableSet;
import com.google.inject.AbstractModule;
import com.google.inject.Binding;
import com.google.inject.BindingAnnotation;
import com.google.inject.Inject;
import com.google.inject.Injector;
import com.google.inject.Key;
import com.google.inject.Provides;
import com.google.inject.TypeLiteral;
import io.confluent.command.CommandStore;
import io.confluent.command.record.Command.CommandKey;
import io.confluent.command.record.Command.CommandMessage;
import io.confluent.controlcenter.ControlCenterConfig;
import io.confluent.controlcenter.alert.record.Alert.AlertInfo;
import io.confluent.controlcenter.keys.Keys.KeyType;
import io.confluent.controlcenter.record.Controlcenter.TriggerEvent;
import io.confluent.controlcenter.record.Controlcenter.ClientGroupProductionAggregate;
import io.confluent.controlcenter.record.Controlcenter.Cluster;
import io.confluent.controlcenter.record.Controlcenter.ClusterContext;
import io.confluent.controlcenter.record.Controlcenter.TopicPartition;
import io.confluent.controlcenter.record.Controlcenter.TriggerActions;
import io.confluent.controlcenter.record.Controlcenter.TriggerMeasurement;
import io.confluent.controlcenter.record.Controlcenter.VerifiableMonitoringMessage;
import io.confluent.controlcenter.record.Controlcenter.VerifiableTriggerInfo;
import io.confluent.controlcenter.record.Controlcenter.WindowedClusterGroup;
import io.confluent.controlcenter.record.Controlcenter.WindowedGrouping;
import io.confluent.controlcenter.serialization.OrderedKeyPrefixedSerdeSupplier;
import io.confluent.controlcenter.serialization.SerializationModule.VerifierKeySerde;
import io.confluent.controlcenter.serialization.SerializationModule.MonitoringTriggerKeySerde;
import io.confluent.controlcenter.serialization.WindowStoreSerde;
import io.confluent.controlcenter.streams.TopicStoreMaster.Store;
import io.confluent.controlcenter.streams.TopicStoreMaster.Topic;
import io.confluent.controlcenter.streams.verify.MonitoringVerifier;
import io.confluent.monitoring.record.Monitoring.MonitoringMessage;
import io.confluent.serializers.OrderedKeyUberSerde;
import io.confluent.serializers.UberSerde;
import org.apache.kafka.common.utils.Bytes;
import org.apache.kafka.streams.StreamsConfig;

import java.lang.annotation.ElementType;
import java.lang.annotation.Retention;
import java.lang.annotation.Target;
import java.lang.reflect.ParameterizedType;
import java.lang.reflect.Type;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import static java.lang.annotation.RetentionPolicy.RUNTIME;

public class TopicStoreModule extends AbstractModule {

  private static final String VERIFIER_STORE_NAME = "MonitoringVerifierStore";
  private static final String GROUP_STORE_NAME = "Group";
  private static final String CLUSTER_STORE_NAME = "Cluster";
  private static final String MONITORING_STREAM_STORE_NAME = "MonitoringStream";
  private static final String MONITORING_AGGREGATOR_WINDOWS_STORE_NAME = "MonitoringMessageAggregatorWindows";
  private static final String AGG_TOPIC_PARTITION_TABLE_WINDOWS_STORE_NAME = "aggregatedTopicPartitionTableWindows";
  private static final String MONITORING_TRIGGER_STORE_NAME = "MonitoringTriggerStore";
  private static final String TRIGGER_ACTIONS_STORE_NAME = "TriggerActionsStore";
  private static final String TRIGGER_EVENTS_STORE_NAME = "TriggerEventsStore";
  private static final String ALERT_HISTORY_STORE_NAME = "AlertHistoryStore";
  private static final String KSTREAM_OUTEROTHER_90 = "KSTREAM-OUTEROTHER-0000000090-store";
  private static final String KSTREAM_OUTERTHIS_89 = "KSTREAM-OUTERTHIS-0000000089-store";

  /**
   * List of through(), to(), and from() topics that needs to be created before hand
   * Note that actual topic name will also have rollup appended to the end of the name, one topic
   * per rollup
   */
  private static final String CONSUMER_GROUP_AGGREGATE_TP_TOPIC = "aggregate-topic-partition";
  private static final String CONSUMER_GROUP_GROUP_AGGREGATE_TOPIC = "group-aggregate-topic";
  private static final String MONITORING_MESSAGE_ERROR_TOPIC = "error-topic";
  private static final String MONITORING_MESSAGE_REKEY_TOPIC = "monitoring-message-rekey";
  private static final String MONITORING_MESSAGE_EXTENSION_TOPIC = "monitoring-aggregate-rekey";
  private static final String GROUP_STREAM_EXTENSION_TOPIC = "group-stream-extension-rekey";
  private static final String CLUSTER_STREAM_EXTENSION_TOPIC = "cluster-rekey";
  private static final String MONITORING_TRIGGER_EVENT_REKEY_TOPIC = "monitoring-trigger-event-rekey";
  // TODO(apovzner) temporary topics until we make same keys for expected and actual consumption streams
  private static final String EXPECTED_GROUP_CONSUMPTION_AGGREGATE_TOPIC = "expected-group-consumption-rekey";
  private static final String ACTUAL_GROUP_CONSUMPTION_AGGREGATE_TOPIC = "actual-group-consumption-rekey";

  @Override
  protected void configure() {
  }

  @BindingAnnotation
  @Retention(RUNTIME)
  @Target({ElementType.FIELD, ElementType.PARAMETER, ElementType.METHOD})
  public @interface MonitoringTopic {

  }

  @Provides
  @Inject
  @MonitoringTopic
  protected Topic<Void, MonitoringMessage, Void, Void> getMonitoringTopic(ControlCenterConfig cfg, UberSerde<MonitoringMessage> monitoringMessageSerde) {
    return new Topic<>(
        cfg.getString(ControlCenterConfig.CONTROL_CENTER_TOPIC_CONFIG),
        null,
        null,
        monitoringMessageSerde,
        false
    );
  }

  @BindingAnnotation
  @Retention(RUNTIME)
  @Target({ElementType.FIELD, ElementType.PARAMETER, ElementType.METHOD})
  public @interface CommandTopic {

  }

  @Provides
  @Inject
  @CommandTopic
  protected Topic<CommandKey, CommandMessage, Void, Void> getCommandTopic(
      ControlCenterConfig config
  ) {
    return new Topic<>(
        config.getString(ControlCenterConfig.CONTROL_CENTER_COMMAND_TOPIC),
        null,
        CommandStore.getCommandKeySerde(),
        CommandStore.getCommandMessageSerde(),
        false
    );
  }

  @BindingAnnotation
  @Retention(RUNTIME)
  @Target({ElementType.FIELD, ElementType.PARAMETER, ElementType.METHOD})
  public @interface ReKeyedMonitoringTopic {

  }

  @Provides
  @Inject
  @ReKeyedMonitoringTopic
  protected Topic<Void, MonitoringMessage, Void, Void> getRekeyedMonitoringTopic(StreamsConfig streamsCfg, UberSerde<MonitoringMessage> monitoringMessageSerde) {
    return new Topic<>(
        TopicStoreMaster.getInternalTopicName(streamsCfg, MONITORING_MESSAGE_REKEY_TOPIC),
        null,
        null,
        monitoringMessageSerde,
        false
    );
  }

  @BindingAnnotation
  @Retention(RUNTIME)
  @Target({ElementType.FIELD, ElementType.PARAMETER, ElementType.METHOD})
  public @interface MonitoringErrorTopic {

  }

  @Provides
  @Inject
  @MonitoringErrorTopic
  protected Topic<Void, MonitoringMessage, Void, Void> getMonitoringMessageErrorTopic(StreamsConfig streamsCfg, UberSerde<MonitoringMessage> monitoringMessageSerde) {
    return new Topic<>(
        TopicStoreMaster.getInternalTopicName(streamsCfg, MONITORING_MESSAGE_ERROR_TOPIC),
        null,
        null,
        monitoringMessageSerde,
        false
    );
  }

  @BindingAnnotation
  @Retention(RUNTIME)
  @Target({ElementType.FIELD, ElementType.PARAMETER, ElementType.METHOD})
  public @interface ReKeyedTriggerEventTopic {

  }

  @Provides
  @Inject
  @ReKeyedTriggerEventTopic
  protected Topic<Bytes, TriggerEvent, Void, Void> getRekeyedTriggerEventTopic(
      StreamsConfig streamsCfg,
      @MonitoringTriggerKeySerde OrderedKeyUberSerde<WindowedClusterGroup> keySerde,
      UberSerde<TriggerEvent> triggerEventSerde) {
    return new Topic<>(
        TopicStoreMaster.getInternalTopicName(streamsCfg, MONITORING_TRIGGER_EVENT_REKEY_TOPIC),
        null,
        keySerde,
        triggerEventSerde,
        false
    );
  }

  @BindingAnnotation
  @Retention(RUNTIME)
  @Target({ElementType.FIELD, ElementType.PARAMETER, ElementType.METHOD})
  public @interface MonitoringMsgExtensionTopic {

  }

  @Provides
  @Inject
  @MonitoringMsgExtensionTopic
  protected Topic<Bytes, VerifiableMonitoringMessage, String, VerifiableMonitoringMessage> getMonitoringMessageExtensionTopic(
      StreamsConfig streamsConfig,
      UberSerde<String> stringSerde,
      OrderedKeyPrefixedSerdeSupplier<KeyType, MonitoringMessage> keySerdeSupplier,
      UberSerde<VerifiableMonitoringMessage> verifiableMonitoringMessageUberSerde
  ) {
    boolean rollup = false;
    return new Topic<>(
        TopicStoreMaster.getInternalTopicName(streamsConfig, MONITORING_MESSAGE_EXTENSION_TOPIC),
        new Store<>(MONITORING_MESSAGE_EXTENSION_TOPIC, stringSerde, verifiableMonitoringMessageUberSerde, verifiableMonitoringMessageUberSerde, rollup),
        keySerdeSupplier.get(),
        verifiableMonitoringMessageUberSerde,
        rollup
    );
  }

  @BindingAnnotation
  @Retention(RUNTIME)
  @Target({ElementType.FIELD, ElementType.PARAMETER, ElementType.METHOD})
  public @interface GroupStreamExtensionTopic {

  }

  @Provides
  @Inject
  @GroupStreamExtensionTopic
  protected Topic<Bytes, WindowedGrouping, Void, WindowedGrouping> getGroupStreamExtensionTopic(
      StreamsConfig streamsConfig,
      OrderedKeyPrefixedSerdeSupplier<KeyType, MonitoringMessage> keySerdeSupplier,
      UberSerde<WindowedGrouping> windowedGroupingUberSerde
  ) {
    boolean rollup = false;
    return new Topic<>(
        TopicStoreMaster.getInternalTopicName(streamsConfig, GROUP_STREAM_EXTENSION_TOPIC),
        null,
        keySerdeSupplier.get(),
        windowedGroupingUberSerde,
        rollup
    );
  }

  @BindingAnnotation
  @Retention(RUNTIME)
  @Target({ElementType.FIELD, ElementType.PARAMETER, ElementType.METHOD})
  public @interface ClusterStreamExtensionTopic {

  }

  @Provides
  @Inject
  @ClusterStreamExtensionTopic
  protected Topic<Bytes, Cluster, Void, Cluster> getClusterStreamExtensionTopic(
      StreamsConfig streamsConfig,
      OrderedKeyPrefixedSerdeSupplier<ClusterContext, Cluster> keySerdeSupplier,
      UberSerde<Cluster> clusterSerde
  ) {
    return new Topic<>(
        TopicStoreMaster.getInternalTopicName(streamsConfig, CLUSTER_STREAM_EXTENSION_TOPIC),
        null,
        keySerdeSupplier.get(),
        clusterSerde,
        false
    );
  }

  @BindingAnnotation
  @Retention(RUNTIME)
  @Target({ElementType.FIELD, ElementType.PARAMETER, ElementType.METHOD})
  public @interface ConsumerGroupAggTPTopic {

  }

  @Provides
  @Inject
  @ConsumerGroupAggTPTopic
  protected Topic<TopicPartition, VerifiableMonitoringMessage, String, VerifiableMonitoringMessage> getConsumerGroupAggrTopicPartitionTopic(
      StreamsConfig streamsConfig,
      UberSerde<String> stringSerde,
      UberSerde<VerifiableMonitoringMessage> verifiableMonitoringMessageUberSerde,
      UberSerde<TopicPartition> topicPartitionSerde
  ) {
    boolean rollup = false;
    return new Topic<>(
        TopicStoreMaster.getInternalTopicName(streamsConfig, CONSUMER_GROUP_AGGREGATE_TP_TOPIC),
        new Store<>(CONSUMER_GROUP_AGGREGATE_TP_TOPIC, stringSerde, verifiableMonitoringMessageUberSerde, verifiableMonitoringMessageUberSerde, rollup),
        topicPartitionSerde,
        verifiableMonitoringMessageUberSerde,
        rollup
    );
  }

  @BindingAnnotation
  @Retention(RUNTIME)
  @Target({ElementType.FIELD, ElementType.PARAMETER, ElementType.METHOD})
  public @interface ConsumerGroupGroupAggTopic {

  }

  @Provides
  @Inject
  @ConsumerGroupGroupAggTopic
  protected Topic<Bytes, VerifiableMonitoringMessage, String, VerifiableMonitoringMessage> getConsumerGroupGroupAggregateTopic(
      StreamsConfig streamsConfig,
      UberSerde<String> stringSerde,
      OrderedKeyPrefixedSerdeSupplier<KeyType, MonitoringMessage> keySerdeSupplier,
      UberSerde<VerifiableMonitoringMessage> verifiableMonitoringMessageUberSerde
  ) {
    boolean rollup = true;
    return new Topic<>(
        TopicStoreMaster.getInternalTopicName(streamsConfig, CONSUMER_GROUP_GROUP_AGGREGATE_TOPIC),
        new Store<>(CONSUMER_GROUP_GROUP_AGGREGATE_TOPIC, stringSerde, verifiableMonitoringMessageUberSerde, verifiableMonitoringMessageUberSerde, rollup),
        keySerdeSupplier.get(),
        verifiableMonitoringMessageUberSerde,
        rollup
    );
  }

  @BindingAnnotation
  @Retention(RUNTIME)
  @Target({ElementType.FIELD, ElementType.PARAMETER, ElementType.METHOD})
  public @interface ExpectedGroupConsumptionAggregateTopic {

  }

  @Provides
  @Inject
  @ExpectedGroupConsumptionAggregateTopic
  protected Topic<WindowedClusterGroup, MonitoringMessage, Void, Void> getExpectedGroupConsumptionAggregateTopic(
      StreamsConfig streamsConfig,
      UberSerde<WindowedClusterGroup> windowedClusterGroupSerde,
      UberSerde<MonitoringMessage> monitoringMessageSerde
  ) {
    return new Topic<>(
        TopicStoreMaster.getInternalTopicName(streamsConfig, EXPECTED_GROUP_CONSUMPTION_AGGREGATE_TOPIC),
        null,
        windowedClusterGroupSerde,
        monitoringMessageSerde,
        false
    );
  }

  @BindingAnnotation
  @Retention(RUNTIME)
  @Target({ElementType.FIELD, ElementType.PARAMETER, ElementType.METHOD})
  public @interface ActualGroupConsumptionAggregateTopic {

  }

  @Provides
  @Inject
  @ActualGroupConsumptionAggregateTopic
  protected Topic<WindowedClusterGroup, MonitoringMessage, Void, Void> getActualGroupConsumptionAggregateTopic(
      StreamsConfig streamsConfig,
      UberSerde<WindowedClusterGroup> windowedClusterGroupSerde,
      UberSerde<MonitoringMessage> monitoringMessageSerde
  ) {
    return new Topic<>(
        TopicStoreMaster.getInternalTopicName(streamsConfig, ACTUAL_GROUP_CONSUMPTION_AGGREGATE_TOPIC),
        null,
        windowedClusterGroupSerde,
        monitoringMessageSerde,
        false
    );
  }

  @BindingAnnotation
  @Retention(RUNTIME)
  @Target({ElementType.FIELD, ElementType.PARAMETER, ElementType.METHOD})
  public @interface VerifierStore {

  }

  @Provides
  @Inject
  @VerifierStore
  protected Store<Bytes, MonitoringVerifier, MonitoringVerifier> getVerifierStore(@VerifierKeySerde OrderedKeyUberSerde<MonitoringMessage> keySerde, UberSerde<MonitoringVerifier> monitoringVerifierUberSerde) {
    return new Store<>(
        VERIFIER_STORE_NAME,
        keySerde,
        monitoringVerifierUberSerde,
        monitoringVerifierUberSerde,
        false
    );
  }

  @BindingAnnotation
  @Retention(RUNTIME)
  @Target({ElementType.FIELD, ElementType.PARAMETER, ElementType.METHOD})
  public @interface MonitoringTriggerStore {

  }

  @Provides
  @Inject
  @MonitoringTriggerStore
  protected Store<Bytes, TriggerMeasurement, TriggerMeasurement> getMonitoringTriggerStore(
      OrderedKeyPrefixedSerdeSupplier<KeyType, MonitoringMessage> keySerdeSupplier,
      UberSerde<TriggerMeasurement> triggerMeasurementUberSerde) {
    return new Store<>(
        MONITORING_TRIGGER_STORE_NAME,
        keySerdeSupplier.get(),
        triggerMeasurementUberSerde,
        triggerMeasurementUberSerde,
        false
    );
  }

  @BindingAnnotation
  @Retention(RUNTIME)
  @Target({ElementType.FIELD, ElementType.PARAMETER, ElementType.METHOD})
  public @interface TriggerActionsStore {

  }

  @Provides
  @Inject
  @TriggerActionsStore
  protected Store<Bytes, TriggerActions, TriggerActions> getTriggerActionsStore(OrderedKeyPrefixedSerdeSupplier<KeyType, MonitoringMessage> keySerdeSupplier, UberSerde<TriggerActions> triggerActionsUberSerde) {
    return new Store<>(
        TRIGGER_ACTIONS_STORE_NAME,
        keySerdeSupplier.get(),
        triggerActionsUberSerde,
        triggerActionsUberSerde,
        false
    );
  }

  @BindingAnnotation
  @Retention(RUNTIME)
  @Target({ElementType.FIELD, ElementType.PARAMETER, ElementType.METHOD})
  public @interface TriggerEventsStore {

  }

  @Provides
  @Inject
  @TriggerEventsStore
  protected Store<Bytes, VerifiableTriggerInfo, VerifiableTriggerInfo> getTriggerEventsStore(OrderedKeyPrefixedSerdeSupplier<KeyType, MonitoringMessage> keySerdeSupplier,
                                                                                              UberSerde<VerifiableTriggerInfo> verifiableTriggerUberSerde) {
    return new Store<>(
        TRIGGER_EVENTS_STORE_NAME,
        keySerdeSupplier.get(),
        verifiableTriggerUberSerde,
        verifiableTriggerUberSerde,
        false
    );
  }


  @BindingAnnotation
  @Retention(RUNTIME)
  @Target({ElementType.FIELD, ElementType.PARAMETER, ElementType.METHOD})
  public @interface KSTREAM_OUTEROTHER_90 {
  }

  // TODO(norwood): fix this once we can name our internal join topics
  @Provides
  @Inject
  @KSTREAM_OUTEROTHER_90
  protected Store<WindowedClusterGroup, MonitoringMessage, TriggerMeasurement> getKstreamOuterOther(
      StreamsConfig streamsConfig,
      UberSerde<WindowedClusterGroup> windowedClusterGroupSerde,
      UberSerde<MonitoringMessage> monitoringMessageSerde,
      UberSerde<TriggerMeasurement> triggerMeasurementUberSerde
  ) {
    return new Store<>(
        KSTREAM_OUTEROTHER_90,
        windowedClusterGroupSerde,
        monitoringMessageSerde,
        triggerMeasurementUberSerde,
        new WindowStoreSerde<>(windowedClusterGroupSerde), // this is to get us compact,delete on the created topic
        false
    );
  }

  @BindingAnnotation
  @Retention(RUNTIME)
  @Target({ElementType.FIELD, ElementType.PARAMETER, ElementType.METHOD})
  public @interface KSTREAM_OUTHERTHIS_89 {
  }

  // TODO(norwood): fix this once we can name our internal join topics
  @Provides
  @Inject
  @KSTREAM_OUTHERTHIS_89
  protected Store<WindowedClusterGroup, MonitoringMessage, TriggerMeasurement>  getKstreamOuterthis(
      StreamsConfig streamsConfig,
      UberSerde<WindowedClusterGroup> windowedClusterGroupSerde,
      UberSerde<MonitoringMessage> monitoringMessageSerde,
      UberSerde<TriggerMeasurement> triggerMeasurementUberSerde
  ) {
    return new Store<>(
        KSTREAM_OUTERTHIS_89,
        windowedClusterGroupSerde,
        monitoringMessageSerde,
        triggerMeasurementUberSerde,
        new WindowStoreSerde<>(windowedClusterGroupSerde), // this is to get us compact,delete on the created topic
        false
    );
  }

  @BindingAnnotation
  @Retention(RUNTIME)
  @Target({ElementType.FIELD, ElementType.PARAMETER, ElementType.METHOD})
  public @interface AlertHistoryStore {

  }

  @Provides
  @Inject
  @AlertHistoryStore
  protected Store<Bytes, AlertInfo, AlertInfo> getAlertHistoryStore(OrderedKeyPrefixedSerdeSupplier<KeyType, MonitoringMessage> keySerdeSupplier,
                                                                    UberSerde<AlertInfo> alertInfoUberSerde) {
    return new Store<>(
        ALERT_HISTORY_STORE_NAME,
        keySerdeSupplier.get(),
        alertInfoUberSerde,
        alertInfoUberSerde,
        false
    );
  }

  @BindingAnnotation
  @Retention(RUNTIME)
  @Target({ElementType.FIELD, ElementType.PARAMETER, ElementType.METHOD})
  public @interface GroupStore {

  }

  @Provides
  @Inject
  @GroupStore
  protected Store<Bytes, WindowedGrouping, WindowedGrouping> getGroupStore(OrderedKeyPrefixedSerdeSupplier<KeyType, MonitoringMessage> keySerdeSupplier, UberSerde<WindowedGrouping> windowedGroupingUberSerde) {
    return new Store<>(
        GROUP_STORE_NAME,
        keySerdeSupplier.get(),
        windowedGroupingUberSerde,
        windowedGroupingUberSerde,
        new WindowStoreSerde<>(keySerdeSupplier.get()),
        true
    );
  }

  @BindingAnnotation
  @Retention(RUNTIME)
  @Target({ElementType.FIELD, ElementType.PARAMETER, ElementType.METHOD})
  public @interface ClusterStore {

  }

  @Provides
  @Inject
  @ClusterStore
  protected Store<Bytes, Cluster, Cluster> getClusterStore(
      OrderedKeyPrefixedSerdeSupplier<ClusterContext, Cluster> keySerdeSupplier,
      UberSerde<Cluster> clusterSerde
  ) {
    return new Store<>(
        CLUSTER_STORE_NAME,
        keySerdeSupplier.get(),
        clusterSerde,
        clusterSerde,
        null,
        false
    );
  }

  @BindingAnnotation
  @Retention(RUNTIME)
  @Target({ElementType.FIELD, ElementType.PARAMETER, ElementType.METHOD})
  public @interface MonitoringStreamStore {

  }

  @Provides
  @Inject
  @MonitoringStreamStore
  protected Store<Bytes, MonitoringMessage, MonitoringMessage> getMonitoringStreamStore(
      OrderedKeyPrefixedSerdeSupplier<KeyType, MonitoringMessage> keySerdeSupplier,
      UberSerde<MonitoringMessage> monitoringMessageSerde
  ) {
    return new Store<>(
        MONITORING_STREAM_STORE_NAME,
        keySerdeSupplier.get(),
        monitoringMessageSerde,
        monitoringMessageSerde,
        new WindowStoreSerde<>(keySerdeSupplier.get()),
        true
    );
  }

  @BindingAnnotation
  @Retention(RUNTIME)
  @Target({ElementType.FIELD, ElementType.PARAMETER, ElementType.METHOD})
  public @interface MonitoringMessageAggregatorWindowsStore {

  }

  @Provides
  @Inject
  @MonitoringMessageAggregatorWindowsStore
  protected Store<Bytes, MonitoringMessage, MonitoringMessage> getMonitoringMessageAggregatorWindowsStore(
      OrderedKeyPrefixedSerdeSupplier<KeyType, MonitoringMessage> keySerdeSupplier,
      UberSerde<MonitoringMessage> monitoringMessageSerde
  ) {
    return new Store<>(
        MONITORING_AGGREGATOR_WINDOWS_STORE_NAME,
        keySerdeSupplier.get(),
        monitoringMessageSerde,
        monitoringMessageSerde,
        new WindowStoreSerde<>(keySerdeSupplier.get()),
        true
    );
  }

  @BindingAnnotation
  @Retention(RUNTIME)
  @Target({ElementType.FIELD, ElementType.PARAMETER, ElementType.METHOD})
  public @interface AggregatedTopicPartitionTableWindowsStore {

  }

  @Provides
  @Inject
  @AggregatedTopicPartitionTableWindowsStore
  protected Store<TopicPartition, MonitoringMessage, ClientGroupProductionAggregate> getAggregatedTopicPartitionTableWindowsStore(
      UberSerde<TopicPartition> topicPartitionSerde,
      UberSerde<MonitoringMessage> monitoringMessageSerde,
      UberSerde<ClientGroupProductionAggregate> clientGroupProductionAggregateUberSerde
  ) {
    return new Store<>(
        AGG_TOPIC_PARTITION_TABLE_WINDOWS_STORE_NAME,
        topicPartitionSerde,
        monitoringMessageSerde,
        clientGroupProductionAggregateUberSerde,
        new WindowStoreSerde<>(topicPartitionSerde),
        true
    );
  }

  @Provides
  protected Set<Topic<?, VerifiableMonitoringMessage, String, VerifiableMonitoringMessage>> getVerifiableTopics(final Injector injector) {
    Set<Topic<?, VerifiableMonitoringMessage, String, VerifiableMonitoringMessage>> result = new HashSet<>();
    for (Map.Entry<Key<?>, Binding<?>> e : injector.getAllBindings().entrySet()) {
      TypeLiteral<?> typeLiteral = e.getKey().getTypeLiteral();
      if (typeLiteral.getRawType().equals(Topic.class)) {
        assert (typeLiteral.getType() instanceof ParameterizedType);
        ParameterizedType type = (ParameterizedType) typeLiteral.getType();
        Type[] pTypes = type.getActualTypeArguments();
        assert pTypes.length == 3;
        if (pTypes[1].equals(VerifiableMonitoringMessage.class) && pTypes[2].equals(String.class)) {
          result.add((Topic<?, VerifiableMonitoringMessage, String, VerifiableMonitoringMessage>) e.getValue().getProvider().get());
        }
      }
    }
    return result;
  }

  @BindingAnnotation
  @Retention(RUNTIME)
  @Target({ElementType.FIELD, ElementType.PARAMETER, ElementType.METHOD})
  public @interface AllTopics {

  }

  @Provides
  @AllTopics
  protected ImmutableSet<Topic> getTopics(final Injector injector) {
    ImmutableSet.Builder<Topic> builder = ImmutableSet.builder();
    //rhoover: iterating through all bindings because findBindingsByType is implemented by a multimap that only checks exact matches
    for (Map.Entry<Key<?>, Binding<?>> e : injector.getAllBindings().entrySet()) {
      if (e.getKey().getTypeLiteral().getRawType().equals(Topic.class)) {
        builder.add((Topic) e.getValue().getProvider().get());
      }
    }
    return builder.build();
  }

  @BindingAnnotation
  @Retention(RUNTIME)
  @Target({ElementType.FIELD, ElementType.PARAMETER, ElementType.METHOD})
  public @interface AllStores {

  }

  @Provides
  @AllStores
  protected ImmutableSet<Store> getStores(final Injector injector) {
    ImmutableSet.Builder<Store> builder = ImmutableSet.builder();
    //rhoover: iterating through all bindings because findBindingsByType is implemented by a multimap that only checks exact matches
    for (Map.Entry<Key<?>, Binding<?>> e : injector.getAllBindings().entrySet()) {
      if (e.getKey().getTypeLiteral().getRawType().equals(Store.class)) {
        builder.add((Store) e.getValue().getProvider().get());
      }
    }
    return builder.build();
  }
}
