// (Copyright) [2016 - 2016] Confluent, Inc.
package io.confluent.controlcenter.streams;

import com.google.common.cache.CacheBuilder;
import com.google.common.cache.CacheLoader;
import com.google.common.cache.LoadingCache;
import com.google.common.collect.Maps;
import com.google.common.util.concurrent.RateLimiter;

import org.apache.kafka.clients.consumer.ConsumerRecord;
import org.apache.kafka.common.TopicPartition;
import org.apache.kafka.streams.processor.TimestampExtractor;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.Map;
import java.util.concurrent.TimeUnit;

import io.confluent.controlcenter.record.Controlcenter.TriggerEvent;
import io.confluent.controlcenter.record.Controlcenter.ClientGroup;
import io.confluent.controlcenter.record.Controlcenter.VerifiableMonitoringMessage;
import io.confluent.controlcenter.record.Controlcenter.WindowedGrouping;
import io.confluent.monitoring.record.Monitoring.MonitoringMessage;

/**
 * tries to extract any window as timestamp
 */
public class WindowExtractor implements TimestampExtractor {

  private static final Logger log = LoggerFactory.getLogger(WindowExtractor.class);

  //used for logging
  private final LoadingCache<TopicPartition, RateLimiter> rateLimiter = CacheBuilder.newBuilder()
      .build(new CacheLoader<TopicPartition, RateLimiter>() {
        @Override
        public RateLimiter load(TopicPartition key) throws Exception {
          return RateLimiter.create(1.0/60.0);
        }
      });
  private final Map<TopicPartition, PartitionCommitInfo> partitionCommitInfo = Maps.newHashMap();

  private class PartitionCommitInfo {
    public long eventTsMs;
    public long offset;
    public long processTimeMs;

    public PartitionCommitInfo(long offset, long eventTsMs) {
      this.offset = offset;
      this.eventTsMs = eventTsMs;
      this.processTimeMs = System.currentTimeMillis();
    }
  }

  @Override
  public long extract(ConsumerRecord<Object, Object> record) {
    long timestamp = 0;
    try {
      if (record.value() instanceof MonitoringMessage) {
        timestamp = ((MonitoringMessage) record.value()).getWindow();
      } else if (record.value() instanceof ClientGroup) {
        timestamp = ((ClientGroup) record.value()).getWindow();
      } else if (record.value() instanceof WindowedGrouping) {
        timestamp = ((WindowedGrouping) record.value()).getWindow();
      } else if (record.value() instanceof VerifiableMonitoringMessage) {
        timestamp = ((VerifiableMonitoringMessage) record.value()).getMonitoringMessage().getWindow();
      } else if (record.value() instanceof TriggerEvent) {
        timestamp = ((TriggerEvent) record.value()).getWindow();
      } else {
        log.warn("unable to extract message timestamp");
      }
    } catch (Exception e) {
       log.warn("unable to extract message timestamp");
    }
    TopicPartition tp = new TopicPartition(record.topic(), record.partition());
    if (rateLimiter.getUnchecked(tp).tryAcquire()) {
      double rate = 0;
      if (partitionCommitInfo.containsKey(tp)) {
        PartitionCommitInfo pci = partitionCommitInfo.get(tp);
        long diff = record.offset() - pci.offset;
        rate = diff / ((double)(System.currentTimeMillis() - pci.processTimeMs) / TimeUnit.SECONDS.toMillis(1));
      }
      log.info("extracted={} topic={} partition={} offset={} timestamp={} rate={} msgs/s", timestamp, record.topic(), record.partition(), record.offset(), record.timestamp(), String.format("%.2f", rate));

      partitionCommitInfo.put(tp, new PartitionCommitInfo(record.offset(), record.timestamp()));
    } else if (log.isTraceEnabled()) {
      log.trace("extracted={} topic={} partition={} offset={} timestamp={}", timestamp, record.topic(), record.partition(), record.offset(), record.timestamp());
    }

    return timestamp;
  }
}
