package io.confluent.controlcenter.streams.alert;

import org.apache.kafka.streams.KeyValue;
import org.apache.kafka.streams.processor.ProcessorSupplier;
import org.apache.kafka.streams.processor.AbstractProcessor;
import org.apache.kafka.streams.processor.Processor;
import org.apache.kafka.streams.processor.ProcessorContext;
import org.apache.kafka.streams.state.KeyValueIterator;
import org.apache.kafka.streams.state.KeyValueStore;

import io.confluent.controlcenter.alert.record.Alert;
import io.confluent.controlcenter.util.KVQuery;
import io.confluent.serializers.OrderedKeyUberSerde;
import org.apache.kafka.common.utils.Bytes;
import io.confluent.controlcenter.alert.record.Alert.AlertInfo;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.concurrent.TimeUnit;

/**
 * Stores Alert history in a state store
 */
public class AlertHistoryProcessorSupplier implements
                                           ProcessorSupplier<Bytes, AlertInfo>{

  private static final Logger log = LoggerFactory.getLogger(AlertHistoryProcessorSupplier.class);

  private static final long DEFAULT_RETAIN_MS = TimeUnit.DAYS.toMillis(7);
  private static final long DEFAULT_PUNCTUATE_INTERVAL_MS = TimeUnit.MINUTES.toMillis(10);
  private static final long MAX_PUNCTUATE_ITEMS = 1000;

  private final String storeName;
  private final OrderedKeyUberSerde<AlertInfo> alertKeySerde;
  private final long storeRetainMs;
  private final long punctuateIntervalMs;
  private final long maxPunctuateItems;
  private final OrderedKeyUberSerde<AlertInfo> alertRangeSerde;

  public AlertHistoryProcessorSupplier(String storeName,
                                       long storeRetainMs,
                                       long punctuateIntervalMs,
                                       long maxPunctuateItems,
                                       OrderedKeyUberSerde<AlertInfo> alertKeySerde,
                                       OrderedKeyUberSerde<AlertInfo> alertRangeSerde) {
    this.storeName = storeName;
    this.storeRetainMs = storeRetainMs;
    this.punctuateIntervalMs = punctuateIntervalMs;
    this.maxPunctuateItems = maxPunctuateItems;
    this.alertKeySerde = alertKeySerde;
    this.alertRangeSerde = alertRangeSerde;
  }

  public AlertHistoryProcessorSupplier(String storeName, OrderedKeyUberSerde<AlertInfo> alertKeySerde, OrderedKeyUberSerde<AlertInfo> alertRangeSerde) {
    this(storeName, DEFAULT_RETAIN_MS, DEFAULT_PUNCTUATE_INTERVAL_MS, MAX_PUNCTUATE_ITEMS,
         alertKeySerde, alertRangeSerde);
  }

  @Override
  public Processor<Bytes, AlertInfo> get() {
    return new AbstractProcessor<Bytes, AlertInfo>() {

      // key: timestamp, alert info
      private KeyValueStore<Bytes, AlertInfo> kvStore;
      private long lastPunctuateTimeMs = -1;

      @Override
      public void init(ProcessorContext context) {
        super.init(context);
        this.kvStore = (KeyValueStore<Bytes, AlertInfo>) context.getStateStore(storeName);
      }

      @Override
      public void process(Bytes key, AlertInfo value) {
        log.trace("Store alert key={}  value={}", key, value);
        kvStore.put(key, value);

        // punctuate on alert timestamp
        punctuate(value.getTimestamp());
      }

      @Override
      public void punctuate(long timestamp) {
        if (lastPunctuateTimeMs < 0) {
          // we will do the first punctuate after 'punctuate interval'
          lastPunctuateTimeMs = timestamp;
        }
        if (timestamp - lastPunctuateTimeMs < punctuateIntervalMs) {
          // we don't want to punctuate more often than 'punctuate interval'
          return;
        }
        lastPunctuateTimeMs = timestamp;

        AlertInfo alertInfoStopKey = AlertInfo.newBuilder()
            .setTimestamp(timestamp - storeRetainMs)
            .build();
        Bytes stop = alertRangeSerde.key(alertInfoStopKey);

        try (KeyValueIterator<Bytes, AlertInfo> iter = KVQuery.rangeFrom(kvStore, stop)) {
          int count = 0;
          while (iter.hasNext()) {
            KeyValue<Bytes, AlertInfo> entry = iter.next();
            log.debug("Removing alert info key={}, timestamp={}, guid={} from store={}",
                      entry.key, entry.value.getTimestamp(), entry.value.getGuid(), storeName);
            kvStore.delete(entry.key);
            count++;
            if (count >= maxPunctuateItems) {
              log.debug("stopping punctuate (after removing max items) at key={} store={}",
                        entry.key, storeName);
              break;
            }
            if (!iter.hasNext()) {
              log.debug("finished all punctuation for store={}", storeName);
            }
          }
        } catch (Exception e) {
          log.error("Range query failed for removing items from store " + storeName, e);
        }
      }
    };
  }
}
