package io.confluent.controlcenter.streams.alert;

import com.google.common.collect.Iterables;
import com.google.inject.Provider;

import org.apache.kafka.common.utils.Bytes;
import org.apache.kafka.streams.KeyValue;
import org.apache.kafka.streams.kstream.Transformer;
import org.apache.kafka.streams.kstream.TransformerSupplier;
import org.apache.kafka.streams.processor.ProcessorContext;
import org.apache.kafka.streams.state.KeyValueIterator;
import org.apache.kafka.streams.state.KeyValueStore;
import org.apache.kafka.streams.state.ReadOnlyKeyValueStore;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.SortedMap;
import java.util.TreeMap;
import java.util.UUID;
import java.util.concurrent.TimeUnit;

import io.confluent.command.record.Command;
import io.confluent.command.record.Command.CommandKey;
import io.confluent.command.record.Command.CommandMessage;
import io.confluent.command.record.alert.CommandAlert.ActionConfig;
import io.confluent.command.record.alert.CommandAlert.MonitoringTriggerConfig;
import io.confluent.controlcenter.alert.record.Alert.ActionInfo;
import io.confluent.controlcenter.alert.record.Alert.AlertInfo;
import io.confluent.controlcenter.alert.record.Alert.TriggerInfo;
import io.confluent.controlcenter.record.Controlcenter.TriggerAction;
import io.confluent.controlcenter.record.Controlcenter.TriggerActions;
import io.confluent.controlcenter.record.Controlcenter.TriggerEvent;
import io.confluent.controlcenter.record.Controlcenter.TriggerEventKey;
import io.confluent.controlcenter.record.Controlcenter.VerifiableTriggerInfo;
import io.confluent.controlcenter.util.KVQuery;
import io.confluent.monitoring.common.Clock;
import io.confluent.monitoring.common.TimeBucket;
import io.confluent.serializers.OrderedKeyUberSerde;

/**
 * This transformer takes the stream of trigger events and emits alerts
 */
public class AlertTransformerSupplier implements
    TransformerSupplier<Bytes, TriggerEvent, KeyValue<Bytes, Iterable<KeyValue<Bytes, AlertInfo>>>>{

  private static final Logger log = LoggerFactory.getLogger(AlertTransformerSupplier.class);

  private static final long PUNCTUATE_INTERVAL_MS = TimeBucket.SIZE;
  private static final long DEFAULT_MAX_PUNCTUATE_MS = TimeUnit.SECONDS.toMillis(1);

  private final long maxTriggerEventsInOneAlert;
  private final String storeName;
  private final String triggerEventsStoreName;
  private final Clock clock;
  private final OrderedKeyUberSerde<TriggerEventKey> actionsKeySerde;
  private final OrderedKeyUberSerde<AlertInfo> eventsKeySerde;

  private final Provider<? extends ReadOnlyKeyValueStore<CommandKey, CommandMessage>> triggerConfigsProvider;
  private final Provider<? extends ReadOnlyKeyValueStore<CommandKey, CommandMessage>> actionConfigsProvider;

  public AlertTransformerSupplier(
      String storeName,
      String triggerEventsStoreName,
      int maxTriggerEventsInOneAlert,
      Clock clock,
      OrderedKeyUberSerde<TriggerEventKey> actionsKeySerde,
      OrderedKeyUberSerde<AlertInfo> eventsKeySerde,
      Provider<? extends ReadOnlyKeyValueStore<CommandKey, CommandMessage>> triggerConfigsProvider,
      Provider<? extends ReadOnlyKeyValueStore<CommandKey, CommandMessage>> actionConfigsProvider
  ) {
    this.storeName = storeName;
    this.triggerEventsStoreName = triggerEventsStoreName;
    this.clock = clock;
    this.maxTriggerEventsInOneAlert = maxTriggerEventsInOneAlert;
    this.actionsKeySerde = actionsKeySerde;
    this.eventsKeySerde = eventsKeySerde;
    this.triggerConfigsProvider = triggerConfigsProvider;
    this.actionConfigsProvider = actionConfigsProvider;
  }

  @Override
  public Transformer<Bytes, TriggerEvent, KeyValue<Bytes, Iterable<KeyValue<Bytes, AlertInfo>>>> get() {
    return new AlertTransformer(PUNCTUATE_INTERVAL_MS,
                                DEFAULT_MAX_PUNCTUATE_MS);
  }

  private class AlertTransformer implements
    Transformer<Bytes, TriggerEvent, KeyValue<Bytes, Iterable<KeyValue<Bytes, AlertInfo>>>> {
    private final long maxPunctuateMs;
    private final long punctuateIntervalMs;

    private ProcessorContext context;

    /**
     * Approximated C3 time is tracked based on timestamps in incoming trigger events (including
     * heartbeat trigger event). The time is calculated in MonitoringTriggerTransformerSupplier
     */
    private long approxC3Time = 0;

    // Stores alert state for all the actions defined for a trigger
    // key = trigger guid
    private KeyValueStore<Bytes, TriggerActions> kvStore;
    // Stores trigger events that are waiting for action timeout
    // key = trigger guid, timestamp, window
    private KeyValueStore<Bytes, VerifiableTriggerInfo> kvTriggerEventStore;

    private long lastPunctuateTimeMs = -1;
    private Bytes lastPunctuateKey = null;
    private final OrderedKeyUberSerde<TriggerEventKey> actionStartKeySerde = actionsKeySerde.prefixKeySerde(2);
    private final OrderedKeyUberSerde<TriggerEventKey> actionPrefixKeySerde = actionsKeySerde.prefixKeySerde(1);

    public AlertTransformer(long punctuateIntervalMs, long maxPunctuateMs) {
      this.punctuateIntervalMs = punctuateIntervalMs;
      this.maxPunctuateMs = maxPunctuateMs;
    }

    @SuppressWarnings("unchecked")
    @Override
    public void init(ProcessorContext context) {
      this.context = context;
      kvStore = (KeyValueStore<Bytes, TriggerActions>) this.context.getStateStore(storeName);
      kvTriggerEventStore = (KeyValueStore<Bytes, VerifiableTriggerInfo>)
          this.context.getStateStore(triggerEventsStoreName);
    }

    @Override
    public KeyValue<Bytes, Iterable<KeyValue<Bytes, AlertInfo>>> transform(
        Bytes key, TriggerEvent triggerEvent
    ) {
      // move C3 clock based on time set in trigger events
      if (triggerEvent.getTimestamp() > approxC3Time) {
        approxC3Time = triggerEvent.getTimestamp();
      }

      if (triggerEvent.getHeartbeat()) {
        log.trace("Alert heartbeat with timestamp={}", triggerEvent.getTimestamp());
        return punctuate(approxC3Time);
      }

      TriggerEventKey triggerEventKey = TriggerEventKey.newBuilder()
          .setTriggerGuid(triggerEvent.getTriggerGuid())
          .build();
      Bytes storeKey = actionsKeySerde.prefixKeySerde(1).key(triggerEventKey);

      log.trace("Alert transform storeKey={} trigger event={}", storeKey, triggerEvent);

      List<KeyValue<Bytes, AlertInfo>> ret = new ArrayList<>();

      try {
        TriggerActions storedActions = kvStore.get(storeKey);
        TriggerActions actions = updateTriggerActionsFromConfig(triggerEvent.getTriggerGuid(),
                                                                storedActions);
        if (actions == null) {
          log.trace("Ignoring trigger={}, no actions/trigger defined", triggerEvent.getTriggerGuid());
          if (storedActions != null) {
            // trigger/action used to be defined but not anymore, so drop queued trigger events
            cleanupTriggerState(triggerEvent.getTriggerGuid(), storeKey);
          }
          return null;
        }

        log.trace("Actions for trigger={}: {}", triggerEvent.getTriggerGuid(), actions);
        actions = maybeCreateAlerts(approxC3Time, triggerEvent, actions, ret);
        if (!actions.equals(storedActions)) {
          log.trace("Store trigger guid={}, actions={}", triggerEvent.getTriggerGuid(), actions);
          kvStore.put(storeKey, actions);
        }
      } catch (Exception e) {
        log.warn("Failed to update trigger event store. May not send an alert when needed.", e);
      }

      KeyValue<Bytes, Iterable<KeyValue<Bytes, AlertInfo>>> pair = punctuate(approxC3Time);
      if (pair != null) {
        Iterables.addAll(ret, pair.value);
      }
      if (ret.size() == 0) {
        return null;
      }

      return KeyValue.pair(
          Bytes.wrap(new byte[]{}),
          (Iterable<KeyValue<Bytes, AlertInfo>>) ret
      );
    }

    @Override
    public KeyValue<Bytes, Iterable<KeyValue<Bytes, AlertInfo>>> punctuate(long timestamp) {
      if (lastPunctuateTimeMs < 0) {
        // we will do the first punctuate after 'punctuate interval'
        lastPunctuateTimeMs = timestamp;
      }
      if (timestamp - lastPunctuateTimeMs < punctuateIntervalMs) {
        // we don't want to punctuate more often than 'punctuate interval'
        return null;
      }

      long now = clock.currentTimeMillis();
      lastPunctuateTimeMs = timestamp;
      long punctuateTimeout = now + maxPunctuateMs;

      List<KeyValue<Bytes, AlertInfo>> alertList = new ArrayList<>();
      try (KeyValueIterator<Bytes, TriggerActions> iter = KVQuery.rangeFrom(kvStore, lastPunctuateKey)) {
        long count = 0;
        while (iter.hasNext()) {
          KeyValue<Bytes, TriggerActions> entry = iter.next();
          if (entry == null) {
            continue;
          }
          if (entry.value == null) {
            log.debug("removing null value key={} from store={}", entry.key, storeName);
            kvStore.delete(entry.key);
          } else {
            log.trace("checking key={} trigger guid={} store={}",
                      entry.key, entry.value.getTriggerGuid(), storeName);

            // update actions if configs changed
            TriggerActions modifiedActions =
                updateTriggerActionsFromConfig(entry.value.getTriggerGuid(), entry.value);
            if (modifiedActions != null) {
              modifiedActions = maybeCreateAlerts(timestamp, null, modifiedActions, alertList);
              if (!modifiedActions.equals(entry.value)) {
                log.trace("Update key={} trigger actions={}", entry.key, modifiedActions);
                kvStore.put(entry.key, modifiedActions);
              }
            } else {
              // trigger/action used to be defined but not anymore, so drop queued trigger events
              cleanupTriggerState(entry.value.getTriggerGuid(), entry.key);
            }
          }
          lastPunctuateKey = entry.key;
          count++;
          if (clock.currentTimeMillis() > punctuateTimeout) {
            log.debug("stopping punctuate at key={} store={}", lastPunctuateKey, storeName);
            break;
          }
          if (!iter.hasNext()) {
            log.debug("finished all punctuation for store={}", storeName);
            lastPunctuateKey = null;
          }
          log.debug("punctuated on count={} records", count);
        }
      } catch (Exception e) {
        log.error("failure during punctuate for store={}", storeName, e);
      }
      if (alertList.isEmpty()) {
        return null;
      }


      // I think we don't care about the key of kv pair
      return KeyValue.pair(
          Bytes.wrap(new byte[] {}),
          (Iterable<KeyValue<Bytes, AlertInfo>>) alertList
      );
    }

    @Override
    public void close() {
      log.info("closing store={}", storeName);
    }


    /**
     * Removes trigger entry from kvStore and all triggerEvents from kvTriggerEventStore
     * associated with given trigger guid
     * @param triggerGuid trigger GUID
     * @param kvStoreKey key to this.kvstore
     */
    private void cleanupTriggerState(String triggerGuid, Bytes kvStoreKey) {
      // remove all trigger entries from kvTriggerEventStore (from timestamp 0 to last timestamp)
      TriggerEventKey triggerEventStartKey = getTriggerEventStartKey(triggerGuid, 0);
      Bytes start = actionStartKeySerde.key(triggerEventStartKey);
      Bytes prefix = actionPrefixKeySerde.key(triggerEventStartKey);
      try (KeyValueIterator<Bytes, VerifiableTriggerInfo> iter = KVQuery.prefixScan(kvTriggerEventStore, prefix, start)) {
        while (iter.hasNext()) {
          KeyValue<Bytes, VerifiableTriggerInfo> entry = iter.next();
          kvTriggerEventStore.delete(entry.key);
          log.trace("Successfuly removed trigger event={}, trigger guild={} for removed/disabled trigger/action",
                   entry.value, triggerGuid);
        }
        kvStore.delete(kvStoreKey);
        log.trace("Removed trigger state in store={} for trigger={}", storeName, triggerGuid);
      } catch (Exception e) {
        log.error("Failed to remove all trigger events for trigger={}", triggerGuid, e);
      }
    }

    /**
     * Creates alerts when actions time out and adds to 'alertList', and returns updated trigger
     * actions (with updated state).
     */
    private TriggerActions maybeCreateAlerts(long currentTime,
                                             TriggerEvent triggerEvent,
                                             TriggerActions actions,
                                             List<KeyValue<Bytes, AlertInfo>> alertList) {
      // make sure to count triggerEvent (if given) as 'queued'
      long nextTriggerInfoSequence = triggerEvent == null ?
                                     actions.getNextTriggerInfoSequence() :
                                     actions.getNextTriggerInfoSequence() + 1;

      // get the map of sequence to TriggerInfo for all trigger events for which we could send at
      // least one alert
      SortedMap<Long, TriggerInfo> sequenceToTriggerMap = getActionableTriggerEvents(currentTime,
                                                                                     actions,
                                                                                     nextTriggerInfoSequence);
      if (triggerEvent == null &&
          (sequenceToTriggerMap == null || sequenceToTriggerMap.isEmpty())) {
        log.trace("No actionable trigger events for trigger guid={}", actions.getTriggerGuid());
        return actions;
      }

      TriggerActions.Builder actionsBuilder = TriggerActions.newBuilder(actions);
      boolean triggerEventInMap = false;
      if (triggerEvent != null && needKeepTriggerEvent(currentTime, actions)) {
        long nextSequence = actions.getNextTriggerInfoSequence();
        keepTriggerEvent(currentTime, triggerEvent, nextSequence);
        actionsBuilder.setNextTriggerInfoSequence(nextSequence + 1);
        sequenceToTriggerMap.put(nextSequence, triggerEvent.getInfo());
        triggerEventInMap = true;
      }

      long lastSequenceOfTriggerEvents = sequenceToTriggerMap.isEmpty() ?
                                         -1 : sequenceToTriggerMap.lastKey();
      for (int i = 0; i < actions.getActionsCount(); i++) {
        TriggerAction action = actions.getActions(i);
        if (canSendAlert(action, currentTime, nextTriggerInfoSequence)) {
          // can send alert
          Collection<TriggerInfo> triggerList =
              sequenceToTriggerMap.tailMap(action.getLastTriggerInfoSequence() + 1).values();
          if (triggerList.isEmpty() && triggerEvent == null) {
            // no trigger events for this action
            continue;
          }

          CommandMessage triggerMessage = triggerConfigsProvider.get().get(
              CommandKey.newBuilder()
                  .setGuid(actions.getTriggerGuid())
                  .setConfigType(Command.CommandConfigType.MONITORING_TRIGGER_CONFIG)
                  .build()
          );
          if (triggerMessage == null) {
            log.warn("unable to find trigger config with guid={}", actions.getTriggerGuid());
            continue;
          }
          MonitoringTriggerConfig monitoringTriggerConfig = triggerMessage.getMonitoringTriggerConfig();
          if (!monitoringTriggerConfig.getGuid().equals(actions.getTriggerGuid())) {
            log.warn("mismatching guids action={} trigger={}", actions.getTriggerGuid(), monitoringTriggerConfig.getGuid());
            continue;
          }
          AlertInfo.Builder alertBuilder = AlertInfo.newBuilder()
              .setGuid(UUID.randomUUID().toString())
              .setTimestamp(currentTime)
              .setMonitoringTrigger(monitoringTriggerConfig)
              .addAllActions(action.getActionsList())
              .addAllTriggers(triggerList);

          if (triggerEvent != null && !triggerEventInMap) {
            alertBuilder.addTriggers(TriggerInfo.newBuilder(triggerEvent.getInfo()));
          }

          AlertInfo alert = alertBuilder.build();
          log.debug("Created alert={}", alert);
          alertList.add(KeyValue.pair(eventsKeySerde.key(alert), alert));

          // once we send an alert, calculate next time we can send an alert
          // but if we did not send alerts for a long time, we don't want to send a burst of
          // alerts just because we accumulated time; so next time is one 'send period' after the
          // threshold just before the current time
          long nextActionTimestamp = currentTime;  // if period <= 0
          if (action.getSendPeriodMs() > 0) {
            long numPeriods = currentTime / action.getSendPeriodMs();
            nextActionTimestamp = Math.min(action.getSendPeriodMs() * (numPeriods + 2),
                                           currentTime + action.getSendPeriodMs());
          }
          // however, if we are sending alert because we reached max events, this alert is free
          if (triggerList.size() >= maxTriggerEventsInOneAlert) {
            nextActionTimestamp = action.getNextActionTimestamp();
          }

          long lastSequence = lastSequenceOfTriggerEvents >= 0 ?
                              lastSequenceOfTriggerEvents : action.getLastTriggerInfoSequence();
          actionsBuilder.setActions(i, action.toBuilder()
                                        .setLastTriggerInfoSequence(lastSequence)
                                        .setNextActionTimestamp(nextActionTimestamp));
        }
      }

      return actionsBuilder.build();
    }

    /**
     * Stores trigger event with key = trigger guid, timestamp, trigger event window
     * @param currentTime time the trigger event arrived to this transformer
     * @param triggerEvent trigger event to store
     * @param sequence  sequence number to store with the trigger event (in a value)
     */
    private void keepTriggerEvent(long currentTime,
                                  TriggerEvent triggerEvent,
                                  long sequence) {
      TriggerEventKey triggerEventKey = TriggerEventKey.newBuilder()
          .setTriggerGuid(triggerEvent.getTriggerGuid())
          .setTimestamp(currentTime)
          .setWindow(triggerEvent.getInfo().getWindow())
          .build();
      Bytes key = actionsKeySerde.key(triggerEventKey);

      VerifiableTriggerInfo value = VerifiableTriggerInfo.newBuilder()
          .setSequence(sequence)
          .setInfo(TriggerInfo.newBuilder(triggerEvent.getInfo()))
          .build();

      log.trace("Store trigger event={}", value);

      // If there is already a value with the same key, it could be for the same trigger event,
      // so we don't care to check
      kvTriggerEventStore.put(key, value);
    }

    // return true if at at least one action need to timeout before sending alert for a trigger
    // event
    private boolean needKeepTriggerEvent(long currentTime,
                                         TriggerActions actions) {
      for (int i = 0; i < actions.getActionsCount(); i++) {
        TriggerAction action = actions.getActions(i);
        if (currentTime < action.getNextActionTimestamp()) {
          // cannot send alert yet, so have to keep the trigger event
          return true;
        }
      }
      return false;
    }

    /**
     * Return true if we can send an alert for 'action', either because:
     *  1) Frequency allows is
     *  2) Number of trigger events that are queued for this action reached the maximim
     */
    private boolean canSendAlert(TriggerAction action, long currentTime, long nextSequence) {
      return currentTime >= action.getNextActionTimestamp() ||
             (nextSequence - action.getLastTriggerInfoSequence() - 1) >= maxTriggerEventsInOneAlert;
    }

    /**
     * Returns a map of sequence to TriggerInfo for events that are actionable.
     */
    private SortedMap<Long, TriggerInfo> getActionableTriggerEvents(long currentTime,
                                                                    TriggerActions actions,
                                                                    long nextTriggerInfoSequence) {
      // the smallest trigger event timestamp that was last checked by currently alertable trigger
      // actions
      long earliestTimestamp = Long.MAX_VALUE;
      // the earliest last sequence that was last checked by currently alertable trigger actions
      long earliestLastTriggerInfoSequence = Long.MAX_VALUE;
      // the earliest last sequence that was last checked by currently NON-alertable trigger actions
      long earliestNonAlertableLastTriggerSequence = Long.MAX_VALUE;
      for (int i = 0; i < actions.getActionsCount(); i++) {
        TriggerAction action = actions.getActions(i);
        if (canSendAlert(action, currentTime, nextTriggerInfoSequence)) {
          // can send alert (alertable)
          long ts = Math.max(0, action.getNextActionTimestamp() - action.getIntervalMs());
          earliestTimestamp = Math.min(earliestTimestamp, ts);
          earliestLastTriggerInfoSequence = Math.min(earliestLastTriggerInfoSequence,
                                                     action.getLastTriggerInfoSequence());
        } else {
          // cannot send alert (non-alertable)
          earliestNonAlertableLastTriggerSequence = Math.min(earliestNonAlertableLastTriggerSequence,
                                                             action.getLastTriggerInfoSequence());
        }
      }


      // we start with earliest timestamp that was checked by currently alertable trigger actions
      // even though that timestamp was already checked, new trigger events could have arrived
      // with the same timestamp; we will later filter older events by sequence numbers
      TriggerEventKey triggerEventStartKey = getTriggerEventStartKey(actions.getTriggerGuid(), earliestTimestamp);
      Bytes start = actionStartKeySerde.key(triggerEventStartKey);
      Bytes prefix = actionPrefixKeySerde.key(triggerEventStartKey);

      SortedMap<Long, TriggerInfo> sequenceToTriggerMap = new TreeMap<>();
      try (KeyValueIterator<Bytes, VerifiableTriggerInfo> iter = KVQuery.prefixScan(kvTriggerEventStore, prefix, start)) {
        while (iter.hasNext()) {
          KeyValue<Bytes, VerifiableTriggerInfo> entry = iter.next();

          // since trigger events may have the same timestamp and we did range lookup by
          // timestamp only, we need to make sure we don't duplicate same trigger events in
          // alerts by also checking sequence numbers
          long sequence = entry.value.getSequence();
          if (sequence > earliestLastTriggerInfoSequence) {
            sequenceToTriggerMap.put(sequence, entry.value.getInfo());
          }
          // we can remove all trigger events with sequence up to and including earliest
          // non-alertable last checked sequence (since non-alertable actions are the ones that
          // are holding the events till timeouts)
          if (sequence <= earliestNonAlertableLastTriggerSequence) {
            // We may lose some alerts if we control center crashes before we send an alert; we
            // could instead keep events longer and remove older items on punctuate separately.
            kvTriggerEventStore.delete(entry.key);
          }
        }
      } catch (NullPointerException npe) {
        log.error("Unexpected null value in a trigger event store for trigger={}",
                  actions.getTriggerGuid(), npe);
      } catch (Exception e) {
        log.error("Failed to retrieve trigger events, may fail to create alert for trigger={}",
                  actions.getTriggerGuid(), e);
      }

      return sequenceToTriggerMap;
    }

    private TriggerEventKey getTriggerEventStartKey(String triggerGuid, long earliestTimestamp) {
      TriggerEventKey.Builder triggerEventKeyBuilder = TriggerEventKey.newBuilder();
      TriggerEventKey key = triggerEventKeyBuilder
              .setTriggerGuid(triggerGuid)
              .setTimestamp(earliestTimestamp)
              .build();
      triggerEventKeyBuilder.clear();
      return key;
    }

    /**
     * Returns initialized trigger action list for a specified trigger GUID
     * @param triggerGuid trigger GUID
     * @return list of trigger actions
     */
    private List<TriggerAction> actionListForTrigger(String triggerGuid) {
      List<TriggerAction> actionList = new ArrayList<>();
      try (KeyValueIterator<CommandKey, CommandMessage> actionConfigIterator = actionConfigsProvider.get().all()) {
        while (actionConfigIterator.hasNext()) {
          KeyValue<CommandKey, CommandMessage> actionConfigKV = actionConfigIterator.next();
          ActionConfig actionConfig = actionConfigKV.value.getActionConfig();
          if (actionConfig == null) {
            log.warn("actionConfig=null for key={}", actionConfigKV.key);
            continue;
          }

          // consider only actions that include this trigger guid
          if (!actionConfig.getEnabled() ||
              !actionConfig.getTriggerGuidList().contains(triggerGuid)) {
            continue;
          }

          long sendPeriodMs = actionConfig.getMaxSendRateNumerator() <= 0 ?
                              0 : actionConfig.getIntervalMs() / actionConfig.getMaxSendRateNumerator();

          // do not consider actionConfig that we already included earlier in this iteration
          boolean found = false;
          for (TriggerAction action : actionList) {
            if (action.getIntervalMs() == actionConfig.getIntervalMs() &&
                action.getSendPeriodMs() == sendPeriodMs) {
              found = true;
            }
          }
          if (found) {
            continue;
          }

          // check if this action can be merged in the existing TriggerAction
          TriggerAction.Builder triggerActionBuilder = TriggerAction.newBuilder()
              .setIntervalMs(actionConfig.getIntervalMs())
              .setSendPeriodMs(sendPeriodMs)
              .addActions(ActionInfo.newBuilder()
                              .setGuid(actionConfig.getGuid())
                              .setName(actionConfig.getName())
                              .setEmail(actionConfig.getEmail())
              )
              .setNextActionTimestamp(0)
              .setLastTriggerInfoSequence(-1);

          try (KeyValueIterator<CommandKey, CommandMessage> actionConfigIterator2 = actionConfigsProvider.get().all()) {
            while (actionConfigIterator2.hasNext()) {
              KeyValue<CommandKey, CommandMessage> actionConfigKV2 = actionConfigIterator2.next();
              ActionConfig actionConfig2 = actionConfigKV2.value.getActionConfig();
              if (actionConfig2 == null) {
                log.warn("actionConfig2=null for key={}", actionConfigKV2.key);
                continue;
              }
              if (!actionConfig.getGuid().equals(actionConfig2.getGuid()) &&
                  actionConfig2.getTriggerGuidList().contains(triggerGuid) &&
                  hasSameFrequency(actionConfig, actionConfig2)) {
                triggerActionBuilder.addActions(ActionInfo.newBuilder()
                                                    .setGuid(actionConfig2.getGuid())
                                                    .setName(actionConfig2.getName())
                                                    .setEmail(actionConfig2.getEmail())
                );
              }
            }
            actionList.add(triggerActionBuilder.build());
          } catch (Exception e) {
            log.warn("Action list for trigger={} maybe incomplete", triggerGuid, e);
          }
        }
      } catch (Exception e) {
        log.error("Failed to build action list for trigger={}", triggerGuid, e);
      }

      return actionList.isEmpty() ? null : actionList;
    }


    /**
     * Update 'actions' from config
     * @param triggerGuid trigger guid of trigger actions
     * @param actions actions currently stored in kvstpre
     * @return up-to-date actions (could be equal to 'actions')
     */
    private TriggerActions updateTriggerActionsFromConfig(String triggerGuid,
                                                          TriggerActions actions) {
      // first make sure that trigger is defined
      CommandMessage triggerMessage = triggerConfigsProvider.get().get(
          CommandKey.newBuilder()
              .setGuid(triggerGuid)
              .setConfigType(Command.CommandConfigType.MONITORING_TRIGGER_CONFIG)
              .build()
      );

      if (triggerMessage == null) {
        // this can only happen if trigger was removed while triggered event was in flight
        log.debug("Unable to find trigger config with guid={}", triggerGuid);
        return null;
      }
      MonitoringTriggerConfig monitoringTriggerConfig = triggerMessage.getMonitoringTriggerConfig();
      if (!monitoringTriggerConfig.getGuid().equals(triggerGuid)) {
        log.warn("mismatching guids key={} trigger={}", triggerGuid, monitoringTriggerConfig.getGuid());
        return null;
      }

      // build the list of actions for this trigger
      List<TriggerAction> actionList = actionListForTrigger(triggerGuid);
      if (actionList != null && actions == null) {
        // first time initializing trigger actions
        log.trace("Initializing actions for trigger={}", triggerGuid);
        return TriggerActions.newBuilder()
            .setTriggerGuid(triggerGuid)
            .setNextTriggerInfoSequence(0)
            .addAllActions(actionList)
            .build();
      } else if (actionList == null) {
        log.trace("No actions defined for trigger={}", triggerGuid);
        return null;
      }

      // compare currently stored actions to what's currently configured
      List<TriggerAction> updatedActionList = new ArrayList<>();
      for (TriggerAction configuredAction: actionList) {
        boolean found = false;
        for (TriggerAction storedAction: actions.getActionsList()) {
          if (hasSameFrequency(configuredAction, storedAction)) {
            // already have action with the same frequency
            updatedActionList.add(TriggerAction.newBuilder(configuredAction)
                                      .setNextActionTimestamp(storedAction.getNextActionTimestamp())
                                      .setLastTriggerInfoSequence(storedAction.getLastTriggerInfoSequence())
                                      .build());
            found = true;
            break;
          }
        }
        if (!found) {
          // new action or action changed frequency
          updatedActionList.add(configuredAction);
        }
        // if action was removed, we are just not adding it to updatedActionList
      }

      return TriggerActions.newBuilder()
          .setTriggerGuid(actions.getTriggerGuid())
          .setNextTriggerInfoSequence(actions.getNextTriggerInfoSequence())
          .addAllActions(updatedActionList)
          .build();
    }


    private boolean hasSameFrequency(TriggerAction action1, TriggerAction action2) {
      return (action1.getIntervalMs() == action2.getIntervalMs() &&
              action1.getSendPeriodMs() == action2.getSendPeriodMs());
    }

    private boolean hasSameFrequency(ActionConfig actionConfig1, ActionConfig actionConfig2) {
      return (actionConfig1.getIntervalMs() == actionConfig2.getIntervalMs() &&
              actionConfig1.getMaxSendRateNumerator() == actionConfig2.getMaxSendRateNumerator());
    }

  }
}
