package io.confluent.controlcenter.streams.alert;

import org.apache.kafka.streams.kstream.ValueJoiner;

import io.confluent.controlcenter.record.Controlcenter.ClientTriggerMeasurement;
import io.confluent.controlcenter.record.Controlcenter.TriggerMeasurement;
import io.confluent.monitoring.record.Monitoring.ClientType;
import io.confluent.monitoring.record.Monitoring.MessageType;
import io.confluent.monitoring.record.Monitoring.MonitoringMessage;

/**
 * Joins values from expected consumption and actual consumption streams
 */
public class MonitoringTriggerMeasurementJoiner
  implements ValueJoiner<MonitoringMessage, MonitoringMessage, TriggerMeasurement> {

  @Override
  public TriggerMeasurement apply(MonitoringMessage actualConsumptionValue,
                                  MonitoringMessage expectedConsumptionValue) {
    if (actualConsumptionValue == null) {
      return null;
    }
    if (expectedConsumptionValue == null) {
      expectedConsumptionValue = MonitoringMessage.newBuilder(actualConsumptionValue)
          .clearCount()
          .clearTotalLatency()
          .clearMaxLatency()
          .build();
    }
    if (expectedConsumptionValue.getClientType() == ClientType.CONTROLCENTER &&
        actualConsumptionValue.getClientType() == ClientType.CONTROLCENTER) {
      return TriggerMeasurement.newBuilder()
          .setHeartbeat(true)
          .build();
    }

    // If both actual and expected consumption have zero count and at least one side has an error
    // state, ignore this measurement. This covers the following use-cases:
    // 1) Either consumer or producer had unclean shutdown, and a window after we stopped getting
    // monitoring data was set in an error state. However, eg. if that happened to a consumer but
    // there was no production, then most likely we did not actually lose any monitoring data;
    // similar if both consumer and producer shut down uncleanly, the state is unknown and
    // possibly nothing was produced and consumed.
    // 2) However, if we identified incomplete monitoring data before we got any monitoring data
    // on either expected or actual consumption side and we filter that measurement out; and then
    // we get some monitoring data for the same window, we will get accumulated measurement with
    // an error state here and pass it downstream. So, the initial filtering does not effect this
    // use-case.
    if ((expectedConsumptionValue.getType() == MessageType.ERROR ||
         actualConsumptionValue.getType() == MessageType.ERROR) &&
        (expectedConsumptionValue.getCount() == 0 && actualConsumptionValue.getCount() == 0)) {
      return null;
    }

    ClientTriggerMeasurement.Builder clientMeasurementBuilder = ClientTriggerMeasurement.newBuilder()
        .setArrivalTime(Math.max(expectedConsumptionValue.getArrivalTime(), actualConsumptionValue
            .getArrivalTime()))
        .setConsumptionDiff(expectedConsumptionValue.getCount() - actualConsumptionValue.getCount())
        .setMaxLatency(actualConsumptionValue.getMaxLatency())
        .setAvgLatency(actualConsumptionValue.getCount() > 0 ?
                       actualConsumptionValue.getTotalLatency() / actualConsumptionValue.getCount() : 0)
        .setHasError(expectedConsumptionValue.getType() == MessageType.ERROR ||
                     actualConsumptionValue.getType() == MessageType.ERROR);

    return TriggerMeasurement.newBuilder()
        .setClusterId(expectedConsumptionValue.getClusterId())
        .setGroup(expectedConsumptionValue.getGroup())
        .setWindow(-1)     // window is set downstream
        .setClientMeasurement(clientMeasurementBuilder)
        .setHeartbeat(false)
        .build();
  }
}
