// (Copyright) [2016 - 2016] Confluent, Inc.
package io.confluent.controlcenter.streams.cluster;

import com.google.inject.Inject;
import io.confluent.controlcenter.record.Controlcenter.Cluster;
import io.confluent.controlcenter.record.Controlcenter.ClusterContext;
import io.confluent.controlcenter.record.Controlcenter.ClusterType;
import io.confluent.controlcenter.serialization.OrderedKeyPrefixedSerdeSupplier;
import io.confluent.controlcenter.streams.TopicStoreMaster.Store;
import io.confluent.controlcenter.streams.TopicStoreMaster.Topic;
import io.confluent.controlcenter.streams.TopicStoreModule.ClusterStore;
import io.confluent.controlcenter.streams.TopicStoreModule.ClusterStreamExtensionTopic;
import io.confluent.monitoring.record.Monitoring.MonitoringMessage;
import io.confluent.serializers.OrderedKeyPrefixedSerde;
import org.apache.kafka.common.utils.Bytes;
import org.apache.kafka.streams.KeyValue;
import org.apache.kafka.streams.kstream.Initializer;
import org.apache.kafka.streams.kstream.KStream;
import org.apache.kafka.streams.kstream.KTable;
import org.apache.kafka.streams.kstream.KeyValueMapper;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 *
 * Calculates the set of clusters we know about
 *
 */
public class ClusterStreamExtension {

  private static final Logger log = LoggerFactory.getLogger(ClusterStreamExtension.class);
  private final Store<Bytes, Cluster, Cluster> clusterStore;
  private final Topic<Bytes, Cluster, Void, Cluster> clusterRekeyTopic;
  private final OrderedKeyPrefixedSerde<ClusterContext, Cluster> clusterKeySerde;

  @Inject
  public ClusterStreamExtension(
      @ClusterStore Store<Bytes, Cluster, Cluster> clusterStore,
      @ClusterStreamExtensionTopic Topic<Bytes, Cluster, Void, Cluster> clusterRekeyTopic,
      OrderedKeyPrefixedSerdeSupplier<ClusterContext, Cluster> keySerdeSupplier
  ) {
    this.clusterStore = clusterStore;
    this.clusterRekeyTopic = clusterRekeyTopic;
    this.clusterKeySerde = keySerdeSupplier.get(ClusterContext.STREAM_MONITORING);
  }

  public KTable<Bytes, Cluster> extend(KStream<byte[], MonitoringMessage> stream) {
    return stream
        .filter(new ClusterListCheck())
        // Convert monitoring message to a Cluster message
        .map(new KeyValueMapper<byte[], MonitoringMessage, KeyValue<Bytes, Cluster>>() {
          private final Cluster.Builder builder = Cluster.newBuilder();

          @Override
          public KeyValue<Bytes, Cluster> apply(byte[] bytes, MonitoringMessage monitoringMessage) {
            builder
                .setClusterType(ClusterType.KAFKA)
                .setClusterId(monitoringMessage.getClusterId())
                .setClusterName(monitoringMessage.getClusterName())
                .build();
            Cluster cluster = builder.build();
            builder.clear();
            return KeyValue.pair(clusterKeySerde.key(cluster), cluster);
          }
        })
        // Re-key by cluster id - duplicate production to this topic is ok because we're computing a set
        .through(clusterRekeyTopic.keySerde, clusterRekeyTopic.valueSerde, clusterRekeyTopic.name)
        // Group records for each cluster
        .groupByKey(clusterStore.keySerde, clusterStore.valueSerde)
        // Combine records for each cluster
        .aggregate(
            new Initializer<Cluster>() {
              @Override
              public Cluster apply() {
                return Cluster.getDefaultInstance();
              }
            },
            new ClusterAggregator(),
            clusterStore.aggregateSerde,
            clusterStore.name
        );
  }

}
