// (Copyright) [2016 - 2016] Confluent, Inc.
package io.confluent.controlcenter.streams.group;

import com.google.common.collect.ImmutableList;
import com.google.inject.Inject;

import org.apache.kafka.common.utils.Bytes;
import org.apache.kafka.streams.kstream.KStream;
import org.apache.kafka.streams.kstream.KTable;
import org.apache.kafka.streams.kstream.Predicate;
import org.apache.kafka.streams.kstream.Window;
import org.apache.kafka.streams.kstream.Windowed;
import org.apache.kafka.streams.kstream.Windows;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import io.confluent.controlcenter.keys.Keys.KeyType;
import io.confluent.controlcenter.record.Controlcenter.WindowedGrouping;
import io.confluent.controlcenter.serialization.OrderedKeyPrefixedSerdeSupplier;
import io.confluent.controlcenter.streams.AbstractMonitoringMessageRollupStreamExtension;
import io.confluent.controlcenter.streams.TopicStoreMaster.Store;
import io.confluent.controlcenter.streams.TopicStoreMaster.Topic;
import io.confluent.controlcenter.streams.TopicStoreModule.GroupStore;
import io.confluent.controlcenter.streams.TopicStoreModule.GroupStreamExtensionTopic;
import io.confluent.controlcenter.streams.TypeMapper;
import io.confluent.monitoring.record.Monitoring.MonitoringMessage;

public class GroupStreamExtension extends
                                  AbstractMonitoringMessageRollupStreamExtension<WindowedGrouping> {

  private static final Logger log = LoggerFactory.getLogger(GroupStreamExtension.class);

  private final Store<Bytes, WindowedGrouping, WindowedGrouping> groupStore;
  private final Topic<Bytes, WindowedGrouping, Void, WindowedGrouping> groupStreamExtentionTopic;
  private final OrderedKeyPrefixedSerdeSupplier<KeyType, MonitoringMessage> orderedKeyPrefixedSerdeSupplier;

  @Inject
  public GroupStreamExtension(
      @GroupStore Store<Bytes, WindowedGrouping, WindowedGrouping> groupStore,
      @GroupStreamExtensionTopic Topic<Bytes, WindowedGrouping, Void, WindowedGrouping> groupStreamExtentionTopic,
      OrderedKeyPrefixedSerdeSupplier<KeyType, MonitoringMessage> orderedKeyPrefixedSerdeSupplier
  ) {
    this.groupStore = groupStore;
    this.groupStreamExtentionTopic = groupStreamExtentionTopic;
    this.orderedKeyPrefixedSerdeSupplier = orderedKeyPrefixedSerdeSupplier;
  }

  @Override
  protected String name() {
    return groupStore.name;
  }

  @Override
  protected Iterable<? extends TypeMapper<WindowedGrouping>> getMappers() {
    return ImmutableList.of(
        new WindowedGroupingTypeMapper(orderedKeyPrefixedSerdeSupplier.get(KeyType.INFO_CLIENTIDS_IN_GROUP)),
        new WindowedGroupingTypeMapper(orderedKeyPrefixedSerdeSupplier.get(KeyType.INFO_TOPICPARTITIONS_IN_CLIENT)),
        new WindowedGroupingTypeMapper(orderedKeyPrefixedSerdeSupplier.get(KeyType.INFO_GROUPS_IN_CLIENTTYPE)),
        new WindowedGroupingTypeMapper(orderedKeyPrefixedSerdeSupplier.get(KeyType.INFO_TOPICS_IN_CLIENTTYPE)),
        new WindowedGroupingTypeMapper(orderedKeyPrefixedSerdeSupplier.get(KeyType.INFO_TOPICPARTITIONS_IN_GROUP))
    );
  }

  /**
   * this will filter out any window that is outside of our retention interval then aggregate the
   * remaining messages to compute a set of members (producers, topics, consumers, etc.)
   *
   * @param stream  base stream to filter and aggregate on
   * @param windows named windows that we want to roll up
   */
  protected KTable<Windowed<Bytes>, WindowedGrouping> filterAndAggregateRollup(
      KStream<Bytes, WindowedGrouping> stream,
      final Windows<Window> windows,
      final String name
  ) {
    return stream
        // Keep only messages that are within the retention time
        .filter(new Predicate<Bytes, WindowedGrouping>() {
          @Override
          public boolean test(Bytes key, WindowedGrouping value) {
            return (value.getWindow() + windows.maintainMs()) > System.currentTimeMillis();
          }
        })
        // Compute the set of "members" for each "group by" key
        .groupByKey(
            groupStore.keySerde,
            groupStore.valueSerde
        ).aggregate(
          new WindowedGroupingInitializer(),
          new WindowedGroupingAggregator(),
          windows,
          groupStore.aggregateSerde,
          name
        );
  }

  @Override
  protected KStream<Bytes, WindowedGrouping> shuffle(KStream<Bytes, WindowedGrouping> stream) {
    //Duplicate production to this topic is ok because we're computing a set
    return stream.through(
        groupStreamExtentionTopic.keySerde,
        groupStreamExtentionTopic.valueSerde,
        groupStreamExtentionTopic.name
    );
  }
}
