// (Copyright) [2016 - 2016] Confluent, Inc.
package io.confluent.controlcenter.streams.verify;

import com.google.common.collect.BoundType;
import com.google.common.collect.Range;
import com.google.common.collect.RangeSet;
import com.google.common.collect.TreeRangeSet;

import io.confluent.monitoring.common.MonitoringMessageUtil;
import io.confluent.monitoring.common.TimeBucket;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Objects;
import java.util.Set;

import io.confluent.controlcenter.errors.AlreadyShutdownException;
import io.confluent.controlcenter.errors.DuplicateSequenceNumberException;
import io.confluent.controlcenter.errors.SequenceAfterShutdownException;
import io.confluent.controlcenter.errors.ShutdownSequenceNotHighestException;
import io.confluent.controlcenter.record.Controlcenter.SequenceInfo;
import io.confluent.controlcenter.record.Controlcenter.SequenceInfoPair;
import io.confluent.controlcenter.record.Controlcenter.VerifierInfo;
import io.confluent.monitoring.record.Monitoring.MessageType;
import io.confluent.monitoring.record.Monitoring.MonitoringMessage;

/**
 * This class manages sequence numbers for (Cluster, Client, Topic, Partition, Session)
 */
public class MonitoringVerifier {

  /**
   * Expected distance in number of time windows between two consecutive sequences that we use when
   * there is no history about window distances. E.g. we got first sequence, and then we don't get
   * any more sequences so we have to timeout and there is no sample points about min and max
   * window.
   */
  private static final long DEFAULT_NORMALIZED_WINDOW_DISTANCE = 10;
  private static final long MIN_NORMALIZED_WINDOW_DISTANCE = 1;
  private static final long MAX_CONSECUTIVE_REPORTED_MISSING_SEQUENCES = 20;

  private static final Logger log = LoggerFactory.getLogger(MonitoringVerifier.class);

  /**
   * Monitoring message containing group, client type, client, session, topic, partition, sample
   * period, for this verifier that we will use as a base to create an error monitoring message.
   */
  private final MonitoringMessage.Builder baseMonitoringMessageBuilder;

  /**
   * Number of ms we wait before a sequence number can be declared missing
   */
  private final long timeout;
  private final long windowHistorySize;

  /**
   * Ranges of sequences that are received or already declared missing
   */
  private RangeSet<SequenceWithMetadata> ranges = TreeRangeSet.create();
  private boolean shutdownReceived = false;
  private boolean sessionEnded = false;
  private boolean maybeSessionEnded = false;
  private long lastSequenceTime = -1;
  /**
   * Last received sequence which has any info about a window: either min and max window, and if min
   * and max windows are not set, then sequence's time window. Null, if no sequence received so far
   * has a window info (rare case),
   */
  private SequenceWithMetadata lastSequenceWithWindow = null;

  /**
   * Expected window distance in ms for 'windowHistorySize' consecutive sequences. Currently, we are
   * using maximum to be on a safer side. Set to -1 if unknown.
   */
  private long expectedWindowDistance = -1;

  /**
   * Constructs the object from the first monitoring message received for this client, session,
   * topic, partition.
   */
  public MonitoringVerifier(MonitoringMessage firstMonitoringMessage,
                            long timeout) {
    this(firstMonitoringMessage, timeout, TimeBucket.HISTORY_SIZE);
  }

  public MonitoringVerifier(
      MonitoringMessage firstMonitoringMessage,
      long timeout,
      int windowHistorySize
  ) {
    baseMonitoringMessageBuilder = MonitoringMessage.newBuilder(MonitoringMessageUtil.baseMonitoringMessage())
        .setClusterId(firstMonitoringMessage.getClusterId())
        .setWindow(firstMonitoringMessage.getWindow())
        .setSession(firstMonitoringMessage.getSession())
        .setGroup(firstMonitoringMessage.getGroup())
        .setClientId(firstMonitoringMessage.getClientId())
        .setClientType(firstMonitoringMessage.getClientType())
        .setTopic(firstMonitoringMessage.getTopic())
        .setPartition(firstMonitoringMessage.getPartition())
        .setMinWindow(firstMonitoringMessage.getMinWindow())
        .setMaxWindow(firstMonitoringMessage.getMaxWindow())
        .setSamplePeriod(firstMonitoringMessage.getSamplePeriod());
    // todo(apovzner) would be good to ensure that if interceptors do not use their custom sample
    // period, and set sample period to null, they are using TimeBucket.SIZE (or a default value we
    // would configure).
    baseMonitoringMessageBuilder.setSamplePeriod(
        firstMonitoringMessage.getSamplePeriod() < TimeBucket.SIZE ?
        TimeBucket.SIZE : firstMonitoringMessage.getSamplePeriod());
    this.timeout = timeout;
    this.windowHistorySize = windowHistorySize;
  }

  /**
   * Constructs the object from VerifierInfo
   */
  public MonitoringVerifier(VerifierInfo info) {
    this.baseMonitoringMessageBuilder = MonitoringMessage.newBuilder(info.getBaseMonitoringMessage());
    this.timeout = info.getTimeout();
    this.shutdownReceived = info.getShutdownReceived();
    this.sessionEnded = info.getSessionEnded();
    this.maybeSessionEnded = info.getMaybeSessionEnded();
    this.lastSequenceTime = info.getLastSequenceTime();
    this.windowHistorySize = info.getWindowHistorySize();
    this.expectedWindowDistance = info.getExpectedWindowDistance();
    if (info.getLastSequenceWithWindow() != null &&
        info.getLastSequenceWithWindow().getSequence() > -1) {
      this.lastSequenceWithWindow = new SequenceWithMetadata(info.getLastSequenceWithWindow());
    }

    // build ranges from list in the VerifierInfo
    for (SequenceInfoPair pair : info.getRangesList()) {
      // all ranges are closed-open
      this.ranges.add(Range.closedOpen(new SequenceWithMetadata(pair.getLower()),
                                       new SequenceWithMetadata(pair.getUpper())));
    }
  }

  public VerifierInfo createVerifierInfo() {
    List<SequenceInfoPair> rangeList = new ArrayList<>();
    for (Range<SequenceWithMetadata> range : this.ranges.asRanges()) {
      rangeList.add(
          SequenceInfoPair.newBuilder()
              .setLower(range.lowerEndpoint().toSequenceInfo())
              .setUpper(range.upperEndpoint().toSequenceInfo())
              .build()
      );
    }
    VerifierInfo.Builder builder = VerifierInfo.newBuilder()
        .setBaseMonitoringMessage(baseMonitoringMessageBuilder)
        .setTimeout(this.timeout)
        .setShutdownReceived(this.shutdownReceived)
        .setSessionEnded(this.sessionEnded)
        .setMaybeSessionEnded(this.maybeSessionEnded)
        .setLastSequenceTime(this.lastSequenceTime)
        .setWindowHistorySize(this.windowHistorySize)
        .setExpectedWindowDistance(this.expectedWindowDistance)
        .addAllRanges(rangeList);
    if (this.lastSequenceWithWindow != null) {
      builder.setLastSequenceWithWindow(this.lastSequenceWithWindow.toSequenceInfo());
    } else {
      builder.setLastSequenceWithWindow(SequenceInfo.newBuilder().setSequence(-1).build());
    }

    return builder.build();
  }

  public String getClientId() {
    return baseMonitoringMessageBuilder.getClientId();
  }

  public String getTopic() {
    return baseMonitoringMessageBuilder.getTopic();
  }

  public int getPartition() {
    return baseMonitoringMessageBuilder.getPartition();
  }

  public long getSamplePeriod() {
    return baseMonitoringMessageBuilder.getSamplePeriod();
  }

  public String getSession() {
    return baseMonitoringMessageBuilder.getSession();
  }

  public MonitoringMessage.Builder getCopyOfBaseMonitoringMessage() {
    return MonitoringMessage.newBuilder(baseMonitoringMessageBuilder.build());
  }

  public long lastActivityTimeMs() {
    return lastSequenceTime;
  }

  /**
   * Records the sequence number as received.
   *
   * @param monitoringMessage monitoring message
   * @param now current time or -1 if time is unknown
   * @throws DuplicateSequenceNumberException    If the sequence number of this message has been
   *                                             already recorded.
   * @throws AlreadyShutdownException            If the monitoring message has shutdown flag set,
   *                                             but shutdown was already received with another
   *                                             monitoring message.
   * @throws ShutdownSequenceNotHighestException If the monitoring message has shutdown flag set,
   *                                             but we already received higher sequence numbers.
   * @throws SequenceAfterShutdownException      If an monitoring message with shutdown flag was
   *                                             already received or if this message has a shutdown
   *                                             flag but we already received a message with a
   *                                             higher sequence number.
   * @throws IllegalStateException               If this method is called after all sequence numbers
   *                                             were verified (e.g., shutdown flag received and all
   *                                             sequences arrived before the shutdown flag)
   */
  public void addSequence(MonitoringMessage monitoringMessage, long now) {
    if (isDone()) {
      throw new IllegalStateException(
          "Received monitoring message after all expected sequences were "
          + "received. Sequence=" + monitoringMessage.getSequence());
    }
    maybeSessionEnded = false;

    SequenceWithMetadata seq = new SequenceWithMetadata(monitoringMessage, now);

    ensureSequenceNotDuplicate(seq);

    // Add this sequence to the range set, but make sure that ranges get merged so that we don't
    // keep info about sequence numbers that are not needed for verification
    SequenceWithMetadata nextSeq = seq.nextWithTimestamp(seq.timestamp());

    // if we received a message with shutdown flag, the message's sequence number is expected to
    // be the highest sequence number we would receive.
    if (monitoringMessage.getShutdown()) {
      if (shutdownReceived) {
        // Not a duplicate monitoring message but the second message with a shutdown flag -- bad
        throw new AlreadyShutdownException("One more non-duplicate sequence with shutdown. " + seq);
      }
      // if we received a sequence number with shutdown flag while we have a higher sequence
      // number -- something went wrong, so not recording this sequence
      if (!ranges.isEmpty() &&
          (ranges.span().contains(seq) ||
           seq.sequence() < ranges.span().lowerEndpoint().sequence())) {
        throw new ShutdownSequenceNotHighestException("Sequence with shutdown flag is not highest"
                                                      + ": " + seq);
      }
      shutdownReceived = true;
    } else if (shutdownReceived) {
      // also bad if we receive a message with a sequence number higher than any existing one, but
      // we already received a shutdown flag.
      if (ranges.span().intersection(Range.atLeast(seq)).isEmpty()) {
        throw new SequenceAfterShutdownException(
            "Sequence number after shutdown is higher than any other sequence number " + seq);
      }
    }

    // actually add the sequence to the range set
    ranges.add(Range.closedOpen(seq, nextSeq));
    // the caller may not know about the current time, in which case we will not change last
    // sequence time until we get a valid timestamp
    if (now >= 0) {
      lastSequenceTime = now;
    }
    updateWindowDistanceAndLastSequence(seq,
                                        monitoringMessage.getType() == MessageType.NORMAL
                                        ? monitoringMessage.getWindow() : -1);

    // a common case would be receiving monotonically increasing sequence numbers or with small
    // variation and then receiving a message with shutdown flag with no missing messages
    // so if we received all sequence numbers and a shutdown message, then set state as done
    verifySequencesBeforeShutdownAndMaybeSetDone();
  }

  /**
   * Returns rages of time windows that have missing monitoring data. If nothing is missing, returns
   * null.
   */
  public RangeSet<Long> getMissingMonitoringData(long now) {
    // if there are no sequences received or we are done (sequences cleared), then we don't have
    // any missing sequence numbers
    if (!ranges.isEmpty() && now >= 0) {
      // if we did not record valid 'lastSequenceTime', we should approximate it to 'now',
      // because otherwise we will never expire a sequence missing after the sequence that
      // arrived before the first C3 heartbeat
      if (lastSequenceTime < 0) {
        lastSequenceTime = now;
      }

      final SequenceWithMetadata seqAfterHighest =
          ranges.asDescendingSetOfRanges().iterator().next().upperEndpoint();
      final long delta = lastSequenceTime <= 0 ? 0 : now - lastSequenceTime;
      final long thresholdTimestamp = seqAfterHighest.timestamp() - timeout + delta;
      log.trace("thresholdTimestamp=" + thresholdTimestamp + ", local delta = " + delta);

      // even if enough time passed to timeout a sequence after last seen (if missing), we still
      // have to look and timeout any missing sequences in between, because we want to collect
      // all time windows which could potentially have missing monitoring data

      Iterator<Range<SequenceWithMetadata>> it = ranges.asRanges().iterator();
      RangeSet<SequenceWithMetadata> verifiedMissedRanges = null;
      RangeSet<Long> missingWindows = null;
      Range<SequenceWithMetadata> prevRange = null;
      Long missingSeq = null;

      while (it.hasNext()) {
        Range<SequenceWithMetadata> nextRange = it.next();
        SequenceWithMetadata nextLowerEndpoint = nextRange.lowerEndpoint();

        // find first missing sequence number
        if (prevRange == null && nextLowerEndpoint.sequence() > 0L) {
          missingSeq = 0L;
        } else if (prevRange != null) {
          if (prevRange.upperBoundType() == BoundType.CLOSED) {
            throw new IllegalStateException("Upper bound of any range must always be open");
          }
          missingSeq = prevRange.upperEndpoint().sequence();
        }

        // check if we should timeout any sequence numbers starting with 'missingSeq'
        // We timeout all sequences in the missing range at the same time
        Long endMissingSeq = nextLowerEndpoint.sequence() - 1L;
        if (missingSeq != null && missingSeq <= endMissingSeq) {
          long estimatedTimestamp =
              getTimedOutSequenceTimestamp(now,
                                           prevRange == null ? null : prevRange.upperEndpoint(),
                                           nextLowerEndpoint.timestamp());

          if (estimatedTimestamp >= 0) {
            log.warn("Missing sequence numbers=[{}, {}] from client={}, topic={}, partition={}, estimated timestamp={}, next timestamp={}",
                     missingSeq, endMissingSeq, getClientId(), getTopic(), getPartition(),
                     estimatedTimestamp, nextLowerEndpoint.timestamp());

            // collect time windows that may have missing monitoring data
            long firstReportedMissingSeq = missingSeq;
            if (endMissingSeq - firstReportedMissingSeq >= MAX_CONSECUTIVE_REPORTED_MISSING_SEQUENCES) {
              log.warn("Missing {} consecutive sequences; will report {} highest sequences for client={},  topic={}, partition=={}",
                       endMissingSeq - firstReportedMissingSeq + 1,
                       MAX_CONSECUTIVE_REPORTED_MISSING_SEQUENCES,
                       getClientId(), getTopic(), getPartition());
              firstReportedMissingSeq = endMissingSeq - MAX_CONSECUTIVE_REPORTED_MISSING_SEQUENCES + 1;
            }
            for (long seq = firstReportedMissingSeq; seq <= endMissingSeq; seq++) {
              missingWindows = addMissingWindowRange(seq, nextLowerEndpoint, missingWindows,
                                                     seq == firstReportedMissingSeq);
            }

            // we verified the sequence number, add it to the ranges
            if (verifiedMissedRanges == null) {
              verifiedMissedRanges = TreeRangeSet.create();
            }
            verifiedMissedRanges.add(Range.closedOpen(
                                         new SequenceWithMetadata(missingSeq, estimatedTimestamp),
                                         nextLowerEndpoint)
            );
          }
        } else if (missingSeq != null) {
          throw new IllegalStateException("Ranges must be in increasing sequence order");
        }

        if (missingSeq != null && missingWindows == null) {
          // nothing timed out in this range, so will not timeout in the next
          break;
        }

        prevRange = nextRange;
      }

      if (verifiedMissedRanges != null && !verifiedMissedRanges.isEmpty()) {
        ranges.addAll(verifiedMissedRanges);
      }

      // now see if we should timeout a sequence after last seen, which means there was a long delay
      // since the last sequence number arrived
      missingWindows = maybeTimeoutSequenceAfterHighest(
          seqAfterHighest,
          thresholdTimestamp,
          missingWindows
      );

      // if we timed out at least one missing sequence number, check if we are done
      if (missingWindows != null) {
        // first see if we should also timeout a sequence after the last seen sequence
        verifySequencesBeforeShutdownAndMaybeSetDone();
      }


      return missingWindows;
    }

    return null;
  }

  /**
   * Notification that session has changed. This verifier will stop timing out sequence numbers >
   * one after the highest received sequence number.
   */
  public void notifySessionEnded() {
    sessionEnded = true;
  }

  /**
   * Return true if all sequence numbers were verified and no new sequence numbers are excepted.
   */
  public boolean isDone() {
    return shutdownReceived && ranges.isEmpty();
  }

  /**
   * Return true is session if session is not expected to receive new sequences, but it might.
   */
  public boolean isSessionNotActive() {
    return isDone() || sessionEnded || maybeSessionEnded;
  }

  /**
   * Used for tests only. Returns true if notifySessionEnded() was called.
   */
  protected boolean isSessionEnded() {
    return sessionEnded;
  }

  /**
   * Checks sequence for duplicate and throws DuplicateSequenceNumberException if sequence is
   * already received.
   */
  private void ensureSequenceNotDuplicate(SequenceWithMetadata seq) {
    if (ranges.contains(seq)) {
      throw new DuplicateSequenceNumberException("Duplicate sequence number: " + seq.sequence());
    }
  }

  /**
   * Check if we should timeout sequence number after the highest seen sequence number. This is the
   * case where we did not receive message with shutdown flag and had long delay since we got a
   * sequence number.
   *
   * @param sequenceAfterHighest The upper endpoint of the last range which is open endpoint, the
   *                             first missing sequence after the highest sequence.
   * @param thresholdTimestamp   Timestamp before which all missing sequence numbers timeout
   * @return true if a sequence timed out
   */
  private RangeSet<Long> maybeTimeoutSequenceAfterHighest(SequenceWithMetadata sequenceAfterHighest,
                                                          long thresholdTimestamp,
                                                          RangeSet<Long> missingWindows) {
    if (!shutdownReceived) {
      // if we did not receive session end notification, then use a bit longer timeout, because
      // we do not know the window range in this case, and have to show all windows potentially
      // incomplete
      // TODO: do we want a longer timeout and what should be the timeout value?
      long estimatedTimestamp = sequenceAfterHighest.timestamp() + getSamplePeriod();
      long missingSeq = sequenceAfterHighest.sequence();

      if (sessionEnded && estimatedTimestamp <= thresholdTimestamp) {
        log.info("Session ended and timed out at least one sequence number ({}) after "
                 + "the highest received sequence number. Done", missingSeq);
        ranges.clear();
        shutdownReceived = true;
        return addMissingWindowRange(missingSeq, missingWindows, true);
      } else if (!maybeSessionEnded && !sessionEnded
                 && (estimatedTimestamp + 5 * getSamplePeriod()) <= thresholdTimestamp) {
        // since we waited for long time to expire sequence after the last seen, we can assume
        // that sequences not going to arrive, and client dropped out. Plus, we are using C3
        // heartbeats for time, so even if monitoring topic partition is slow, if client is still
        // alive, we should have received at least a client heartbeat message by now
        log.info("Timed out sequence={} after the highest received sequence for"
                 + " topic={}, partition={}, clientID={}. Session maybe ended.",
                 missingSeq, getTopic(), getPartition(), getClientId());
        // since we timed out the sequence after highest, all previous sequence numbers also
        // timeout; so add the whole range starting from 0
        ranges.add(Range.closedOpen(new SequenceWithMetadata(0L),
                                    sequenceAfterHighest.nextWithTimestamp(estimatedTimestamp)));
        maybeSessionEnded = true;
        return addMissingWindowRange(missingSeq, missingWindows, true);
      }
    }
    return missingWindows;
  }

  /**
   * If we received shutdown and have a contiguous set of sequence numbers (without holes), then we
   * are done. We set state to 'done' by clearing ranges, so that we can easier verify 'done' state
   * later.
   */
  private void verifySequencesBeforeShutdownAndMaybeSetDone() {
    if (shutdownReceived) {
      Set<Range<SequenceWithMetadata>> rangeSet = ranges.asRanges();
      // we should have one contiguous range in the set if we received all sequence numbers
      // without any ones missing
      if (rangeSet.size() == 1 &&
          rangeSet.iterator().next().lowerEndpoint().sequence() == 0L) {
        log.debug("MonitoringVerifier got all sequence numbers and a shutdown flag!");
        ranges.clear();
      }
    }
  }

  /**
   * Check if sequence number timed out.
   *
   * We don't have to be very precise with timing out, and we better timeout later than earlier. We
   * estimate the timestamp of a missing sequence number as max time of seen sequences around the
   * missing sequence.
   *
   * @param currentWallclockTime            Min of all partition timestamps
   * @param prevUpperEndpoint               The upper endpoint of the range preceding the sequence
   *                                        number. This is an open end-point that actually contains
   *                                        the first missing sequence number in this range. The
   *                                        timestamp assigned to the endpoint is the timestamp of
   *                                        the previous (non-missing) sequence.
   * @param nextNonMissingSequenceTimestamp Timestamp of the next closest non-missing sequence.
   * @return Estimated timestamp of the sequence number if it timed out, otherwise null.
   */
  private long getTimedOutSequenceTimestamp(long currentWallclockTime,
                                            SequenceWithMetadata prevUpperEndpoint,
                                            long nextNonMissingSequenceTimestamp) {
    long thresholdTimestamp = currentWallclockTime - timeout;
    long estimatedTimestamp = nextNonMissingSequenceTimestamp;
    if (prevUpperEndpoint != null && prevUpperEndpoint.timestamp() > 0) {
      estimatedTimestamp = Math.max(estimatedTimestamp, prevUpperEndpoint.timestamp());
    }
    if (estimatedTimestamp <= thresholdTimestamp) {
      // timeout!
      return estimatedTimestamp;
    }

    return -1L;
  }

  /**
   * Finds a range of missing time windows for a given sequence number, when there is a non-missing
   * sequence after 'missingSequence'. Adds missing time window range to 'missingWindows'.
   */
  private RangeSet<Long> addMissingWindowRange(long missingSequence,
                                               SequenceWithMetadata nextLowerEndpoint,
                                               RangeSet<Long> missingWindows,
                                               boolean firstInMissingRange) {
    long delta = nextLowerEndpoint.sequence() - missingSequence;
    if (delta < 1) {
      throw new IllegalArgumentException("Invalid missing sequence");
    }
    if (nextLowerEndpoint.minWindow() < 0 && delta <= windowHistorySize) {
      // min & max window == null, and we are within history range, this means that all 'history
      // size' windows before 'nextLowerEndpoint' are heartbeat windows, so no monitoring data is
      // lost
      return missingWindows;
    }

    // if we are here, we definitely lost monitoring data
    RangeSet<Long> retWindows = missingWindows;
    if (retWindows == null) {
      retWindows = TreeRangeSet.create();
    }

    if (nextLowerEndpoint.minWindow() >= 0) {
      // we have info about min and max window
      long minWindow = nextLowerEndpoint.minWindow();
      long maxWindow = nextLowerEndpoint.maxWindow();

      if (delta > this.windowHistorySize) {
        // we have to estimate the range, so log warning
        log.info("Missed too many consecutive sequence numbers for client=" + getClientId() +
                 ", topic=" + getTopic() + ", partition=" + getPartition() + ". Will estimate " +
                 "time window range with missed monitoring data.");
        // we need to get estimate for the missed sequences outside of window history size covered
        // by min and max window
        long expectedDistance = getExpectedDistanceMs(delta - windowHistorySize);
        minWindow = Math.max(0, minWindow - expectedDistance);
        maxWindow += expectedDistance;
      }
      retWindows.add(Range.closed(minWindow, maxWindow));
      log.info("Missing sequence=" + missingSequence + " window range=[" + minWindow +
               "..." + maxWindow + "] for clientId=" + getClientId() + ", topic=" + getTopic() +
               ", partition=" + getPartition());
      return retWindows;
    }

    // There is no window info in the next non-missing sequence.
    // Example: 0 1 2 3 4 5 6 7 8 9 10 11 12 13 14 15
    // Suppose we lost sequence 6 and we have sequence 10, and 7,8,9 = heartbeat
    // This means that sequence 10 is the next non-missing sequence, but min and max window == null
    // TODO(apovzner) In a more general case, we should find the lower endpoint in the next range,
    // but for now we are just getting window range relative to the last seen sequence -- which
    // works in a common case when we have one or more contiguous missing sequence vs. many
    // missing sequences in between of non-missing sequences
    return addMissingWindowRange(missingSequence, missingWindows, firstInMissingRange);
  }

  /**
   * Finds a range of missing time windows for a given sequence number when there is no higher
   * non-missing sequence. Or, finds a range of missing time windows for a lower sequence since last
   * seen when we need to get missing window based on info in the last seen sequence. Adds missing
   * time window range to 'missingWindows'.
   */
  private RangeSet<Long> addMissingWindowRange(long missingSequence,
                                               RangeSet<Long> missingWindows,
                                               boolean firstInMissingRange) {
    RangeSet<Long> retWindows = missingWindows;
    if (retWindows == null) {
      retWindows = TreeRangeSet.create();
    }
    if (
        lastSequenceWithWindow == null ||
        lastSequenceWithWindow.minWindow() < 0 ||
        lastSequenceWithWindow.maxWindow() < 0 ||
        lastSequenceWithWindow.sequence() < 0
        ) {
      // No sequence arrived so far with a known window (either min/max or time window).
      // Several failures must happen at the same time to get to this state
      // e.g., first monitoring message of the session which always has to have at least window
      // set (see updateWindowDistanceAndLastSequence() will set min and max window to window if
      // they are null) is lost, and then 'windowHistorySize' number of heartbeat messages also
      // lost, and then there are possibly heartbeat messages that gets delivered (but they don't
      // have window set), and then client crashes without sending a shutdown message
      if (firstInMissingRange) {
        log.warn("No information about windows received for this session for client=" +
                 getClientId() + "topic=" + getTopic() + ", partition=" + getPartition());
      }
      // TODO(apovzner) need info from previous session ? or maybe next?
      // for now set window ranges to 0 (will address this next time)
      retWindows.add(Range.closed(0L, 0L));
      return retWindows;
    }
    long delta = Math.abs(missingSequence - lastSequenceWithWindow.sequence());
    if (delta < 1) {
      throw new IllegalArgumentException(
          "Missing sequence expected to be different from last seen"
      );
    }

    // since min/max window info is for the windows prior the last seen sequence, we are actually
    // interested in distance for delta + 1
    // also, since we cannot assume which sequences have min and max windows, we just in case go
    // back from minWindow and forward from maxWindow the same distance
    long expectedDistance = getExpectedDistanceMs(delta + 1);
    long minWindow = Math.max(0, lastSequenceWithWindow.minWindow() - expectedDistance);
    long maxWindow = lastSequenceWithWindow.maxWindow() + expectedDistance;
    retWindows.add(Range.closed(minWindow, maxWindow));
    log.warn("No window info in the next non-missing sequence. Estimated window range for missing "
             + "sequence=" + missingSequence + " [" + minWindow +
             "..." + maxWindow + "], lastSequenceWithWindow=" + lastSequenceWithWindow +
             " for clientId=" + getClientId() +
             ", topic=" + getTopic() + ", partition=" + getPartition());
    return retWindows;
  }

  /**
   * For cases when we don't know window ranges for missing sequences, we need to make a guess. We
   * collect expected window distance that we see within 'windowHistorySize' number of messages.
   * Distance = maxWindow - minWindow. Currently, expected distance == max distance.
   *
   * For cases when we need to estimate time window ranges for missing sequence where there is no
   * sequences after a missing one, we also keep last non-missing sequence info that has either
   * valid min and max window or a valid window. So that we have some starting point to estimate a
   * range of windows with missing monitoring data.
   *
   * @param seq       sequence info from monitoring message
   * @param seqWindow window of the monitoring message
   */
  private void updateWindowDistanceAndLastSequence(SequenceWithMetadata seq, long seqWindow) {
    if (seq.minWindow() >= 0 && seq.maxWindow() >= 0) {
      lastSequenceWithWindow = seq;
      // update estimated window distance
      long delta = seq.maxWindow() - seq.minWindow();
      if (delta < 0) {
        log.error("maxWindow < minWindow. Not able to record expected window range.");
      } else {
        // lets be pessimistic for now -- recording maximum window range
        expectedWindowDistance = Math.max(expectedWindowDistance, delta);
      }
    } else if (seqWindow >= 0) {
      // min and max window unknown (e.g. first sequence after long idle time)
      // if window is known, then set min and max to window, otherwise, do not update last sequence
      lastSequenceWithWindow = seq.withMinMaxWindow(seqWindow);
    }
  }

  /**
   * Expected distance in ms that 'sequenceDelta' number of sequences are expected to cover. The
   * distance is a multiple of 'sampling period' so that all window ranges align to time windows (to
   * simplify aggregation).
   */
  private long getExpectedDistanceMs(long sequenceDelta) {
    long normalizedDistanceWin =
        Math.max(expectedWindowDistance < 0 || windowHistorySize <= 1 ?
                 DEFAULT_NORMALIZED_WINDOW_DISTANCE :
                 expectedWindowDistance / (windowHistorySize - 1) / getSamplePeriod(),
                 MIN_NORMALIZED_WINDOW_DISTANCE);
    long normalizedDistance = normalizedDistanceWin * getSamplePeriod();
    log.trace("Calculated normalized distance =" + normalizedDistance +
              ", history size=" + windowHistorySize +
              ", expectedWindowDistance=" + expectedWindowDistance +
              ", default normalized distance = " + DEFAULT_NORMALIZED_WINDOW_DISTANCE);
    return normalizedDistance * sequenceDelta;
  }

  @Override
  public boolean equals(Object o) {
    if (this == o) {
      return true;
    }
    if (o == null || getClass() != o.getClass()) {
      return false;
    }
    MonitoringVerifier other = (MonitoringVerifier) o;
    return Objects.equals(baseMonitoringMessageBuilder.buildPartial(), other.baseMonitoringMessageBuilder.buildPartial()) &&
           Objects.equals(lastSequenceTime, other.lastSequenceTime) &&
           Objects.equals(ranges, other.ranges) &&
           Objects.equals(shutdownReceived, other.shutdownReceived) &&
           Objects.equals(sessionEnded, other.sessionEnded) &&
           Objects.equals(maybeSessionEnded, other.maybeSessionEnded) &&
           Objects.equals(timeout, other.timeout) &&
           Objects.equals(windowHistorySize, other.windowHistorySize) &&
           Objects.equals(expectedWindowDistance, other.expectedWindowDistance) &&
           Objects.equals(lastSequenceWithWindow, other.lastSequenceWithWindow);
  }

  @Override
  public int hashCode() {
    return Objects.hash(baseMonitoringMessageBuilder.buildPartial(), lastSequenceTime, ranges,
                        shutdownReceived, sessionEnded, maybeSessionEnded, timeout,
                        windowHistorySize, expectedWindowDistance, lastSequenceWithWindow);
  }

  @Override
  public String toString() {
    return "ClientID=" + getClientId() + ", topic=" + getTopic() + ", partition=" + getPartition() +
           ", session=" + getSession() +
           ", timeout=" + timeout + ", windowHistorySize=" + windowHistorySize +
           ", shutdownReceived=" + shutdownReceived +
           ", maybeSessionEnded=" + maybeSessionEnded +
           ", sessionEnded=" + sessionEnded + ", lastSequenceTime=" + lastSequenceTime +
           ", expectedWindowDistance=" + expectedWindowDistance +
           ", lastSequenceWithWindow=" + lastSequenceWithWindow +
           ", Ranges=" + ranges;
  }
}
