// (Copyright) [2016 - 2016] Confluent, Inc.
package io.confluent.controlcenter.streams.verify;

import com.google.common.base.Joiner;
import com.google.common.collect.Iterables;
import com.google.common.collect.Range;
import com.google.common.collect.RangeSet;
import io.confluent.controlcenter.errors.AlreadyShutdownException;
import io.confluent.controlcenter.errors.DuplicateSequenceNumberException;
import io.confluent.controlcenter.errors.SequenceAfterShutdownException;
import io.confluent.controlcenter.errors.ShutdownSequenceNotHighestException;
import io.confluent.controlcenter.util.KVQuery;
import io.confluent.monitoring.common.Clock;
import io.confluent.monitoring.common.TimeBucket;
import io.confluent.monitoring.record.Monitoring;
import io.confluent.monitoring.record.Monitoring.MonitoringMessage;
import io.confluent.serializers.OrderedBytes;
import io.confluent.serializers.OrderedKeyUberSerde;
import org.apache.kafka.common.utils.Bytes;
import org.apache.kafka.streams.KeyValue;
import org.apache.kafka.streams.kstream.Transformer;
import org.apache.kafka.streams.kstream.TransformerSupplier;
import org.apache.kafka.streams.processor.ProcessorContext;
import org.apache.kafka.streams.state.KeyValueIterator;
import org.apache.kafka.streams.state.KeyValueStore;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;
import java.util.concurrent.TimeUnit;

/**
 * This is the transformer for verifying that we receive all monitoring data correctly. It filters
 * duplicate monitoring messages and writes missing monitoring data into rocksdb.
 *
 * Verification is per clientId, topic, partition, session
 *
 * We are using lower level Kafka Streams API because we want a 'unlimited' window that lives only
 * during a client session.
 *
 * Value: MonitoringMessage Key (output stream) incorporates: clientId, topic, partition, session ID
 * Aggregated Value: MonitoringVerifier
 *
 * Each time monitoring message is processed, we also verify if there are any timed out missing
 * sequence numbers. If so, we write this info into RocksDB (not yet implemented). Once we received
 * monitoring message with a new session ID for clientId, topic, partition, the key-value(s) for the
 * old session ID are removed from the key-value store (either right away or after last sequence
 * times out).
 *
 * Verification is done during process call, which means we do verification each time we process an
 * monitoring message. We may improve this later by doing verification less frequently.
 *
 * This approach works if we continuously receive monitoring messages from client, topic, partition
 * (either same session ID or new session IDs). To address the case when clients stops sending
 * monitoring messages (e.g., shuts down without restart), we also do verification on punctuate.
 * However, we set a long interval for that (good config TBD), since punctuation needs to iterate
 * over the whole store, which will be inefficient.
 */
public class VerifyTransformerSupplier implements
    TransformerSupplier<byte[], MonitoringMessage, KeyValue<byte[], Iterable<KeyValue<byte[], MonitoringMessage>>>> {

  private static final Logger log = LoggerFactory.getLogger(VerifyTransformerSupplier.class);

  private static final long DEFAULT_RETAIN_MS = TimeUnit.HOURS.toMillis(4);
  private static final long PUNCTUATE_INTERVAL_MS = 12 * TimeBucket.SIZE;
  private static final long DEFAULT_MAX_PUNCTUATE_MS = TimeUnit.SECONDS.toMillis(1);

  private final long sequenceTimeout;
  private final String storeName;
  /**
   * This is how long key-values are kept in the state store since verifier is declared 'done'
   * (session ended). Normally, we don't need to keep verifiers after they are done. One reason for
   * keeping them is if the client shuts down and never restarts, so we don't get session end. In
   * that case, we don't need to wait much longer thant a timeout of the sequence after the last
   * one. Another reason is if there is no way to get any window info from the next session and we
   * need to infer from the previous session -- in that case we may want to wait longer.
   */
  private final long storeRetainMs;
  private final long maxPunctuateMs;
  private final long punctuateIntervalMs;
  private final int numSourcePartitions;
  private final Clock clock;
  private final OrderedKeyUberSerde<MonitoringMessage> keySerde;

  public VerifyTransformerSupplier(
      long sequenceTimeout,
      String storeName,
      long storeRetainMs,
      long maxPunctuateMs,
      long punctuateIntervalMs,
      int numSourcePartitions,
      Clock clock,
      OrderedKeyUberSerde<MonitoringMessage> keySerde
  ) {
    this.sequenceTimeout = sequenceTimeout;
    this.storeName = storeName;
    this.storeRetainMs = storeRetainMs;
    this.maxPunctuateMs = maxPunctuateMs;
    this.punctuateIntervalMs = punctuateIntervalMs;
    this.numSourcePartitions = numSourcePartitions;
    this.clock = clock;
    this.keySerde = keySerde;
  }

  public VerifyTransformerSupplier(
      String storeName,
      int numSourcePartitions,
      Clock clock,
      OrderedKeyUberSerde<MonitoringMessage> keySerde
  ) {
    this(TimeBucket.SEQUENCE_TIMEOUT, storeName,
         DEFAULT_RETAIN_MS, DEFAULT_MAX_PUNCTUATE_MS, PUNCTUATE_INTERVAL_MS, numSourcePartitions,
         clock, keySerde);
  }

  @Override
  public Transformer<byte[], MonitoringMessage, KeyValue<byte[], Iterable<KeyValue<byte[], MonitoringMessage>>>> get() {
    return new MonitoringVerifyTransformer(
        punctuateIntervalMs,
        maxPunctuateMs,
        storeName,
        storeRetainMs,
        numSourcePartitions
    );
  }

  private class MonitoringVerifyTransformer implements
      Transformer<byte[], MonitoringMessage, KeyValue<byte[], Iterable<KeyValue<byte[], MonitoringMessage>>>> {

    private final long storeRetainMs;
    private final long maxPunctuateMs;
    private final long punctuateIntervalMs;
    private final String storeName;

    private long lastPunctuateTimeMs = -1;
    private Bytes lastPunctuateKey = null;
    private ProcessorContext context;
    private KeyValueStore<Bytes, MonitoringVerifier> kvStore;

    private int slowestMonitoringPartition;
    private ArrayList<Long> partitionTimestamps;

    public MonitoringVerifyTransformer(long punctuateIntervalMs, long maxPunctuateMs, String
        storeName, long storeRetainMs, int numSourcePartitions) {
      this.storeName = storeName;
      this.storeRetainMs = storeRetainMs;
      this.maxPunctuateMs = maxPunctuateMs;
      this.punctuateIntervalMs = punctuateIntervalMs;
      this.partitionTimestamps = new ArrayList<>(numSourcePartitions);
      for (int i = 0; i < numSourcePartitions; ++i) {
        this.partitionTimestamps.add(i, -1L);
      }
      this.slowestMonitoringPartition = 0;
    }

    @SuppressWarnings("unchecked")
    @Override
    public void init(ProcessorContext context) {
      this.context = context;
      // We do not schedule punctuate to KStreams, which is based on KStream timestamp, which is
      // extracted from monitoring message 'window'. Instead, we implement C3 heartbeat time
      // based punctuate -- punctuate is called only when C3 heartbeat messages are received and
      // missing data is verified based on timestamp in C3 heartbeat message. Since one or more
      // partitions of monitoring topic can become slower than others, or a partition may
      // disappear for some time, we move the clock based on the slowest partition. We still
      // don't want to punctuate too often, so punctuation only happens if enough wall clock time
      // is passed. C3 heartbeat timestamp based punctuate makes sure that if C3 is behind
      // processing monitoring data, we do not start expiring all monitoring messages. If C3 is
      // not behind, C3 heartbeat message timestamp is close to wall clock time.
      kvStore = (KeyValueStore<Bytes, MonitoringVerifier>) this.context.getStateStore(storeName);
    }

    @Override
    public KeyValue<byte[], Iterable<KeyValue<byte[], MonitoringMessage>>> transform(
        byte[] key,
        MonitoringMessage monitoringMessage
    ) {
      if (monitoringMessage == null) {
        return null;
      }

      log.trace("transform store={} message={}", storeName, monitoringMessage);

      long arrivalTime = partitionTimestamps.get(monitoringMessage.getMonitoringTopicPartition());
      List<KeyValue<byte[], MonitoringMessage>> ret = new ArrayList<>();

      if (monitoringMessage.getClientType() == Monitoring.ClientType.CONTROLCENTER) {
        // those are heartbeats to keep streams active
        maybePunctuate(monitoringMessage, ret);
        if (ret.size() == 0) {
          return null;
        } else {
          return KeyValue.pair(null, (Iterable<KeyValue<byte[], MonitoringMessage>>) ret);
        }
      }

      String logKey = logKey(monitoringMessage);
      Bytes storeKey = keySerde.key(monitoringMessage);

      try {
        MonitoringVerifier verifier = kvStore.get(storeKey);

        if (verifier == null) {
          verifier = new MonitoringVerifier(monitoringMessage, sequenceTimeout);
          // check if there are any verifiers for the same client, topic, partition (but different
          // session) and if so, notify each such verifier about session change
          notifySessionChange(storeKey, monitoringMessage, arrivalTime, ret);
        }

        log.trace("store={} {} monitoringVerifier={}", storeName, logKey, verifier);

        if (monitoringMessage.getShutdown()) {
          log.info("closing session={} sequence={} store={}", monitoringMessage.getSession(), monitoringMessage.getSequence(), storeName);
        }

        // notify verifier about this monitoring message
        verifier.addSequence(monitoringMessage, arrivalTime);

        // also check if any sequence numbers timed out
        // we do it here, because doing it in punctuate could be less efficient, because punctuate
        // has to iterate over the whole key-value store
        // on the other hand, we could check less often; we can optimize this later
        long clockTime = partitionTimestamps.get(slowestMonitoringPartition);
        if (clockTime > 0) {
          // we should not expire any sequences until we know the min time
          verifyMonitoringData(verifier, clockTime, ret);
        }

        // update the store with new value
        // even if we received all sequence numbers for the client, topic, partition -- do not
        // remove right away, because the duplicate will create the key-value again; will remove
        // during punctuation
        kvStore.put(storeKey, verifier);
      } catch (DuplicateSequenceNumberException dse) {
        // ignore this message, but log the warning
        log.info("ignoring duplicate sequence={} {} store={}", monitoringMessage.getSequence(), logKey, storeName);
        return null;
      } catch (IllegalStateException ise) {
        // most of times this is ok, e.g., if we timeout all sequence numbers and
        // declared that verifier is done, but a delayed message arrived.
        log.info("unexpected arrival of sequence={} {} store={} ise={}", monitoringMessage.getSequence(), logKey, storeName, ise.getMessage());
        return null;
      } catch (AlreadyShutdownException iae) {
        // this is worse, because monitoring message has some unexpected fields, so logging as error
        log.error("shutdown already received sequence={} {} store={}", monitoringMessage.getSequence(), logKey, storeName, iae.getMessage());
        return null;
      } catch (SequenceAfterShutdownException se) {
        log.error("sequence number higher than shutdown message sequence={} {} store={}", monitoringMessage.getSequence(), logKey, storeName, se.getMessage());
        return null;
      } catch (ShutdownSequenceNotHighestException sse) {
        log.error("sequence number of shutdown message not highest sequence={} {} store={}", monitoringMessage.getSequence(), logKey, storeName, sse.getMessage());
        return null;
      } catch (Exception e) {
        // kvstore exception, which means that the sequence number of this message will be
        // declared missing. This will be false positive, which happens anyway for some of time
        // windows. We will log warning, but let the message go through for aggregation
        log.warn("Failed to update verifier with this message's sequence number. This sequence "
                 + "number maybe declared missing later even thought it is not. Sequence={}.",
                 monitoringMessage.getSequence(), e);
      }

      // we do not need heartbeat messages downstream, plus their 0 window may break log
      // compaction/other stuff for internal topics, so better not to send them downstream
      if (monitoringMessage.getType() == Monitoring.MessageType.NORMAL) {
        ret.add(KeyValue.pair(key, MonitoringMessage.newBuilder(monitoringMessage)
            .setArrivalTime(arrivalTime >= 0 ? arrivalTime : 0).build()));
      }

      if (ret.size() == 0) {
        return null;
      }

      return KeyValue.pair(null, (Iterable<KeyValue<byte[], MonitoringMessage>>) ret);
    }

    @Override
    public KeyValue<byte[], Iterable<KeyValue<byte[], MonitoringMessage>>> punctuate(long timestamp) {
      // check all verifiers if any sequence numbers timed out
      if (lastPunctuateTimeMs < 0) {
        // we will do the first punctuate after 'punctuate interval'
        lastPunctuateTimeMs = timestamp;
      }
      if (timestamp - lastPunctuateTimeMs < punctuateIntervalMs) {
        // we don't want to punctuate more often than 'punctuate interval'
        return null;
      }
      long now = clock.currentTimeMillis();
      lastPunctuateTimeMs = timestamp;
      long punctuateTimeout = now + maxPunctuateMs;

      List<KeyValue<byte[], MonitoringMessage>> missingWindowMsgList = new ArrayList<>();
      try (KeyValueIterator<Bytes, MonitoringVerifier> iter = KVQuery.rangeFrom(kvStore, lastPunctuateKey)) {
        long count = 0;
        while (iter.hasNext()) {
          KeyValue<Bytes, MonitoringVerifier> entry = iter.next();
          if (entry == null) {
            continue;
          }
          if (entry.value == null) {
            log.debug("removing null value key={} from store={}", entry.key, storeName);
            kvStore.delete(entry.key);
          } else {
            log.trace("checking key={} verifier={} store={}", entry.key, entry.value, storeName);
            verifyMonitoringData(entry.value, timestamp, missingWindowMsgList);
            if (entry.value.isSessionNotActive() && (timestamp - entry.value.lastActivityTimeMs() > storeRetainMs)) {
              log.debug("removing verifier key={} lastActivity={} from store={}", entry.key, entry.value.lastActivityTimeMs(), storeName);
              kvStore.delete(entry.key);
            } else {
              // asking verifier to check missing sequence numbers may change verifier's state, so
              // update the kvstore with updated verifier
              kvStore.put(entry.key, entry.value);
            }
          }
          lastPunctuateKey = entry.key;
          count++;
          if (clock.currentTimeMillis() > punctuateTimeout) {
            log.debug("stopping punctuate at key={} store={}", lastPunctuateKey, storeName);
            break;
          }
          if (!iter.hasNext()) {
            log.debug("finished all punctuation for store={}", storeName);
            lastPunctuateKey = null;
          }
          log.debug("punctuated on count={} records", count);
        }
      } catch (Exception e) {
        log.error("failure during punctuate for store={} e={}", storeName, e.getMessage());
      }

      if (missingWindowMsgList.isEmpty()) {
        return null;
      }

      return KeyValue.pair(
          null,
          (Iterable<KeyValue<byte[], MonitoringMessage>>) missingWindowMsgList
      );
    }

    @Override
    public void close() {
      log.info("closing store={}", storeName);
    }

    /**
     * Updates the clock of the corresponding partition and possibly moves clock we use to verify
     * monitoring data (which is the clock of the slowest partition).
     *
     * @param monitoringTopicPartition partition of the monitoring topic where C3 heartbeat message
     *                                 was published
     * @param monitoringMessageTimestamp timestamp of the heartbeat message
     * @return true if slowest partition clock moved
     */
    private boolean updateMonitoringTopicClock(int monitoringTopicPartition,
                                               long monitoringMessageTimestamp) {
      if (partitionTimestamps.get(monitoringTopicPartition) >= monitoringMessageTimestamp) {
        // we want the clock to increase monotonically
        return false;
      }

      long previousMin = partitionTimestamps.get(slowestMonitoringPartition);
      partitionTimestamps.set(monitoringTopicPartition, monitoringMessageTimestamp);
      // we only care if we update slowest partition: if we update another partition, its clock
      // can only grow, so we know that it will not become slowest
      if (monitoringTopicPartition == slowestMonitoringPartition) {
        long min = Collections.min(partitionTimestamps);
        if (min != partitionTimestamps.get(slowestMonitoringPartition)) {
          // another partition is slowest, update slowest partition
          slowestMonitoringPartition = partitionTimestamps.indexOf(min);
        }
        if (previousMin != min) {
          // min changed, move the clock
          log.trace("Moved slowest clock: partition={} timestamp={}",
                    slowestMonitoringPartition, min);
          return true;
        }
      }
      return false;
    }

    /**
     * Calls punctuate() if enough time passed since the last punctuate
     */
    private void maybePunctuate(MonitoringMessage monitoringMessage,
                                List<KeyValue<byte[], MonitoringMessage>> msgList) {
      if (updateMonitoringTopicClock(monitoringMessage.getMonitoringTopicPartition(),
                                     monitoringMessage.getTimestamp())) {
          KeyValue<byte[], Iterable<KeyValue<byte[], MonitoringMessage>>> pair =
              punctuate(partitionTimestamps.get(slowestMonitoringPartition));
          if (pair != null) {
            Iterables.addAll(msgList, pair.value);
          }
      }
    }

    /**
     * Check whether there are missing monitoring data. If there is missing data, adds messages to
     * the 'msgList', one per window that may have a missing monitoring data.
     *
     * @param verifier Monitoring verifier
     * @param now      current local time
     */
    private void verifyMonitoringData(
        MonitoringVerifier verifier,
        long now,
        List<KeyValue<byte[], MonitoringMessage>> msgList
    ) {
      RangeSet<Long> windowRanges = verifier.getMissingMonitoringData(now);
      if (windowRanges != null) {
        log.warn(
            "missing monitoring data for client={} topic={} partition={} session={} window ranges={} {}",
            verifier.getClientId(), verifier.getTopic(), verifier.getPartition(), verifier.getSession(), windowRanges, verifier
        );
        for (Range<Long> range : windowRanges.asRanges()) {
          long win = range.lowerEndpoint();
          while (win <= range.upperEndpoint()) {
            MonitoringMessage.Builder msg = verifier.getCopyOfBaseMonitoringMessage();
            msg.setType(Monitoring.MessageType.ERROR);
            msg.setWindow(win);
            msg.setArrivalTime(now);
            msgList.add(KeyValue.pair(new byte[]{}, msg.build()));
            win += verifier.getSamplePeriod();
          }
        }
      }
    }

    private void notifySessionChange(
        Bytes storeKey,
        MonitoringMessage monitoringMessage, long now,
        List<KeyValue<byte[], MonitoringMessage>> msgList
    ) {
      // find all monitoring verifiers for the same client, topic, partition
      String logKey = logKey(monitoringMessage);
      log.debug("new session={} {}", monitoringMessage.getSession(), logKey);

      //Iterate over all sessions for this (cluster, type, client, topic, partition)
      Bytes prefixKey = keySerde.prefixKeySerde(keySerde.numFields() - 1).key(monitoringMessage);

      try (KeyValueIterator<Bytes, MonitoringVerifier> iter = KVQuery.prefixScan(kvStore, prefixKey)) {
        while (iter.hasNext()) {
          KeyValue<Bytes, MonitoringVerifier> entry = iter.next();
          log.trace("Will verify key={}, Verifier={}", entry.key, entry.value);
          if (entry.value != null && !storeKey.equals(entry.key)) {
            entry.value.notifySessionEnded();
            log.debug("Notified session ended key={}, Verifier={}", entry.key, entry.value);
            verifyMonitoringData(entry.value, now, msgList);
            kvStore.put(entry.key, entry.value);
          } else if (!storeKey.equals(entry.key)) {
            log.debug("Unexpected key, verifier={}", entry.value);
          }
        }
      } catch (Exception e) {
        // Not much we can do, hopefully session will change and we will be able to end
        // session/notify about missing monitoring data next time
        log.error("Failed to record session end for one or more sessions for " + logKey, e);
      }
    }
  }

  private static String logKey(MonitoringMessage message) {
    return Joiner.on(' ').join(
        "clientId=", message.getClientId(),
        "topic=", message.getTopic(),
        "partition=", message.getPartition(),
        "session=", message.getSession()
    );
  }
}
