// (Copyright) [2016 - 2016] Confluent, Inc.
package io.confluent.controlcenter.util;

import io.confluent.serializers.OrderedBytes;
import org.apache.kafka.common.utils.Bytes;
import org.apache.kafka.streams.KeyValue;
import org.apache.kafka.streams.state.KeyValueIterator;
import org.apache.kafka.streams.state.ReadOnlyKeyValueStore;

/**
 *  Helper class for common KVStore query patterns.  Hopefully, this will go away when better
 *  APIs are added to Kafka Streams
 */
public class KVQuery {
  public static final Bytes LAST_KEY = Bytes.wrap(new byte[] { (byte)0xff });
  public static final Bytes FIRST_KEY = Bytes.wrap(new byte[] {});


  public static <T> KeyValueIterator<Bytes, T> rangeFrom(ReadOnlyKeyValueStore<Bytes, T> store, Bytes from) {
    if (from == null) {
      return store.all();
    }
    return store.range(from, LAST_KEY);
  }

  public static <T> KeyValueIterator<Bytes, T> rangeTo(ReadOnlyKeyValueStore<Bytes, T> store, Bytes to) {
    if (to == null) {
      return store.all();
    }
    return store.range(FIRST_KEY, to);
  }

  public static <T> KeyValueIterator<Bytes, T> prefixScan(ReadOnlyKeyValueStore<Bytes, T> store, Bytes prefix) {
    return new PrefixIterator<>(store.range(prefix, LAST_KEY), prefix);
  }

  public static <T> KeyValueIterator<Bytes, T> prefixScan(ReadOnlyKeyValueStore<Bytes, T> store, Bytes prefix, Bytes start) {
    return new PrefixIterator<>(store.range(start, LAST_KEY), prefix);
  }

  static class PrefixIterator<T> implements KeyValueIterator<Bytes, T> {

    private final KeyValueIterator<Bytes, T> iter;
    private final Bytes prefix;
    private KeyValue<Bytes, T> current;

    public PrefixIterator(KeyValueIterator<Bytes, T> iter, Bytes prefix) {
      this.iter = iter;
      this.prefix = prefix;
    }

    @Override
    public void close() {
      current = null;
      iter.close();
    }

    @Override
    public boolean hasNext() {
      if (current != null) {
        return true;
      }
      if (!iter.hasNext()) {
        return false;
      }
      current = iter.next();
      if (!OrderedBytes.startsWith(current.key.get(), prefix.get())) {
        current = null;
        return false;
      }
      return true;
    }

    @Override
    public KeyValue<Bytes, T> next() {
      if (current == null) {
        return iter.next();
      }
      KeyValue<Bytes, T> result = current;
      current = null;
      return result;
    }

    @Override
    public void remove() {
      throw new UnsupportedOperationException("Remove not supported");
    }

  }

}
