// (Copyright) [2016 - 2016] Confluent, Inc.
package io.confluent.controlcenter.data;

import com.google.inject.Provider;
import io.confluent.command.CommandStore;
import io.confluent.command.cluster_metadata.CommandClusterMetadata;
import io.confluent.command.cluster_metadata.CommandClusterMetadata.ClusterMetadata;
import io.confluent.command.cluster_metadata.CommandClusterMetadata.KafkaClusterMetadata;
import io.confluent.command.record.Command.CommandConfigType;
import io.confluent.command.record.Command.CommandKey;
import io.confluent.command.record.Command.CommandMessage;
import io.confluent.controlcenter.record.Controlcenter;
import io.confluent.controlcenter.record.Controlcenter.Cluster;
import io.confluent.controlcenter.record.Controlcenter.ClusterContext;
import io.confluent.controlcenter.record.Controlcenter.ClusterType;
import io.confluent.controlcenter.rest.res.KafkaCluster;
import io.confluent.controlcenter.rest.res.KafkaClusterDisplay;
import io.confluent.controlcenter.rest.res.KafkaClusterListResponse;
import io.confluent.controlcenter.serialization.OrderedKeyPrefixedClusterSerdeSupplier;
import io.confluent.controlcenter.serialization.OrderedKeyPrefixedSerdeSupplier;
import io.confluent.controlcenter.streams.MockReadOnlyKeyValueStore;
import io.confluent.serializers.OrderedKeyPrefixedSerde;
import org.apache.kafka.clients.producer.RecordMetadata;
import org.apache.kafka.common.TopicPartition;
import org.apache.kafka.common.utils.Bytes;
import org.apache.kafka.streams.state.ReadOnlyKeyValueStore;
import org.easymock.EasyMock;
import org.junit.Before;
import org.junit.Test;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.Comparator;
import java.util.List;

import static org.easymock.EasyMock.eq;
import static org.easymock.EasyMock.expect;
import static org.easymock.EasyMock.replay;
import static org.junit.Assert.assertEquals;

public class ClusterMetadataDaoTest {
  private static final OrderedKeyPrefixedSerdeSupplier<ClusterContext, Cluster> serdeSupplier = new OrderedKeyPrefixedClusterSerdeSupplier();
  private MockReadOnlyKeyValueStore<CommandKey, CommandMessage> clusterMetaDataStore;
  private MockReadOnlyKeyValueStore<Bytes, Cluster> clusterStore;
  private ClusterMetadataDao dao;
  private Comparator<KafkaCluster> cmpator = new Comparator<KafkaCluster>() {
    @Override
    public int compare(KafkaCluster o1, KafkaCluster o2) {
      return o1.clusterId.compareTo(o2.clusterId);
    }
  };
  private CommandStore mockCmdStore;

  @Before
  public void setUp() throws Exception {
    clusterStore = new MockReadOnlyKeyValueStore<>();
    OrderedKeyPrefixedSerde<ClusterContext, Cluster> keySerde = serdeSupplier.get(ClusterContext.STREAM_MONITORING);
    Cluster cluster = Cluster.newBuilder()
        .setClusterType(ClusterType.KAFKA)
        .setClusterId("xxxxx")
        .build();
    clusterStore.mockAdd(
        keySerde.key(cluster),
        cluster
    );
    cluster = Cluster.newBuilder()
        .setClusterType(ClusterType.KAFKA)
        .setClusterId("no-meta")
        .setClusterName("Future Cluster Name")
        .build();
    clusterStore.mockAdd(
        keySerde.key(cluster),
        cluster
    );
    cluster = Cluster.newBuilder()
        .setClusterType(ClusterType.KAFKA)
        .build();
    clusterStore.mockAdd(
        keySerde.key(cluster),
        cluster
    );

    clusterMetaDataStore = new MockReadOnlyKeyValueStore<>();
    MockReadOnlyKeyValueStore.mockAddKafkaClusterMetadata(
        clusterMetaDataStore,
        "xxxxx",
        "My Xy Kafka Cluster",
        KafkaClusterMetadata.newBuilder()
            .addAllZookeeperConnect(Arrays.asList("zk1", "zk2"))
            .addAllBootstrapServers(Arrays.asList("k1", "k2", "k3"))
            .build()
    );
    MockReadOnlyKeyValueStore.mockAddKafkaClusterMetadata(
        clusterMetaDataStore,
        "yyyyy",
        "Yay Kafka Cluster",
        KafkaClusterMetadata.newBuilder().build()
    );
    MockReadOnlyKeyValueStore.mockAddConnectClusterMetadata(
        clusterMetaDataStore,
        "PRD Funk Connect",
        CommandClusterMetadata.KafkaConnectClusterMetadata.newBuilder()
            .setKafkaClusterId("xxxxx")
            .addAllRestUrls(Arrays.asList("http://funk.com:9000"))
            .build()
    );
    MockReadOnlyKeyValueStore.mockAddConnectClusterMetadata(
        clusterMetaDataStore,
        "PRD Connect",
        CommandClusterMetadata.KafkaConnectClusterMetadata.newBuilder()
            .setKafkaClusterId("")
            .addAllRestUrls(Arrays.asList("http://empty.com:9000"))
            .build()
    );
    mockCmdStore = EasyMock.mock(CommandStore.class);
    dao = new ClusterMetadataDao(
        clusterMetaDataStore,
        new Provider<ReadOnlyKeyValueStore<Bytes, Cluster>>() {
          @Override
          public ReadOnlyKeyValueStore<Bytes, Cluster> get() {
            return clusterStore;
          }
        },
        serdeSupplier,
        mockCmdStore
    );
  }

  @Test
  public void getKafkaClustersForConnect() throws Exception {
    KafkaClusterListResponse clusters = dao.getKafkaClustersForConnect();
    assertEquals(2, clusters.clusters.size());
    assertEquals("default", clusters.defaultClusterId);
    KafkaClusterDisplay emptyCluster = clusters.clusters.get(0);
    assertEquals("default", emptyCluster.clusterId);
    assertEquals("Unknown", emptyCluster.displayName);

    KafkaClusterDisplay xCluster = clusters.clusters.get(1);
    assertEquals("xxxxx", xCluster.clusterId);
    assertEquals("My Xy Kafka Cluster", xCluster.displayName);
  }

  @Test
  public void getKafkaClustersForStreamMonitoring() throws Exception {
    KafkaClusterListResponse clusters = dao.getKafkaClustersForStreamMonitoring();
    assertEquals(3, clusters.clusters.size());
    assertEquals("xxxxx", clusters.defaultClusterId);

    KafkaClusterDisplay xCluster = clusters.clusters.get(0);
    assertEquals("xxxxx", xCluster.clusterId);
    assertEquals("My Xy Kafka Cluster", xCluster.displayName);

    KafkaClusterDisplay noMetaCluster = clusters.clusters.get(1);
    assertEquals("no-meta", noMetaCluster.clusterId);
    assertEquals("Future Cluster Name", noMetaCluster.displayName);

    KafkaClusterDisplay emptyCluster = clusters.clusters.get(2);
    assertEquals("default", emptyCluster.clusterId);
    assertEquals("Unknown", emptyCluster.displayName);
  }

  @Test
  public void testDefaultNameOverride() throws Exception {
    MockReadOnlyKeyValueStore.mockAddKafkaClusterMetadata(
        clusterMetaDataStore,
        "",
        "My Old Data",
        KafkaClusterMetadata.newBuilder().build()
    );


    KafkaClusterListResponse monClusters = dao.getKafkaClustersForStreamMonitoring();
    assertEquals(3, monClusters.clusters.size());
    KafkaClusterDisplay emptyMonCluster = monClusters.clusters.get(2);
    assertEquals("default", emptyMonCluster.clusterId);
    assertEquals("My Old Data", emptyMonCluster.displayName);

    List<KafkaCluster> clusters = dao.getKafkaClusters();
    Collections.sort(clusters, cmpator);
    assertEquals(4, clusters.size());

    KafkaCluster emptyCluster = clusters.get(0);
    assertEquals("default", emptyCluster.clusterId);
    assertEquals("My Old Data", emptyCluster.displayName);
    assertEquals(0, emptyCluster.zookeeperConnect.size());
    assertEquals(0, emptyCluster.bootstrapServers.size());
  }

  @Test
  public void getKafkaClusters() throws Exception {
    List<KafkaCluster> clusters = dao.getKafkaClusters();
    Collections.sort(clusters, cmpator);
    assertEquals(4, clusters.size());

    KafkaCluster emptyCluster = clusters.get(0);
    assertEquals("default", emptyCluster.clusterId);
    assertEquals("", emptyCluster.displayName);
    assertEquals(0, emptyCluster.zookeeperConnect.size());
    assertEquals(0, emptyCluster.bootstrapServers.size());

    KafkaCluster noMetaCluster = clusters.get(1);
    assertEquals("no-meta", noMetaCluster.clusterId);
    assertEquals("Future Cluster Name", noMetaCluster.displayName);
    assertEquals(0, noMetaCluster.zookeeperConnect.size());
    assertEquals(0, noMetaCluster.bootstrapServers.size());


    KafkaCluster xCluster = clusters.get(2);
    assertEquals("xxxxx", xCluster.clusterId);
    assertEquals("My Xy Kafka Cluster", xCluster.displayName);
    assertEquals(2, xCluster.zookeeperConnect.size());
    assertEquals("zk1", xCluster.zookeeperConnect.get(0));
    assertEquals(3, xCluster.bootstrapServers.size());
    assertEquals("k1", xCluster.bootstrapServers.get(0));

    KafkaCluster yCluster = clusters.get(3);
    assertEquals("yyyyy", yCluster.clusterId);
    assertEquals(0, yCluster.zookeeperConnect.size());
    assertEquals(0, yCluster.bootstrapServers.size());
  }

  @Test
  public void updateClusterFields() throws Exception {
    List<String> newZKList = Arrays.asList("newZK1", "newZK2");
    KafkaCluster updates = new KafkaCluster("xxxxx", "New Name", newZKList, null);
    CommandMessage.Builder builder = CommandMessage.newBuilder();
    dao.updateClusterFields(builder, updates);
    CommandMessage msg = builder.build();
    CommandMessage expected = CommandMessage.newBuilder()
        .setClusterMetadata(ClusterMetadata.newBuilder()
            .setDisplayName("New Name")
            .setKafka(KafkaClusterMetadata.newBuilder()
                .addAllZookeeperConnect(newZKList)
            )
        )
        .build();
    assertEquals(expected, msg);
  }

  @Test
  public void updateClusterFieldsReset() throws Exception {
    KafkaCluster updates = new KafkaCluster("xxxxx", "", new ArrayList<String>(), new ArrayList<String>());
    CommandMessage.Builder builder = CommandMessage.newBuilder()
        .setClusterMetadata(
            ClusterMetadata.newBuilder()
              .setDisplayName("Override this")
              .setKafka(KafkaClusterMetadata.newBuilder()
                  .addAllZookeeperConnect(Arrays.asList("overrideZK1", "overrideZK2"))
              )
        );

    dao.updateClusterFields(builder, updates);
    CommandMessage msg = builder.build();
    CommandMessage expected = CommandMessage.newBuilder()
        .setClusterMetadata(ClusterMetadata.newBuilder()
            .setKafka(KafkaClusterMetadata.newBuilder())
        )
        .build();
    assertEquals(expected, msg);
  }

  @Test
  public void updateClusterFieldsDoNothing() throws Exception {
    KafkaCluster updates = new KafkaCluster("xxxxx", null, null, null);
    CommandMessage.Builder builder = CommandMessage.newBuilder()
        .setClusterMetadata(
            ClusterMetadata.newBuilder()
                .setDisplayName("Keep this")
                .setKafka(KafkaClusterMetadata.newBuilder()
                    .addAllZookeeperConnect(Arrays.asList("keepZk1", "keepZk2"))
                )
        );
    CommandMessage expected = builder.build();

    dao.updateClusterFields(builder, updates);
    CommandMessage msg = builder.build();
    assertEquals(expected, msg);
  }

  @Test
  public void updateKafkaCluster() throws Exception {
    KafkaCluster updates = new KafkaCluster("xxxxx", "New Name", new ArrayList<String>(), new ArrayList<String>());
    CommandKey expectedKey = CommandKey.newBuilder()
        .setConfigType(CommandConfigType.CLUSTER_METADATA)
        .setGuid("xxxxx")
        .build();
    CommandMessage expectedMsg = CommandMessage.newBuilder()
        .setClusterMetadata(ClusterMetadata.newBuilder()
            .setDisplayName("New Name")
            .setKafka(KafkaClusterMetadata.newBuilder())
        )
        .build();

    RecordMetadata recordMeta = new RecordMetadata(new TopicPartition("foo", 0), 0, 0, 0, 0, 0, 0);
    long timeoutMs = 99L;
    expect(mockCmdStore.awaitCommand(eq(expectedKey), eq(expectedMsg), eq(timeoutMs))).andReturn(recordMeta).once();
    replay(mockCmdStore);
    dao.updateKafkaCluster(updates.clusterId, updates, timeoutMs);
  }

  @Test
  public void updateKafkaClusterMissingId() throws Exception {
    KafkaCluster updates = new KafkaCluster("default", "New Name", new ArrayList<String>(), new ArrayList<String>());
    CommandKey expectedKey = CommandKey.newBuilder()
        .setConfigType(CommandConfigType.CLUSTER_METADATA)
        .setGuid("")
        .build();
    CommandMessage expectedMsg = CommandMessage.newBuilder()
        .setClusterMetadata(ClusterMetadata.newBuilder()
            .setDisplayName("New Name")
            .setKafka(KafkaClusterMetadata.newBuilder())
        )
        .build();

    RecordMetadata recordMeta = new RecordMetadata(new TopicPartition("foo", 0), 0, 0, 0, 0, 0, 0);
    long timeoutMs = 99L;
    expect(mockCmdStore.awaitCommand(eq(expectedKey), eq(expectedMsg), eq(timeoutMs))).andReturn(recordMeta).once();
    replay(mockCmdStore);
    dao.updateKafkaCluster(updates.clusterId, updates, timeoutMs);
  }
}