package io.confluent.controlcenter.streams;

import com.google.common.collect.Lists;
import com.google.common.collect.Maps;
import io.confluent.command.cluster_metadata.CommandClusterMetadata.ClusterMetadata;
import io.confluent.command.cluster_metadata.CommandClusterMetadata.KafkaClusterMetadata;
import io.confluent.command.cluster_metadata.CommandClusterMetadata.KafkaConnectClusterMetadata;
import io.confluent.command.record.Command.CommandConfigType;
import io.confluent.command.record.Command.CommandKey;
import io.confluent.command.record.Command.CommandMessage;
import io.confluent.command.record.alert.CommandAlert.ActionConfig;
import io.confluent.command.record.alert.CommandAlert.MonitoringTriggerConfig;
import org.apache.kafka.streams.KeyValue;
import org.apache.kafka.streams.state.KeyValueIterator;
import org.apache.kafka.streams.state.ReadOnlyKeyValueStore;

import java.util.AbstractMap.SimpleEntry;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

public class MockReadOnlyKeyValueStore<K, V> implements ReadOnlyKeyValueStore<K, V> {
  private final Map<K, V> internal = Maps.newHashMap();
  private final List<Entry<K, V>> order = Lists.newArrayList();

  public synchronized void mockAdd(K key, V value) {
    order.add(new SimpleEntry<>(key, value));
    this.internal.put(key, value);
  }

  public synchronized void mockReplace(Integer i, V value) {
    this.order.get(i).setValue(value);
    this.internal.put(this.order.get(i).getKey(), value);
  }

  public V mockGet(Integer i) {
    return this.order.get(i).getValue();
  }

  @Override
  public V get(K key) {
    return internal.get(key);
  }

  @Override
  public KeyValueIterator<K, V> range(K from, K to) {

    final Iterator<Entry<K, V>> messages = order.iterator();
    return new KeyValueIterator<K, V>() {
      @Override
      public void close() {
      }

      @Override
      public boolean hasNext() {
        return messages.hasNext();
      }

      @Override
      public KeyValue<K, V> next() {
        Entry<K, V> entry = messages.next();
        return KeyValue.pair(entry.getKey(), entry.getValue());
      }

      @Override
      public void remove() {
      }
    };
  }

  @Override
  public KeyValueIterator<K, V> all() {
    return range(null, null);
  }

  @Override
  public long approximateNumEntries() {
    return internal.size();
  }

  public static void mockAddMonitoringTriggerConfig(MockReadOnlyKeyValueStore<CommandKey, CommandMessage> store, MonitoringTriggerConfig monitoringTriggerConfig) {
    store.mockAdd(
        CommandKey.newBuilder()
            .setConfigType(CommandConfigType.MONITORING_TRIGGER_CONFIG)
            .setGuid(monitoringTriggerConfig.getGuid())
            .build(),
        CommandMessage.newBuilder()
            .setMonitoringTriggerConfig(monitoringTriggerConfig)
            .build()
    );
  }

  public static void mockAddActionConfig(MockReadOnlyKeyValueStore<CommandKey, CommandMessage> store, ActionConfig actionConfig) {
    store.mockAdd(
        CommandKey.newBuilder()
            .setConfigType(CommandConfigType.ACTION_CONFIG)
            .setGuid(actionConfig.getGuid())
            .build(),
        CommandMessage.newBuilder()
            .setActionConfig(actionConfig)
            .build()
    );
  }

  public static void mockReplaceActionConfig(MockReadOnlyKeyValueStore<CommandKey, CommandMessage> store, Integer index, ActionConfig actionConfig) {
    store.mockReplace(
        index,
        CommandMessage.newBuilder()
            .setActionConfig(actionConfig)
            .build()
    );
  }

  public static void mockAddKafkaClusterMetadata(MockReadOnlyKeyValueStore<CommandKey, CommandMessage> store, String clusterId, String displayName, KafkaClusterMetadata kafkaClusterMetadata) {
    mockAddClusterMetadata(
        store,
        clusterId,
        ClusterMetadata.newBuilder()
          .setDisplayName(displayName)
          .setKafka(kafkaClusterMetadata)
          .build()
    );
  }

  public static void mockAddConnectClusterMetadata(MockReadOnlyKeyValueStore<CommandKey, CommandMessage> store, String displayName, KafkaConnectClusterMetadata connectClusterMetadata) {
    mockAddClusterMetadata(
        store,
        "fake-cluster",
        ClusterMetadata.newBuilder()
            .setDisplayName(displayName)
            .setConnect(connectClusterMetadata)
            .build()
    );
  }

  public static void mockAddClusterMetadata(MockReadOnlyKeyValueStore<CommandKey, CommandMessage> store, String clusterId, ClusterMetadata clusterMetadata) {
    store.mockAdd(
        CommandKey.newBuilder()
            .setConfigType(CommandConfigType.CLUSTER_METADATA)
            .setGuid(clusterId)
            .build(),
        CommandMessage.newBuilder()
            .setClusterMetadata(clusterMetadata)
            .build()
    );
  }
}
