package io.confluent.controlcenter.streams.alert;


import com.google.common.collect.ImmutableList;
import com.google.inject.Guice;
import com.google.inject.Injector;
import com.google.inject.Key;

import junit.framework.AssertionFailedError;

import org.apache.kafka.clients.producer.Producer;
import org.apache.kafka.clients.producer.ProducerRecord;
import org.apache.kafka.common.serialization.Serializer;
import org.apache.kafka.common.utils.Bytes;
import org.apache.kafka.streams.processor.Processor;
import org.apache.kafka.streams.processor.StateStoreSupplier;
import org.apache.kafka.streams.processor.internals.RecordCollector;
import org.apache.kafka.streams.processor.internals.RecordCollectorImpl;
import org.apache.kafka.streams.state.KeyValueIterator;
import org.apache.kafka.streams.state.KeyValueStore;
import org.apache.kafka.streams.state.Stores;
import org.easymock.EasyMock;
import org.junit.Before;
import org.junit.Test;

import java.io.File;
import java.nio.file.Files;
import java.util.List;

import io.confluent.command.record.alert.CommandAlert.EmailAction;
import io.confluent.command.record.alert.CommandAlert.MonitoringTriggerConfig;
import io.confluent.command.record.alert.CommandAlert.TriggerCondition;
import io.confluent.command.record.alert.CommandAlert.TriggerMetricType;
import io.confluent.controlcenter.alert.record.Alert.ActionInfo;
import io.confluent.controlcenter.alert.record.Alert.AlertInfo;
import io.confluent.controlcenter.alert.record.Alert.TriggerInfo;
import io.confluent.controlcenter.serialization.SerializationModule;
import io.confluent.controlcenter.serialization.SerializationModule.TriggerEventsKeyPrefixSerde;
import io.confluent.controlcenter.serialization.SerializationModule.TriggerEventsKeySerde;
import io.confluent.controlcenter.streams.MockProcessorContext;
import io.confluent.controlcenter.util.KVQuery;
import io.confluent.serializers.OrderedKeyUberSerde;
import io.confluent.serializers.ProtoSerde;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertTrue;

public class AlertHistoryProcessorSupplierTest {
  private static final String STORE_NAME = "AlertHistoryTestStore";
  private static final long START_TIME_MS = 360000002L;
  private static final String DEFAULT_CLUSTER_ID = "cluster-0";
  private static final String DEFAULT_CONSUMER_GROUP = "cg-0";
  private static final long retentionPeriodMs = 36000000;  // 10 hours
  private static final long punctuateIntervalMs = 15000;
  private static final long maxPunctuateItems = 5;
  private static final Injector injector = Guice.createInjector(new SerializationModule());
  private static final OrderedKeyUberSerde<AlertInfo> alertKeySerde =
      injector.getInstance(new Key<OrderedKeyUberSerde<AlertInfo>>(TriggerEventsKeySerde.class) {});
  OrderedKeyUberSerde<AlertInfo> alertRangeSerde =
      injector.getInstance(new Key<OrderedKeyUberSerde<AlertInfo>>(TriggerEventsKeyPrefixSerde.class) {});

  private MockProcessorContext context;
  private Producer<byte[], byte[]> mockProducer;
  private KeyValueStore<Bytes, AlertInfo> kvStore;
  private File stateDir;

  private AlertHistoryProcessorSupplier supplier;
  private Processor<Bytes, AlertInfo> processor;

  @Before
  public void setUp() throws Exception {
    StateStoreSupplier storeSupplier = Stores.create(STORE_NAME)
        .withKeys(alertKeySerde)
        .withValues(new ProtoSerde<>(AlertInfo.getDefaultInstance()))
        .persistent()
        .build();

    // The code block is needed to be able to test persistent kv store backed by rocksdb
    stateDir = Files.createTempDirectory("test").toFile();
    stateDir.deleteOnExit();
    mockProducer = EasyMock.createMock(Producer.class);
    RecordCollector recordCollector = new RecordCollectorImpl(mockProducer, "fakeTaskId") {
      @Override
      public <K1, V1> void send(ProducerRecord<K1, V1> record, Serializer<K1> keySerializer,
                                Serializer<V1> valueSerializer) {
        // nothing
      }
    };
    context = new MockProcessorContext(stateDir, recordCollector);

    kvStore = (KeyValueStore<Bytes, AlertInfo>) storeSupplier.get();
    kvStore.init(context, kvStore);
    context.register(kvStore, false, null);

    supplier = new AlertHistoryProcessorSupplier(STORE_NAME, retentionPeriodMs,
                                                 punctuateIntervalMs, maxPunctuateItems,
                                                 alertKeySerde, alertRangeSerde);
    processor = supplier.get();
    processor.init(context);
  }

  private static final AlertInfo makeAlertInfo(String guid, Long timestamp,
                                               MonitoringTriggerConfig triggerCfg,
                                               List<TriggerInfo> triggers) {
    ActionInfo.Builder actionBuilder = ActionInfo.newBuilder().setEmail(
        EmailAction.newBuilder().setSubject(guid).setAddress("a@b").build());

    return AlertInfo.newBuilder()
        .setGuid(guid)
        .setTimestamp(timestamp)
        .setMonitoringTrigger(triggerCfg)
        .addActions(actionBuilder)
        .addAllTriggers(triggers)
        .build();
  }

  private static final AlertInfo makeAlertInfo(String guid, Long timestamp,
                                               String triggerGuid,
                                               Long triggerWindow,
                                               Long triggerConsumptionDiff) {
    return makeAlertInfo(guid, timestamp,
                         makeTriggerConfig(triggerGuid),
                         ImmutableList.of(makeTriggerInfo(triggerWindow, triggerConsumptionDiff)));
  }

  private static final MonitoringTriggerConfig makeTriggerConfig(String guid) {
    return MonitoringTriggerConfig.newBuilder()
        .setClusterId(DEFAULT_CLUSTER_ID)
        .setGroup(DEFAULT_CONSUMER_GROUP)
        .setGuid(guid)
        .setName(guid)
        .setMetric(TriggerMetricType.CONSUMPTION_DIFF)
        .setCondition(TriggerCondition.NOT_EQUAL)
        .setLongValue(0)
        .setLagMs(60000L)
        .build();
  }

  private static final TriggerInfo makeTriggerInfo(Long window, Long consumptionDiff) {
    return TriggerInfo.newBuilder()
        .setWindow(window)
        .setLongValue(consumptionDiff)
        .build();
  }

  @Test
  public void testBasicPutAndGet() {
    final AlertInfo alert = makeAlertInfo("alert1", START_TIME_MS, "trigger1", 100L, 10L);

    processor.process(alertKey(alert), alert);

    AlertInfo retAlert = kvStore.get(alertKey(alert));
    assertEquals(alert, retAlert);
  }

  @Test
  public void testRangeGet() {
    Long timestamp = START_TIME_MS;
    ImmutableList<AlertInfo> alerts =
        ImmutableList.of(makeAlertInfo("alert1", timestamp, "trigger1", 1000L, 10L),
                         makeAlertInfo("alert2", timestamp + 60000L, "trigger1", 2000L, 9L),
                         makeAlertInfo("alert3", timestamp + 120000L, "trigger1", 3000L, 9L),
                         makeAlertInfo("alert4", timestamp + 180000L, "trigger1", 4000L, 1L));

    for (AlertInfo alert : alerts) {
      processor.process(alertKey(alert), alert);
    }
    verifyStoredAlerts(timestamp + 180001L, alerts);
  }

  @Test
  public void testAlertsRemovedAfterRetentionPeriod() {
    AlertInfo alert1 = makeAlertInfo("alert1", START_TIME_MS, "trigger1", 1000L, 10L);
    AlertInfo alert2 = makeAlertInfo("alert2", START_TIME_MS + retentionPeriodMs + 1L,
                                     "trigger1", 2000L, 8L);

    processor.process(alertKey(alert1), alert1);
    verifyStoredAlerts(START_TIME_MS + retentionPeriodMs +2L,
                       ImmutableList.of(alert1));

    processor.process(alertKey(alert2), alert2);
    verifyStoredAlerts(START_TIME_MS + retentionPeriodMs + 2L,
                       ImmutableList.of(alert2));
  }

  @Test
  public void testPunctuateInterval() {
    AlertInfo alert1 = makeAlertInfo("alert1", START_TIME_MS, "trigger1", 1000l, 10L);
    AlertInfo alert2 = makeAlertInfo("alert2", START_TIME_MS + punctuateIntervalMs - 10L,
                                     "trigger1", 2000L, 8L);
    AlertInfo alert3 = makeAlertInfo("alert3", START_TIME_MS + retentionPeriodMs + 1L,
                                     "trigger1", 3000L, 9L);

    processor.process(alertKey(alert1), alert1);
    verifyStoredAlerts(START_TIME_MS + retentionPeriodMs + 2L,
                       ImmutableList.of(alert1));

    processor.process(alertKey(alert2), alert2);
    verifyStoredAlerts(START_TIME_MS + retentionPeriodMs + 2L,
                       ImmutableList.of(alert1, alert2));

    processor.process(alertKey(alert3), alert3);
    verifyStoredAlerts(START_TIME_MS + retentionPeriodMs + 2L,
                       ImmutableList.of(alert2, alert3));
  }

  @Test
  public void testPunctuateStopsAfterMaxPunctuateItems() {
    ImmutableList<AlertInfo> alerts =
        ImmutableList.of(
            makeAlertInfo("alert1", START_TIME_MS, "trigger1", 1000l, 10L),
            makeAlertInfo("alert2", START_TIME_MS + 60000L, "trigger1", 2000L, 9L),
            makeAlertInfo("alert3", START_TIME_MS + 120000L, "trigger1", 3000L, 9L),
            makeAlertInfo("alert4", START_TIME_MS, "trigger1", 4000L, 10L),
            makeAlertInfo("alert5", START_TIME_MS + 60000L, "trigger1", 5000L, 9L),
            makeAlertInfo("alert6", START_TIME_MS + 120000L, "trigger1", 6000l, 9L),
            makeAlertInfo("alert7", START_TIME_MS + 180000L, "trigger1", 7000l, 1L),
            makeAlertInfo("alert8", START_TIME_MS + 240000L, "trigger1", 8000l, 1L),
            makeAlertInfo("alert9", START_TIME_MS + retentionPeriodMs + 240001l, "trigger1", 9000l, 10L)
        );

    for (AlertInfo alert : alerts) {
      processor.process(alertKey(alert), alert);
    }

    verifyStoredAlerts(
        START_TIME_MS + retentionPeriodMs + 240002L,
        ImmutableList.<AlertInfo>builder().add(
            alerts.get(0),
            alerts.get(3),
            alerts.get(4),
            alerts.get(8)
        ).build()
    );

    AlertInfo alert10 = makeAlertInfo("alert10", START_TIME_MS + retentionPeriodMs + 300001L,
                                      "trigger1", 10000L, 10L);
    processor.process(alertKey(alert10), alert10);
    verifyStoredAlerts(START_TIME_MS + retentionPeriodMs + 300002L,
                       ImmutableList.of(alerts.get(alerts.size() - 1), alert10));
  }

  private Bytes alertKey(AlertInfo alertInfo) {
    return alertKeySerde.key(alertInfo);
  }

  private void verifyStoredAlerts(Long stopTimestamp,
                                  List<AlertInfo> alerts) {
    Bytes stop = alertRangeSerde.key(AlertInfo.newBuilder().setTimestamp(stopTimestamp).build());

    try (KeyValueIterator<Bytes, AlertInfo> it = KVQuery.rangeFrom(kvStore, null)) {
      int count = 0;
      while (it.hasNext()) {
        AlertInfo alert = it.next().value;
        assertNotNull(alert);
        assertNotNull(alerts);
        assertTrue(
            String.format("Unexpected alert %s in store", alert.getGuid()),
            alerts.contains(alert)
        );
        count++;
      }
      assertEquals(alerts.size(), count);
    } catch (Exception e) {
      throw new AssertionFailedError("Failed to verify stored alerts: " + e.getMessage());
    }
  }
}
