package io.confluent.controlcenter.streams.alert;


import com.google.common.collect.ImmutableList;
import com.google.inject.Guice;
import com.google.inject.Injector;
import com.google.inject.Key;
import com.google.inject.Provider;
import com.google.inject.util.Providers;

import junit.framework.AssertionFailedError;

import org.apache.kafka.clients.producer.Producer;
import org.apache.kafka.clients.producer.ProducerRecord;
import org.apache.kafka.common.serialization.Serializer;
import org.apache.kafka.common.utils.Bytes;
import org.apache.kafka.streams.KeyValue;
import org.apache.kafka.streams.kstream.Transformer;
import org.apache.kafka.streams.processor.StateStoreSupplier;
import org.apache.kafka.streams.processor.internals.RecordCollector;
import org.apache.kafka.streams.processor.internals.RecordCollectorImpl;
import org.apache.kafka.streams.state.KeyValueIterator;
import org.apache.kafka.streams.state.KeyValueStore;
import org.apache.kafka.streams.state.Stores;
import org.easymock.EasyMock;
import org.junit.Before;
import org.junit.Test;

import java.io.File;
import java.nio.file.Files;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;

import io.confluent.command.record.Command.CommandKey;
import io.confluent.command.record.Command.CommandMessage;
import io.confluent.command.record.alert.CommandAlert.EmailAction;
import io.confluent.command.record.alert.CommandAlert.ActionConfig;
import io.confluent.command.record.alert.CommandAlert.MonitoringTriggerConfig;
import io.confluent.command.record.alert.CommandAlert.TriggerCondition;
import io.confluent.command.record.alert.CommandAlert.TriggerMetricType;
import io.confluent.controlcenter.alert.record.Alert.AlertInfo;
import io.confluent.controlcenter.alert.record.Alert.TriggerInfo;
import io.confluent.controlcenter.record.Controlcenter.TriggerActions;
import io.confluent.controlcenter.record.Controlcenter.TriggerEvent;
import io.confluent.controlcenter.record.Controlcenter.TriggerEventKey;
import io.confluent.controlcenter.record.Controlcenter.VerifiableTriggerInfo;
import io.confluent.controlcenter.serialization.SerializationModule;
import io.confluent.controlcenter.serialization.SerializationModule.TriggerActionsKeySerde;
import io.confluent.controlcenter.serialization.SerializationModule.TriggerEventsKeySerde;
import io.confluent.controlcenter.streams.MockClock;
import io.confluent.controlcenter.streams.MockProcessorContext;
import io.confluent.controlcenter.streams.MockReadOnlyKeyValueStore;
import io.confluent.controlcenter.util.KVQuery;
import io.confluent.monitoring.common.TimeBucket;
import io.confluent.serializers.OrderedBytes;
import io.confluent.serializers.OrderedKeyUberSerde;
import io.confluent.serializers.ProtoSerde;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertNull;
import static org.junit.Assert.assertTrue;

public class AlertTransformerSupplierTest {

  private static final String STORE_NAME = "AlertTransformerTestStore";
  private static final String TRIGGER_EVENT_STORE_NAME = "AlertTransformerTriggerEventsStore";
  private static final int MAX_TRIGGER_EVENTS_IN_ALERT = 10;
  private static final long START_TIME_MS = 36000000L;
  private static final String DEFAULT_CLUSTER_ID = "cluster-0";
  private static final String DEFAULT_CONSUMER_GROUP = "cg-0";
  private static final Bytes EMPTY_BYTE_ARRAY = Bytes.wrap(new byte[] {});
  private static final Injector injector = Guice.createInjector(new SerializationModule());
  private static final OrderedKeyUberSerde<TriggerEventKey> actionsKeySerde =
      injector.getInstance(new Key<OrderedKeyUberSerde<TriggerEventKey>>(TriggerActionsKeySerde.class) {});
  private static final OrderedKeyUberSerde<AlertInfo> eventsKeySerde =
      injector.getInstance(new Key<OrderedKeyUberSerde<AlertInfo>>(TriggerEventsKeySerde.class) {});

  private final MockClock mockClock = new MockClock();

  private MockProcessorContext context;
  private Producer<byte[], byte[]> mockProducer;
  private KeyValueStore<Bytes, TriggerActions> kvStore;
  private KeyValueStore<Bytes, VerifiableTriggerInfo> kvTriggerEventsStore;
  private File stateDir;

  private AlertTransformerSupplier supplier;
  private Transformer<Bytes, TriggerEvent,
      KeyValue<Bytes, Iterable<KeyValue<Bytes, AlertInfo>>>> transformer;

  private MockReadOnlyKeyValueStore<CommandKey, CommandMessage> triggerConfigs;
  private MockReadOnlyKeyValueStore<CommandKey, CommandMessage> actionConfigs;

  private static TriggerEvent makeTriggerEvent(String triggerGuid, Long window, Long timestamp,
                                               Long consumptionDiff) {
    TriggerInfo.Builder builder = TriggerInfo.newBuilder()
        .setWindow(window)
        .setLongValue(consumptionDiff);

    return TriggerEvent.newBuilder()
        .setTimestamp(timestamp)
        .setTriggerGuid(triggerGuid)
        .setInfo(builder)
        .build();
  }

  private static TriggerEvent makeTriggerEvent(
      MonitoringTriggerConfig config,
      Long window,
      Long timestamp
  ) {
    TriggerInfo.Builder builder = TriggerInfo.newBuilder().setWindow(window);
    long value = 0;
    switch (config.getCondition()) {
      case GREATER_THAN:
        value = config.getLongValue() + 10;
        break;
      case LESS_THAN:
        value = config.getLongValue() - 10;
        break;
      case EQUAL:
        value = config.getLongValue() + 1;
        break;
      case NOT_EQUAL:
        value = config.getLongValue();
        break;
    }

    switch (config.getMetric()) {
      case CONSUMPTION_DIFF:
      case AVG_LATENCY:
      case MAX_LATENCY:
        builder.setLongValue(value);
        break;
    }

    return TriggerEvent.newBuilder()
        .setTimestamp(timestamp)
        .setTriggerGuid(config.getGuid())
        .setInfo(builder)
        .build();
  }

  private static MonitoringTriggerConfig makeTriggerConfig(
      String clusterId,
      String group,
      String guid,
      TriggerMetricType metricType,
      TriggerCondition condition,
      long value
  ) {
    return MonitoringTriggerConfig.newBuilder()
        .setClusterId(clusterId)
        .setGroup(group)
        .setGuid(guid)
        .setName(guid)
        .setMetric(metricType)
        .setCondition(condition)
        .setLongValue(value)
        .setLagMs(60000L)
        .build();
  }

  private static ActionConfig makeActionConfig(
      boolean enabled,
      String guid,
      int rate,
      long interval,
      List<String> triggerGuids
  ) {
    return ActionConfig.newBuilder()
        .setEnabled(enabled)
        .setGuid(guid)
        .setEmail(
            EmailAction.newBuilder()
                .setSubject(guid)
                .setAddress("a@b")
                .build()
        )
        .setMaxSendRateNumerator(rate)
        .setIntervalMs(interval)
        .addAllTriggerGuid(triggerGuids)
        .build();
  }

  private static ActionConfig makeActionConfig(
      String guid,
      int rate,
      long interval,
      List<String> triggerGuids
  ) {
    return makeActionConfig(true, guid, rate, interval, triggerGuids);
  }

  private static ActionConfig makeDisabledActionConfig(
      String guid,
      int rate,
      long interval,
      List<String> triggerGuids
  ) {
    return makeActionConfig(false, guid, rate, interval, triggerGuids);
  }

  private static TriggerActions getStoredTriggerActions(KeyValueStore<Bytes, TriggerActions> store,
                                                        TriggerEvent event) {
    TriggerEventKey triggerEventKey = TriggerEventKey.newBuilder()
        .setTriggerGuid(event.getTriggerGuid())
        .build();
    Bytes storeKey = actionsKeySerde.prefixKeySerde(1).key(triggerEventKey);
    return store.get(storeKey);
  }

  private static List<VerifiableTriggerInfo> getStoredTriggerEvents(
      KeyValueStore<Bytes, VerifiableTriggerInfo> store, TriggerEvent event) {

    TriggerEventKey prefixEventKey = TriggerEventKey.newBuilder()
        .setTriggerGuid(event.getTriggerGuid())
        .build();

    Bytes prefixKey = actionsKeySerde.prefixKeySerde(1).key(prefixEventKey);
    List<VerifiableTriggerInfo> triggerInfoList = new ArrayList<>();
    try (KeyValueIterator<Bytes, VerifiableTriggerInfo> iter = KVQuery.prefixScan(store, prefixKey)) {
      while (iter.hasNext()) {
        KeyValue<Bytes, VerifiableTriggerInfo> entry = iter.next();
        if (!OrderedBytes.startsWith(entry.key.get(), prefixKey.get())) {
          break;
        }
        triggerInfoList.add(entry.value);
      }
    }
    return triggerInfoList;
  }

  @Before
  public void setUp() throws Throwable {
    StateStoreSupplier storeSupplier = Stores.create(STORE_NAME)
        .withKeys(actionsKeySerde)
        .withValues(new ProtoSerde<>(TriggerActions.getDefaultInstance()))
        .persistent()
        .build();

    StateStoreSupplier storeSupplier2 = Stores.create(TRIGGER_EVENT_STORE_NAME)
        .withKeys(eventsKeySerde)
        .withValues(new ProtoSerde<>(VerifiableTriggerInfo.getDefaultInstance()))
        .persistent()
        .build();

    // The code block is needed to be able to test persistent kv store backed by rocksdb
    stateDir = Files.createTempDirectory("test").toFile();
    stateDir.deleteOnExit();
    mockProducer = EasyMock.createMock(Producer.class);
    RecordCollector recordCollector = new RecordCollectorImpl(mockProducer, "fakeTaskId") {
      @Override
      public <K1, V1> void send(ProducerRecord<K1, V1> record, Serializer<K1> keySerializer,
                                Serializer<V1> valueSerializer) {
        // nothing
      }
    };
    context = new MockProcessorContext(stateDir, recordCollector);

    kvStore = (KeyValueStore<Bytes, TriggerActions>) storeSupplier.get();
    kvStore.init(context, kvStore);
    context.register(kvStore, false, null);

    kvTriggerEventsStore = (KeyValueStore<Bytes, VerifiableTriggerInfo>) storeSupplier2.get();
    kvTriggerEventsStore.init(context, kvTriggerEventsStore);
    context.register(kvTriggerEventsStore, false, null);

    mockClock.setCurrentTimeMillis(START_TIME_MS);
    this.triggerConfigs = new MockReadOnlyKeyValueStore<>();
    this.actionConfigs = new MockReadOnlyKeyValueStore<>();
    supplier = new AlertTransformerSupplier(
        STORE_NAME, TRIGGER_EVENT_STORE_NAME,
        MAX_TRIGGER_EVENTS_IN_ALERT,
        mockClock,
        actionsKeySerde,
        eventsKeySerde,
        Providers.of(triggerConfigs),
        Providers.of(actionConfigs)
    );
  }

  @Test
  public void testTransformIgnoresNonActionableTrigger() {
    initTransformer();

    final TriggerEvent event = makeTriggerEvent("123", 100L, START_TIME_MS, 1L);

    KeyValue<Bytes, Iterable<KeyValue<Bytes, AlertInfo>>> ret = transformer.transform(EMPTY_BYTE_ARRAY, event);
    assertNull(ret);
    assertNull(getStoredTriggerActions(kvStore, event));
  }

  @Test
  public void testInitDoesNotInitializeStore() {
    initTransformerWithOneTriggerOneAction();

    int count = 0;
    try (KeyValueIterator<Bytes, TriggerActions> it = KVQuery.rangeFrom(kvStore, null)) {
      while (it.hasNext()) {
        count++;
      }
    } catch (Exception e) {
      throw new AssertionFailedError("Failed to verify stored trigger actions: " + e.getMessage());
    }
    assertEquals(0, count);
  }

  @Test
  public void testTransformCreateAlertOnFirstActionableTriggerEvent() {
    initTransformerWithOneTriggerOneAction();

    final TriggerEvent event = makeTriggerEvent(triggerConfigs.mockGet(0).getMonitoringTriggerConfig(), 1000L, START_TIME_MS);

    KeyValue<Bytes, Iterable<KeyValue<Bytes, AlertInfo>>> ret = transformer.transform(EMPTY_BYTE_ARRAY, event);
    assertNotNull(ret);
    Iterator<KeyValue<Bytes, AlertInfo>> it = ret.value.iterator();
    assertTrue(it.hasNext());
    assertEquals("Expect 1 trigger event in the alert", 1, it.next().value.getTriggersCount());

    // all trigger events list in trigger actions must be empty
    List<VerifiableTriggerInfo> triggerEvents = getStoredTriggerEvents(kvTriggerEventsStore, event);
    assertTrue(triggerEvents.isEmpty());
  }

  @Test
  public void testTransformQueueSecondActionableEvent() {
    initTransformerWithOneTriggerOneAction();

    final TriggerEvent event1 = makeTriggerEvent(triggerConfigs.mockGet(0).getMonitoringTriggerConfig(), 1000L, START_TIME_MS);
    final TriggerEvent event2 = makeTriggerEvent(triggerConfigs.mockGet(0).getMonitoringTriggerConfig(), 2000L, START_TIME_MS);

    KeyValue<Bytes, Iterable<KeyValue<Bytes, AlertInfo>>> ret = transformer.transform(EMPTY_BYTE_ARRAY, event1);
    verifyAlerts(ret, 1, 1, 1);

    assertNull(transformer.transform(EMPTY_BYTE_ARRAY, event2));

    // trigger events list in trigger actions must contain 1 event
    List<VerifiableTriggerInfo> triggerEvents = getStoredTriggerEvents(kvTriggerEventsStore, event2);
    assertEquals(1, triggerEvents.size());
    assertEquals(2000L, triggerEvents.get(0).getInfo().getWindow());
  }

  @Test
  public void testMaxRatePreservedOnDifferentEventArrivalRate() {
    initTransformerWithOneTriggerOneAction();

    final TriggerEvent event1 = makeTriggerEvent(triggerConfigs.mockGet(0).getMonitoringTriggerConfig(), 1000L, START_TIME_MS);

    KeyValue<Bytes, Iterable<KeyValue<Bytes, AlertInfo>>> ret = transformer.transform(EMPTY_BYTE_ARRAY, event1);
    verifyAlerts(ret, 1, 1, 1);

    assertNull(transformer.transform(EMPTY_BYTE_ARRAY, makeTriggerEvent(triggerConfigs.mockGet(0).getMonitoringTriggerConfig(),
                                                                        2000L, START_TIME_MS + 45000L)));

    // trigger events list in trigger actions must contain 1 event
    List<VerifiableTriggerInfo> triggerEvents = getStoredTriggerEvents(kvTriggerEventsStore, event1);
    assertEquals(1, triggerEvents.size());
    assertEquals(2000L, triggerEvents.get(0).getInfo().getWindow());

    assertNull(transformer.transform(EMPTY_BYTE_ARRAY, makeTriggerEvent(triggerConfigs.mockGet(0).getMonitoringTriggerConfig(),
                                                                        3000L, START_TIME_MS + 55000L)));
    List<VerifiableTriggerInfo> triggerEvents2 = getStoredTriggerEvents(kvTriggerEventsStore, event1);
    assertEquals(2, triggerEvents2.size());

  }

  @Test
  public void testTransformSendsAlertsWithMaxRate() {
    initTransformerWithOneTriggerOneAction();

    final TriggerEvent event1 = makeTriggerEvent(triggerConfigs.mockGet(0).getMonitoringTriggerConfig(),
                                                 1000L, START_TIME_MS);
    final TriggerEvent event2 = makeTriggerEvent(triggerConfigs.mockGet(0).getMonitoringTriggerConfig(),
                                                 2000L, START_TIME_MS + 10 * actionConfigs.mockGet(0).getActionConfig().getIntervalMs());
    final TriggerEvent event3 = makeTriggerEvent(triggerConfigs.mockGet(0).getMonitoringTriggerConfig(), 3000L,
                                                 START_TIME_MS + 10 * actionConfigs.mockGet(0).getActionConfig().getIntervalMs());

    assertNotNull(transformer.transform(EMPTY_BYTE_ARRAY, event1));
    assertNotNull(transformer.transform(EMPTY_BYTE_ARRAY, event2));

    // this will break if rate of the action changes in testTransformSendsAlertsWithMaxRate
    assertNull(transformer.transform(EMPTY_BYTE_ARRAY, event3));

    // trigger events list in trigger actions must contain 1 event
    List<VerifiableTriggerInfo> triggerEvents = getStoredTriggerEvents(kvTriggerEventsStore, event2);
    assertEquals(1, triggerEvents.size());
    assertEquals(3000L, triggerEvents.get(0).getInfo().getWindow());
  }

  @Test
  public void testCreateActionAfterFirstTransform() {
    // create one trigger, but no actions
    ImmutableList<String> triggers = ImmutableList.of("trigger1");
    MockReadOnlyKeyValueStore.mockAddMonitoringTriggerConfig(
        triggerConfigs,
        makeTriggerConfig(
            DEFAULT_CLUSTER_ID,
            DEFAULT_CONSUMER_GROUP,
            triggers.get(0),
            TriggerMetricType.CONSUMPTION_DIFF,
            TriggerCondition.NOT_EQUAL,
            0
        )
    );
    initTransformer();

    final TriggerEvent event1 = makeTriggerEvent(triggerConfigs.mockGet(0).getMonitoringTriggerConfig(),
                                                 1000L,
                                                 START_TIME_MS);
    final TriggerEvent event2 = makeTriggerEvent(triggerConfigs.mockGet(0).getMonitoringTriggerConfig(),
                                                 2000L,
                                                 START_TIME_MS);

    // no actions, no alerts
    assertNull(transformer.transform(EMPTY_BYTE_ARRAY, event1));

    // add action to the trigger
    MockReadOnlyKeyValueStore.mockAddActionConfig(actionConfigs, makeActionConfig("action1", 1, 60000L, triggers));
    KeyValue<Bytes, Iterable<KeyValue<Bytes, AlertInfo>>> ret = transformer.transform(EMPTY_BYTE_ARRAY, event2);
    verifyAlerts(ret, 1, 1, 1);
  }

  @Test
  public void testAddActionAfterFirstTransform() {
    initTransformerWithOneTriggerOneAction();
    final TriggerEvent event1 = makeTriggerEvent(triggerConfigs.mockGet(0).getMonitoringTriggerConfig(),
                                                 1000L,
                                                 START_TIME_MS);
    final TriggerEvent event2 = makeTriggerEvent(triggerConfigs.mockGet(0).getMonitoringTriggerConfig(),
                                                 2000L,
                                                 START_TIME_MS + 10 * actionConfigs.mockGet(0).getActionConfig().getIntervalMs());

    KeyValue<Bytes, Iterable<KeyValue<Bytes, AlertInfo>>> ret = transformer.transform(EMPTY_BYTE_ARRAY, event1);
    verifyAlerts(ret, 1, 1, 1);

    // add one more action for the trigger
    MockReadOnlyKeyValueStore.mockAddActionConfig(actionConfigs,
                                                  makeActionConfig("action2", 1, 60000L, ImmutableList.of("trigger1")));

    ret = transformer.transform(EMPTY_BYTE_ARRAY, event2);
    // expect two actions
    verifyAlerts(ret, 1, 2, 1);
  }

  @Test
  public void testAddActionDifferentFrequenceAfterFirstTransform() {
    initTransformerWithOneTriggerOneAction();
    final TriggerEvent event1 = makeTriggerEvent(triggerConfigs.mockGet(0).getMonitoringTriggerConfig(),
                                                 1000L,
                                                 START_TIME_MS);
    final TriggerEvent event2 = makeTriggerEvent(triggerConfigs.mockGet(0).getMonitoringTriggerConfig(),
                                                 2000L,
                                                 START_TIME_MS + 10 * actionConfigs.mockGet(0).getActionConfig().getIntervalMs());

    KeyValue<Bytes, Iterable<KeyValue<Bytes, AlertInfo>>> ret = transformer.transform(EMPTY_BYTE_ARRAY, event1);
    verifyAlerts(ret, 1, 1, 1);

    // add one more action for the trigger
    MockReadOnlyKeyValueStore.mockAddActionConfig(actionConfigs,
                                                  makeActionConfig("action2", 2, 60000L, ImmutableList.of("trigger1")));

    ret = transformer.transform(EMPTY_BYTE_ARRAY, event2);
    // expect two actions, two alerts, two trigger events
    verifyAlerts(ret, 2, 2, 2);
  }

  @Test
  public void testInfiniteMaxSendRate() {
    initTransformerWithOneTrigger();

    MockReadOnlyKeyValueStore.mockAddActionConfig(actionConfigs,
                                                  makeActionConfig("action1", 0, 60000L, ImmutableList.of("trigger1")));

    final TriggerEvent event1 = makeTriggerEvent(triggerConfigs.mockGet(0).getMonitoringTriggerConfig(),
                                                 1000L,
                                                 START_TIME_MS);
    final TriggerEvent event2 = makeTriggerEvent(triggerConfigs.mockGet(0).getMonitoringTriggerConfig(),
                                                 2000L,
                                                 START_TIME_MS);

    KeyValue<Bytes, Iterable<KeyValue<Bytes, AlertInfo>>> ret = transformer.transform(EMPTY_BYTE_ARRAY, event1);
    verifyAlerts(ret, 1, 1, 1);

    ret = transformer.transform(EMPTY_BYTE_ARRAY, event2);
    verifyAlerts(ret, 1, 1, 1);
  }

  @Test
  public void testDisabledActionDoesNotSendAlert() {
    initTransformerWithOneTrigger();
    MockReadOnlyKeyValueStore.mockAddActionConfig(actionConfigs,
                                                  makeDisabledActionConfig("action1", 0, 60000L,
                                                                           ImmutableList.of("trigger1")));
    final TriggerEvent event = makeTriggerEvent(triggerConfigs.mockGet(0).getMonitoringTriggerConfig(),
                                                1000L,
                                                START_TIME_MS);
    assertNull(transformer.transform(EMPTY_BYTE_ARRAY, event));
  }

  @Test
  public void testDisableActionOnQueuedTriggerEvent() {
    initTransformerWithOneTrigger();

    MockReadOnlyKeyValueStore.mockAddActionConfig(actionConfigs,
                                                  makeActionConfig("action1", 1, 60000L, ImmutableList.of("trigger1")));

    final TriggerEvent event1 = makeTriggerEvent(triggerConfigs.mockGet(0).getMonitoringTriggerConfig(),
                                                 1000L,
                                                 START_TIME_MS);
    final TriggerEvent event2 = makeTriggerEvent(triggerConfigs.mockGet(0).getMonitoringTriggerConfig(),
                                                 2000L,
                                                 START_TIME_MS);
    final TriggerEvent event3 = makeTriggerEvent(triggerConfigs.mockGet(0).getMonitoringTriggerConfig(),
                                                 3000L,
                                                 START_TIME_MS + 10 * actionConfigs.mockGet(0).getActionConfig().getIntervalMs());
    final TriggerEvent hbEvent = TriggerEvent.newBuilder()
        .setHeartbeat(true)
        .setTimestamp(START_TIME_MS + 10 * actionConfigs.mockGet(0).getActionConfig().getIntervalMs())
        .build();

    KeyValue<Bytes, Iterable<KeyValue<Bytes, AlertInfo>>> ret = transformer.transform(EMPTY_BYTE_ARRAY, event1);
    verifyAlerts(ret, 1, 1, 1);

    // event2 will be queued
    assertNull(transformer.transform(EMPTY_BYTE_ARRAY, event2));

    // disable action
    MockReadOnlyKeyValueStore.mockReplaceActionConfig(actionConfigs, 0,
                                                      makeDisabledActionConfig("action1", 0, 60000L,
                                                                               ImmutableList
                                                                                   .of("trigger1")));

    // expect that we drop queued trigger events when action/trigger disabled/removed
    assertNull(transformer.transform(EMPTY_BYTE_ARRAY, hbEvent));
    List<VerifiableTriggerInfo> triggerEvents = getStoredTriggerEvents(kvTriggerEventsStore, event2);
    assertTrue(triggerEvents.isEmpty());

    // event3 should be ignored
    assertNull(transformer.transform(EMPTY_BYTE_ARRAY, event3));
    triggerEvents = getStoredTriggerEvents(kvTriggerEventsStore, event3);
    assertTrue(triggerEvents.isEmpty());
  }

  @Test
  public void testPunctuateUsesUpdatedTriggerActionConfigs() {
    initTransformerWithOneTriggerOneAction();
    final TriggerEvent event1 = makeTriggerEvent(triggerConfigs.mockGet(0).getMonitoringTriggerConfig(),
                                                 1000L,
                                                 START_TIME_MS);
    final TriggerEvent event2 = makeTriggerEvent(triggerConfigs.mockGet(0).getMonitoringTriggerConfig(),
                                                 2000L,
                                                 START_TIME_MS);
    final TriggerEvent hbEvent = TriggerEvent.newBuilder()
        .setHeartbeat(true)
        .setTimestamp(START_TIME_MS + 10 * actionConfigs.mockGet(0).getActionConfig().getIntervalMs())
        .build();


    makeTriggerEvent(triggerConfigs.mockGet(0).getMonitoringTriggerConfig(),
                                                 3000L,
                                                 START_TIME_MS + 10 * actionConfigs.mockGet(0).getActionConfig().getIntervalMs());

    KeyValue<Bytes, Iterable<KeyValue<Bytes, AlertInfo>>> ret = transformer.transform(EMPTY_BYTE_ARRAY, event1);
    verifyAlerts(ret, 1, 1, 1);

    // next event will be queued
    assertNull(transformer.transform(EMPTY_BYTE_ARRAY, event2));

    // add one more action for the trigger
    MockReadOnlyKeyValueStore.mockAddActionConfig(actionConfigs,
                                                  makeActionConfig("action2", 2, 60000L, ImmutableList.of("trigger1")));

    // should create alert on heartbeat event
    ret = transformer.transform(EMPTY_BYTE_ARRAY, hbEvent);
    // expect two actions, two alerts, two trigger events
    verifyAlerts(ret, 2, 2, 2);
  }

  @Test
  public void testTransformerCombinesActionsWithSameFrequency() {
    initTransformerWithOneTriggerMultipleActions();

    final TriggerEvent event1 = makeTriggerEvent(triggerConfigs.mockGet(0).getMonitoringTriggerConfig(), 1000L, START_TIME_MS);

    KeyValue<Bytes, Iterable<KeyValue<Bytes, AlertInfo>>> ret = transformer.transform(EMPTY_BYTE_ARRAY, event1);
    assertNotNull(ret);
    Iterator<KeyValue<Bytes, AlertInfo>> it = ret.value.iterator();
    // expect two alerts
    assertTrue(it.hasNext());
    it.next();
    assertTrue(it.hasNext());
    it.next();
    assertFalse(it.hasNext());

    // all trigger events list in trigger actions must be empty
    List<VerifiableTriggerInfo> triggerEvents = getStoredTriggerEvents(kvTriggerEventsStore, event1);
    assertTrue(triggerEvents.isEmpty());
  }

  @Test
  public void testSameTriggerEventInDifferentAlerts() {
    initTransformerWithOneTriggerMultipleActions();

    long currentTime = START_TIME_MS;
    long window = 0;

    // first transform will send alerts for all actions
    assertNotNull(transformer.transform(EMPTY_BYTE_ARRAY, makeTriggerEvent(triggerConfigs.mockGet(0).getMonitoringTriggerConfig(),
                                                                           window, currentTime)));

    // next 5 transforms should only create alert for action3 (6 alerts per minute)
    for (int i = 0; i < 5; i++) {
      currentTime += 10000L;
      window += TimeBucket.SIZE;

      TriggerEvent event = makeTriggerEvent(triggerConfigs.mockGet(0).getMonitoringTriggerConfig(), window, currentTime);
      KeyValue<Bytes, Iterable<KeyValue<Bytes, AlertInfo>>> ret = transformer.transform(EMPTY_BYTE_ARRAY, event);
      assertNotNull("Transform is expected to return at least one alert", ret);
      // expect 1 alert
      Iterator<KeyValue<Bytes, AlertInfo>> it = ret.value.iterator();
      assertTrue(it.hasNext());
      AlertInfo alert = it.next().value;
      assertEquals(1, alert.getActionsCount());
      assertEquals("action3", alert.getActions(0).getEmail().getSubject());
      assertEquals(1, alert.getTriggersCount());
      assertFalse("Transform is expected to return exactly one alert", it.hasNext());

      // must queue i + 1 trigger events
      List<VerifiableTriggerInfo> triggerEvents = getStoredTriggerEvents(kvTriggerEventsStore, event);
      assertEquals(i + 1, triggerEvents.size());
    }

    // next transform should return two alerts and all actions
    currentTime += 10000L;
    window += TimeBucket.SIZE;
    TriggerEvent event = makeTriggerEvent(triggerConfigs.mockGet(0).getMonitoringTriggerConfig(), window, currentTime);
    KeyValue<Bytes, Iterable<KeyValue<Bytes, AlertInfo>>> ret =
        transformer.transform(EMPTY_BYTE_ARRAY, makeTriggerEvent(triggerConfigs.mockGet(0).getMonitoringTriggerConfig(), window, currentTime));
    verifyAlerts(ret, 2, 3, 7);

    // all trigger events list in trigger actions must be empty
    List<VerifiableTriggerInfo> triggerEvents = getStoredTriggerEvents(kvTriggerEventsStore, event);
    assertTrue(triggerEvents.isEmpty());
  }

  @Test
  public void testSendHigherFrequencyWhenReachedMaxTriggerEvents() {
    initTransformerWithOneTriggerMultipleActions();

    long currentTime = START_TIME_MS;
    long window = 0;

    // first transform will send alerts for all actions
    assertNotNull(transformer.transform(EMPTY_BYTE_ARRAY, makeTriggerEvent(triggerConfigs.mockGet(0).getMonitoringTriggerConfig(), window, START_TIME_MS)));

    // next 8 transforms should only create alert (every other time) for action3 (6 alerts per
    // minute)
    for (int i = 0; i < 9; i++) {
      currentTime += 5000L;
      window += TimeBucket.SIZE;

      TriggerEvent event = makeTriggerEvent(triggerConfigs.mockGet(0).getMonitoringTriggerConfig(), window, currentTime);
      KeyValue<Bytes, Iterable<KeyValue<Bytes, AlertInfo>>> ret = transformer.transform(EMPTY_BYTE_ARRAY, event);
      if (i % 2 == 1) {
        verifyAlerts(ret, 1, 1, 2);
      }

      // must queue i + 1 trigger events
      List<VerifiableTriggerInfo> triggerEvents = getStoredTriggerEvents(kvTriggerEventsStore, event);
      assertEquals(i + 1, triggerEvents.size());
    }

    // next transform should return one alert for action that did not timeout but reached max
    TriggerEvent event = makeTriggerEvent(triggerConfigs.mockGet(0).getMonitoringTriggerConfig(),
                                          window, currentTime);
    KeyValue<Bytes, Iterable<KeyValue<Bytes, AlertInfo>>> ret =
        transformer.transform(EMPTY_BYTE_ARRAY, makeTriggerEvent(triggerConfigs.mockGet(0).getMonitoringTriggerConfig(), window, currentTime));
    verifyAlerts(ret, 1, 2, 10);

    // we should have 1 trigger event in the queue for action3
    List<VerifiableTriggerInfo> triggerEvents = getStoredTriggerEvents(kvTriggerEventsStore, event);
    assertEquals(1, triggerEvents.size());

    // then after we reach timeout for the action with long period, we should send another alert
    currentTime += 15000L;
    window += TimeBucket.SIZE;
    KeyValue<Bytes, Iterable<KeyValue<Bytes, AlertInfo>>> ret2 =
        transformer.transform(EMPTY_BYTE_ARRAY, makeTriggerEvent(triggerConfigs.mockGet(0).getMonitoringTriggerConfig(), window, currentTime));
    verifyAlerts(ret2, 2, 3, 3);
  }

  @Test
  public void testSameTriggerMultipleActions() {
    initTransformerWithMultipleIntersectingTriggersActions();

    final TriggerEvent event1 = makeTriggerEvent(triggerConfigs.mockGet(1).getMonitoringTriggerConfig(), 1000L, START_TIME_MS);
    KeyValue<Bytes, Iterable<KeyValue<Bytes, AlertInfo>>> ret = transformer.transform(EMPTY_BYTE_ARRAY, event1);
    assertNotNull(ret);
    Iterator<KeyValue<Bytes, AlertInfo>> it = ret.value.iterator();
    // expect two alerts
    assertTrue(it.hasNext());
    assertEquals(1, it.next().value.getActionsCount());
    assertTrue(it.hasNext());
    assertEquals(1, it.next().value.getActionsCount());
    assertFalse(it.hasNext());

    // all trigger events list in trigger actions must be empty
    List<VerifiableTriggerInfo> triggerEvents = getStoredTriggerEvents(kvTriggerEventsStore, event1);
    assertTrue(triggerEvents.isEmpty());
  }

  @Test
  public void testMultipleActionsMultipleTriggers() {
    initTransformerWithMultipleNonIntersectingTriggersActions();

    final TriggerEvent event1 = makeTriggerEvent(triggerConfigs.mockGet(0).getMonitoringTriggerConfig(), 1000L, START_TIME_MS);
    final TriggerEvent event2 = makeTriggerEvent(triggerConfigs.mockGet(1).getMonitoringTriggerConfig(), 2000L, START_TIME_MS);
    final TriggerEvent event3 = makeTriggerEvent(triggerConfigs.mockGet(2).getMonitoringTriggerConfig(), 3000L, START_TIME_MS);
    final TriggerEvent event4 = makeTriggerEvent(triggerConfigs.mockGet(1).getMonitoringTriggerConfig(),
                                                 3000L, START_TIME_MS + 10 * actionConfigs.mockGet(0).getActionConfig().getIntervalMs());

    // each even should trigger one alert, with one action and one trigger
    KeyValue<Bytes, Iterable<KeyValue<Bytes, AlertInfo>>> ret = transformer.transform(EMPTY_BYTE_ARRAY, event1);
    verifyAlerts(ret, 1, 1, 1);

    ret = transformer.transform(EMPTY_BYTE_ARRAY, event2);
    verifyAlerts(ret, 1, 1, 1);

    ret = transformer.transform(EMPTY_BYTE_ARRAY, event3);
    verifyAlerts(ret, 1, 1, 1);

    ret = transformer.transform(EMPTY_BYTE_ARRAY, event4);
    verifyAlerts(ret, 1, 1, 1);
  }

  @Test
  public void testPunctuateCreatesAlerts() {
    initTransformerWithOneTriggerOneAction();

    final TriggerEvent event1 = makeTriggerEvent(triggerConfigs.mockGet(0).getMonitoringTriggerConfig(), 1000L, START_TIME_MS);
    final TriggerEvent event2 = makeTriggerEvent(triggerConfigs.mockGet(0).getMonitoringTriggerConfig(), 2000L, START_TIME_MS);
    // first transform creates a alert
    assertNotNull(transformer.transform(EMPTY_BYTE_ARRAY, event1));
    // second transform should queue the event till timeout
    assertNull(transformer.transform(EMPTY_BYTE_ARRAY, event2));

    mockClock.setCurrentTimeMillis(START_TIME_MS + 60000L + 10);
    KeyValue<Bytes, Iterable<KeyValue<Bytes, AlertInfo>>> ret =
        transformer.punctuate(START_TIME_MS + 60000L + 10);
    // expect one alert
    assertNotNull(ret);
    Iterator<KeyValue<Bytes, AlertInfo>> it = ret.value.iterator();
    assertTrue(it.hasNext());
    AlertInfo alert = it.next().value;
    assertEquals("Expect one action in alert", 1, alert.getActionsCount());
    assertEquals("action1", alert.getActions(0).getEmail().getSubject());
    assertFalse("Punctuate is expected to return exactly one alert", it.hasNext());
  }

  @Test
  public void testPunctuateDoesNotCreateAlertBeforeTimeout() {
    initTransformerWithOneTriggerOneAction();

    // first transform creates a alert
    assertNotNull(transformer.transform(EMPTY_BYTE_ARRAY, makeTriggerEvent(triggerConfigs.mockGet(0).getMonitoringTriggerConfig(), 1000L, START_TIME_MS)));
    // second transform should queue the event till timeout
    assertNull(transformer.transform(EMPTY_BYTE_ARRAY, makeTriggerEvent(triggerConfigs.mockGet(0).getMonitoringTriggerConfig(), 2000L, START_TIME_MS)));

    KeyValue<Bytes, Iterable<KeyValue<Bytes, AlertInfo>>> ret =
        transformer.punctuate(START_TIME_MS);
    assertNull(ret);
  }

  @Test
  public void testTransformPunctuates() {
    initTransformerWithMultipleIntersectingTriggersActions();

    // first transform for all triggers
    assertNotNull(transformer.transform(EMPTY_BYTE_ARRAY, makeTriggerEvent(triggerConfigs.mockGet(0).getMonitoringTriggerConfig(), 1000L, START_TIME_MS)));
    assertNotNull(transformer.transform(EMPTY_BYTE_ARRAY, makeTriggerEvent(triggerConfigs.mockGet(1).getMonitoringTriggerConfig(), 1000L, START_TIME_MS)));
    assertNotNull(transformer.transform(EMPTY_BYTE_ARRAY, makeTriggerEvent(triggerConfigs.mockGet(2).getMonitoringTriggerConfig(), 1000L, START_TIME_MS)));

    // queue triggers for each action
    assertNull(transformer.transform(EMPTY_BYTE_ARRAY, makeTriggerEvent(triggerConfigs.mockGet(0).getMonitoringTriggerConfig(), 2000L, START_TIME_MS)));
    assertNull(transformer.transform(EMPTY_BYTE_ARRAY, makeTriggerEvent(triggerConfigs.mockGet(1).getMonitoringTriggerConfig(), 2000L, START_TIME_MS)));
    assertNull(transformer.transform(EMPTY_BYTE_ARRAY, makeTriggerEvent(triggerConfigs.mockGet(2).getMonitoringTriggerConfig(), 2000L, START_TIME_MS)));

    // send event for one of the actions, but all actions should create alerts
    KeyValue<Bytes, Iterable<KeyValue<Bytes, AlertInfo>>> ret =
        transformer.transform(EMPTY_BYTE_ARRAY, makeTriggerEvent(triggerConfigs.mockGet(0).getMonitoringTriggerConfig(),
                                                                 3000L, START_TIME_MS + 60000L + 10));
    // expect two alerts and 4 actions and 5 trigger events total (one event in two actions)
    verifyAlerts(ret, 4, 4, 5);
  }

  @Test
  public void testDoNotCreateAlertsForUnknownTriggerReferredInAction() {
    initTransformerWithActionRefersUnknownTriggerGuid();

    assertNull(transformer.transform(EMPTY_BYTE_ARRAY,
                                     makeTriggerEvent(makeTriggerConfig(DEFAULT_CLUSTER_ID,
                                                                        DEFAULT_CONSUMER_GROUP,
                                                                        "trigger2",
                                                                        TriggerMetricType.CONSUMPTION_DIFF,
                                                                        TriggerCondition.GREATER_THAN,
                                                                        5), 1000L, START_TIME_MS)));
  }

  @Test
  public void testTriggerEventsWaitTillLatestAction() {
    initTransformerWithOneTriggerMultipleActionsDifferentFrequency();

    long currentTime = START_TIME_MS;
    long window = 1000L;

    // first transform always creates alerts
    assertNotNull(transformer.transform(EMPTY_BYTE_ARRAY, makeTriggerEvent(triggerConfigs.mockGet(0).getMonitoringTriggerConfig(), window, currentTime)));

    // second transform will queue the event
    window += 1000L;
    assertNull(transformer.transform(EMPTY_BYTE_ARRAY, makeTriggerEvent(triggerConfigs.mockGet(0).getMonitoringTriggerConfig(), window, currentTime)));

    for (int i = 0; i < 5; i++) {
      currentTime += 10000L;
      window += 1000L;
      KeyValue<Bytes, Iterable<KeyValue<Bytes, AlertInfo>>> ret =
          transformer.transform(EMPTY_BYTE_ARRAY, makeTriggerEvent(triggerConfigs.mockGet(0).getMonitoringTriggerConfig(), window, currentTime));
      if (i == 0) {
        verifyAlerts(ret, 1, 1, 2);   // one more event before the for loop
      } else if (i == 2) {
        verifyAlerts(ret, 2, 2, 5);
      } else {
        verifyAlerts(ret, 1, 1, 1);
      }
    }

    currentTime += 10000L;
    window += 1000L;

    KeyValue<Bytes, Iterable<KeyValue<Bytes, AlertInfo>>> ret =
        transformer.transform(EMPTY_BYTE_ARRAY, makeTriggerEvent(triggerConfigs.mockGet(0).getMonitoringTriggerConfig(), window, currentTime));
    // expected number of trigger events: 1 for action with 10 sec timeout; 2 + 1 with 30 sec
    // timeout, and 7 (all trigger events) for action with 60 sec timeout.
    verifyAlerts(ret, 3, 3, 11);
  }

  @Test
  public void testAlertsHaveUniqueIds() {
    initTransformerWithOneTriggerOneAction();
    Set<String> guids = new HashSet<>();
    long currentTime = START_TIME_MS;

    for (int i = 0; i < 10; i++) {
      KeyValue<Bytes, Iterable<KeyValue<Bytes, AlertInfo>>>
          ret =
          transformer.transform(EMPTY_BYTE_ARRAY, makeTriggerEvent(triggerConfigs.mockGet(0).getMonitoringTriggerConfig(), (i + 1) * 1000L, currentTime));
      verifyAlerts(ret, 1, 1, 1);
      assertFalse("Alert guid is expected to be unique",
                  guids.contains(ret.value.iterator().next().value.getGuid()));
      guids.add(ret.value.iterator().next().value.getGuid());
      currentTime = START_TIME_MS + (i + 1) * 60000L;
    }
  }

  private void initTransformer() {
    transformer = supplier.get();
    transformer.init(context);
  }

  private void initTransformerWithOneTrigger() {
    ImmutableList<String> triggers = ImmutableList.of("trigger1");
    MockReadOnlyKeyValueStore.mockAddMonitoringTriggerConfig(
        triggerConfigs,
        makeTriggerConfig(
            DEFAULT_CLUSTER_ID,
            DEFAULT_CONSUMER_GROUP,
            triggers.get(0),
            TriggerMetricType.CONSUMPTION_DIFF,
            TriggerCondition.NOT_EQUAL,
            0
        )
    );
    initTransformer();
  }

  private void initTransformerWithOneTriggerOneAction() {
    ImmutableList<String> triggers = ImmutableList.of("trigger1");
    MockReadOnlyKeyValueStore.mockAddMonitoringTriggerConfig(
        triggerConfigs,
        makeTriggerConfig(
            DEFAULT_CLUSTER_ID,
            DEFAULT_CONSUMER_GROUP,
            triggers.get(0),
            TriggerMetricType.CONSUMPTION_DIFF,
            TriggerCondition.NOT_EQUAL,
            0
        )
    );
    MockReadOnlyKeyValueStore.mockAddActionConfig(actionConfigs, makeActionConfig("action1", 1, 60000L, triggers));
    initTransformer();
  }

  // 3 actions, each has same trigger. Two actions have same frequency.
  private void initTransformerWithOneTriggerMultipleActions() {
    ImmutableList<String> triggers = ImmutableList.of("trigger1");
    MockReadOnlyKeyValueStore.mockAddMonitoringTriggerConfig(
        triggerConfigs,
        makeTriggerConfig(
            DEFAULT_CLUSTER_ID,
            DEFAULT_CONSUMER_GROUP,
            triggers.get(0),
            TriggerMetricType.CONSUMPTION_DIFF,
            TriggerCondition.NOT_EQUAL,
            0
        )
    );
    MockReadOnlyKeyValueStore.mockAddActionConfig(actionConfigs, makeActionConfig("action1", 1, 60000L, triggers));
    MockReadOnlyKeyValueStore.mockAddActionConfig(actionConfigs, makeActionConfig("action2", 1, 60000L, triggers));
    MockReadOnlyKeyValueStore.mockAddActionConfig(actionConfigs, makeActionConfig("action3", 6, 60000L, triggers));
    initTransformer();
  }

  // 2 actions, each has two triggers, one trigger is common
  private void initTransformerWithMultipleIntersectingTriggersActions() {
    ImmutableList<String> triggersForAction1 = ImmutableList.of("trigger1", "trigger2");
    ImmutableList<String> triggersForAction2 = ImmutableList.of("trigger2", "trigger3");
    MockReadOnlyKeyValueStore.mockAddMonitoringTriggerConfig(
        triggerConfigs,
        makeTriggerConfig(
            DEFAULT_CLUSTER_ID,
            DEFAULT_CONSUMER_GROUP,
            "trigger1",
            TriggerMetricType.CONSUMPTION_DIFF,
            TriggerCondition.NOT_EQUAL,
            0
        )
    );
    MockReadOnlyKeyValueStore.mockAddMonitoringTriggerConfig(
        triggerConfigs,
        makeTriggerConfig(
            DEFAULT_CLUSTER_ID,
            DEFAULT_CONSUMER_GROUP,
            "trigger2",
            TriggerMetricType.CONSUMPTION_DIFF,
            TriggerCondition.GREATER_THAN,
            5
    ));
    MockReadOnlyKeyValueStore.mockAddMonitoringTriggerConfig(
        triggerConfigs,
        makeTriggerConfig(
            DEFAULT_CLUSTER_ID,
            DEFAULT_CONSUMER_GROUP,
            "trigger3",
            TriggerMetricType.CONSUMPTION_DIFF,
            TriggerCondition.LESS_THAN,
            10
      )
    );
    MockReadOnlyKeyValueStore.mockAddActionConfig(actionConfigs, makeActionConfig("action1", 1, 60000L, triggersForAction1));
    MockReadOnlyKeyValueStore.mockAddActionConfig(actionConfigs, makeActionConfig("action2", 2, 60000L, triggersForAction2));
    initTransformer();
  }

  // 2 actions, no common triggers
  private void initTransformerWithMultipleNonIntersectingTriggersActions() {
    ImmutableList<String> triggersForAction1 = ImmutableList.of("trigger1", "trigger2");
    ImmutableList<String> triggersForAction2 = ImmutableList.of("trigger3");
    MockReadOnlyKeyValueStore.mockAddMonitoringTriggerConfig(
        triggerConfigs,
        makeTriggerConfig(
            DEFAULT_CLUSTER_ID,
            DEFAULT_CONSUMER_GROUP,
            "trigger1",
            TriggerMetricType.CONSUMPTION_DIFF,
            TriggerCondition.NOT_EQUAL,
            0
        )
    );
    MockReadOnlyKeyValueStore.mockAddMonitoringTriggerConfig(
        triggerConfigs,
        makeTriggerConfig(
            DEFAULT_CLUSTER_ID,
            DEFAULT_CONSUMER_GROUP,
            "trigger2",
            TriggerMetricType.CONSUMPTION_DIFF,
            TriggerCondition.GREATER_THAN,
            5
        ));
    MockReadOnlyKeyValueStore.mockAddMonitoringTriggerConfig(
        triggerConfigs,
        makeTriggerConfig(
            DEFAULT_CLUSTER_ID,
            DEFAULT_CONSUMER_GROUP,
            "trigger3",
            TriggerMetricType.CONSUMPTION_DIFF,
            TriggerCondition.LESS_THAN,
            10
        )
    );
    MockReadOnlyKeyValueStore.mockAddActionConfig(actionConfigs, makeActionConfig("action1", 1, 60000L, triggersForAction1));
    MockReadOnlyKeyValueStore.mockAddActionConfig(actionConfigs, makeActionConfig("action2", 1, 60000L, triggersForAction2));
    initTransformer();
  }


  // 3 actions, each has same trigger and different frequency
  private void initTransformerWithOneTriggerMultipleActionsDifferentFrequency() {
    ImmutableList<String> triggers = ImmutableList.of("trigger1");
    MockReadOnlyKeyValueStore.mockAddMonitoringTriggerConfig(
        triggerConfigs,
        makeTriggerConfig(
            DEFAULT_CLUSTER_ID,
            DEFAULT_CONSUMER_GROUP,
            triggers.get(0),
            TriggerMetricType.CONSUMPTION_DIFF,
            TriggerCondition.NOT_EQUAL,
            0
        )
    );
    MockReadOnlyKeyValueStore.mockAddActionConfig(actionConfigs, makeActionConfig("action1", 1, 60000L, triggers));
    MockReadOnlyKeyValueStore.mockAddActionConfig(actionConfigs, makeActionConfig("action2", 2, 60000L, triggers));
    MockReadOnlyKeyValueStore.mockAddActionConfig(actionConfigs, makeActionConfig("action3", 6, 60000L, triggers));
    initTransformer();
  }

  // 1 action, one trigger, but action refers to trigger that is not defined
  private void initTransformerWithActionRefersUnknownTriggerGuid() {
    ImmutableList<String> triggersForAction = ImmutableList.of("trigger2");
    MockReadOnlyKeyValueStore.mockAddMonitoringTriggerConfig(
        triggerConfigs,
        makeTriggerConfig(
            DEFAULT_CLUSTER_ID,
            DEFAULT_CONSUMER_GROUP,
            "trigger1",
            TriggerMetricType.CONSUMPTION_DIFF,
            TriggerCondition.NOT_EQUAL,
            0
        )
    );
    MockReadOnlyKeyValueStore.mockAddActionConfig(actionConfigs, makeActionConfig("action1", 1, 60000L, triggersForAction));
    initTransformer();
  }

  private void verifyAlerts(KeyValue<Bytes, Iterable<KeyValue<Bytes, AlertInfo>>> alerts,
                            int expectedAlertCount,
                            int expectedTotalActionCount,
                            int expectedTotalTriggerEventCount) {
    int alertCount = 0;
    int actionsCount = 0;
    int triggerEventCount = 0;
    assertNotNull("Expect returned key-value for alerts not null", alerts);
    assertNotNull("Expected returned alert list not null", alerts.value);
    Iterator<KeyValue<Bytes, AlertInfo>> it = alerts.value.iterator();
    while (it.hasNext()) {
      alertCount++;
      AlertInfo alert = it.next().value;
      actionsCount += alert.getActionsCount();
      triggerEventCount += alert.getTriggersCount();
    }
    assertEquals("Expected number of alerts", expectedAlertCount, alertCount);
    assertEquals("Expected total action count in all alerts",
                 expectedTotalActionCount, actionsCount);
    assertEquals("Expected total trigger event count in all alerts",
                 expectedTotalTriggerEventCount, triggerEventCount);
  }

}
