package io.confluent.controlcenter.streams.alert;

import com.google.inject.Guice;
import com.google.inject.Injector;
import com.google.inject.Key;
import com.google.inject.util.Providers;

import org.apache.kafka.clients.producer.Producer;
import org.apache.kafka.clients.producer.ProducerRecord;
import org.apache.kafka.common.serialization.Serializer;
import org.apache.kafka.common.utils.Bytes;
import org.apache.kafka.streams.KeyValue;
import org.apache.kafka.streams.kstream.Transformer;
import org.apache.kafka.streams.processor.StateStoreSupplier;
import org.apache.kafka.streams.processor.internals.RecordCollector;
import org.apache.kafka.streams.processor.internals.RecordCollectorImpl;
import org.apache.kafka.streams.state.KeyValueStore;
import org.apache.kafka.streams.state.Stores;
import org.easymock.EasyMock;
import org.junit.Before;
import org.junit.Test;

import java.io.File;
import java.nio.file.Files;
import java.util.Iterator;

import io.confluent.command.record.Command.CommandKey;
import io.confluent.command.record.Command.CommandMessage;
import io.confluent.command.record.alert.CommandAlert.TriggerCondition;
import io.confluent.command.record.alert.CommandAlert.TriggerMetricType;
import io.confluent.controlcenter.record.Controlcenter.ClientTriggerMeasurement;
import io.confluent.controlcenter.record.Controlcenter.TriggerEvent;
import io.confluent.controlcenter.record.Controlcenter.TriggerMeasurement;
import io.confluent.controlcenter.record.Controlcenter.WindowedClusterGroup;
import io.confluent.controlcenter.serialization.SerializationModule;
import io.confluent.controlcenter.serialization.SerializationModule.MonitoringTriggerKeySerde;
import io.confluent.controlcenter.streams.MockClock;
import io.confluent.controlcenter.streams.MockProcessorContext;
import io.confluent.controlcenter.streams.MockReadOnlyKeyValueStore;
import io.confluent.monitoring.common.TimeBucket;
import io.confluent.serializers.OrderedKeyUberSerde;
import io.confluent.serializers.ProtoSerde;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertNull;
import static org.junit.Assert.assertTrue;

public class MonitoringTriggerTransformerSupplierTest {

  private static final String STORE_NAME = "MonitoringTriggerTransformerTestStore";
  private static final Long START_TIME_MS = 36000000L;
  private static final String DEFAULT_CLUSTER_ID = "cluster-0";
  private static final String DEFAULT_CONSUMER_GROUP = "cg-0";

  private final MockClock mockClock = new MockClock();
  private final Injector injector = Guice.createInjector(new SerializationModule());
  private final OrderedKeyUberSerde<WindowedClusterGroup> keySerde =
      injector.getInstance(new Key<OrderedKeyUberSerde<WindowedClusterGroup>>(MonitoringTriggerKeySerde.class) {});


  private MockProcessorContext context;
  private Producer<byte[], byte[]> mockProducer;
  private KeyValueStore<Bytes, TriggerMeasurement> kvStore;
  private File stateDir;

  private MonitoringTriggerTransformerSupplier supplier;
  private Transformer<WindowedClusterGroup,
      TriggerMeasurement,
      KeyValue<Bytes, Iterable<KeyValue<Bytes, TriggerEvent>>>> transformer;
  private MockReadOnlyKeyValueStore<CommandKey, CommandMessage> triggerConfigs;

  private static final ClientTriggerMeasurement makeClientTriggerMeasurement(Long timestamp,
                                                                             Long consumptionDiff,
                                                                             Long avgLatency,
                                                                             Long maxLatency,
                                                                             boolean hasError) {
    return ClientTriggerMeasurement.newBuilder()
        .setArrivalTime(timestamp)
        .setConsumptionDiff(consumptionDiff)
        .setAvgLatency(avgLatency)
        .setMaxLatency(maxLatency)
        .setHasError(hasError)
        .build();
  }

  private static final TriggerMeasurement makeTriggerMeasurement(Long timestamp,
                                                                 Long consumptionDiff,
                                                                 Long avgLatency,
                                                                 Long maxLatency,
                                                                 boolean hasError) {
    ClientTriggerMeasurement.Builder builder = ClientTriggerMeasurement.newBuilder()
        .setArrivalTime(timestamp)
        .setConsumptionDiff(consumptionDiff)
        .setAvgLatency(avgLatency)
        .setMaxLatency(maxLatency)
        .setHasError(hasError);

    return TriggerMeasurement.newBuilder()
        .setClusterId(DEFAULT_CLUSTER_ID)
        .setGroup(DEFAULT_CONSUMER_GROUP)
        .setWindow(-1)
        .setClientMeasurement(builder)
        .build();
  }

  private static final TriggerMeasurement makeHeartbeatTriggerMeasurement() {
    return TriggerMeasurement.newBuilder()
        .setHeartbeat(true)
        .build();
  }

  private static final TriggerMeasurement makeTriggerMeasurement(ClientTriggerMeasurement clientTriggerMeasurement) {
    return TriggerMeasurement.newBuilder()
        .setClusterId(DEFAULT_CLUSTER_ID)
        .setGroup(DEFAULT_CONSUMER_GROUP)
        .setWindow(-1)
        .setClientMeasurement(clientTriggerMeasurement)
        .build();
  }

  private static final WindowedClusterGroup makeKey(Long window) {
    return WindowedClusterGroup.newBuilder()
        .setClusterId(DEFAULT_CLUSTER_ID)
        .setGroup(DEFAULT_CONSUMER_GROUP)
        .setWindow(window)
        .build();
  }

  private static final WindowedClusterGroup makeKey(String clusterId, String cg, Long window) {
    return WindowedClusterGroup.newBuilder()
        .setClusterId(clusterId)
        .setGroup(cg)
        .setWindow(window)
        .build();
  }

  private TriggerMeasurement getStoredTriggerMeasurement(
      KeyValueStore<Bytes, TriggerMeasurement> store,
      WindowedClusterGroup key) {
    return store.get(keySerde.key(key));
  }

  @Before
  public void setUp() throws Exception {
    StateStoreSupplier storeSupplier = Stores.create(STORE_NAME)
        .withKeys(keySerde)
        .withValues(new ProtoSerde<>(TriggerMeasurement.getDefaultInstance()))
        .persistent()
        .build();

    // The code block is needed to be able to test persistent kv store backed by rocksdb
    stateDir = Files.createTempDirectory("test").toFile();
    stateDir.deleteOnExit();
    mockProducer = EasyMock.createMock(Producer.class);
    RecordCollector recordCollector = new RecordCollectorImpl(mockProducer, "fakeTaskId") {
      @Override
      public <K1, V1> void send(ProducerRecord<K1, V1> record, Serializer<K1> keySerializer,
                                Serializer<V1> valueSerializer) {
        // nothing
      }
    };
    context = new MockProcessorContext(stateDir, recordCollector);

    kvStore = (KeyValueStore<Bytes, TriggerMeasurement>) storeSupplier.get();
    kvStore.init(context, kvStore);
    context.register(kvStore, false, null);

    mockClock.setCurrentTimeMillis(START_TIME_MS);
    this.triggerConfigs = new MockReadOnlyKeyValueStore();

    supplier = new MonitoringTriggerTransformerSupplier(STORE_NAME, mockClock, keySerde,
                                                        Providers.of(triggerConfigs), 1);
    transformer = supplier.get();
    transformer.init(context);
  }

  @Test
  public void testTransformDoesNotStoreNonTriggerableMeasurement() {
    MockReadOnlyKeyValueStore.mockAddMonitoringTriggerConfig(
        triggerConfigs,
        CommandAlertUtils.makeTriggerConfig(
            DEFAULT_CLUSTER_ID,
            DEFAULT_CONSUMER_GROUP,
            "trigger1",
            TriggerMetricType.CONSUMPTION_DIFF,
            TriggerCondition.NOT_EQUAL,
            0L
        )
    );

    final WindowedClusterGroup key = makeKey(100L);
    TriggerMeasurement triggerMeasurement = makeTriggerMeasurement(START_TIME_MS + 10L,
                                                                   0L, 10L, 100L, false);
    assertNull(transformer.transform(key, triggerMeasurement));
    assertNull(getStoredTriggerMeasurement(kvStore, key));
  }

  @Test
  public void testConsumerGroupMustMatch() {
    MockReadOnlyKeyValueStore.mockAddMonitoringTriggerConfig(
        triggerConfigs,
        CommandAlertUtils.makeTriggerConfig(
            DEFAULT_CLUSTER_ID,
            "different-consumer-group",
            "trigger1",
            TriggerMetricType.CONSUMPTION_DIFF,
            TriggerCondition.NOT_EQUAL,
            0L
        )
    );

    final WindowedClusterGroup key = makeKey(100L);
    TriggerMeasurement triggerMeasurement = makeTriggerMeasurement(START_TIME_MS + 10L,
                                                                   1L, 10L, 100L, false);
    assertNull(transformer.transform(key, triggerMeasurement));
    assertNull(getStoredTriggerMeasurement(kvStore, key));
  }

  @Test
  public void testConsumerGroupSetCorrectlyFromTheKey() {
    MockReadOnlyKeyValueStore.mockAddMonitoringTriggerConfig(
        triggerConfigs,
        CommandAlertUtils.makeTriggerConfig(
            DEFAULT_CLUSTER_ID,
            "different-consumer-group",
            "trigger1",
            TriggerMetricType.CONSUMPTION_DIFF,
            TriggerCondition.NOT_EQUAL,
            0L
        )
    );

    final WindowedClusterGroup key = makeKey(DEFAULT_CLUSTER_ID, "different-consumer-group", 100L);
    TriggerMeasurement triggerMeasurement = makeTriggerMeasurement(START_TIME_MS + 10L,
                                                                   1L, 10L, 100L, false);
    assertNull(transformer.transform(key, triggerMeasurement));

    TriggerMeasurement storedTriggerMeasurement = getStoredTriggerMeasurement(kvStore, key);
    assertNotNull(storedTriggerMeasurement);
    assertEquals(key.getGroup(), storedTriggerMeasurement.getGroup());
  }

  @Test
  public void testClusterIdMustMatch() {
    MockReadOnlyKeyValueStore.mockAddMonitoringTriggerConfig(
        triggerConfigs,
        CommandAlertUtils.makeTriggerConfig(
            "different-cluster-id",
            DEFAULT_CONSUMER_GROUP,
            "trigger1",
            TriggerMetricType.CONSUMPTION_DIFF,
            TriggerCondition.NOT_EQUAL,
            0L
        )
    );

    final WindowedClusterGroup key = makeKey(100L);
    TriggerMeasurement triggerMeasurement = makeTriggerMeasurement(START_TIME_MS + 10L,
                                                                   1L, 10L, 100L, false);
    assertNull(transformer.transform(key, triggerMeasurement));
    assertNull(getStoredTriggerMeasurement(kvStore, key));
  }

  @Test
  public void testClusterIdSetCorrectlyFromTheKey() {
    MockReadOnlyKeyValueStore.mockAddMonitoringTriggerConfig(
        triggerConfigs,
        CommandAlertUtils.makeTriggerConfig(
            "different-cluster-id",
            DEFAULT_CONSUMER_GROUP,
            "trigger1",
            TriggerMetricType.CONSUMPTION_DIFF,
            TriggerCondition.NOT_EQUAL,
            0L
        )
    );

    final WindowedClusterGroup key = makeKey("different-cluster-id", DEFAULT_CONSUMER_GROUP, 100L);
    TriggerMeasurement triggerMeasurement = makeTriggerMeasurement(START_TIME_MS + 10L,
                                                                   1L, 10L, 100L, false);
    assertNull(transformer.transform(key, triggerMeasurement));

    TriggerMeasurement storedTriggerMeasurement = getStoredTriggerMeasurement(kvStore, key);
    assertNotNull(storedTriggerMeasurement);
    assertEquals(key.getClusterId(), storedTriggerMeasurement.getClusterId());
  }

  @Test
  public void testTransformStoresNotEqualConsumptionDiff() {
    final TriggerMetricType metricType = TriggerMetricType.CONSUMPTION_DIFF;
    final TriggerCondition condition = TriggerCondition.NOT_EQUAL;
    final Long triggerValue = 0L;
    final Long consumptionDiff = 1L;
    final Long avgLatency = 10L;
    final Long maxLatency = 100L;

    testTransformStoresTriggerableMeasurement(metricType, condition, triggerValue,
                                              consumptionDiff, avgLatency, maxLatency);
  }

  @Test
  public void testTransformStoresEqualConsumptionDiff() {
    final TriggerMetricType metricType = TriggerMetricType.CONSUMPTION_DIFF;
    final TriggerCondition condition = TriggerCondition.EQUAL;
    final Long triggerValue = 10L;
    final Long consumptionDiff = triggerValue;
    final Long avgLatency = 10L;
    final Long maxLatency = 100L;

    testTransformStoresTriggerableMeasurement(metricType, condition, triggerValue,
                                              consumptionDiff, avgLatency, maxLatency);
  }

  @Test
  public void testTransformStoresLessThanConsumptionDiff() {
    final TriggerMetricType metricType = TriggerMetricType.CONSUMPTION_DIFF;
    final TriggerCondition condition = TriggerCondition.LESS_THAN;
    final Long triggerValue = 0L;
    final Long consumptionDiff = triggerValue - 1L;
    final Long avgLatency = 10L;
    final Long maxLatency = 100L;

    testTransformStoresTriggerableMeasurement(metricType, condition, triggerValue,
                                              consumptionDiff, avgLatency, maxLatency);
  }

  @Test
  public void testTransformStoresGreaterThanConsumptionDiff() {
    final TriggerMetricType metricType = TriggerMetricType.CONSUMPTION_DIFF;
    final TriggerCondition condition = TriggerCondition.GREATER_THAN;
    final Long triggerValue = 0L;
    final Long consumptionDiff = triggerValue + 1L;
    final Long avgLatency = 10L;
    final Long maxLatency = 100L;

    testTransformStoresTriggerableMeasurement(metricType, condition, triggerValue,
                                              consumptionDiff, avgLatency, maxLatency);
  }

  @Test
  public void testTransformStoresNotEqualAvgLatency() {
    final TriggerMetricType metricType = TriggerMetricType.AVG_LATENCY;
    final TriggerCondition condition = TriggerCondition.NOT_EQUAL;
    final Long triggerValue = 10L;
    final Long consumptionDiff = 1L;
    final Long avgLatency = 11L;
    final Long maxLatency = 100L;

    testTransformStoresTriggerableMeasurement(metricType, condition, triggerValue,
                                              consumptionDiff, avgLatency, maxLatency);
  }

  @Test
  public void testTransformStoresEqualAvgLatency() {
    final TriggerMetricType metricType = TriggerMetricType.AVG_LATENCY;
    final TriggerCondition condition = TriggerCondition.EQUAL;
    final Long triggerValue = 10L;
    final Long consumptionDiff = 1L;
    final Long avgLatency = triggerValue;
    final Long maxLatency = 100L;

    testTransformStoresTriggerableMeasurement(metricType, condition, triggerValue,
                                              consumptionDiff, avgLatency, maxLatency);
  }

  @Test
  public void testTransformStoresLessThanAvgLatency() {
    final TriggerMetricType metricType = TriggerMetricType.AVG_LATENCY;
    final TriggerCondition condition = TriggerCondition.LESS_THAN;
    final Long triggerValue = 10L;
    final Long consumptionDiff = 1L;
    final Long avgLatency = triggerValue - 5L;
    final Long maxLatency = 100L;

    testTransformStoresTriggerableMeasurement(metricType, condition, triggerValue,
                                              consumptionDiff, avgLatency, maxLatency);
  }

  @Test
  public void testTransformStoresGreaterThanAvgLatency() {
    final TriggerMetricType metricType = TriggerMetricType.AVG_LATENCY;
    final TriggerCondition condition = TriggerCondition.GREATER_THAN;
    final Long triggerValue = 10L;
    final Long consumptionDiff = 1L;
    final Long avgLatency = triggerValue + 5L;
    final Long maxLatency = 100L;

    testTransformStoresTriggerableMeasurement(metricType, condition, triggerValue,
                                              consumptionDiff, avgLatency, maxLatency);
  }

  @Test
  public void testTransformStoresNotEqualMaxLatency() {
    final TriggerMetricType metricType = TriggerMetricType.MAX_LATENCY;
    final TriggerCondition condition = TriggerCondition.NOT_EQUAL;
    final Long triggerValue = 0L;
    final Long consumptionDiff = 1L;
    final Long avgLatency = 10L;
    final Long maxLatency = 100L;

    testTransformStoresTriggerableMeasurement(metricType, condition, triggerValue,
                                              consumptionDiff, avgLatency, maxLatency);
  }

  @Test
  public void testTransformStoresEqualMaxLatency() {
    final TriggerMetricType metricType = TriggerMetricType.MAX_LATENCY;
    final TriggerCondition condition = TriggerCondition.EQUAL;
    final Long triggerValue = 0L;
    final Long consumptionDiff = 1L;
    final Long avgLatency = 10L;

    testTransformStoresTriggerableMeasurement(metricType, condition, triggerValue,
                                              consumptionDiff, avgLatency, triggerValue);
  }

  @Test
  public void testTransformStoresLessThanMaxLatency() {
    final TriggerMetricType metricType = TriggerMetricType.MAX_LATENCY;
    final TriggerCondition condition = TriggerCondition.LESS_THAN;
    final Long triggerValue = 10L;
    final Long consumptionDiff = 1L;
    final Long avgLatency = 10L;
    final Long maxLatency = triggerValue - 5L;

    testTransformStoresTriggerableMeasurement(metricType, condition, triggerValue,
                                              consumptionDiff, avgLatency, maxLatency);
  }

  @Test
  public void testTransformStoresGreaterThanMaxLatency() {
    final TriggerMetricType metricType = TriggerMetricType.MAX_LATENCY;
    final TriggerCondition condition = TriggerCondition.GREATER_THAN;
    final Long triggerValue = 10L;
    final Long consumptionDiff = 1L;
    final Long avgLatency = 10L;
    final Long maxLatency = triggerValue + 10L;

    testTransformStoresTriggerableMeasurement(metricType, condition, triggerValue,
                                              consumptionDiff, avgLatency, maxLatency);
  }

  @Test
  public void testTransformDoesNotIgnoreErrorMeasurements() {
    MockReadOnlyKeyValueStore.mockAddMonitoringTriggerConfig(
        triggerConfigs,
        CommandAlertUtils.makeTriggerConfig(
            DEFAULT_CLUSTER_ID,
            DEFAULT_CONSUMER_GROUP,
            "trigger1",
            TriggerMetricType.MAX_LATENCY,
            TriggerCondition.NOT_EQUAL,
            19000L
        )
    );
    final WindowedClusterGroup key = makeKey(100L);
    Long measurementTimestamp = 10L;
    TriggerMeasurement triggerMeasurement = makeTriggerMeasurement(measurementTimestamp, 1L, 10L, 9999L, true);
    assertNull(transformer.transform(key, triggerMeasurement));

    Long punctuateInterval = TimeBucket.SIZE;
    TriggerMeasurement tm2 = makeTriggerMeasurement(measurementTimestamp + CommandAlertUtils.DEFAULT_TRIGGER_LAG_MS + punctuateInterval,
                                                    1L, 10L, 100L, false);
    KeyValue<Bytes, Iterable<KeyValue<Bytes, TriggerEvent>>> ret = transformer.transform(makeKey(300L), tm2);
    assertNotNull(ret);
    TriggerEvent triggerEvent = ret.value.iterator().next().value;
    assertEquals("trigger1", triggerEvent.getTriggerGuid());
    assertEquals(9999L, triggerEvent.getInfo().getLongValue());
    assertNull(getStoredTriggerMeasurement(kvStore, key));
  }

  @Test
  public void testPunctuateCreatesTriggerEventAndRemovesMeasurementFromStore() {
    MockReadOnlyKeyValueStore.mockAddMonitoringTriggerConfig(
        triggerConfigs,
        CommandAlertUtils.makeTriggerConfig(
            DEFAULT_CLUSTER_ID,
            DEFAULT_CONSUMER_GROUP,
            "trigger1",
            TriggerMetricType.CONSUMPTION_DIFF,
            TriggerCondition.NOT_EQUAL,
            0L
        )
    );

    final WindowedClusterGroup key = makeKey(100L);
    Long measurementTimestamp = 10L;  // in the past
    TriggerMeasurement triggerMeasurement = makeTriggerMeasurement(measurementTimestamp,
                                                                   1L, 10L, 100L, false);
    assertNull(transformer.transform(key, triggerMeasurement));

    // triggers are fired based on timestamp we pass to punctuate
    TriggerMeasurement tm2 = makeTriggerMeasurement(measurementTimestamp + CommandAlertUtils.DEFAULT_TRIGGER_LAG_MS,
                                                    1L, 10L, 100L, false);
    KeyValue<Bytes, Iterable<KeyValue<Bytes, TriggerEvent>>> ret = transformer.transform(makeKey(200L), tm2);
    assertNotNull(ret);
    TriggerEvent triggerEvent = ret.value.iterator().next().value;
    assertEquals("trigger1", triggerEvent.getTriggerGuid());
    assertEquals(1L, triggerEvent.getInfo().getLongValue());
    assertNull(getStoredTriggerMeasurement(kvStore, key));
  }

  @Test
  public void testTriggersAreFiredOnMeasurementTimestamp() {
    MockReadOnlyKeyValueStore.mockAddMonitoringTriggerConfig(
        triggerConfigs,
        CommandAlertUtils.makeTriggerConfig(
            DEFAULT_CLUSTER_ID,
            DEFAULT_CONSUMER_GROUP,
            "trigger1",
            TriggerMetricType.CONSUMPTION_DIFF,
            TriggerCondition.NOT_EQUAL,
            0L
        )
    );

    final WindowedClusterGroup key = makeKey(100L);
    // in the future
    Long measurementTimestamp = START_TIME_MS + 5 * CommandAlertUtils.DEFAULT_TRIGGER_LAG_MS;
    TriggerMeasurement triggerMeasurement = makeTriggerMeasurement(measurementTimestamp,
                                                                   1L, 10L, 100L, false);
    assertNull(transformer.transform(key, triggerMeasurement));

    // move current time way ahead
    mockClock.setCurrentTimeMillis(measurementTimestamp + 5 * CommandAlertUtils.DEFAULT_TRIGGER_LAG_MS);

    // since we move arrival time forward, will not fire trigger, since C3 clock did not move
    // past lag
    TriggerMeasurement tm2 = makeTriggerMeasurement(measurementTimestamp + CommandAlertUtils.DEFAULT_TRIGGER_LAG_MS - 10L,
                                                    1L, 10L, 100L, false);
    assertNull(transformer.transform(makeKey(200L), tm2));

    // but should fire trigger once C3 time moves
    Long punctuateInterval = TimeBucket.SIZE;
    TriggerMeasurement tm3 = makeTriggerMeasurement(measurementTimestamp + CommandAlertUtils.DEFAULT_TRIGGER_LAG_MS + punctuateInterval,
                                                    1L, 10L, 100L, false);
    KeyValue<Bytes, Iterable<KeyValue<Bytes, TriggerEvent>>> ret = transformer.transform(makeKey(300L), tm3);
    assertNotNull(ret);
    TriggerEvent triggerEvent = ret.value.iterator().next().value;
    assertEquals("trigger1", triggerEvent.getTriggerGuid());
    assertEquals(1L, triggerEvent.getInfo().getLongValue());
    assertNull(getStoredTriggerMeasurement(kvStore, key));
  }

  @Test
  public void testTriggersAreFiredOnLagSinceFirstMeasurement() {
    MockReadOnlyKeyValueStore.mockAddMonitoringTriggerConfig(
        triggerConfigs,
        CommandAlertUtils.makeTriggerConfig(
            DEFAULT_CLUSTER_ID,
            DEFAULT_CONSUMER_GROUP,
            "trigger1",
            TriggerMetricType.CONSUMPTION_DIFF,
            TriggerCondition.NOT_EQUAL,
            0L
        )
    );

    // will update a measurement for the same key, and expect trigger fired after the time of the
    // first message (START_TIME_MS) + CommandAlertUtils.DEFAULT_TRIGGER_LAG_MS
    final WindowedClusterGroup key = makeKey(100L);
    Long measurementTimestamp = START_TIME_MS;
    TriggerMeasurement triggerMeasurement = makeTriggerMeasurement(measurementTimestamp,
                                                                   1L, 10L, 100L, false);
    assertNull(transformer.transform(key, triggerMeasurement));

    measurementTimestamp += (CommandAlertUtils.DEFAULT_TRIGGER_LAG_MS - 10);
    TriggerMeasurement triggerMeasurement2 = makeTriggerMeasurement(measurementTimestamp,
                                                                    1L, 10L, 100L, false);
    assertNull(transformer.transform(key, triggerMeasurement2));

    // c3 time moves with measurement arrival time, so should fire trigger
    measurementTimestamp += TimeBucket.SIZE;
    TriggerMeasurement triggerMeasurement3 = makeTriggerMeasurement(measurementTimestamp,
                                                                    1L, 10L, 100L, false);
    KeyValue<Bytes, Iterable<KeyValue<Bytes, TriggerEvent>>> ret =
        transformer.transform(key, triggerMeasurement3);
    assertNotNull(ret);
    TriggerEvent triggerEvent = ret.value.iterator().next().value;
    assertEquals("trigger1", triggerEvent.getTriggerGuid());
    assertEquals(1L, triggerEvent.getInfo().getLongValue());
    assertNull(getStoredTriggerMeasurement(kvStore, key));
  }

  @Test
  public void testTriggersAreFiredOnWallclockWhenC3ClockDoesNotMove() {
      MockReadOnlyKeyValueStore.mockAddMonitoringTriggerConfig(
          triggerConfigs,
          CommandAlertUtils.makeTriggerConfig(
              DEFAULT_CLUSTER_ID,
              DEFAULT_CONSUMER_GROUP,
              "trigger1",
              TriggerMetricType.CONSUMPTION_DIFF,
              TriggerCondition.NOT_EQUAL,
              0L
          )
      );

      final WindowedClusterGroup key = makeKey(100L);
      // in the future
      Long measurementTimestamp = START_TIME_MS + 5 * CommandAlertUtils.DEFAULT_TRIGGER_LAG_MS;
      TriggerMeasurement triggerMeasurement = makeTriggerMeasurement(measurementTimestamp,
                                                                     1L, 10L, 100L, false);
      assertNull(transformer.transform(key, triggerMeasurement));

      // move current time way ahead
      mockClock.setCurrentTimeMillis(measurementTimestamp + 5 * CommandAlertUtils.DEFAULT_TRIGGER_LAG_MS);

      // message with arrival time smaller than arrival time of previous message
      TriggerMeasurement tm2 = makeTriggerMeasurement(measurementTimestamp - 10L,
                                                      1L, 10L, 100L, false);
      KeyValue<Bytes, Iterable<KeyValue<Bytes, TriggerEvent>>> ret = transformer.transform(makeKey(200L), tm2);
      assertNotNull(ret);
      TriggerEvent triggerEvent = ret.value.iterator().next().value;
      assertEquals("trigger1", triggerEvent.getTriggerGuid());
      assertEquals(1L, triggerEvent.getInfo().getLongValue());
      assertNull(getStoredTriggerMeasurement(kvStore, key));
  }

  @Test
  public void testTriggersAreFiredOnC3Heartbeat() {
    MockReadOnlyKeyValueStore.mockAddMonitoringTriggerConfig(
        triggerConfigs,
        CommandAlertUtils.makeTriggerConfig(
            DEFAULT_CLUSTER_ID,
            DEFAULT_CONSUMER_GROUP,
            "trigger1",
            TriggerMetricType.CONSUMPTION_DIFF,
            TriggerCondition.NOT_EQUAL,
            0L
        )
    );

    final WindowedClusterGroup key = makeKey(100L);
    // in the future
    Long measurementTimestamp = START_TIME_MS + 5 * CommandAlertUtils.DEFAULT_TRIGGER_LAG_MS;
    TriggerMeasurement triggerMeasurement = makeTriggerMeasurement(measurementTimestamp,
                                                                   1L, 10L, 100L, false);
    assertNull(transformer.transform(key, triggerMeasurement));

    // move current time way ahead
    mockClock.setCurrentTimeMillis(measurementTimestamp + 5 * CommandAlertUtils.DEFAULT_TRIGGER_LAG_MS);

    // message with arrival time smaller than arrival time of previous message
    TriggerMeasurement tm2 = makeHeartbeatTriggerMeasurement();
    KeyValue<Bytes, Iterable<KeyValue<Bytes, TriggerEvent>>> ret = transformer.transform(makeKey(200L), tm2);
    assertNotNull(ret);

    // expect heartbeat and trigger event
    Iterator<KeyValue<Bytes, TriggerEvent>> it = ret.value.iterator();
    assertTrue(it.hasNext());
    TriggerEvent heartbeatTriggerEvent = it.next().value;
    assertTrue(heartbeatTriggerEvent.getHeartbeat());
    assertTrue(it.hasNext());
    TriggerEvent triggerEvent = it.next().value;
    assertEquals("trigger1", triggerEvent.getTriggerGuid());
    assertEquals(1L, triggerEvent.getInfo().getLongValue());
    assertNull(getStoredTriggerMeasurement(kvStore, key));
  }

  @Test
  public void testMultipleTriggerEventsFromSameMeasurement() {
    MockReadOnlyKeyValueStore.mockAddMonitoringTriggerConfig(
        triggerConfigs,
        CommandAlertUtils.makeTriggerConfig(
            DEFAULT_CLUSTER_ID,
            DEFAULT_CONSUMER_GROUP,
            "trigger1",
            TriggerMetricType.CONSUMPTION_DIFF,
            TriggerCondition.NOT_EQUAL,
            0L
        )
    );
    MockReadOnlyKeyValueStore.mockAddMonitoringTriggerConfig(
        triggerConfigs,
        CommandAlertUtils.makeTriggerConfig(
            DEFAULT_CLUSTER_ID,
            DEFAULT_CONSUMER_GROUP,
            "trigger2",
            TriggerMetricType.MAX_LATENCY,
            TriggerCondition.GREATER_THAN,
            1000L
        )
    );
    MockReadOnlyKeyValueStore.mockAddMonitoringTriggerConfig(
        triggerConfigs,
        CommandAlertUtils.makeTriggerConfig(
            DEFAULT_CLUSTER_ID,
            DEFAULT_CONSUMER_GROUP,
            "trigger3",
            TriggerMetricType.AVG_LATENCY,
            TriggerCondition.GREATER_THAN,
            100L
        )
    );

    final WindowedClusterGroup key = makeKey(100L);
    Long measurementTimestamp = START_TIME_MS + 10L;
    TriggerMeasurement triggerMeasurement = makeTriggerMeasurement(measurementTimestamp,
                                                                   1L, 105L, 1005L, false);
    assertNull(transformer.transform(key, triggerMeasurement));

    mockClock.setCurrentTimeMillis(measurementTimestamp + CommandAlertUtils.DEFAULT_TRIGGER_LAG_MS);

    TriggerMeasurement tm2 = makeTriggerMeasurement(measurementTimestamp + CommandAlertUtils.DEFAULT_TRIGGER_LAG_MS,
                                                    1L, 105L, 1005L, false);
    KeyValue<Bytes, Iterable<KeyValue<Bytes, TriggerEvent>>> ret =
        transformer.transform(makeKey(200L), tm2);
    assertNotNull(ret);
    int count = 0;
    Iterator<KeyValue<Bytes, TriggerEvent>> it = ret.value.iterator();
    while (it.hasNext()) {
      TriggerEvent event = it.next().value;
      if (event.getTriggerGuid().equals("trigger1")) {
        assertEquals(1L, event.getInfo().getLongValue());
      } else if (event.getTriggerGuid().equals("trigger2")) {
        assertEquals(1005L, event.getInfo().getLongValue());
      } else if (event.getTriggerGuid().equals("trigger3")) {
        assertEquals(105L, event.getInfo().getLongValue());
      }
      count++;
    }
    assertEquals(3, count);
    assertNull(getStoredTriggerMeasurement(kvStore, key));
  }

  @Test
  public void testMultipleTriggerEventsWithDifferentLagsFromSameMeasurement() {
    final Long trigger1Lag = 600000L;
    final Long trigger2Lag = trigger1Lag + 2 * TimeBucket.SIZE;
    final Long trigger3Lag = trigger2Lag + 2 * TimeBucket.SIZE;
    MockReadOnlyKeyValueStore.mockAddMonitoringTriggerConfig(
        triggerConfigs,
        CommandAlertUtils.makeTriggerConfig(
            DEFAULT_CLUSTER_ID,
            DEFAULT_CONSUMER_GROUP,
            "trigger1",
            TriggerMetricType.CONSUMPTION_DIFF,
            TriggerCondition.NOT_EQUAL,
            0L,
            trigger1Lag
        )
    );
    MockReadOnlyKeyValueStore.mockAddMonitoringTriggerConfig(
        triggerConfigs,
        CommandAlertUtils.makeTriggerConfig(
            DEFAULT_CLUSTER_ID,
            DEFAULT_CONSUMER_GROUP,
            "trigger2",
            TriggerMetricType.MAX_LATENCY,
            TriggerCondition.GREATER_THAN,
            1000L,
            trigger2Lag
        )
    );
    MockReadOnlyKeyValueStore.mockAddMonitoringTriggerConfig(
        triggerConfigs,
        CommandAlertUtils.makeTriggerConfig(
            DEFAULT_CLUSTER_ID,
            DEFAULT_CONSUMER_GROUP,
            "trigger3",
            TriggerMetricType.AVG_LATENCY,
            TriggerCondition.GREATER_THAN,
            100L,
            trigger3Lag
        )
    );

    final WindowedClusterGroup key = makeKey(100L);
    Long measurementTimestamp = START_TIME_MS + 10L;
    TriggerMeasurement triggerMeasurement = makeTriggerMeasurement(measurementTimestamp,
                                                                   1L, 105L, 1005L, false);
    assertNull(transformer.transform(key, triggerMeasurement));

    TriggerMeasurement tm2 = makeTriggerMeasurement(measurementTimestamp + trigger1Lag,
                                                    1L, 105L, 1005L, false);
    KeyValue<Bytes, Iterable<KeyValue<Bytes, TriggerEvent>>> ret =
        transformer.transform(makeKey(200L), tm2);
    assertNotNull(ret);
    Iterator<KeyValue<Bytes, TriggerEvent>> it = ret.value.iterator();
    assertTrue(it.next().value.getTriggerGuid().equals("trigger1"));
    assertFalse(it.hasNext());

    TriggerMeasurement tm3 = makeTriggerMeasurement(measurementTimestamp + trigger2Lag,
                                                    1L, 105L, 1005L, false);
    ret = transformer.transform(makeKey(300L), tm3);
    assertNotNull(ret);
    it = ret.value.iterator();
    assertTrue(it.next().value.getTriggerGuid().equals("trigger2"));
    assertFalse(it.hasNext());

    TriggerMeasurement tm4 = makeTriggerMeasurement(measurementTimestamp + trigger3Lag,
                                                    1L, 105L, 1005L, false);
    ret = transformer.transform(makeKey(400L), tm4);
    assertNotNull(ret);
    it = ret.value.iterator();
    assertTrue(it.next().value.getTriggerGuid().equals("trigger3"));
    assertFalse(it.hasNext());
  }

  private void testTransformStoresTriggerableMeasurement(TriggerMetricType metricType,
                                                         TriggerCondition condition,
                                                         Long triggerValue,
                                                         Long consumptionDiff,
                                                         Long avgLatency,
                                                         Long maxLatency) {
    MockReadOnlyKeyValueStore.mockAddMonitoringTriggerConfig(
        triggerConfigs,
        CommandAlertUtils.makeTriggerConfig(
            DEFAULT_CLUSTER_ID,
            DEFAULT_CONSUMER_GROUP,
            "trigger1",
            metricType,
            condition,
            triggerValue
        )
    );

    final Long measurementTimestamp = START_TIME_MS + 10L;
    final ClientTriggerMeasurement measurement = makeClientTriggerMeasurement(measurementTimestamp,
                                                                              consumptionDiff,
                                                                              avgLatency,
                                                                              maxLatency,
                                                                              false);
    final WindowedClusterGroup key = makeKey(100L);

    TriggerMeasurement triggerMeasurement = makeTriggerMeasurement(measurement);
    KeyValue<Bytes, Iterable<KeyValue<Bytes, TriggerEvent>>> ret =
        transformer.transform(key, triggerMeasurement);
    assertNull(ret);

    TriggerMeasurement storedTriggerMeasurement = getStoredTriggerMeasurement(kvStore, key);
    assertNotNull(storedTriggerMeasurement);
    assertEquals(key.getClusterId(), storedTriggerMeasurement.getClusterId());
    assertEquals(key.getGroup(), storedTriggerMeasurement.getGroup());
    assertEquals(key.getWindow(), storedTriggerMeasurement.getWindow());
  }

}
