// (Copyright) [2016 - 2016] Confluent, Inc.
package io.confluent.controlcenter.streams.verify;

import com.google.common.collect.ImmutableList;
import com.google.common.collect.Range;
import com.google.common.collect.RangeSet;

import io.confluent.monitoring.common.MonitoringMessageUtil;
import io.confluent.monitoring.common.TimeBucket;
import junit.framework.AssertionFailedError;

import org.junit.Before;
import org.junit.Test;

import java.util.Iterator;

import io.confluent.controlcenter.errors.AlreadyShutdownException;
import io.confluent.controlcenter.errors.DuplicateSequenceNumberException;
import io.confluent.controlcenter.errors.SequenceAfterShutdownException;
import io.confluent.controlcenter.errors.ShutdownSequenceNotHighestException;
import io.confluent.controlcenter.record.Controlcenter.VerifierInfo;
import io.confluent.monitoring.record.Monitoring.MessageType;
import io.confluent.monitoring.record.Monitoring.ClientType;
import io.confluent.monitoring.record.Monitoring.MonitoringMessage;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertNotEquals;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertNull;
import static org.junit.Assert.assertTrue;

public class MonitoringVerifierTest {

  private final int historySize = 5;
  private final String clientId = "test-client";
  private final String topic = "test-topic";
  private final int partition = 0;
  private final long startSequence = 0;
  private MonitoringMessage monitoringMessage;

  @Before
  public void setUp() throws Exception {
    monitoringMessage = MonitoringMessage.newBuilder(MonitoringMessageUtil.baseMonitoringMessage())
        .setSession("test-session")
        .setGroup("test-group")
        .setClientId(clientId)
        .setClientType(ClientType.PRODUCER)
        .setTopic(topic)
        .setPartition(partition)
        .build();
  }

  @Test
  public void testFields() {
    MonitoringVerifier
        verifier =
        new MonitoringVerifier(monitoringMessage, 0, historySize);
    assertEquals(clientId, verifier.getClientId());
    assertEquals(topic, verifier.getTopic());
    assertEquals(partition, verifier.getPartition());
    assertEquals(TimeBucket.SIZE, verifier.getSamplePeriod());
  }

  @Test
  public void testMonotonicallyIncreasingSequence() {
    MonitoringVerifier
        verifier =
        new MonitoringVerifier(monitoringMessage, 0, historySize);
    final long startTime = 1100;
    final long window = 1000;
    final long now = startTime + 20;
    final long count = 9;

    for (int i = 0; i < count; i++) {
      addSequence(verifier, startSequence + i, startTime + i, window, -1, -1, now);
    }
    addSequenceWithShutdown(verifier, startSequence + count, startTime + count, window, -1,
                            -1, now);
    // no missing sequence numbers and shutdown means done with this verifier
    assertTrue(verifier.isDone());
    verifyNoMissingSequenceNumbers(verifier, now);
  }

  @Test
  public void testMonotonicallyDecreasingSequence() {
    MonitoringVerifier
        verifier =
        new MonitoringVerifier(monitoringMessage, 0, historySize);
    final long startTime = 1100;
    final long window = 1000;
    final long now = 1010;
    final int count = 9;

    addSequenceWithShutdown(verifier, startSequence + count, startTime + count, window, -1,
                            -1, now);
    for (int i = count - 1; i >= 0; i--) {
      addSequence(verifier, startSequence + i, startTime + i, window, -1, -1, now);
    }
    assertTrue(verifier.isDone());
    verifyNoMissingSequenceNumbers(verifier, now);
  }

  @Test
  public void testOutOfOrderSequence() {
    MonitoringVerifier
        verifier =
        new MonitoringVerifier(monitoringMessage, 0, historySize);
    final long startTime = 1100;
    final long window = 1000;
    final long now = 1220;

    addSequence(verifier, startSequence + 1, startTime + 1, window, -1, -1, now);
    addSequence(verifier, startSequence + 3, startTime + 3, window, -1, -1, now);
    addSequence(verifier, startSequence + 2, startTime + 2, window, -1, -1, now);
    addSequenceWithShutdown(verifier, startSequence + 5, startTime + 5, window, -1, -1, now);
    addSequence(verifier, startSequence, startTime, window, -1, -1, now);
    addSequence(verifier, startSequence + 4, startTime + 4, window, -1, -1, now);

    assertTrue(verifier.isDone());
    verifyNoMissingSequenceNumbers(verifier, now);
  }

  @Test
  public void testGetMissingDataAfterCreate() {
    MonitoringVerifier verifier =
        new MonitoringVerifier(monitoringMessage, 0, historySize);
    verifyNoMissingSequenceNumbers(verifier, 1000);
  }

  @Test(expected = DuplicateSequenceNumberException.class)
  public void testDuplicateSequenceNumber() {
    MonitoringVerifier
        verifier =
        new MonitoringVerifier(monitoringMessage, 0, historySize);
    final long startTime = 1100;
    final long window = 1000;
    final long now = 1220;

    addSequence(verifier, startSequence, startTime, window, -1, -1, now);
    addSequence(verifier, startSequence + 1, startTime + 1, window, -1, -1, now);
    addSequence(verifier, startSequence, startTime, window, -1, -1, now);
  }

  @Test(expected = SequenceAfterShutdownException.class)
  public void testHigherSequenceAfterShutdown() {
    MonitoringVerifier
        verifier =
        new MonitoringVerifier(monitoringMessage, 0, historySize);
    final long startTime = 1100;
    final long window = 1000;
    final long now = 1210;

    addSequence(verifier, startSequence, startTime, window, -1, -1, now);
    addSequenceWithShutdown(verifier, startSequence + 2, startTime + 2, window, -1, -1, now);
    assertFalse(verifier.isDone());
    addSequence(verifier, startSequence + 3, startTime + 3, window, -1, -1, now);
  }

  @Test(expected = ShutdownSequenceNotHighestException.class)
  public void testShutdownLowerSequenceNumber() {
    MonitoringVerifier
        verifier =
        new MonitoringVerifier(monitoringMessage, 0, historySize);
    final long startTime = 1200;
    final long window = 1000;
    final long now = 1210;

    addSequence(verifier, startSequence, startTime, window, -1, -1, now);
    addSequence(verifier, startSequence + 3, startTime + 3, window, -1, -1, now);
    addSequenceWithShutdown(verifier, startSequence + 2, startTime + 2, window, -1, -1, now);
  }

  @Test(expected = AlreadyShutdownException.class)
  public void testSecondShutdownMessage() {
    MonitoringVerifier
        verifier =
        new MonitoringVerifier(monitoringMessage, 0, historySize);
    final long startTime = 1200;
    final long window = 1000;
    final long now = 1210;

    addSequence(verifier, startSequence, startTime, window, -1, -1, now);
    addSequence(verifier, startSequence + 1, startTime + 1, window, -1, -1, now);
    addSequenceWithShutdown(verifier, startSequence + 3, startTime + 3, window, -1, -1, now);
    addSequenceWithShutdown(verifier, startSequence + 2, startTime + 2, window, -1, -1, now);
  }

  @Test(expected = IllegalStateException.class)
  public void testAddSequenceAfterDone() {
    MonitoringVerifier
        verifier =
        new MonitoringVerifier(monitoringMessage, 0, historySize);
    final long startTime = 1200;
    final long window = 1000;
    final long now = 1210;

    addSequence(verifier, startSequence, startTime, window, -1, -1, now);
    addSequenceWithShutdown(verifier, startSequence + 1, startTime + 1, window, -1, -1, now);
    assertTrue(verifier.isDone());
    addSequence(verifier, startSequence + 2, startTime + 2, window, -1, -1, now);
  }

  @Test
  public void testContiguousRangeNoMissingSequence() {
    MonitoringVerifier
        verifier =
        new MonitoringVerifier(monitoringMessage, 0, historySize);
    final long startTime = 1200;
    final long window = 1000;
    final long now = 1210;
    final long count = 9;

    for (int i = 0; i < count; i++) {
      addSequence(verifier, startSequence + i, startTime + i, window, -1, -1, now);
    }
    assertFalse("isDone returned true, but there was no shutdown message",
                verifier.isDone());
    verifyNoMissingSequenceNumbers(verifier, now);
    assertFalse("isDone returned true after verification, even thought there is no shutdown "
                + "message", verifier.isDone());
  }

  @Test
  public void testMissingFirstSequence() {
    ImmutableList<Long> sequences = ImmutableList.of(startSequence + 1,
                                                     startSequence + 2,
                                                     startSequence + 3,
                                                     startSequence + 4);
    doTestMissingSequencesWithZeroTimeout(sequences);
  }

  @Test
  public void testMissingFirstSequenceSet() {
    ImmutableList<Long> sequences = ImmutableList.of(startSequence + 2,
                                                     startSequence + 3,
                                                     startSequence + 4,
                                                     startSequence + 5);
    doTestMissingSequencesWithZeroTimeout(sequences);
  }

  @Test
  public void testMissingMiddleSequence() {
    ImmutableList<Long> sequences = ImmutableList.of(startSequence,
                                                     startSequence + 1,
                                                     startSequence + 3);
    doTestMissingSequencesWithZeroTimeout(sequences);
  }

  @Test
  public void testMissingMiddleSequenceSet() {
    ImmutableList<Long> sequences = ImmutableList.of(startSequence,
                                                     startSequence + 1,
                                                     startSequence + 5);
    doTestMissingSequencesWithZeroTimeout(sequences);
  }

  @Test
  public void testNoFirstSequenceNumberNoTimeout() {
    MonitoringVerifier
        verifier =
        new MonitoringVerifier(monitoringMessage, TimeBucket.SIZE, historySize);
    final long startTime = 1200;
    final long window = 1000;
    final long now = startTime + 30;

    addSequence(verifier, startSequence + 1, startTime + 10, window, -1, -1, now);
    addSequence(verifier, startSequence + 2, startTime + 20, window, -1, -1, now);
    verifyNoMissingSequenceNumbers(verifier, now);
  }

  @Test
  public void testNoMiddleSequenceNumberNoTimeout() {
    MonitoringVerifier
        verifier =
        new MonitoringVerifier(monitoringMessage, TimeBucket.SIZE, historySize);
    final long startTime = 1200;
    final long window = 1000;
    final long now = startTime + 400;

    addSequence(verifier, startSequence, startTime, window, -1, -1, now);
    addSequence(verifier, startSequence + 2, startTime + 200, window, -1, -1, now);
    addSequence(verifier, startSequence + 7, startTime + 300, window, -1, -1, now);
    addSequence(verifier, startSequence + 4, startTime + 300, window, -1, -1, now);
    verifyNoMissingSequenceNumbers(verifier, now);
  }

  @Test
  public void testNoMiddleSequenceNumbersSomeMissing() {
    MonitoringVerifier
        verifier =
        new MonitoringVerifier(monitoringMessage, 2 * TimeBucket.SIZE, historySize);
    final long startTime = 0;
    final long window = 1000;
    final long now = startTime + 10 * TimeBucket.SIZE;

    addSequence(verifier, startSequence, startTime, window, -1, -1, now);
    addSequence(verifier, startSequence + 3, startTime + 3 * TimeBucket.SIZE, window, window,
                window + 2 * TimeBucket.SIZE, now + 3 * TimeBucket.SIZE);
    // since the number of missing sequences < history size, we know exactly the window range
    // with potentially missing monitoring data
    verifyMissingMonitoringDataOneRange(verifier, now + 5 * TimeBucket.SIZE,
                                        window, window + 2 * TimeBucket.SIZE);
  }

  @Test
  public void testMultipleMissingSequenceNumbers() {
    MonitoringVerifier
        verifier =
        new MonitoringVerifier(monitoringMessage, 2 * TimeBucket.SIZE, historySize);
    final long startTime = 0;
    final long window = 1000;
    final long now = startTime + 10 * TimeBucket.SIZE;

    addSequence(verifier, startSequence, startTime, window, -1, -1, now);
    addSequence(verifier, startSequence + 3, startTime + 3 * TimeBucket.SIZE,
                window + 3 * TimeBucket.SIZE, window, window + 2 * TimeBucket.SIZE,
                now + 2 * TimeBucket.SIZE);

    // should not timeout if we advanced < timeout time
    verifyNoMissingSequenceNumbers(verifier, now + 3 * TimeBucket.SIZE);

    addSequence(verifier, startSequence + 4, startTime + 4 * TimeBucket.SIZE,
                window + 4 * TimeBucket.SIZE, window, startTime + 3 * TimeBucket.SIZE,
                now + 3 * TimeBucket.SIZE);
    // sequence 1 and 2 should timeout, with the same window range
    verifyMissingMonitoringDataOneRange(verifier, now + 4 * TimeBucket.SIZE,
                                        window, window + 2 * TimeBucket.SIZE);
    addSequence(verifier, startSequence + 6, startTime + 6 * TimeBucket.SIZE,
                window + 6 * TimeBucket.SIZE,
                window + TimeBucket.SIZE, window + 5 * TimeBucket.SIZE, now + 5 * TimeBucket.SIZE);
    verifyNoMissingSequenceNumbers(verifier, now + 4 * TimeBucket.SIZE);
    addSequence(verifier, startSequence + 7, startTime + 7 * TimeBucket.SIZE,
                window + 7 * TimeBucket.SIZE,
                window + 2 * TimeBucket.SIZE, window + 6 * TimeBucket.SIZE, now + 6 * TimeBucket.SIZE);
    // sequence 5 should timeout, with one window range specified in sequence startSequence + 6
    verifyMissingMonitoringDataOneRange(verifier, now + 7 * TimeBucket.SIZE,
                                        window + TimeBucket.SIZE, window + 5 * TimeBucket.SIZE);
  }

  @Test
  public void testNoMissingNumbersOnBatchArrival() {
    MonitoringVerifier
        verifier =
        new MonitoringVerifier(monitoringMessage, 2 * TimeBucket.SIZE, historySize);
    final long startTime = 0;
    final long window = 1000;
    final long now = startTime + 10 * TimeBucket.SIZE;

    addSequence(verifier, startSequence, startTime, window, -1, -1, now);
    addSequence(verifier, startSequence + 3, startTime + 3 * TimeBucket.SIZE,
                window + 3 * TimeBucket.SIZE, window, window + 2 * TimeBucket.SIZE,
                now);

    // if we call on the same timestamp, sequences should not expire
    verifyNoMissingSequenceNumbers(verifier, now);

    addSequence(verifier, startSequence + 4, startTime + 4 * TimeBucket.SIZE,
                window + 4 * TimeBucket.SIZE, window, window + 3 * TimeBucket.SIZE,
                now);
    verifyNoMissingSequenceNumbers(verifier, now);

    // should expire after timeout
    verifyMissingMonitoringDataOneRange(verifier, now + 2 * TimeBucket.SIZE,
                                        window, window + 2 * TimeBucket.SIZE);
  }

  @Test
  public void testMissingMiddleSequenceNumbersAfterActivityEnd() {
    MonitoringVerifier
        verifier =
        new MonitoringVerifier(monitoringMessage, 2 * TimeBucket.SIZE, historySize);
    final long startTime = 0;
    final long window = 1000;
    final long delay = 10;

    // add sequences where local time is slightly lagging behind client time
    addSequence(verifier, startSequence, startTime, window, -1, -1, startTime + delay);
    addSequence(verifier, startSequence + 1, startTime + 1000,
                window, window, window, startTime + 1000 + delay);
    addSequence(verifier, startSequence + 2,
                startTime + TimeBucket.SIZE,
                window + TimeBucket.SIZE, window, window,
                startTime + TimeBucket.SIZE + delay);
    // missed 2 sequence numbers
    addSequenceWithShutdown(verifier, startSequence + 5,
                            startTime + 3 * TimeBucket.SIZE,
                            window + 3 * TimeBucket.SIZE,
                            window, window + 2 * TimeBucket.SIZE,
                            startTime + 3 * TimeBucket.SIZE + delay);
    // if we verify right away, should not timeout yet
    verifyNoMissingSequenceNumbers(verifier, startTime + 3 * TimeBucket.SIZE + delay);

    // suppose we waited 2 sampling intervals and checked again
    // should timeout sequence 3 & 4 with exactly one window range specified in startSequence + 5
    verifyMissingMonitoringDataOneRange(verifier, startTime + 5 * TimeBucket.SIZE + delay,
                                        window, window + 2 * TimeBucket.SIZE);

    // all sequence numbers are verified and no more are expected
    assertTrue(verifier.isDone());
  }

  @Test
  public void testTimeoutWaitingForNextSequence() {
    MonitoringVerifier
        verifier =
        new MonitoringVerifier(monitoringMessage, 2 * TimeBucket.SIZE, historySize);
    final long startTime = 0;
    final long window = 1000;
    final long delay = 10;

    // add sequences where local time is slightly lagging behind client time
    addSequence(verifier, startSequence, startTime, window, -1, -1, startTime + delay);
    addSequence(verifier, startSequence + 1, startTime + 1000,
                window, window, window, startTime + 1000 + delay);
    addSequence(verifier, startSequence + 2,
                startTime + TimeBucket.SIZE,
                window + TimeBucket.SIZE, window, window,
                startTime + TimeBucket.SIZE + delay);
    verifyNoMissingSequenceNumbers(verifier, startTime + TimeBucket.SIZE + delay);
    assertFalse(verifier.isDone());

    // once session ends, next time sequence is expired, verifier stops waiting for sequence numbers
    verifier.notifySessionEnded();
    verifyMissingMonitoringData(verifier, startTime + 11 * TimeBucket.SIZE + delay,
                                ImmutableList.of(window,
                                                 window + TimeBucket.SIZE,
                                                 window + 2 * TimeBucket.SIZE));
    assertTrue(verifier.isDone());
  }

  @Test(expected = DuplicateSequenceNumberException.class)
  public void testMonitoringDataLongDelay() {
    MonitoringVerifier
        verifier =
        new MonitoringVerifier(monitoringMessage, 2 * TimeBucket.SIZE, historySize);
    final long startTime = 0;
    final long window = 1000;

    addSequence(verifier, startSequence, startTime, window, -1, -1, startTime);
    addSequence(verifier, startSequence + 1, startTime + 1000,
                window, window, window, startTime + 1000);
    addSequence(verifier, startSequence + 2,
                startTime + TimeBucket.SIZE,
                window + TimeBucket.SIZE, window, window,
                startTime + TimeBucket.SIZE);
    verifyNoMissingSequenceNumbers(verifier, startTime + TimeBucket.SIZE);
    assertFalse(verifier.isDone());

    // if we check much later without sending any more sequence numbers, a sequence number after
    // the highest sequence should timeout and verifier stops waiting for any more sequences
    verifyMissingMonitoringData(verifier, startTime + 9 * TimeBucket.SIZE, ImmutableList.of(window));
    assertTrue(verifier.isSessionNotActive());

    // if we start sending a new sequence (the one that expired), we should get a dup
    addSequence(verifier, startSequence + 3,
                startTime + 2 * TimeBucket.SIZE,
                window + 2 * TimeBucket.SIZE, window, window,
                startTime + 10 * TimeBucket.SIZE);
  }

  @Test
  public void testMaybeEndSessionIfClientDropsOut() {
    MonitoringVerifier
        verifier =
        new MonitoringVerifier(monitoringMessage, 2 * TimeBucket.SIZE, historySize);
    final long startTime = 0;
    final long window = 1000;

    addSequence(verifier, startSequence, startTime, window, -1, -1, startTime);
    addSequence(verifier, startSequence + 1, startTime + 1000,
                window, window, window, startTime + 1000);
    verifyNoMissingSequenceNumbers(verifier, startTime + TimeBucket.SIZE);
    assertFalse(verifier.isDone());

    // if we check much later without sending any more sequence numbers, a sequence number after
    // the highest sequence should timeout and verifier stops waiting for any more sequences
    verifyMissingMonitoringData(verifier, startTime + 9 * TimeBucket.SIZE, ImmutableList.of(window));
    assertTrue(verifier.isSessionNotActive());

    // after that, the verifier should stop expiring any more sequences
    verifyNoMissingSequenceNumbers(verifier, startTime + 10 * TimeBucket.SIZE);
    verifyNoMissingSequenceNumbers(verifier, startTime + 11 * TimeBucket.SIZE);
    verifyNoMissingSequenceNumbers(verifier, startTime + 12 * TimeBucket.SIZE);

    // if we start sending a new sequence (suppose we just skip the dup), session should continue
    addSequence(verifier, startSequence + 3,
                startTime + 3 * TimeBucket.SIZE,
                window + 3 * TimeBucket.SIZE, window, window,
                startTime + 13 * TimeBucket.SIZE);
    assertFalse(verifier.isSessionNotActive());

    // one more
    addSequence(verifier, startSequence + 4,
                startTime + 3 * TimeBucket.SIZE,
                window + 3 * TimeBucket.SIZE, window, window,
                startTime + 14 * TimeBucket.SIZE);
    assertFalse(verifier.isSessionNotActive());
  }

  @Test(expected = DuplicateSequenceNumberException.class)
  public void testAddSequenceAfterSequenceTimeout() {
    MonitoringVerifier
        verifier =
        new MonitoringVerifier(monitoringMessage, 2 * TimeBucket.SIZE, historySize);
    final long startTime = 0;
    final long window = 1000;
    final long delay = 10;

    // add sequences where local time is slightly lagging behind client time
    addSequence(verifier, startSequence, startTime, window, -1, -1, startTime + delay);
    addSequence(verifier, startSequence + 2, startTime + 2000,
                window, window, window, startTime + 2000 + delay);
    // should expire sequence number 1 after long delay (but not too long, so that it ends the
    // session) and a sequence after the last recorded window range for missing monitoring data
    // should include window
    verifyMissingMonitoringData(verifier, startTime + 4 * TimeBucket.SIZE + delay,
                                ImmutableList.of(window));
    addSequence(verifier, startSequence + 1, startTime + 1000, window, window, window,
                startTime + 9 * TimeBucket.SIZE + delay);
  }

  @Test
  public void testAddSequenceAfterLongDelayButBeforeVerification() {
    MonitoringVerifier
        verifier =
        new MonitoringVerifier(monitoringMessage, 2 * TimeBucket.SIZE, historySize);
    final long startTime = 0;
    final long window = 1000;
    final long delay = 10;

    // add sequences where local time is slightly lagging behind client time
    addSequence(verifier, startSequence, startTime, window, -1, -1, startTime + delay);
    addSequence(verifier, startSequence + 2, startTime + 2000,
                window, window, window, startTime + 2000 + delay);

    // if we don't check for sequence numbers expirations (by calling getMissingMonitoringData),
    // sequence numbers don't expire. So, if sequence number arrives even after long delay, but
    // before the check, then it is not considered as timed out.
    addSequence(verifier, startSequence + 1, startTime + 1000, window, window, window,
                startTime + 9 * TimeBucket.SIZE + delay);
    verifyNoMissingSequenceNumbers(verifier, startTime + 9 * TimeBucket.SIZE + delay);
  }

  @Test
  public void testMissingConsecutiveSequencesMoreThanHistory() {
    MonitoringVerifier
        verifier =
        new MonitoringVerifier(monitoringMessage, 2 * TimeBucket.SIZE, 3);
    final long startTime = 0;
    final long window = 1000;
    final long delay = 10;

    // add sequences where local time is slightly lagging behind client time
    addSequence(verifier, startSequence, startTime, window, -1, -1, startTime + delay);
    addSequence(verifier, startSequence + 5, startTime + 3 * TimeBucket.SIZE,
                window + 3 * TimeBucket.SIZE, window, window + 3 * TimeBucket.SIZE,
                startTime + 3 * TimeBucket.SIZE + delay);
    // sequence 1 is outside of range recorded in startSequence + 5
    // when we check, we should get sequences 1, 2, 3 missing
    verifyMissingMonitoringData(verifier, startTime + 9 * TimeBucket.SIZE + delay,
                                ImmutableList.of(window,
                                                 window + TimeBucket.SIZE,
                                                 window + 2 * TimeBucket.SIZE,
                                                 window + 2 * TimeBucket.SIZE,
                                                 window + 3 * TimeBucket.SIZE));
  }

  @Test
  public void testMissingConsecutiveSequencesMoreThanHistoryNoWindowInfo() {
    // same as above, but the missing sequences inside history range are heartbeat sequences, so
    // the next non-missing sequence does not have min and max window info
    MonitoringVerifier
        verifier =
        new MonitoringVerifier(monitoringMessage, 2 * TimeBucket.SIZE, 3);
    final long startTime = 0;
    final long window = 1000;
    final long delay = 10;

    // add sequences where local time is slightly lagging behind client time
    addSequence(verifier, startSequence, startTime, window, -1, -1, startTime + delay);
    addSequence(verifier, startSequence + 5, startTime + 3 * TimeBucket.SIZE,
                window + 3 * TimeBucket.SIZE, -1, -1, startTime + 3 * TimeBucket.SIZE + delay);
    // sequence 1 is outside of history range for startSequence + 5, and startSequence + 5 does
    // not have history info because sequences 2,3,4 are heartbeat sequences
    // sequence startSequence + 5 has window info and is also a last sequence, the window range
    // will be calculated based on window of sequence startSequence + 5 (for sequence 1)
    verifyMissingMonitoringData(verifier, startTime + 9 * TimeBucket.SIZE + delay,
                                ImmutableList.of(window,
                                                 window + TimeBucket.SIZE,
                                                 window + 2 * TimeBucket.SIZE,
                                                 window + 3 * TimeBucket.SIZE));
  }

  @Test
  public void testMissingConsecutiveSequencesMoreThanMaxReported() {
    MonitoringVerifier
        verifier =
        new MonitoringVerifier(monitoringMessage, 2 * TimeBucket.SIZE, 50);
    final long startTime = 0;
    final long window = 1000;
    final long delay = 10;

    // add sequences where local time is slightly lagging behind client time
    addSequence(verifier, startSequence, startTime, window, -1, -1, startTime + delay);
    addSequence(verifier, startSequence + 30, startTime + 3 * TimeBucket.SIZE,
                window + 3 * TimeBucket.SIZE, window, window + 3 * TimeBucket.SIZE,
                startTime + 3 * TimeBucket.SIZE + delay);
    // sequence 1 is outside of range recorded in startSequence + 5
    // when we check, we should get sequences 1, 2, 3 missing
    verifyMissingMonitoringData(verifier, startTime + 9 * TimeBucket.SIZE + delay,
                                ImmutableList.of(window,
                                                 window + TimeBucket.SIZE,
                                                 window + 2 * TimeBucket.SIZE,
                                                 window + 3 * TimeBucket.SIZE));
  }

  @Test
  public void testMissingSomeHeartbeatMsgInConsecutiveSequenceDoesNotDetectLostMonitoringData() {
    MonitoringVerifier
        verifier =
        new MonitoringVerifier(monitoringMessage, 2 * TimeBucket.SIZE, 3);
    final long startTime = 0;
    final long window = 1000;
    final long delay = 10;

    addSequence(verifier, startSequence, startTime, window, -1, -1, startTime + delay);
    addSequence(verifier, startSequence + 4, startTime + 3 * TimeBucket.SIZE,
                window + 3 * TimeBucket.SIZE, -1, -1, startTime + 3 * TimeBucket.SIZE + delay);
    // all lost sequences are heartbeat messages,
    // because min and max window is null in startSequence + 4
    verifyNoMissingSequenceNumbers(verifier, startTime + 9 * TimeBucket.SIZE + delay);
  }

  @Test
  public void testMissingSequenceWhenNoWindowInfo() {
    MonitoringVerifier verifier =
        new MonitoringVerifier(monitoringMessage, 2 * TimeBucket.SIZE, 3);
    final long startTime = 0;
    final long delay = 10;

    // lost sequence 0, and 3 heartbeat messages, and next sequence that got delivered is also
    // a heartbeat msg
    addHeartbeatSequence(verifier, startSequence + 4, startTime + 1000,
                         -1, -1, startTime + 1000 + delay);

    // if we check much later without sending any more sequence numbers, a sequence number after
    // the highest sequence should timeout and verifier stops waiting for any more sequences
    // for now this use-case returns [0, 0] range
    verifyMissingMonitoringDataOneRange(verifier, startTime + 9 * TimeBucket.SIZE + delay, 0L, 0L);
  }

  @Test
  public void testSerializeDeserialize() {
    MonitoringVerifier
        verifier =
        new MonitoringVerifier(monitoringMessage, 3 * TimeBucket.SIZE, historySize);
    final long startTime = 1000;
    final long window = 1000;
    final long now = startTime + 40;

    addSequence(verifier, startSequence, startTime, window, -1, -1, now);
    addSequence(verifier, startSequence + 1, startTime, window, window, window, now);
    addSequence(verifier, startSequence + 10, startTime, window, window, window, now);
    verifyToFromVerifierInfo(verifier);
  }

  @Test
  public void testSerializeDeserializeOneRange() {
    MonitoringVerifier
        verifier =
        new MonitoringVerifier(monitoringMessage, 2 * TimeBucket.SIZE, historySize);
    final long startTime = 1400;
    final long now = startTime + 80;
    final long window = 1000;

    addSequence(verifier, startSequence, startTime, window, 10, 100, now);
    addSequence(verifier, startSequence + 1, startTime, window, window, window, now);
    addSequence(verifier, startSequence + 2, startTime, window, window, window, now);
    verifyToFromVerifierInfo(verifier);
  }

  @Test
  public void testSerializeDeserializeInitializedVerifier() {
    MonitoringVerifier
        verifier =
        new MonitoringVerifier(monitoringMessage, 2 * TimeBucket.SIZE, historySize);
    verifyToFromVerifierInfo(verifier);
  }


  @Test
  public void testSerializeDeserializeDoneVerifier() {
    MonitoringVerifier
        verifier =
        new MonitoringVerifier(monitoringMessage, 3 * TimeBucket.SIZE, historySize);
    final long startTime = 1000;
    final long now = startTime + 40;
    final long window = 1000;

    addSequence(verifier, startSequence, startTime, window, -1, -1, now);
    addSequenceWithShutdown(verifier, startSequence + 1, startTime, window, window, window, now);
    assertTrue(verifier.isDone());
    verifyToFromVerifierInfo(verifier);
  }

  @Test
  public void testEquals() {
    MonitoringVerifier
        verifier1 =
        new MonitoringVerifier(monitoringMessage, 3 * TimeBucket.SIZE, historySize);
    MonitoringVerifier
        verifier2 =
        new MonitoringVerifier(monitoringMessage, 3 * TimeBucket.SIZE, historySize);
    assertEquals(verifier1, verifier2);
    assertEquals(verifier2, verifier1);
    assertEquals(verifier1.hashCode(), verifier2.hashCode());
  }

  @Test
  public void testNotEquals() {
    MonitoringVerifier
        verifier1 =
        new MonitoringVerifier(monitoringMessage, 2 * TimeBucket.SIZE, historySize);
    MonitoringVerifier
        verifier2 =
        new MonitoringVerifier(monitoringMessage, 2 * TimeBucket.SIZE, historySize);
    assertEquals(verifier1, verifier2);

    addSequence(verifier1, 0l, 1000l, 1000l, -1, -1, 1010l);
    assertNotEquals(verifier1, verifier2);
    assertNotEquals(verifier2, verifier1);
    assertNotEquals(verifier1, null);

    addSequence(verifier2, 0l, 1000l, 1000l, -1, -1, 1010l);
    assertEquals(verifier1, verifier2);
    assertEquals(verifier1.hashCode(), verifier2.hashCode());

    addSequence(verifier2, 3l, 1000l, 1000l, 1000l, 1000l, 1010l);
    addSequence(verifier2, 10l, 1000l, 1000l, 1000l, 1000l, 1010l);
    assertNotEquals(verifier1, verifier2);
    assertNotEquals(verifier1.hashCode(), verifier2.hashCode());

    addSequence(verifier1, 3l, 1000l, 1000l, 1000l, 1000l, 1010l);
    addSequence(verifier1, 10l, 1000l, 1000l, 1000l, 1000l, 1010l);
    assertEquals(verifier1, verifier2);
  }

  private void verifyToFromVerifierInfo(MonitoringVerifier verifier) {
    VerifierInfo info = verifier.createVerifierInfo();
    assertNotNull(info);
    MonitoringVerifier newMonitoringVerifier;
    try {
      newMonitoringVerifier = new MonitoringVerifier(info);
    } catch (Exception e) {
      throw new AssertionFailedError("Failed to construct monitoring verifier from VerifierInfo");
    }
    assertTrue(verifier.equals(newMonitoringVerifier));
  }

  /**
   * Basic test that adds sequences in 'sequenceList', which has missing sequence numbers. Since
   * timeout is zero, all sequence numbers that are missing from the list should also be declared
   * missing by the verifier after after the first call to getMissingMonitoringData()
   *
   * At least one sequence number must be missing in 'sequenceList'.
   *
   * @param sequenceList List of sequence numbers to add to the verifier. The timestamps would be
   *                     assigned as starting timestamp + sequence number.
   */
  private void doTestMissingSequencesWithZeroTimeout(ImmutableList<Long> sequenceList) {
    MonitoringVerifier
        verifier =
        new MonitoringVerifier(monitoringMessage, 0, historySize);
    final long startTime = 1000;
    final long now = startTime + 20;
    final long window = 1000;

    Long lastSequence = sequenceList.get(sequenceList.size() - 1);
    for (Long sequence : sequenceList) {
      if (!sequence.equals(lastSequence)) {
        addSequence(verifier, sequence, startTime + sequence, window, window, window, now);
      } else {
        addSequenceWithShutdown(verifier, sequence, startTime + sequence, window, window,
                                window, now);
      }
    }
    assertFalse("isDone returned true, but we have an un-processed missing sequence number",
                verifier.isDone());

    // every sequence had the same window, so verifier should be able to detect the range that
    // contains that window
    verifyMissingMonitoringData(verifier, now, ImmutableList.of(window));
    assertTrue("Expected verifier to be done after finding a missing sequence", verifier.isDone());
  }

  private static void addHeartbeatSequence(MonitoringVerifier verifier, long sequence,
                                           long timestamp, long minWindow, long maxWindow,
                                           long now) {
    addSequence(verifier, MessageType.HEARTBEAT, sequence,
                timestamp, 0, 0, minWindow, maxWindow, now);
  }

  private static void addSequence(MonitoringVerifier verifier, long sequence, long timestamp,
                                  long window, long minWindow, long maxWindow, long now) {
    addSequence(verifier, MessageType.NORMAL, sequence,
                timestamp, window, 1, minWindow, maxWindow, now);
  }

  private static void addSequence(MonitoringVerifier verifier, MessageType type,
                                  long sequence, long timestamp,
                                  long window, long count, long minWindow,
                                  long maxWindow, long now) {
    MonitoringMessage monitoringMessage = MonitoringMessage.newBuilder(MonitoringMessageUtil.baseMonitoringMessage())
        .setClientId(verifier.getClientId())
        .setTopic(verifier.getTopic())
        .setPartition(verifier.getPartition())
        .setSequence(sequence)
        .setTimestamp(timestamp)
        .setWindow(window)
        .setMinWindow(minWindow)
        .setMaxWindow(maxWindow)
        .setSamplePeriod(TimeBucket.SIZE)
        .setCount(count)
        .setType(type)
        .build();
    verifier.addSequence(monitoringMessage, now);
  }

  private static void addSequenceWithShutdown(MonitoringVerifier verifier, long sequence,
                                              long timestamp, long window, long minWindow,
                                              long maxWindow, long now) {
    addSequenceWithShutdown(verifier, MessageType.NORMAL, sequence, timestamp, window,
                            TimeBucket.SIZE, 1, minWindow, maxWindow, now);
  }

  private static void addSequenceWithShutdown(MonitoringVerifier verifier,
                                              MessageType type, long sequence,
                                              long timestamp, long window, long samplePeriod,
                                              long count, long minWindow, long maxWindow, long now) {
    MonitoringMessage monitoringMessage = MonitoringMessage.newBuilder(MonitoringMessageUtil.baseMonitoringMessage())
        .setClientId(verifier.getClientId())
        .setTopic(verifier.getTopic())
        .setPartition(verifier.getPartition())
        .setSequence(sequence)
        .setTimestamp(timestamp)
        .setWindow(window)
        .setMinWindow(minWindow)
        .setMaxWindow(maxWindow)
        .setSamplePeriod(samplePeriod)
        .setCount(count)
        .setType(type)
        .setShutdown(Boolean.TRUE)
        .build();
    verifier.addSequence(monitoringMessage, now);
  }

  private static void verifyNoMissingSequenceNumbers(MonitoringVerifier verifier, long now) {
    assertNull(verifier.getMissingMonitoringData(now));
  }

  /**
   * Verifies that 'getMissingMonitoringData' returns ranges that include every window in
   * 'windowList'.
   */
  private static void verifyMissingMonitoringData(MonitoringVerifier verifier, long now,
                                                  ImmutableList<Long> windowList) {
    RangeSet<Long> ranges = verifier.getMissingMonitoringData(now);
    assertNotNull(ranges);
    for (Long window : windowList) {
      assertTrue(String.format("Expected window %d to be in ranges returned from"
                               + "getMissingMonitoringData()", window), ranges.contains(window));
    }
  }

  /**
   * Verifies that 'getMissingMonitoringData' returns one range with lower endpoint ==
   * expectedMinWindow and upper endpoint == expectedMaxWindow
   */
  private static void verifyMissingMonitoringDataOneRange(MonitoringVerifier verifier, long now,
                                                          Long expectedMinWindow,
                                                          Long expectedMaxWindow) {
    RangeSet<Long> ranges = verifier.getMissingMonitoringData(now);
    assertNotNull(ranges);
    Iterator<Range<Long>> it = ranges.asRanges().iterator();
    assertTrue("Expected at least one range to be returned from getMissingMonitoringData",
               it.hasNext());
    Range<Long> winRange = it.next();
    assertEquals(expectedMinWindow, winRange.lowerEndpoint());
    assertEquals(expectedMaxWindow, winRange.upperEndpoint());
    assertFalse("Expected exactly one range to be returned from getMissingMonitoringData",
                it.hasNext());
  }
}
