// @flow
// (Copyright) [2016 - 2016] Confluent, Inc.
import { makeAction } from 'actions/util';

import type { Action } from 'types';

export type ConnectorConfig = {
  'connector.class': string,
  name: string
};

export type ConnectorType = 'SINK' | 'SOURCE';

export type Connector = {
  config: ConnectorConfig,
  name: string,
  type: ConnectorType,
  tasks: Array<Object>
};

export type ConnectorOrTaskStatus = 'FAILED' | 'PAUSED' | 'RUNNING' | 'UNASSIGNED';

export type ConnectorList = Array<{
  name: string,
  type: ConnectorType,
  tasks: number,
  topics: string
}>;

export type ConnectorListState = {
  items: ConnectorList,
  isFresh: boolean
};

export type ConnectorStatus = {
  name: string,
  connector: {
    state: ConnectorOrTaskStatus,
    worker_id: string
  },
  tasks: Array<Object>
};

// Request actions
export const CONNECTOR_DESTROY_FAILURE = 'CONNECTOR_DESTROY_FAILURE';
export type ConnectorDestroyFailureAction = Action<'CONNECTOR_DESTROY_FAILURE', { name: string, error: Object }>;
export const CONNECTOR_DESTROY_REQUEST = 'CONNECTOR_DESTROY_REQUEST';
export type ConnectorDestroyRequestAction = Action<'CONNECTOR_DESTROY_REQUEST', { name: string }>;
export const CONNECTOR_DESTROY_SUCCESS = 'CONNECTOR_DESTROY_SUCCESS';
export type ConnectorDestroySuccessAction = Action<'CONNECTOR_DESTROY_SUCCESS', { name: string }>;
export const destroy = {
  failure: (name: string, error: Error): ConnectorDestroyFailureAction => makeAction(CONNECTOR_DESTROY_FAILURE, { name, error }),
  request: (name: string): ConnectorDestroyRequestAction => makeAction(CONNECTOR_DESTROY_REQUEST, { name }),
  success: (name: string): ConnectorDestroySuccessAction => makeAction(CONNECTOR_DESTROY_SUCCESS, { name })
};

export const CONNECTOR_FETCH_FAILURE = 'CONNECTOR_FETCH_FAILURE';
export type ConnectorFetchFailureAction = Action<'CONNECTOR_FETCH_FAILURE', { name: string, error: Object }>;
export const CONNECTOR_FETCH_REQUEST = 'CONNECTOR_FETCH_REQUEST';
export type ConnectorFetchRequestAction = Action<'CONNECTOR_FETCH_REQUEST', { name: string }>;
export const CONNECTOR_FETCH_SUCCESS = 'CONNECTOR_FETCH_SUCCESS';
export type ConnectorFetchSuccessAction = Action<'CONNECTOR_FETCH_SUCCESS', { connector: Connector }>;
export const fetch = {
  failure: (name: string, error: Error): ConnectorFetchFailureAction => makeAction(CONNECTOR_FETCH_FAILURE, { name, error }),
  request: (name: string): ConnectorFetchRequestAction => makeAction(CONNECTOR_FETCH_REQUEST, { name }),
  success: (name: string, connector: Connector): ConnectorFetchSuccessAction => makeAction(CONNECTOR_FETCH_SUCCESS, { name, connector })
};

export const CONNECTOR_LIST_FETCH_FAILURE = 'CONNECTOR_LIST_FETCH_FAILURE';
export type ConnectorListFetchFailureAction = Action<'CONNECTOR_LIST_FETCH_FAILURE', { error: Object }>;
export const CONNECTOR_LIST_FETCH_REQUEST = 'CONNECTOR_LIST_FETCH_REQUEST';
export type ConnectorListFetchRequestAction = Action<'CONNECTOR_LIST_FETCH_REQUEST', void>;
export const CONNECTOR_LIST_FETCH_SUCCESS = 'CONNECTOR_LIST_FETCH_SUCCESS';
export type ConnectorListFetchSuccessAction = Action<'CONNECTOR_LIST_FETCH_SUCCESS', { connectors: ConnectorList }>;
export const listFetch = {
  failure: (error: Error): ConnectorListFetchFailureAction => makeAction(CONNECTOR_LIST_FETCH_FAILURE, { error }),
  request: (): ConnectorListFetchRequestAction => makeAction(CONNECTOR_LIST_FETCH_REQUEST),
  success: (connectors: ConnectorList): ConnectorListFetchSuccessAction => makeAction(CONNECTOR_LIST_FETCH_SUCCESS, { connectors })
};

export const CONNECTOR_PAUSE_FAILURE = 'CONNECTOR_PAUSE_FAILURE';
export type ConnectorPauseFailureAction = Action<'CONNECTOR_PAUSE_FAILURE', { error: Object }>;
export const CONNECTOR_PAUSE_REQUEST = 'CONNECTOR_PAUSE_REQUEST';
export type ConnectorPauseRequestAction = Action<'CONNECTOR_PAUSE_REQUEST', { name: string }>;
export const CONNECTOR_PAUSE_SUCCESS = 'CONNECTOR_PAUSE_SUCCESS';
export type ConnectorPauseSuccessAction = Action<'CONNECTOR_PAUSE_SUCCESS', { name: string }>;
export const pause = {
  failure: (name: string, error: Error): ConnectorPauseFailureAction => makeAction(CONNECTOR_PAUSE_FAILURE, { name, error }),
  request: (name: string): ConnectorPauseRequestAction => makeAction(CONNECTOR_PAUSE_REQUEST, { name }),
  success: (name: string): ConnectorPauseSuccessAction => makeAction(CONNECTOR_PAUSE_SUCCESS, { name })
};

export const CONNECTOR_RESUME_FAILURE = 'CONNECTOR_RESUME_FAILURE';
export type ConnectorResumeFailureAction = Action<'CONNECTOR_RESUME_FAILURE', { error: Object }>;
export const CONNECTOR_RESUME_REQUEST = 'CONNECTOR_RESUME_REQUEST';
export type ConnectorResumeRequestAction = Action<'CONNECTOR_RESUME_REQUEST', { name: string }>;
export const CONNECTOR_RESUME_SUCCESS = 'CONNECTOR_RESUME_SUCCESS';
export type ConnectorResumeSuccessAction = Action<'CONNECTOR_RESUME_SUCCESS', { name: string }>;
export const resume = {
  failure: (name: string, error: Error): ConnectorResumeFailureAction => makeAction(CONNECTOR_RESUME_FAILURE, { name, error }),
  request: (name: string): ConnectorResumeRequestAction => makeAction(CONNECTOR_RESUME_REQUEST, { name }),
  success: (name: string): ConnectorResumeSuccessAction => makeAction(CONNECTOR_RESUME_SUCCESS, { name })
};

export const CONNECTOR_SAVE_FAILURE = 'CONNECTOR_SAVE_FAILURE';
export type ConnectorSaveFailureAction = Action<'CONNECTOR_SAVE_FAILURE', { name: string, error: Object }>;
export const CONNECTOR_SAVE_REQUEST = 'CONNECTOR_SAVE_REQUEST';
export type ConnectorSaveRequestAction = Action<'CONNECTOR_SAVE_REQUEST', { name: string }>;
export const CONNECTOR_SAVE_SUCCESS = 'CONNECTOR_SAVE_SUCCESS';
export type ConnectorSaveSuccessAction = Action<'CONNECTOR_SAVE_SUCCESS', { connector: Connector }>;
export const save = {
  failure: (name: string, error: Error): ConnectorSaveFailureAction => makeAction(CONNECTOR_SAVE_FAILURE, { name, error }),
  request: (name: string): ConnectorSaveRequestAction => makeAction(CONNECTOR_SAVE_REQUEST, { name }),
  success: (name: string, connector: Connector): ConnectorSaveSuccessAction => makeAction(CONNECTOR_SAVE_SUCCESS, { name, connector })
};

export const CONNECTOR_STATUS_FAILURE = 'CONNECTOR_STATUS_FAILURE';
export type ConnectorStatusFailureAction = Action<'CONNECTOR_STATUS_FAILURE', { error: Object }>;
export const CONNECTOR_STATUS_REQUEST = 'CONNECTOR_STATUS_REQUEST';
export type ConnectorStatusRequestAction = Action<'CONNECTOR_STATUS_REQUEST', { name: string }>;
export const CONNECTOR_STATUS_SUCCESS = 'CONNECTOR_STATUS_SUCCESS';
export type ConnectorStatusSuccessAction = Action<'CONNECTOR_STATUS_SUCCESS', { name: string, connectorStatus: ConnectorStatus }>;
export const status = {
  failure: (name: string, error: Error): ConnectorStatusFailureAction => makeAction(CONNECTOR_STATUS_FAILURE, { name, error }),
  request: (name: string): ConnectorStatusRequestAction => makeAction(CONNECTOR_STATUS_REQUEST, { name }),
  success: (name: string, connectorStatus: ConnectorStatus): ConnectorStatusSuccessAction => makeAction(CONNECTOR_STATUS_SUCCESS, { name, connectorStatus })
};

// UI actions
export const DELETE_CONNECTOR = 'DELETE_CONNECTOR';
export type DeleteConnectorAction = Action<'DELETE_CONNECTOR', { name: string }>;
export const deleteConnector = (name: string): DeleteConnectorAction => makeAction(DELETE_CONNECTOR, { name });

export const LOAD_CONNECTOR = 'LOAD_CONNECTOR';
export type LoadConnectorAction = Action<'LOAD_CONNECTOR', { name: string }>;
export const loadConnector = (name: string): LoadConnectorAction => makeAction(LOAD_CONNECTOR, { name });

export const LOAD_CONNECTOR_STATUS = 'LOAD_CONNECTOR_STATUS';
export type LoadConnectorStatusAction = Action<'LOAD_CONNECTOR_STATUS', { name: string }>;
export const loadConnectorStatus = (name: string): LoadConnectorStatusAction => makeAction(LOAD_CONNECTOR_STATUS, { name });

export const LOAD_CONNECTOR_LIST = 'LOAD_CONNECTOR_LIST';
export type LoadConnectorListAction = Action<'LOAD_CONNECTOR_LIST', void>;
export const loadConnectorList = (): LoadConnectorListAction => makeAction(LOAD_CONNECTOR_LIST);

export const PAUSE_CONNECTOR = 'PAUSE_CONNECTOR';
export type PauseConnectorAction = Action<'PAUSE_CONNECTOR', { name: string }>;
export const pauseConnector = (name: string): PauseConnectorAction => makeAction(PAUSE_CONNECTOR, { name });

export const RESUME_CONNECTOR = 'RESUME_CONNECTOR';
export type ResumeConnectorAction = Action<'RESUME_CONNECTOR', { name: string }>;
export const resumeConnector = (name: string): ResumeConnectorAction => makeAction(RESUME_CONNECTOR, { name });

export const SAVE_CONNECTOR = 'SAVE_CONNECTOR';
export type SaveConnectorAction = Action<'SAVE_CONNECTOR', { config: ConnectorConfig }>;
export const saveConnector = (config: ConnectorConfig): SaveConnectorAction => makeAction(SAVE_CONNECTOR, { config });
