// @flow
// (Copyright) [2016 - 2016] Confluent, Inc.

import debug from 'debug';

import {
  getJson,
  makeAction,
  encodeQueryData
} from '../util';
import { idNameFromDataLayoutType } from 'utils/monitoring/data';
import { setPageScrollInfo } from 'actions/monitoring/layout';
import {
  allowDataFetch,
  disallowDataFetch,
  incrementConnectionCount,
  decrementConnectionCount
} from './dataManager';
import { getSelectedClusterIdSelector } from 'selectors/clusters';

import type { DataLayoutType, RequestType } from 'types/monitoring/data';
import type {
  Action,
  Dispatchable,
  Dispatcher
} from 'types';

import type { ClusterId } from 'types/clusters';

export function constructMonitoringDataRequest(clusterId: ClusterId, dataLayoutType: DataLayoutType, params: Object) {
  return `/2.0/monitoring/${clusterId == null ? 'null' : clusterId}/${dataLayoutType.toLowerCase()}?${encodeQueryData(params)}`;
}

const log = debug('controlcenter:actions:monitoring');

const MINIMUM_WAIT_AFTER_ERROR_MS = 5000;

export const RECEIVE_DATA = 'RECEIVE_DATA';
export type ReceiveDataActionPayload = { data: Object, requestType: RequestType };
export type ReceiveDataAction = Action<'RECEIVE_DATA', ReceiveDataActionPayload>;

export const CLEAR_DATA = 'CLEAR_DATA';
export type ClearDataAction = Action<'CLEAR_DATA', null>;
export const clearData = (): ClearDataAction => makeAction(CLEAR_DATA, null);

export type LoadDataFunction = (
  dataLayoutType: DataLayoutType,
  id: ?string,
  startTimeMs: number,
  stopTimeMs: number,
  rollup: string,
  requestType: RequestType,
  extraArguments: ?Object,
  isFullPeriod: boolean
) => Dispatchable;

export function loadData(
  dataLayoutType: DataLayoutType,
  id: ?string,
  startTimeMs: number,
  stopTimeMs: number,
  rollup: string,
  requestType: RequestType,
  extraArguments: ?Object,
  isFullPeriod: boolean
): Dispatchable {

  return (dispatch: Dispatcher, getState: Function) => {
    const idParam = {};
    const idName = idNameFromDataLayoutType(dataLayoutType);
    if (idName) {
      idParam[idName] = id;
    }
    const extraArgs = extraArguments ? extraArguments : {};

    const params = {
      startTimeMs,
      stopTimeMs,
      rollup,
      type: requestType,
      ...idParam,
      ...extraArgs
    };

    const clusterId = String(getSelectedClusterIdSelector('MONITORING')(getState()));

    const reqPath = constructMonitoringDataRequest(clusterId, dataLayoutType, params);

    dispatch(incrementConnectionCount());
    return getJson(reqPath)
      .then(data => {
        dispatch(
          makeAction(
            RECEIVE_DATA,
            { requestType, data, isFullPeriod }
          )
        );
        dispatch(
          setPageScrollInfo(
            window.scrollY,
            window.document.body.scrollHeight,
            window.innerHeight
          )
        );
      })
      .catch((e) => {
        log(`error in loadData: ${e}`);
        dispatch(disallowDataFetch());
        setTimeout(
          () => {
            dispatch(allowDataFetch());
          }, MINIMUM_WAIT_AFTER_ERROR_MS
        );
      })
      .then(() => {
        dispatch(decrementConnectionCount());
      });
  };

}
