// @flow
// (Copyright) [2016 - 2016] Confluent, Inc.
import { connect } from 'react-redux';
import React, { Component } from 'react';
import { createSelector } from 'reselect';
import { sumArrayUntil } from 'utils';
import { Tooltip } from 'components/tooltip';

import styles from 'components/aggregate-stats/AggregateStats.css';

import {
  producedCountDataSelector,
  consumedCountDataSelector,
  ignoreDeliveryDiscrepancyAfterIndexSelector
} from 'selectors/monitoring/data';

import type { TimeSeries } from 'types/monitoring/data';


export function calculateAggregateConsumption(
  producedCounts: TimeSeries,
  consumedCounts: TimeSeries,
  ignoreAfterIndex: ?number
): ?number {

  if (!producedCounts || !consumedCounts || !ignoreAfterIndex) {
    return null;
  }

  const totalProduced = sumArrayUntil(producedCounts.values, ignoreAfterIndex);
  const totalConsumed = sumArrayUntil(consumedCounts.values, ignoreAfterIndex);

  return totalProduced > 0
    ? totalConsumed / totalProduced * 100
    : null;
}

const aggregateConsumptionSelector = createSelector(
  [consumedCountDataSelector, producedCountDataSelector, ignoreDeliveryDiscrepancyAfterIndexSelector],
  (consumedCounts: Object, producedCounts: Object, ignoreAfterIndex: ?number): ?number => {
    return calculateAggregateConsumption(producedCounts, consumedCounts, ignoreAfterIndex);
  }
);


export function calculateConsumptionWarning(
  producedCounts: TimeSeries,
  consumedCounts: TimeSeries,
  ignoreAfterIndex: ?number
): ?string {

  if (!producedCounts || !consumedCounts || !ignoreAfterIndex) {
    return null;
  }

  // Check to see if there are any instances of over or underconsumption in the current data
  let overageDetected = false;
  let lossDetected = false;

  for (let i=0; i<producedCounts.values.length && i<ignoreAfterIndex; ++i) {
    const pVal = producedCounts.values[i];
    const cVal = consumedCounts.values[i];
    if (pVal == null || cVal == null) {
      continue;
    }
    if (pVal > cVal) {
      lossDetected = true;
      if (overageDetected) {
        break;
      }
    } else if (pVal < cVal) {
      overageDetected = true;
      if (lossDetected) {
        break;
      }
    }
  }

  // If needed, set a proper warning message to be displayed in the tooltip
  let warnMessage = '';
  if (overageDetected && lossDetected) {
    warnMessage = 'Over and under consumption have been detected during this time window.';
  } else if (overageDetected && !lossDetected) {
    warnMessage = 'Over consumption has been detected during this time window.';
  } else if (!overageDetected && lossDetected) {
    warnMessage = 'Under consumption has been detected during this time window.';
  }

  return (sumArrayUntil(producedCounts.values, ignoreAfterIndex) > 0) && (overageDetected || lossDetected)
    ? warnMessage
    : null;
}

const consumptionWarningSelector = createSelector(
  [consumedCountDataSelector, producedCountDataSelector, ignoreDeliveryDiscrepancyAfterIndexSelector],
  (consumedCounts: Object, producedCounts: Object, ignoreAfterIndex: ?number): ?string => {
    return calculateConsumptionWarning(producedCounts, consumedCounts, ignoreAfterIndex);
  }
);


export class AggregateConsumption extends Component {
  props: {
    aggregateConsumption: ?number,
    consumptionWarning: ?string
  };

  render() {
    return (
      <div>
        {this.props.aggregateConsumption != null &&
          <div className={styles[this.props.consumptionWarning ? 'item-warn' : 'item']}>
            {this.props.consumptionWarning && (
              <span className={styles['tip']}>
                <Tooltip tipBody={this.props.consumptionWarning}>
                  <span className={styles['tip-inner']}>!</span>
                </Tooltip>
              </span>
            )}
            <span className={styles['item-stat']}>{`${this.props.aggregateConsumption.toFixed(2)}%`}</span>
            <span>Overall completion</span>
          </div>
        }
      </div>
    );
  }
}


function mapStateToProps(state) {
  return {
    aggregateConsumption: aggregateConsumptionSelector(state),
    consumptionWarning: consumptionWarningSelector(state)
  };
}

export default connect(mapStateToProps)(AggregateConsumption);
