// @flow
// (Copyright) [2016 - 2016] Confluent, Inc.
import React, { Component } from 'react';
import { connect } from 'react-redux';
import moment from 'moment';
import { Link } from 'react-router';
import { FlexColumn } from 'react-virtualized';
import { TableContentPanel } from 'components/layout';
import { CellOverflowHidden } from 'components/table';
import HistoryItemModal from './HistoryItemModal.jsx';
import {
  alertsHistory,
  alertsPrependNewHistory
} from 'actions/alerts';
import { HISTORY_INDEX_PATHNAME } from 'constants/alerts';
import { parseTimeGuid } from 'utils/alerts';
import type { AlertSummary } from 'types/alerts';
import type { Dispatchable } from 'types';
import type { FetchHistorySpecType } from 'actions/alerts';
import {
  mostRecentHistoryItemSelector,
  earliestHistoryItemSelector
} from 'selectors/alerts';


export class History extends Component {

  props: {
    routeParams: { timeguid?: string },
    history: Array<AlertSummary>,
    newCount: number,
    fetchHistoryPeriodMs: number,
    mostRecentItem: AlertSummary,
    earliestItem: AlertSummary,
    requestHistory: (spec: FetchHistorySpecType) => Dispatchable,
    alertsPrependNewHistory: () => Dispatchable
  }

  _historyIntervalId: number;

  static defaultProps = {
    fetchHistoryPeriodMs: 4000
  };

  fetchHistory() {
    this.props.requestHistory({
      type: 'NEW',
      itemId: null,
      count: 40
    });
  }

  handleNotificationClick() {
    this.props.alertsPrependNewHistory();
  }

  handleHitBottom() {
    if (this.props.earliestItem) {
      this.props.requestHistory({
        type: 'MAIN',
        itemId: this.props.earliestItem,
        count: 40
      });
    }
  }

  componentDidMount() {
    if (this.props.routeParams != null && this.props.routeParams.timeguid != null) {
      const timeguid = parseTimeGuid(this.props.routeParams.timeguid);
      if (timeguid === null) {
        return;
      }
      this.props.requestHistory({
        type: 'SPECIFIC',
        itemId: { timestamp: timeguid.timestamp, guid: timeguid.guid },
        count: 1
      });
    }

    this.props.requestHistory({
      type: 'MAIN',
      itemId: null,
      count: 40
    });

    this._historyIntervalId = setInterval(
      this.fetchHistory.bind(this),
      this.props.fetchHistoryPeriodMs
    );
  }

  componentWillUnmount() {
    clearInterval(this._historyIntervalId);
  }

  render() {
    const rowGetter = ({ index }) => this.props.history[index];
    return (
      <div>
        <TableContentPanel
          rowData={this.props.history}
          notificationText={
            this.props.newCount === 0
              ? null
              : `${this.props.newCount} new alert${this.props.newCount === 1 ? '' : 's'}`
          }
          onNotificationClick={this.handleNotificationClick.bind(this)}
          onHitBottom={this.handleHitBottom.bind(this)}
        >
          <FlexColumn
            label="Time"
            dataKey="timestamp"
            rowGetter={rowGetter}
            width={110}
            cellRenderer={(item) => moment(item.cellData).fromNow()}
          />
          <FlexColumn
            label="Trigger"
            dataKey="triggerName"
            rowGetter={rowGetter}
            width={50}
            flexGrow={1}
            cellRenderer={CellOverflowHidden}
          />
          <FlexColumn
            label="Action Count"
            dataKey="actionCount"
            rowGetter={rowGetter}
            width={100}
          />
          <FlexColumn
            dataKey="link"
            cellRenderer={(item) => <Link to={`${HISTORY_INDEX_PATHNAME}/${item.cellData}`}>view</Link>}
            width={50}
          />
        </TableContentPanel>
        {this.props.routeParams.timeguid
          ? <HistoryItemModal itemId={this.props.routeParams.timeguid} />
          : null}
      </div>
    );
  }
}

export default connect(
  (state) => ({
    // TODO(mhowlett): consider using memoizing here.
    history: state.alerts.history.main.map(item => (
      {
        link: `${item.timestamp}-${item.guid}`,
        timestamp: item.timestamp,
        triggerName: item.monitoringTrigger.name,
        actionCount: item.actions.length
      }
    )),
    newCount: state.alerts.history.new.length,
    mostRecentItem: mostRecentHistoryItemSelector(state),
    earliestItem: earliestHistoryItemSelector(state)
  }),
  {
    requestHistory: alertsHistory.request,
    alertsPrependNewHistory
  }
)(History);
