// @flow
// (Copyright) [2016 - 2016] Confluent, Inc.
import React, { Component } from 'react';

import MouseCatcher from '../util/MouseCatcher';
import styles from './ButtonPopup.css';

export const Alignments = {
  LEFT: 'LEFT',
  RIGHT: 'RIGHT'
};

export type Alignment = $Enum<typeof Alignments>;

export type ButtonPopupProps = {
  rootElementId: string,
  children?: any,
  alignment: Alignment,
  flag: boolean,
  label: string,
  popupHeight: number,
  onMenuClosed?: () => void,
  onMenuOpened?: () => void
};

// ButtonPopup encapsulates the following functionality:
//  - a small circular button that responds to mouse clicks.
//    - which (optionally) contains a label in the middle of the circle.
//    - the flag parameter allows the color of the circle to act as a warning signal.
//  - functionality related to displaying / hiding a popup:
//    - this popup displayed is the child of the element.
//    - clicking anywhere outside the popup area closes the popup.
//    - the popup location can either be aligned left or right of the circle.
export class ButtonPopup extends Component {

  props: ButtonPopupProps;
  state: {
    isOpen: boolean
  };

  outerDiv: HTMLElement;

  constructor() {
    super();

    this.state = {
      isOpen: false
    };
  }

  handleClick(ev: SyntheticMouseEvent) {
    if (ev.button !== 0) {
      // ignore click if not left button.
      return;
    }

    if (this.state.isOpen) {
      if (this.props.onMenuClosed) {
        this.props.onMenuClosed();
      }
    } else {
      if (this.props.onMenuOpened) {
        this.props.onMenuOpened();
      }
    }

    this.setState({
      isOpen: !this.state.isOpen
    });
  }

  close() {
    this.setState({ isOpen: false });
  }

  handleCatcherClick() {
    if (this.props.onMenuClosed) {
      this.props.onMenuClosed();
    }
    this.setState({
      isOpen: false
    });
  }

  render() {
    const { children, alignment, flag, label, popupHeight } = this.props;
    const { isOpen } = this.state;

    const buttonSize = 18;
    return (
      <div ref={el => this.outerDiv = el}>
        {isOpen
          ? <div className={styles['popup']}>
              <div
                className={
                  styles[alignment === Alignments.LEFT ? 'tip-left' : 'tip-right'] +
                  (flag ? ` ${styles['flag']}` : '')
                }
              >
                <span className={styles[alignment === Alignments.LEFT ? 'tip-inner-left' : 'tip-inner-right']}>{label}</span>
              </div>
              <div className={styles.positioning} style={{ left: `${buttonSize/2}px`, top: `${-popupHeight/2}px` }}>
                <MouseCatcher rootElementId={this.props.rootElementId} onClose={this.handleCatcherClick.bind(this)}>
                  <div className={styles[alignment === Alignments.LEFT ? 'popup-left' : 'popup-right']}>
                    {children}
                  </div>
                </MouseCatcher>
              </div>
            </div>
          : <div className={styles['button-root']}>
              <div className={styles.circle + (flag ? ` ${styles['flag']}` : '')} onMouseDown={this.handleClick.bind(this)}>
                <div className={styles[alignment === Alignments.LEFT ? 'circle-inner-left' : 'circle-inner-right']}>{label}</div>
              </div>
            </div>
        }
      </div>
    );
  }

}

export default ButtonPopup;
