// @flow
// (Copyright) [2016 - 2016] Confluent, Inc.

import React, { Component } from 'react';
import { timeFormat } from 'd3-time-format';

import styles from './Axis.css';

import type {
  TimeScale,
  TimeSpan
} from 'types/monitoring';

import {
  dropLeadingZero,
  formatTime
} from 'utils';

const utcOffsetMs = (new Date()).getTimezoneOffset()*60*1000;

const tickFormatFunction = (dateMs: number): string =>
  ((dateMs-utcOffsetMs) % 86400000) === 0 // day boundary
    ? timeFormat('%b %e')(dateMs).toLowerCase()
    : dropLeadingZero(timeFormat('%I:%M%p')(dateMs).toLowerCase());

const mouseOverFormatFunction = formatTime;

const calculateTicksCountHint = (scale: Object): number =>
  Math.min(6, Math.floor((scale.range()[1] - scale.range()[0])/100));

const transparentSeparation = 60;
const opaqueSeparation = 75;
const tickOpacity = (x1, x2) =>
  Math.max(0,
    Math.min(1,
      (Math.abs(x1 - x2) - transparentSeparation) / (opaqueSeparation - transparentSeparation)));

type AxisDefaultProps = {
  mouseOverTime: ?Date,
  missingRegions: Array<Object>,
  disabled: boolean,
  styleClassName: string
};

export class Axis extends Component {
  props: {
    xScale: TimeScale,
    mouseOverTime: ?number,
    missingRegions: Array<TimeSpan>,
    disabled: boolean,
    styleClassName: string
  };

  static defaultProps: AxisDefaultProps = {
    mouseOverTime: null,
    missingRegions: [],
    disabled: false,
    styleClassName: 'axis-styling'
  };

  render() {
    const { xScale, mouseOverTime, missingRegions } = this.props;
    const ticks = xScale.ticks(calculateTicksCountHint(xScale))
      .filter(date => date.getSeconds() === 0) // TODO: work out best way to format sub 1 minute tick dates.
      .map(date => ({ date, screenX: xScale(date) }));

    const axisStylingMarginLeft = 10;
    const mouseOverX = mouseOverTime ? xScale(mouseOverTime): null;
    return (
      <div className={this.props.disabled ? styles['root-disabled'] : styles.root}>
        <div className={styles[this.props.styleClassName]}>
          {
            missingRegions.map((region, i) => {
              const leftPx = xScale(region.startTime);
              const widthPx = xScale(region.endTime) - leftPx;
              return (
                <div
                  key={i}
                  style={{ left: `${leftPx-axisStylingMarginLeft}px`, width: `${widthPx}px` }}
                  className={styles['missing-data']}
                />
              );
            })
          }
        </div>
        {
          ticks.map(
            (tick) => {
              const opacity = mouseOverX
                ? tickOpacity(mouseOverX, tick.screenX)
                : 1.0;
              return (
                <div
                  key={tick.date.getTime()}
                  className={styles['axis-label']}
                  style={{ left: tick.screenX - 50, top: 0, opacity: opacity }}
                >
                  <span className={styles[mouseOverTime ? 'axis-tick-label' : 'axis-tick-label-strong']}>
                    {tickFormatFunction(tick.date.getTime())}
                  </span>
                </div>
              );
            }
          )
        }
        {
          mouseOverX &&
            <div className={styles['axis-label']} style={{ left: mouseOverX - 50, top: 0 }}>
              <div className={styles['axis-tick']} />
              <span>{mouseOverFormatFunction(mouseOverTime)}</span>
            </div>
        }
      </div>
    );
  }
}

export default Axis;
