// @flow
// (Copyright) [2016 - 2016] Confluent, Inc.
import React, { Component } from 'react';
import lazyCache from 'react-lazy-cache';
import { format, formatPrefix, precisionPrefix } from 'd3-format';

import ChartArea from '../ChartArea';
import styles from './AreaLineChart.css';

import AreaLineChartInner from './AreaLineChartInner';

import type { RangeType } from 'constants/monitoring/timeframe';

import {
  constructChartYScale,
  calculateDataBounds,
  aggregateCountToSlots,
  xBoundsMemoized
} from '../../util';

import type { TimeSeries } from 'types/monitoring/data';
import type { MinMax } from 'types';

import type {
  TimeScale,
  NumberScale,
  SlotWidth,
  TimeSpan
} from 'types/monitoring';

import type {
  HoverFormatter
} from '../hoverUtil';

import { PositionFreezablePopupMenu } from 'components/charting/popup/PositionFreezablePopupMenu';
import { Alignments } from 'components/button-popup';


type AreaLineChartProps = {
  lineData: ?TimeSeries,
  areaData: ?TimeSeries,
  height: number,
  xScale: TimeScale,
  slotWidth: SlotWidth,
  legend: ?any,
  mouseOverTime: ?Date,
  hoverFormatter: ?HoverFormatter,
  timeframeRangeType?: RangeType,
  leftOverlayItems?: any,
  rightOverlayItems?: any,
  solidLeadingGuide?: boolean,
  missingRegions: Array<TimeSpan>,
  disabled: boolean,
  onFocusComponent?: Function
};

type AreaLineChartDefaultProps = {
  missingRegions: Array<TimeSpan>
};

const aggregateDataToSlots = (
  data: ?TimeSeries,
  slotWidth: SlotWidth,
  xScale: TimeScale
): ?TimeSeries => {
  return aggregateCountToSlots(data ? data : null, slotWidth.ms, xBoundsMemoized(xScale));
};

const defaultFormatter = format('3.3s');

export function createAxisCountTextFormatter(gridLines: Array<Object>): (num: number) => string {
  return (gridLines.length >= 2)
    ? (() => {
      const highest = gridLines[gridLines.length-1].y;
      const separation = highest - gridLines[gridLines.length-2].y;
      const pfx = precisionPrefix(separation, highest);
      return (num) => {
        return num === 0
          ? '0'
          : formatPrefix(`.${pfx}`, highest)(num);
      };
    })()
    : defaultFormatter;
}

export class AreaLineChart extends Component {

  props: AreaLineChartProps;

  cache: {
    componentWillReceiveProps: (nextProps: AreaLineChartProps) => void,
    legendVisible: boolean,
    titleVisible: boolean,
    lineSlotData: ?TimeSeries,
    areaSlotData: ?TimeSeries,
    yScale: NumberScale,
    timeframeRangeType?: RangeType
  };

  static defaultProps: AreaLineChartDefaultProps = {
    missingRegions: []
  };

  handleFocus(mousePageX: number, mousePageY: number, mouseComponentX: number, mouseComponentY: number) {
    const { mouseOverTime, onFocusComponent } = this.props;

    if (onFocusComponent != null && this.props.hoverFormatter != null) {

      if (mouseOverTime == null) {
        return;
      }

      const hoverInfo = this.props.hoverFormatter(
        this.cache.lineSlotData,
        this.cache.areaSlotData,
        undefined,
        mouseOverTime,
        this.props.xScale,
        this.cache.yScale);

      if (!hoverInfo) {
        return;
      }

      const { ptX, ptY } = hoverInfo;

      const mouseOverTimeSlotQuantized = this.props.xScale.invert(ptX);

      const adjX =
        this.props.xScale(mouseOverTime) -
        this.props.xScale(mouseOverTimeSlotQuantized);

      const adjY =
        ptY - mouseComponentY;

      onFocusComponent(mousePageX - adjX, mousePageY + adjY);
    }
  }

  componentWillMount() {
    this.cache = lazyCache(this, {
      legendVisible: {
        params: ['legend'],
        fn: (legend: any): boolean => legend != null
      },
      titleVisible: {
        params: ['title'],
        fn: (title: any): boolean => title !== null
      },
      lineSlotData: {
        params: ['lineData', 'slotWidth', 'xScale'],
        fn: aggregateDataToSlots
      },
      areaSlotData: {
        params: ['areaData', 'slotWidth', 'xScale'],
        fn: aggregateDataToSlots
      },
      yScale: {
        params: ['lineSlotData', 'areaSlotData', 'legendVisible', 'titleVisible', 'height'],
        fn: (lineSlotData: ?TimeSeries, areaSlotData: ?TimeSeries, legendVisible: boolean, titleVisible: boolean, height: number): NumberScale => {
          const pMinMax: MinMax = calculateDataBounds(lineSlotData ? lineSlotData.values : []);
          const cMinMax: MinMax = calculateDataBounds(areaSlotData ? areaSlotData.values : []);
          const minDataY: number = Math.min(pMinMax.min, cMinMax.min);
          const maxDataY: number = Math.max(pMinMax.max, cMinMax.max);
          return constructChartYScale(legendVisible, height, minDataY, maxDataY);
        }
      }
    });
  }

  componentWillReceiveProps(nextProps: AreaLineChartProps) {
    this.cache.componentWillReceiveProps(nextProps);
  }

  render() {

    const {
      height,
      xScale,
      slotWidth,
      legend,
      mouseOverTime,
      hoverFormatter,
      timeframeRangeType,
      leftOverlayItems,
      rightOverlayItems,
      solidLeadingGuide,
      missingRegions,
      disabled
    } = this.props;

    const {
      lineSlotData,
      areaSlotData,
      yScale
    } = this.cache;

    function calcOverlayPosY(index: number) {
      if (lineSlotData) {
        if (lineSlotData.values[index] == null) {
          return yScale.range()[0];
        }
        return yScale(lineSlotData.values[index]);
      }
      if (areaSlotData) {
        if (areaSlotData.values[index] == null) {
          return yScale.range()[0];
        }
        return yScale(areaSlotData.values[index]);
      }
      return yScale.range()[0];
    }

    const leftOverlayPos = leftOverlayItems
      ? { x: xScale.range()[0],
          y: calcOverlayPosY(0)
        }
      : null;

    const rightOverlayPos = rightOverlayItems
      ? { x: xScale.range()[1],
          y: calcOverlayPosY(lineSlotData ? lineSlotData.values.length-1 : 0)
        }
      : null;

    return (
      <div className={disabled ? styles['root-disabled'] : styles.root}>
        <ChartArea
          xScale={xScale}
          yScale={yScale}
          componentHeight={height}
          legend={legend}
          lineSlotData={lineSlotData}
          areaSlotData={areaSlotData}
          hoverFormatter={hoverFormatter}
          mouseOverTime={mouseOverTime}
          solidLeadingGuide={solidLeadingGuide}
          tickTextFormatterCreator={createAxisCountTextFormatter}
          onFocusComponent={this.handleFocus.bind(this)}
        >
          <g>
            {(lineSlotData || areaSlotData) &&
              <AreaLineChartInner
                lineSlotData={lineSlotData}
                areaSlotData={areaSlotData}
                xScale={xScale}
                yScale={yScale}
                slotWidth={slotWidth}
                timeframeRangeType={timeframeRangeType}
                missingRegions={missingRegions}
              />
            }
          </g>
        </ChartArea>
        {leftOverlayPos &&
          <PositionFreezablePopupMenu
            x={leftOverlayPos.x}
            y={leftOverlayPos.y}
            items={leftOverlayItems}
            alignment={Alignments.LEFT}
          />
        }
        {rightOverlayPos &&
          <PositionFreezablePopupMenu
            x={rightOverlayPos.x}
            y={rightOverlayPos.y}
            items={rightOverlayItems}
            alignment={Alignments.RIGHT}
          />
        }
      </div>
    );
  }
}

export default AreaLineChart;
