// @flow
// (Copyright) [2016 - 2016] Confluent, Inc.
import React from 'react';

import PureComponent from 'components/util/PureComponent';

import FillGradient, { GradientType } from '../FillGradient';
import styles from '../ChartArea.css';

import {
  timeSeriesSlice,
  computeIgnoreDeliveryDiscrepancyAfterIndex
} from 'utils/monitoring/data';

import { assert, rangeLimit } from 'utils';

import type {
  TimeSeries
} from 'types/monitoring/data';

import type {
  TimeScale,
  NumberScale,
  SlotWidth,
  TimeSpan
} from 'types/monitoring';

import type { RangeType } from 'constants/monitoring/timeframe';
import { RangeTypes } from 'constants/monitoring/timeframe';


const PATH_AREA = Symbol('PATH_AREA');
const PATH_LINE = Symbol('PATH_LINE');

function constructPath(
  xScale: TimeScale,
  yScale: NumberScale,
  pathType: Symbol,
  slotWidthPx: number,
  data: TimeSeries
): string {
  const startScreenX = xScale(data.startTimeMs);
  const indexToScreenX = (offset: number): number => startScreenX + slotWidthPx*(offset);

  if (!data || data.values.length === 0) {
    return '';
  }

  let firstIncludedPointIndex = -1;
  let linePath = [];
  for (let i = 0; i<data.values.length; ++i) {
    const missingValue = data.values[i] === null || data.values[i] === undefined;
    if (missingValue && firstIncludedPointIndex < 0) {
      continue;
    }
    const value = data.values[i] || 0;
    if (firstIncludedPointIndex < 0 && value > 0) {
      firstIncludedPointIndex = i;
    }
    linePath.push(`${indexToScreenX(i)},${yScale(value)} `);
    linePath.push(`${indexToScreenX(i + 1)},${yScale(value)} `);
  }
  let lastPoint = undefined;
  linePath = linePath.reduce((accum, point) => {
    if (point !== lastPoint) {
      accum.push(point);
    }
    lastPoint = point;
    return accum;
  }, []).join('');
  if (pathType === PATH_LINE) {
    return linePath;
  }

  const areaPath = data.values.length === 0 || firstIncludedPointIndex < 0
    ? linePath
    : `${indexToScreenX(firstIncludedPointIndex)},${yScale.range()[0]} ${linePath} ${indexToScreenX(data.values.length)},${yScale.range()[0]}`;

  return areaPath;
}

function splitOnMissing(lineData: TimeSeries, areaData: TimeSeries, missingRegions: Array<TimeSpan>) {
  const notMissing = [];
  const missing = [];
  const length = lineData.values.length;
  const startTimeMs = lineData.startTimeMs;
  const stepMs = lineData.stepMs;

  let previousEndIndex = 0;
  missingRegions.forEach((timeSpan: TimeSpan) => {
    // Note: explicit integer alignment should not actually be necessary here.
    const startIndex = rangeLimit(Math.floor((timeSpan.startTime.getTime() - startTimeMs) / stepMs), length);
    const endIndex = rangeLimit(Math.floor((timeSpan.endTime.getTime() - startTimeMs) / stepMs), length);

    if (previousEndIndex !== startIndex) {
      notMissing.push({
        lineData: timeSeriesSlice(lineData, previousEndIndex, startIndex),
        areaData: timeSeriesSlice(areaData, previousEndIndex, startIndex)
      });
    }

    missing.push({
      lineData: timeSeriesSlice(lineData, startIndex, endIndex),
      areaData: timeSeriesSlice(areaData, startIndex, endIndex)
    });

    previousEndIndex = endIndex;
  });

  if (previousEndIndex < length) {
    notMissing.push({
      lineData: timeSeriesSlice(lineData, previousEndIndex, length),
      areaData: timeSeriesSlice(areaData, previousEndIndex, length)
    });
  }

  return {
    notMissing,
    missing
  };
}

function splitContiguousNonzeroSegments(data: TimeSeries): Array<TimeSeries> {
  const out = [];
  let open = false;
  let startTimeMs: number = 0;
  let startIndex: number = 0;
  data.values.forEach((datum: ?number = 0, i: number): void => {
    const currentTimeMs = data.startTimeMs + (i * data.stepMs);
    if (!datum) {
      if (open) {
        open = false;
        out.push({
          startTimeMs: startTimeMs,
          stepMs: data.stepMs,
          values: data.values.slice(startIndex, i)
        });
      }
    } else if (!open) {
      open = true;
      startTimeMs = currentTimeMs;
      startIndex = i;
    }
  });
  if (open) {
    out.push({
      startTimeMs: startTimeMs,
      stepMs: data.stepMs,
      values: data.values.slice(startIndex)
    });
  }
  return out;
}


function selectWhereDifferent(expectedSeries: TimeSeries, actualSeries: TimeSeries, ignoreDiffForLast: number): TimeSeries {
  assert(
    expectedSeries.values.length === actualSeries.values.length,
    'selectWhereDifferent expecting similar sized Timeseries'
  );
  return {
    startTimeMs: expectedSeries.startTimeMs,
    stepMs: expectedSeries.stepMs,
    values: expectedSeries.values.map((expected: ?number, i: number) => {
      const expectedVal = expected || 0;
      const actual = actualSeries.values[i] || 0;
      const diff = expectedVal - actual;
      return (diff === 0 || i >= expectedSeries.values.length - ignoreDiffForLast) ? 0 : expectedVal;
    })
  };
}

const selectAreaDataWhereDifferent = (lineData, areaData, ignoreFailureForLast) => selectWhereDifferent(areaData, lineData, ignoreFailureForLast);
const selectLineDataWhereDifferent = (lineData, areaData, ignoreFailureForLast) => selectWhereDifferent(lineData, areaData, ignoreFailureForLast);

export type AreaLineChartInnerProps = {
  lineSlotData: TimeSeries,
  areaSlotData: TimeSeries,
  xScale: TimeScale,
  yScale: NumberScale,
  slotWidth: SlotWidth,
  timeframeRangeType?: RangeType,
  missingRegions: Array<TimeSpan>
}

export class AreaLineChartInner extends PureComponent {
  props: AreaLineChartInnerProps;

  // TODO(mhowlett): this is to solve a problem whereby the linear gradient is not showing correctly after
  // navigating between pages, even though the correct DOM elements exist. Maybe this will not be a problem
  // on upgrading to React v15?
  static gradientIdCountHack: number = 0;

  render() {
    const { lineSlotData, areaSlotData, xScale, yScale, slotWidth, timeframeRangeType, missingRegions } = this.props;

    AreaLineChartInner.gradientIdCountHack += 1;

    const ignoreFailureForLast = timeframeRangeType
      ? (() => {
          const ignoreAfterIndex = computeIgnoreDeliveryDiscrepancyAfterIndex(
            lineSlotData,
            areaSlotData,
            timeframeRangeType === RangeTypes.ROLLING
              ? null
              : new Date()
          );
          return lineSlotData.values.length - ignoreAfterIndex - 1;
        })()
      : 0;

    const { notMissing, missing } = splitOnMissing(lineSlotData, areaSlotData, missingRegions);

    const constructLine = constructPath.bind(undefined, xScale, yScale, PATH_LINE, slotWidth.px);
    const constructArea = constructPath.bind(undefined, xScale, yScale, PATH_AREA, slotWidth.px);

    const renderAreaData = (lSlotData, aSlotData, isMissing, isLast) =>
      [
        aSlotData &&
        splitContiguousNonzeroSegments(aSlotData).map((splitSlotData) => (
          <polyline
            key={`success-fill-${splitSlotData.startTimeMs}`}
            points={constructArea(splitSlotData)}
            fill={`url(#${(isMissing ? GradientType.missing : GradientType.expected) + AreaLineChartInner.gradientIdCountHack})`}
          />
        )),

        lSlotData && aSlotData &&
        splitContiguousNonzeroSegments(
          selectAreaDataWhereDifferent(lSlotData, aSlotData, isLast ? ignoreFailureForLast : 0)).map((splitSlotData) => (
            <polyline
              key={`error-fill-${splitSlotData.startTimeMs}`}
              points={constructArea(splitSlotData)}
              fill={`url(#${(isMissing ? GradientType.missing : GradientType.failure) + AreaLineChartInner.gradientIdCountHack})`}
            />
        ))
      ];

    const renderLineData = (lSlotData, aSlotData, isMissing, isLast) =>
      [
        lSlotData &&
        splitContiguousNonzeroSegments(lSlotData).map((splitSlotData) => (
          <polyline
            key={`success-line-${splitSlotData.startTimeMs}`}
            points={constructLine(splitSlotData)}
            className={styles[isMissing ? 'line-style-missing' : 'line-style']}
          />
        )),
        lSlotData && aSlotData &&
        splitContiguousNonzeroSegments(
          selectLineDataWhereDifferent(lSlotData, aSlotData, isLast ? ignoreFailureForLast : 0)).map((splitSlotData) => (
            <polyline
              key={`error-line-${splitSlotData.startTimeMs}`}
              points={constructLine(splitSlotData)}
              className={styles[isMissing ? 'line-style-missing' : 'diff-line-style']}
            />
        ))
      ];

    return (
      <g>
        <defs>
          <FillGradient type={GradientType.expected} idCountHack={AreaLineChartInner.gradientIdCountHack} />
          <FillGradient type={GradientType.failure} idCountHack={AreaLineChartInner.gradientIdCountHack} />
          <FillGradient type={GradientType.missing} idCountHack={AreaLineChartInner.gradientIdCountHack} />
        </defs>
        {notMissing.map((data, i, arr) => renderAreaData(notMissing[i].lineData, notMissing[i].areaData, false, i===arr.length-1))}
        {missing.map((data, i, arr) => renderAreaData(missing[i].lineData, missing[i].areaData, true, i===arr.length-1))}
        {notMissing.map((data, i, arr) => renderLineData(notMissing[i].lineData, notMissing[i].areaData, false, i === arr.length-1))}
        {missing.map((data, i, arr) => renderLineData(missing[i].lineData, missing[i].areaData, true, i === arr.length-1))}
      </g>
    );
  }
}

export default AreaLineChartInner;
