// @flow
// (Copyright) [2016 - 2016] Confluent, Inc.
import React, { Component } from 'react';
import lazyCache from 'react-lazy-cache';
import styles from './RangeChart.css';

import RangeChartInner from './RangeChartInner';
import ChartArea from '../ChartArea';

import {
  constructChartYScale,
  calculateDataBounds,
  xBoundsMemoized,
  aggregateTotalCountMinMaxToSlots
} from '../../util';

import type {
  MultiTimeSeries,
  LatencyData,
  RawLatencyData
} from 'types/monitoring/data';

import type { MinMax } from 'types';

import type {
  TimeScale,
  NumberScale,
  SlotWidth,
  TimeSpan
} from 'types/monitoring';

import { latencyFormatter } from 'components/charting/charts/hoverUtil';
import type { HoverFormatter } from 'components/charting/charts/hoverUtil';


type RangeChartProps = {
  data: ?MultiTimeSeries<RawLatencyData>,
  height: number,
  xScale: TimeScale,
  slotWidth: SlotWidth,
  legend: any,
  mouseOverTime: ?Date,
  hoverFormatter?: ?HoverFormatter,
  missingRegions: Array<TimeSpan>,
  disabled: boolean,
  onFocusComponent?: Function
};

type RangeChartDefaultProps = {
  missingRegions: Array<TimeSpan>
}

export function createAxisDurationTextFormatter(): (num: number) => string {
  return (num) => {
    if (num === 0) {
      return '0';
    }
    if (num < 1000) {
      return `${num}ms`;
    }
    return `${num/1000}s`;
  };
}

export class RangeChart extends Component {

  props: RangeChartProps;
  cache: any;

  static defaultProps: RangeChartDefaultProps = {
    missingRegions: [],
    hoverFormatter: latencyFormatter
  };

  handleFocus(mousePageX: number, mousePageY: number, mouseComponentX: number, mouseComponentY: number) {
    const { mouseOverTime, onFocusComponent, hoverFormatter } = this.props;

    if (onFocusComponent) {

      if (mouseOverTime == null) {
        return;
      }

      if (!hoverFormatter) {
        return null;
      }

      const hoverInfo = hoverFormatter(
        undefined,
        undefined,
        this.cache.slotData,
        mouseOverTime,
        this.props.xScale,
        this.cache.yScale);

      if (!hoverInfo) {
        return;
      }

      const { ptX, ptY } = hoverInfo;

      const mouseOverTimeSlotQuantized = this.props.xScale.invert(ptX);

      const adjX =
        this.props.xScale(mouseOverTime) -
        this.props.xScale(mouseOverTimeSlotQuantized);

      const adjY =
        ptY - mouseComponentY;

      onFocusComponent(mousePageX - adjX, mousePageY + adjY);
    }
  }

  componentWillMount() {
    this.cache = lazyCache(this, {
      legendVisible: {
        params: ['legend'],
        fn: (legend: any): boolean => legend !== null && legend !== undefined
      },
      slotData: {
        params: ['data', 'slotWidth', 'xScale'],
        fn: (data: ?MultiTimeSeries<RawLatencyData>, slotWidth: SlotWidth, xScale: TimeScale): ?MultiTimeSeries<LatencyData> => {
          return aggregateTotalCountMinMaxToSlots(data ? data : null, slotWidth.ms, xBoundsMemoized(xScale));
        }
      },
      yScale: {
        params: ['slotData', 'data', 'legendVisible', 'height'],
        fn: (slotData: ?MultiTimeSeries<LatencyData>, data: MultiTimeSeries<RawLatencyData>, legendVisible: boolean, height: number): NumberScale => {
          const { minLatency, maxLatency } = slotData ? slotData.values : { minLatency: [], maxLatency: [] };
          const lMinMax: MinMax = calculateDataBounds(data ? minLatency : []);
          const hMinMax: MinMax = calculateDataBounds(data ? maxLatency : []);
          const minDataY: number = Math.min(lMinMax.min, hMinMax.min);
          const maxDataY: number = Math.max(lMinMax.max, hMinMax.max);
          return constructChartYScale(legendVisible, height, minDataY, maxDataY);
        }
      }
    });
  }

  componentWillReceiveProps(nextProps: RangeChartProps) {
    this.cache.componentWillReceiveProps(nextProps);
  }

  render() {
    const { height, xScale, slotWidth, legend, hoverFormatter, mouseOverTime, missingRegions, disabled } = this.props;
    const { slotData, yScale } = this.cache;

    return (
      <div className={disabled ? styles['root-disabled'] : styles['root']}>
        <ChartArea
          xScale={xScale}
          yScale={yScale}
          componentHeight={height}
          legend={legend}
          title={null}
          multiSlotData={slotData}
          hoverFormatter={hoverFormatter}
          mouseOverTime={mouseOverTime}
          tickTextFormatterCreator={createAxisDurationTextFormatter}
          onFocusComponent={this.handleFocus.bind(this)}
        >
          <g>
            <RangeChartInner
              slotData={slotData}
              slotWidthPx={slotWidth.px}
              xScale={xScale}
              yScale={yScale}
              missingRegions={missingRegions}
            />
          </g>
        </ChartArea>
      </div>
    );
  }
}

export default RangeChart;
