// @flow
// (Copyright) [2016 - 2016] Confluent, Inc.
import { assert } from 'utils';
import { createSelector } from 'reselect';
import d3scale from 'd3-scale';

import { argSelectors } from 'utils/memoize';

import type {
  TimeSeries,
  MultiTimeSeries,
  RawLatencyData,
  LatencyData
} from 'types/monitoring/data';

import type { MinMax } from 'types';

import type {
  TimeScale,
  NumberScale
} from 'types/monitoring';

import {
  CHART_PADDING_TOP_LG,
  CHART_PADDING_TOP_BASE,
  CHART_PADDING_BOTTOM
} from 'constants/monitoring/layout';


// Note: the purpose of this function is really to ensure that we are not
// relying on the internals of xScale to always return the same instance
// of the domain (though it probably does). In practice, this'll actually
// be more inefficient calling this than just calling domain() on xScale.
export const xBoundsMemoized = createSelector(
  argSelectors(1),
  (xScale: TimeScale): Array<number> => xScale.domain()
);

export const aggregateCountToSlots = (
  data: ?TimeSeries,
  slotWidthMs: number,
  bounds: Array<Date>
): ?TimeSeries => {
  if (data == null) {
    return null;
  }
  const multiData = {
    startTimeMs: data.startTimeMs,
    stepMs: data.stepMs,
    values: { count: data.values }
  };
  const resampled = aggregateToSlots(multiData, slotWidthMs, bounds);
  // this case can actually never happen, but flow can't work this out.
  // flow also requires a check here for undefined to be happy (hence the ==)
  if (resampled == null) {
    return null;
  }
  return {
    startTimeMs: resampled.startTimeMs,
    stepMs: resampled.stepMs,
    values: resampled.values.count
  };
};

export const aggregateTotalCountMinMaxToSlots = (
  data: ?MultiTimeSeries<RawLatencyData>,
  slotWidthMs: number,
  bounds: Array<Date>
): ?MultiTimeSeries<LatencyData> => {
  if (data == null) {
    return null;
  }
  return aggregateToSlots(data, slotWidthMs, bounds);
};

const sumTwo = (v1: ?number, v2: ?number) => !v1 ? v2 : (!v2 ? v1 : v1 + v2);

// this method handles data in two different forms (described below)
// - not the best factoring of this functionality however it isn't
// exposed, so I've put fixing this up on the todo list.
//
// data in:      [ [...totalValues], [...countValues], [...minValues], [...maxValues] ]
// return data:  [ [...avgValues], [...minValues], [...maxValues] ]
//
// data in:      [ [...totalValues] ]
// return data:  [ [...totalValues] ]
function aggregateToSlots(data: MultiTimeSeries<Object>, slotWidthMs: number, bounds: Array<Date>) : ?MultiTimeSeries<Object> {

  const lowerBoundMs = bounds[0].getTime();
  const upperBoundMs = bounds[1].getTime();

  if (data === null) {
    return null;
  }

  const keys = Object.keys(data.values);

  assert(
    keys.length === 1 ||
    keys.length === 4,
      `MultiTimeSeries has ${keys.length} timeseries, 1 or 4 expected.`
  );

  assert(
    slotWidthMs % data.stepMs === 0,
    `slotWidthMs=${slotWidthMs} should be multiple of stepms=${data.stepMs}`
  );

  for (let i=1; i<keys.length; ++i) {
    assert(
      data.values[keys[i-1]].length === data.values[keys[i]].length,
      'MultiTimeSeries values lengths are inconsistent'
    );
  }

  const dataLength = data.values[keys[0]].length;

  const resampleCount = slotWidthMs / data.stepMs;
  const stepMs = resampleCount * data.stepMs;
  let startTimeMs = Math.ceil(data.startTimeMs / stepMs) * stepMs;

  if (startTimeMs < lowerBoundMs) {
    startTimeMs = Math.ceil(lowerBoundMs / stepMs) * stepMs;
  }

  const calcMinMax = keys.length === 4;

  const result: MultiTimeSeries<Object> = {
    stepMs,
    startTimeMs,
    values: calcMinMax
      ? { averageLatency: [], minLatency: [], maxLatency: [] }
      : { count: [] }
  };

  // excludes first aggregate bin if it's not complete.
  // includes last aggregate bin even if it's not complete.
  for (let i=(startTimeMs - data.startTimeMs) / data.stepMs;
       i<dataLength && (data.startTimeMs+i*data.stepMs < upperBoundMs);
       i += resampleCount) {
    if (calcMinMax) {
      const tot = data.values.totalLatency.slice(i, i+resampleCount).reduce(sumTwo);
      const cnt = data.values.count.slice(i, i+resampleCount).reduce(sumTwo);
      const min = Math.min.apply(null, data.values.minLatency.slice(i, i+resampleCount).filter(val => val !== null));
      const max = Math.max.apply(null, data.values.maxLatency.slice(i, i+resampleCount));
      result.values.averageLatency.push(cnt ? tot/cnt : null);
      result.values.minLatency.push(cnt ? min : null);
      result.values.maxLatency.push(cnt ? max : null);
    } else {
      const tot = data.values.count.slice(i, i+resampleCount).reduce(sumTwo);
      result.values.count.push(tot);
    }
  }

  return result;
}


export function calculateDataBounds(data: Array<?number>): MinMax {

  // $FlowFixMe
  const nonNullData: Array<number> = data.filter(val => val !== null);

  if (!nonNullData || nonNullData.length === 0) {
    return { min: Infinity, max: -Infinity };
  }

  let max: number = Math.max(...nonNullData);
  let min: number = Math.min(...nonNullData);
  const extent: number = max - min;
  max = max + extent/50.0;
  min = min - extent/10.0;
  min = Math.max(min, 0);
  return { min, max };
}

export function constructChartYScale(
  showLegend: bool,
  height: number,
  minDataY: number,
  maxDataY: number
): NumberScale {

  const paddingTop: number = showLegend
    ? CHART_PADDING_TOP_LG
    : CHART_PADDING_TOP_BASE;

  assert(height - paddingTop - CHART_PADDING_BOTTOM > 0, `chart height of ${height} is too small`);

  // Domain (Data set)
  // Range (Charting area) is set inversed, i.e relative to top of charting areaData, bottom point first
  const yScale: NumberScale = d3scale.scaleLinear()
    .domain([0, maxDataY])
    .range([height - CHART_PADDING_BOTTOM, paddingTop])
    .nice(2);

  return yScale;
}
