// @flow
// (Copyright) [2016 - 2016] Confluent, Inc.
import React, { Component } from 'react';
import { FlexColumn } from 'react-virtualized';
import { connect } from 'react-redux';
import { throttle } from 'lodash';

import { deleteConnector, loadConnectorList } from 'actions/connectors';
import { ContentPanel } from 'components/layout';
import EmptyListPrompt from 'components/EmptyListPrompt';
import { FluidTable, CellActivity, CellLink, LinkIconTypes } from 'components/table';
import { STEP1, NEW_SINK_PATHNAME, NEW_SOURCE_PATHNAME } from 'constants/management';

import type { ConnectorListState, ConnectorList, ConnectorType } from 'actions/connectors';

type ConnectorIndexProps = {
  connectorType: ConnectorType,
  connectors: ConnectorList,
  dispatchDeleteConnector?: (name: string) => void,
  isFresh: boolean,
  loadConnectorList?: () => void,
  rowHeight: number,
  showAddPrompt: boolean,
  throttleInterval: number
};

export const POLLING_INTERVAL = 1000;
export const tableHeaderHeight = 52;
export const tablePaddingHeight = 70;

const EmptyListStates = {
  SINK: {
    buttonRoute: {
      pathname: NEW_SINK_PATHNAME,
      query: {
        step: STEP1
      }
    },
    buttonText: 'Add a sink',
    header: 'You haven\'t added any sinks yet',
    text: 'Pipe data out of your cluster'
  },
  SOURCE: {
    buttonRoute: {
      pathname: NEW_SOURCE_PATHNAME,
      query: {
        step: STEP1
      }
    },
    buttonText: 'Add a source',
    header: 'You haven\'t added any sources yet',
    text: 'Bring data into your cluster'
  }
};

export class ConnectorIndex extends Component {
  props: ConnectorIndexProps;
  state: {
    height: number
  };

  static defaultProps = {
    isFresh: false,
    rowHeight: 50,
    showAddPrompt: false,
    throttleInterval: 100
  };

  pollingIntervalId: number;
  rootEl: HTMLDivElement;

  setHeight: () => void = () => {
    const {
      connectors,
      rowHeight
    } = this.props;

    const { top } = this.rootEl.getBoundingClientRect();

    const allItems = connectors.length * rowHeight + tableHeaderHeight + tablePaddingHeight;
    const spaceAvailable = window.innerHeight;

    const height = Math.max(0, Math.min(allItems, spaceAvailable - top));

    this.setState({ height });
  };

  constructor(props: ConnectorIndexProps) {
    super();
    this.state = { height: 0 };
    this.setHeight = throttle(this.setHeight.bind(this), props.throttleInterval);
  }

  poll() {
    if (this.props.loadConnectorList && !document.hidden) {
      this.props.loadConnectorList();
    }
  }

  componentWillMount() {
    this.pollingIntervalId = setInterval(this.poll.bind(this), POLLING_INTERVAL);
    window.addEventListener('resize', this.setHeight);
  }

  componentDidMount() {
    this.setHeight();
  }

  componentDidUpdate(prevProps: ConnectorIndexProps) {
    if ((prevProps.connectors.length !== this.props.connectors.length) || (prevProps.rowHeight !== this.props.rowHeight)) {
      this.setHeight();
    }
  }

  componentWillUnmount() {
    this.setHeight.cancel();
    clearInterval(this.pollingIntervalId);

    window.removeEventListener('resize', this.setHeight);
  }

  render() {
    const {
      connectorType,
      connectors,
      dispatchDeleteConnector,
      rowHeight,
      showAddPrompt
    } = this.props;

    const { height } = this.state;

    const isSink = connectorType === 'SINK';

    return (
      <div ref={el => this.rootEl = el}>
        <ContentPanel height={height}>
          {connectors.length === 0 ? (
            showAddPrompt && (
              <EmptyListPrompt {...EmptyListStates[connectorType]} />
            )
          ) : (
            <FluidTable
              rowData={connectors}
              rowHeight={rowHeight}
            >
              <FlexColumn
                label="Status"
                dataKey="status"
                cellRenderer={CellActivity}
                width={100}
              />
              <FlexColumn
                label="Active Tasks"
                dataKey="tasks"
                width={130}
              />
              <FlexColumn
                label="Name"
                dataKey="name"
                width={100}
                flexGrow={1}
              />
              {isSink && (
                <FlexColumn
                  label="Topics"
                  dataKey="topics"
                  width={100}
                  flexGrow={5}
                />
              )}
              <FlexColumn
                dataKey="name"
                cellRenderer={CellLink}
                cellDataGetter={({ rowData }) => ({
                  linkTo: `/management/connect/${isSink ? 'sinks' : 'sources'}/${encodeURIComponent(rowData.name)}/edit`,
                  icon: LinkIconTypes.EDIT,
                  text: 'Edit'
                })}
                style={{
                  textAlign: 'right'
                }}
                width={50}
              />

              <FlexColumn
                dataKey="name"
                cellRenderer={CellLink}
                cellDataGetter={({ rowData }) => ({
                  icon: LinkIconTypes.DELETE,
                  linkTo: `/management/connect/${isSink ? 'sinks' : 'sources'}`,
                  onClick: (e: SyntheticMouseEvent) => {
                    e.preventDefault();
                    if (dispatchDeleteConnector) {
                      dispatchDeleteConnector(rowData.name);
                    }
                  },
                  text: 'Delete'
                })}
                style={{
                  textAlign: 'right'
                }}
                width={60}
              />
            </FluidTable>
          )}
        </ContentPanel>
      </div>
    );
  }
}

function mapStateToProps(state: { management: { connectors: { list: ConnectorListState } } }, ownProps: Object): $Shape<ConnectorIndexProps> {
  const { connectorType } = ownProps.route;
  const { items: connectors, isFresh } = state.management.connectors.list;
  return {
    connectorType,
    connectors: connectors.filter(connector => connector.type === connectorType),
    isFresh
  };
}

const mapDispatchToProps = {
  dispatchDeleteConnector: deleteConnector,
  loadConnectorList
};

function mergeProps(stateProps, dispatchProps, ownProps) {
  const showAddPrompt = stateProps.isFresh && stateProps.connectors.length === 0;

  return {
    ...ownProps,
    ...stateProps,
    ...dispatchProps,
    showAddPrompt
  };
}


export default connect(mapStateToProps, mapDispatchToProps, mergeProps)(ConnectorIndex);
