// @flow
// (Copyright) [2016 - 2016] Confluent, Inc.
import React, { Component } from 'react';
import { connect } from 'react-redux';
import { timeFormat } from 'd3-time-format';
import InlineSVG from 'svg-inline-react';
import { get } from 'lodash';

import { assert, formatTime } from 'utils';

import styles from './AlertSetupPopover.css';

import {
  aComponentIsFocusedSelector,
  focusedComponentSelector,
  focusedConsumerGroupSelector,
  popoverPositionSelector
} from 'selectors/layout';

import { MonitoringDataTypes } from 'types/monitoring';

import {
  ComponentTypes,
  unsetFocusedComponent
} from 'actions/layout';

import downIcon from 'static/icons/arrow-down-solid.svg';
import eyeIcon from 'static/icons/sm/eye.svg';

import { Button, ButtonClassTypes } from 'components/button';

import { slotWidthFormatter } from 'utils/monitoring';

import {
  slotWidthMsSelector,
  mouseOverTimeSlotQuantizedSelector
} from 'selectors/monitoring';

import type { FocusedComponentSlotAggregatedData } from 'selectors/monitoring/data';
import { focusedChartSlotAggregatedSelector } from 'selectors/monitoring/data';

import type { FocusedSpec } from 'reducers/layout';

import { alertsPrefillNewTrigger } from 'actions/alerts';

import { getSelectedClusterIdSelector } from 'selectors/clusters';

import type { ClusterId } from 'types/clusters';


export type AlertSetupPopoverProps = {
  unsetFocusedComponent: Function,
  alertsPrefillNewTrigger: Function,
  focusedComponent: ?FocusedSpec,
  focusedConsumerGroup: ?string,
  pageX: number,
  pageY: number,
  isOpen: boolean,
  rootElementId: string,
  children?: any,
  slotWidthMs: number,
  clusterId: ClusterId,
  mouseOverTimeSlotQuantized: ?number,
  focusedChartSlotAggregatedData: ?FocusedComponentSlotAggregatedData
};

export class AlertSetupPopover extends Component {

  props: AlertSetupPopoverProps;

  handleMouseDown() {
    this.props.unsetFocusedComponent();
  }

  handlePrefillClick() {
    this.props.unsetFocusedComponent();
    this.props.alertsPrefillNewTrigger(this.props.focusedConsumerGroup, this.props.clusterId, get(this.props.focusedComponent, 'id.qualifier'));
  }

  render() {
    const {
      focusedComponent,
      focusedChartSlotAggregatedData,
      mouseOverTimeSlotQuantized
    } = this.props;

    if (!this.props.isOpen) {
      return null;
    }

    if (focusedComponent == null) {
      return null;
    }

    if (mouseOverTimeSlotQuantized == null) {
      return null;
    }

    if (focusedChartSlotAggregatedData == null) {
      return null;
    }

    const mouseOverSlotNumber = Math.floor(
      (mouseOverTimeSlotQuantized - focusedChartSlotAggregatedData.produced.startTimeMs) /
      focusedChartSlotAggregatedData.produced.stepMs
    );

    assert(
      mouseOverSlotNumber >= 0,
      'expecting mouseOverSlotNumber to be positive'
    );

    assert(
      mouseOverSlotNumber < focusedChartSlotAggregatedData.produced.values.length,
      'expecting mouseOverSlotNumber to be within range'
    );

    const description = focusedComponent.id.qualifier === MonitoringDataTypes.DELIVERY
      ? (() => {
          const numberConsumed = focusedChartSlotAggregatedData.consumed.values[mouseOverSlotNumber];
          const numberProduced = focusedChartSlotAggregatedData.produced.values[mouseOverSlotNumber];

          const consumption = numberConsumed != null && numberProduced != null && numberConsumed !== 0
            ? (numberConsumed / numberProduced) * 100.0
            : 100.0;

          return `Message consumption is ${consumption.toFixed(2)}% of expected for this consumer group here.`;
        })()
      : (() => {
          const averageLatencyMaybe = focusedChartSlotAggregatedData.latency.values.averageLatency[mouseOverSlotNumber];
          const averageLatency = averageLatencyMaybe ? averageLatencyMaybe : 0;
          const maximumLatencyMaybe = focusedChartSlotAggregatedData.latency.values.maxLatency[mouseOverSlotNumber];
          const maximumLatency = maximumLatencyMaybe ? maximumLatencyMaybe : 0;
          return `Average / maximum latency is ${Math.round(averageLatency)}ms / ${Math.round(maximumLatency)}ms for this consumer group here.`;
        })();

    // The popover has a fixed width as specified in CSS, the following const is based on that.
    // TODO(mhowlett): this will break if the page has been scaled.
    const popoverHorizontalMinSpace = 310;
    const alignmentClass = this.props.pageX > window.innerWidth - popoverHorizontalMinSpace ? 'root-right' : 'root-left';

    return (
      <div>
        <div className={styles.catcher} onMouseDown={this.handleMouseDown.bind(this)} />
        <div className={`${styles['root']} ${styles[alignmentClass]}`} style={{ left: `${this.props.pageX}px`, top: `${this.props.pageY}px` }}>
          <div className={styles.header}>
            <div className={styles.date}>
              {timeFormat('%B %d, ')(mouseOverTimeSlotQuantized) + formatTime(mouseOverTimeSlotQuantized)}
            </div>
            <div className={styles.duration}>
              {slotWidthFormatter(this.props.slotWidthMs)} window
            </div>
          </div>
          <div className={styles.body}>
            <div className={styles.arrow}>
              <InlineSVG src={downIcon} className={styles.icon} />
            </div>
            <div className={styles.description}>
              {description}
            </div>
          </div>
          <div className={styles.footer}>
            <Button
              type={ButtonClassTypes.PRIMARY}
              text={"Setup an alert"}
              icon={eyeIcon}
              onClick={this.handlePrefillClick.bind(this)}
            />
          </div>
        </div>
      </div>
    );
  }

}

export default connect(
  (state) => ({
    isOpen: aComponentIsFocusedSelector(state) && focusedComponentSelector(state).type === ComponentTypes.MONITORING_CHART,
    focusedComponent: focusedComponentSelector(state),
    focusedConsumerGroup: focusedConsumerGroupSelector(state),
    pageX: popoverPositionSelector(state).pageX,
    pageY: popoverPositionSelector(state).pageY,
    slotWidthMs: slotWidthMsSelector(state),
    mouseOverTimeSlotQuantized: mouseOverTimeSlotQuantizedSelector(state),
    focusedChartSlotAggregatedData: focusedChartSlotAggregatedSelector(state),
    clusterId: getSelectedClusterIdSelector('MONITORING')(state)
  }),
  {
    unsetFocusedComponent,
    alertsPrefillNewTrigger
  }
)(AlertSetupPopover);
