// @flow
// (Copyright) [2016 - 2016] Confluent, Inc.
import debug from 'debug';
import { connect } from 'react-redux';
import { Component } from 'react';

import { assert } from 'utils';
import {
  startTimeSlotQuantizedSelector,
  fetchStopTimeMsSelector,
  desiredRollup,
  desiredDataQuantizationMsSelector,
  visibleMembersSelector,
  chartAreaWidthPxSelector
} from 'selectors/monitoring';
import { loadData, clearData } from 'actions/monitoring/data';
import { RequestTypes } from 'types/monitoring/data';

import {
  currentDataLayoutTypeSelector,
  currentDataLayoutIdSelector,
  currentDataBinSizeMs,
  lastMemberListFetchTimeMs,
  lastOverviewDataFetchTimeMs,
  dataSelector
} from 'selectors/monitoring/data';

import type { RequestType, DataLayoutType } from 'types/monitoring/data';
import type { LoadDataFunction } from 'actions/monitoring/data';
import type { RangeType } from 'constants/monitoring/timeframe';
import type { Rollup } from 'types/monitoring';


const log = debug('controlcenter:components:monitoring:DataManager');

export const MAX_CONNECTION_COUNT = 4;

export type DataManagerProps = {
  startTimeMs: number,
  fetchStopTimeMs: number,
  loadData: LoadDataFunction,
  clearData: Function,
  routedDataLayoutId: ?string,
  routedDataLayoutType: ?DataLayoutType,
  currentDataLayoutId: ?string,
  currentDataLayoutType: ?DataLayoutType,
  rangeType: RangeType,
  currentDataBinSizeMs: ?number,
  desiredRollup: Rollup,
  desiredDataQuantizationMs: number,
  chartAreaWidthPx: number,
  preventFetch: boolean,
  lastMemberListFetchTimeMs: number,
  lastOverviewDataFetchTimeMs: number,
  visibleMembers: { startIndex: number, endIndex: number },
  data: ?Object,
  connectionCount: number,

  // interval at which we see if a new data fetch (or more)
  // should be initiated, and initiate them.
  cyclePeriodMs: number,

  // interval at which the list of members should be re-acquired.
  memberListRefreshMs: number,

  // interval at which full data set is requested from server.
  // Applies to all timeframe update modes.
  fullRefreshMs: number,

  // when partial data requested, the length of time (ending
  // in current time) that is requested.
  updateWindowSizeMs: number,

  // maximum frequency at which (partial) data is requested.
  maxFetchIntervalMs: number
}


export function computePriority(
  memberIndex: number,
  partialUpdateEpoch?: number,
  fullUpdateEpoch?: number,
  visibleMembers: ?Object,
  fullRefreshMs: number,
  maxFetchIntervalMs: number,
  requestType: RequestType
) {
  assert(
    (partialUpdateEpoch == null && fullUpdateEpoch == null) ||
    (partialUpdateEpoch != null && fullUpdateEpoch != null &&
     partialUpdateEpoch >= fullUpdateEpoch),
    'partialUpdateEpoch should always be greater than fullUpdateEpoch'
  );

  if (fullUpdateEpoch != null && partialUpdateEpoch == null) {
    log('invalid state: partialUpdateEpoch is null and fullUpdateEpoch isn\'t');
    return 0;
  }

  if (partialUpdateEpoch && fullUpdateEpoch) {
    if (Date.now() - partialUpdateEpoch < maxFetchIntervalMs) {
      return 0;
    }
  }

  const offscreenRenderWidth = 2;

  let score = 0;

  if (visibleMembers &&
      memberIndex >= visibleMembers.startIndex - offscreenRenderWidth &&
      memberIndex <= visibleMembers.endIndex + offscreenRenderWidth) {

    if (memberIndex >= visibleMembers.startIndex &&
        memberIndex <= visibleMembers.endIndex) {
      // visible members get automatic priority boost.
      score += 50;
    }

    if (fullUpdateEpoch == null) {
      // never obtained any data - increase priority quite a lot.
      score += 100;
    } else {
      // already have data. priorty to re-get depends on how long ago we last fetched
      // If a partial update is all that is required, then priority is lower.
      if (fullUpdateEpoch < Date.now() - fullRefreshMs) {
        score += (Date.now() - fullUpdateEpoch)/1000;
      } else {
        // $FlowFixMe - flow believes partialUpdateEpoch can be undefined, but that it can't is enforced above.
        score += ((Date.now() - partialUpdateEpoch)/1000)/2;
      }
    }

    if (requestType === RequestTypes.OVERVIEW) {
      // there is a preference for overview data first on first page load.
      score += 1;
    }
  }

  return score;
}


export class DataManager extends Component {
  props: DataManagerProps;

  _dataFetchIntervalId: number;
  _retrieveCount: number;

  static defaultProps = {
    cyclePeriodMs: 300,
    memberListRefreshMs: 30000,
    maxFetchIntervalMs: 2000,
    updateWindowSizeMs: 60000,
    fullRefreshMs: 30000
  };

  createPrioritizedTaskList(): Array<Object> {

    const {
      data,
      visibleMembers,
      fullRefreshMs,
      maxFetchIntervalMs,
      startTimeMs
    } = this.props;

    const priorityList = [];

    // Priority #1 is making sure there is a member list and that it is up to date.

    if (this.props.currentDataLayoutType === null ||
        !this.props.lastMemberListFetchTimeMs ||
        this.props.lastMemberListFetchTimeMs < Date.now() - this.props.memberListRefreshMs) {
      priorityList.push(
        {
          type: RequestTypes.MEMBER_LIST,
          priority: Number.MAX_SAFE_INTEGER
        }
      );
    }

    // Now prioritize member + overview data.

    if (data != null) {

      const priority = computePriority(
        -1,
        data.overview ? data.overview.partialEpoch : undefined,
        data.overview ? data.overview.fullEpoch : undefined,
        visibleMembers,
        fullRefreshMs,
        maxFetchIntervalMs,
        RequestTypes.OVERVIEW
      );
      if (priority > 0) {
        priorityList.push({
          type: RequestTypes.OVERVIEW,
          priority,
          fullRefresh: data.overview
            ? (Date.now() - data.overview.fullEpoch) > fullRefreshMs ||
              data.overview.startTimeMs > startTimeMs
            : true
        });
      }

      if (data.members) {
        data.members.forEach((member, i) => {
          const priority = computePriority(
            i,
            member.partialEpoch,
            member.fullEpoch,
            visibleMembers,
            fullRefreshMs,
            maxFetchIntervalMs,
            RequestTypes.MEMBER
          );
          if (priority > 0) {
            priorityList.push({
              type: RequestTypes.MEMBER,
              priority,
              member,
              fullRefresh: member.fullEpoch
                ? (Date.now() - member.fullEpoch) > fullRefreshMs ||
                  member.startTimeMs > startTimeMs
                : true
            });
          }
        });
      }

    }

    const highPriorityFirstComparitor =
      (t1, t2) => t1.priority < t2.priority ? 1 : (t1.priority === t2.priority ? 0 : -1);

    priorityList.sort(highPriorityFirstComparitor);

    return priorityList;
  }


  fetchNextData() {

    const {
      currentDataBinSizeMs,
      routedDataLayoutType,
      routedDataLayoutId
    } = this.props;

    // 1. first check if we are in a state where we want to fetch data at all.
    if (this.props.preventFetch) {
      return;
    }

    if (!routedDataLayoutType) {
      return;
    }

    if (this.props.connectionCount >= MAX_CONNECTION_COUNT) {
      return;
    }

    // 2. now check to see whether data format has changed such that we should
    //    ditch the whole lot and start again from scratch.
    if ((routedDataLayoutType !== this.props.currentDataLayoutType ||
         routedDataLayoutId !== this.props.currentDataLayoutId)) {
      if (this.props.currentDataLayoutType !== null) {
        this.props.clearData();
        return;
      }
    }

    if (currentDataBinSizeMs !== this.props.desiredDataQuantizationMs) {
      if (currentDataBinSizeMs != null) {
        this.props.clearData();
        return;
      }
    }

    // 3. if neither of those case, get a prioritized list of tasks and initiate.
    const tasks = this.createPrioritizedTaskList();

    const numberToInitiate = MAX_CONNECTION_COUNT - this.props.connectionCount;
    for (let i=0; i<numberToInitiate; ++i) {
      if (i > tasks.length-1) {
        break;
      }
      const task = tasks[i];

      let isFullRefresh;
      let fetchStartTimeMs;
      if (task.fullRefresh || currentDataBinSizeMs == null) {
        fetchStartTimeMs = this.props.startTimeMs;
        isFullRefresh = true;
      } else {
        fetchStartTimeMs =
          Math.floor(
            (this.props.fetchStopTimeMs - this.props.updateWindowSizeMs)/currentDataBinSizeMs
          ) * currentDataBinSizeMs;
        isFullRefresh = false;
        if (fetchStartTimeMs < this.props.startTimeMs) {
          fetchStartTimeMs = this.props.startTimeMs;
          isFullRefresh = true;
        }
      }

      switch (task.type) {
        case RequestTypes.MEMBER_LIST:
          this.props.loadData(
            routedDataLayoutType,
            routedDataLayoutId,
            this.props.startTimeMs,
            this.props.fetchStopTimeMs,
            this.props.desiredRollup,
            RequestTypes.MEMBER_LIST,
            null,
            true
          );
          break;
        case RequestTypes.OVERVIEW:
          this.props.loadData(
            routedDataLayoutType,
            routedDataLayoutId,
            fetchStartTimeMs,
            this.props.fetchStopTimeMs,
            this.props.desiredRollup,
            RequestTypes.OVERVIEW,
            null,
            isFullRefresh
          );
          break;
        case RequestTypes.MEMBER:
          const memberData = {};
          if (task.member.partition !== undefined) {
            memberData.memberPartition = task.member.partition;
          }
          if (task.member.topic !== undefined) {
            memberData.memberTopic = task.member.topic;
          }
          if (task.member.group !== undefined) {
            memberData.memberGroup = task.member.group;
          }
          if (task.member.clientId !== undefined) {
            memberData.memberClientId = task.member.clientId;
          }
          this.props.loadData(
            routedDataLayoutType,
            routedDataLayoutId,
            fetchStartTimeMs,
            this.props.fetchStopTimeMs,
            this.props.desiredRollup,
            RequestTypes.MEMBER,
            memberData,
            isFullRefresh
          );
          break;
        default:
          assert(
            false,
            `unknown request type: ${task.type}`
          );
          break;
      }
    }
  }

  componentWillMount() {
    this._retrieveCount = 0;
    this._dataFetchIntervalId = setInterval(
      this.fetchNextData.bind(this),
      this.props.cyclePeriodMs
    );
  }

  componentWillUnmount() {
    clearInterval(this._dataFetchIntervalId);
    this.props.clearData();
  }

  render(): any {
    return null;
  }
}

function mapStateToProps(state, props) {
  return {
    startTimeMs: startTimeSlotQuantizedSelector(state).getTime(),
    fetchStopTimeMs: fetchStopTimeMsSelector(state),
    currentDataLayoutId: currentDataLayoutIdSelector(state),
    currentDataLayoutType: currentDataLayoutTypeSelector(state),
    routedDataLayoutType: props.route.routedDataLayoutType,
    routedDataLayoutId: props.routeParams.id === undefined ? null : props.routeParams.id,
    rangeType: state.monitoring.timeframe.rangeType,
    currentDataBinSizeMs: currentDataBinSizeMs(state),
    desiredDataQuantizationMs: desiredDataQuantizationMsSelector(state),
    desiredRollup: desiredRollup(state),
    chartAreaWidthPx: chartAreaWidthPxSelector(state),
    preventFetch: state.monitoring.dataManager.preventFetch,
    connectionCount: state.monitoring.dataManager.connectionCount,
    visibleMembers: visibleMembersSelector(state),
    lastMemberListFetchTimeMs: lastMemberListFetchTimeMs(state),
    lastOverviewDataFetchTimeMs: lastOverviewDataFetchTimeMs(state),
    data: dataSelector(state)
  };
}

export default connect(
  mapStateToProps,
  {
    loadData,
    clearData
  }
)(DataManager);
