// @flow
// (Copyright) [2016 - 2016] Confluent, Inc.
import React, { Component } from 'react';

import Axis from 'components/charting/Axis';
import AreaLineChart from 'components/charting/charts/area-line-chart';
import { connect } from 'react-redux';
import DeliveryLegend from 'components/charting/legend/DeliveryLegend';
import { dataSelector, currentDataLayoutTypeSelector } from 'selectors/monitoring/data';
import LatencyLegend from 'components/charting/legend/LatencyLegend';
import { produceConsumeFormatter } from 'components/charting/charts/hoverUtil';
import RangeChart from 'components/charting/charts/range-chart';

import type { RangeType } from 'constants/monitoring/timeframe';

import type { Dispatchable } from 'types';

import {
  OVERVIEW_MESSAGE_DELIVERY_CHART_HEIGHT,
  OVERVIEW_LATENCY_CHART_HEIGHT
} from 'constants/monitoring/layout';

import {
  chartXScaleSelector,
  slotWidth,
  mouseOverTimeSlotQuantizedSelector
} from 'selectors/monitoring';

import type {
  TimeScale,
  SlotWidth,
  MonitoringDataType
} from 'types/monitoring';

import type {
  DataLayoutType,
  TimeSeries,
  MultiTimeSeries
} from 'types/monitoring/data';

import {
  ComponentTypes,
  setFocusedComponent,
  setPopoverPosition
} from 'actions/layout';

import { focusedComponentSelector } from 'selectors/layout';

import { MonitoringDataTypes } from 'types/monitoring';
import type { ComponentType } from 'actions/layout';
import { DataLayoutTypes } from 'types/monitoring/data';

import {
  isChartDisabled,
  isAxisDisabled
} from './DeliveryLatencyMemberCharts';

import type { FocusedSpec } from 'reducers/layout';


export type DeliveryLatencyChartsProps = {
  data: ?Object,
  xScale: TimeScale,
  slotWidth: SlotWidth,
  mouseOverTime: ?Date,
  mouseOverTimeSlotQuantized: ?Date,
  timeframeRangeType: RangeType,
  producedCounts: ?TimeSeries,
  consumedCounts: ?TimeSeries,
  consumedLatency: ?MultiTimeSeries<*>,
  topChartHeight: number,
  bottomChartHeight: number,
  setFocusedComponent?: (type: ComponentType, id: Object) => Dispatchable,
  setPopoverPosition?: (pageX: number, pageY: number) => Dispatchable,
  routedDataLayoutType: ?DataLayoutType,
  focusedComponent: ?FocusedSpec
}

export class DeliveryLatencyCharts extends Component {
  props: DeliveryLatencyChartsProps;

  static defaultProps = {
    topChartHeight: OVERVIEW_MESSAGE_DELIVERY_CHART_HEIGHT,
    bottomChartHeight: OVERVIEW_LATENCY_CHART_HEIGHT
  };

  handleChartFocus(qualifier: MonitoringDataType, pageX: number, pageY: number) {
    const { setFocusedComponent, setPopoverPosition } = this.props;

    if (this.props.routedDataLayoutType !== DataLayoutTypes.CONSUMER_GROUP &&
        this.props.routedDataLayoutType !== DataLayoutTypes.CONSUMER_GROUP_TOPICS) {
      return;
    }

    if (setFocusedComponent && setPopoverPosition) {
      setPopoverPosition(pageX, pageY);
      setFocusedComponent(ComponentTypes.MONITORING_CHART, { componentId: -1, qualifier: qualifier });
    }
  }

  render() {

    const {
      data,
      xScale,
      slotWidth,
      mouseOverTime,
      mouseOverTimeSlotQuantized,
      timeframeRangeType,
      topChartHeight,
      bottomChartHeight,
      focusedComponent
    } = this.props;

    const missingRegions = data && data.overview ? data.overview.missing : [];

    return (
      <div>
        {data && data.overview && data.overview.produced && data.overview.consumed &&
         data.overview.produced.count && data.overview.consumed.count
          ? <AreaLineChart
              height={topChartHeight}
              xScale={xScale}
              slotWidth={slotWidth}
              lineData={data.overview.produced.count}
              areaData={data.overview.consumed.count}
              legend={<DeliveryLegend slotWidthMs={slotWidth.ms} />}
              mouseOverTime={mouseOverTime}
              hoverFormatter={produceConsumeFormatter}
              timeframeRangeType={timeframeRangeType}
              leftOverlayItems={data.destinations}
              rightOverlayItems={data.sources}
              solidLeadingGuide={true}
              missingRegions={missingRegions}
              onFocusComponent={this.handleChartFocus.bind(this, MonitoringDataTypes.DELIVERY)}
              disabled={isChartDisabled(focusedComponent, -1, MonitoringDataTypes.DELIVERY)}
            />
          : <div style={{ height: `${topChartHeight}px` }} />
        }
        <Axis
          xScale={xScale}
          mouseOverTime={mouseOverTimeSlotQuantized}
          missingRegions={missingRegions}
          disabled={isAxisDisabled(focusedComponent, -1)}
        />
        {data && data.overview && data.overview.consumed && data.overview.consumed.latency
          ? <RangeChart
              height={bottomChartHeight}
              xScale={xScale}
              slotWidth={slotWidth}
              data={data.overview.consumed.latency}
              legend={<LatencyLegend slotWidthMs={slotWidth.ms} />}
              mouseOverTime={mouseOverTime}
              missingRegions={missingRegions}
              onFocusComponent={this.handleChartFocus.bind(this, MonitoringDataTypes.LATENCY)}
              disabled={isChartDisabled(focusedComponent, -1, MonitoringDataTypes.LATENCY)}
            />
          : <div style={{ height: `${bottomChartHeight}px` }} />
        }
      </div>
    );
  }
}

function mapStateToProps(state) {
  return {
    data: dataSelector(state),
    xScale: chartXScaleSelector(state),
    slotWidth: slotWidth(state),
    mouseOverTime: state.monitoring.layout.mouseOverTime,
    mouseOverTimeSlotQuantized: mouseOverTimeSlotQuantizedSelector(state),
    timeframeRangeType: state.monitoring.timeframe.rangeType,
    focusedComponent: focusedComponentSelector(state),
    routedDataLayoutType: currentDataLayoutTypeSelector(state)
  };
}

export default connect(
  mapStateToProps,
  {
    setFocusedComponent,
    setPopoverPosition
  }
)(DeliveryLatencyCharts);
