// @flow
// (Copyright) [2016 - 2016] Confluent, Inc.
import { connect } from 'react-redux';
import debug from 'debug';
import React, { Component } from 'react';

import styles from './DeliveryLatencyMemberCharts.css';

import AreaLineChart from 'components/charting/charts/area-line-chart';
import RangeChart from 'components/charting/charts/range-chart';
import Axis from 'components/charting/Axis';
import { MemberHeader } from './MemberHeader';

import { produceConsumeFormatter } from 'components/charting/charts/hoverUtil';
import { dataSelector } from 'selectors/monitoring/data';
import { DataLayoutTypes } from 'types/monitoring/data';
import { memberIdNameFromDataLayoutType } from 'utils/monitoring/data';

import type { Dispatchable } from 'types';

import type { DataLayoutType } from 'types/monitoring/data';
import type { RangeType } from 'constants/monitoring/timeframe';

import {
  MEMBER_MESSAGE_DELIVERY_CHART_HEIGHT,
  MEMBER_LATENCY_CHART_HEIGHT
} from 'constants/monitoring/layout';

import {
  chartXScaleSelector,
  slotWidth,
  mouseOverTimeSlotQuantizedSelector
} from 'selectors/monitoring';

import {
  ComponentTypes,
  setFocusedComponent,
  setPopoverPosition
} from 'actions/layout';

import type {
  TimeScale,
  SlotWidth,
  MonitoringDataType
} from 'types/monitoring';

import { focusedComponentSelector } from 'selectors/layout';

import {
  MonitoringDataTypes,
  MemberTitleIconTypes
} from 'types/monitoring';
import type { ComponentType } from 'actions/layout';

import type { FocusedSpec } from 'reducers/layout';


const log = debug('controlcenter:components:monitoring:DeliveryLatencyMemberCharts');

export const isChartDisabled = (focusedComponent: ?FocusedSpec, chartIdx: number, qualifier: MonitoringDataType) => {
  if (focusedComponent == null) {
    return false;
  }
  return (
    focusedComponent.type === ComponentTypes.MONITORING_CHART &&
    !(focusedComponent.id.componentId === chartIdx && focusedComponent.id.qualifier === qualifier)
  );
};

export const isAxisDisabled = (focusedComponent: ?FocusedSpec, chartIdx: number) => {
  if (focusedComponent == null) {
    return false;
  }
  return (
    focusedComponent.type === ComponentTypes.MONITORING_CHART &&
    focusedComponent.id.componentId !== chartIdx
  );
};

export type DeliveryLatencyMemberChartsProps = {
  data: ?Object,
  xScale: TimeScale,
  slotWidth: SlotWidth,
  mouseOverTime: ?Date,
  mouseOverTimeSlotQuantized: ?Date,
  timeframeRangeType: RangeType,
  routedDataLayoutType: DataLayoutType,
  topChartHeight: number,
  bottomChartHeight: number,
  focusedComponent: ?FocusedSpec,
  setFocusedComponent?: (type: ComponentType, id: Object) => Dispatchable,
  setPopoverPosition?: (pageX: number, pageY: number) => Dispatchable
}

export class DeliveryLatencyMemberCharts extends Component {
  props: DeliveryLatencyMemberChartsProps;

  static defaultProps = {
    topChartHeight: MEMBER_MESSAGE_DELIVERY_CHART_HEIGHT,
    bottomChartHeight: MEMBER_LATENCY_CHART_HEIGHT
  };

  handleChartFocus(qualifier: MonitoringDataType, index: number, pageX: number, pageY: number) {
    const { setPopoverPosition, setFocusedComponent } = this.props;

    if (this.props.routedDataLayoutType !== DataLayoutTypes.CONSUMER_GROUPS &&
        this.props.routedDataLayoutType !== DataLayoutTypes.CONSUMER_GROUPS_TOPICS) {
      return;
    }

    if (setFocusedComponent && setPopoverPosition) {
      setPopoverPosition(pageX, pageY);
      setFocusedComponent(ComponentTypes.MONITORING_CHART, { componentId: index, qualifier: qualifier });
    }
  }

  render() {
    const {
      data,
      xScale,
      slotWidth,
      mouseOverTime,
      mouseOverTimeSlotQuantized,
      timeframeRangeType,
      routedDataLayoutType,
      topChartHeight,
      bottomChartHeight,
      focusedComponent
    } = this.props;

    if (!data || !data.members) {
      return null;
    }

    if (routedDataLayoutType !== data.type) {
      return null;
    }
    const memberIdName = memberIdNameFromDataLayoutType(routedDataLayoutType);

    const {
      latencyDataGetter,
      lineDataGetter,
      areaDataGetter,
      missingRegionGetter,
      titleGetter,
      pathGetter,
      iconType
    } = (memberIdName.length > 0)
      ? { iconType: MemberTitleIconTypes.CONSUMER_GROUP,
          latencyDataGetter: member => member.consumed ? member.consumed.latency : null,
          areaDataGetter: member => member.consumed ? member.consumed.count : null,
          lineDataGetter: member => member.produced ? member.produced.count : null,
          missingRegionGetter: member => member.missing ? member.missing : [],
          pathGetter: member => (() => {
            switch (routedDataLayoutType) {
              case DataLayoutTypes.CONSUMER_GROUP:
                return `/monitoring/consumers/${member.clientId}`;
              case DataLayoutTypes.CONSUMER_GROUPS:
                 return `/monitoring/consumer-groups/${member.group}`;
              default:
                return null;
            }
          })(),
          titleGetter: member => memberIdName.map((id, idx) => idx === 0 ? member[id] : `${id}: ${member[id]}`)
        }
      : (() => {
          log(`unexpected data layout type: ${routedDataLayoutType}`);
          return {
            iconType: null,
            latencyDataGetter: () => null,
            areaDataGetter: () => null,
            lineDataGetter: () => null,
            missingRegionGetter: () => [],
            titleGetter: () => null,
            pathGetter: () => null
          };
        })();

    return (
      <div className={styles['items']}>
        {
          data.members.map((member, i) =>
            <div key={i} className={styles.item}>
              <MemberHeader title={titleGetter(member)} iconType={iconType} detailPath={pathGetter(member)} disabled={this.props.focusedComponent} />
              <AreaLineChart
                height={topChartHeight}
                xScale={xScale}
                slotWidth={slotWidth}
                areaData={areaDataGetter(member)}
                lineData={lineDataGetter(member)}
                legend={null}
                hoverFormatter={produceConsumeFormatter}
                timeframeRangeType={timeframeRangeType}
                mouseOverTime={mouseOverTime}
                leftOverlayItems={member.destinations}
                rightOverlayItems={member.sources}
                solidLeadingGuide={true}
                missingRegions={missingRegionGetter(member)}
                onFocusComponent={this.handleChartFocus.bind(this, MonitoringDataTypes.DELIVERY, i)}
                disabled={isChartDisabled(focusedComponent, i, MonitoringDataTypes.DELIVERY)}
              />
              <Axis
                xScale={xScale}
                mouseOverTime={mouseOverTimeSlotQuantized}
                missingRegions={missingRegionGetter(member)}
                disabled={isAxisDisabled(focusedComponent, i)}
              />
              <RangeChart
                height={bottomChartHeight}
                xScale={xScale}
                slotWidth={slotWidth}
                data={latencyDataGetter(member)}
                legend={null}
                mouseOverTime={mouseOverTime}
                missingRegions={missingRegionGetter(member)}
                onFocusComponent={this.handleChartFocus.bind(this, MonitoringDataTypes.LATENCY, i)}
                disabled={isChartDisabled(focusedComponent, i, MonitoringDataTypes.LATENCY)}
              />
            </div>
          )
        }
      </div>
    );
  }
}

function mapStateToProps(state, props) {
  return {
    data: dataSelector(state),
    xScale: chartXScaleSelector(state),
    slotWidth: slotWidth(state),
    mouseOverTime: state.monitoring.layout.mouseOverTime,
    mouseOverTimeSlotQuantized: mouseOverTimeSlotQuantizedSelector(state),
    routedDataLayoutType: props.route.routedDataLayoutType,
    timeframeRangeType: state.monitoring.timeframe.rangeType,
    focusedComponent: focusedComponentSelector(state)
  };
}

export default connect(
  mapStateToProps,
  {
    setFocusedComponent,
    setPopoverPosition
  }
)(DeliveryLatencyMemberCharts);

