// @flow
// (Copyright) [2016 - 2016] Confluent, Inc.
import { connect } from 'react-redux';
import React, { Component, cloneElement } from 'react';

import {
  setChartAreaWidth,
  setPageScrollInfo
} from 'actions/monitoring/layout';
import {
  updateTimeRange,
  updateEndTime,
  updateFetchExtendedEndTime
} from 'actions/monitoring/timeframe';
import { Header, Breadcrumbs } from 'components/header';
import { MonitoringDataTypes } from 'types/monitoring';
import { RangeTypes } from 'constants/monitoring/timeframe';
import { TimePicker } from 'components/time-picker';
import { consumerGroupsSelector } from 'selectors/monitoring/data';
import { chartAreaWidthPxSelector } from 'selectors/monitoring';
import { showLicenseExpiredSelector } from 'selectors/license';
import { aComponentIsFocusedSelector } from 'selectors/layout';
import ClusterSelectorDropdown from 'components/ClusterSelectorDropdown';
import LicenseExpired from 'components/LicenseExpired';
import AlertSetupPopover from './AlertSetupPopover';

import styles from './Monitoring.css';

import type { Dispatchable } from 'types';
import type { RangeType } from 'constants/monitoring/timeframe';

const aggStats = [MonitoringDataTypes.DELIVERY, MonitoringDataTypes.LATENCY];

export function timeframeSpecFromQueryParams(queryParams: Object): ?Object {
  const { s, e, g, r, growing, rolling } = queryParams;
  const fixedStart = queryParams['fixed-start'];
  const fixedEnd = queryParams['fixed-end'];

  if (rolling || r) {
    return {
      rangeType: RangeTypes.ROLLING,
      lengthMs: (rolling !== undefined ? rolling : r) * 60 * 1000
    };
  }

  if (growing || g) {
    const growingType = growing !== undefined ? growing : g;
    if (growingType === 'day') {
      return { rangeType: RangeTypes.GROWING_DAY };
    }
    if (growingType === 'week') {
      return { rangeType: RangeTypes.GROWING_WEEK };
    }
    if (growingType === 'month') {
      return { rangeType: RangeTypes.GROWING_MONTH };
    }
    return null;
  }

  if (s || fixedStart) {
    const startTimeMs = fixedStart !== undefined ? fixedStart : s;
    if (e || fixedEnd) {
      const endTimeMs = fixedEnd !== undefined ? fixedEnd : e;
      return {
        rangeType: RangeTypes.FIXED,
        startTimeMs: parseInt(startTimeMs, 10),
        stopTimeMs: parseInt(endTimeMs, 10)
      };
    }
  }

  return null;
}


export type MonitoringProps = {
  chartAreaWidthPx: number,

  dataFetchRefreshInterval: number,
  displayTickRefreshInterval: number,

  setTimeRange: (startTimeMs: number, stopTimeMs: number, rangeType: RangeType, rangeTypeRolling: number, rangeTypeGrowing: number) => Dispatchable,
  updateEndTime: (endTimeMs: number) => Dispatchable,
  setChartAreaWidth: (width: number) => Dispatchable,

  consumerGroups: ?Array<string>,

  top: any,
  bottom: any,

  params: Object,
  routes: Array<Object>,

  showLicenseExpired: boolean
}

export class Monitoring extends Component {
  monitoringContainer: HTMLElement;
  _measure: Function;
  handleScroll: Function;
  _displayTickIntervalId: number;

  static defaultProps = {
    displayTickRefreshInterval: 1000
  };

  _measure() {
    // TODO: we could calculate chartAreaWidthPx on startup as well as via the
    // resize method in this component (though it's all a bit of a hack).
    // currently, this is not set for the first two renders, screwing up
    // positioning of the labels and other things we might add. alternatively
    // we could hack the various child elements so they don't display if
    // width is 0 (currently not implemented - would prefer first alternative).

    // better still (in my opinion), the we could mange the large scale layout
    // by hand and have everything determined from window size. we only really
    // have the side nav extra to worry about (not a big deal).

    const elem = this.monitoringContainer;
    if (
      elem &&
      elem.offsetWidth !== undefined &&
      elem.offsetWidth !== this.props.chartAreaWidthPx
    ) {
      const width = elem.offsetWidth || this.props.chartAreaWidthPx;
      this.props.setChartAreaWidth(width);
    }
  }

  constructor() {
    super();
    this._measure = this._measure.bind(this);
    this.handleScroll = this.handleScroll.bind(this);
  }

  handleScroll() {
    this.props.setPageScrollInfo(
      window.scrollY,
      window.document.body.scrollHeight,
      window.innerHeight
     );
  }

  componentWillMount() {
    const timeframeSpec = timeframeSpecFromQueryParams(this.props.location.query);
    if (timeframeSpec) {
      this.props.updateTimeRange(timeframeSpec);
    }

    this.props.updateEndTime(Date.now());

    this._displayTickIntervalId = setInterval(
      () => {
        if (!this.props.aComponentIsFocused) {
          this.props.updateEndTime(Date.now());
        } else {
          this.props.updateFetchExtendedEndTime(Date.now());
        }
      },
      this.props.displayTickRefreshInterval
    );
  }

  componentDidMount() {
    window.addEventListener('resize', this._measure);
    window.addEventListener('scroll', this.handleScroll);
  }

  componentDidUpdate() {
    this._measure();
  }

  componentWillUnmount() {
    clearInterval(this._displayTickIntervalId);
    window.removeEventListener('resize', this._measure);
    window.removeEventListener('scroll', this.handleScroll);
  }

  render() {
    const {
      params,
      routes,
      location,
      consumerGroups,
      showLicenseExpired
    } = this.props;

    return (
      <div className={styles['monitoring-root']}>
        {!showLicenseExpired
          ? <div>
              {this.props.dataManager}
              <Header
                bottomLeft={(
                  <Breadcrumbs
                    routes={routes}
                    params={params}
                    location={location}
                    consumerGroups={consumerGroups}
                  />
                )}
                aggStats={aggStats}
                topRight={[
                  <ClusterSelectorDropdown
                    clusterType="MONITORING"
                    key="clusterSelector"
                  />,
                  <TimePicker key="timepicker" />
                ]}
              />
              <div ref={el => this.monitoringContainer = el} className={styles['chartarea-root']}>
                {this.props.top}
              </div>
              {
                this.props.children
                  ? cloneElement(this.props.bottom, {}, this.props.children)
                  : this.props.bottom
              }
              <AlertSetupPopover />
            </div>
          : <LicenseExpired />
        }
      </div>
    );
  }
}

function mapStateToProps(state) {
  return {
    chartAreaWidthPx: chartAreaWidthPxSelector(state),
    // although not explicitly required, this forces the component to update
    // when data changes, which is required to measure the width correctly.
    data: state.monitoring.data,
    startTime: state.monitoring.timeframe.startTimeMs/1000,
    endTime: state.monitoring.timeframe.stopTimeMs/1000,
    rangeType: state.monitoring.timeframe.rangeType,
    consumerGroups: consumerGroupsSelector(state),
    showLicenseExpired: showLicenseExpiredSelector(state),
    aComponentIsFocused: aComponentIsFocusedSelector(state)
  };
}

export default connect(
  mapStateToProps,
  {
    updateTimeRange,
    updateEndTime,
    updateFetchExtendedEndTime,
    setChartAreaWidth,
    setPageScrollInfo
  }
)(Monitoring);
