// @flow
// (Copyright) [2016 - 2016] Confluent, Inc.
import React, { Component } from 'react';
import { throttle, debounce } from 'lodash';

import Pill from './Pill';
import PopoverMenu from 'components/PopoverMenu';
import styles from './PillBox.css';

export type PillBoxProps = {
  children?: any, // Pills only
  hasFlexibleWidth: boolean,
  maxVisiblePillGuestimate: number
};

class PillBox extends Component {
  props: PillBoxProps;
  state: {
    lastAvailableWidth: number | null,
    visiblePillCount: number,
    measuringStatus: 'INIT' | 'INCREASING' | 'DECREASING' | 'DONE'
  };

  static defaultProps = {
    maxVisiblePillGuestimate: 3,
    hasFlexibleWidth: false
  };

  rootEl: HTMLDivElement;
  innerEl: HTMLDivElement;
  handleResize: Function;
  measureFit: Function;

  constructor(props: PillBoxProps) {
    super(props);

    this.state = {
      lastAvailableWidth: null,
      visiblePillCount: Math.min(React.Children.count(props.children), props.maxVisiblePillGuestimate),
      measuringStatus: 'INIT'
    };

    this.handleResize = throttle(this.handleResize.bind(this), 1000 / 60);
    this.measureFit = debounce(this.measureFit.bind(this));
  }

  handleResize() {
    const { hasFlexibleWidth } = this.props;
    const availableWidth = this.rootEl.offsetWidth;

    // note: the lastAvailableWidth optimization won't work if the width is flexible
    // from such factors as parent width using % or media queries
    if (hasFlexibleWidth || availableWidth !== this.state.lastAvailableWidth) {
      this.setState({
        lastAvailableWidth: availableWidth,
        measuringStatus: 'INIT',
        visiblePillCount: Math.max(
          Math.min(this.props.maxVisiblePillGuestimate, React.Children.count(this.props.children)),
          this.state.visiblePillCount
        )
      });
    }
  }

  measureFit() {
    const { visiblePillCount, measuringStatus } = this.state;

    if (measuringStatus === 'DONE') {
      return;
    }

    if (visiblePillCount === 0) {
      this.setState({
        measuringStatus: 'DONE'
      });
      return;
    }

    const innerWidth = this.innerEl.offsetWidth;
    const availableWidth = this.rootEl.offsetWidth;

    const isSpaceAvailable = innerWidth < availableWidth;

    const childCount = React.Children.count(this.props.children);

    if (isSpaceAvailable) {
      switch (measuringStatus) {
        case 'INIT':
          this.setState({
            measuringStatus: visiblePillCount === childCount ? 'DONE' : 'INCREASING'
          });
          break;
        case 'INCREASING':
          if (visiblePillCount === childCount) {
            this.setState({
              measuringStatus: 'DONE'
            });
          } else {
            this.setState({
              visiblePillCount: Math.min(visiblePillCount + this.props.maxVisiblePillGuestimate, childCount)
            });
          }
          break;
        case 'DECREASING':
          this.setState({
            measuringStatus: 'DONE'
          });
          break;
      }
    } else {
      switch (measuringStatus) {
        case 'INIT':
          this.setState({
            measuringStatus: 'DECREASING'
          });
          break;
        case 'INCREASING':
          this.setState({
            measuringStatus: 'DECREASING'
          });
          break;
        case 'DECREASING':
          this.setState({
            visiblePillCount: visiblePillCount - 1
          });
          break;
      }
    }
  }

  componentDidMount() {
    this.measureFit();
    window.addEventListener('resize', this.handleResize);
  }

  componentDidUpdate() {
    this.measureFit();
  }

  componentWillUnmount() {
    window.removeEventListener('resize', this.handleResize);
    this.handleResize.cancel();
    this.measureFit.cancel();
  }

  render() {
    const { children } = this.props;
    const { measuringStatus, visiblePillCount } = this.state;

    const totalPillCount = React.Children.count(children);
    const renderMore = visiblePillCount < totalPillCount;
    const childrenArray = React.Children.toArray(children);

    return (
      <div
        className={styles.root}
        style={{ visibility: measuringStatus === 'DONE' ? null : 'hidden' }}
        ref={el => this.rootEl = el}
      >
        <div className={styles.inner} ref={el => this.innerEl = el}>
          {renderMore ? childrenArray.slice(0, visiblePillCount) : children}
          {renderMore && (
            <div className={styles.morePill}>
              <PopoverMenu
                menuItems={childrenArray.slice(visiblePillCount).map(pill => ({
                  label: pill.props.children
                }))}
              >
                <Pill>
                  {visiblePillCount === totalPillCount ? `${totalPillCount} items` : `+ ${totalPillCount - visiblePillCount} more`}
                </Pill>
              </PopoverMenu>
            </div>
          )}
        </div>
      </div>
    );
  }
}


export default PillBox;
