// @flow
// (Copyright) [2016 - 2016] Confluent, Inc.
import React, { Component } from 'react';
import cn from 'classnames';
import { AutoSizer, FlexTable, FlexColumn, ScrollSync } from 'react-virtualized';
import { findDOMNode } from 'react-dom';
import { throttle } from 'lodash';

import styles from './BigDataTable.css';


const additionalColumnHeightRequired = 19;

const dividerHeaderRenderer = ({ label }) => (
  <div className={styles['column-header']}>
    {label}
  </div>
);

const dividerColumn = ({ isFirst = false, groupName, rowHeader = false }: {
  isFirst: boolean,
  groupName: string,
  rowHeader: boolean
}) => {
  return (
    <FlexColumn
      key={groupName}
      rowHeader={rowHeader}
      className={styles[isFirst ? 'column-divider-cell--first' : 'column-divider-cell']}
      headerClassName={styles[isFirst ? 'column-divider-header--first' : 'column-divider-header']}
      headerRenderer={dividerHeaderRenderer}
      label={groupName}
      dataKey="none"
      flexShrink={0}
      flexGrow={0}
      width={isFirst ? 0 : 1}
    />
  );
};

type Column = {
  props: {
    columnGroup: string,
    rowHeader: boolean
  }
};

const injectColumnGroups = (newArr: Array<Column>, column: Column, index: number, origArr: Array<Column>) => {
  if (column.props.columnGroup) {
    const prevColumn = origArr[index - 1];

    const isFirst = index === 0;
    // if last is in the same column group then we don't need to do anything
    if (isFirst
      || (prevColumn.props && prevColumn.props.columnGroup !== column.props.columnGroup && typeof column.props.columnGroup !== 'undefined')
    ) {
      newArr.push(dividerColumn({
        groupName: column.props.columnGroup,
        isFirst,
        rowHeader: column.props.rowHeader
      }));
    }
  }

  newArr.push(column);

  return newArr;
};

const updateBodyColumnWidths = (rowBodyColumns, bodyAvailableWidth) => {

  if (bodyAvailableWidth > 0) {
    const flexWidth = rowBodyColumns.reduce((accum, curr) => curr.props.flexGrow + accum, 0);

    // flexGrow not specified anywhere, add more width to the last body item.
    if (flexWidth <= 0) {
      const lastIndex = rowBodyColumns.length - 1;
      const endColumn = rowBodyColumns[lastIndex];
      const updatedEndColumn = React.cloneElement(endColumn, { width: bodyAvailableWidth + endColumn.props.width });
      rowBodyColumns.splice(lastIndex, 1, updatedEndColumn);
      return rowBodyColumns;
    }

    // else, don't no action required, FlexTable will automatically handle columns with flexGrow.
  }

  return rowBodyColumns;
};


/*
 * TODO(@dylanharrington): support this:
 *
 * <BigDataTable {...}>
 *   <FlexHeaderColumnGroup name="Connector">
 *     <FlexColumn label="Status" {...} />
 *     <FlexColumn label="Name" {...} />
 *   </FlexHeaderColumnGroup>
 *   <FlexBodyColumnGroup name="Cluster">
 *     <FlexColumn label="Name" {...} />
 *   </FlexBodyColumnGroup>
 *   ...
 * </BigDataTable>
 */
export class BigDataTable extends Component {
  props: {
    bodyTableProps: Object,
    children?: any,
    headerHeight: number,
    headerTableProps: Object,
    height: number,
    rowData: Array<any>,
    rowHeight: number,
    scrollbarSize: number,
    onHitBottom?: Function
  };

  state: {
    availableWidth: number,
    scrollLeft: number
  };

  static defaultProps = {
    bodyTableProps: {},
    headerHeight: 35,
    headerTableProps: {},
    rowHeight: 40,
    scrollbarSize: 10
  };

  bodyTable: HTMLElement;

  handleScroll: () => void;
  handleAutosizerResize: (dimensions: { width: number }) => void;

  constructor() {
    super();

    this.state = {
      availableWidth: 0,
      scrollLeft: 0
    };

    this.handleScroll = throttle(this.handleScroll.bind(this), 20);
    this.handleAutosizerResize = throttle(this.handleAutosizerResize.bind(this), 20);
  }

  handleHitBottom() {
    if (this.props.onHitBottom) {
      this.props.onHitBottom();
    }
  }

  handleScroll() {
    const { scrollLeft } = this.bodyTable;
    this.setState({
      scrollLeft
    });
  }

  handleAutosizerResize({ width }: { width: number }) {
    this.setState({
      availableWidth: width
    });
  }

  componentDidMount() {
    this.bodyTable.addEventListener('scroll', this.handleScroll, ({
      passive: true
    }: any));
  }

  componentWillUnmount() {
    this.bodyTable.removeEventListener('scroll', this.handleScroll, ({
      passive: true
    }: any));
    this.handleScroll.cancel();
    this.handleAutosizerResize.cancel();
  }

  render() {
    const {
      bodyTableProps,
      children,
      headerHeight,
      headerTableProps,
      height,
      rowData,
      rowHeight,
      scrollbarSize
    } = this.props;

    const { availableWidth } = this.state;

    const childrenArr = React.Children.toArray(children);

    const hasColumnGroups = childrenArr.findIndex(el => typeof el.props.columnGroup !== 'undefined') > -1;

    let rowBodyColumns = childrenArr.filter(column => Boolean(!column.props.rowHeader));
    let rowHeaderColumns = childrenArr.filter(column => Boolean(column.props.rowHeader));

    if (hasColumnGroups) {
      rowBodyColumns = rowBodyColumns.reduce(injectColumnGroups, []);
      rowHeaderColumns = rowHeaderColumns.reduce(injectColumnGroups, []);
    }

    const accumulateWidths = (currWidth, column) => currWidth + column.props.width;

    const rowBodyWidth = rowBodyColumns.reduce(accumulateWidths, 0);
    const rowHeaderWidth = rowHeaderColumns.reduce(accumulateWidths, 0);

    const sharedProps = {
      rowWrapperClassName: ({ index }) => `Grid__cell-${(index + 1) % 2 === 0 ? 'even' : 'odd'}`,
      headerHeight: hasColumnGroups ? headerHeight + additionalColumnHeightRequired : headerHeight,
      rowGetter: ({ index }) => rowData[index],
      rowData,
      rowHeight,
      estimatedRowHeight: rowHeight,
      rowCount: rowData.length,
      overscanRowCount: 0,
      height: height - scrollbarSize
    };

    const virtualWidth = rowHeaderWidth + rowBodyWidth;
    const maxWidth = Math.max(virtualWidth, availableWidth);

    rowBodyColumns = updateBodyColumnWidths(rowBodyColumns, availableWidth - rowHeaderWidth);

    const { scrollLeft } = this.state;

    return (
      <div className={styles.root}>
        <AutoSizer
          children={() => null}
          disableHeight={true}
          onResize={this.handleAutosizerResize}
        />
        <ScrollSync>
          {({ scrollTop, onScroll }) => (
            <div
              className={cn(styles.wrapper, {
                // [styles.hiddenContentRight]: maxWidth - availableWidth > scrollLeft,
                [styles.hiddenContentLeft]: scrollLeft > 0
                // [styles.hiddenContentBelow]: scrollHeight > scrollTop + height,
                // [styles.hiddenContentAbove]: scrollTop > 0
              })}
              style={{ height }}
            >
              {rowHeaderColumns.length === 0
                ? null
                : <FlexTable
                    className={styles['header-table']}
                    onScroll={onScroll}
                    scrollTop={scrollTop}
                    {...sharedProps}
                    {...headerTableProps}
                    width={rowHeaderWidth}
                  >
                    {rowHeaderColumns}
                  </FlexTable>
              }
              <FlexTable
                ref={el => this.bodyTable = findDOMNode(el)}
                className={styles['body-table']}
                style={{
                  left: rowHeaderWidth,
                  maxWidth: availableWidth - rowHeaderWidth
                }}
                onScroll={(params) => {
                  const { clientHeight, scrollHeight, scrollTop } = params;
                  if (clientHeight + scrollTop === scrollHeight) {
                    this.handleHitBottom();
                  }
                  onScroll(params);
                }}
                scrollTop={scrollTop}
                {...sharedProps}
                {...bodyTableProps}
                width={maxWidth - rowHeaderWidth}
              >
                {rowBodyColumns}
              </FlexTable>
            </div>
          )}
        </ScrollSync>
      </div>
    );
  }
}

export default BigDataTable;
