// @flow
// (Copyright) [2016 - 2016] Confluent, Inc.
import React, { Component } from 'react';
import { connect } from 'react-redux';
import moment from 'moment';

import clockIcon from 'static/icons/sm/clock.svg';
import styles from './TimePicker.css';
import { ButtonSecondaryMedium, ButtonPrimaryMedium, ButtonClassTypes } from 'components/button';
import { DefaultRanges, Calendar, Clock, MainControls } from 'components/time-picker/index';
import { MouseCatcher } from 'components/util/MouseCatcher';
import { PickerTargets, DateTimeFormats } from 'constants/timePicker';
import { RangeTypes } from 'constants/monitoring/timeframe';
import { assert } from 'utils';
import { rootElementId } from 'constants';
import { setTimePickerVisibility } from 'actions/monitoring/layout';
import { setTimeRange, updateTimeRangeQueryParams } from 'actions/monitoring/timeframe';
import { defaultState as timeframeDefaultState } from 'reducers/monitoring/timeframe';
import { defaultState as layoutDefaultState } from 'reducers/monitoring/layout';

import type { RangeType } from 'constants/monitoring/timeframe';
import type { ButtonClassType } from 'components/button';

export const PickerModes = {
  DEFAULTS: 'DEFAULTS',
  CALENDAR: 'CALENDAR',
  CLOCK: 'CLOCK'
};

export type PickerMode = $Enum<typeof PickerModes>;

// Method used to figure out how to label a given time range
// Specifically used for labeling the picker button
export const getPickerLabel = (
  startTime: number,
  endTime: number,
  rangeType: RangeType
): string => {
  const start = moment.unix(startTime);
  const end = moment.unix(endTime);

  const startDay = moment.unix(startTime).format(DateTimeFormats.DATE);
  const endDay = moment.unix(endTime).format(DateTimeFormats.DATE);

  const startDayDisplay = moment.unix(startTime).format(DateTimeFormats.DATE_SHORT);
  const endDayDisplay = moment.unix(endTime).format(DateTimeFormats.DATE_SHORT);

  const startTimeDisplay = moment.unix(startTime).format(DateTimeFormats.TIME);
  const endTimeDisplay = moment.unix(endTime).format(DateTimeFormats.TIME);

  let label = endDayDisplay;
  let timeDifference = start.to(end, true);
  const isFullDayRange = Math.round(end.diff(start, 'minutes', true)) === 1440;

  // Manipulate timeDifference returned by moment js for all day ranges
  if (isFullDayRange) {
    if (rangeType === RangeTypes.ROLLING) {
      timeDifference = '24 hours';
    } else if (rangeType === RangeTypes.FIXED) {
      timeDifference = 'Full day';
    }
  }

  const labelPatterns = {
    simple: `${label} - ${timeDifference}`,
    relative: `${label} - Last ${timeDifference}`,
    verbose: `${startDayDisplay} - ${endDayDisplay}`,
    veryVerbose: `${startDayDisplay} ${startTimeDisplay} - ${endDayDisplay} ${endTimeDisplay}`
  };

  // Customize default timeDifference from moment
  if (startDay !== endDay) {
    // start and end days are different
    if (rangeType === RangeTypes.ROLLING) {
      label = labelPatterns.relative;
    } else if (rangeType === RangeTypes.FIXED) {
      if (startTimeDisplay === endTimeDisplay) {
        label = labelPatterns.verbose;
      } else {
        label = labelPatterns.veryVerbose;
      }
    } else if (rangeType === RangeTypes.GROWING_DAY ||
               rangeType === RangeTypes.GROWING_WEEK ||
               rangeType === RangeTypes.GROWING_MONTH) {
      label = labelPatterns.veryVerbose;
    }
  } else {
    // start and end days are the same
    if (rangeType === RangeTypes.ROLLING) {
      label = labelPatterns.relative;
    } else if (rangeType === RangeTypes.FIXED) {
      if (isFullDayRange) {
        label = labelPatterns.simple;
      } else {
        if (startTimeDisplay === endTimeDisplay) {
          label = labelPatterns.verbose;
        } else {
          label = labelPatterns.veryVerbose;
        }
      }
    } else if (rangeType === RangeTypes.GROWING_DAY ||
               rangeType === RangeTypes.GROWING_WEEK ||
               rangeType === RangeTypes.GROWING_MONTH) {
      label = labelPatterns.relative;
    }
  }

  return label;
};

// Method used to figure out how to label a given time range
export const getRangeLabel = (
  startTime: number,
  endTime: number,
  rangeType: RangeType
): string => {
  const start = moment.unix(startTime);
  const end = moment.unix(endTime);

  let label = start.to(end, true);

  // Customize default timeDifference wording from moment js
  if (rangeType === RangeTypes.FIXED) {
    if (Math.round(end.diff(start, 'days', true)) === 1) {
      // All full day fixed ranges
      if (moment().diff(start, 'days') === 0) {
        label = 'Today';
      } else if (moment().diff(start, 'days') === 1) {
        label = 'Yesterday';
      } else if (moment().diff(start, 'weeks') === 1) {
        label = 'Today last week';
      } else if (moment().diff(start, 'months') === 1) {
        label = 'Today last month';
      }
    } else if (Math.round(end.diff(start, 'weeks', true)) === 1) {
      // All full week fixed ranges
      label = 'This week';
    } else if (Math.round(end.diff(start, 'months', true)) === 1) {
      // All full month fixed ranges
      label = 'This month';
    }
  }

  if (rangeType === RangeTypes.ROLLING) {
    // All rolling ranges being with 'Last '
    // Refer to a day as '24 hours'
    if (Math.round(end.diff(start, 'hours', true)) === 24) {
      label = '24 hours';
    }
    label = `Last ${label}`;
  }

  if (rangeType === RangeTypes.GROWING_DAY) {
    assert(
      start.diff(moment().startOf('day'), 'days') === 0,
      'unexpected start time for rangetype: growingDay'
    );
    label = 'Day so far';
  } else if (rangeType === RangeTypes.GROWING_WEEK) {
    assert(
      start.diff(moment().startOf('week'), 'days') === 0,
      'unexpected start time for rangetype: growingWeek'
    );
    label = 'Week so far';
  } else if (rangeType === RangeTypes.GROWING_MONTH) {
    assert(
      start.diff(moment().startOf('month'), 'days') === 0,
      'unexpected start time for rangetype: growingMonth'
    );
    label = 'Month so far';
  }

  return label;
};

type TimePickerProps = {
  endTime: number,
  rangeType: RangeType,
  startTime: number,
  timePickerVisible: boolean,
  setTimePickerVisibility: Function,
  setTimeRange: Function,
  updateTimeRangeQueryParams: Function
};

export class TimePicker extends Component {
  props: TimePickerProps;
  state: {
    endDayButtonType: ButtonClassType,
    endTime: number,
    endTimeButtonType: ButtonClassType,
    hovered: boolean,
    hoveredEndTime: number,
    hoveredRangeType: RangeType,
    hoveredStartTime: number,
    mode: PickerMode,
    pickerLabel: string,
    rangeLabel: string,
    rangeType: RangeType,
    startDayButtonType: ButtonClassType,
    startTime: number,
    startTimeButtonType: ButtonClassType
  };

  static defaultProps = {
    endTime: timeframeDefaultState.stopTimeMs / 1000,
    rangeType: timeframeDefaultState.rangeType,
    startTime: timeframeDefaultState.startTimeMs / 1000,
    timePickerVisible: layoutDefaultState.timePickerVisible,
    setTimePickerVisibility: () => {},
    setTimeRange: () => {},
    updateTimeRangeQueryParams: () => {}
  };

  constructor(props: TimePickerProps) {
    super();
    this.state = {
      startTime: props.startTime,
      endTime: props.endTime,
      pickerLabel: getPickerLabel(props.startTime, props.endTime, props.rangeType),
      rangeLabel: getRangeLabel(props.startTime, props.endTime, props.rangeType),
      rangeType: props.rangeType,
      hoveredStartTime: props.startTime,
      hoveredEndTime: props.endTime,
      hoveredRangeType: props.rangeType,
      hovered: false,
      mode: PickerModes.DEFAULTS,  // defaults, calendar, clock
      startDayButtonType: ButtonClassTypes.SECONDARY,
      startTimeButtonType: ButtonClassTypes.SECONDARY,
      endDayButtonType: ButtonClassTypes.SECONDARY,
      endTimeButtonType: ButtonClassTypes.SECONDARY
    };
  }

  // Handles clicks on the main timepicker menu controls
  handleMenuButtonClick: Function = (target: number): void => {
    let mode = PickerModes.DEFAULTS;
    let startDayButtonType = ButtonClassTypes.SECONDARY;
    let startTimeButtonType = ButtonClassTypes.SECONDARY;
    let endDayButtonType = ButtonClassTypes.SECONDARY;
    let endTimeButtonType = ButtonClassTypes.SECONDARY;

    if (target === PickerTargets.START_DAY && this.state.startDayButtonType !== ButtonClassTypes.PRIMARY) {
      mode = PickerModes.CALENDAR;
      startDayButtonType = ButtonClassTypes.PRIMARY;
    } else if (target === PickerTargets.START_TIME && this.state.startTimeButtonType !== ButtonClassTypes.PRIMARY) {
      mode = PickerModes.CLOCK;
      startTimeButtonType = ButtonClassTypes.PRIMARY;
    } else if (target === PickerTargets.END_DAY && this.state.endDayButtonType !== ButtonClassTypes.PRIMARY) {
      mode = PickerModes.CALENDAR;
      endDayButtonType = ButtonClassTypes.PRIMARY;
    } else if (target === PickerTargets.END_TIME && this.state.endTimeButtonType !== ButtonClassTypes.PRIMARY) {
      mode = PickerModes.CLOCK;
      endTimeButtonType = ButtonClassTypes.PRIMARY;
    }

    this.setState({
      mode: mode,
      startDayButtonType: startDayButtonType,
      startTimeButtonType: startTimeButtonType,
      endDayButtonType: endDayButtonType,
      endTimeButtonType: endTimeButtonType
    });
  };

  // Puts the component back to the proper startup state
  reset: () => void = () => {
    this.setState({
      mode: PickerModes.DEFAULTS,
      startDayButtonType: ButtonClassTypes.SECONDARY,
      startTimeButtonType: ButtonClassTypes.SECONDARY,
      endDayButtonType: ButtonClassTypes.SECONDARY,
      endTimeButtonType: ButtonClassTypes.SECONDARY
    });
    this.props.setTimePickerVisibility(false);
  };

  // Handle clicks on any of the default ranges
  handleDefaultRangeClick: (
    startTime: number,
    endTime: number,
    rangeType: RangeType
  ) => void = (startTime, endTime, rangeType) => this.setState({
    startTime: startTime,
    endTime: endTime,
    rangeType: rangeType,
    pickerLabel: getPickerLabel(startTime, endTime, rangeType),
    rangeLabel: getRangeLabel(startTime, endTime, rangeType)
  });

  // Handle hovering on any of the default ranges
  handleDefaultRangeHover: (
    hoveredStartTime: number,
    hoveredEndTime: number,
    hoveredRangeType: RangeType,
    hovered: boolean
  ) => void = (hoveredStartTime, hoveredEndTime, hoveredRangeType, hovered) => this.setState({
    hovered,
    hoveredEndTime,
    hoveredRangeType,
    hoveredStartTime
  });

  handleDefaultRangeBlur: () => void = () => this.setState({
    hovered: false
  });

  handleStartDateChange: (date: number) => void = date => this.setState({
    pickerLabel: getPickerLabel(date, this.state.endTime, RangeTypes.FIXED),
    rangeLabel: getRangeLabel(date, this.state.endTime, RangeTypes.FIXED),
    rangeType: RangeTypes.FIXED,
    startTime: date
  });

  handleEndDateChange: (date: number) => void = date => this.setState({
    endTime: date,
    pickerLabel: getPickerLabel(this.state.startTime, date, RangeTypes.FIXED),
    rangeLabel: getRangeLabel(this.state.startTime, date, RangeTypes.FIXED),
    rangeType: RangeTypes.FIXED
  });

  componentWillReceiveProps(props: TimePickerProps) {
    if (!props.timePickerVisible) {
      this.setState({
        startTime: props.startTime,
        endTime: props.endTime,
        pickerLabel: getPickerLabel(props.startTime, props.endTime, props.rangeType),
        rangeLabel: getRangeLabel(props.startTime, props.endTime, props.rangeType),
        rangeType: props.rangeType
      });
    }
  }

  render() {
    return (
      <div className={styles['time-picker-root']}>
        <ButtonSecondaryMedium
          text={this.state.pickerLabel}
          icon={clockIcon}
          onClick={() => this.props.setTimePickerVisibility(!this.props.timePickerVisible)}
        />
        {this.props.timePickerVisible && (
          <MouseCatcher rootElementId={rootElementId}
            onClose={() => {
              this.props.setTimePickerVisibility(false);
            }}
          >
            <div className={styles['menu']}>
              <MainControls
                startDayButtonType={this.state.startDayButtonType}
                startTimeButtonType={this.state.startTimeButtonType}
                endDayButtonType={this.state.endDayButtonType}
                endTimeButtonType={this.state.endTimeButtonType}
                startTime={this.state.hovered
                  ? this.state.hoveredStartTime
                  : this.state.startTime
                }
                endTime={this.state.hovered
                  ? this.state.hoveredEndTime
                  : this.state.endTime
                }
                onHandleMenuButtonClick={this.handleMenuButtonClick}
              />
              <div className={styles['defaults']}>
                <div className={styles['inner']}>
                  {this.state.mode === PickerModes.DEFAULTS && (
                    <DefaultRanges
                      selectedRangeLabel={this.state.rangeLabel}
                      onHandleDefaultRangeHover={this.handleDefaultRangeHover}
                      onHandleDefaultRangeClick={this.handleDefaultRangeClick}
                      onHandleDefaultRangeBlur={this.handleDefaultRangeBlur}
                      getRangeLabel={getRangeLabel}
                    />
                  )}
                  {this.state.mode === PickerModes.CALENDAR && (
                    <div>
                      <Calendar
                        minDate={this.state.startDayButtonType === ButtonClassTypes.PRIMARY
                          ? 0
                          : this.state.startTime
                        }
                        onDateChange={this.state.startDayButtonType === ButtonClassTypes.PRIMARY
                          ? this.handleStartDateChange
                          : this.handleEndDateChange
                        }
                        currDate={this.state.startDayButtonType === ButtonClassTypes.PRIMARY
                          ? this.state.startTime
                          : this.state.endTime
                        }
                      />
                    </div>
                  )}
                  {this.state.mode === PickerModes.CLOCK && (
                    <div>
                      <Clock
                        onDateChange={this.state.startTimeButtonType === ButtonClassTypes.PRIMARY
                          ? this.handleStartDateChange
                          : this.handleEndDateChange
                        }
                        currTime={this.state.startTimeButtonType === ButtonClassTypes.PRIMARY
                          ? this.state.startTime
                          : this.state.endTime
                        }
                      />
                    </div>
                  )}
                </div>
              </div>
              <div className={styles['actions']}>
                <div className={styles['inner']}>
                  <ButtonSecondaryMedium
                    text="Cancel"
                    onClick={() => {
                      this.setState({
                        startTime: this.props.startTime,
                        endTime: this.props.endTime,
                        rangeType: this.props.rangeType,
                        pickerLabel: getPickerLabel(this.props.startTime, this.props.endTime, this.props.rangeType),
                        rangeLabel: getRangeLabel(this.props.startTime, this.props.endTime, this.props.rangeType)
                      });
                      this.reset();
                    }}
                  />
                  <ButtonPrimaryMedium
                    text="Apply"
                    onClick={() => {
                      this.props.updateTimeRangeQueryParams({
                        startTimeMs: this.state.startTime * 1000,
                        stopTimeMs: this.state.endTime * 1000,
                        rangeType: this.state.rangeType
                      });
                      this.reset();
                    }}
                  />
                </div>
              </div>
            </div>
          </MouseCatcher>
        )}
      </div>
    );
  }
}

function mapStateToProps(state) {
  return {
    timePickerVisible: state.monitoring.layout.timePickerVisible,
    startTime: state.monitoring.timeframe.startTimeMs/1000,
    endTime: state.monitoring.timeframe.stopTimeMs/1000,
    rangeType: state.monitoring.timeframe.rangeType
  };
}

export default connect(
  mapStateToProps,
  {
    setTimePickerVisibility,
    setTimeRange,
    updateTimeRangeQueryParams
  }
)(TimePicker);
