// @flow
// (Copyright) [2016 - 2016] Confluent, Inc.
import debug from 'debug';
import { browserHistory } from 'react-router';
import { call, put, race, select, take, fork } from 'redux-saga/effects';
import { change, formValueSelector } from 'redux-form';
import { takeLatest } from 'redux-saga';

import * as actions from 'actions/clusters';
import { CLUSTERS_INDEX_PATHNAME } from 'constants/clusters';
import { FormTypes } from 'types/form';
import { clearData } from 'actions/monitoring/data';
import { getJson, patchJson } from 'actions/util';
import { locationBeforeTransitionsSelector } from 'selectors/routing';

import type {
  ClusterSelectedAction,
  LoadClusterEditFormAction,
  LoadClustersAction
} from 'actions/clusters';

import type {
  Cluster,
  AllClustersResponse,
  ClusterType,
  ClustersResponse
} from 'types/clusters';

import type { FormChangeAction } from 'types/form';

const log = debug('controlcenter:sagas:clusters');

const getClusterQueryParam = (clusterType: ClusterType) => clusterType === 'MONITORING' ? 'monitoringClusterId' : 'connectClusterId';

export function constructClusterPath(clusterType: ClusterType) {
  return `/2.0/clusters/kafka/display/${clusterType === 'CONNECT' ? 'kafka-connect-ui' : 'stream-monitoring'}`;
}

export function* fetchClusters(action: LoadClustersAction): Generator<*, *, *> {
  const { clusterType } = action.payload;
  const reqPath = constructClusterPath(clusterType);

  yield put(actions.clustersFetch.request(clusterType));

  try {
    const response: ClustersResponse | any = yield call(getJson, reqPath);
    const { clusters, defaultClusterId } = response;

    yield put(actions.clustersFetch.success(clusterType, defaultClusterId, clusters));

    const location: any = yield select(locationBeforeTransitionsSelector);

    const { query } = location;

    // if there isn't a clusterId param set, it should be the default
    if (typeof query[getClusterQueryParam(clusterType)] === 'undefined') {
      yield put(actions.selectCluster(clusterType, defaultClusterId));
    }

  } catch (error) {
    log(`error looking up path=${reqPath} status=${error.message}`);
    yield put(actions.clustersFetch.failure(error));
  }
}

export function* fetchAllClusters(): Generator<*, *, *> {
  const reqPath = '/2.0/clusters/kafka';
  yield put(actions.allClustersFetch.request());

  try {
    const clusters: AllClustersResponse | any = yield call(getJson, reqPath);
    yield put(actions.allClustersFetch.success(clusters));
  } catch (error) {
    log(`error looking up path=${reqPath} status=${error.message}`);
    yield put(actions.allClustersFetch.failure(error));
  }
}

export function* selectCluster(action: ClusterSelectedAction): Generator<*, *, *> {
  const { clusterType, selectedClusterId } = action.payload;

  const location: any = yield select(locationBeforeTransitionsSelector);

  const { query, pathname } = location;

  const clusterQueryParam = getClusterQueryParam(clusterType);

  if (selectedClusterId === query[clusterQueryParam]) {
    return;
  }

  if (clusterType === 'MONITORING') {
    yield put(clearData());
  }

  yield call(browserHistory.push, {
    pathname,
    query: {
      [clusterQueryParam]: selectedClusterId
    }
  });
}

export function* loadClusterEditForm(action: LoadClusterEditFormAction): Generator<*, *, *> {
  const { clusterId } = action.payload;

  yield put(actions.loadClusterIndex());

  const clusterResponse: any = yield race({
    success: take(actions.ALL_CLUSTERS_FETCH_SUCCESS),
    failure: take(actions.ALL_CLUSTERS_FETCH_FAILURE)
  });

  if (clusterResponse.failure) {
    return false;
  }

  const { clusters } = clusterResponse.success.payload;

  const cluster = clusters.find(item => item.clusterId === clusterId);

  if (typeof cluster === 'undefined') {
    return false;
  }

  yield put(change(FormTypes.CLUSTER_FORM, 'displayName', cluster.displayName || cluster.clusterId));
  yield put(change(FormTypes.CLUSTER_FORM, 'clusterId', cluster.clusterId));
  yield put(change(FormTypes.CLUSTER_FORM, 'zookeeperConnect', cluster.zookeeperConnect));
  yield put(change(FormTypes.CLUSTER_FORM, 'bootstrapServers', cluster.bootstrapServers));
}

const clusterFormValueSelector = formValueSelector(FormTypes.CLUSTER_FORM);

export function* saveClusterEditForm(): Generator<*, *, *> {

  const values: Cluster | any = yield select(state => clusterFormValueSelector(state, 'displayName', 'clusterId'));

  const { clusterId, ...clusterUpdates } = values;

  const reqPath = `/2.0/clusters/kafka/${String(clusterId)}`;

  yield put(actions.saveCluster.request(clusterUpdates));

  try {
    yield call(patchJson, reqPath, clusterUpdates);
    yield put(actions.saveCluster.success(clusterUpdates));

    yield call(browserHistory.push, CLUSTERS_INDEX_PATHNAME);
  } catch (error) {
    log(`error looking up path=${reqPath} status=${error.message}`);
    yield put(actions.saveCluster.failure(error));
  }
}

export function* clusterFormChangeSaga(action: FormChangeAction): Generator<*, *, *> {
  const { meta: { field }, payload: value } = action;
  yield put(change(FormTypes.CLUSTER_FORM, field, value));
}

export function* loadClustersSaga(): Generator<*, *, *> {
  yield call(takeLatest, actions.LOAD_CLUSTERS, fetchClusters);
}

export function* loadClusterIndexSaga(): Generator<*, *, *> {
  yield call(takeLatest, actions.LOAD_CLUSTER_INDEX, fetchAllClusters);
}

export function* selectClusterSaga(): Generator<*, *, *> {
  yield call(takeLatest, actions.SELECT_CLUSTER, selectCluster);
}

export function* loadClusterEditFormSaga(): Generator<*, *, *> {
  yield call(takeLatest, actions.LOAD_CLUSTER_EDIT_FORM, loadClusterEditForm);
}

export function* saveClusterEditFormSaga(): Generator<*, *, *> {
  yield call(takeLatest, actions.SAVE_CLUSTER_EDIT_FORM, saveClusterEditForm);
}

export function* clustersSaga(): Generator<*, *, *> {
  yield fork(loadClusterEditFormSaga);
  yield fork(loadClusterIndexSaga);
  yield fork(loadClustersSaga);
  yield fork(saveClusterEditFormSaga);
  yield fork(selectClusterSaga);
}

export default clustersSaga;
