// @flow
// (Copyright) [2016 - 2016] Confluent, Inc.
import debug from 'debug';
import { actionTypes, stopAsyncValidation, startAsyncValidation } from 'redux-form';
import { call, put, race, select, take, fork } from 'redux-saga/effects';
import { delay, takeLatest } from 'redux-saga';

import * as ConnectorActions from 'actions/connectors';
import * as ConnectorFormActions from 'actions/connectorForm';
import { getSelectedClusterIdSelector } from 'selectors/clusters';
import * as ConnectorPluginActions from 'actions/connectorPlugins';
import { createDebounceSaga } from '../util';
import { currentFormTypeSelector } from 'selectors/management';
import { encodeFieldName, serializeValues } from 'utils/connector';
import { getJson, makeAction } from 'actions/util';
import { InputTypes } from 'constants/management';

import { FormTypes } from 'types/form';

import type { Action } from 'types';
import type { FormBlurAction, FormChangeAction } from 'types/form';
import type {
  ConnectorConfig,
  ConnectorFetchSuccessAction
} from 'actions/connectors';
import type {
  LoadConnectorForEditingAction,
  SubmitFormAction
} from 'actions/connectorForm';
import type {
  ConnectorPluginValidateSuccessAction
} from 'actions/connectorPlugins';

import { get } from 'lodash';

const log = debug('controlcenter:sagas:management:connectorForm');

type SearchOption = {
  reqPath: string,
  formatter: (item: Object) => {
    label: string,
    value: string
  }
};

type SearchOptions = { [key: string]: SearchOption };

const TOPICS_BY_CLUSTER_PATH = 'TOPICS_BY_CLUSTER_PATH';

const topicSearchOption: SearchOption = {
  reqPath: TOPICS_BY_CLUSTER_PATH,
  formatter: item => ({
    label: item.name,
    value: item.name
  })
};

const SearchOptionsByFieldType: SearchOptions = {
  [encodeFieldName('topic')]: topicSearchOption,
  [encodeFieldName('topics')]: topicSearchOption,
  [encodeFieldName('connector.class')]: {
    reqPath: '/2.0/management/connect/connector-plugins',
    formatter: item => ({
      label: item.class.slice(item.class.lastIndexOf('.') + 1),
      value: item.class
    })
  }
};

const connectorBlacklist = ['org.apache.kafka.connect.runtime.rest.resources.ConnectorPluginsResourceTest.ConnectorPluginsResourceTestConnector'];

export function* search(action: FormBlurAction): Generator<*, *, *> {
  const { field: fieldName, form } = action.meta;
  const field = encodeFieldName(fieldName);

  if (form !== FormTypes.CONNECTOR_FORM || !(field in SearchOptionsByFieldType)) {
    return;
  }

  const options = SearchOptionsByFieldType[field];

  let { reqPath } = options;

  yield put(makeAction(ConnectorFormActions.CONNECTOR_FORM_SEARCH_REQUESTED, {
    field,
    form
  }));

  try {
    if (reqPath === TOPICS_BY_CLUSTER_PATH) {
      const clusterId: any = yield select(getSelectedClusterIdSelector('CONNECT'));
      reqPath = `/2.0/kafka/${clusterId}/topics`;
    }

    let results: any = yield call(getJson, reqPath);

    if (field === encodeFieldName('connector.class')) {
      const formType = yield select(currentFormTypeSelector);
      results = results.filter(option => option.type === formType && !connectorBlacklist.includes(option.class));
    }

    if (results.length) {
      yield put(makeAction(ConnectorFormActions.CONNECTOR_FORM_SEARCH_SUCCEEDED, {
        field,
        options: results.map(options.formatter)
      }));
    }
  } catch (error) {
    log(`error looking up path=${reqPath} status=${error.message}`);
    yield put(makeAction(ConnectorFormActions.CONNECTOR_FORM_SEARCH_FAILED, error));
  }
}

export function* loadConnectorForEditing(action: LoadConnectorForEditingAction): Generator<*, *, *> {
  const { name, form } = action.payload;

  yield put(ConnectorFormActions.fetchEdit.request(name, form));
  yield put(ConnectorActions.loadConnector(name));

  const fetchAction: ConnectorFetchSuccessAction | any = yield take(ConnectorActions.CONNECTOR_FETCH_SUCCESS);
  const { connector } = fetchAction.payload;

  yield put(ConnectorPluginActions.validateConfig(connector.config));

  const validateSuccessAction: ConnectorPluginValidateSuccessAction | any = yield take(ConnectorPluginActions.CONNECTOR_PLUGIN_VALIDATE_SUCCESS);
  const { validationData } = validateSuccessAction.payload;

  validationData.configs = validationData.configs.map((config: Object) => {
    const { name } = config.value;

    if (config.definition.type !== InputTypes.PASSWORD || !(name in connector.config)) {
      return config;
    }

    return {
      ...config,
      value: {
        ...config.value,
        value: connector.config[name]
      }
    };
  });

  yield put(ConnectorFormActions.fetchEdit.success(name, validationData, connector));
}

export function* validateForm(action: Action<*, *>): Generator<*, *, *> {
  const { field, form, value } = action.payload;

  const values: any = yield select(state => serializeValues(state.form[form].values));
  const name = values.name;

  yield put(startAsyncValidation(form));
  yield put(ConnectorFormActions.validate.request(name, form));

  const breakingErrors = {};

  if (!values['connector.class']) {
    breakingErrors['connector.class'] = 'Required';
  }

  if (!name) {
    breakingErrors.name = 'Required';
  }

  if (field === 'name' && value) {
    yield put(ConnectorActions.loadConnector(value));

    const connector: any = yield race({
      notfound: take(ConnectorActions.CONNECTOR_FETCH_FAILURE),
      found: take(ConnectorActions.CONNECTOR_FETCH_SUCCESS)
    });

    if (connector.found) {
      breakingErrors.name = `A connector named "${value}" already exists`;
    }
  }

  if (Object.keys(breakingErrors).length) {
    return yield put(stopAsyncValidation(form, breakingErrors));
  }

  yield put(ConnectorPluginActions.validateConfig(values));

  const validateSuccessAction: any = yield take(ConnectorPluginActions.CONNECTOR_PLUGIN_VALIDATE_SUCCESS);

  const { validationData } = validateSuccessAction.payload;

  const errors = validationData.configs.reduce((accum, fieldConfig) => {
    const { errors, name } = fieldConfig.value;

    if (errors.length) {
      accum[encodeFieldName(name)] = errors[0];
    }

    return accum;
  }, {});

  // TODO(@dylanharrington): conditionally start/stop async if a delay of ~1000ms finishes before validation
  yield put(ConnectorFormActions.validate.success(name, validationData));
  yield put(stopAsyncValidation(form, errors));
}

export function* submit(action: SubmitFormAction): Generator<*, *, *> {
  const { form } = action.payload;

  const config : any | ConnectorConfig = yield select(state => serializeValues(state.form[form].values));

  yield put(ConnectorFormActions.submit.request(config.name, form));
  yield put(ConnectorActions.saveConnector(config));

  // TODO(@dylanharrington): handle FAILURE
  yield take(ConnectorActions.CONNECTOR_SAVE_SUCCESS);
  yield put(ConnectorFormActions.submit.success(config.name, form));
}

// this could be abstracted, just needs to handle the action transform
export function* formChange(action: FormChangeAction): Generator<*, *, *> {
  const { meta: { field, form }, payload: value } = action;
  try {
    yield call(delay, 100);
    yield fork(validateForm, {
      payload: {
        field,
        form,
        value
      }
    });
  } catch (error) {
    log(error);
  }
}

export function* loadConnectorForEditingSaga(): Generator<*, *, *> {
  yield* takeLatest(ConnectorFormActions.LOAD_CONNECTOR_FOR_EDITING, loadConnectorForEditing);
}

export function* searchSaga(): Generator<*, *, *> {
  yield* takeLatest(actionTypes.FOCUS, search);
}

export function* submitSaga(): Generator<*, *, *> {
  yield* takeLatest(ConnectorFormActions.START_SUBMIT, submit);
}

const formChangeWatcherSaga: Generator<*, *, *> = createDebounceSaga(
  actionTypes.CHANGE,
  formChange,
  action => get(action, 'meta.form') === FormTypes.CONNECTOR_FORM
);

export default function* connectorFormSaga(): Generator<*, *, *> {
  yield fork(submitSaga);
  yield fork(formChangeWatcherSaga);
  yield fork(loadConnectorForEditingSaga);
  yield fork(searchSaga);
}
