// @flow
// (Copyright) [2016 - 2016] Confluent, Inc.
import { createSelector } from 'reselect';
import debug from 'debug';


import { DataLayoutTypes } from 'types/monitoring/data';

import type {
  DataLayoutType,
  MultiTimeSeries,
  TimeSeries
} from 'types/monitoring/data';

import {
  chartXScaleSelector,
  desiredDataQuantizationMsSelector,
  slotWidthMsSelector,
  startTimeSlotQuantizedSelector,
  stopTimeSlotQuantizedSelector
} from 'selectors/monitoring';

import { computeIgnoreDeliveryDiscrepancyAfterIndex } from 'utils/monitoring/data';
import { focusedComponentSelector } from 'selectors/layout';
import { ComponentTypes } from 'actions/layout';
import type { FocusedSpec } from 'reducers/layout';
import {
  aggregateCountToSlots,
  aggregateTotalCountMinMaxToSlots
} from 'components/charting/util';
import type { TimeScale } from 'types/monitoring';

const log = debug('controlcenter:selectors:monitoring:data');

const stateDataSelector = state => state.monitoring.data;

export const currentDataBinSizeMs = (state: any): ?number => state.monitoring.data ? state.monitoring.data.stepMs : null;

// note: -ve response -> remove this number of elements from start
export function numberToExtendArrayAtStart(
  dataStartTimeMs: number,
  dataStepMs: number,
  uiStartTimeMs: number
): number {
  return Math.floor((dataStartTimeMs - uiStartTimeMs) / dataStepMs);
}

// note: -ve response -> remove this number of elements from end.
export function numberToExtendArrayAtEnd(
  numberOfValues: number,
  dataStartTimeMs: number,
  dataStepMs: number,
  uiStopTimeMs: number
): number {
  const lastDataBinStartTimeMs = dataStartTimeMs + dataStepMs * numberOfValues;
  return Math.floor((uiStopTimeMs - lastDataBinStartTimeMs) / dataStepMs);
}

export function alignDataAndChartStartStopTimes<T:Object>(
  values: T,
  dataStartTimeMs: number,
  dataStepMs: number,
  uiStartTimeMs: number,
  uiStopTimeMs: number
): MultiTimeSeries<T> {

  const adjValues: T = Object.assign({}, values);
  const keys = Object.keys(adjValues);
  const arrayLen = adjValues[keys[0]].length;

  const numberToAddAtStart = numberToExtendArrayAtStart(dataStartTimeMs, dataStepMs, uiStartTimeMs);
  const numberToAddAtEnd = numberToExtendArrayAtEnd(arrayLen, dataStartTimeMs, dataStepMs, uiStopTimeMs);

  keys.forEach(key => {
    if (numberToAddAtStart > 0) {
      adjValues[key] = Array(numberToAddAtStart).fill(null).concat(adjValues[key]);
    } else if (numberToAddAtStart < 0) {
      adjValues[key] = adjValues[key].slice(-numberToAddAtStart);
    }

    if (numberToAddAtEnd > 0) {
      adjValues[key] = adjValues[key].concat(Array(numberToAddAtEnd).fill(null));
    } else if (numberToAddAtEnd < 0) {
      adjValues[key] = adjValues[key].slice(0, numberToAddAtEnd);
    }
  });

  return {
    startTimeMs: dataStartTimeMs - numberToAddAtStart * dataStepMs,
    stepMs: dataStepMs,
    values: adjValues
  };
}

export function parseData(data: any, startTimeMs: ?number, stepMs: number, uiStartTimeMs: number, uiStopTimeMs: number): any {
  if (Array.isArray(data)) {
    return data.map(datum => parseData(datum, null, stepMs, uiStartTimeMs, uiStopTimeMs));
  }
  const result = {};
  Object.keys(data).forEach(k => {
    switch (k) {
      case 'overview':
        result.overview =
          parseData(data.overview, data.overview.startTimeMs, stepMs, uiStartTimeMs, uiStopTimeMs);
        break;
      case 'members':
        result.members = data.members.map(
          member => parseData(member, member.startTimeMs, stepMs, uiStartTimeMs, uiStopTimeMs)
        );
        break;
      case 'error':
        // note: missing data gets consolidated (and this property removed) later.
        if (startTimeMs != null) {
          result.missing = alignDataAndChartStartStopTimes(
            { missing: data.error },
            startTimeMs, stepMs,
            uiStartTimeMs, uiStopTimeMs
          ).values.missing;
        }
        break;
      case 'crc':
        if (startTimeMs != null) {
          const crcResult = alignDataAndChartStartStopTimes(
            { crc: data.crc },
            startTimeMs, stepMs,
            uiStartTimeMs, uiStopTimeMs
          );
          // convert MultiTimeSeries to TimeSeries
          result.crc = {
            startTimeMs: crcResult.startTimeMs,
            stepMs: crcResult.stepMs,
            values: crcResult.values.crc
          };
        }
        break;
      case 'group':
        if (Array.isArray(data[k])) {
          if (startTimeMs != null) {
            // group can either be a sequence if this is the primary message delivery
            // data for a particular consumer.
            const groupResult = alignDataAndChartStartStopTimes(
              { group: data.group },
              startTimeMs, stepMs,
              uiStartTimeMs, uiStopTimeMs
            );
            result.group = {
              startTimeMs: groupResult.startTimeMs,
              stepMs: groupResult.stepMs,
              values: groupResult.values.group
            };
          }
        } else {
          // otherwise it's part of an identity.
          result[k] = data[k];
        }
        break;
      case 'count':
        if (startTimeMs != null) {
          const countResult = alignDataAndChartStartStopTimes(
            { count: data.count },
            startTimeMs, stepMs,
            uiStartTimeMs, uiStopTimeMs
          );
          // TODO(mhowlett): We'd probably be better off without the TimeSeries type.
          // convert MultiTimeSeries to TimeSeries
          result.count = {
            startTimeMs: countResult.startTimeMs,
            stepMs: countResult.stepMs,
            values: countResult.values.count
          };
        }
        break;
      case 'totalLatency':
        if (startTimeMs != null) {
          result.latency =
            alignDataAndChartStartStopTimes(
              {
                totalLatency: data.totalLatency,
                count: data.count,
                minLatency: data.minLatency,
                maxLatency: data.maxLatency
              },
              startTimeMs, stepMs,
              uiStartTimeMs, uiStopTimeMs
            );
        }
        break;
      case 'minLatency':
      case 'maxLatency':
        break;
      case 'consumed':
      case 'produced':
      default:
        // includes children of members.
        if (typeof data[k] === 'object' && data[k]) {
          result[k] = parseData(data[k], startTimeMs, stepMs, uiStartTimeMs, uiStopTimeMs);
        } else {
          result[k] = data[k];
        }
        break;
    }
  });
  return result;
}

export function transformMissingData(data: Object): void {
  const pMissing = data.produced && data.produced.missing
    ? data.produced.missing
    : null;

  const cMissing = data.consumed && data.consumed.missing
    ? data.consumed.missing
    : null;

  if (!pMissing && !cMissing) {
    data.missing = [];
    return;
  }

  let startTimeMs;
  let stepMs;
  let missing;
  if (pMissing && cMissing) {
    missing = pMissing.map((isMissing, i) => isMissing || cMissing[i]);
    delete data.produced.missing;
    delete data.consumed.missing;
    startTimeMs = data.produced.count.startTimeMs;
    stepMs = data.produced.count.stepMs;
  } else if (pMissing) {
    missing = pMissing;
    delete data.produced.missing;
    startTimeMs = data.produced.count.startTimeMs;
    stepMs = data.produced.count.stepMs;
  } else {
    missing = cMissing;
    delete data.consumed.missing;
    startTimeMs = data.consumed.count.startTimeMs;
    stepMs = data.consumed.count.stepMs;
  }

  if (!missing) {
    // this can never happen, it's to help flow.
    data.missing = [];
    return;
  }

  const ranges = [];
  let isOpen = false;
  let openIndex = -1;
  missing.forEach((isMissing, i) => {
    if (isMissing && !isOpen) {
      isOpen = true;
      openIndex = i;
    }
    if (!isMissing && isOpen) {
      isOpen = false;
      ranges.push([openIndex, i]);
    }
  });

  if (isOpen) {
    ranges.push([openIndex, missing.length]);
  }

  // Note: the end time refers to end of the last bin (not the start of it)
  data.missing = ranges.map(range => ({
    startTime: new Date(startTimeMs + range[0]*stepMs),
    endTime: new Date(startTimeMs + range[1]*stepMs)
  }));
}

export const dataSelector = createSelector(
  [stateDataSelector, startTimeSlotQuantizedSelector, stopTimeSlotQuantizedSelector, currentDataBinSizeMs, desiredDataQuantizationMsSelector],
  (data: Object, startTimeSlotQuantized: Date, stopTimeSlotQuantized: Date, currentDataBinSizeMs: ?number, desiredDataQuantizationMs: number): ?Object => {

    if (!data) {
      return null;
    }

    // TODO(mhowlett): something more clever instead of this.
    if (currentDataBinSizeMs !== desiredDataQuantizationMs) {
      return null;
    }

    const result = parseData(
      data,
      null,
      data.stepMs,
      startTimeSlotQuantized.getTime(),
      stopTimeSlotQuantized.getTime()
    );

    if (!result.members) {
      result.members = [];
    }

    if (result.overview) {
      transformMissingData(result.overview);
    }
    result.members.forEach((member) => transformMissingData(member));

    return result;
  }
);

export const currentDataLayoutTypeSelector = createSelector(
  [stateDataSelector],
  (data: any): ?DataLayoutType => {

    if (!data) {
      return null;
    }

    if (!data.type) {
      return null;
    }

    if (Object.keys(DataLayoutTypes).indexOf(data.type) === -1) {
      log(`unknown data layout type: ${data.type}`);
      return null;
    }

    return data.type;
  }
);

export const currentDataLayoutIdSelector = createSelector(
  [stateDataSelector, currentDataLayoutTypeSelector],
  (data: any, dataLayoutType: ?DataLayoutType): ?string => {
    switch (dataLayoutType) {
      case 'CONSUMER_GROUP':
      case 'CONSUMER_GROUP_TOPICS':
        return data.group;
      case 'CONSUMER':
        return data.clientId;
      default:
        return null;
    }
  }
);

export const producedCountDataSelector = createSelector(
  [dataSelector],
  (data) => {
    if (!data || !data.overview || !data.overview.produced || !data.overview.produced.count) {
      return null;
    }
    return data.overview.produced.count;
  }
);

export const consumedCountDataSelector = createSelector(
  [dataSelector],
  (data) => {
    if (!data || !data.overview || !data.overview.consumed || !data.overview.consumed.count) {
      return null;
    }
    return data.overview.consumed.count;
  }
);

export const consumedLatencyDataSelector = createSelector(
  [dataSelector],
  (data) => {
    if (!data || !data.overview || !data.overview.consumed || !data.overview.consumed.latency) {
      return null;
    }
    return data.overview.consumed.latency;
  }
);

export const ignoreDeliveryDiscrepancyAfterIndexSelector = createSelector(
  [producedCountDataSelector, consumedCountDataSelector],
  (producedCounts: ?Object, consumedCounts: ?Object): ?number => {
    return computeIgnoreDeliveryDiscrepancyAfterIndex(producedCounts, consumedCounts);
  }
);

export const consumerGroupsSelector = createSelector(
  [dataSelector],
  (data: ?Object): ?Array<string> => {
    if (!data || !data.overview || !data.overview.consumed || !data.overview.consumed.group) {
      return null;
    }
    return data.overview.consumed.group.values.reduce((accum, current) => {
      // finding more than one is a rare event, so this is perfectly acceptable.
      if (accum.indexOf(current) === -1 && current != null) {
        accum.push(current);
      }
      return accum;
    }, []).sort();
  }
);

export const lastMemberListFetchTimeMs = (state: Object) =>
  state.monitoring.data && state.monitoring.data.listEpoch
    ? state.monitoring.data.listEpoch
    : 0;

export const lastOverviewDataFetchTimeMs = (state: Object) =>
  state.monitoring.data && state.monitoring.data.overviewEpoch
    ? state.monitoring.data.overviewEpoch
    : 0;

export type FocusedComponentSlotAggregatedData = {
  produced: TimeSeries,
  consumed: TimeSeries,
  latency: MultiTimeSeries<*>
}

export const focusedChartSlotAggregatedSelector = createSelector(
  [dataSelector, focusedComponentSelector, slotWidthMsSelector, chartXScaleSelector],
  (data: ?Object,
  focusedComponent: FocusedSpec,
  slotWidthMs: number,
  chartXScale: TimeScale
): ?FocusedComponentSlotAggregatedData => {

    if (!focusedComponent) {
      return null;
    }

    if (focusedComponent.type !== ComponentTypes.MONITORING_CHART) {
      return null;
    }

    if (data == null) {
      return null;
    }

    const isOverviewComponent = focusedComponent.id.componentId === -1;

    const produced = isOverviewComponent
      ? data.overview.produced.count
      : data.members[focusedComponent.id.componentId].produced.count;

    const consumed = isOverviewComponent
      ? data.overview.consumed.count
      : data.members[focusedComponent.id.componentId].consumed.count;

    const latency = isOverviewComponent
      ? data.overview.consumed.latency
      : data.members[focusedComponent.id.componentId].consumed.latency;

    const domain = chartXScale.domain();

    const producedAggregated = aggregateCountToSlots(produced, slotWidthMs, domain);
    const consumedAggregated = aggregateCountToSlots(consumed, slotWidthMs, domain);
    const latencyAggregated = aggregateTotalCountMinMaxToSlots(latency, slotWidthMs, domain);

    if (producedAggregated == null || consumedAggregated == null || latencyAggregated == null) {
      return null;
    }

    return {
      produced: producedAggregated,
      consumed: consumedAggregated,
      latency: latencyAggregated
    };
  }
);
