// @flow
// (Copyright) [2016 - 2016] Confluent, Inc.
import d3scale from 'd3-scale';
import { createSelector } from 'reselect';
import { assert } from 'utils';

import { Rollups } from 'types/monitoring';
import type { TimeScale, Rollup } from 'types/monitoring';

import {
  Y_AXIS_WIDTH_PX,
  OVERVIEW_MESSAGE_DELIVERY_CHART_HEIGHT,
  OVERVIEW_LATENCY_CHART_HEIGHT
} from 'constants/monitoring/layout';

const MIN_SLOT_WIDTH_PX = 4;


// Note: - the start-stop time ranges below should generally be considered
//         closed-open intervals [...). This is the case because we
//         identify time windows by a start time and duration. i.e. for
//         a given quantized start-stop interval end point, we are usually not
//         interested in the window with that start time.

// Note: - Some functions are memoized, some are not. Generally there are two
//         reasons to memoize:
//         1. efficiency. if the calculation is relatively slow there is
//            benefit to caching results.
//         2. object identity can be used to check whether the procuced value
//            has changed. never necessary if the returned value is, for example,
//            of type number. but if it's object, memoizing conveniently allows
//            use of === to check if the value has changed (provided the same
//            applies to inputs to the function). Note that memoizing will
//            actually be less efficient for some of the trivial functions.

export const desiredRollupNoCountLimit = (state: any): Rollup => {

  // TODO: this can be negative
  const axisWidthPx = state.monitoring.layout.chartAreaWidthPx - Y_AXIS_WIDTH_PX;
  const axisWidthMs = state.monitoring.timeframe.stopTimeMs - state.monitoring.timeframe.startTimeMs;
  const minSlotWidthMs = (axisWidthMs / axisWidthPx) * MIN_SLOT_WIDTH_PX;
  const minDataResolutionMs = minSlotWidthMs / 2; // 2 is somewhat arbitrary.

  if (minDataResolutionMs > Rollups.toMs(Rollups.ONE_WEEK)) {
    return Rollups.ONE_WEEK;
  }

  if (minDataResolutionMs > Rollups.toMs(Rollups.ONE_HOUR)) {
    return Rollups.ONE_HOUR;
  }

  return Rollups.FIFTEEN_SECONDS;
};


export function desiredRollup(state: any): Rollup {
  const rollupNoCountLimit = desiredRollupNoCountLimit(state);

  const nPts = (state.monitoring.timeframe.stopTimeMs - state.monitoring.timeframe.startTimeMs) / Rollups.toMs(rollupNoCountLimit);

  if (nPts < 6000) {
    return rollupNoCountLimit;
  }
  if (rollupNoCountLimit === Rollups.FIFTEEN_SECONDS) {
    return Rollups.ONE_HOUR;
  }
  if (rollupNoCountLimit === Rollups.ONE_HOUR) {
    return Rollups.ONE_WEEK;
  }
  return rollupNoCountLimit;
}


export const desiredDataQuantizationMsSelector = (state: any): number => Rollups.toMs(desiredRollup(state));


// - dataQuantized start and end times are the most fine grained.
// - currently the quantization amount is a constant, eventually it will be
//   derived from actual data.
// - round the values up or down?
//   - rounding up the stop time seems most natural as it matches what we want to
//     do in the case the range is tracking current time.
//   - I've rounded up the start time as well to match. This means if a "nice"
//     time range was provided (eg 4 hours), that the time range will stay the
//     same even if the end points are not on quantization points.

const timeDataQuantizedMs = (td: number, binSize: number): number => Math.ceil(td/binSize)*binSize;
const toDate = (millis: number): Date => new Date(millis);
export const startTimeDataQuantizedMsSelector =
  (state: any): number => timeDataQuantizedMs(
    state.monitoring.timeframe.startTimeMs,
    desiredDataQuantizationMsSelector(state)
  );
export const stopTimeDataQuantizedMsSelector =
  (state: any) : number => timeDataQuantizedMs(
    state.monitoring.timeframe.stopTimeMs,
    desiredDataQuantizationMsSelector(state)
  );
export const startTimeDataQuantizedSelector = createSelector(
  [startTimeDataQuantizedMsSelector],
  (startTimeDataQuantizedMs: number): Date => toDate(startTimeDataQuantizedMs)
);
export const stopTimeDataQuantizedSelector = createSelector(
  [stopTimeDataQuantizedMsSelector],
  (stopTimeDataQuantizedMs: number): Date => toDate(stopTimeDataQuantizedMs)
);


// - a slot is the unit of quantization in the charting area.
// - it is greater than 1 pixel primarily because of styling requirements for
//   range area chart.
// - slot width and positioning is invariant over the entire charting area.
//   line charts are quantized to this in addition to range charts. this is
//   important as it is imporant for all charts to appear to update in lock-step.

export const calcSlotWidthMs = (
  widthPx: number,
  desiredDataQuantizationMs: number,
  startTimeDataQuantizedMs: number,
  stopTimeDataQuantizedMs: number
) => {
  const axisWidthPixels = widthPx - Y_AXIS_WIDTH_PX;
  const quantizedTimeExtentMs = stopTimeDataQuantizedMs - startTimeDataQuantizedMs;
  const numberOfDataPointsOverTimeExtent = quantizedTimeExtentMs / desiredDataQuantizationMs;
  const maxSlots = Math.floor(axisWidthPixels / MIN_SLOT_WIDTH_PX);
  const resampleCount = Math.ceil(numberOfDataPointsOverTimeExtent/maxSlots);
  return resampleCount * desiredDataQuantizationMs;
  // TODO: actually, because this calculation is based on the dataQuantization
  // times not slotQuantization times (which will be a slightly different range),
  // it's possible for this calculation to slightly under or over estimate
  // resampleCount which could result in a slotWidthPx value that is slightly <
  // MIN_SLOT_WIDTH_PX. I don't think this is worth worrying about.
};
export const slotWidthMsSelector = (state: any) : number => calcSlotWidthMs(
  chartAreaWidthPxSelector(state),
  desiredDataQuantizationMsSelector(state),
  startTimeDataQuantizedMsSelector(state),
  stopTimeDataQuantizedMsSelector(state)
);

// - start time - round up or down?
//   - we don't ever want partial bins at the start or a range, so round up
//     in case the data we have is determined from the input time range.
// - stop time - round up or down?
//   - we DO want partial bins at the end of range, so round up
//     in case the data we have is determined from the input time range.
const timeSlotQuantizedMs = (slotWidthMs: number, timeDataQuantizedMs: number) : number => Math.ceil(timeDataQuantizedMs / slotWidthMs) * slotWidthMs;

export const calcTimeSlotQuantizedMs = (
  timeMs: number,
  widthPx: number,
  desiredDataQuantizationMs: number,
  startTimeDataQuantizedMs: number,
  stopTimeDataQuantizedMs: number
) => timeSlotQuantizedMs(
  calcSlotWidthMs(widthPx, desiredDataQuantizationMs, startTimeDataQuantizedMs, stopTimeDataQuantizedMs),
  timeDataQuantizedMs(timeMs, desiredDataQuantizationMs)
);

export const startTimeSlotQuantizedMsSelector = (state: any) : number => timeSlotQuantizedMs(slotWidthMsSelector(state), startTimeDataQuantizedMsSelector(state));
export const stopTimeSlotQuantizedMsSelector = (state: any) : number => timeSlotQuantizedMs(slotWidthMsSelector(state), stopTimeDataQuantizedMsSelector(state));
export const startTimeSlotQuantizedSelector = createSelector(
  [startTimeSlotQuantizedMsSelector],
  (startTimeSlotQuantizedMs: number): Date => toDate(startTimeSlotQuantizedMs)
);
export const stopTimeSlotQuantizedSelector = createSelector(
  [stopTimeSlotQuantizedMsSelector],
  (stopTimeSlotQuantizedMs: number): Date => toDate(stopTimeSlotQuantizedMs)
);

export const fetchStopTimeMsSelector = (state: any): number => {
  const stopTimeSlotQuantizedMs = stopTimeSlotQuantizedMsSelector(state);
  const fetchExtendedStopTimeMs = state.monitoring.timeframe.fetchExtendedStopTimeMs;
  if (!fetchExtendedStopTimeMs) {
    return stopTimeSlotQuantizedMs;
  }
  if (fetchExtendedStopTimeMs > stopTimeSlotQuantizedMs) {
    return fetchExtendedStopTimeMs;
  }
  return stopTimeSlotQuantizedMs;
};

export const calcSlotWidthPx = (
  yAxisWidth: number,
  widthPx: number,
  slotWidthMs: number,
  startTimeSlotQuantizedMs: number,
  stopTimeSlotQuantizedMs: number
): number => {
  const axisWidthMs = stopTimeSlotQuantizedMs - startTimeSlotQuantizedMs;
  assert(axisWidthMs % slotWidthMs === 0, 'expecting integer # slots in axis extent');
  const numberSlots = axisWidthMs / slotWidthMs;
  const axisWidthPx = widthPx - yAxisWidth;
  return axisWidthPx / numberSlots;
};
export const slotWidthPxSelector = (state: any) : number => calcSlotWidthPx(
  Y_AXIS_WIDTH_PX,
  chartAreaWidthPxSelector(state),
  slotWidthMsSelector(state),
  startTimeSlotQuantizedMsSelector(state),
  stopTimeSlotQuantizedMsSelector(state)
);

export const slotWidth = createSelector(
  [slotWidthMsSelector, slotWidthPxSelector],
  (slotWidthMs, slotWidthPx) => ({ ms: slotWidthMs, px: slotWidthPx })
);


export const chartAreaWidthPxSelector = (state: any): number => state.monitoring.layout.chartAreaWidthPx;

export const chartXScaleSelector = createSelector(
  [startTimeSlotQuantizedSelector, stopTimeSlotQuantizedSelector, chartAreaWidthPxSelector],
  (startTimeSlotQuantized: Date, stopTimeSlotQuantized: Date, chartAreaWidthPx: number): TimeScale => {
    return d3scale.scaleTime()
      .domain([startTimeSlotQuantized, stopTimeSlotQuantized])
      .range([Y_AXIS_WIDTH_PX, chartAreaWidthPx]);
  }
);


export const mouseOverTimeSlotQuantizedSelector = (state: any): ?number => {
  const mouseOverTime = state.monitoring.layout.mouseOverTime;
  const slotWidthMs = slotWidthMsSelector(state);
  return mouseOverTime
    ? timeSlotQuantizedMs(slotWidthMs, timeDataQuantizedMs(mouseOverTime.getTime(), slotWidthMs)) - slotWidthMs / 2
    : null;
};


export const visibleMembersSelector = (state: any): { startIndex: number, endIndex: number } => {
  const { pageScrollOffsetY, pageHeight, windowHeight } = state.monitoring.layout;

  const numberOfMembers = state.monitoring && state.monitoring.data && state.monitoring.data.members
    ? state.monitoring.data.members.length
    : 0;

  if (numberOfMembers === 0) {
    return { startIndex: -1, endIndex: -1 };
  }

  // Some calculations below are not ideal because they are not robust to changes in styling.
  // (that is, some dimension information is currently defined in CSS so not available here.
  //  I don't feel like changing that).
  // Also, the sticky nav header is not considered correctly.
  // However, the calculation is approximate only and constructed in such a way that I believe
  // it is never likely to miss members that should be currently visible.

  const HEADER_HEIGHT = 60;
  const STICKY_HEADER_HEIGHT = 37;
  const AXIS_HEIGHT = 20;

  const effectiveWindowHeight = Math.max(10, windowHeight - HEADER_HEIGHT - STICKY_HEADER_HEIGHT);

  const overviewChartHeight = OVERVIEW_MESSAGE_DELIVERY_CHART_HEIGHT + OVERVIEW_LATENCY_CHART_HEIGHT + AXIS_HEIGHT;

  const memberAreaTotalHeight = pageHeight - overviewChartHeight - STICKY_HEADER_HEIGHT;
  const heightPerMember = memberAreaTotalHeight / numberOfMembers;

  const adjPageScrollOffset = pageScrollOffsetY - overviewChartHeight;

  // expand range by 1 on either side. This should be enough to ensure a member
  // that should be visible is never excluded from the range. In general, the range
  // will be two too wide though.
  const arbitraryRangeExpand = 1;

  const startIndex = (() => {
    const idx = adjPageScrollOffset / heightPerMember;
    return idx < 0 ? -1 : Math.floor(idx) - arbitraryRangeExpand;
  })();

  const endIndex = (() => {
    const idx =
      startIndex +
      Math.ceil(effectiveWindowHeight / heightPerMember) +
      arbitraryRangeExpand*2;
    return idx > numberOfMembers-1 ? numberOfMembers-1 : idx;
  })();

  return { startIndex, endIndex };
};
