// @flow
// (Copyright) [2016 - 2016] Confluent, Inc.
import React from 'react';
import { mount, shallow } from 'enzyme';
import { spy } from 'sinon';

import { Switch } from 'components/form/Switch';
import styles from 'components/form/Switch.css';

import { getBaseFormField } from './util';

describe('<Switch />', () => {
  let props;
    const { reduxFormInputProps } = getBaseFormField('switch');

  beforeEach(() => {
    props = {
      ...reduxFormInputProps,
      offText: 'disabled!',
      onText: 'enabled!',
      value: true
    };
  });

  it('renders', () => {
    const wrapper = mount(<Switch {...props} />);
    expect(wrapper.find({ className: styles['switch-root'] })).to.be.present();
    wrapper.unmount();
  });

  it('responds correctly when checked', () => {
    props = { ...props, value: true };
    const wrapper = shallow(<Switch {...props} />);
    expect(wrapper.find({ className: styles['bauble'] })).to.not.be.present();
    expect(wrapper.find({ className: styles['bauble--checked'] })).to.be.present();
    expect(wrapper.find({ className: styles['slider-container'] })).to.not.be.present();
    expect(wrapper.find({ className: styles['slider-container--checked'] })).to.be.present();
    expect(wrapper.find({ className: styles.text }).text()).to.equal(props.onText);
    wrapper.unmount();
  });

  it('responds correctly when not checked', () => {
    props = { ...props, value: false };
    const wrapper = shallow(<Switch {...props} />);
    expect(wrapper.find({ className: styles['bauble'] })).to.be.present();
    expect(wrapper.find({ className: styles['bauble--checked'] })).to.not.be.present();
    expect(wrapper.find({ className: styles['slider-container'] })).to.be.present();
    expect(wrapper.find({ className: styles['slider-container--checked'] })).to.not.be.present();
    expect(wrapper.find({ className: styles.text }).text()).to.equal(props.offText);
    wrapper.unmount();
  });

  describe('responds correctly when disabled', () => {
    let wrapper;
    let checkboxFocus;
    let checkboxClick;

    beforeEach(() => {
      props = { ...props, disabled: true };
      wrapper = mount(<Switch {...props} />);
      const instance = wrapper.instance();
      checkboxFocus = spy(instance.checkbox, 'focus');
      checkboxClick = spy(instance.checkbox, 'click');
    });

    afterEach(() => {
      wrapper.unmount();
    });

    it('the correct elements are present', () => {
      expect(wrapper.find({ className: styles['switch-container--disabled'] })).to.be.present();
      expect(wrapper.find({ className: styles['switch-container'] })).to.not.be.present();
    });

    it('clicks and focuses checkbox when non-checkbox element is click target', () => {
      expect(checkboxFocus).to.not.be.called;
      expect(checkboxClick).to.not.be.called;

      wrapper.find({ className: styles['switch-container--disabled'] }).not('[type="checkbox"]').simulate('click');
      expect(checkboxFocus).to.be.calledOnce;
      expect(checkboxClick).to.be.calledOnce;
    });

    it('does not click or focus checkbox when checkbox is click target', () => {
      expect(checkboxFocus).to.not.be.called;
      expect(checkboxClick).to.not.be.called;
      wrapper.find({ className: styles['switch-container--disabled'] }).children('[type="checkbox"]').simulate('click');
      expect(checkboxClick).to.not.be.called;
      expect(checkboxFocus).to.not.be.called;
    });
  });

  it('responds correctly when not disabled', () => {
    props = { ...props, disabled: false };
    const wrapper = mount(<Switch {...props} />);
    const handleClick = spy(wrapper.instance(), 'handleClick');

    expect(wrapper.find({ className: styles['switch-container--disabled'] })).to.not.be.present();
    expect(wrapper.find({ className: styles['switch-container'] })).to.be.present();

    wrapper.find({ className: styles['switch-container'] }).simulate('click');
    expect(handleClick).to.not.be.calledOnce;

    wrapper.unmount();
  });
});
