// @flow
// (Copyright) [2016 - 2016] Confluent, Inc.
import React from 'react';
import { mount } from 'enzyme';
import { spy } from 'sinon';

import {
  DataManager,
  computePriority,
  MAX_CONNECTION_COUNT
} from 'components/monitoring/DataManager';
import { RangeTypes } from 'constants/monitoring/timeframe';

import { RequestTypes, DataLayoutTypes } from 'types/monitoring/data';


const dataManagerStaticConfig = {
  cyclePeriodMs: 300,
  memberListRefreshMs: 30000,
  maxFetchIntervalMs: 2000,
  updateWindowSizeMs: 60000,
  fullRefreshMs: 30000
};


let loadDataCallCount;

function resetLoadDataCallCount() {
  loadDataCallCount = 0;
}

function loadDataMock() {
  // ignore all arguments.
  loadDataCallCount += 1;
}

describe('<DataManager />', () => {

  describe('computePriority', () => {

    let now;
    beforeEach((): void => { now = Date.now(); });

    it('previously fetched before maxFetchIntervalMs', () => {
      expect(computePriority(
        -1, now-2000, now-2000,
        { startIndex: -1, endIndex: 1 },
        30000,
        20000,
        RequestTypes.OVERVIEW
      )).to.eql(0);
    });

    it('not retrieved, visible', () => {
      expect(computePriority(
        0, undefined, undefined,
        { startIndex: -1, endIndex: 1 },
        30000,
        5000,
        RequestTypes.MEMBER
      )).to.eql(150);
    });

    it('not retrieved, not visible', () => {
      expect(computePriority(
        10, undefined, undefined,
        { startIndex: -1, endIndex: 1 },
        30000,
        5000,
        RequestTypes.MEMBER
      )).to.eql(0);
    });

    it('not yet retrieved, visible, overview', () => {
      expect(computePriority(
        -1, undefined, undefined,
        { startIndex: -1, endIndex: 1 },
        30000,
        5000,
        RequestTypes.OVERVIEW
      )).to.eql(151);
    });

    it('need full update, not visible', () => {
      const priority = computePriority(
        -1, now-20000, now-40000,
        { startIndex: 0, endIndex: 2 },
        30000,
        5000,
        RequestTypes.OVERVIEW
      );
      expect(priority).to.be.above(40);
      expect(priority).to.be.below(43);
    });

    it('need partial update, not visible', () => {
      const priority = computePriority(
        -1, now-20000, now-20000,
        { startIndex: 0, endIndex: 2 },
        30000,
        5000,
        RequestTypes.OVERVIEW
      );
      expect(priority).to.be.above(9);
      expect(priority).to.be.below(12);
    });

    it('need full update, visible', () => {
      const priority = computePriority(
        -1, now-20000, now-40000,
        { startIndex: -1, endIndex: 2 },
        30000,
        5000,
        RequestTypes.OVERVIEW
      );
      expect(priority).to.be.above(90);
      expect(priority).to.be.below(93);
    });

    it('need partial update, visible', () => {
      const priority = computePriority(
        -1, now-20000, now-20000,
        { startIndex: -1, endIndex: 2 },
        30000,
        5000,
        RequestTypes.OVERVIEW
      );
      expect(priority).to.be.above(59);
      expect(priority).to.be.below(62);
    });

  });

  describe('createPrioritizedTaskList', () => {
    it('no previous fetch', () => {
      const context = {
        props: {
          ...dataManagerStaticConfig,
          currentDataLayoutType: null,
          lastMemberListFetchTimeMs: null,
          data: null
        }
      };
      const tasks = (new DataManager).createPrioritizedTaskList.bind(context)();
      expect(tasks.length).to.eql(1);
      expect(tasks[0].type).to.eql(RequestTypes.MEMBER_LIST);
      expect(tasks[0].priority).to.be.above(10000);
    });

    it('a scenario immediately after MEMBER_LIST fetch.', () => {
      const context = {
        props: {
          ...dataManagerStaticConfig,
          currentDataLayoutType: DataLayoutTypes.CONSUMER_GROUPS,
          currentDataLayoutId: null,
          routedDataLayoutType: DataLayoutTypes.CONSUMER_GROUPS,
          routedDataLayoutId: null,
          visibleMembers: { startIndex: 0, endIndex: 1 },
          lastMemberListFetchTimeMs: Date.now(),
          data: {
            destinations: [],
            sources: [],
            members: [
              {
                group: 'my-group1'
              },
              {
                group: 'my-group2'
              },
              {
                group: 'my-group3'
              }
            ],
            rollup: 'FIFTEEN_SECONDS',
            stepMs: 15000,
            type: DataLayoutTypes.CONSUMER_GROUPS
          }
        }
      };
      const tasks = (new DataManager).createPrioritizedTaskList.bind(context)();
      expect(tasks.length).to.eql(4);
      expect(tasks[0].fullRefresh).to.eql(true);
      expect(tasks[1].fullRefresh).to.eql(true);
      expect(tasks[2].fullRefresh).to.eql(true);
      expect(tasks[3].fullRefresh).to.eql(true);
      expect(tasks[0].member.group).to.eql('my-group1');
      expect(tasks[1].member.group).to.eql('my-group2');
      expect(tasks[2].type).to.eql('OVERVIEW');
      expect(tasks[3].member.group).to.eql('my-group3');
    });
  });

  describe('fetchNextData', () => {
    it('dataManager.preventFetch === true', () => {
      resetLoadDataCallCount();
      const context = {
        props: {
          preventFetch: true,
          routedDataLayoutType: DataLayoutTypes.CONSUMER_GROUPS,
          connectionCount: 0,
          loadData: loadDataMock
        }
      };
      (new DataManager).fetchNextData.bind(context)();
      expect(loadDataCallCount).to.eql(0);
    });

    it('no routedDataLayoutType', () => {
      resetLoadDataCallCount();
      const context = {
        props: {
          preventFetch: false,
          routedDataLayoutType: null,
          connectionCount: 0,
          loadData: loadDataMock
        }
      };
      (new DataManager).fetchNextData.bind(context)();
      expect(loadDataCallCount).to.eql(0);
    });

    it('connectionCount exceeded', () => {
      resetLoadDataCallCount();
      const context = {
        props: {
          preventFetch: false,
          routedDataLayoutType: null,
          connectionCount: MAX_CONNECTION_COUNT,
          loadData: loadDataMock
        }
      };
      (new DataManager).fetchNextData.bind(context)();
      expect(loadDataCallCount).to.eql(0);
    });
  });

  it('mounts', () => {
    spy(DataManager.prototype, 'componentWillMount');
    spy(DataManager.prototype, 'componentWillUnmount');

    const wrapper = mount(
      <DataManager
        preventFetch={false}
        currentlyInError={false}
        chartAreaWidthPx={1000}
        currentDataBinSizeMs={null}
        desiredDataQuantizationMs={15000}
        desiredRollup={'FIFTEEN_SECONDS'}
        routeDataLayoutId={null}
        currentDataLayoutId={null}
        routedDataLayoutId={null}
        currentDataLayoutType={null}
        routedDataLayoutType={null}
        startTimeMs={new Date(0)}
        fetchStopTimeMs={new Date(1000)}
        rangeType={RangeTypes.FIXED}
        refreshCycle={1}
        connectionCount={1}
        data={null}
        lastMemberListFetchTimeMs={0}
        lastOverviewDataFetchTimeMs={0}
        visibleMembers={{ startIndex: 0, endIndex: 0 }}
        loadData={() => {
          return () => {
            return ({ a: null });
          };
        }}
        clearData={() => {}}
      />
    );
    expect(DataManager.prototype.componentWillMount.calledOnce).to.equal(true);
    wrapper.unmount();
    expect(DataManager.prototype.componentWillUnmount.calledOnce).to.equal(true);
  });

});
