// @flow
// (Copyright) [2016 - 2016] Confluent, Inc.

import { constructMembers } from 'reducers/monitoring/data';

describe('constructMembers', () => {
  it('no members', () => {
    const members = [];
    const updateMembers = [];
    const memberIdNames = ['clientId'];
    const result = constructMembers(members, updateMembers, memberIdNames);
    expect(result).to.not.be.null;
    expect(result.length).to.eql(0);
  });

  it('equal members', () => {
    const members = [
      {
        clientId: 'myclient',
        produced: { count: [0, 1, 2, 3] },
        consumed: { count: [10, 20, 30, 40] }
      }
    ];
    const updateMembers = [
      {
        clientId: 'myclient'
      }
    ];
    const memberIdNames = ['clientId'];
    const result = constructMembers(members, updateMembers, memberIdNames);
    expect(result).to.not.be.null;
    expect(result.length).to.eql(1);
    expect(result[0].clientId).to.eql('myclient');
  });

  it('different members', () => {
    const members = [
      {
        clientId: 'myclient1',
        produced: { count: [0, 1, 2, 3] },
        consumed: { count: [10, 20, 30, 40] }
      }
    ];
    const updateMembers = [
      {
        clientId: 'myclient2'
      }
    ];
    const memberIdNames = ['clientId'];
    const result = constructMembers(members, updateMembers, memberIdNames);
    expect(result).to.not.be.null;
    expect(result.length).to.eql(1);
    // myclient2 is positioned first because there is no member before it in the update.
    expect(result[0].clientId).to.eql('myclient2');
    expect(result[0].produced).to.not.exist;
  });

  it('empty updateMembers', () => {
    const members = [
      {
        clientId: 'myclient1',
        produced: { count: [0, 1, 2, 3] },
        consumed: { count: [10, 20, 30, 40] }
      }
    ];
    const updateMembers = [];
    const memberIdNames = ['clientId'];
    const result = constructMembers(members, updateMembers, memberIdNames);
    expect(result).to.not.be.null;
    expect(result.length).to.eql(0);
  });


  it('empty members', () => {
    const members = [];
    const updateMembers = [
      {
        clientId: 'myclient1'
      }
    ];
    const memberIdNames = ['clientId'];
    const result = constructMembers(members, updateMembers, memberIdNames);
    expect(result).to.not.be.null;
    expect(result.length).to.eql(1);
    expect(result[0].produced).to.not.exist;
  });

  it('swapped', () => {
    const members = [
      { clientId: 'myclient1', produced: { count: [1, 2, 3, 4] }, consumed: { count: [10, 20, 30, 40] } },
      { clientId: 'myclient2', produced: { count: [2, 4, 6, 8] }, consumed: { count: [20, 40, 60, 80] } }
    ];
    const updateMembers = [
      { clientId: 'myclient2' },
      { clientId: 'myclient1' }
    ];
    const memberIdNames = ['clientId'];
    const result = constructMembers(members, updateMembers, memberIdNames);
    expect(result).to.not.be.null;
    expect(result.length).to.eql(2);
    expect(result[0].clientId).to.eql('myclient2');
    expect(result[1].clientId).to.eql('myclient1');
    expect(result[0].produced.count[0]).to.eql(2);
    expect(result[0].produced.count[1]).to.eql(4);
    expect(result[1].produced.count[0]).to.eql(1);
    expect(result[1].produced.count[1]).to.eql(2);
    expect(result[0].consumed.count[0]).to.eql(20);
    expect(result[0].consumed.count[1]).to.eql(40);
    expect(result[1].consumed.count[0]).to.eql(10);
    expect(result[1].consumed.count[1]).to.eql(20);
  });

  it('order test 1', () => {
    const members = [
      { clientId: 'myclient1', produced: { count: [0, 1, 2, 3] }, consumed: { count: [10, 20, 30, 40] } },
      { clientId: 'myclient2', produced: { count: [0, 1, 2, 3] }, consumed: { count: [10, 20, 30, 40] } },
      { clientId: 'myclient3', produced: { count: [0, 1, 2, 3] }, consumed: { count: [10, 20, 30, 40] } }
    ];
    const updateMembers = [
      { clientId: 'myclient1' },
      { clientId: 'myclient4' }
    ];
    const memberIdNames = ['clientId'];
    const result = constructMembers(members, updateMembers, memberIdNames);
    expect(result).to.not.be.null;
    expect(result.length).to.eql(2);
    expect(result[0].clientId).to.eql('myclient1');
    expect(result[1].clientId).to.eql('myclient4');
  });

  it('order test 2', () => {
    const members = [
      { clientId: 'myclient1', produced: { count: [0, 1, 2, 3] }, consumed: { count: [10, 20, 30, 40] } },
      { clientId: 'myclient2', produced: { count: [0, 1, 2, 3] }, consumed: { count: [10, 20, 30, 40] } },
      { clientId: 'myclient3', produced: { count: [0, 1, 2, 3] }, consumed: { count: [10, 20, 30, 40] } }
    ];
    const updateMembers = [
      { clientId: 'myclient1' },
      { clientId: 'myclient4' },
      { clientId: 'myclient3' }
    ];
    const memberIdNames = ['clientId'];
    const result = constructMembers(members, updateMembers, memberIdNames);
    expect(result).to.not.be.null;
    expect(result.length).to.eql(3);
    expect(result[0].clientId).to.eql('myclient1');
    expect(result[1].clientId).to.eql('myclient4');
    expect(result[2].clientId).to.eql('myclient3');
  });

  it('order test 3', () => {
    const members = [
      { clientId: 'myclient3', produced: { count: [0, 1, 2, 3] }, consumed: { count: [10, 20, 30, 40] } }
    ];
    const updateMembers = [
      { clientId: 'myclient1' },
      { clientId: 'myclient2' },
      { clientId: 'myclient3' }
    ];
    const memberIdNames = ['clientId'];
    const result = constructMembers(members, updateMembers, memberIdNames);
    expect(result).to.not.be.null;
    expect(result.length).to.eql(3);
    expect(result[0].clientId).to.eql('myclient1');
    expect(result[1].clientId).to.eql('myclient2');
    expect(result[2].clientId).to.eql('myclient3');
  });

  it('multiple memberIdNames', () => {
    const members = [
      { topic: 'topic1', partition: 0, produced: { count: [0, 1, 2, 3] }, consumed: { count: [10, 20, 30, 40] } },
      { topic: 'topic1', partition: 1, produced: { count: [4, 1, 2, 3] }, consumed: { count: [100, 200, 300, 400] } },
      { topic: 'topic2', partition: 0, produced: { count: [5, 1, 2, 3] }, consumed: { count: [1000, 2000, 3000, 4000] } }
    ];
    const updateMembers = [
      { topic: 'topic1', partition: 0 }
    ];
    const memberIdNames = ['topic', 'partition'];
    const result = constructMembers(members, updateMembers, memberIdNames);
    expect(result).to.not.be.null;
    expect(result.length).to.eql(1);
    expect(result[0].topic).to.eql('topic1');
    expect(result[0].partition).to.eql(0);
    expect(result[0].produced.count[0]).to.eql(0);
  });

});
