// @flow
// (Copyright) [2016 - 2016] Confluent, Inc.
import { call, fork, put } from 'redux-saga/effects';
import { takeLatest } from 'redux-saga';

import * as Sagas from 'sagas/management/connectors';
import * as ConnectorActions from 'actions/connectors';

const {
  connectorsSaga,
  deleteConnectorSaga,
  fetchConnector,
  fetchConnectorList,
  loadConnectorListSaga,
  loadConnectorSaga,
  loadConnectorStatusSaga,
  pauseConnectorSaga,
  resumeConnectorSaga,
  saveConnector,
  saveConnectorSaga
} = Sagas;

describe('sagas/management/connectors', () => {
  const name = 'my-connector';
  const connectorResponse = {
    config: {
      'connector.class': 'some-connector-class',
      name
    },
    name,
    tasks: [],
    type: 'SINK'
  };
  const error = new Error('dope error');
  let generator;
  let next;

  describe('saveConnector', () => {
    const connectorConfig = connectorResponse.config;

    beforeEach(() => {
      const initialAction = ConnectorActions.saveConnector(connectorConfig);
      generator = saveConnector(initialAction);
      next = generator.next();
    });

    it('should yield a request action', () => {
      expect(
        next.value
      ).to.deep.equal(
        put(ConnectorActions.save.request(connectorConfig.name))
      );
    });

    it('should yield a failure action', () => {
      // move past request
      generator.next();

      expect(
        generator.throw(error).value
      ).to.deep.equal(
        put(ConnectorActions.save.failure(connectorConfig.name, error))
      );
    });

    it('should yield a success action', () => {
      // move past request
      generator.next();

      expect(
        generator.next(connectorResponse).value
      ).to.deep.equal(
        put(ConnectorActions.save.success(connectorResponse.name, connectorResponse))
      );
    });

    it('should finish after success or error', () => {
      // move past request
      generator.next();

      // move past failure and success
      generator.next();

      expect(
        generator.next()
      ).to.deep.equal({
        done: true,
        value: undefined
      });
    });
  });

  describe('fetchConnector', () => {

    beforeEach(() => {
      const initialAction = ConnectorActions.loadConnector(name);
      generator = fetchConnector(initialAction);
      next = generator.next();
    });

    it('should yield a request action', () => {
      expect(
        next.value
      ).to.deep.equal(
        put(ConnectorActions.fetch.request(name))
      );
    });

    it('should yield a failure action', () => {
      // move past request
      generator.next();

      expect(
        generator.throw(error).value
      ).to.deep.equal(
        put(ConnectorActions.fetch.failure(name, error))
      );
    });

    it('should yield a success action', () => {

      // move past request
      generator.next();

      expect(
        generator.next(connectorResponse).value
      ).to.deep.equal(
        put(ConnectorActions.fetch.success(name, connectorResponse))
      );
    });

    it('should finish after success or error', () => {
      // move past request
      generator.next();

      // move past failure and success
      generator.next();

      expect(
        generator.next()
      ).to.deep.equal({
        done: true,
        value: undefined
      });
    });
  });

  describe('fetchConnectorList', () => {
    beforeEach(() => {
      const initialAction = ConnectorActions.loadConnectorList();
      generator = fetchConnectorList(initialAction);
      next = generator.next();
    });

    it('should yield a request action', () => {
      expect(
        next.value
      ).to.deep.equal(
        put(ConnectorActions.listFetch.request())
      );
    });

    it('should yield a failure action', () => {
      // move past request
      generator.next();

      expect(
        generator.throw(error).value
      ).to.deep.equal(
        put(ConnectorActions.listFetch.failure(error))
      );
    });

    it('should yield a success action', () => {
      const connectorList = [{
        name: 'a connector',
        topics: 'my,topics',
        tasks: 2,
        type: 'SINK'
      }];

      // move past request
      generator.next();

      expect(
        generator.next(connectorList).value
      ).to.deep.equal(
        put(ConnectorActions.listFetch.success(connectorList))
      );
    });

    it('should finish after success or error', () => {
      // move past request
      generator.next();

      // move past failure and success
      generator.next();

      expect(
        generator.next()
      ).to.deep.equal({
        done: true,
        value: undefined
      });
    });
  });

  describe('saveConnectorSaga', () => {
    beforeEach(() => {
      generator = saveConnectorSaga();
      next = generator.next();
    });

    it('takes latest SAVE_CONNECTOR action and start `saveConnector`', () => {
      expect(
        next.value
      ).to.deep.equal(call(takeLatest, ConnectorActions.SAVE_CONNECTOR, saveConnector));
    });

    it('should finish', () => {
      next = generator.next();

      expect(
        next
      ).to.deep.equal({
        done: true,
        value: undefined
      });
    });
  });

  describe('loadConnectorSaga', () => {
    beforeEach(() => {
      generator = loadConnectorSaga();
      next = generator.next();
    });

    it('takes latest LOAD action and start `fetchConnector`', () => {
      expect(
        next.value
      ).to.deep.equal(call(takeLatest, ConnectorActions.LOAD_CONNECTOR, fetchConnector));
    });

    it('should finish', () => {
      next = generator.next();

      expect(
        next
      ).to.deep.equal({
        done: true,
        value: undefined
      });
    });
  });

  describe('loadConnectorListSaga', () => {
    beforeEach(() => {
      generator = loadConnectorListSaga();
      next = generator.next();
    });

    it('takes latest LOAD_CONNECTOR_LIST action and starts `fetchConnectorList`', () => {
      expect(
        next.value
      ).to.deep.equal(call(takeLatest, [
        ConnectorActions.CONNECTOR_DESTROY_SUCCESS,
        ConnectorActions.LOAD_CONNECTOR_LIST
      ], fetchConnectorList));
    });

    it('should finish', () => {
      next = generator.next();

      expect(
        next
      ).to.deep.equal({
        done: true,
        value: undefined
      });
    });
  });

  describe('connectorsSaga', () => {
    beforeEach(() => {
      generator = connectorsSaga();
    });

    it('forks correct sagas', () => {
      next = generator.next();
      expect(
        next.value,
        'forks `deleteConnectorSaga`'
      ).to.deep.equal(fork(deleteConnectorSaga));


      next = generator.next();
      expect(
        next.value,
        'forks `loadConnectorListSaga`'
      ).to.deep.equal(fork(loadConnectorListSaga));


      next = generator.next();
      expect(
        next.value,
        'forks `loadConnectorSaga`'
      ).to.deep.equal(fork(loadConnectorSaga));

      next = generator.next();
      expect(
        next.value,
        'forks `loadConnectorStatusSaga`'
      ).to.deep.equal(fork(loadConnectorStatusSaga));

      next = generator.next();
      expect(
        next.value,
        'forks `pauseConnectorSaga`'
      ).to.deep.equal(fork(pauseConnectorSaga));

      next = generator.next();
      expect(
        next.value,
        'forks `resumeConnectorSaga`'
      ).to.deep.equal(fork(resumeConnectorSaga));

      next = generator.next();
      expect(
        next.value,
        'forks `saveConnectorSaga`'
      ).to.deep.equal(fork(saveConnectorSaga));

      next = generator.next();

      expect(
        next,
        'should be finished'
      ).to.deep.equal({
        done: true,
        value: undefined
      });
    });
  });
});
