// @flow
// (Copyright) [2016 - 2016] Confluent, Inc.
import {
  dataSelector,
  numberToExtendArrayAtStart,
  numberToExtendArrayAtEnd,
  alignDataAndChartStartStopTimes,
  transformMissingData
} from 'selectors/monitoring/data';

import { RangeTypes } from 'constants/monitoring/timeframe';


const state = {
  monitoring: {
    layout: {
      chartAreaWidthPx: 1000
    },
    timeframe: {
      rangeType: RangeTypes.ROLLING,
      startTimeMs: 14587511015000,
      stopTimeMs: 14587511015000 + (240*60*1000)
    },
    data: {
      overview: {
        startTimeMs: 14587511015000,
        consumed: {
          count: [
            46,
            48,
            46,
            45,
            50,
            53,
            48,
            50,
            50
          ],
          maxLatency: [
            5865,
            6302,
            7056,
            6889,
            8894,
            8110,
            5558,
            8737,
            8279
          ],
          minLatency: [
            2902,
            816,
            1578,
            2406,
            4788,
            4279,
            1203,
            4484,
            4193
          ],
          totalLatency: [
            191314,
            185107,
            185990,
            210387,
            329308,
            309756,
            159149,
            322337,
            308883
          ]
        },
        produced: {
          count: [
            46,
            48,
            46,
            45,
            50,
            53,
            48,
            50,
            50
          ]
        }
      },
      group: 'other-group',
      members: [
        {
          clientId: 'other-id-1',
          startTimeMs: 14587511015000,
          consumed: {
            count: [
              46,
              48,
              46,
              45,
              50,
              53,
              48,
              50,
              50
            ],
            maxLatency: [
              5865,
              6302,
              7056,
              6889,
              8894,
              8110,
              5558,
              8737,
              8279
            ],
            minLatency: [
              2902,
              816,
              1578,
              2406,
              4788,
              4279,
              1203,
              4484,
              4193
            ],
            totalLatency: [
              191314,
              185107,
              185990,
              210387,
              329308,
              309756,
              159149,
              322337,
              308883
            ]
          }
        }
      ],
      rollup: 'FIFTEEN_SECONDS',
      stepMs: 15000,
      dataLayoutType: 'CONSUMER_GROUP'
    }
  }
};

describe('metric data', () => {
  it('data', () => {
    const data = dataSelector(state);
    expect(data).to.exist;
    expect(data.overview.consumed).to.exist;
    expect(data.overview.produced).to.exist;
    expect(data.members).to.exist;
    expect(data.members[0]).to.exist;
    expect(data.members[0].consumed.count).to.exist;
    expect(data.members[0].consumed.latency).to.exist;
    expect(data.members[0].consumed.latency.values).to.exist;
    expect(Object.keys(data.members[0].consumed.latency.values).length).to.eql(4);
  });
});

describe('numberToExtendArrayAtStart', () => {
  it('none', () => {
    expect(numberToExtendArrayAtStart(1000000, 15000, 1000000)).to.eql(0);
  });

  it('extend', () => {
    expect(numberToExtendArrayAtStart(1015000, 15000, 1000000)).to.eql(1);
  });

  it('contract', () => {
    expect(numberToExtendArrayAtStart(985000, 15000, 1000000)).to.eql(-1);
  });
});

describe('numberToExtendArrayAtEnd', () => {
  it('none', () => {
    expect(numberToExtendArrayAtEnd(1, 1000000, 15000, 1015000)).to.eql(0);
  });

  it('extend', () => {
    expect(numberToExtendArrayAtEnd(2, 1000000, 15000, 1045000)).to.eql(1);
  });

  it('contract', () => {
    expect(numberToExtendArrayAtEnd(2, 1000000, 15000, 1015000)).to.eql(-1);
  });
});

describe('alignDataAndChartStartStopTimes', () => {
  it('already aligned', () => {
    const input = {
      count: [1, 2, 3, 4, 5]
    };

    const output = alignDataAndChartStartStopTimes(input, 0, 15000, 0, 75000);

    expect(output.values.count.length).to.eql(5);
    expect(output.values.count[0]).to.eql(1);
    expect(output.values.count[4]).to.eql(5);
  });

  it('shifted', () => {
    const input = {
      count: [1, 2, 3, 4, 5]
    };

    const output = alignDataAndChartStartStopTimes(input, 0, 15000, 15000, 30000);

    expect(output.values.count.length).to.eql(1);
    expect(output.values.count[0]).to.eql(2);
  });

  it('extend, shift', () => {
    const input = {
      count: [1, 2, 3, 4, 5]
    };

    const output = alignDataAndChartStartStopTimes(input, 15000, 15000, 0, 115000);

    expect(output.values.count.length).to.eql(7);
    expect(output.values.count[0]).to.eql(null);
    expect(output.values.count[1]).to.eql(1);
    expect(output.values.count[6]).to.eql(null);
  });

});

function makeData(produced, consumed) {
  const data = {};
  if (produced) {
    data.produced = {
      missing: [false, false, true, true, false, false],
      count: {
        startTimeMs: 0,
        stepMs: 15000,
        values: [0, 0, 0, 0, 0, 0]
      }
    };
  }
  if (consumed) {
    data.consumed = {
      missing: [false, false, true, true, false, false],
      count: {
        startTimeMs: 0,
        stepMs: 15000,
        values: [0, 0, 0, 0, 0, 0]
      }
    };
  }
  return data;
}

describe('transformMissingData', () => {

  it('missing arrays not present', () => {
    const data = {};
    transformMissingData(data);
    expect(data.missing).to.exist;
    expect(data.missing).to.have.length(0);
  });

  it('only produced present', () => {
    const data = makeData(true, false);
    transformMissingData(data);
    expect(data.missing).to.exist;
    expect(data.missing).to.have.length(1);
  });

  it('only consumed present', () => {
    const data = makeData(false, true);
    transformMissingData(data);
    expect(data.missing).to.exist;
    expect(data.missing).to.have.length(1);
  });

  it('only produced missing present', () => {
    const data = makeData(true, true);
    delete data.consumed.missing;
    transformMissingData(data);
    expect(data.missing).to.exist;
    expect(data.missing).to.have.length(1);
  });

  it('only consumed missing present', () => {
    const data = makeData(true, true);
    delete data.produced.missing;
    transformMissingData(data);
    expect(data.missing).to.exist;
    expect(data.missing).to.have.length(1);
  });

  it('produced and consumed have missing, data OR\'d (1)', () => {
    const data = makeData(true, true);
    data.produced.missing = [false, false, true, true, false, false];
    data.consumed.missing = [false, false, false, false, false, false];
    transformMissingData(data);
    expect(data.missing).to.exist;
    expect(data.missing).to.have.length(1);
    expect(data.missing[0].endTime.getTime() - data.missing[0].startTime.getTime()).to.eql(30000);
  });

  it('produced and consumed have missing, data OR\'d (2)', () => {
    const data = makeData(true, true);
    data.produced.missing = [false, false, true, false, false, false];
    data.consumed.missing = [false, false, false, true, false, false];
    transformMissingData(data);
    expect(data.missing).to.exist;
    expect(data.missing).to.have.length(1);
    expect(data.missing[0].endTime.getTime() - data.missing[0].startTime.getTime()).to.eql(30000);
  });

  it('produced and consumed have missing, data OR\'d (3)', () => {
    const data = makeData(true, true);
    data.produced.missing = [false, false, true, false, false, false];
    data.consumed.missing = [false, false, true, false, false, false];
    transformMissingData(data);
    expect(data.missing).to.exist;
    expect(data.missing).to.have.length(1);
    expect(data.missing[0].startTime.getTime()).to.eql(30000);
    expect(data.missing[0].endTime.getTime() - data.missing[0].startTime.getTime()).to.eql(15000);
  });

  it('missing at start', () => {
    const data = makeData(true, true);
    data.produced.missing = [true, true, false, false, false, false];
    data.consumed.missing = [false, false, false, false, false, false];
    transformMissingData(data);
    expect(data.missing).to.exist;
    expect(data.missing).to.have.length(1);
    expect(data.missing[0].startTime.getTime()).to.eql(0);
    expect(data.missing[0].endTime.getTime() - data.missing[0].startTime.getTime()).to.eql(30000);
  });

  it('missing at end', () => {
    const data = makeData(true, true);
    data.produced.missing = [false, false, false, true, false, true];
    data.consumed.missing = [false, false, false, false, true, false];
    transformMissingData(data);
    expect(data.missing).to.exist;
    expect(data.missing).to.have.length(1);
    expect(data.missing[0].endTime.getTime() - data.missing[0].startTime.getTime()).to.eql(45000);
  });

  it('missing full range', () => {
    const data = makeData(true, true);
    data.produced.missing = [true, true, true, true, true, true];
    transformMissingData(data);
    expect(data.missing).to.exist;
    expect(data.missing).to.have.length(1);
    expect(data.missing[0].endTime.getTime() - data.missing[0].startTime.getTime()).to.eql(90000);
  });

  it('more than one missing region', () => {
    const data = makeData(true, true);
    data.produced.missing = [true, false, true, true, false, true];
    data.consumed.missing = [false, false, false, false, false, false];
    transformMissingData(data);
    expect(data.missing).to.exist;
    expect(data.missing).to.have.length(3);
    expect(data.missing[0].endTime.getTime() - data.missing[0].startTime.getTime()).to.eql(15000);
    expect(data.missing[1].endTime.getTime() - data.missing[1].startTime.getTime()).to.eql(30000);
    expect(data.missing[2].endTime.getTime() - data.missing[2].startTime.getTime()).to.eql(15000);
  });

  it('missing region only 1 bin long', () => {
    const data = makeData(true, true);
    data.produced.missing = [false, false, false, true, false, false];
    data.consumed.missing = [false, false, false, false, false, false];
    transformMissingData(data);
    expect(data.missing).to.exist;
    expect(data.missing).to.have.length(1);
    expect(data.missing[0].endTime.getTime() - data.missing[0].startTime.getTime()).to.eql(15000);
  });

  it('no missing', () => {
    const data = makeData(true, true);
    data.produced.missing = [false, false, false, false, false, false];
    data.consumed.missing = [false, false, false, false, false, false];
    transformMissingData(data);
    expect(data.missing).to.exist;
    expect(data.missing).to.have.length(0);
  });
});
