// @flow
// (Copyright) [2016 - 2016] Confluent, Inc.
import { Y_AXIS_WIDTH_PX } from 'constants/monitoring/layout';

import {
  startTimeDataQuantizedSelector,
  stopTimeDataQuantizedSelector,
  slotWidthMsSelector,
  slotWidthPxSelector,
  startTimeSlotQuantizedSelector,
  stopTimeSlotQuantizedSelector,
  desiredRollupNoCountLimit,
  visibleMembersSelector
} from 'selectors/monitoring';

import {
  Rollups
} from 'types/monitoring';

const testState1 = {
  monitoring: {
    timeframe: {
      startTimeMs: 0,
      stopTimeMs: (10*60*1000),
      rangeType: null
    },
    layout: {
      chartAreaWidthPx: 1000 + Y_AXIS_WIDTH_PX
    }
  }
};

const testState2 = {
  monitoring: {
    timeframe: {
      startTimeMs: 4,
      stopTimeMs: (10*60*1000 + 4),
      rangeType: null
    },
    layout: {
      chartAreaWidthPx: 75 + Y_AXIS_WIDTH_PX
    }
  }
};

describe('data quantized', () => {

  it('startTimeDataQuantized quantized input', () => {
    expect(startTimeDataQuantizedSelector(testState1)).to.eql(new Date(0));
  });

  it('startTimeDataQuantized general input', () => {
    expect(startTimeDataQuantizedSelector(testState2)).to.eql(new Date(15000));
  });

  it('stopTimeDataQuantized quantized input', () => {
    expect(stopTimeDataQuantizedSelector(testState1)).to.eql(new Date(10*60*1000));
  });

  it('stopTimeDataQuantized general input', () => {
    expect(stopTimeDataQuantizedSelector(testState2)).to.eql(new Date(10*60*1000 + 15000));
  });

  it('slotWidthMs simple', () => {
    // axis width is 1000px -> maximum 250 slots.
    // interval is 10 minutes -> 40 seconds per slot.
    // but quantization is 15s, so slot width will be 15s.
    expect(slotWidthMsSelector(testState1)).to.eql(15000);
  });

  it('slotWidthPx simple', () => {
    // axis width is 1000px -> maximum 250 slots.
    // interval is 10 minutes -> 40 seconds per slot.
    // but quantization is 15s, so slot width will be 15s.
    // => 40 slots
    // => 25
    expect(slotWidthPxSelector(testState1)).to.equal(25);
  });

  it('slotWidthMs aggregated', () => {
    // axis width is 75px -> maximum 18 slots.
    // interval is 10 minutes -> 33.333s per slotQuantization
    // but quantization is 15s, so slot width will be 30s.
    expect(slotWidthMsSelector(testState2)).to.eql(45000);
  });

  it('slotWidthPx aggregated', () => {
    // axis width is 195px -> maximum 48 slots.
    // interval is 10 minutes -> 33.333s per slotQuantization
    // but quantization is 15s, so slot width will be 15s.
    // => 18 slots
    // => 5.769230769230769
    expect(slotWidthPxSelector(testState2)).to.almost.equal(5.769230769230769);
  });

  it('startTimeSlotQuantized quantized input', () => {
    expect(startTimeSlotQuantizedSelector(testState1)).to.eql(new Date(0));
  });

  it('startTimeSlotQuantized general input', () => {
    expect(startTimeSlotQuantizedSelector(testState2)).to.eql(new Date(45000));
  });

  it('stopTimeSlotQuantized quantized input', () => {
    expect(stopTimeSlotQuantizedSelector(testState1)).to.eql(new Date(10*60*1000));
  });

  it('stopTimeSlotQuantized general input', () => {
    expect(stopTimeSlotQuantizedSelector(testState2)).to.eql(new Date(10*60*1000 + 30000));
  });

  it('desiredRollup tiny range', () => {
    const state = {
      monitoring: {
        layout: {
          chartAreaWidthPx: 1000
        },
        timeframe: {
          startTimeMs: 0,
          stopTimeMs: 100000
        }
      }
    };
    expect(desiredRollupNoCountLimit(state)).to.eql(Rollups.FIFTEEN_SECONDS);
  });

  it('desiredRollup 5 seconds', () => {
    const state = {
      monitoring: {
        layout: {
          chartAreaWidthPx: 1000
        },
        timeframe: {
          startTimeMs: 0,
          stopTimeMs: 2000000
        }
      }
    };
    expect(desiredRollupNoCountLimit(state)).to.eql(Rollups.FIFTEEN_SECONDS);
  });

  it('desiredRollup 1 hour', () => {
    const state = {
      monitoring: {
        layout: {
          chartAreaWidthPx: 1000
        },
        timeframe: {
          startTimeMs: 0,
          stopTimeMs: 24000000000
        }
      }
    };
    expect(desiredRollupNoCountLimit(state)).to.eql(Rollups.ONE_HOUR);
  });

  it('desiredRollup 1 week', () => {
    const state = {
      monitoring: {
        layout: {
          chartAreaWidthPx: 100
        },
        timeframe: {
          startTimeMs: 0,
          stopTimeMs: 24000000000
        }
      }
    };
    expect(desiredRollupNoCountLimit(state)).to.eql(Rollups.ONE_WEEK);
  });
});

describe('visibleMembers', () => {

  it('undefined members', () => {
    const state = {
      monitoring: {
        layout: {
          pageScrollOffsetY: 0,
          pageHeight: 1000,
          windowHeight: 500
        },
        data: {
        }
      }
    };
    const range = visibleMembersSelector(state);
    expect(range.startIndex).to.eql(-1);
    expect(range.endIndex).to.eql(-1);
  });

  it('no members', () => {
    const state = {
      monitoring: {
        layout: {
          pageScrollOffsetY: 0,
          pageHeight: 1000,
          windowHeight: 500
        },
        data: {
          members: []
        }
      }
    };
    const range = visibleMembersSelector(state);
    expect(range.startIndex).to.eql(-1);
    expect(range.endIndex).to.eql(-1);
  });

  it('many members, start range', () => {
    const state = {
      monitoring: {
        layout: {
          pageScrollOffsetY: 0,
          pageHeight: 24000,
          windowHeight: 500
        },
        data: {
        }
      }
    };
    state.monitoring.data.members = new Array(100).fill(null);

    const range = visibleMembersSelector(state);
    expect(range.startIndex).to.eql(-1);
    expect(range.endIndex).to.eql(3);
  });

  it('many members, mid range', () => {
    const state = {
      monitoring: {
        layout: {
          pageScrollOffsetY: 4000,
          pageHeight: 24000,
          windowHeight: 500
        },
        data: {
        }
      }
    };
    state.monitoring.data.members = new Array(100).fill(null);

    const range = visibleMembersSelector(state);
    expect(range.startIndex).to.eql(14);
    expect(range.endIndex).to.eql(18);
  });

  it('many members, end range', () => {
    const state = {
      monitoring: {
        layout: {
          pageScrollOffsetY: 23500,
          pageHeight: 24000,
          windowHeight: 500
        },
        data: {
        }
      }
    };
    state.monitoring.data.members = new Array(100).fill(null);

    const range = visibleMembersSelector(state);
    expect(range.startIndex).to.eql(97);
    expect(range.endIndex).to.eql(99);
  });

});
