// @flowfixme
// (Copyright) [2016 - 2016] Confluent, Inc.
import autoprefixer from 'autoprefixer';
import path from 'path';
import postcssModulesValues from 'postcss-modules-values';
import HtmlWebpackPlugin from 'html-webpack-plugin';
import ExtractTextPlugin from 'extract-text-webpack-plugin';
import webpack from 'webpack';
import pkg from './package.json';

export const PATHS = {
  buildDir: path.join(__dirname, 'dist'),
  staticDir: path.join(__dirname, 'src', 'static'),
  srcDir: path.join(__dirname, 'src'),
  testDir: path.join(__dirname, 'tests')
};

const { EDITOR_SCHEME, NODE_ENV, USE_REDBOX } = process.env;

const env = NODE_ENV || 'development';

export const envGlobals = {
  'process.env': {
    NODE_ENV: JSON.stringify(env),
    EDITOR_SCHEME: JSON.stringify(EDITOR_SCHEME)
  },
  NODE_ENV: env,
  __USE_REDBOX__: Boolean(USE_REDBOX) || Boolean(EDITOR_SCHEME),
  __DEV__: env === 'development',
  __PROD__: env === 'production',
  __TEST__: env === 'test',
  __VERSION__: JSON.stringify(pkg.version)
};

function createLoaderString(loaderName: string, options?: {}) {
  if (!options) {
    return loaderName;
  } else {
    return `${loaderName}?${JSON.stringify(options)}`;
  }
}

const urlLoaderBaseOptions = {
  digest: 'hex',
  hash: 'sha512',
  limit: 100000,
  name: '[name]-[hash].[ext]'
};

const cssLoaderString = createLoaderString('css', {
  minimize: true,
  modules: true,
  importLoaders: 1,
  localIdentName: '[name]__[local]___[hash:base64:5]',
  sourceMap: true
});

const cssLoaders = [
  cssLoaderString,
  'postcss'
];

const lessLoaders = [
  cssLoaderString,
  createLoaderString('less', {
    sourceMap: true
  }),
  'postcss'
];

export const loaders = {
  babel: {
    test: /\.jsx?$/,
    loaders: [
      'babel'
    ],
    include: [
      PATHS.srcDir,
      PATHS.testDir
    ],
    exclude: /node_modules/
  },
  fonts: {
    test: /\.woff2?(\?v=\d+\.\d+\.\d+)?$/,
    loaders: [
      createLoaderString('url', {
        ...urlLoaderBaseOptions,
        limit: 100000,
        mimetype: 'application/font-woff'
      })
    ]
  },

  // these get inlined
  svgs: {
    test: /\.svg$/,
    loaders: [
      'raw',
      'image-webpack'
    ]
  },
  images: {
    test: /\.(gif|jpe?g)$/,
    loaders: [
      createLoaderString('url', urlLoaderBaseOptions),
      'image-webpack'
    ]
  },
  png: {
    test: /\.(png)$/,
    loaders: [
      createLoaderString('url', urlLoaderBaseOptions)
    ]
  },
  json: {
    test: /\.json$/,
    loaders: [
      'json'
    ]
  },
  less: {
    test: /\.less$/,
    loaders: lessLoaders
  },
  extractLess: {
    test: /\.less$/,
    loader: ExtractTextPlugin.extract('style', lessLoaders)
  },
  css: {
    test: /\.css$/,
    loaders: cssLoaders
  },
  extractCss: {
    test: /\.css$/,
    loader: ExtractTextPlugin.extract('style', cssLoaders)
  }
};

export const baseWebpackConfig = {
  cache: true,
  entry: [
    path.resolve('./src/client.jsx')
  ],
  output: {
    filename: 'app.js',
    path: PATHS.buildDir,
    publicPath: '/dist/'
  },
  module: {
    loaders: [
      // babel should remain first
      loaders.babel,
      loaders.extractLess,
      loaders.extractCss,
      loaders.images,
      loaders.png,
      loaders.svgs,
      loaders.json,
      loaders.fonts
    ]
  },
  resolve: {
    root: [
      path.resolve('./src')
    ],
    extensions: ['', '.jsx', '.js', '.json'],
    modulesDirectories: ['node_modules', 'src']
  },
  plugins: [
    new ExtractTextPlugin('app.css', {
      allChunks: true
    }),
    new webpack.DefinePlugin(envGlobals),
    new webpack.optimize.OccurenceOrderPlugin(),
    new HtmlWebpackPlugin({
      title: 'Control Center',
      hash: false,
      template: './index.html',
      favicon: path.join(PATHS.staticDir, 'favicon.ico'),
      inject: 'body',
      minify: {
        collapseWhitespace: true
      }
    })
  ],
  postcss() {
    return [postcssModulesValues, autoprefixer];
  },
  externals: {},
  imageWebpackLoader: {
    pngquant: {},
    svgo: {
      plugins: [
        {
          removeTitle: {}
        },
        {
          removeDesc: {}
        },
        {
          removeDimensions: {}
        }
      ]
    }
  }
};
