package io.confluent.controlcenter.license;

import org.joda.time.format.ISODateTimeFormat;
import org.jose4j.jwt.JwtClaims;

import java.io.FileInputStream;
import java.security.PrivateKey;
import java.security.PublicKey;
import java.util.concurrent.TimeUnit;

import io.airlift.airline.Cli;
import io.airlift.airline.Command;
import io.airlift.airline.Help;
import io.airlift.airline.Option;
import io.airlift.airline.ParseOptionMissingException;
import io.confluent.license.License;

public class LicenseCli {
  public static void main(String[] args) throws Exception {
    Cli.CliBuilder<Runnable> builder = Cli.<Runnable>builder("license")
        .withDefaultCommand(Help.class)
        .withCommands(Help.class, Create.class, Verify.class);
    Cli<Runnable> license = builder.build();

    try {
      license.parse(args).run();
    } catch (ParseOptionMissingException e) {
      System.err.println(e.getMessage());
    }
  }

  @Command(name = "create", description = "Create a license")
  public static class Create implements Runnable {
    @Option(name = "--monitoring", description = "Enable Monitoring")
    public boolean monitoring;

    @Option(name = "--audience", description = "Audience for this signature", required = true)
    public String audience;

    @Option(name = "--private-key", description = "Key to sign with", required = true)
    public String keyFile;

    @Option(name = "--expiration", description = "ISO 8601 formatted date of expiration.")
    public String expiration;

    public void run() {
      try {
        System.out.println("now=" + System.currentTimeMillis());
        long expiry = System.currentTimeMillis() + TimeUnit.DAYS.toMillis(365);
        if (expiration != null) {
          expiry = ISODateTimeFormat.dateTimeNoMillis().parseDateTime(expiration).toDateTime().getMillis();
        }
        System.out.println("expiry=" + expiry);
        PrivateKey privateKey = License.loadPrivateKey(new FileInputStream(keyFile));
        String out = License.sign(
            privateKey,
            audience,
            expiry,
            monitoring
        );
        System.out.println(out);
      } catch (Exception e) {
        System.err.println(e.getMessage());
      }
    }
  }

  @Command(name = "verify", description = "Verify a license")
  public static class Verify implements Runnable {
    @Option(name = "--license", description = "License to verify", required = true)
    public String license;

    @Option(name = "--public-key", description = "Key to sign with", required = true)
    public String keyFile;

    public void run() {
      try {
        PublicKey publicKey = License.loadPublicKey(new FileInputStream(keyFile));
        JwtClaims claims = License.verify(publicKey, license);
        System.out.println(claims.toJson());
      } catch (Exception e) {
        System.err.println(e.getMessage());
      }
    }
  }
}
