// (Copyright) [2016 - 2016] Confluent, Inc.
package io.confluent.monitoring.clients.interceptor;


import com.google.common.base.Preconditions;

import org.apache.kafka.clients.consumer.ConsumerConfig;
import org.apache.kafka.clients.consumer.ConsumerInterceptor;
import org.apache.kafka.clients.consumer.ConsumerRecord;
import org.apache.kafka.clients.consumer.ConsumerRecords;
import org.apache.kafka.clients.consumer.OffsetAndMetadata;
import org.apache.kafka.common.ClusterResource;
import org.apache.kafka.common.ClusterResourceListener;
import org.apache.kafka.common.TopicPartition;

import java.util.Map;

/**
 * Collects monitoring metrics on consumer side.
 */
public class MonitoringConsumerInterceptor<K, V> implements ConsumerInterceptor<K, V>,
                                                            ClusterResourceListener {

  private volatile ClusterResource clusterResource;
  private volatile MonitoringInterceptor monitoringInterceptor;
  private String groupId = "";
  private Map<String, ?> configs;

  /**
   * Gets Consumer configuration and initializes monitoringInterceptor and starts its background
   * publishing thread.
   */
  public void configure(Map<String, ?> configs) {
    this.configs = configs;
    Object groupIdValue = configs.get(ConsumerConfig.GROUP_ID_CONFIG);
    if (groupIdValue != null && groupIdValue instanceof String) {
      groupId = (String) groupIdValue;
    }
  }

  /**
   * This is called just before the records are returned by KafkaConsumer.poll()
   *
   * @param records records to be consumed by the client.
   * @return unmodified records
   */
  public ConsumerRecords<K, V> onConsume(ConsumerRecords<K, V> records) {
    if (monitoringInterceptor == null) {
      synchronized (this) {
        if (monitoringInterceptor == null) {
          Preconditions.checkState(clusterResource != null, "clusterResource is not defined");
          Preconditions.checkState(configs != null, "producer interceptor is not configured");

          monitoringInterceptor = MonitoringInterceptor.createForConsumer(
              clusterResource.clusterId(),
              groupId,
              configs
          );
          monitoringInterceptor.start();
        }
      }
    }

    for (ConsumerRecord<K, V> record : records) {
      // serializedKeySize and serializedValueSize are set to -1 if key and valure are null
      int size =
          Math.max(record.serializedKeySize(), 0) + Math.max(record.serializedValueSize(), 0);
      long latency = System.currentTimeMillis() - record.timestamp();
      monitoringInterceptor.recordMessageMetric(
          record.topic(),
          record.partition(),
          record.timestamp(),
          size,
          record.checksum(),
          latency
      );
    }
    return records;
  }

  /**
   * This is called when offsets get committed. This method does nothing.
   */
  public void onCommit(Map<TopicPartition, OffsetAndMetadata> offsets) {
  }

  /**
   * This is called when interceptor is closed
   */
  public void close() {
    if (monitoringInterceptor != null) {
      monitoringInterceptor.shutdown();
    }
  }

  @Override
  public void onUpdate(ClusterResource clusterResource) {
    this.clusterResource = clusterResource;
  }
}
