// (Copyright) [2016 - 2016] Confluent, Inc.
package io.confluent.monitoring.clients.interceptor;


import io.confluent.monitoring.common.MonitoringMessageUtil;
import io.confluent.monitoring.common.TimeBucket;
import org.apache.kafka.clients.CommonClientConfigs;
import org.apache.kafka.clients.producer.KafkaProducer;
import org.apache.kafka.clients.producer.ProducerRecord;
import org.apache.kafka.common.TopicPartition;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.Map;
import java.util.Properties;
import java.util.UUID;
import java.util.concurrent.ConcurrentHashMap;

import io.confluent.monitoring.record.Monitoring.ClientType;
import io.confluent.monitoring.record.Monitoring.MonitoringMessage;
import io.confluent.serializers.ProtoSerde;

/**
 * This is the common class that will be used by Monitoring Producer and Consumer Interceptors
 * implementing ProducerInterceptor and ConsumerInterceptor interfaces.
 *
 * MonitoringInteceptor is thread-safe.
 *
 * TODO: We currently never remove topic/partition entries from buckets map. Probably implement
 * expiring of topic-partition to cleanup.
 */
public class MonitoringInterceptor implements Runnable {

  private static final Logger log = LoggerFactory.getLogger(MonitoringInterceptor.class);

  private final long samplePeriodMs;
  private final long publishPeriodMs;
  private final int windowHistorySize;
  private final MonitoringMessage baseMonitoringMessage;
  private final MonitoringInterceptorConfig interceptorConfig;
  private final String publishTopic;
  private final ProtoSerde<MonitoringMessage> monitoringMessageSerde;

  private ConcurrentHashMap<TopicPartition, MonitoringTimeBuckets> buckets;
  private Thread reporterThread;
  private volatile boolean publishing;
  private volatile boolean closed = false;
  private KafkaProducer<byte[], byte[]> monitoringProducer;

  /**
   * @param clientType Currently either producer or consumer
   * @param configs    This is either Producer or Consumer Config (filled in with clientId if
   *                   automatically generated) passed from KafkaProducer or KafkaConsumer
   */
  public MonitoringInterceptor(ClientType clientType, String clusterId, String groupId, Map<String, ?> configs) {
    this.interceptorConfig = MonitoringInterceptorConfig.getConfig(configs);
    this.samplePeriodMs = TimeBucket.SIZE;                   // not-configurable in 3.0
    this.publishPeriodMs = interceptorConfig.getLong(MonitoringInterceptorConfig.PUBLISH_PERIOD);
    this.windowHistorySize = TimeBucket.HISTORY_SIZE;        // not-configurable in 3.0
    this.publishTopic = interceptorConfig.getString(MonitoringInterceptorConfig.TOPIC_CONFIG);

    Properties producerProperties = MonitoringInterceptorConfig.getProducerProperties(configs);

    baseMonitoringMessage = MonitoringMessage.newBuilder(MonitoringMessageUtil.baseMonitoringMessage())
        .setSession((UUID.randomUUID().toString()))
        .setClusterId(clusterId)
        .setGroup(groupId)
        .setClientId(configs.get(CommonClientConfigs.CLIENT_ID_CONFIG).toString())
        .setClientType(clientType)
        .setWindow(MonitoringTimeBuckets.UNKNOWN_WINDOW)
        .buildPartial();

    buckets = new ConcurrentHashMap<>();
    String interceptorClientId = producerProperties.getProperty(CommonClientConfigs.CLIENT_ID_CONFIG);
    reporterThread = new Thread(this, "confluent-metrics-reporter-" + interceptorClientId);
    this.publishing = false;
    log.info("creating producer for client={} {}", interceptorClientId, producerProperties);
    this.monitoringProducer = new KafkaProducer<>(producerProperties);
    this.monitoringMessageSerde = new ProtoSerde<MonitoringMessage>(MonitoringMessage.getDefaultInstance());
  }

  public static MonitoringInterceptor createForProducer(String clusterId, Map<String, ?> configs) {
    // 3.0 does not use group for producers, but we may want to change that later.
    return new MonitoringInterceptor(ClientType.PRODUCER, clusterId, "", configs);
  }

  public static MonitoringInterceptor createForConsumer(String clusterId, String groupId,
                                                        Map<String, ?> configs) {
    return new MonitoringInterceptor(ClientType.CONSUMER, clusterId, groupId, configs);
  }

  /**
   * Starts the background thread that publishes monitoring metrics to MonitoringTopic. Since the use-case
   * right now is to start on interceptor configure, and shutdown on producer/consumer close, the
   * method throws IllegalStateException if start() is called after shutdown.
   *
   * @throws IllegalStateException After shutdown() method is called.
   */
  public void start() {
    if (!this.closed) {
      reporterThread.start();
      this.publishing = true;
    } else {
      throw new IllegalStateException("Monitoring Interceptor was already shut down");
    }
  }

  /**
   * Record message metric into monitoring metrics for the given topic-partition.
   *
   * @param topic     Topic
   * @param partition Partition
   * @param timestamp Message timestamp
   * @param size      Message size (key plus value, pre-compression)
   * @param checksum  CRC of the message
   * @param latency   Latency of the message if available, otherwise non-positive value
   * @throws IllegalStateException If a background publishing thread is not running.
   */
  public void recordMessageMetric(
      String topic,
      int partition,
      long timestamp,
      int size,
      long checksum,
      long latency
  ) {
    // to prevent quiet recording of metrics without publishing them, which is also important to
    // not take all the memory to store the metrics we never consume.
    if (this.publishing == false) {
      throw new IllegalStateException(
          "Publishing thread must be running to enable recording of metrics");
    }

    try {
      TopicPartition tp = new TopicPartition(topic, partition);
      MonitoringTimeBuckets tpBuckets = buckets.get(tp);
      if (tpBuckets == null) {
        buckets.putIfAbsent(tp, new MonitoringTimeBuckets(this.samplePeriodMs,
                                                          this.windowHistorySize));
        tpBuckets = buckets.get(tp);
      }
      tpBuckets.record(timestamp, size, checksum, latency);
    } catch (Exception e) {
      // since interceptor does not propagate exceptions, we should not propagate this one as well
      // most likely, either timestamp is invalid or we are out of memory. If timestamp is
      // invalid, most likely message is corrupted and monitoring app will detect this with CRC
      // mismatch. Out of memory or other failure results in monitoring data loss that we will detect
      // with sequence numbers.
      log.error("Monitoring Interceptor failed to record message metrics for topic: " + topic + ", "
                + "partition: " + partition, e);
    }
  }

  @Override
  public void run() {
    // periodically publish to monitoring topic
    try {
      while (!Thread.currentThread().isInterrupted()) {
        Thread.sleep(this.publishPeriodMs);
        publishMetrics();
      }
    } catch (InterruptedException ie) {
      log.info("Publish thread interrupted!");
    } catch (Exception e) {
      log.error("Terminating publishing and collecting monitoring metrics", e);
    } finally {
      // try to publish all cached metrics, but first prevent metrics recording
      this.publishing = false;
      try {
        // make sure to set 'shutdown' attribute in each last message sent for a topic/partition
        publishMetrics(true);
      } catch (Exception e) {
        log.error("Failed to publish all cached metrics on termination", e);
      }
      log.info("Publishing Monitoring Metrics stopped for clientID={}", baseMonitoringMessage.getClientId());
    }
  }

  /**
   * Shuts down metrics reporter thread. All cached metrics are flushed to MonitoringTopic.
   */
  public void shutdown() {
    // stop thread but make sure all metrics get flushed.
    if (!this.closed) {
      try {
        reporterThread.interrupt();
        reporterThread.join(2000);
      } catch (Exception e) {
        log.error("Failed to shutdown metrics reporting thread", e);
      } finally {
        try {
          monitoringProducer.close();
        } catch (Exception e) {
          log.error("Failed to close monitoring interceptor", e);
        }
        this.closed = true;
        log.info("Closed monitoring interceptor for client ID={}", baseMonitoringMessage.getClientId());
      }
    }
  }

  /**
   * Publish current snapshot of all metrics to Monitoring Topic.
   */
  private void publishMetrics() {
    publishMetrics(false);
  }

  /**
   * Publish current snapshot of all metrics to Monitoring Topic. If 'shutdown' is true, set 'shutdown'
   * bit in the attribute field of the last monitoring message of each topic/partition.
   */
  private void publishMetrics(boolean shutdown) {
    // get a snapshot of the map (copy?) or at least all topic/partitions
    // and publish all metrics
    long now = System.currentTimeMillis();
    for (ConcurrentHashMap.Entry<TopicPartition, MonitoringTimeBuckets> entry : buckets.entrySet()) {
      MonitoringMessage.Builder publishMessage = entry.getValue().getNextMetrics(this.baseMonitoringMessage, now);
      if (publishMessage == null) {
        // topic/partition did not get any recorded metrics in this sampling period.
        // send one heartbeat message for this topic-partition so that the collector will be able
        // to identify missed monitoring data if any
        MonitoringMessage.Builder heartbeatMsg = entry.getValue().getEmptyMetrics(this.baseMonitoringMessage, now);
        heartbeatMsg.setTopic(entry.getKey().topic());
        heartbeatMsg.setPartition(entry.getKey().partition());
        if (shutdown) {
          heartbeatMsg.setShutdown(Boolean.TRUE);
        }
        publishMonitoringMessage(heartbeatMsg.build());
      } else {
        // topic/partition got at least one window with metrics; publish all recorded windows
        while (publishMessage != null) {
          publishMessage.setTopic(entry.getKey().topic());
          publishMessage.setPartition(entry.getKey().partition());
          MonitoringMessage.Builder next = entry.getValue().getNextMetrics(this.baseMonitoringMessage, now);
          // only set shutdown flag in the last message of topic-partition
          if (shutdown && next == null) {
            publishMessage.setShutdown(Boolean.TRUE);
          }
          publishMonitoringMessage(publishMessage.build());
          publishMessage = next;
        }
      }
    }
  }

  /**
   * Publishes give monitoring message to Monitoring Topic. Method is protected for testing.
   *
   * @param monitoringMessage Monitoring message to publish
   */
  protected void publishMonitoringMessage(MonitoringMessage monitoringMessage) {
    log.debug("Publish {}", monitoringMessage);
    try {
      this.monitoringProducer.send(
          new ProducerRecord<byte[], byte[]>(
              this.publishTopic,
              MonitoringMessageUtil.partitionFor(monitoringMessage, this.monitoringProducer.partitionsFor(this.publishTopic).size()),
              null,
              this.monitoringMessageSerde.serialize(monitoringMessage)
          )
      );
    } catch (Throwable t) {
      log.warn("failed to publish monitoring message", t);
    }
  }
}
