// (Copyright) [2016 - 2016] Confluent, Inc.
package io.confluent.monitoring.clients.interceptor;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import io.confluent.monitoring.record.Monitoring.MonitoringMessage;

/**
 * This class encapsulates all the metrics that we want to put into one time bucket Currently, it
 * has all metrics we want to publish to monitoring topic. In the future, if we want to have longer time
 * intervals for collecting latency and message size distributions, we will need to create a
 * different class, and move them to that class from here.
 */
public class MonitoringMetrics {

  private static final Logger log = LoggerFactory.getLogger(MonitoringMetrics.class);

  private long messageCount = 0L;
  private long byteCount = 0L;
  private long aggregateCrc = 0;
  private long totalLatency = 0;
  private long minLatency = Long.MAX_VALUE;
  private long maxLatency = 0;
  private long minMessageSize = Long.MAX_VALUE;
  private long maxMessageSize = 0;

  public MonitoringMetrics() {
  }

  /**
   * Record message metrics and aggregate with existing metrics.
   *
   * If the method gets invalid 'size', the method updates only message count, crc, and latency
   * metrics. The reason for not failing is that if message is recorded on producer side, then
   * publishing a message succeeded (which most like will not happen anyway). On consumer side, the
   * message most likely got corrupted, which will be accounted for in CRC, and which will lead to
   * CRC mismatch.
   *
   * If the method gets invalid latency, it most likely means that consumer clock is behind producer
   * clock. In that case, we count it as zero latency and don't record the message latency into
   * latency metrics.
   *
   * @param size     Size of the message
   * @param checksum Checksum of the message
   * @param latency  Latency of the message
   */
  public void recordMessageEvent(int size, long checksum, long latency) {
    messageCount++;
    aggregateCrc ^= checksum;
    if (latency >= 0) {
      totalLatency += latency;
      minLatency = Math.min(latency, minLatency);
      maxLatency = Math.max(latency, maxLatency);
    } else {
      log.warn("Negative message latency");
    }
    if (size > 0) {
      byteCount += size;
      minMessageSize = Math.min(size, minMessageSize);
      maxMessageSize = Math.max(size, maxMessageSize);
    } else {
      log.warn("Nonpositive message size");
    }
  }

  public void recordMessageEvent(int size, long checksum) {
    recordMessageEvent(size, checksum, 0);
  }

  /**
   * Reset all metrics
   */
  public void reset() {
    messageCount = 0L;
    byteCount = 0L;
    aggregateCrc = 0;
    totalLatency = 0L;
    minLatency = Long.MAX_VALUE;
    maxLatency = 0L;
    minMessageSize = Long.MAX_VALUE;
    maxMessageSize = 0L;
  }

  public long averageLatencyMs() {
    if (messageCount == 0) {
      return 0;
    }
    return totalLatency / messageCount;
  }

  /**
   * Set relevant fields to the given monitoring message from metrics in this class.
   */
  public void fillMonitoringMessage(MonitoringMessage.Builder monitoringMessage) {
    monitoringMessage.setAggregateCrc((int) aggregateCrc);
    monitoringMessage.setAggregateBytes(byteCount);
    monitoringMessage.setCount(messageCount);
    monitoringMessage.setTotalLatency(totalLatency);
    monitoringMessage.setMinLatency(totalLatency == 0 ? 0 : minLatency);
    monitoringMessage.setMaxLatency(maxLatency);
  }

  @Override
  public String toString() {
    return "MonitoringMetrics{" +
           "messageCount=" + messageCount +
           ", byteCount=" + byteCount +
           ", aggregateCrc=" + aggregateCrc +
           ", totalLatency=" + totalLatency +
           ", minLatency=" + minLatency +
           ", maxLatency=" + maxLatency +
           ", minMessageSize=" + minMessageSize +
           ", maxMessageSize=" + maxMessageSize +
           '}';
  }
}
