// (Copyright) [2016 - 2016] Confluent, Inc.
package io.confluent.monitoring.clients.interceptor;


import com.google.common.base.Preconditions;

import org.apache.kafka.clients.producer.ProducerInterceptor;
import org.apache.kafka.clients.producer.ProducerRecord;
import org.apache.kafka.clients.producer.RecordMetadata;
import org.apache.kafka.common.ClusterResource;
import org.apache.kafka.common.ClusterResourceListener;

import java.util.Map;


public class MonitoringProducerInterceptor<K, V> implements ProducerInterceptor<K, V>,
                                                            ClusterResourceListener {

  private Map<String, ?> configs;

  private volatile ClusterResource clusterResource;
  private volatile MonitoringInterceptor monitoringInterceptor;

  /**
   * Gets producer configuration and initializes monitoringInterceptor and starts its background
   * publishing thread.
   */
  public void configure(Map<String, ?> configs) {
    this.configs = configs;
  }

  /**
   * This is called just after client calls KafkaProducer.send(). This method does nothing.
   *
   * @param record the record from client
   * @return unmodified 'record' passed to this method
   */
  public ProducerRecord<K, V> onSend(ProducerRecord<K, V> record) {
    return record;
  }

  /**
   * This method is called when the record sent to the server has been acknowledged, or when sending
   * the record fails before it gets sent to the server.
   *
   * Currently, this implementation records only successful acks.
   *
   * @param metadata  The metadata for the record that was sent (i.e. the partition and offset). If
   *                  error occurred, metadata is only partially filled (topic, and maybe
   *                  partition). RecordMetadata.serializedKeySize and serializedValueSize will be
   *                  set to -1 if key and value are null.
   * @param exception The exception thrown during processing of this record. Null if no error
   *                  occurred.
   */
  public void onAcknowledgement(RecordMetadata metadata, Exception exception) {
    if(monitoringInterceptor == null) {
      synchronized(this) {
        if(monitoringInterceptor == null) {
          Preconditions.checkState(clusterResource != null, "clusterResource is not defined");
          Preconditions.checkState(configs != null, "producer interceptor is not configured");

          monitoringInterceptor = MonitoringInterceptor.createForProducer(
              clusterResource.clusterId(),
              configs
          );
          monitoringInterceptor.start();
        }
      }
    }

    if (metadata != null && exception == null) {
      int size =
          Math.max(metadata.serializedKeySize(), 0) + Math.max(metadata.serializedValueSize(), 0);

      monitoringInterceptor.recordMessageMetric(
          metadata.topic(),
          metadata.partition(),
          metadata.timestamp(),
          size,
          metadata.checksum(),
          0
      );
    }
  }


  @Override
  public void onUpdate(ClusterResource clusterResource) {
      this.clusterResource = clusterResource;
  }

  /**
   * This is called when interceptor is closed
   */
  public void close() {
    if (monitoringInterceptor != null) {
      monitoringInterceptor.shutdown();
    }
  }
}
