// (Copyright) [2016 - 2016] Confluent, Inc.
package io.confluent.monitoring.clients.interceptor;

import com.google.common.collect.EvictingQueue;

import java.util.Queue;
import java.util.concurrent.atomic.AtomicLong;

import io.confluent.monitoring.record.Monitoring;

/**
 * This is a collection of time buckets containing monitoring metrics. Each topic/partition will
 * have its own instance of this class.
 *
 * This class is used by producer and consumer interceptors to record monitoring metrics into
 * time buckets of configurable size. A background thread will periodically retrieve time buckets
 * with metrics and publish them to the Monitoring Topic. When metrics are retrieved, the
 * corresponding bucket holding the metrics is cleared to make sure that same metrics are published
 * only once.
 *
 * For (bucket) timestamps that do not have any message events recorded, this object will not return
 * any metrics, but the caller will publish a 'heartbeat' monitoring message with empty metrics to
 * the Monitoring Topic.
 */
public class MonitoringTimeBuckets extends TimeBuckets<MonitoringMetrics> {

  public static final long UNKNOWN_WINDOW = -1;

  private final AtomicLong curSequenceNumber = new AtomicLong(0);

  private TimeWindowHistory windowHistory;

  public MonitoringTimeBuckets(long bucketSizeMs, int windowHistorySize) {
    super(bucketSizeMs);
    windowHistory = new TimeWindowHistory(windowHistorySize);
  }

  /**
   * This method is called to record message monitoring metrics. The method accumulates message
   * metrics into the appropriate time bucket, identified based on 'messageTimestamp'.
   *
   * The method can throw an exception either because messageTimestamp is invalid or time bucket
   * allocation has failed. The caller (interceptor) will catch all the exceptions thrown by thin
   * method, log, and not propagate further. This means that metrics will not be recorded for this
   * message. This is OK for the following reasons. 1) If message timestamp is invalid (negative),
   * then the message is corrupted or there is a bug in the producer (producer sets timestamp field
   * to local time if the message timestamp provided by the client is negative). The corrupted
   * message will be accounted for by CRC mismatch. If there is a bug in the producer, we should fix
   * it. 2) If we get out of memory error, then the client we are intercepting also not doing well.
   * This will be similar to the case of client failure, which we will address with sequence
   * numbers.
   *
   * @param messageTimestamp Timestamp of the message
   * @param size             Message size
   * @param checksum         Message checksum
   * @param latency          Message latency, if available, or 0.
   * @throws IllegalArgumentException If messageTimestamp is invalid
   */
  public void record(long messageTimestamp, int size, long checksum, long latency) {
    synchronized (this) {
      Bucket<MonitoringMetrics> bucket = get(messageTimestamp);
      if (bucket.data == null) {
        bucket.data = new MonitoringMetrics();
      }
      bucket.data.recordMessageEvent(size, checksum, latency);
    }
  }

  /**
   * Returns monitoring message filled with metrics from the bucket with the oldest timestamp, or
   * null if there is no metrics to publish.
   *
   * @param base MonitoringMessage filled in with common fields
   * @param now  Current time to timestamp the message
   * @return MonitoringMessage or null if there are no metrics to publish
   */
  public Monitoring.MonitoringMessage.Builder getNextMetrics(Monitoring.MonitoringMessage base, long now) {
    Monitoring.MonitoringMessage.Builder monitoringMessageBuilder = null;

    synchronized (this) {
      Bucket<MonitoringMetrics> bucket = removeOldestBucket();
      if (bucket != null) {
        monitoringMessageBuilder = Monitoring.MonitoringMessage.newBuilder(base);
        monitoringMessageBuilder.setSequence(curSequenceNumber.getAndIncrement());
        monitoringMessageBuilder.setWindow(bucket.timestamp);
        monitoringMessageBuilder.setSamplePeriod(getBucketSizeMs());
        monitoringMessageBuilder.setTimestamp(now);
        monitoringMessageBuilder.setType(Monitoring.MessageType.NORMAL);
        windowHistory.fillMinMaxWindowsAndRecord(monitoringMessageBuilder);
        bucket.data.fillMonitoringMessage(monitoringMessageBuilder);
      }
    }

    return monitoringMessageBuilder;
  }

  /**
   * Returns monitoring message filled with zeros for all metrics and null timestamp ID for the
   * window. This is used for publishing empty monitoring messages for topic/partitions that do now
   * have any activity during a sampling period. Sequence number is set and incremented.
   *
   * This method does not check whether there are any metrics collected for publishing. It is up to
   * the caller to use getEmptyMetrics and getNextMetrics methods correctly.
   *
   * @param base MonitoringMessage filled in with common fields
   * @param now  Current time to timestamp the message
   * @return Monitoring message with zero metrics but sequence number set.
   */
  public Monitoring.MonitoringMessage.Builder getEmptyMetrics(Monitoring.MonitoringMessage base, long now) {
    Monitoring.MonitoringMessage.Builder monitoringMessageBuilder = Monitoring.MonitoringMessage.newBuilder(base);
    monitoringMessageBuilder.setTimestamp(now);
    monitoringMessageBuilder.setSequence(curSequenceNumber.getAndIncrement());
    monitoringMessageBuilder.setSamplePeriod(getBucketSizeMs());
    windowHistory.fillMinMaxWindowsAndRecord(monitoringMessageBuilder);
    // make sure that we set window to 0 after calling fillMinMaxWindowsAndRecord method, since
    // it expects unset window to be -1. However, we have to set window to 0 before publishing
    // message to the monitoring topic since we are using window for KStreams timestamp, and it
    // does not except timestamps < 0; sample period is null
    monitoringMessageBuilder.setWindow(0);
    monitoringMessageBuilder.setType(Monitoring.MessageType.HEARTBEAT);
    MonitoringMetrics metrics = new MonitoringMetrics();
    metrics.fillMonitoringMessage(monitoringMessageBuilder);
    return monitoringMessageBuilder;
  }

  /**
   * Maintains history of time windows seen in last 'size' monitoring messages returned by
   * MonitoringTimeBuckets (corresponds to last 'size' sequence numbers).
   *
   * If MonitoringTimeBuckets do not have any metrics recorded and return empty Monitoring Messages,
   * then a 'seen' time window in this case is null.
   *
   * We don't need to make this class thread-safe, because it is always called on metrics retrieval
   * which is always called from one thread (publishing metrics thread in the interceptor).
   */
  private class TimeWindowHistory {

    Queue<Long> history;

    public TimeWindowHistory(int size) {
      history = EvictingQueue.create(size);
    }

    /**
     * Fill in min and max window fields in monitoring message with min and max windows recorded in
     * the history. Min and max windows can be null if no windows were recorded in the last
     * 'size' of sequence numbers (sending empty metrics).
     *
     * Also adds window from monitoring message to the history.
     */
    public void fillMinMaxWindowsAndRecord(Monitoring.MonitoringMessage.Builder monitoringMessage) {
      for (Long window : history) {
        if (window != UNKNOWN_WINDOW) {
          monitoringMessage.setMinWindow(
              Math.min(
                  window,
                  monitoringMessage.getMinWindow() == UNKNOWN_WINDOW ? window : monitoringMessage.getMinWindow()
              )
          );
          monitoringMessage.setMaxWindow(
              Math.max(
                  window,
                  monitoringMessage.getMaxWindow() == UNKNOWN_WINDOW ? window : monitoringMessage.getMaxWindow()
              )
          );
        }
      }
      // we want to record null windows, so that we always use history of last 'size sequence
      // numbers, independent of how many of them are heartbeat messages without a window
      history.add(monitoringMessage.getWindow() == UNKNOWN_WINDOW ? UNKNOWN_WINDOW : monitoringMessage.getWindow());
    }
  }
}
