// (Copyright) [2016 - 2016] Confluent, Inc.
package io.confluent.monitoring.clients.interceptor;

import java.util.Map;
import java.util.NavigableMap;
import java.util.TreeMap;

/**
 * This is a collection of time buckets containing a set of metrics. Each bucket is of the same
 * configurable time interval. Each bucket is identified by a bucket timestamp. An event falls into
 * a certain time bucket based on its own timestamp.
 *
 * The collection of time buckets is represented as a map, where key is a bucket timestamp. If
 * recorded metrics sample falls into a time bucket with timestamp not currently in the map, a new
 * map entry is created. The buckets get removed from the map, when the caller (normally the class
 * derived from TimeBuckets) retrieves the bucket.
 *
 * The use case for this class is to record the metrics into time buckets based on provided event
 * timestamps. Event timestamps are not guaranteed to be monotonically increasing, and in worse
 * case, can come in any order and time distance. Also, once metrics are retrieved, they need to be
 * cleared to make sure that same metrics are not counted twice. So API support the case where
 * retrieving metrics also removes metrics from the map (but more API could be added to support
 * other use-cases).
 *
 * The intended usage of the class is to derive the class using the type of metrics needed
 * collecting. Metrics will be recorded on the caller thread. Normally, there will be a background
 * thread (but not required) that will periodically retrieve bucketed metrics. Once metrics are
 * retrieved, buckets are always cleared and removed from the map.
 *
 * Possible issues that may need addressing in the future: 1. Every time a metric is recorded to the
 * timebucket not in the map, a new time bucket is allocated and added to the map. In the use-case
 * where metrics are periodically retrieved and new buckets are allocated in the next time interval,
 * this adds overhead of object allocation in every time interval. We will do some performance
 * testing and possibly improve with pre-allocating buckets and maintaining a free list of buckets.
 * 2. In a very bad case, timestamps may have large jitter causing the map to grow a lot. We may
 * want to bound the memory space usage (especially because we are going to maintain a map per
 * topic/partition). We can bound map size by dynamically changing metrics retrieval interval.
 *
 * @param <T> the class representing a metric or a set of metrics in each time bucket. We currently
 *            have one such type. However, we made it generic in case we want to use different time
 *            intervals for latency and message size distributions. In that case, we will separate
 *            MonitoringMetrics class into two, and create another class that derives from TimeBuckets
 */
public class TimeBuckets<T> {

  private final long bucketSizeMs;
  private NavigableMap<Long, Bucket<T>> buckets;   // timestamp -> time bucket

  public TimeBuckets(long bucketSizeMs) {
    this.bucketSizeMs = bucketSizeMs;
    this.buckets = new TreeMap<>();
  }

  /**
   * Returns time bucket that corresponds to an event timestamp. if timebucket does not exists for a
   * corresponding bucket timestamp, the new time bucket is created and added to the map. The caller
   * must allocate the data part of the time bucket.
   *
   * @param timeMs Event timestamp
   * @return Time bucket containing the data stored in the bucket that corresponds to a given
   * timestamp.
   */
  public Bucket<T> get(long timeMs) {
    if (timeMs < 0) {
      throw new IllegalArgumentException("Invalid timestamp " + timeMs);
    }

    long timestamp = getTimestamp(timeMs);
    Bucket<T> bucket = buckets.get(timestamp);
    if (bucket == null) {
      // start a new bucket
      bucket = new Bucket<>(timestamp);
      buckets.put(timestamp, bucket);
    }
    return bucket;
  }

  /**
   * Returns bucket with the oldest timestamp and removes this bucket from the list. If the list is
   * empty, returns null.
   */
  public Bucket<T> removeOldestBucket() {
    Map.Entry<Long, Bucket<T>> entry = this.buckets.pollFirstEntry();
    return entry != null ? entry.getValue() : null;
  }

  /**
   * Returns bucket timestamp that corresponds to the given time in milliseconds
   */
  public long getTimestamp(long timeMs) {
    return timeMs - (timeMs % bucketSizeMs);
  }

  /**
   * Returns number of time buckets
   */
  public long getNumberOfBuckets() {
    return buckets.size();
  }

  public long getBucketSizeMs() {
    return bucketSizeMs;
  }


  public class Bucket<T> {

    public long timestamp;            // timestamp that identifies this time bucket
    public T data;                    // data recorded in this bucket

    public Bucket(long timestamp) {
      this.timestamp = timestamp;
      this.data = null;
    }

    /**
     * Returns true if given timestamp identifies with this time bucket
     */
    public boolean corresponds(long timestamp) {
      return timestamp == this.timestamp;
    }
  }
}
