// (Copyright) [2016 - 2016] Confluent, Inc.
package io.confluent.monitoring.clients.interceptor;

import com.google.protobuf.InvalidProtocolBufferException;

import io.confluent.monitoring.common.TimeBucket;
import junit.framework.AssertionFailedError;

import org.apache.kafka.clients.CommonClientConfigs;
import org.apache.kafka.clients.producer.KafkaProducer;
import org.apache.kafka.clients.producer.ProducerConfig;
import org.apache.kafka.clients.producer.ProducerRecord;
import org.apache.kafka.clients.producer.RecordMetadata;
import org.apache.kafka.clients.producer.internals.FutureRecordMetadata;
import org.apache.kafka.clients.producer.internals.ProduceRequestResult;
import org.apache.kafka.common.PartitionInfo;
import org.apache.kafka.common.TopicPartition;
import org.apache.kafka.common.config.ConfigException;
import org.easymock.EasyMock;
import org.easymock.IAnswer;
import org.junit.Before;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.powermock.api.mockito.PowerMockito;
import org.powermock.core.classloader.annotations.PrepareForTest;
import org.powermock.modules.junit4.PowerMockRunner;

import java.io.IOException;
import java.util.*;
import java.util.concurrent.CopyOnWriteArrayList;
import java.util.concurrent.Future;

import io.confluent.monitoring.record.Monitoring;
import io.confluent.monitoring.record.Monitoring.ClientType;
import io.confluent.monitoring.record.Monitoring.MonitoringMessage;
import io.confluent.serializers.ProtoSerde;
import io.confluent.serializers.UberSerde;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertNull;
import static org.junit.Assert.assertTrue;

@RunWith(PowerMockRunner.class)
@PrepareForTest(MonitoringInterceptor.class)
public class MonitoringInterceptorTest {

  private final Long samplePeriod = new Long(TimeBucket.SIZE);
  private final String clientId = "Test Client";
  private final String clusterId = "Test Cluster";
  private final String groupId = "Test Group";
  private final ClientType clientType = ClientType.CONSUMER;
  private final TopicPartition topicPartition = new TopicPartition("test", 0);
  private final Map<String, Object> configs = new HashMap<>();
  private KafkaProducer<String, byte[]> monitoringProducer;
  private MonitoringInterceptor interceptor;
  // this is used in the main thread and also in a publish thread (from mocked producer send)
  private List<ProducerRecord<String, byte[]>> sent = new CopyOnWriteArrayList<>();
  private UberSerde<MonitoringMessage> monitoringMessageUberSerde = new ProtoSerde<>(MonitoringMessage.getDefaultInstance());

  @Before
  public void setUp() throws Exception {
    configs.put(CommonClientConfigs.CLIENT_ID_CONFIG, clientId);
    configs.put(CommonClientConfigs.BOOTSTRAP_SERVERS_CONFIG, "bootstrap");

    monitoringProducer = EasyMock.createMock(KafkaProducer.class);
    PowerMockito.whenNew(KafkaProducer.class).withAnyArguments().thenReturn(monitoringProducer);
  }

  @Test(expected = ConfigException.class)
  public void testCreateWithoutClientIdFails() {
    EasyMock.replay(monitoringProducer);
    interceptor = new MonitoringInterceptor(clientType, clusterId, groupId, new HashMap<String,
        Object>());
  }

  @Test
  public void testPreferInterceptorClientIdConfig() throws InvalidProtocolBufferException {
    final String otherClientId = "other-client";

    expectMonitoringProducerSendMetrics(false);
    expectMonitoringProducerClose();
    EasyMock.replay(monitoringProducer);

    Map<String, Object> otherConfigs = new HashMap<>();
    otherConfigs.put(ProducerConfig.CLIENT_ID_CONFIG, otherClientId);
    otherConfigs.put(ProducerConfig.BOOTSTRAP_SERVERS_CONFIG, "bootstraps");

    interceptor = new MonitoringInterceptor(clientType, clusterId, groupId, otherConfigs);
    interceptor.start();
    interceptor.recordMessageMetric(topicPartition.topic(), topicPartition.partition(),
                                    0, 100, 10, 100);
    interceptor.shutdown();
    EasyMock.verify(monitoringProducer);

    // make sure that's the client id used in the monitoring message
    for (ProducerRecord<String, byte[]> record : sent) {
      MonitoringMessage monitoringMessage = monitoringMessageUberSerde.deserialize(record.value());
      assertEquals("Invalid client ID in published monitoring message",
                   otherClientId, monitoringMessage.getClientId());
    }
  }

  @Test(expected = IllegalStateException.class)
  public void testRecordBeforeStartFails() {
    EasyMock.replay(monitoringProducer);
    interceptor = new MonitoringInterceptor(clientType, clusterId, groupId, configs);

    interceptor.recordMessageMetric("test", 0, 0, 0, 0, 0);
  }

  @Test(expected = IllegalStateException.class)
  public void testStartAfterShutdownFails() {
    expectMonitoringProducerClose();
    EasyMock.replay(monitoringProducer);

    interceptor = new MonitoringInterceptor(clientType, clusterId, groupId, configs);
    interceptor.start();
    interceptor.shutdown();
    interceptor.start();
  }

  @Test
  public void testPartitioningOfMetricData() throws InvalidProtocolBufferException {
    expectMonitoringProducerSendMetrics(false, 10);
    expectMonitoringProducerClose();
    EasyMock.replay(monitoringProducer);

    interceptor = new MonitoringInterceptor(clientType, clusterId, groupId, configs);
    interceptor.start();
    interceptor.recordMessageMetric(topicPartition.topic(), topicPartition.partition(),
            0, 10, 10, 10);
    interceptor.recordMessageMetric(topicPartition.topic(), topicPartition.partition(),
            samplePeriod, 10, 10, 10);
    interceptor.recordMessageMetric(topicPartition.topic(), topicPartition.partition(),
            3 * samplePeriod, 10, 10, 10);
    interceptor.shutdown();

    EasyMock.verify(monitoringProducer);
    assertEquals(3, sent.size());
    for (ProducerRecord<String, byte[]> record: sent) {
      assertEquals(4, record.partition().intValue());
    }
  }

  @Test
  public void testRecordAfterStartPublishesMetrics() throws InvalidProtocolBufferException {
    expectMonitoringProducerSendMetrics(false);
    expectMonitoringProducerClose();
    EasyMock.replay(monitoringProducer);

    interceptor = new MonitoringInterceptor(clientType, clusterId, groupId, configs);
    interceptor.start();
    interceptor.recordMessageMetric(topicPartition.topic(), topicPartition.partition(),
                                    0, 100, 10, 100);
    interceptor.shutdown();

    EasyMock.verify(monitoringProducer);
    verifyPublishedMetricsSameTopicPartition(topicPartition, 1L, true);
  }

  @Test
  public void testStartShutdownWithoutRecordDoesNotPublishToMonitoringTopic() {
    // use-case: client starts, nothing happens, client shuts down. Interceptor does not know
    // about any topic/partitions, so no monitoring messages get published
    // so do not expect that monitoring producer send() is called
    expectMonitoringProducerClose();
    EasyMock.replay(monitoringProducer);

    interceptor = new MonitoringInterceptor(clientType, clusterId, groupId, configs);
    interceptor.start();
    interceptor.shutdown();
    EasyMock.verify(monitoringProducer);
    assertEquals(0, sent.size());
  }

  @Test
  public void testPublishHeartbeatMessagesForKnownTopicPartition() throws InvalidProtocolBufferException {
    expectMonitoringProducerSendMetrics(false);
    expectMonitoringProducerClose();
    EasyMock.replay(monitoringProducer);

    interceptor = new MonitoringInterceptor(clientType, clusterId, groupId, configs);
    interceptor.start();
    try {
      // we need to record at least one metric, so that Interceptor knows about topic-partition
      interceptor.recordMessageMetric(topicPartition.topic(), topicPartition.partition(),
                                      samplePeriod, 10, 10, 10);
      Thread.sleep(samplePeriod * 3);
    } catch (InterruptedException ie) {
      throw new AssertionFailedError("Did not finish waiting for publish");
    }
    interceptor.shutdown();
    EasyMock.verify(monitoringProducer);
    // we should see at least three monitoring messages
    assertTrue(sent.size() >= 3);
    verifyPublishedMetricsSameTopicPartition(topicPartition, 1L, true);
  }

  @Test(expected = IllegalStateException.class)
  public void testRecordAfterShutdownFails() {
    expectMonitoringProducerClose();
    EasyMock.replay(monitoringProducer);

    interceptor = new MonitoringInterceptor(clientType, clusterId, groupId, configs);
    interceptor.start();
    interceptor.shutdown();
    interceptor.recordMessageMetric("test", 0, 0, 100, 10, 100);
  }

  @Test
  public void testRecordToInvalidTimestampDoesNotRecordMetric() throws InvalidProtocolBufferException {
    // if we record metric with invalid timestamp, there is no exception, but metric should not
    // be recorded. Interceptor will still publish a heartbeat message, but total message count is
    // expected to be zero
    expectMonitoringProducerSendMetrics(false);
    expectMonitoringProducerClose();
    EasyMock.replay(monitoringProducer);

    interceptor = new MonitoringInterceptor(clientType, clusterId, groupId, configs);
    interceptor.start();
    // record with invalid timestamp
    interceptor.recordMessageMetric("test", 0, -1, 100, 10, 100);
    interceptor.shutdown();

    EasyMock.verify(monitoringProducer);
    verifyPublishedMetricsSameTopicPartition(topicPartition, 0L, true);
  }

  @Test
  public void testRecordMetricWithInvalidLatencyRecordsMetric() throws InvalidProtocolBufferException {
    expectMonitoringProducerSendMetrics(false);
    expectMonitoringProducerClose();
    EasyMock.replay(monitoringProducer);

    interceptor = new MonitoringInterceptor(clientType, clusterId, groupId, configs);
    interceptor.start();
    interceptor.recordMessageMetric(topicPartition.topic(), topicPartition.partition(),
                                    0, 100, 10, -10);
    interceptor.shutdown();

    EasyMock.verify(monitoringProducer);
    verifyPublishedMetricsSameTopicPartition(topicPartition, 1L, true);
  }

  @Test
  public void testRecordMetricsWithInvalidBytesRecordsMetric() throws InvalidProtocolBufferException {
    expectMonitoringProducerSendMetrics(false);
    expectMonitoringProducerClose();
    EasyMock.replay(monitoringProducer);

    interceptor = new MonitoringInterceptor(clientType, clusterId, groupId, configs);
    interceptor.start();
    interceptor.recordMessageMetric(topicPartition.topic(), topicPartition.partition(),
                                    0, -10, 10, 10);
    interceptor.shutdown();

    EasyMock.verify(monitoringProducer);
    verifyPublishedMetricsSameTopicPartition(topicPartition, 1L, true);
  }

  @Test
  public void testRecordMultipleWindowsPublishesMultipleMonitoringMessages() throws InvalidProtocolBufferException {
    expectMonitoringProducerSendMetrics(false);
    expectMonitoringProducerClose();
    EasyMock.replay(monitoringProducer);

    interceptor = new MonitoringInterceptor(clientType, clusterId, groupId, configs);
    interceptor.start();
    interceptor.recordMessageMetric(topicPartition.topic(), topicPartition.partition(),
                                    0, 10, 10, 10);
    interceptor.recordMessageMetric(topicPartition.topic(), topicPartition.partition(),
                                    samplePeriod, 10, 10, 10);
    interceptor.recordMessageMetric(topicPartition.topic(), topicPartition.partition(),
                                    3 * samplePeriod, 10, 10, 10);
    interceptor.shutdown();

    EasyMock.verify(monitoringProducer);
    assertEquals(3, sent.size());
    verifyPublishedMetricsSameTopicPartition(topicPartition, 3L, true);
  }

  @Test
  public void testRecordMultiplePartitionsPublishesMultipleMonitoringMessages() throws InvalidProtocolBufferException {
    expectMonitoringProducerSendMetrics(false);
    expectMonitoringProducerClose();
    EasyMock.replay(monitoringProducer);

    interceptor = new MonitoringInterceptor(clientType, clusterId, groupId, configs);
    interceptor.start();
    interceptor.recordMessageMetric(topicPartition.topic(), topicPartition.partition(),
                                    0, 10, 10, 10);
    interceptor.recordMessageMetric("test", 1, 0, 10, 10, 10);
    interceptor.recordMessageMetric("test", 2, 0, 10, 10, 10);
    interceptor.shutdown();

    EasyMock.verify(monitoringProducer);
    assertEquals(3, sent.size());
    Map<TopicPartition, Long> recordedMetrics = new HashMap<>();
    recordedMetrics.put(topicPartition, 1L);
    recordedMetrics.put(new TopicPartition("test", 1), 1L);
    recordedMetrics.put(new TopicPartition("test", 2), 1L);
    verifyPublishedMetrics(recordedMetrics, true);
  }

  @Test
  public void testRecordMultipleTopicsPublishesMultipleMonitoringMessages() throws InvalidProtocolBufferException {
    expectMonitoringProducerSendMetrics(false);
    expectMonitoringProducerClose();
    EasyMock.replay(monitoringProducer);

    interceptor = new MonitoringInterceptor(clientType, clusterId, groupId, configs);
    interceptor.start();
    interceptor.recordMessageMetric("test1", 0, 0, 10, 10, 10);
    interceptor.recordMessageMetric("test2", 0, 0, 10, 10, 10);
    interceptor.recordMessageMetric("test3", 0, 0, 10, 10, 10);
    interceptor.shutdown();

    EasyMock.verify(monitoringProducer);
    assertEquals(3, sent.size());
    Map<TopicPartition, Long> recordedMetrics = new HashMap<>();
    recordedMetrics.put(new TopicPartition("test1", 0), 1L);
    recordedMetrics.put(new TopicPartition("test2", 0), 1L);
    recordedMetrics.put(new TopicPartition("test3", 0), 1L);
    verifyPublishedMetrics(recordedMetrics, true);
  }

  @Test
  public void testRecordSameTopicPartitionWindowPublishesOneMonitoringMessages() throws InvalidProtocolBufferException {
    expectMonitoringProducerSendMetrics(false);
    expectMonitoringProducerClose();
    EasyMock.replay(monitoringProducer);

    interceptor = new MonitoringInterceptor(clientType, clusterId, groupId, configs);
    interceptor.start();
    interceptor.recordMessageMetric(topicPartition.topic(), topicPartition.partition(),
                                    0, 10, 10, 10);
    interceptor.recordMessageMetric(topicPartition.topic(), topicPartition.partition(),
                                    1, 10, 10, 10);
    interceptor.recordMessageMetric(topicPartition.topic(), topicPartition.partition(),
                                    2, 10, 10, 10);
    interceptor.shutdown();

    EasyMock.verify(monitoringProducer);
    assertEquals(1, sent.size());
    verifyPublishedMetricsSameTopicPartition(topicPartition, 3L, true);
  }

  @Test
  public void testPublishBeforeShutdownMultipleWindows() throws InvalidProtocolBufferException {
    expectMonitoringProducerSendMetrics(false);
    monitoringProducer.close();
    EasyMock.expectLastCall().anyTimes();
    EasyMock.replay(monitoringProducer);

    interceptor = new MonitoringInterceptor(clientType, clusterId, groupId, configs);
    interceptor.start();
    interceptor.recordMessageMetric(topicPartition.topic(), topicPartition.partition(),
                                    0, 10, 10, 10);
    interceptor.recordMessageMetric(topicPartition.topic(), topicPartition.partition(),
                                    100 * samplePeriod, 10, 10, 10);
    interceptor.recordMessageMetric(topicPartition.topic(), topicPartition.partition(),
                                    2 * samplePeriod, 10, 10, 10);
    interceptor.recordMessageMetric(topicPartition.topic(), topicPartition.partition(),
                                    10 * samplePeriod, 10, 10, 10);
    try {
      Thread.sleep(samplePeriod * 2);
    } catch (InterruptedException ie) {
      throw new AssertionFailedError("Did not finishing waiting for publish");
    }

    EasyMock.verify(monitoringProducer);
    // the monitoring topic producer may continue sending heartbeat messages, so we just need to
    // check no less than 4
    assertTrue(4 >= sent.size());
    verifyPublishedMetricsSameTopicPartition(topicPartition, 4L, false);

    interceptor.shutdown();
  }

  @Test
  public void testProducerSendFailureDoesNotStopPublishing() throws InvalidProtocolBufferException {
    final int count = 10;
    expectMonitoringProducerSendMetrics(true);
    expectMonitoringProducerClose();
    EasyMock.replay(monitoringProducer);

    interceptor = new MonitoringInterceptor(clientType, clusterId, groupId, configs);
    interceptor.start();
    try {
      for (int i = 0; i < count; i++) {
        interceptor.recordMessageMetric(topicPartition.topic(), topicPartition.partition(),
                                        i * samplePeriod, 10, 10, 10);
        Thread.sleep(1000);
      }
    } catch (InterruptedException ie) {
      throw new AssertionFailedError("Did not finish waiting for publish");
    }
    interceptor.shutdown();

    EasyMock.verify(monitoringProducer);
    assertTrue(count <= sent.size());
    verifyPublishedMetricsSameTopicPartition(topicPartition, new Long(count), true);
  }

  @Test
  public void testHeartbeatMessage() throws InvalidProtocolBufferException {
    expectMonitoringProducerSendMetrics(false);
    monitoringProducer.close();
    EasyMock.expectLastCall().anyTimes();
    EasyMock.replay(monitoringProducer);

    interceptor = new MonitoringInterceptor(clientType, clusterId, groupId, configs);
    interceptor.start();
    interceptor.recordMessageMetric(topicPartition.topic(), topicPartition.partition(),
                                    0, 10, 10, 10);
    try {
      // wait so that interceptor publishes heartbeat messages
      Thread.sleep(samplePeriod * 3);
    } catch (InterruptedException ie) {
      throw new AssertionFailedError("Did not finishing waiting for publish");
    }

    EasyMock.verify(monitoringProducer);
    // only the first message is expected to be non-heartbeat, and expect at least one heartbeat
    // message after that
    assertTrue(2 >= sent.size());
    ProducerRecord<String, byte[]> record = sent.get(1);
    MonitoringMessage monitoringMessage = monitoringMessageUberSerde.deserialize(record.value());
    assertEquals(Monitoring.MessageType.HEARTBEAT, monitoringMessage.getType());
    assertEquals(0, monitoringMessage.getWindow());
    assertTrue(monitoringMessage.getSamplePeriod() > 0);
    assertEquals(0, monitoringMessage.getCount());

    interceptor.shutdown();
  }

  private void verifyPublishedMetricsSameTopicPartition(TopicPartition tp, Long messageCount,
                                                        boolean expectShutdown) throws InvalidProtocolBufferException {
    Map<TopicPartition, Long> recordedMetrics = new HashMap<>();
    recordedMetrics.put(tp, messageCount);
    verifyPublishedMetrics(recordedMetrics, expectShutdown);
  }

  private void verifyPublishedMetrics(
      Map<TopicPartition, Long> recordedMetrics,
      boolean expectShutdown
  ) throws InvalidProtocolBufferException {
    Map<TopicPartition, Long> publishedMetrics = new HashMap<>();
    Map<TopicPartition, Set<Long>> seqNumbers = new HashMap<>();
    Map<TopicPartition, Long> seqNumShutdownFlag = new HashMap<>();
    for (ProducerRecord<String, byte[]> record : sent) {
      MonitoringMessage monitoringMessage = monitoringMessageUberSerde.deserialize(record.value());
      assertNotNull(monitoringMessage.getClientId());
      assertEquals(clientId, monitoringMessage.getClientId());

      assertNotNull(monitoringMessage.getGroup());
      assertEquals(groupId, monitoringMessage.getGroup());

      assertNotNull(monitoringMessage.getClusterId());
      assertEquals(clusterId, monitoringMessage.getClusterId());

      // window must never be -1
      assertTrue(monitoringMessage.getWindow() >= 0);

      // sample period must be set
      assertTrue(monitoringMessage.getSamplePeriod() > 0);

      if (monitoringMessage.getType() == Monitoring.MessageType.HEARTBEAT) {
        assertEquals(0, monitoringMessage.getCount());
        assertEquals(0, monitoringMessage.getAggregateBytes());
      }

      // interceptor does not send 'error' messages (yet)
      assertTrue(monitoringMessage.getType() != Monitoring.MessageType.ERROR);

      TopicPartition tp = new TopicPartition(monitoringMessage.getTopic(), monitoringMessage.getPartition());
      Long count = publishedMetrics.get(tp);
      Set<Long> tpSeqNums = seqNumbers.get(tp);
      if (count == null) {
        count = new Long(0);
        assertNull(tpSeqNums);
        tpSeqNums = new HashSet<>();
        seqNumbers.put(tp, tpSeqNums);
      }
      count += monitoringMessage.getCount();
      publishedMetrics.put(tp, count);
      assertFalse(
          String.format("Repeating sequence number %d", monitoringMessage.getSequence()),
          tpSeqNums.contains(monitoringMessage.getSequence())
      );
      tpSeqNums.add(monitoringMessage.getSequence());

      if (expectShutdown && monitoringMessage.getShutdown()) {
        assertFalse(
            String.format(
                "Got more than one monitoring message with shutdown flag set for topic %s, partition %d.",
                monitoringMessage.getTopic(), monitoringMessage.getPartition()
            ),
            seqNumShutdownFlag.containsKey(tp)
        );
        seqNumShutdownFlag.put(tp, monitoringMessage.getSequence());
      }
    }
    assertEquals(recordedMetrics, publishedMetrics);

    // verify that sequence numbers are from 0 to num of monitoring messages per topic-partition
    for (Map.Entry<TopicPartition, Set<Long>> entry : seqNumbers.entrySet()) {
      for (long seq = 0; seq < entry.getValue().size(); seq++) {
        assertTrue(String.format("Missing sequence number %d", seq),
                   entry.getValue().contains(seq));
      }
      // verify that shutdown flag is set in the message with the highest sequence number
      if (expectShutdown) {
        assertTrue("Topic-partition is missing an monitoring message with shutdown flag",
                   seqNumShutdownFlag.containsKey(entry.getKey()));
        assertEquals("Shutdown flag is set in monitoring message with unexpected sequence number.",
                     entry.getValue().size() - 1,
                     seqNumShutdownFlag.get(entry.getKey()).intValue());
      }
    }
  }

  private void expectMonitoringProducerSendMetrics(final boolean failFirstSend) {
    expectMonitoringProducerSendMetrics(failFirstSend, 1);
  }

  private void expectMonitoringProducerSendMetrics(final boolean failFirstSend, int numPartitions) {
    expectPartitions(numPartitions);
    expectSend(failFirstSend);
  }

  private void expectPartitions(final int numPartitions) {
    EasyMock.expect(monitoringProducer.partitionsFor(EasyMock.anyObject(String.class))).andAnswer(
            new IAnswer<List<PartitionInfo>>() {
              @Override
              public List<PartitionInfo> answer() throws Throwable {
                ArrayList<PartitionInfo> partionsInfoList = new ArrayList<>();
                for (int i = 0; i < numPartitions; i++) {
                  partionsInfoList.add(EasyMock.createMock(PartitionInfo.class));
                }
                return partionsInfoList;
              }
            }
    ).anyTimes();
  }

  private void expectSend(final boolean failFirstSend) {
    EasyMock.expect(monitoringProducer.send(EasyMock.anyObject(ProducerRecord.class))).andAnswer(
        new IAnswer<Future<RecordMetadata>>() {
          @Override
          public Future<RecordMetadata> answer() throws Throwable {
            ProducerRecord<String, byte[]>
                record =
                (ProducerRecord<String, byte[]>) EasyMock.getCurrentArguments()[0];
            ProduceRequestResult result = new ProduceRequestResult();
            FutureRecordMetadata future = new FutureRecordMetadata(result, 0, 0, 0, 0, 0);
            TopicPartition topicPartition = new TopicPartition(record.topic(), 0);
            sent.add(record);
            if (sent.size() == 1 && failFirstSend == true) {
              throw new IOException("Injected exception");
            }
            result.done(topicPartition, 0, null);
            return future;
          }
        }).anyTimes();
  }


  private void expectMonitoringProducerClose() {
    monitoringProducer.close();
    EasyMock.expectLastCall().once();
  }
}
