// (Copyright) [2016 - 2016] Confluent, Inc.
package io.confluent.monitoring.clients.interceptor;

import org.junit.Test;

import io.confluent.monitoring.record.Monitoring.MonitoringMessage;

import static org.junit.Assert.assertEquals;

public class MonitoringMetricsTest {

  private final Long samplePeriod = 60000L;  // 1 minute

  private static void verifyValues(
      MonitoringMetrics metrics,
      long count,
      long bytes,
      long aggrCrc,
      long totalLatency,
      long minLatency,
      long maxLatency
  ) {
    MonitoringMessage.Builder monitoringMessage = MonitoringMessage.newBuilder();
    metrics.fillMonitoringMessage(monitoringMessage);

    // verify all fields are initial value
    assertEquals(count, monitoringMessage.getCount());
    assertEquals(bytes, monitoringMessage.getAggregateBytes());
    assertEquals(aggrCrc, monitoringMessage.getAggregateCrc());
    assertEquals(totalLatency, monitoringMessage.getTotalLatency());
    assertEquals(minLatency, monitoringMessage.getMinLatency());
    assertEquals(maxLatency, monitoringMessage.getMaxLatency());
  }

  private static void verifyInitialValues(MonitoringMetrics metrics) {
    verifyValues(metrics, 0, 0, 0, 0, 0, 0);
  }

  @Test
  public void testInitializedMetrics() {
    MonitoringMetrics metrics = new MonitoringMetrics();
    verifyInitialValues(metrics);
  }

  @Test
  public void testRecordInvalidSizeUpdatesMetrics() {
    MonitoringMetrics metrics = new MonitoringMetrics();
    long msgChecksum = 8371;
    long msgLatency = 20;
    metrics.recordMessageEvent(-1, msgChecksum, msgLatency);
    verifyValues(metrics, 1, 0, msgChecksum, msgLatency, msgLatency, msgLatency);
  }

  @Test
  public void testRecordInvalidLatencyUpdatesMetrics() {
    MonitoringMetrics metrics = new MonitoringMetrics();
    long msgChecksum = 8371;
    int msgBytes = 8900;
    metrics.recordMessageEvent(msgBytes, msgChecksum, -2);
    verifyValues(metrics, 1, msgBytes, msgChecksum, 0, 0, 0);
  }

  @Test
  public void testRecordedMetrics() {
    MonitoringMetrics metrics = new MonitoringMetrics();
    long checksum = 8371;
    int msgBytes = 8900;
    long count = 1;
    long latency = 20l;

    metrics.recordMessageEvent(msgBytes, checksum, latency);
    verifyValues(metrics, count, msgBytes, checksum, latency, latency, latency);
  }

  @Test
  public void testAggregatedMetrics() {
    MonitoringMetrics metrics = new MonitoringMetrics();
    long checksum = 8371;
    int msgBytes = 8900;
    long msgLatency = 30l;

    metrics.recordMessageEvent(msgBytes, checksum, msgLatency);
    metrics.recordMessageEvent(msgBytes, checksum * 10, msgLatency + 10);
    metrics.recordMessageEvent(msgBytes, checksum * 100, msgLatency + 20);

    // verify that values are as expected
    long expectedCount = 3l;
    long expectedChecksum = checksum ^ (checksum * 10) ^ (checksum * 100);
    long expectedBytes = 3 * msgBytes;
    long expectedTotalLatency = 3 * msgLatency + 30;
    long expectedMinLatency = msgLatency;
    long expectedMaxLatency = msgLatency + 20;
    verifyValues(metrics, expectedCount, expectedBytes, expectedChecksum, expectedTotalLatency,
                 expectedMinLatency, expectedMaxLatency);
  }

  /**
   * If latency is not recorded, totalLatency, minLatency, and maxLatency fields are expected to be zero.
   */
  @Test
  public void testAggregatedMetricsWithoutLatency() {
    MonitoringMetrics metrics = new MonitoringMetrics();
    long checksum = 8371;
    int msgBytes = 8900;

    // one message event
    metrics.recordMessageEvent(msgBytes, checksum);
    verifyValues(metrics, 1, msgBytes, checksum, 0, 0, 0);

    // one more
    metrics.recordMessageEvent(msgBytes + 2, checksum * 10);
    verifyValues(metrics, 2, 2 * msgBytes + 2, checksum ^ (checksum * 10), 0, 0, 0);
  }
}
