// (Copyright) [2016 - 2016] Confluent, Inc.
package io.confluent.monitoring.clients.interceptor;

import io.confluent.monitoring.common.MonitoringMessageUtil;
import org.junit.Test;

import java.util.Random;

import io.confluent.monitoring.record.Monitoring;
import io.confluent.monitoring.record.Monitoring.MonitoringMessage;
import io.confluent.monitoring.record.Monitoring.MonitoringMessageOrBuilder;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertNull;
import static org.junit.Assert.assertTrue;

public class MonitoringTimeBucketsTest {

  private final long samplePeriod = 60000L;  // 1 minute
  private final int windowHistorySize = 4;
  private final long fakeCurrentTime = samplePeriod * 10 + 123;

  // this is background thread for testing "publishing' metrics
  private class Publisher implements Runnable {

    private volatile boolean shutdown;
    private MonitoringTimeBuckets monitoringTimeBuckets;
    private long shortSamplePeriod;
    private long publishCount;
    private long totalMessageCount;
    private long totalBytesPublished;

    public Publisher(MonitoringTimeBuckets monitoringTimeBuckets, long period) {
      this.shutdown = false;
      this.monitoringTimeBuckets = monitoringTimeBuckets;
      this.shortSamplePeriod = period;
      this.publishCount = 0;
      this.totalMessageCount = 0;
      this.totalBytesPublished = 0;
    }

    private void pretendPublishMetrics() {
      long now = System.currentTimeMillis();
      MonitoringMessage.Builder msg =
          monitoringTimeBuckets.getNextMetrics(MonitoringMessage.getDefaultInstance(), now);
      while (msg != null) {
        this.publishCount++;
        this.totalMessageCount += msg.getCount();
        this.totalBytesPublished += msg.getAggregateBytes();
        msg = monitoringTimeBuckets.getNextMetrics(
            MonitoringMessage.getDefaultInstance(),
            now
        );
      }
    }

    public void run() {
      // run until close is called
      while (!shutdown) {
        try {
          Thread.sleep(this.shortSamplePeriod);
          pretendPublishMetrics();
        } catch (Exception e) {
          System.out.println("Error while testing publishing monitoring metrics" + e);
          // we will fail the test when verifying recorded vs. published
          shutdown = true;
        }
      }
      // get all remaining metrics
      pretendPublishMetrics();

      // the last bucket may not be published, try one more time
      long now = System.currentTimeMillis();
      MonitoringMessageOrBuilder msg =
          monitoringTimeBuckets.getNextMetrics(MonitoringMessage.getDefaultInstance(), now);
      if (msg != null) {
        this.publishCount++;
        this.totalMessageCount += msg.getCount();
        this.totalBytesPublished += msg.getAggregateBytes();
      }
    }

    public void close() {
      this.shutdown = true;
    }

    public long getPublishCount() {
      return this.publishCount;
    }

    public long getTotalMessageCountPublished() {
      return this.totalMessageCount;
    }

    public long getTotalBytesPublished() {
      return this.totalBytesPublished;
    }
  }

  private void retrieveAndVerifyMonitoringMessage(
      MonitoringTimeBuckets monitoringTimeBuckets,
      long expectedWindow,
      long expectedMinWindow,
      long expectedMaxWindow,
      long expectedSequenceNumber,
      long expectedCount,
      long expectedBytes,
      long expectedCrc,
      long expectedTotalLatency,
      long expectedMinLatency,
      long expectedMaxLatency
  ) {
    MonitoringMessageOrBuilder publishMsg =
        monitoringTimeBuckets.getNextMetrics(
            MonitoringMessageUtil.baseMonitoringMessage(),
            fakeCurrentTime
        );
    assertNotNull(publishMsg);
    assertEquals(fakeCurrentTime, publishMsg.getTimestamp());
    assertEquals(samplePeriod, publishMsg.getSamplePeriod());
    assertEquals("Unexpected message type", Monitoring.MessageType.NORMAL, publishMsg.getType());
    assertEquals("Unexpected window", expectedWindow, publishMsg.getWindow());
    assertEquals("Unexpected minWindow", expectedMinWindow, publishMsg.getMinWindow());
    assertEquals("Unexpected maxWindow", expectedMaxWindow, publishMsg.getMaxWindow());
    assertEquals("Unexpected sequence number",
                 expectedSequenceNumber, publishMsg.getSequence());
    assertEquals("Unexpected count", expectedCount, publishMsg.getCount());
    assertEquals("Unexpected bytes", expectedBytes, publishMsg.getAggregateBytes());
    assertEquals("Unexpected aggr crc", expectedCrc, publishMsg.getAggregateCrc());
    assertEquals("Unexpected total latency",
                 expectedTotalLatency, publishMsg.getTotalLatency());
    assertEquals("Unexpected min latency",
                 expectedMinLatency, publishMsg.getMinLatency());
    assertEquals("Unexpected max latency",
                 expectedMaxLatency, publishMsg.getMaxLatency());
  }

  private void retrieveAndVerifyNullMessage(MonitoringTimeBuckets monitoringTimeBuckets) {
    MonitoringMessageOrBuilder publishMsg =
        monitoringTimeBuckets.getNextMetrics(
            MonitoringMessage.getDefaultInstance(),
            fakeCurrentTime
        );
    assertNull(publishMsg);
  }

  private void retrieveAndVerifyHeartbeatMessage(
      MonitoringTimeBuckets monitoringTimeBuckets,
      long expectedSequenceNumber,
      long expectedMinWindow,
      long expectedMaxWindow
  ) {
    MonitoringMessageOrBuilder publishMsg =
        monitoringTimeBuckets.getEmptyMetrics(
            MonitoringMessageUtil.baseMonitoringMessage(),
            fakeCurrentTime
        );
    assertNotNull(publishMsg);
    assertEquals(fakeCurrentTime, publishMsg.getTimestamp());
    assertEquals(expectedSequenceNumber, publishMsg.getSequence());
    assertEquals(0, publishMsg.getWindow());
    assertEquals("Sample period must be set", samplePeriod, publishMsg.getSamplePeriod());
    assertEquals("Unexpected message type",
                 Monitoring.MessageType.HEARTBEAT, publishMsg.getType());
    assertEquals(expectedMinWindow, publishMsg.getMinWindow());
    assertEquals(expectedMaxWindow, publishMsg.getMaxWindow());
    assertEquals(0, publishMsg.getCount());
  }

  @Test
  public void testRecordMetrics() {
    MonitoringTimeBuckets monitoringTimeBuckets = new MonitoringTimeBuckets(samplePeriod,
                                                                            windowHistorySize);
    int numberOfBuckets = 10;
    int msgBytes = 300;
    long msgLatency = 1000l;
    long msgChecksum = 123;

    // record one message metric per bucket
    long msgTimestamp = 2 * samplePeriod;
    for (int i = 0; i < numberOfBuckets; ++i) {
      monitoringTimeBuckets.record(msgTimestamp, msgBytes, msgChecksum, msgLatency);
      msgTimestamp += samplePeriod;
    }
    assertEquals(numberOfBuckets, monitoringTimeBuckets.getNumberOfBuckets());

    // record few more metrics to the next bucket (all metrics fall into the same bucket)
    for (int i = 0; i < numberOfBuckets; ++i) {
      monitoringTimeBuckets.record(msgTimestamp, msgBytes, msgChecksum, msgLatency);
    }
    assertEquals(numberOfBuckets + 1, monitoringTimeBuckets.getNumberOfBuckets());

    // record one more metrics in the past
    monitoringTimeBuckets.record(0, msgBytes, msgChecksum, msgLatency);
    assertEquals(numberOfBuckets + 2, monitoringTimeBuckets.getNumberOfBuckets());
  }

  @Test
  public void testRecordInvalidBytes() {
    MonitoringTimeBuckets monitoringTimeBuckets = new MonitoringTimeBuckets(samplePeriod,
                                                                            windowHistorySize);
    monitoringTimeBuckets.record(0, -200, 8493, 100);
    retrieveAndVerifyMonitoringMessage(
        monitoringTimeBuckets, 0, -1, -1, 0, 1, 0, 8493, 100, 100, 100
    );
  }

  @Test
  public void testRecordInvalidLatency() {
    MonitoringTimeBuckets monitoringTimeBuckets = new MonitoringTimeBuckets(samplePeriod,
                                                                            windowHistorySize);
    monitoringTimeBuckets.record(0, 200, 3847, -120);
    retrieveAndVerifyMonitoringMessage(
        monitoringTimeBuckets, 0, -1, -1, 0, 1, 200, 3847, 0, 0, 0
    );
  }

  @Test(expected = IllegalArgumentException.class)
  public void testRecordToInvalidTimestamp() {
    MonitoringTimeBuckets monitoringTimeBuckets = new MonitoringTimeBuckets(samplePeriod,
                                                                            windowHistorySize);
    monitoringTimeBuckets.record(-20, 200, 2134, 2003);
  }

  @Test
  public void testHeartbeatMonitoringMessageSequenceNumbers() {
    MonitoringTimeBuckets monitoringTimeBuckets = new MonitoringTimeBuckets(samplePeriod,
                                                                            windowHistorySize);
    retrieveAndVerifyHeartbeatMessage(monitoringTimeBuckets, 0, -1, -1);
    retrieveAndVerifyHeartbeatMessage(monitoringTimeBuckets, 1, -1, -1);
    retrieveAndVerifyHeartbeatMessage(monitoringTimeBuckets, 2, -1, -1);
  }

  @Test
  public void testHeartbeatMonitoringMessageWithRecordedMetrics() {
    MonitoringTimeBuckets monitoringTimeBuckets = new MonitoringTimeBuckets(samplePeriod,
                                                                            windowHistorySize);
    // even if we record a metrics, we can still retrieve heartbeat message
    // it is up to the caller to use that method correctly
    monitoringTimeBuckets.record(0, 200, 8493, 100);
    // note that since we got heartbeat messages only, and they always have null window, expected
    // min and max windows are also null
    retrieveAndVerifyHeartbeatMessage(monitoringTimeBuckets, 0, -1, -1);
    retrieveAndVerifyHeartbeatMessage(monitoringTimeBuckets, 1, -1, -1);
  }

  @Test
  public void testSequenceNumbersMonotonicallyIncrease() {
    MonitoringTimeBuckets monitoringTimeBuckets = new MonitoringTimeBuckets(samplePeriod, 5);
    // test a mix of filled in and heartbeat messages get assigned correct sequence numbers
    long msgTimestamp = 2 * samplePeriod;
    int msgBytes = 350;
    long msgLatency = 1000l;
    long msgChecksum = 12364;
    retrieveAndVerifyHeartbeatMessage(monitoringTimeBuckets, 0, -1, -1);
    monitoringTimeBuckets.record(msgTimestamp, msgBytes, msgChecksum, msgLatency);
    monitoringTimeBuckets.record(msgTimestamp + samplePeriod, msgBytes, msgChecksum, msgLatency);
    monitoringTimeBuckets
        .record(msgTimestamp + 2 * samplePeriod, msgBytes, msgChecksum, msgLatency);
    retrieveAndVerifyMonitoringMessage(monitoringTimeBuckets,
                                       msgTimestamp, -1, -1, 1, 1, msgBytes, msgChecksum,
                                       msgLatency, msgLatency, msgLatency);
    retrieveAndVerifyMonitoringMessage(monitoringTimeBuckets,
                                       msgTimestamp + samplePeriod, msgTimestamp, msgTimestamp,
                                       2, 1, msgBytes, msgChecksum,
                                       msgLatency, msgLatency, msgLatency);
    retrieveAndVerifyMonitoringMessage(monitoringTimeBuckets,
                                       msgTimestamp + 2 * samplePeriod,
                                       msgTimestamp, msgTimestamp + samplePeriod, 3, 1, msgBytes,
                                       msgChecksum, msgLatency, msgLatency, msgLatency);
    retrieveAndVerifyHeartbeatMessage(monitoringTimeBuckets, 4,
                                      msgTimestamp, msgTimestamp + 2 * samplePeriod);
  }

  @Test
  public void testBasicMetricsRetrieval() {
    MonitoringTimeBuckets monitoringTimeBuckets = new MonitoringTimeBuckets(samplePeriod,
                                                                            windowHistorySize);

    // verify we get null when no metrics are recorded
    retrieveAndVerifyNullMessage(monitoringTimeBuckets);

    // record one metrics and make sure we can retrieve it
    final int msgBytes = 200;
    final long msgLatency = 220;
    final long msgChecksum = 90;
    final long msgTimestamp = samplePeriod + 3;
    monitoringTimeBuckets.record(msgTimestamp, msgBytes, msgChecksum, msgLatency);
    retrieveAndVerifyMonitoringMessage(monitoringTimeBuckets,
                                       samplePeriod, -1, -1, 0, 1, msgBytes, msgChecksum,
                                       msgLatency, msgLatency, msgLatency);

    // verify that once we retrieve time bucket, it's not there anymore
    assertEquals(0, monitoringTimeBuckets.getNumberOfBuckets());
    retrieveAndVerifyNullMessage(monitoringTimeBuckets);
  }

  @Test
  public void testPublishingMetricsFromOneBucket() {
    MonitoringTimeBuckets monitoringTimeBuckets = new MonitoringTimeBuckets(samplePeriod,
                                                                            windowHistorySize);
    int msgBytes = 300;
    long msgLatency = 1000l;
    long msgChecksum = 123l;

    // record metrics that fall into the same time bucket
    final long firstBucketTimestamp = 2 * samplePeriod;
    long msgTimestamp = firstBucketTimestamp;
    int count = 82;
    long expectedChecksum = 0;
    for (int i = 0; i < count; i++) {
      monitoringTimeBuckets.record(msgTimestamp, msgBytes, msgChecksum, msgLatency);
      msgTimestamp += 1l;
      expectedChecksum = expectedChecksum ^ msgChecksum;
    }
    assertEquals(1, monitoringTimeBuckets.getNumberOfBuckets());

    // we should only get one bucket to publish
    retrieveAndVerifyMonitoringMessage(monitoringTimeBuckets,
                                       firstBucketTimestamp, -1, -1, 0, count,
                                       count * msgBytes,
                                       expectedChecksum, msgLatency * count, msgLatency,
                                       msgLatency);
    assertNull(monitoringTimeBuckets.getNextMetrics(MonitoringMessage.getDefaultInstance(),
                                                    fakeCurrentTime));
  }

  @Test
  public void testPublishMetricsFromMultipleBuckets() {
    MonitoringTimeBuckets monitoringTimeBuckets = new MonitoringTimeBuckets(samplePeriod,
                                                                            windowHistorySize);
    long baseTimestamp = 5 * samplePeriod;
    int msgBytes = 300;
    long msgLatency = 1000l;
    long msgChecksum = 123l;

    // each recording should fall into a separate bucket
    monitoringTimeBuckets.record(baseTimestamp + 1, msgBytes, msgChecksum, msgLatency);
    monitoringTimeBuckets.record(baseTimestamp - 1, msgBytes, msgChecksum, msgLatency);
    monitoringTimeBuckets.record(baseTimestamp + samplePeriod, msgBytes, msgChecksum, msgLatency);

    // verify that time buckets get published as expected
    retrieveAndVerifyMonitoringMessage(monitoringTimeBuckets,
                                       baseTimestamp - samplePeriod, -1, -1, 0, 1, msgBytes,
                                       msgChecksum, msgLatency, msgLatency, msgLatency);
    retrieveAndVerifyMonitoringMessage(monitoringTimeBuckets,
                                       baseTimestamp,
                                       baseTimestamp - samplePeriod, baseTimestamp - samplePeriod,
                                       1, 1, msgBytes, msgChecksum,
                                       msgLatency, msgLatency, msgLatency);
    retrieveAndVerifyMonitoringMessage(monitoringTimeBuckets,
                                       baseTimestamp + samplePeriod,
                                       baseTimestamp - samplePeriod, baseTimestamp,
                                       2, 1, msgBytes,
                                       msgChecksum, msgLatency, msgLatency, msgLatency);
  }

  @Test
  public void testMinMaxWindowNoEmptyBuckets() {
    final int historySize = 3;
    final long msgTimestamp = 2 * samplePeriod;
    final int msgBytes = 350;
    final long msgLatency = 1000l;
    final long msgChecksum = 12364;

    MonitoringTimeBuckets timeBuckets = new MonitoringTimeBuckets(samplePeriod, historySize);

    // one stat per bucket
    int count = 2 * historySize;
    for (int i = 0; i < count; i++) {
      timeBuckets.record(msgTimestamp + i * samplePeriod, msgBytes, msgChecksum, msgLatency);
    }

    // first min and max window is always null
    retrieveAndVerifyMonitoringMessage(timeBuckets,
                                       msgTimestamp, -1, -1, 0, 1, msgBytes, msgChecksum,
                                       msgLatency, msgLatency, msgLatency);

    for (int i = 1; i < count; i++) {
      long expectedMinWindow = msgTimestamp;
      if (i >= historySize) {
        expectedMinWindow = msgTimestamp + (i - historySize) * samplePeriod;
      }
      retrieveAndVerifyMonitoringMessage(timeBuckets,
                                         msgTimestamp + i * samplePeriod,
                                         expectedMinWindow, msgTimestamp + (i - 1) * samplePeriod,
                                         i, 1, msgBytes, msgChecksum,
                                         msgLatency, msgLatency, msgLatency);
    }
  }

  @Test
  public void testMinMaxWindowSomeEmptyBuckets() {
    // this test assumes that history size is set to 3
    final int historySize = 3;
    final long msgTimestamp = 2 * samplePeriod;
    final int msgBytes = 350;
    final long msgLatency = 1000l;
    final long msgChecksum = 12364;

    MonitoringTimeBuckets timeBuckets = new MonitoringTimeBuckets(samplePeriod, historySize);

    // one stat per bucket
    int count = 2 * historySize;
    for (int i = 0; i < count; i++) {
      timeBuckets.record(msgTimestamp + 2 * i * samplePeriod, msgBytes, msgChecksum, msgLatency);
    }

    // first min and max window is always null
    retrieveAndVerifyMonitoringMessage(timeBuckets,
                                       msgTimestamp, -1, -1, 0, 1, msgBytes, msgChecksum,
                                       msgLatency, msgLatency, msgLatency);
    retrieveAndVerifyHeartbeatMessage(timeBuckets, 1, msgTimestamp, msgTimestamp);
    retrieveAndVerifyMonitoringMessage(timeBuckets,
                                       msgTimestamp + 2 * samplePeriod, msgTimestamp, msgTimestamp,
                                       2, 1, msgBytes, msgChecksum,
                                       msgLatency, msgLatency, msgLatency);
    retrieveAndVerifyHeartbeatMessage(timeBuckets,
                                      3, msgTimestamp, msgTimestamp + 2 * samplePeriod);
    retrieveAndVerifyMonitoringMessage(timeBuckets, msgTimestamp + 4 * samplePeriod,
                                       msgTimestamp + 2 * samplePeriod,
                                       msgTimestamp + 2 * samplePeriod,
                                       4, 1, msgBytes, msgChecksum,
                                       msgLatency, msgLatency, msgLatency);
    retrieveAndVerifyHeartbeatMessage(timeBuckets,
                                      5, msgTimestamp + 2 * samplePeriod,
                                      msgTimestamp + 4 * samplePeriod);
    retrieveAndVerifyHeartbeatMessage(timeBuckets,
                                      6, msgTimestamp + 4 * samplePeriod,
                                      msgTimestamp + 4 * samplePeriod);
    retrieveAndVerifyHeartbeatMessage(timeBuckets,
                                      7, msgTimestamp + 4 * samplePeriod,
                                      msgTimestamp + 4 * samplePeriod);
    retrieveAndVerifyHeartbeatMessage(timeBuckets, 8, -1, -1);
  }

  private void runMultithreadRecordPublish(long samplePeriodMs, long publishPeriodMs,
                                           int maxJitterMs) throws Exception {
    MonitoringTimeBuckets monitoringTimeBuckets = new MonitoringTimeBuckets(samplePeriodMs,
                                                                            windowHistorySize);
    Random rand = new Random();
    int msgBytes = 100;
    long msgLatency = 1000l;
    long msgChecksum = 123l;

    // start publish thread
    final Publisher publisher = new Publisher(monitoringTimeBuckets, publishPeriodMs);
    final Thread publishThread = new Thread(publisher);
    publishThread.start();

    // record
    long startTimestamp = -1l;
    long endTimestamp = -1l;
    long totalMessages = 2000000;
    for (long i = 0; i < totalMessages; i++) {
      long now = System.currentTimeMillis();
      int jitterMs = rand.nextInt(2 * maxJitterMs + 1);
      now = now + maxJitterMs - jitterMs;
      monitoringTimeBuckets.record(now, msgBytes, msgChecksum, msgLatency);
      if (startTimestamp < 0) {
        startTimestamp = now;
      }
      endTimestamp = now;
    }

    // finish
    publisher.close();
    publishThread.join();

    // verify we 'published' everything we recorded
    assertEquals(totalMessages, publisher.getTotalMessageCountPublished());
    assertEquals(totalMessages * msgBytes, publisher.getTotalBytesPublished());

    long timediff = endTimestamp - startTimestamp;
    long totalTimeBuckets = timediff / samplePeriodMs;
    long totalPublishCount = publisher.getPublishCount();

    // we can potentially use this test as micro perf benchmark to compare different policies for
    // evicting buckets
    System.out.println("Total messages recorded: " + totalMessages +
                       ", runtime=" + timediff + "ms" +
//                ", msg recorded/sec (including get system time)=" + (totalMessages/timediff)*1000 +
                       ", sample period=" + samplePeriodMs +
                       ", publish period=" + publishPeriodMs +
                       ", jitter=" + maxJitterMs + "ms" +
                       ", time buckets = " + totalTimeBuckets +
                       ", total publish count=" + totalPublishCount);
  }

  /**
   * The following four tests test recording of message metrics with monotonically increasing
   * timestamps, and verify that all metrics get publish independent of sample period vs. publish
   * period settings.
   */
  @Test
  public void testMultithreadRecordPublishSamePeriod() throws Exception {
    runMultithreadRecordPublish(10, 10, 0);
  }

  @Test
  public void testMultithreadRecordPublishWithShorterPublishPeriod() throws Exception {
    runMultithreadRecordPublish(10, 5, 0);
  }

  @Test
  public void testMultithreadRecordPublishWithLongerPublishPeriod() throws Exception {
    runMultithreadRecordPublish(10, 15, 0);
  }

  @Test
  public void testMultithreadRecordPublishWithLongPublishPeriod() throws Exception {
    runMultithreadRecordPublish(10, 50, 0);
  }

  /**
   * The following tests test recording of message metrics with out-of-order timestamps
   */
  @Test
  public void testMultithreadRecordPublishWithTinyJitter() throws Exception {
    runMultithreadRecordPublish(10, 10, 5);
  }

  @Test
  public void testMultithreadRecordPublishWithSmallJitter() throws Exception {
    runMultithreadRecordPublish(10, 10, 15);
  }

  @Test
  public void testMultithreadRecordPublishWithMediumJitter() throws Exception {
    runMultithreadRecordPublish(10, 10, 30);
  }

  @Test
  public void testMultithreadRecordPublishWithLargeJitter() throws Exception {
    runMultithreadRecordPublish(10, 10, 50);
  }
}
