// (Copyright) [2016 - 2016] Confluent, Inc.
package io.confluent.monitoring.clients.interceptor;

import com.google.common.collect.ImmutableList;

import org.junit.Test;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertNull;
import static org.junit.Assert.assertTrue;

public class TimeBucketsTest {

  public final long BUCKET_SIZE_MS = 1000L;   // 1 second

  // bucket must already exist
  public static void recordToExistingBucket(long messageTimestamp, int value,
                                            int expectedBucketValue, TimeBuckets<Integer> buckets) {
    long timestamp = buckets.getTimestamp(messageTimestamp);
    TimeBuckets<Integer>.Bucket<Integer> bucket = buckets.get(messageTimestamp);
    assertNotNull(bucket.data);
    assertTrue(bucket.corresponds(timestamp));
    assertEquals(expectedBucketValue, bucket.data.intValue());
    bucket.data = value;
  }

  // bucket must be available
  public static void recordToNonExistingBucket(long messageTimestamp, int value,
                                               TimeBuckets<Integer> buckets) {
    long timestamp = buckets.getTimestamp(messageTimestamp);
    TimeBuckets<Integer>.Bucket<Integer> bucket = buckets.get(messageTimestamp);
    assertNull(bucket.data);
    bucket.data = new Integer(value);
    assertTrue(bucket.corresponds(timestamp));
  }

  public static void verifyRemoveOldestBucket(boolean expectExists, int expectedValue,
                                              TimeBuckets<Integer> buckets) {
    TimeBuckets<Integer>.Bucket<Integer> bucket = buckets.removeOldestBucket();
    if (expectExists) {
      assertNotNull(bucket);
      assertEquals(expectedValue, bucket.data.intValue());
    } else {
      assertNull(bucket);
    }
  }

  @Test
  public void testEmptyBuckets() {
    TimeBuckets<Integer> buckets = new TimeBuckets(BUCKET_SIZE_MS);
    assertEquals(0, buckets.getNumberOfBuckets());
  }

  @Test
  public void testBucketSize() {
    TimeBuckets<Integer> buckets = new TimeBuckets<>(BUCKET_SIZE_MS);
    assertEquals(BUCKET_SIZE_MS, buckets.getBucketSizeMs());
  }

  /**
   * Test record values using given timestamps. Make sure that timestamps passed to this test map to
   * different buckets
   */
  private void doTestBasicRecordValues(ImmutableList<Long> timestamps) {
    // we should have the same number of buckets as timestamps, so we always map one bucket to one timestamp
    TimeBuckets<Integer> buckets = new TimeBuckets(BUCKET_SIZE_MS);

    // first record value to every bucket, which we except to be non-existing with unallocated
    // metric data
    int i = 1;
    for (Long msgTimestamp : timestamps) {
      recordToNonExistingBucket(msgTimestamp.longValue(), i, buckets);
      assertEquals(i, buckets.getNumberOfBuckets());
      i++;
    }

    // record new value to each bucket again, now buckets must exist
    i = 1;
    for (Long msgTimestamp : timestamps) {
      recordToExistingBucket(msgTimestamp.longValue(), i + timestamps.size(), i, buckets);
      i++;
    }
    assertEquals(timestamps.size(), buckets.getNumberOfBuckets());
  }

  /**
   * This test is independent of how we implement selecting bucket for eviction
   */
  @Test
  public void testRecordValuesOrderedTimestamps() {
    long msgTimestamp = 0;
    ImmutableList<Long> timestamps = ImmutableList.of(msgTimestamp,
                                                      msgTimestamp + BUCKET_SIZE_MS,
                                                      msgTimestamp + 2 * BUCKET_SIZE_MS,
                                                      msgTimestamp + 3 * BUCKET_SIZE_MS,
                                                      msgTimestamp + 4 * BUCKET_SIZE_MS);

    doTestBasicRecordValues(timestamps);
  }

  /**
   * This test is independent of how we implement selecting bucket for eviction
   */
  @Test
  public void testRecordValuesNotOrderedTimestamps() {
    long msgTimestamp = 0;
    ImmutableList<Long> timestamps = ImmutableList.of(msgTimestamp + 4 * BUCKET_SIZE_MS,
                                                      msgTimestamp + 2 * BUCKET_SIZE_MS,
                                                      msgTimestamp + BUCKET_SIZE_MS,
                                                      msgTimestamp + 3 * BUCKET_SIZE_MS,
                                                      msgTimestamp);

    doTestBasicRecordValues(timestamps);
  }

  @Test
  public void testNonAlignedTimestampsInOneBucket() {
    TimeBuckets<Integer> buckets = new TimeBuckets(BUCKET_SIZE_MS);

    long msgTimestamp = 10;
    recordToNonExistingBucket(msgTimestamp, 7, buckets);

    // increase timestamp, but stay in the same bucket
    msgTimestamp += 10;
    recordToExistingBucket(msgTimestamp, 8, 7, buckets);

    // decrease timestamp, but stay in the same bucket
    msgTimestamp -= 5;
    recordToExistingBucket(msgTimestamp, 9, 8, buckets);

    // same timestamp
    recordToExistingBucket(msgTimestamp, 10, 9, buckets);

    // verify that we have only one bucket
    assertEquals(1, buckets.getNumberOfBuckets());
  }

  public void testBucketBoundaries() {
    TimeBuckets<Integer> buckets = new TimeBuckets(BUCKET_SIZE_MS);

    long msgTimestamp = 0;
    recordToNonExistingBucket(msgTimestamp, 7, buckets);
    assertEquals(1, buckets.getNumberOfBuckets());

    // write the last timestamp that could fit into this bucket
    recordToExistingBucket(msgTimestamp + BUCKET_SIZE_MS - 1, 13, 7, buckets);
    assertEquals(1, buckets.getNumberOfBuckets());

    // the next timestamp should start a new bucket
    recordToNonExistingBucket(msgTimestamp + BUCKET_SIZE_MS, 14, buckets);
    assertEquals(2, buckets.getNumberOfBuckets());
  }

  public void testBucketRetrievalZeroOrOneBucket() {
    TimeBuckets<Integer> buckets = new TimeBuckets(BUCKET_SIZE_MS);
    long msgTimestamp = BUCKET_SIZE_MS + 12;

    // there should be no buckets to retrieve
    verifyRemoveOldestBucket(false, 0, buckets);

    // verify with one valid bucket
    int value = 7;
    recordToNonExistingBucket(msgTimestamp, value, buckets);
    assertEquals(1, buckets.getNumberOfBuckets());

    verifyRemoveOldestBucket(true, value, buckets);
    assertEquals(0, buckets.getNumberOfBuckets());
  }

  @Test
  public void testBucketRetrievalIncreasingTimestamps() {
    TimeBuckets<Integer> buckets = new TimeBuckets(BUCKET_SIZE_MS);
    long msgTimestamp = 0;

    // record values to allocate some buckets
    final int count = 9;
    for (int i = 0; i < count; ++i) {
      recordToNonExistingBucket(msgTimestamp, i, buckets);
      msgTimestamp += BUCKET_SIZE_MS;
    }
    assertEquals(count, buckets.getNumberOfBuckets());

    // get all the buckets, should be in the same order as we created them
    for (int i = 0; i < count; ++i) {
      verifyRemoveOldestBucket(true, i, buckets);
    }
    assertEquals(0, buckets.getNumberOfBuckets());
  }

  @Test
  public void testBucketRetrievalNonOrderedTimestamps() {
    TimeBuckets<Integer> buckets = new TimeBuckets(BUCKET_SIZE_MS);
    long msgTimestamp = BUCKET_SIZE_MS + 12;

    // create some buckets
    int value = 7;
    recordToNonExistingBucket(msgTimestamp, value, buckets);
    recordToNonExistingBucket(msgTimestamp + 10 * BUCKET_SIZE_MS, value + 1, buckets);
    recordToNonExistingBucket(msgTimestamp + BUCKET_SIZE_MS, value + 2, buckets);
    recordToNonExistingBucket(msgTimestamp + 5 * BUCKET_SIZE_MS, value + 3, buckets);

    // verify order of retrieval
    verifyRemoveOldestBucket(true, value, buckets);
    verifyRemoveOldestBucket(true, value + 2, buckets);
    verifyRemoveOldestBucket(true, value + 3, buckets);
    verifyRemoveOldestBucket(true, value + 1, buckets);

    // verify no more buckets
    verifyRemoveOldestBucket(false, 0, buckets);
  }
}
