// (Copyright) [2016 - 2016] Confluent, Inc.
package io.confluent.controlcenter.example;

import com.google.common.base.Joiner;
import com.google.common.base.MoreObjects;
import com.google.common.collect.Lists;
import com.google.inject.Inject;

import org.apache.kafka.clients.consumer.ConsumerConfig;
import org.apache.kafka.clients.consumer.ConsumerRebalanceListener;
import org.apache.kafka.clients.consumer.ConsumerRecord;
import org.apache.kafka.clients.consumer.ConsumerRecords;
import org.apache.kafka.clients.consumer.KafkaConsumer;
import org.apache.kafka.common.TopicPartition;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.Collection;
import java.util.List;
import java.util.Properties;
import java.util.concurrent.TimeUnit;

import io.confluent.controlcenter.ControlCenterConfig;
import io.confluent.monitoring.clients.interceptor.MonitoringInterceptorConfig;
import io.confluent.monitoring.common.TimeBucket;

final class FakeConsumer implements Runnable {

  private static final Logger log = LoggerFactory.getLogger(FakeConsumer.class);

  private final List<String> topics;
  private final ControlCenterConfig config;
  private String group = "fakeConsumerGroupId";
  private String id = "myconsumer";

  private enum State {
    RUNNING, STOPPED;
  }

  @Inject
  public FakeConsumer(List<String> topics, ControlCenterConfig config) {
    this.topics = topics;
    this.config = config;
  }

  public void setGroup(String group) {
    this.group = group;
  }

  public void setId(String id) {
    this.id = id;
  }

  public void run() {
    Properties properties = new Properties();
    properties.put(ConsumerConfig.BOOTSTRAP_SERVERS_CONFIG, config.getList("bootstrap.servers"));
    properties.put(ConsumerConfig.AUTO_OFFSET_RESET_CONFIG, "earliest");
    properties.put(ConsumerConfig.GROUP_ID_CONFIG, group);
    properties.put(ConsumerConfig.CLIENT_ID_CONFIG, id);
    properties.put(ConsumerConfig.SESSION_TIMEOUT_MS_CONFIG, "30000");
    properties.put(ConsumerConfig.ENABLE_AUTO_COMMIT_CONFIG, "true");
    properties.put(ConsumerConfig.AUTO_COMMIT_INTERVAL_MS_CONFIG, "10000");
    properties.put(
        ConsumerConfig.VALUE_DESERIALIZER_CLASS_CONFIG,
        "org.apache.kafka.common.serialization.LongDeserializer"
    );
    properties.put(
        ConsumerConfig.KEY_DESERIALIZER_CLASS_CONFIG,
        "org.apache.kafka.common.serialization.StringDeserializer"
    );
    properties.put(
        ConsumerConfig.INTERCEPTOR_CLASSES_CONFIG,
        "io.confluent.monitoring.clients.interceptor.MonitoringConsumerInterceptor"
    );
    for (String s : Lists.newArrayList("", MonitoringInterceptorConfig.MONITORING_INTERCEPTOR_PREFIX)) {
//      properties.put(s + CommonClientConfigs.SECURITY_PROTOCOL_CONFIG, "SSL");
//      properties.put(s + SslConfigs.SSL_TRUSTSTORE_LOCATION_CONFIG, "/Users/norwood/clients/kafka/server.truststore.jks");
//      properties.put(s + SslConfigs.SSL_TRUSTSTORE_PASSWORD_CONFIG, "password");
//      properties.put(s + SslConfigs.SSL_KEYSTORE_LOCATION_CONFIG, "/Users/norwood/clients/kafka/server.keystore.jks");
//      properties.put(s + SslConfigs.SSL_KEYSTORE_PASSWORD_CONFIG, "password");
    }

    KafkaConsumer<String, Long> consumer = null;
    log.info("starting {}", this);
    long runTime = Math.round(Math.random() * TimeUnit.SECONDS.toMillis(300));
    long sleepTime = 0;
    long startTime = System.currentTimeMillis();
    State currentState = State.STOPPED;
    try {
      while(!Thread.currentThread().isInterrupted()) {
        long now = System.currentTimeMillis();
        switch (currentState) {
          case RUNNING:
            if (startTime + runTime < now) {
              if (consumer != null) {
                log.info("shutting down consumer={} group={}", id, group);
                consumer.close();
              }
              consumer = null;
              startTime = now;
              sleepTime = Math.round(Math.random() * TimeUnit.SECONDS.toMillis(30));
              currentState = State.STOPPED;
            } else {
              try {
                ConsumerRecords<String, Long> records = consumer.poll(500);
                for (ConsumerRecord<String, Long> record : records) {
                  log.trace("read={} consumer={} group={}", record.value().toString(), id, group);
                }
              } catch (Exception e) {
                log.error("consuming", e);
              }
              Thread.sleep(Math.round(Math.random() * TimeUnit.SECONDS.toMillis(3)));
            }
            break;
          case STOPPED:
            if (startTime + sleepTime < now) {
              consumer = new KafkaConsumer<String, Long>(properties);
              final FakeConsumer owner = this;
              consumer.subscribe(topics, new ConsumerRebalanceListener() {
                @Override
                public void onPartitionsRevoked(Collection<TopicPartition> partitions) {
                  log.warn("partitions revoked {} {}", owner, partitions);
                }

                @Override
                public void onPartitionsAssigned(Collection<TopicPartition> partitions) {
                  log.info("partitions assigned {} {}", owner, partitions);
                }
              });
              startTime = now;
              runTime = Math.round(Math.random() * TimeUnit.SECONDS.toMillis(300));
              currentState = State.RUNNING;
            }
            break;
        }
      }
    } catch (InterruptedException ie) {
      log.info("interrupted {}", this);
    } finally {
      log.info("Will close {}", this);
      if (consumer != null) {
        consumer.close();
      }
    }
  }

  public List<String> getTopics() {
    return topics;
  }

  @Override
  public String toString() {
    return MoreObjects.toStringHelper(this)
        .add("clientId", id)
        .add("group", group)
        .add("topics", "[" + Joiner.on(',').join(topics) + "]")
        .toString();
  }
}
