// (Copyright) [2016 - 2016] Confluent, Inc.
package io.confluent.controlcenter.example;

import com.google.inject.Guice;
import com.google.inject.Injector;

import org.apache.commons.cli.CommandLine;
import org.apache.commons.cli.DefaultParser;
import org.apache.commons.cli.Option;
import org.apache.commons.cli.Options;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.concurrent.Executors;
import java.util.concurrent.ScheduledExecutorService;
import java.util.concurrent.TimeUnit;

import io.confluent.controlcenter.ControlCenterConfig;

public class Rock {

  private static final Logger log = LoggerFactory.getLogger(Rock.class);
  private static final String TEST_TOPIC = "test-topic";

  /**
   * The first prototype is tailored to query number of messages produced vs. number of messages
   * consumed that could be filtered per topic key: first 4 letters of type, first 20 letters of
   * topic, ts, id (consumer or producer id) This is an easy way to get fixed size key prefix.
   * Another option to consider is to encode type, topic with SHA-256. Getting SHA-256 every time we
   * read/write to the DB is CPU consuming, so we need to investigate which one is the best approach
   * In either case, we will store actual type and topic in a value.
   */
  public static void main(String[] args) throws Exception {
    Options options = new Options()
        .addOption(
            Option.builder("config")
                .hasArg()
                .longOpt("config")
                .desc("control center properties file")
                .build()
        );

    CommandLine cmd = new DefaultParser().parse(options, args);

    ControlCenterConfig config = new ControlCenterConfig(cmd.getOptionValue("config"));

    final Injector injector = Guice.createInjector(
        new ExampleModule(TEST_TOPIC, config)
    );

    ScheduledExecutorService ses = Executors.newScheduledThreadPool(6);

    try {
      // create a consumer group=fakeConsumerGroupId id=myconsumer
      FakeConsumer consumerOffTwoTopics = injector.getInstance(FakeConsumer.class);
      consumerOffTwoTopics.getTopics().add("another-topic");
      ses.submit(consumerOffTwoTopics);

      // create a consumer group=other-group id=other-id-1
      FakeConsumer fc = injector.getInstance(FakeConsumer.class);
      fc.setGroup("other-group");
      fc.setId("other-id-1");
      ses.submit(fc);

      // create a consumer group=other-group id=other-id-2
      FakeConsumer fc2 = injector.getInstance(FakeConsumer.class);
      fc2.setGroup("other-group");
      fc2.setId("other-id-2");
      ses.submit(fc2);

      // create a producer id=myproducer
      ses.submit(injector.getInstance(FakeProducer.class));

      // create a producer id=producer-id-2
      FakeProducer fp2 = injector.getInstance(FakeProducer.class);
      fp2.setId("producer-id-2");
      fp2.getTopics().add("another-topic");
      ses.submit(fp2);

      log.warn("await termination");
      ses.awaitTermination(100, TimeUnit.DAYS);
      log.warn("finished");
    } catch (Exception e) {
      log.error("who knows", e);
    } finally {
      ses.shutdownNow();
    }
  }
}
